/* 1strs8.c
 * Calculate the "ones string" length for primes.
 * arguments: [-r<minlen> <maxlen>]: smallest and largest Mersenne exponent to print.
 *            [-f]: find the first factor (of an exponent in the range) that's larger than each input number.
 *            <start>: first prime to find smallest Mersenne number which it factors.
 *            [<stop>]: largest value to calculate length of ones string of.
 * stdout: primes and the smallest Mersenne number each factors.
 */
const char *program_name = "1strs8"; /* for perror() and similar */
#include "version.h"

#define MERS_1STRS

#ifndef FACTOR_PACKAGE
#define FACTOR_PACKAGE FACTOR_LONGLONG
#endif

#include "setup.h"
#include "factor.h"

FACTOR savef, numer; /* globals used by some factor.h routines */

/*!!If these are changed, the second argument to /usr/games/primes should be adjusted to match. */
/*!!  Doing that adjustment in code would be painful; I don't want to take the time presently. */
/*!!  Leaving off the second argument to /usr/games/primes seems to stall the pclose() call rather than killing the child process early on Ubuntu Gutsy Gibbon. */
#define SIEVE_SIZE (65528)
#ifndef NUM_SMALL_PRIMES
#define NUM_SMALL_PRIMES (6540)
#endif /* NUM_SMALL_PRIMES */

unsigned char sieve[SIEVE_SIZE/NBBY];
#define CLEAR_BIT(x) (sieve[x/NBBY] &= (unsigned char)(~(1 << (x & (NBBY - 1)))))
#define TEST_BIT(x) (sieve[x/NBBY] & (1 << (x & (NBBY - 1))))
#define SET_VALUE ((1 << NBBY) - 1)

/* attempt to use smaller types to save memory */
/*  subscript 0 is prime 3 gives subscript 6541 is prime 65537; 3511 is 32771; 53 is 257 */
/* we want to avoid signed types, however */
#if (NUM_SMALL_PRIMES < 53) && (SIEVE_SIZE < 256)
typedef unsigned char INTEGER;
#else
# if (NUM_SMALL_PRIMES < 6541) && (SIEVE_SIZE < 65536)
typedef unsigned short INTEGER;
# else
typedef unsigned long INTEGER;
# endif
#endif

struct small_prime_info
 {
  INTEGER prime;
  INTEGER bit_to_clear;
 } small_primes[NUM_SMALL_PRIMES];

/*!!Next is lifted from testerl3.c v1.209, but note that num2 is always 2 here, */
/*!! so it is _much_ simpler. */
/* Compute x such that 2*x = -1 MOD small_primes[which].prime. */
/* We can then use x to compute the first bit in the sieve array that to be cleared. */
static INTEGER euclid(FACTOR p, UL which)
 {
  INTEGER x = small_primes[which].prime/2;
  UL temp = fac_mod_int(p, (UL)small_primes[which].prime);

  /*!!note the next line's product when choosing the types (& their lengths) involved */
  return((x*temp) % small_primes[which].prime);
 }

int main(int argc, char **argv)
 {
  register UL factor, /* factor of length (of 1s string) */
    minlen = 601, /* smallest exponent we care about; all smaller than M601 are completely factored */
    maxlen = ~(UL)0; /* largest exponent we care about */
  UL minlentmp, maxlentmp; /* temporaries that can be passed, by address, to sscanf() */
  FACTOR p, /* prime for which we are calculating the 1s string length */
    pshift, /* p, shifted; note that first shifted bit must be a 1 */
    acc, /* accumulator: power of two mod p */
    length, length2, /* length of 1s string so far */
    cofactor, /* cofactor of length after dividing out factors as they're found */
    bit, /* power of two just larger than pshift */
    stop; /* largest p to try */
  UL small; /* used to read small primes and then as temporary */
  INTEGER which, next; /* subscripts for small_primes array */
  FILE *primesfp; /* for reading the small primes from /usr/games/primes */

  fac_init();
  fac_init_var(p);
  fac_init_var(pshift);
  fac_init_var(acc);
  fac_init_var(length);
  fac_init_var(length2);
  fac_init_var(cofactor);
  fac_init_var(bit);
  fac_init_var(stop);
  if (argc < 2 || !isascii(argv[1][0]))
   {
    fprintf(stderr, "\tUsage: %s [-r<minlen> <maxlen>] [-f | <start> [<stop>] ]\n", program_name);
    return(1);
   }
  if (argv[1][0] == '-' && argv[1][1] == 'v')
   {
    fprintf(stdout, VERSION);
    return(0);
   }
  if (argc > 3 && argv[1][0] == '-' && argv[1][1] == 'r')
   {
    if (sscanf((++argv)[0] + 2, SCANF_FMT_UL, &minlentmp) != 1)
     {
      fprintf(stderr, "%s: invalid start of range: '%s'\n", program_name, argv[0] + 2);
      fprintf(stderr, "\tUsage: %s [-r<minlen> <maxlen>] [-f | <start> [<stop>] ]\n", program_name);
      return(1);
     }
    if (sscanf((++argv)[0], SCANF_FMT_UL, &maxlentmp) != 1)
     {
      fprintf(stderr, "%s: invalid end of range: '%s'\n", program_name, argv[0] + 2);
      fprintf(stderr, "\tUsage: %s [-r<minlen> <maxlen>] [-f | <start> [<stop>] ]\n", program_name);
      return(1);
     }
    minlen = minlentmp;
    maxlen = maxlentmp;
    argc -= 2;
   }
  if (minlen < 601)
    minlen = 601;
  if (minlen >= maxlen)
    maxlen = ~(UL)0;
  if (maxlen < minlen)
   {
    fprintf(stderr, "%s: start of range cannot be larger than end\n", program_name);
    fprintf(stderr, "\tUsage: %s [-r<minlen> <maxlen>] [-f | <start> [<stop>] ]\n", program_name);
    return(1);
   }
  if (argc < 2 || sscan_fac(*++argv, &p) != 1)
   {
    fprintf(stderr, "%s: first non-flag argument, the number to start at, is required\n", program_name);
    fprintf(stderr, "\tUsage: %s [-r<minlen> <maxlen>] [-f | <start> [<stop>] ]\n", program_name);
    return(1);
   }
  if ((fac_mod8(p) & 1) == 0)
    fac_add_int(p, (UL)1, &p);
  if (--argc < 2 || sscan_fac(*++argv, &stop) != 1)
    fac_set_int((UL)0, &stop);
  if (!fac_less_than_int(stop, (UL)1) && !fac_less_than(p, stop))
   {
    fprintf(stderr, "%s: start must be less than stop\n", program_name);
    fprintf(stderr, "\tUsage: %s [-r<minlen> <maxlen>] [-f | <start> [<stop>] ]\n", program_name);
    return(1);
   }
  /* read small primes rather than duplicating code to find them */
  /*!!the second argument to /usr/games/primes needs to be large enough to generate at least NUM_SMALL_PRIMES prime numbers */
  primesfp = popen("/usr/games/primes 3 400000", "r");
  for (which = 0; which < NUM_SMALL_PRIMES; which++)
    if (fscanf(primesfp, PRINTF_FMT_UL, &small) < 1)
      //|| (small >> NBBY*sizeof(INTEGER)) != 0, but gcc v7.5 claims that is always false
      break;
    else
      small_primes[which].prime = (INTEGER)small;
  (void)pclose(primesfp);
  if (which < NUM_SMALL_PRIMES)
   {
    fprintf(stderr, "%s: problems reading small primes\n", program_name);
    fprintf(stderr, " which = %ld, small = " PRINTF_FMT_UL ", NUM_SMALL_PRIMES = %d\n", which, small, NUM_SMALL_PRIMES);
    return(1);
   }
  if (fac_less_than_int(p, (UL)small_primes[NUM_SMALL_PRIMES - 1].prime))
   {
    fprintf(stderr, "%s: starting number too small; must be at least %lu\n", program_name,
            (unsigned long)small_primes[NUM_SMALL_PRIMES - 1].prime);
    return(1);
   }
  /* set bits of sieve for numbers starting with p */
  for (which = 0; which < NUM_SMALL_PRIMES; which++)
    small_primes[which].bit_to_clear = euclid(p, (UL)which);
  memset(sieve, SET_VALUE, sizeof(sieve));
  next = SIEVE_SIZE;
  for (setup(); !shouldTerminate; next++, fac_add_int(p, (UL)2, &p))
   { /* step to next number not sieved out */
    for ( ; next < SIEVE_SIZE && !TEST_BIT(next); next++)
      fac_add_int(p, (UL)2, &p);
    while (next >= SIEVE_SIZE && !shouldTerminate)
     { /* do next sieve */
      memset(sieve, SET_VALUE, sizeof(sieve));
      for (which = 0; which < NUM_SMALL_PRIMES; which++)
       {
        for (small = small_primes[which].bit_to_clear; small < SIEVE_SIZE;
             small += small_primes[which].prime)
          CLEAR_BIT(small);
        small_primes[which].bit_to_clear = small - SIEVE_SIZE;
       }
      for (next = 0; next < SIEVE_SIZE && !TEST_BIT(next); next++)
        fac_add_int(p, (UL)2, &p);
     }
    if (shouldTerminate)
      break;
    fac_prepare_denominator(p);
    for (fac_div_int(p, (UL)2, &pshift); (fac_mod8(pshift) & 1) == 0; fac_div_int(pshift, (UL)2, &pshift))
      ;
    for (small = 0, fac_set_int((UL)1, &bit); !fac_less_than(pshift, bit); small++)
      fac_times2(bit, &bit);
    small--;
    fac_set_int((UL)1, &length);
    fac_set_int((UL)2, &acc);
    fac_set_int((UL)3, &cofactor); /* acc + 1, for comparing against p */
    while (small > 0 && (fac_less_than_int(acc,(UL)1) || !fac_less_than_int(acc,(UL)2))
           && fac_less_than(cofactor, p) && !shouldTerminate)
     {
      fac_square_mod(acc, &acc);
      fac_times2(length, &length);
      --small;
      if (fac_bit(pshift, small) != 0)
       {
        fac_times2_mod(acc, &acc);
        fac_add_int(length, (UL)1, &length);
       }
      fac_prepare_numerator(acc);
      fac_mod(&acc);
      fac_add_int(acc, (UL)1, &cofactor);
     }
    fac_set(length, &cofactor);
    fac_add_int(length, (UL)2, &bit);
    while (!fac_less_than(p, bit) && (fac_less_than_int(acc,(UL)1) || !fac_less_than_int(acc,(UL)2))
           && !shouldTerminate)
     {
      fac_square_mod(acc, &acc);
      fac_mul_int(length, (UL)2, &length);
      fac_add_int(length, (UL)2, &bit);
     }
    if (fac_less_than_int(acc, (UL)1) || !fac_less_than_int(acc, (UL)2) /* 2^length != 1 (mod p): p !prime */
        || fac_less_than_int(length, minlen) || shouldTerminate) /* too small or need to exit */
      continue;

    factor = (UL)small_primes[which = 0].prime;
    while (!fac_less_than_int(cofactor, factor) && !shouldTerminate)
     {
      if (fac_less_than_int(cofactor, (UL)1) || !fac_less_than_int(cofactor, (UL)2))
       {
        fac_div_int(cofactor, factor, &bit);
        if (fac_less_than_int(bit, factor))
         {
          if (fac_less_than_int(cofactor, ~(UL)0))
            factor = fac_mod_int(cofactor, ~(UL)0);
          else
            factor = 0; /* should become the cofactor, but that is probably too large: fake it */
         }
       }
      while (factor != 0 && fac_mod_int(cofactor, factor) == 0)
        (void)fac_div_int(cofactor, factor, &cofactor);
      fac_set(length, &length2);
      while ((factor == 0 || fac_mod_int(length, factor) == 0) &&
             !fac_less_than(length, length2) && !fac_less_than(length2, length) &&
             !shouldTerminate)
       {
        if (factor == 0)
          fac_div(length, cofactor, &pshift, &bit); /* remainder, in bit, must be zero */
        else
          (void)fac_div_int(length, factor, &pshift); /* remainder, returned, must be zero */
        for (small = 0, fac_set_int((UL)1, &bit); !fac_less_than(pshift, bit); small++)
          fac_times2(bit, &bit);
        small--;
        fac_set_int((UL)1, &length2);
        fac_set_int((UL)2, &acc);
        while (small > 0 && (fac_less_than_int(acc, (UL)1) || !fac_less_than_int(acc, (UL)2)) &&
               !shouldTerminate)
         {
          fac_square_mod(acc, &acc);
          fac_times2(length2, &length2);
          --small;
          if (fac_bit(pshift, small) != 0)
           {
            fac_times2_mod(acc, &acc);
            fac_add_int(length2, (UL)1, &length2);
           }
          fac_prepare_numerator(acc);
          fac_mod(&acc);
         }
        if (!fac_less_than_int(acc, (UL)1) && fac_less_than_int(acc, (UL)2)) /* 2^length2 == 1 (mod p) */
         {
          factor = small_primes[which = 0].prime;
          fac_set(length2, &length);
          fac_set(length2, &cofactor);
          while (fac_mod8(cofactor) % 2 == 0)
            fac_div_int(cofactor, (UL)2, &cofactor);
          while (fac_mod_int(cofactor, (UL)3) == 0)
            fac_div_int(cofactor, (UL)3, &cofactor);
         }
       }
      if (factor == 0)
        fac_set_int((UL)1, &cofactor);
      if (which < NUM_SMALL_PRIMES - 1)
        factor = small_primes[++which].prime;
      else
        factor += 2;
     }
    fac_prepare_numerator(p);
    fac_prepare_denominator(length);
    fac_mod(&acc);
    if (fac_less_than_int(length, maxlen) && !fac_less_than_int(acc, (UL)1) && fac_less_than_int(acc, (UL)2)
        && !fac_less_than_int(length, minlen) && !shouldTerminate)
     {
      printf("M( ");
      fprint_fac(stdout, length);
      printf(" )C: ");
      fprint_fac(stdout, p);
      printf("\n");
      if (!fac_less_than_int(stop, (UL)1) && fac_less_than(stop, p))
        break;
      else
        continue;
     }
    if ((fac_mod8(p) & 1) == 0)
      fac_add_int(p, (UL)1, &p);
    if (!fac_less_than_int(stop, (UL)1) && fac_less_than(stop, p))
      break;
   }
  printf("M( " PRINTF_FMT_UL " )X: ", maxlen);
  fac_sub_int(p, (UL)4, &p);
  fprint_fac(stdout, p);
  printf(" 1strs8 %s end\n", VERSION);
  return(0);
 }
