#!/usr/bin/perl -Tw
# ECMSave.pm: functions to read and compare (P-1 and P+1) save files of the (Linux) ecm program.

package ECMSave;

#!!Now that gmp-ecm checksums can be calculated, add support to convert
#!!  binary Prime95 P+-1 save files to gmp-ecm save files.
#!!    See extract.c and the gmp-ecm source for details.

#!!Implement covers() that returns true (1) only if the argument ECMSave instance is "covered"
#!!  by the caller instance - i.e., same exponentm, method, cofactor (L vs. M), and the argument's
#!!  bounds are the same or smaller.
#!!coversFile() that calls new() based on the file/HASH argument?
#!!  or just implement as part of covers() based on the type of the arguments?

use strict;
use strict 'vars';
use File::stat;
use Math::BigInt try => 'GMP';

use parent 'Exporter';

BEGIN { push(@INC, "$ENV{'PWD'}") if -s('Mersenne.pm'); }
use Mersenne;
use MersInfo;

our ($VERSION) = MersInfo::version();

our @EXPORT_OK = qw();

# Constants related to gmp-ecm save files.
# In gmp-ecm save files, these keys always have numeric values
#   (except N may be a simple "formula" but only numbers are supported here).
#   'X', unlike the others, is in hexadecimal (base 16) with a leading 0x.
my (%ECMnumerics) = ( 'N' => '([0-9]+)', 'B1' => '([0-9]+)', 'CHECKSUM' => '([0-9]+)',
                      'SIGMA' => '([0-9]+)', 'A' => '([0-9]+)', 'X' => '(0[xX][A-Fa-f0-9]+)',
                      'Z' => '([0-9]+)', 'PARAM' => '([0-9]+)', 'ETYPE' => '([0-9]+)', );
my ($ECMvaluePattern) = '([-+.:@ A-Za-z0-9]+)';
# From gmp-ecm/ecm-ecm.h:
# /* The checksum for savefile is the product of all mandatory fields, modulo
#    the greatest prime below 2^32 */
#define CHKSUMMOD 4294967291U
my ($ECMchecksumMod) = Math::BigInt->new("4294967291");

# Could get this from Perl's Socket module, but that is not needed here otherwise.
my ($CRLF) = "\015\012"; # record separator on sockets; strip from input

# All (new) file names should start with this.
#   All file names should be "ecm.$exponent.$method$M" where $method is 'pm1' or 'pp1'
#     and M is '' (for c and L cofactors) or '.M' (for M cofactors).
#   An older file naming convention that had $method and $exponent swapped is still
#     supported for existing files but not for new files.
our ($filePrefix) = 'ecm.';

# These patterns apply to the final component of the pathname in the current naming style.
our (@filePatterns) = ('ecm[.]([1-9][0-9]+)[.]',
                       'ecm[.]([1-9][0-9]+)[.](p[mp]1)',
                       'ecm[.]([1-9][0-9]+)[.](p[mp]1)([.]M|)',
  );

# Directories used by scripts for ecm save files plus '.' (the current directory).
my (%subdirectories) = ( 'doP+-1.pl' => 'p+-1.sav',
                         'mersServer.pl' => 'serv.sav/ecm',
                         'mersServer' => 'serv.sav/ecm',
                         '.' => '.', );
# Same directories in the order to search them for ecm save files.
my (@subdirectories) = ( 'p+-1.sav', 'serv.sav/ecm', '.', );

# There are a lot of ecm save files (9 per 8 Mersenne numbers), so split them into
#   subdirectories based on the exponent % $expMod.  Since this requires the exponent,
#   new() requires it via the 'exp' entry of the parameter hash.
our ($expMod) = 9941;

sub subdirs()
{ # Prevent callers from modifying @subdirectories.
  my (@list);
  push(@list, @subdirectories);
  return(@list)
    if wantarray();
  return($list[0]);
}

sub subdir
{ # Return the directory related to the script with the UUID.
  my ($uuid) = shift();
  my ($exp) = shift(); # Mersenne exponent ...
  return(undef) # ... which must be a positive number if given.
    if defined($exp) && $exp !~ m,^([1-9][0-9]+)$,;
  return("$subdirectories{$uuid}" . '/' . ($exp % $expMod))
    if $uuid && exists $subdirectories{$uuid} && $exp;
  return("$subdirectories{$uuid}" . '')
    if $uuid && exists $subdirectories{$uuid};
  # Return undef for unknown arguments or UUIDs.
  return(undef)
    if $uuid;
  # Return the list (or first entry) if no UUID or exponent was given.
  my (@list);
  push(@list, $subdirectories[0]);
  return(@list)
    if wantarray();
  return($list[0]);
}

# 'file' => relative-path-to-save-file but possibly in a subdirectory per 'dir',
#     %subdirectories, and/or $expMod.
#   If 'file' is given, only 'exp' is required.
#   If 'file' is not given, the others except 'dir' are required.
# 'exp' => Mersenne exponent.
# 'method' => 'P-1', 'P+1', 'pm1', 'pp1', 'o', 'q', etc.
# 'M' => '' or '0' (c or l cofactor) or '1' or '.M' (m cofactor).
# 'dir' => subdirectory (string) or subdirectories (ARRAY ref) to look in.
sub new
{
  my $class = shift;
  my $self = {};
  local $_;
  my ($key);
  while (@_)
    {
      $key = shift;
      return("ECMSave::new given odd number of arguments - should be key/value pairs of a HASH - " . join(' ', caller()))
        unless @_;
      return("ECMSave::new given unexpected key '$key' - only 'file', 'exp', 'method', 'M', and 'dir' are supported - " . join(' ', caller()))
        unless (grep { $key eq $_ } ('file', 'exp', 'method', 'M', 'dir'));
      $self->{$key} = shift;
      $key = undef;
    }
  return("ECMSave::new requires the Mersenne exponent as value of key 'exp' - " . join(' ', caller()))
    unless exists $self->{'exp'};
  return("ECMSave::new given non-numeric value '$self->{'exp'}' for 'exp' - " . join(' ', caller()))
    unless $self->{'exp'} =~ m,^([1-9][0-9]*)$,;
  $self->{'exp'} = $1;
  $self->{'M'} = ''
    if !exists($self->{'M'}) || $self->{'exp'} % 8 != 4;
  $self->{'M'} = ''
    unless $self->{'M'} =~ m,^([.]?M|1)$,;
  $self->{'M'} = '.M'
    if $self->{'M'};
  return("ECMSave::new given 'M' of '$self->{'M'}' that is not '' nor '.M'" . join(' ', caller()))
    unless $self->{'M'} eq '' || $self->{'M'} eq '.M';
  return("ECMSave::new - mkdir " . subdir() . " failed, $!")
    unless -d subdir() || mkdir(subdir());
  return("ECMSave::new - mkdir " . subdir() . "/" . ($self->{'exp'} % $expMod) . " failed, $!")
    unless (-d(subdir() . "/" . ($self->{'exp'} % $expMod))
            || mkdir(subdir() . "/" . ($self->{'exp'} % $expMod)));
  if (!exists($self->{'file'}))
    { # Construct the usual file name if the needed info was provided.
      return("ECMSave::new needs 'M' or 'file' or both - " . join(' ', caller()))
        unless exists $self->{'M'};
      return("ECMSave::new needs 'method' or 'file' or both - " . join(' ', caller()))
        unless exists $self->{'method'};
      # Standardize 'method' to match the option given to the ecm program.
      $self->{'method'} = 'pm1' if $self->{'method'} eq 'P-1';
      $self->{'method'} = 'pm1' if $self->{'method'} eq 'o'; # letter in mers format
      $self->{'method'} = 'pp1' if $self->{'method'} eq 'P+1';
      $self->{'method'} = 'pp1' if $self->{'method'} eq 'q'; # letter in mers format
      return("ECMSave::new given 'method' of $self->{'method'} that is not 'pm1' nor 'pp1' - " . join(' ', caller()))
        unless $self->{'method'} =~ m,^(p[mp]1)$,;
      $self->{'method'} = "$1";
      $self->{'file'} = "$filePrefix" . "$self->{'exp'}.$self->{'method'}" . $self->{'M'};
      $self->{'file'} = subdir() . "/" . ($self->{'exp'} % $expMod) . "/$self->{'file'}";
    }
  else
    {
      my ($pattern) = $filePatterns[0];
      my ($final) = $self->{'file'};
      $final =~ s,^.*/,,;
      return("ECMSave::new given unexpected 'file' name - did not start with 'ecm.', the exponent, and '.' - " . join(' ', caller()))
        unless $final =~ s,^$pattern,,;
      return("ECMSave::new given 'file' that did not include the method - pm1 or pp1 - " . join(' ', caller()))
        unless $final =~ s,^(p[mp]1),, || exists $self->{'method'};
      my ($match) = $1;
      $self->{'method'} = "$match"
        unless exists $self->{'method'};
      return("ECMSave::new given 'method' $self->{'method'} that does not match 'file' $self->{'file'} - " . join(' ', caller()))
        unless $self->{'method'} eq "$match";
      $match = ($final =~ s,^[.]M,,) ? '.M' : '';
      $self->{'M'} = $match
        unless exists $self->{'M'};
      return("ECMSave::new given 'M' $self->{'M'} that does not match 'file' $self->{'file'} - " . join(' ', caller()))
        unless $self->{'M'} eq "$match";
    }
  return("ECMSave::new needs 'file' or all three of 'exp', 'method', and 'M' to create 'file' - " . join(' ', caller()))
    unless exists $self->{'file'};
  return("ECMSave::new - 'file' must be a relative path (no leading /) - " . join(' ', caller()))
    if $self->{'file'} =~ m,^/,;
  $self->{'file'} =~ s,^[.]/,,
    while $self->{'file'} =~ m,^[.]/,;
  $self->{'file'} =~ s,//+,/,g;
  $self->{'file'} =~ s,/[.]/,/,g;
  return("new - 'file' cannot include '..' - " . join(' ', caller()))
    if $self->{'file'} =~ m,[/.][.][/.],;
  return("ECMSAVE::new - 'file' should be alphanumeric perhaps with -+_./ and spaces - " . join(' ', caller()))
    unless $self->{'file'} =~ m,^([-+_./ A-Za-z0-9]*)$,;
  $self->{'file'} = "$1";
  if (exists($self->{'dir'}) && !ref($self->{'dir'}))
    {
      if (-d($self->{'dir'}))
        { # Turn it (as a string) into an ARRAY ref so the safety checks do not have to be duplicated.
          $self->{'dir'} = [ "$self->{'dir'}", ];
        }
      else
        {
          delete $self->{'dir'};
        }
    }
  push(@{$self->{'dir'}}, subdirs());
  if (!-f($self->{'file'}) || !-s($self->{'file'}) || -l($self->{'file'}))
    { # Check under each 'dir' for the 'file', including the directory of the 'file' if not '.'.
      my ($file) = $self->{'file'};
      my ($dir);
      if ($file =~ m,/,)
        {
          $file =~ s,^(.*)/,,;
          $dir = $1;
        }
      my (%list);
      $list{$dir} = 1 if $dir;
      $dir = undef;
      foreach (@{$self->{'dir'}})
        {
          next if m,^/,;
          s,^[.]/,,
            while m,^[.]/,;
          s,//+,/,g;
          next if m,[/.][.][/.],;
          return("ECMSave::new - 'dir' entries should be alphanumeric perhaps with -+_./ and spaces - " . join(' ', caller()))
            unless m,^([-+_./ A-Za-z0-9]*)$,;
          $list{$_} = 1;
        }
      $list{$_} = 1 foreach (subdirs());
      # There are a lot of these files, so use subdirs based on the
      #   exponent to try to split them up into smaller sets.
      $dir = $self->{'exp'} % $expMod;
      foreach (keys(%list))
        {
          next unless -d;
          if (-d("$_/$dir"))
            {
              next unless -f("$_/$dir/$file");
              next unless -s("$_/$dir/$file");
              next if -l("$_/$dir/$file");
              $self->{'file'} = "$_/$dir/$file";
              last;
            }
          next unless -f("$_/$file");
          next unless -s("$_/$file");
          next if -l("$_/$file");
          $self->{'file'} = "$_/$dir/$file",
            last
            if ($dir && (-d("$_/$dir") || mkdir("$_/$dir"))
                && rename("$_/$file", "$_/$dir/$file"));
          $self->{'file'} = "$_/$file";
          last;
        }
      if ((!-f($self->{'file'}) || !-s($self->{'file'}) || -l($self->{'file'}))
          &&  $self->{'method'} =~ m,^p[mp]1$,)
        { # Last chance - look under the directories used by scripts that doP+-1.pl replaced.
          $dir = ($self->{'method'} eq 'pm1') ? 'p-1.sav' : 'p+1.sav';
          if (-d($dir) && !-l($dir))
            { # The older scripts named the files a different way.
              my ($oldFile) = "$dir/$filePrefix$self->{'method'}.$self->{'exp'}$self->{'M'}";
              my ($newFile) = "$filePrefix$self->{'exp'}.$self->{'method'}$self->{'M'}";
              if (-f($oldFile) && -s($oldFile) && !-l($oldFile))
                {
                  $self->{'file'} = subdir() . "/" . ($self->{'exp'} % $expMod) . "/$newFile";
                  return("ECMSave::new - could not rename '$oldFile' to '$self->{'file'}', $! - " . join(' ', caller()))
                    unless rename($oldFile, $self->{'file'});
                }
            }
        }
      return("ECMSave::new - logic error - 'file' not in self - " . join(' ', caller()))
        unless exists $self->{'file'};
      return("ECMSave::new - 'file' '$self->{'file'}' is empty - " . join(' ', caller()))
        if -f $self->{'file'} && ! -s $self->{'file'};
      return("ECMSave::new - 'file' '$self->{'file'}' is a soft link or directory, not supported - " . join(' ', caller()))
        if -l $self->{'file'} || -d $self->{'file'};
    }
  bless $self, $class;
  return($self);
}

sub errors()
{
  my ($self) = shift;
  return undef unless $self->{'error'};
  return $self->{'error'};
}

# Read the save file, check the content in minor ways, and retain and return the content.
sub content
{
  my ($self) = shift;
  local $_;
  my ($fh);
  return($self->{'content'})
    if exists $self->{'content'};
  $self->{'error'} = "ECMSave::content - self has no 'file' - " . join(' ', caller()),
    return(undef)
    unless exists $self->{'file'};
  $self->{'error'} = "ECMSave::content - 'file' is empty or does not exist, errno $! - " . join(' ', caller()),
    return(undef)
    unless -s $self->{'file'};
  $self->{'error'} = "ECMSave::content - could not read $self->{'file'}, errno $! - " . join(' ', caller()),
    return(undef)
    unless open($fh, '<', "$self->{'file'}");
  my ($save);
  while (<$fh>)
    { # make sure each var=value pair is separated by ';' and at least one following space
      s/([$CRLF])+/ /g;
      $save = $_, next unless defined $save;
      $save .= "; " unless $save =~ m/; +$/;
      $save .= "$_";
    }
  $self->{'error'} = "ECMSave::content - file $self->{'file'} had no data - " . join(' ', caller()),
    return undef
    unless $save;
  # Drop redundant, leading, and trailing separators and spaces.
  $save =~ s/; +; */; /g;
  $save =~ s/^[; ]+//;
  $save =~ s/[; ]+$//;
  $save =~ s/([$CRLF])+/ /g;
  $save =~ s/ $//;
  # But the ecm program requires a trailing ';', so ...
  $self->{'content'} = $save . '; ';
  return($self->{'content'});
}

# Parse the content into a HASH of var=value (each var a key) and
#   retain the HASH in $self->{'data'}.
sub parse
{
  my ($self) = shift;
  local $_;
  return(1) if exists $self->{'data'}; # already called
  $self->content()
    unless exists $self->{'content'};
  return undef
    unless exists $self->{'content'};
  my ($varHash);
  my (@vars) = split(/; +/, $self->{'content'});
  while (@vars)
    {
      my (@kv) = split(/=/, shift(@vars));
      if ($#kv == 1 && $kv[0] =~ m/^([A-Z0-9]+)$/)
        {
          my ($key) = "$1";
          $varHash->{$key} = Math::BigInt->new("$1"),
            next
            if exists($ECMnumerics{$key}) && $kv[1] =~ m/^$ECMnumerics{$key}$/;
          $varHash->{$key} = "$1",
            next
            if $kv[1] =~ m/^$ECMvaluePattern$/;
        }
      $self->{'error'} = "ECMSave::parse - unexpected content in ecm save file - " . join('=', @kv);
      $self->{'error'} = "ECMSave::parse - keys and values empty - " . join(' ', @vars)
        unless @kv;
      $self->{'error'} = "ECMSave::parse - vars empty - " . $self->{'content'}
        unless @kv || @vars;
      $self->{'error'} .= " - " . join(' ', caller());
      return(undef); # unexpected content
    }
  $self->{'data'} = $varHash;
  return(1);
}

# Return the Mersenne's exponent.
sub exp
{
  my ($self) = shift;
  return(undef)
    unless exists $self->{'exp'};
  return($self->{'exp'});
}

# Return which cofactor the save file is for.
sub M
{
  my ($self) = shift;
  return(undef)
    unless exists $self->{'M'};
  return($self->{'M'});
}

# Return the directory(ies) given to new().
sub dirs
{
  my ($self) = shift;
  my (@dirs);
  return(@dirs) unless $self->{'dir'};
  return(@{$self->{'dir'}}) if ref($self->{'dir'}) eq 'ARRAY';
  return(@dirs);
}

# Return an arbitrary value from the saved data.
sub get
{
  my ($self) = shift;
  my ($key) = shift;
  $self->{'error'} = "ECMSave::get given invalid key '$key' - " . join(' ', caller()),
    return(undef)
    unless $key =~ m/^([A-Z0-9]+)$/;
  $key = "$1";
  return(undef)
    unless $self->parse();
  $self->{'error'} = "ECMSave::get given unexpected key '$key' - " . join(' ', caller()),
    return(undef)
    unless exists $self->{'data'}{"$key"};
  return $self->{'data'}{"$key"};
}

# Return the saved P-1 or P+1 stage 1 bound, 'B1' in the HASH from parse().
sub bound
{
  my ($self) = shift;
  return(undef)
    unless $self->get('B1');
  return Math::BigInt->new($self->get('B1'));
}

# Return the METHOD, either 'P-1' or 'P+1' for the current (Jan 2019) ecm program.
sub method
{
  my ($self) = shift;
  return $self->{'method'} if $self->{'method'};
  return $self->get('METHOD');
}

# Return the number the ecm run tried to factor.
sub n
{
  my ($self) = shift;
  return $self->get('N');
}

# Return the file name.
sub file
{
  my ($self) = shift;
  return undef unless exists $self->{'file'} && ! -l $self->{'file'};
  return $self->{'file'};
}

# Return the checksum.
sub checksum
{
  my ($self) = shift;
  return $self->{'checksum'} if exists $self->{'checksum'};
  $self->{'checksum'} = $self->get('CHECKSUM'),
    return($self->{'checksum'})
    if $self->get('CHECKSUM');
  return undef
    unless $self->calcChecksum();
  $self->{'checksum'} = $self->calcChecksum();
  return $self->{'checksum'};
}

sub calcChecksum()
{ # Calculate the checksum from other info per gmp-ecm/resume.c read_resumefile_line().
  my ($self) = shift;
  my ($val) = $self->bound() % $ECMchecksumMod;
  $val *= ($self->get('SIGMA') % $ECMchecksumMod)
    if $self->get('SIGMA');
  $val *= ($self->get('A') % $ECMchecksumMod)
    if $self->get('A');
  $val *= $self->n() % $ECMchecksumMod;
  $val *= (Math::BigInt->new('' . $self->get('X')) % $ECMchecksumMod)
    if $self->get('X');
  $val *= ($self->get('Z') % $ECMchecksumMod)
    if $self->get('Z');
  $val *= (1 + $self->get('PARAM')) % $ECMchecksumMod
    if defined $self->get('PARAM');
  return($val % $ECMchecksumMod);
}

# Change the file name.
sub setFile
{
  my ($self) = shift;
  my ($newfn) = shift;
  return undef unless -f($newfn) && -s($newfn) && ! -l $newfn;
  $self->{'file'} = $newfn;
  return($self);
}

# Return the content of the save file with a prefix line and suffix line ...
#   ... containing the "generic" file name, not the actual one, ...
#   ... so that other scripts don't have to allow arbitrary names.
sub str
{
  my ($self) = shift;
  return undef unless $self && ref($self) eq 'ECMSave';
  return undef unless $filePrefix;
  return undef unless exists $self->{'exp'};
  return undef unless exists $self->{'method'};
  return undef unless exists $self->{'M'};
  my ($content) = $self->content();
  return undef unless $content;
  my ($name) = "$filePrefix$self->{'exp'}.$self->{'method'}$self->{'M'}";
  return("save file begin $name\n"
         . "$content\n"
         . "save file end $name\n");
}

# Print the save file with a prefix line and suffix line ...
#   ... containing the "generic" file name, not the actual one, ...
#   ... so that other scripts don't have to allow arbitrary names.
sub print
{
  my ($self) = shift;
  my ($str) = $self->str();
  return undef unless $str;
  my ($fh) = shift;
  $fh = *STDOUT unless $fh;
  return(print($fh "$str"));
}

sub check
{
  my ($self) = shift;
  my ($mers) = shift;
  $mers = Mersenne->new('n' => $self->{'exp'})
    unless $mers;
  $self->{'error'} = "ECMSave::check - argument is not a Mersenne reference",
    return(undef)
    unless ref($mers) eq 'Mersenne';
  $self->{'error'} = "ECMSave::check - exponent $self->{'exp'} does not have an 'M' cofactor",
    return(undef)
    unless $self->{'M'} eq '' || $self->{'exp'} % 8 == 4;
  $self->{'error'} = ("ECMSave::check - exponent $self->{'exp'} does not match Mersenne ref exponent "
                      . $mers->n()),
    return(undef)
    unless $mers->n() eq $self->{'exp'};
  my ($calcCS) = $self->calcChecksum();
  $self->{'error'} = ("ECMSave::check - CHECKSUM " . $self->get('CHECKSUM') . " does not match calculated value " . $calcCS . "\n"),
    return(undef)
    if defined($calcCS) && $self->checksum() != $calcCS;
  my (@f) = $mers->f();
  # $save->n() is almost always a cofactor, so compare with those first.
  unshift(@f, $mers->c());
  my ($found);
  while (@f)
    {
      return(1) if $self->n() == $f[0];
      $found = 1, last if $self->n() % $f[0] == 0;
      $found = 2, last if $f[0] % $self->n() == 0;
      shift(@f);
    }
  $self->{'error'} = "ECMSave::check - N in $self->{'file'} is not a factor or cofactor of M$self->{'exp'}",
    return(undef)
    unless $found;
  $mers->f($self->n())
    if !$#f || $f[0] ne $self->n();
  return(1);
}

# Create a save file from the data in the given hash, mostly 'content' and 'file'.
#   Note - this is not a member sub.
sub createSaveFile(%)
{
  my (%saveInfo) = @_;
  local $_;
  return("ECMSave::createSaveFile - no 'file' given")
    unless exists $saveInfo{'file'};
  return("ECMSave::createSaveFile - unexpected 'file' name - must be 'ecm.<exp>.<method>[.M]'")
    unless $saveInfo{'file'} =~ m,^(ecm[.]([1-9][0-9]+)[.](p[mp]1)([.]M)?)$,;
  my ($file) = $1 if $1;
  my ($exp) = $2 ? $2 : $saveInfo{'exp'};
  my ($method) = $3 ? $3 : $saveInfo{'method'};
  my ($M) = $4 ? $4 : $saveInfo{'M'};
  return("ECMSave::createSaveFile - missing 'exp', 'method', and/or 'M'")
    unless $exp && $method && defined($M);
  return("ECMSave::createSaveFile - 'exp' must be a number")
    unless $exp =~ m,^([1-9][0-9]*)$,;
  $exp = $1;
  return("ECMSave::createSaveFile - 'method' must be 'pm1' or 'pp1'")
    unless $method =~ m,^(p[mp]1)$,;
  $method = $1;
  return("ECMSave::createSaveFile - 'M' must be '' or '.M'")
    unless $M =~ m,^(|[.]M)$,;
  $M = $1;
  $file = "$filePrefix$exp.$method$M" unless $file;
  my ($subdir) = subdir();
  return("ECMSave::createSaveFile - mkdir $subdir failed, $!")
    if ! -d $subdir && !mkdir($subdir);
  $subdir .= "/" . ($exp % $expMod);
  return("ECMSave::createSaveFile - mkdir $subdir failed, $!")
    if ! -d $subdir && !mkdir($subdir);
  $subdir .= '/' if -d $subdir;
  $subdir = '' unless -d $subdir && ! -l $subdir;
  $file = "$subdir$file";
  my ($cnt) = 0;
  if (-s("$file") && -f("$file"))
    {
      ++$cnt while -s "$file.$cnt" && -f "$file.$cnt";
      $file = "$file.$cnt";
    }
  return("ECMSave::createSaveFile - could not create '$file'")
    unless open(my $saveFH, '>', "$file");
  print($saveFH $saveInfo{'content'});
  unlink("$file"),
    return("ECMSave::createSaveFile - could not close '$file'")
    unless close($saveFH);
  my ($save) = ECMSave->new('exp' => $exp, 'file' => "$file", 'M' => $M);
  unlink("$file"),
    return($save)
    unless ref($save);
  my ($bound) = $save->bound();
  unlink("$file"),
    return("ECMSave::createSaveFile - unrecognized stage1 bound written to '$file'")
    unless $bound;
  return $save;
}

# Find the "best"/"furthest" save file for the given exponent, method, and cofactor.
# If $mers, it should be a Mersenne instance/object/ref for M($exp).
# $warn is a flag to print warnings if something unexpected is found.
# If $delete, then other matching files that are found are removed to save disk space
#   _if_ they are old enough.  But note that soft links are always removed if
#   $delete is 'true' - these files and directories should never be soft links.
#   Note that $delete should be a time() value - seconds since the UNIX/Linux Epoch.
#     Files modified since that time() will never be deleted here to avoid deleting
#       a recently created file the caller doesn't know about.
sub findBestSaveFile
{
  my ($exp, $method, $M, $mers, $warn, $delete) = @_;
  local $_;
  my ($save, $stats, $stats2);
  while (!defined($save))
    { # If there's a problem with the save file found by new(),
      #   try again if it is removed here.
      $save = ECMSave->new('exp' => $exp, 'method' => $method, 'M' => $M);
      return($save) # this is (should be) a string ...
        unless $save && ref($save); # ... if this condition is true
      if ($save->errors())
        {
          print("# ECMSave::new - " . $save->errors() . "\n")
            if $warn;
          return(undef);
        }
      return($save) # not an actual file, just an ECMSave object
        if !$save->file() || !-s $save->file();
      $mers = undef if $mers && ref($mers) ne 'Mersenne';
      $mers = Mersenne->new('n' => $exp) unless $mers;
      $stats = stat($save->file())
        if $delete;
      last # found a file that "makes sense" and matches the Mersenne, $exp, $method, etc.
        if $save->check($mers);
      print("# ECMSave::findBestSaveFile - " . $save->file() . " failed check() - " . $save->errors());
      return(undef)
        unless $delete && $stats->mtime() < $delete;
      unlink($save->file());
      $save = undef;
    }
  my ($pattern) = '^' . "$filePrefix";
  $pattern =~ s,[.],[.],g; # this looks like but is not a no-op - literal '.'s are changed to '[.]'
  $pattern .= $save->exp() . '[.]' . $save->method() . $save->M();
  my ($expDir) = $exp % $expMod;
  my (@dirs, %dirs); # preserve the order but skip duplicates within the next loop
  push(@dirs, "$_/$expDir", "$_")
    foreach (subdirs(), $save->dirs());
  foreach my $d (@dirs)
    {
      next
        if exists $dirs{$d};
      $dirs{$d} = 1;
      print("# ECMSave::findBestSaveFile - directory '$d' is a soft link; removing it\n"),
        unlink($d),
        next
        if $delete && -l($d);
      print("# ECMSave::findBestSaveFile - directory '$d' is a soft link; ignoring it\n")
        if $warn && -l($d);
      next
        unless -d($d) && !-l($d);
      my ($dfh);
      if (!opendir($dfh, $d))
        {
          print("# ECMSave::findBestSaveFile - could not opendir '$d', $!\n")
            if $warn;
          next;
        }
      my (@files) = grep { -f("$d/$_") && -s(_) && m,$pattern, && !m,[.](log|cofactor[M]?)$, } readdir($dfh);
      closedir($dfh)
        || !$warn
        || print("# ECMSave::findBestSaveFile - could not closedir '$d', $!\n");
      if ($save->exp() % 8 == 4 && $save->M() eq '')
        { # Filter out any '.M' files that $pattern matched.
          my ($pattern2) = "$pattern" . '[.]M';
          @files = grep { !m,$pattern2, } (@files, );
        }
      next unless @files;
      foreach my $f (@files)
        {
          print("# ECMSave::findBestSaveFile - '$d/$f' is a soft link; removing it\n"),
            unlink("$d/$f"),
            next
            if $delete && -l("$d/$f");
          print("# ECMSave::findBestSaveFile - '$d/$f' is a soft link; ignoring it\n")
            if $warn && -l("$d/$f");
          next
            if -l("$d/$f");
          next
            if $save->file() eq "$d/$f";
          my ($save2) = ECMSave->new('exp' => $exp, 'file' => "$d/$f", 'M' => $save->M());
          if (!$save2 || !ref($save2))
            {
              print('# ' . $save2 . "\n")
                if $save2 && $warn;
              next;
            }
          if (!$save2->check($mers))
            {
              print("# ECMSave::findBestSaveFile - " . $save2->file() . " failed check() - " . $save2->errors() . "\n")
                if $warn;
              if ($delete)
                {
                  $stats2 = stat($save2->file());
                  unlink($save2->file())
                    if $stats2->mtime() < $delete;
                }
              next;
            }
          if ($save->method() ne $save2->method())
            {
              print("# ECMSave::findBestSaveFile - '$d/$f' has method " . $save2->method() . " rather than " . $save->method() . "\n")
                if $warn;
              next;
            }
          if ($save->M() ne $save2->M())
            {
              print("# ECMSave::findBestSaveFile - '$d/$f' has M " . $save2->M() . " rather than " . $save->M() . "\n")
                if $warn && $save->M();
              next;
            }
          if ($save->bound() > $save2->bound())
            {
              print("# ECMSave::findBestSaveFile - '" . $save2->file() . "' has smaller stage1 bound\n")
                if $warn;
              if ($delete)
                {
                  $stats2 = stat($save2->file());
                  unlink($save2->file())
                    if $stats2->mtime() < $delete;
                }
              next;
            }
          if ($save->bound() < $save2->bound())
            {
              print("# ECMSave::findBestSaveFile - '" . $save->file() . "' has smaller stage1 bound\n")
                if $warn;
              unlink($save->file())
                if $delete && $stats->mtime() < $delete;
              $save = $save2;
              $stats = stat($save->file());
              next;
            }
          if ($save->n() < $save2->n())
            {
              print("# ECMSave::findBestSaveFile - '" . $save2->file() . "' has larger N (cofactor)\n")
                if $warn;
              if ($delete)
                {
                  $stats2 = stat($save2->file());
                  unlink($save2->file())
                    if $stats2->mtime() < $delete;
                }
              next;
            }
          if ($save->n() > $save2->n())
            {
              print("# ECMSave::findBestSaveFile - '" . $save->file() . "' has larger N (cofactor)\n")
                if $warn;
              unlink($save->file())
                if $delete && $stats->mtime() < $delete;
              $save = $save2;
              $stats = stat($save->file());
              next;
            }
          #!!incomplete - other comparisons?
          print("# ECMSave::findBestSaveFile - '" . $save2->file() . "' may duplicate '" . $save->file() . "'\n")
            if $warn;
          if ($delete)
            {
              $stats2 = stat($save2->file());
              unlink($save2->file())
                if $stats2->mtime() < $delete - (-f('mysql.devel') ? 0 : 7*24*60*60);
            }
        }
    }
  my ($savefn) = $save->file();
  return($save)
    unless -f($savefn) && -s($savefn) && !-l($savefn);
  my ($dir) = subdir() . "/" . ($save->exp() % $expMod) . "/";
  if (substr($savefn, 0, length($dir)) ne $dir)
    { # The file isn't in the "usual" place, so move it there ... but don't remove an existing file.
      my (@pieces) = split(/\//, $savefn);
      my ($last) = pop(@pieces);
      my ($newfn) = "$dir$last";
      if (-f($newfn) && -s($newfn))
        {
          my ($cnt) = 0;
          ++$cnt
            while -f("$newfn.$cnt") || -s("$newfn.$cnt") || -l("$newfn.$cnt");
          # If this rename() fails, just leave the file in the non-"usual" place.
          return($save)
            unless rename($newfn, "$newfn.$cnt");
        }
      $save->{'error'} = "ECMSave::findBestSaveFile - could not rename '$savefn' to '$newfn', $! - " . join(' ', caller()),
        return($save)
        unless rename($savefn, $newfn);
      $save->setFile($newfn);
    }
  return($save);
}

1;
