#!/usr/bin/perl -Tw
# An interface to SQL for mers format data.

package MersDB;

use strict;
use strict 'vars';
use Math::BigInt try => 'GMP';
use Math::Prime::Util;
use DBI;

use parent 'Exporter';

# Functions and variables which are exported by default
#our @EXPORT      = qw();
# Functions and variables which can be optionally exported
our @EXPORT_OK   = qw();

use Class::Tiny qw ( dbuser dbpassword );

BEGIN { push(@INC, '.') if -s('Mersenne.pm'); }

use MersInfo;
use Mersenne;

our ($VERSION) = MersInfo::version();

# Debug info - sometimes, DBI complains:
#   Issuing rollback() due to DESTROY without explicit disconnect() of DBD::mysql::db handle mersdevel at ./fetch.pl line 108.
# ... without any info about which SQL statement created the handle.
# See the end of execute() and calls to DBI::finish().
my (%executions);

END
{
  print("list of unfinished statement handles\n")
    if scalar(keys(%executions));
  print("DBI::finish not called for '" . join(' ', @{$executions{$_}}) . "'\n")
    foreach (keys(%executions));
}

# Valid letters of the mers format; constant after this.
#   PDSLM must be before clm or logic in find() needs to be modified.
our ($validLetters) = 'CDEGHJLMPSXUgclmoq';

# Letters that indicate the Mersenne is completely factored.
our ($doneLetters) = 'PD';
# Pattern for them - $1 needs to be set to the one letter.  Constant after this.
our ($donePattern) = '^([' . $doneLetters . '])$';

# These next two values are used to store numbers larger than SQL supports directly.
#   Only currently used for factors ('C').
# The largest number of bits in a single number that SQL supports.
our ($SQLmaxBits) = Math::BigInt->new('64');
# The largest value SQL can store - note the decrement.
our ($SQLmaxInt) = Math::BigInt->bone()->blsft($SQLmaxBits)->bdec();

# These are default values for attributes that can be given to DBI::connect().
#   'RaiseError' is disabled in some cases, notably in reserve() as noted there.
my (%dbiAttrs) = ( 'RaiseError' => 1,
                   'AutoCommit' => 0, #!!commenting enables it despite warning about deprecation
                   'AutoInactiveDestroy' => 1,
                   'ShowErrorStatement' => 1, 'FetchHashKeyName' => 'NAME_lc',
                   'PrintError' => 0, 'Taint' => 1,
                   #!!latest mysql says this is deprecated - 2023 Sep 1st - but
                   #'mysql_auto_reconnect' => 0, # ... mysql.pm still has it set to 1
  );

# A list of special UUIDs, not the usual random ones, used by some of the scripts.
#   The mersServer and runSPRP reserve the related special exponent to prevent
#      running more than one at a time.
#!!only "local" (special?) UUIDs should be able to modify some tables
#!!only "allowed" UUIDs should be able to modify data in the db
my (@specialUUIDs) = ('mersServer', 'runSPRP', 'release', 'reserve', 'save', 'fetch', );
my (%specialExponents);
$specialExponents{$specialUUIDs[$_]} = $_,
  $specialExponents{$_} = $specialUUIDs[$_]
  foreach (0 .. $#specialUUIDs);

my ($bigintPattern) = MersInfo::patterns('bigint');
my ($floatPattern) = MersInfo::patterns('float');
my ($uuidPattern) = MersInfo::patterns('uuid');

# Each special UUID has an associated exponent for reservations.
sub specialExponent($)
{
  my ($arg) = shift;
  return(undef)
    unless exists $specialExponents{$arg};
  return($specialExponents{$arg});
}

# If a table's 'value' hash has more entries than this, reduceMemory() will trim
#   some values.  Nothing here calls reduceMemory().
my ($maxValues) = 5000;

# This maps each mers letter with a database table to the related SQL (e.g., to create the table).
#   Many of the database table names are the mers letter; the others are fairly obvious.
our (%tables) = 
  ( 'X' =>
    { # mers letter X is for extent of 1strs factoring
      #   An extent of 0 means all Mersennes with that exponent and smaller are completely factored.
      'columns' => [ 'x', 'extent', ],
      'patterns' => [ MersInfo::patterns('bigint'), MersInfo::patterns('bigint'), ],
      'create' => q{ create table if not exists `X` (`x` BIGINT UNSIGNED PRIMARY KEY UNIQUE, `extent` BIGINT UNSIGNED NOT NULL) },
      'replace' => q{ replace X (x, extent) values(?, ?) },
      'fetch' => q{ select x, extent from X },
      'query' => q{ select x, extent from X where x = ? },
      'delete' => q{ delete from X where x = ? },
      'deleteIndex' => 1,
      'queryIndex' => 1,
      'keyIndex' => 1, # field in mers format - array subscript after split(/ /, 'M( 1260 )X: 0'), e.g.; 1 is the exponent
      'valueIndex' => 3, # field in mers format; 3 is the extent, stage 1 bound, effort, digits in cofactor, etc., depending on the mers letter ('X' for this table)
      'value' => { 4 => 0, }, # M(1) thru M(4) are completely factored or prime
      # index of ? in SQL prepare statement mapped to field (array index after split()) of mers format
      1 => 1, # exponent (of the Mersenne number, M(n) - the power of 2, n, within "2^n - 1")
      2 => 3, # extent (of 1strs trial factoring)
    },
    'C' =>
    { # mers letter C has factors.  Large factors are split into 'limbs' (ala libgmp) of 64 bits (BIGINT UNSIGNED), low bits in limb 0, next 64 in limb 1, etc.
      'columns' => [ 'exp', 'ord', 'limb', 'factor', ],
      'patterns' => [ MersInfo::patterns('bigint'), MersInfo::patterns('bigint'), MersInfo::patterns('bigint'), MersInfo::patterns('bigint'), ],
      'create' => q{ create table if not exists `C` (`exp` BIGINT UNSIGNED NOT NULL, `ord` INT UNSIGNED NOT NULL, `limb` INT UNSIGNED NOT NULL, `factor` BIGINT UNSIGNED NOT NULL, UNIQUE ( exp, ord, limb )) },
      'insert' => q{ insert C (exp, ord, limb, factor) values (?, ?, ?, ?) },
      'delete' => q{ delete from C where exp = ? },
      'deleteIndex' => 1,
      # 'fetch' would retrieve way too much data
      'query' => q{ select exp, ord, limb, factor from C where exp = ? order by ord, limb },
      'queryIndex' => 1,
      'keyIndex' => 1,
      1 => 1, # exponent
      2 => 3, # ord (# of factor of this Mersenne, starting at 0)
      3 => 4, # limb # (limb 0's value is the factor % 2^64)
      4 => 5, # limb value
    },
    'c' =>
    { # mers letter c has cofactor digits
      'columns' => [ 'exp', 'digits', ],
      'patterns' => [ MersInfo::patterns('bigint'), MersInfo::patterns('bigint'), ],
      'create' => q{ create table if not exists `c` (`exp` BIGINT UNSIGNED PRIMARY KEY UNIQUE, `digits` BIGINT UNSIGNED NOT NULL) },
      'replace' => q{ replace c (exp, digits) values (?, ?) },
      'delete' => q{ delete from c where exp = ? },
      'deleteIndex' => 1,
        #!!This table is far too large to 'fetch', though not as much as for 'C'
        #!!'fetch' => q{ select exp, digits from c },
      # use a join because the sprp table tracks whether the cofactor needs an SPRP check (clm work)
      'fetchByPriority' => q{ select c.exp, c.digits from c left join (sprp) on (sprp.exp = c.exp) where sprp.needed != 0 and c.digits > ? and not exists (select * from reservations where reservations.exp = c.exp and reservations.kind = 'c') and not exists (select * from done where done.exp = c.exp) order by c.digits asc limit ? },
      'fetchByPriorityPatterns' => [ MersInfo::patterns('bigint'), MersInfo::patterns('bigint'), ],
      # ' limit 1 *$' is required so that next() can return the exponent rather than the statement handle
      'next' => q{ select c.exp from c left join (sprp) on (sprp.exp = c.exp) where sprp.needed != 0 and not exists (select * from reservations where reservations.exp = c.exp and reservations.kind = 'c') and not exists (select * from done where done.exp = c.exp) order by c.exp asc limit 1 },
      'query' => q{ select exp, digits from c where exp = ? },
      'queryIndex' => 1,
      'keyIndex' => 1,
      'valueIndex' => 3,
      'value' => { },
      1 => 1, # exponent, cannot be 4 % 8
      2 => 3, # digits in (composite) cofactor
    },
    'l' =>
    { # mers letter l has cofactor digits
      'columns' => [ 'exp', 'digits', ],
      'patterns' => [ MersInfo::patterns('bigint'), MersInfo::patterns('bigint'), ],
      'create' => q{ create table if not exists `l` (`exp` BIGINT UNSIGNED PRIMARY KEY UNIQUE, `digits` BIGINT UNSIGNED NOT NULL) },
      'replace' => q{ replace l (exp, digits) values (?, ?) },
      'delete' => q{ delete from l where exp = ? },
      'deleteIndex' => 1,
        #!!This table is far too large to 'fetch'
        #!!'fetch' => q{ select exp, digits from l },
      'fetchByPriority' => q{ select l.exp, l.digits from l left join (sprp) on (sprp.exp = l.exp) where sprp.needed != 0 and l.digits > ? and not exists (select * from reservations where reservations.exp = l.exp and reservations.kind = 'c') and not exists (select * from done where done.exp = l.exp) order by l.digits asc limit ? },
      'fetchByPriorityPatterns' => [ MersInfo::patterns('bigint'), MersInfo::patterns('bigint'), ],
      'next' => q{ select l.exp from l left join (sprp) on (sprp.exp = l.exp) where sprp.needed != 0 and not exists (select * from reservations where reservations.exp = l.exp and reservations.kind = 'c') and not exists (select * from done where done.exp = l.exp) order by l.exp asc limit 1 },
      'query' => q{ select exp, digits from l where exp = ? },
      'queryIndex' => 1,
      'keyIndex' => 1,
      'valueIndex' => 3,
      'value' => { },
      1 => 1, # exponent, must be 4 % 8
      2 => 3, # digits in (composite) cofactor
    },
    'm' =>
    { # mers letter m has cofactor digits
      'columns' => [ 'exp', 'digits', ],
      'patterns' => [ MersInfo::patterns('bigint'), MersInfo::patterns('bigint'), ],
      'create' => q{ create table if not exists `m` (`exp` BIGINT UNSIGNED PRIMARY KEY UNIQUE, `digits` BIGINT UNSIGNED NOT NULL) },
      'replace' => q{ replace m (exp, digits) values (?, ?) },
      'delete' => q{ delete from m where exp = ? },
      'deleteIndex' => 1,
        #!!This table is far too large to 'fetch'
        #!!'fetch' => q{ select exp, digits from m },
      'fetchByPriority' => q{ select m.exp, m.digits from m left join (sprp) on (sprp.exp = m.exp) where sprp.needed != 0 and m.digits > ? and not exists (select * from reservations where reservations.exp = m.exp and reservations.kind = 'c') and not exists (select * from done where done.exp = m.exp) order by m.digits asc limit ? },
      'fetchByPriorityPatterns' => [ MersInfo::patterns('bigint'), MersInfo::patterns('bigint'), ],
      'next' => q{ select m.exp from m left join (sprp) on (sprp.exp = m.exp) where sprp.needed != 0 and not exists (select * from reservations where reservations.exp = m.exp and reservations.kind = 'c') and not exists (select * from done where done.exp = m.exp) order by m.exp asc limit 1 },
      'query' => q{ select exp, digits from m where exp = ? },
      'queryIndex' => 1,
      'keyIndex' => 1,
      'valueIndex' => 3,
      'value' => { },
      1 => 1, # exponent, must be 4 % 8
      2 => 3, # digits in (composite) cofactor
    },
    'S' =>
    { # mers letter S indicates the single cofactor is SPRP with the cofactor digit count
      'columns' => [ 'exp', 'digits', ],
      'patterns' => [ MersInfo::patterns('bigint'), MersInfo::patterns('bigint'), ],
      'create' => q{ create table if not exists `S` (`exp` BIGINT UNSIGNED PRIMARY KEY UNIQUE, `digits` BIGINT UNSIGNED NOT NULL) },
      'replace' => q{ replace S (exp, digits) values (?, ?) },
      'delete' => q{ delete from S where exp = ? },
      'deleteIndex' => 1,
      'fetch' => q{ select exp, digits from S },
      'fetchByPriority' => q{ select S.exp, S.digits from S left join (sprp) on (sprp.exp = S.exp) where sprp.needed = 0 and S.digits > ? and not exists (select * from reservations where reservations.exp = S.exp and reservations.kind = 'S') and not exists (select * from done where done.exp = S.exp) order by S.exp asc limit ? },
      'fetchByPriorityPatterns' => [ MersInfo::patterns('bigint'), MersInfo::patterns('bigint'), ],
      'next' => q{ select S.exp from S left join (sprp) on (sprp.exp = S.exp) where sprp.needed = 0 and not exists (select * from reservations where reservations.exp = S.exp and reservations.kind = 'S') and not exists (select * from done where done.exp = S.exp) order by S.exp asc limit 1 },
      'query' => q{ select exp, digits from S where exp = ? },
      'queryIndex' => 1,
      'keyIndex' => 1,
      'valueIndex' => 3,
      'value' => { },
      1 => 1, # exponent, cannot be 4 % 8
      2 => 3, # digits in (SPRP) cofactor
    },
    'L' =>
    { # mers letter L has cofactor digits of SPRP cofactor
      'columns' => [ 'exp', 'digits', ],
      'patterns' => [ MersInfo::patterns('bigint'), MersInfo::patterns('bigint'), ],
      'create' => q{ create table if not exists `L` (`exp` BIGINT UNSIGNED PRIMARY KEY UNIQUE, `digits` BIGINT UNSIGNED NOT NULL) },
      'replace' => q{ replace L (exp, digits) values (?, ?) },
      'delete' => q{ delete from L where exp = ? },
      'deleteIndex' => 1,
      'fetch' => q{ select exp, digits from L },
      'fetchByPriority' => q{ select L.exp, L.digits from L left join (M, sprp) on (M.exp = L.exp and sprp.exp = L.exp) where sprp.needed = 0 and L.exp > ? and not exists (select * from reservations where reservations.exp = L.exp and reservations.kind = 'S') and not exists (select * from done where done.exp = L.exp) order by L.exp asc limit ? },
      'fetchByPriorityPatterns' => [ MersInfo::patterns('bigint'), MersInfo::patterns('bigint'), ],
      'next' => q{ select L.exp from L left join (M, sprp) on (M.exp = L.exp and sprp.exp = L.exp) where sprp.needed = 0 and not exists (select * from reservations where reservations.exp = L.exp and reservations.kind = 'S') and not exists (select * from done where done.exp = L.exp) order by L.exp asc limit 1 },
      'query' => q{ select exp, digits from L where exp = ? },
      'queryIndex' => 1,
      'keyIndex' => 1,
      'valueIndex' => 3,
      'value' => { },
      1 => 1, # exponent, must be 4 % 8
      2 => 3, # digits in (SPRP) cofactor
    },
    'M' =>
    { # mers letter M has cofactor digits of SPRP cofactor
      'columns' => [ 'exp', 'digits', ],
      'patterns' => [ MersInfo::patterns('bigint'), MersInfo::patterns('bigint'), ],
      'create' => q{ create table if not exists `M` (`exp` BIGINT UNSIGNED PRIMARY KEY UNIQUE, `digits` BIGINT UNSIGNED NOT NULL) },
      'replace' => q{ replace M (exp, digits) values (?, ?) },
      'delete' => q{ delete from M where exp = ? },
      'deleteIndex' => 1,
      'fetch' => q{ select exp, digits from M },
      'fetchByPriority' => q{ select M.exp, M.digits from M left join (L, sprp) on (L.exp = M.exp and sprp.exp = M.exp) where sprp.needed = 0 and M.exp > ? and not exists (select * from reservations where reservations.exp = M.exp and reservations.kind = 'S') and not exists (select * from done where done.exp = M.exp) order by M.exp asc limit ? },
      'fetchByPriorityPatterns' => [ MersInfo::patterns('bigint'), MersInfo::patterns('bigint'), ],
      'next' => q{ select M.exp from M left join (L, sprp) on (L.exp = M.exp and sprp.exp = M.exp) where sprp.needed = 0 and not exists (select * from reservations where reservations.exp = M.exp and reservations.kind = 'S') and not exists (select * from done where done.exp = M.exp) order by M.exp asc limit 1 },
      'query' => q{ select exp, digits from M where exp = ? },
      'queryIndex' => 1,
      'keyIndex' => 1,
      'valueIndex' => 3,
      'value' => { },
      1 => 1, # exponent, must be 4 % 8
      2 => 3, # digits in (SPRP) cofactor
    },
    'o' =>
    { # mers letter o has P-1 stage 1 and stage 2 bounds and perhaps priority (which can be calculated from the others)
      'columns' => [ 'exp', 'b1', 'b2', 'pri', ],
      'patterns' => [ MersInfo::patterns('bigint'), MersInfo::patterns('bigint'), MersInfo::patterns('bigint'), MersInfo::patterns('float'), ],
      'create' => q{ create table if not exists `o` (`exp` BIGINT UNSIGNED PRIMARY KEY UNIQUE, `b1` BIGINT UNSIGNED NOT NULL, `b2` BIGINT UNSIGNED NOT NULL, `pri` DOUBLE NOT NULL) },
      'replace' => q{ replace o (exp, b1, b2, pri) values (?, ?, ?, ?) },
      'delete' => q{ delete from o where exp = ? },
      'deleteIndex' => 1,
        #!!This table is far too large to 'fetch'
        #!!'fetch' => q{ select exp, b1, b2, pri from o },
      'fetchByPriority' => q{ select o.exp, o.b1, o.b2, o.pri from o left join (sprp) on (sprp.exp = o.exp) where sprp.needed = 0 and o.pri > ? and not exists (select * from reservations where reservations.exp = o.exp) and not exists (select * from done where done.exp = o.exp) and not exists (select * from S where S.exp = o.exp) order by o.pri asc limit ? },
      'fetchByPriorityPatterns' => [ MersInfo::patterns('float'), MersInfo::patterns('bigint'), ],
      'next' => q{ select o.exp from o left join (sprp) on (sprp.exp = o.exp) where sprp.needed = 0 and not exists (select * from reservations where reservations.exp = o.exp) and not exists (select * from done where done.exp = o.exp) and not exists (select * from S where S.exp = o.exp) order by o.pri asc limit 1 },
      'query' => q{ select exp, b1, b2, pri from o where exp = ? },
      'queryIndex' => 1,
      'keyIndex' => 1,
      'valueIndex' => 3,
      'value' => { },
      1 => 1, # exponent
      2 => 3, # stage 1 bound
      3 => 4, # stage 2 bound
      4 => 5, # priority
      'pri' => 4, # index of the priority in SQL statments
    },
    'q' =>
    { # mers letter q has P+1 stage 1 and stage 2 bounds
      'columns' => [ 'exp', 'b1', 'b2', 'pri', ],
      'patterns' => [ MersInfo::patterns('bigint'), MersInfo::patterns('bigint'), MersInfo::patterns('bigint'), MersInfo::patterns('float'), ],
      'create' => q{ create table if not exists `q` (`exp` BIGINT UNSIGNED PRIMARY KEY UNIQUE, `b1` BIGINT UNSIGNED NOT NULL, `b2` BIGINT UNSIGNED NOT NULL, `pri` DOUBLE NOT NULL) },
      'replace' => q{ replace q (exp, b1, b2, pri) values (?, ?, ?, ?) },
      'delete' => q{ delete from q where exp = ? },
      'deleteIndex' => 1,
        #!!This table is far too large to 'fetch'
        #!!'fetch' => q{ select exp, b1, b2, pri from q },
      'fetchByPriority' => q{ select q.exp, q.b1, q.b2, q.pri from q left join (sprp) on (sprp.exp = q.exp) where sprp.needed = 0 and q.pri > ? and not exists (select * from reservations where reservations.exp = q.exp) and not exists (select * from done where done.exp = q.exp) and not exists (select * from S where S.exp = q.exp) order by q.pri asc limit ? },
      'fetchByPriorityPatterns' => [ MersInfo::patterns('float'), MersInfo::patterns('bigint'), ],
      'next' => q{ select q.exp from q left join (sprp) on (sprp.exp = q.exp) where sprp.needed = 0 and not exists (select * from reservations where reservations.exp = q.exp) and not exists (select * from done where done.exp = q.exp) and not exists (select * from S where S.exp = q.exp) order by q.pri asc limit 1 },
      'query' => q{ select exp, b1, b2, pri from q where exp = ? },
      'queryIndex' => 1,
      'keyIndex' => 1,
      'valueIndex' => 3,
      'value' => { },
      1 => 1, # exponent
      2 => 3, # stage 1 bound
      3 => 4, # stage 2 bound
      4 => 5, # priority
      'pri' => 4,
    },
    'E' =>
    { # mers letter E has ECM "effort" and largest bound
      'columns' => [ 'exp', 'b1', 'b2', 'pri', ],
      'patterns' => [ MersInfo::patterns('bigint'), MersInfo::patterns('bigint'), MersInfo::patterns('bigint'), MersInfo::patterns('float'), ],
      'create' => q{ create table if not exists `E` (`exp` BIGINT UNSIGNED PRIMARY KEY UNIQUE, `effort` BIGINT UNSIGNED NOT NULL, `bound` BIGINT UNSIGNED NOT NULL, `pri` DOUBLE NOT NULL) },
      'replace' => q{ replace E (exp, effort, bound, pri) values (?, ?, ?, ?) },
      'delete' => q{ delete from E where exp = ? },
      'deleteIndex' => 1,
        #!!This table is far too large to 'fetch'
        #!!'fetch' => q{ select exp, effort, bound, pri from E },
      'fetchByPriority' => q{ select E.exp, E.effort, E.bound, E.pri from E left join (sprp) on (sprp.exp = E.exp) where sprp.needed = 0 and E.pri > ? and not exists (select * from reservations where reservations.exp = E.exp) and not exists (select * from done where done.exp = E.exp) and not exists (select * from S where S.exp = E.exp) order by E.pri asc limit ? },
      'fetchByPriorityPatterns' => [ MersInfo::patterns('float'), MersInfo::patterns('bigint'), ],
      'next' => q{ select E.exp from E left join (sprp) on (sprp.exp = E.exp) where sprp.needed = 0 and not exists (select * from reservations where reservations.exp = E.exp) and not exists (select * from done where done.exp = E.exp) and not exists (select * from S where S.exp = E.exp) order by E.pri asc limit 1 },
      'query' => q{ select exp, effort, bound, pri from E where exp = ? },
      'queryIndex' => 1,
      'keyIndex' => 1,
      'valueIndex' => 3,
      'value' => { },
      1 => 1, # exponent
      2 => 3, # effort (sum of the bounds of all curves done)
      3 => 4, # largest bound
      4 => 5, # priority
      'pri' => 4,
    },
    'done' =>
    { # the Mersenne number is "done", completely factored: D (all factors proven prime) or P (Mersenne itself is prime)
      'columns' => [ 'exp', 'm', ],
      'patterns' => [ MersInfo::patterns('bigint'), $donePattern, ],
      'create' => q{ create table if not exists `done` (`exp` BIGINT UNSIGNED PRIMARY KEY UNIQUE, `m` enum ('D', 'P') NOT NULL) },
      'replace' => q{ replace done (exp, m) values(?, ?) },
      'fetch' => q{ select exp, m from done },
      'fetchByPriority' => q{ select exp, m from done where m = ? and exp >= ? order by exp limit ? },
      'fetchByPriorityPatterns' => [ $donePattern, MersInfo::patterns('bigint'), MersInfo::patterns('bigint'), ],
      'keyIndex' => 1, # field # in mers format; 1 is the exponent
      'query' => q{ select exp, m from done where exp = ? },
      'queryIndex' => 1,
      'valueIndex' => 2, # field 2 is the mers letter; in this case P or D
      'value' => # 'keyIndex' is the index of the key here; 'valueIndex' is the index of the value here
        { # This list needs to match the 'value's in 'missing', just below, to initialize a new db correctly.
          2 => 'P', # M(2) = 3 is 'P'rime
          3 => 'P', # M(3) = 7 is 'P'rime
          4 => 'D', # M(4) = 15 is completely 'D'one, all the factors are proven to be prime
        },
      1 => 1, # first ? in replace or fetch is mers format field 1, the exponent
      2 => 2, # second ? is mers format field 2, the mers letter, D or P
    },
    'missing' =>
    { # This tracks the smallest exponent of each mers letter with no relevant data in the db
      'columns' => [ 'm', 'exp', ],
      'patterns' => [ q{^([cEoqU])$}, MersInfo::patterns('bigint'), ],
      'create' => q{ create table if not exists `missing` (`m` enum ('c', 'E', 'o', 'q', 'U') PRIMARY KEY UNIQUE, `exp` BIGINT UNSIGNED NOT NULL) },
      'replace' => q{ replace missing (m, exp) values(?, ?) },
      'fetch' => q{ select m, exp from missing },
      'keyIndex' => 1,
      'valueIndex' => 2,
      'value' =>
        { # This maps each mers letter to the smallest exponent with no relevant mers data.
          'c' => 5, 'E' => 5, 'o' => 5, 'q' => 5, 'U' => 5,
        },
      'm' => # This maps each mers letter to the value of `m` to use in this table of the database and in the 'value' hash.
        {    #   E.g., mers letters cdlLmMDSP are all cofactor information (well, DP aren't but they are relevant in this context).
          'c' => 'c', 'E' => 'E', 'e' => 'E', 'g' => 'U', 'G' => 'U', 'H' => 'U', 'I' => 'U', 'J' => 'U',
            'K' => 'U', 'l' => 'c', 'L' => 'c', 'm' => 'c', 'M' => 'c', 'o' => 'o', 'q' => 'q', 'U' => 'U',
            'sprp' => 'c',
          # These apply to all data for an exponent, not just to one letter's data.
          'S' => 'all', 'D' => 'all', 'P' => 'all', 'd' => 'all', # 'd' changed to 'S' in Jan 2019
        },
      1 => 1, # mers letter
      2 => 2, # exponent of 2
    },
    'sprp' =>
    { # This tracks whether the cofactor(s) need SPRP checks using values of 'needed':
      #   0, already done;
      #   1, needed;
      #   2, missing factors - cofactor digit count is higher than clmSLM value;
      #   3, some of the factors might be composite.
      'columns' => [ 'exp', 'needed', ],
      'patterns' => [ MersInfo::patterns('bigint'), MersInfo::patterns('bigint'), ],
      'create' => q{ create table if not exists `sprp` (`exp` BIGINT UNSIGNED PRIMARY KEY UNIQUE, `needed` TINYINT UNSIGNED NOT NULL) },
      'replace' => q{ replace sprp (exp, needed) values (?, ?) },
        #!!this table should never be "fetched" in this sense - too much data, perhaps even if needed != 0 is added
        #!!  and needed != 0 and limit 10 or similar?  and set 'fetched' to 0 or the limit count?
        #!!'fetch' => q{ select exp, needed from sprp },
      'fetchByPriority' => q{ select exp, needed from sprp where needed != 0 and exp > ? and not exists (select * from reservations where reservations.exp = sprp.exp and reservations.kind = 'c') and not exists (select * from done where done.exp = sprp.exp) order by exp limit ? },
      'fetchByPriorityPatterns' => [ MersInfo::patterns('bigint'), MersInfo::patterns('bigint'), ],
      'next' => q{ select sprp.exp from sprp where needed != 0 and not exists (select * from reservations where reservations.exp = sprp.exp and reservations.kind = 'c') and not exists (select * from done where done.exp = sprp.exp) order by sprp.exp limit 1 },
      'query' => q{ select exp, needed from sprp where exp = ? },
      'delete' => q{ delete from sprp where exp = ? }, # an entry with needed > 0 should be replaced with needed = 0 before deletion
      'deleteIndex' => 1,
      'queryIndex' => 1,
      'value' => { },
      'keyIndex' => 1,
      'valueIndex' => 2, # the value here is not actually part of mers format
      1 => 1, # exponent
      2 => 2, # see 'sprp' initial comment
    },
    'reservations' =>
    { # This records which clients (by UUID) are doing tasks for which exponents
      # Note that 'kind' is not restricted by the code to the mers work letters and 'uuid' does not have to be a UUID or hostname.
      #   'uuid's of 'mersServer' and 'runSPRP' are used by those Perl scripts to reserve 'c' (SPRP checking) work, which is not given to clients.
      'columns' => [ 'exp', 'uuid', 'kind', ],
      'patterns' => [ MersInfo::patterns('bigint'), MersInfo::patterns('uuid'), '^([?!c' . "$Mersenne::workLetters" . '])$' , ],
      'create' => q{ create table if not exists `reservations` (`exp` BIGINT UNSIGNED PRIMARY KEY UNIQUE, `uuid` VARCHAR(40) NOT NULL, `kind` VARCHAR(1) NOT NULL) },
      'fetch' => q{ select exp, uuid, kind from reservations },
      'insert' => q{ insert into reservations (`exp`, `uuid`, `kind`) values(?, ?, ?) },
      'delete' => q{ delete from reservations where exp = ? },
      'delete1' => q{ delete from reservations where exp = ? },
      'delete2' => q{ delete from reservations where uuid = ? },
      'exists' => q{ select exp, uuid, kind from reservations where exp = ? and uuid = ? and kind = ? },
      'existsPatterns' => [ MersInfo::patterns('bigint'), MersInfo::patterns('uuid'), '^([?!c' . "$Mersenne::workLetters" . '])$' , ],
      'query' => q{ select exp, uuid, kind from reservations where exp = ? },
      'query1' => q{ select exp, uuid, kind from reservations where exp = ? },
      'query2' => q{ select exp, uuid, kind from reservations where uuid = ? },
      'query3' => q{ select exp, uuid, kind from reservations where kind = ? },
      'count' => q{ select count(*) from reservations where uuid = ? and kind = ? },
      'countPatterns' => [ MersInfo::patterns('uuid'), '^([?!c' . "$Mersenne::workLetters" . '])$' , ],
      'insertIndex' => 1,
      'keyIndex' => 1,
      'deleteIndex' => 1,
      'delete1Index' => 1,
      'delete2Index' => 2,
      'queryIndex' => 1,
      'query1Index' => 1,
      'query2Index' => 2,
      'query3Index' => 3,
      1 => 1, # exponent
      2 => 2, # uuid of client doing some work related to the Mersenne
      3 => 3, # mers letter of the kind of work the client is doing
      'uuid' => 2, # UUID can be in fields of SQL; only needed for a safety check in execute()
    },
    'aliases' =>
    { # This records UUIDs, host names, IP addresses, and aliases
      'columns' => [ 'uuid', 'alias', ],
      'patterns' => [ MersInfo::patterns('uuid'), MersInfo::patterns('uuid'), ],
      'create' => q{ create table if not exists `aliases` (`uuid` VARCHAR(255) NOT NULL PRIMARY KEY UNIQUE, `alias` VARCHAR(40) NOT NULL) },
      'fetch' => q{ select uuid, alias from aliases },
      'replace' => q{ replace aliases (`uuid`, `alias`) values(?, ?) },
      'delete' => q{ delete from aliases where uuid = ? },
      'query' => q{ select uuid, alias from aliases where uuid = ? },
      'keyIndex' => 1,
      'queryIndex' => 1,
      'deleteIndex' => 1,
      'valueIndex' => 2, # the value here is the UUID for an alias, which is often a hostname or IP address
      'value' => { },
      1 => 1, # an alias of that UUID, including the UUID itself
      2 => 2, # UUID of client doing some work related to the Mersenne
      'uuid' => 2,
    },
    'allowed' =>
    { # Lists the mers letters a UUID is allowed to work on
      'columns' => [ 'uuid', 'letters', 'tasks', ],
      'patterns' => [ MersInfo::patterns('uuid'), '^([' . "$Mersenne::workLetters" . ']+)$', MersInfo::patterns('bigint'), ],
      'create' => q{ create table if not exists `allowed` (`uuid` VARCHAR(40) NOT NULL PRIMARY KEY UNIQUE, `letters` VARCHAR(40) NOT NULL, `tasks` INT UNSIGNED NOT NULL) },
      'fetch' => q{ select uuid, letters, tasks from allowed },
      'replace' => q{ replace allowed (uuid, letters, tasks) values(?, ?, ?) },
      'delete' => q{ delete from allowed where uuid = ? },
      'query' => q{ select uuid, letters, tasks from allowed where uuid = ? },
      'keyIndex' => 1,
      'queryIndex' => 1,
      'deleteIndex' => 1,
      'valueIndex' => 2, # the value here is the letters of mers format the clients can work on
      'value' => { },
      1 => 1, # uuid of client doing some work related to the Mersenne
      2 => 2, # letters of mers format the clients can work on
      3 => 3, # maximum tasks the UUID may reserve at a time
      'uuid' => 1,
    },
    'contact' =>
    { # tracks time of last connect by a uuid to the server or db
      'columns' => [ 'uuid', 'time', ],
      'patterns' => [ MersInfo::patterns('uuid'), MersInfo::patterns('bigint'), ],
      'create' => q{ create table if not exists `contact` (`uuid` VARCHAR(40) NOT NULL PRIMARY KEY UNIQUE, `time` BIGINT UNSIGNED NOT NULL) },
      'fetch' => q{ select uuid, time from contact },
      'replace' => q{ replace contact (`uuid`, `time`) values(?, ?) },
      'delete' => q{ delete from contact where uuid = ? },
      'query' => q{ select uuid, time from contact where uuid = ? },
      'keyIndex' => 1,
      'queryIndex' => 1,
      'deleteIndex' => 1,
      'valueIndex' => 2, # the value here is a time(), not part of mers format
      'value' => { },
      1 => 1, # uuid of client doing some work related to Mersenne numbers
      2 => 2, # time of the last contact by the uuid
      'uuid' => 1,
    },
  );   

sub new
{
  my $class = shift;
  my (@args) = @_;
  my $self = { };
  my (@files) = ('done', 'primes', 'gaps', 'results', 'bsearch', 'allowed', 'aliases', );
  # This file contains data for "done" Mersennes - D or S - and their factors (C); the cofactor(s) are implied by the D or S.
  ${$self}{'file'}{'done'} = 'M/factors'
    if -f 'M/factors' && -s 'M/factors';
  # This file contains the Mersenne primes (P in mers format).
  ${$self}{'file'}{'primes'} = 'M/primesP'
    if -f 'M/primesP' && -s 'M/primesP';
  # This file contains the trial factoring gaps and joins (G and J).
  ${$self}{'file'}{'gaps'} = 'results.gaps'
    if -f 'results.gaps' && -s 'results.gaps';
  # This directory contains all the Mersenne data not in the prior three files.  The files in the directory are created using
  #   the split program so they are named 'aaa', 'aab', ... 'baa', 'bab', etc., though perhaps with a different number of letters.
  #   The bsearch (Perl) script used to find data in them sorts the file names so it doesn't matter what the file names are as
  #     long as the sort of the individual lines in the files is preserved if 'cat *' is run in the directory.
  ${$self}{'file'}{'results'} = 'results'
    if -d 'results';
  # This file (or directory) contains the data for Mersennes with prime exponents not in the first three files.
  ${$self}{'file'}{'results'} = 'results.primes'
    if !-d('results') && (-s('results.primes') || -d('results.primes'));
  # This (Perl) script implements binary search over sorted flat files, including files created by split saved in a single directory.
  #   It is only used if there are flat (text) files with additional data in the 'resultsDir' or there are 'sortedFiles'.
  #     Since creating and updating those flat files needs other files not in this (mers) package, bsearch is not in the mers package.
  ${$self}{'file'}{'bsearch'} = '/home/wedgingt/bin/bsearch'
    if -x('/home/wedgingt/bin/bsearch') && ! -l('/home/wedgingt/bin/bsearch');
  # This file contains the UUIDs of clients that are allowed to ask mersServer.pl for tasks.
  ${$self}{'file'}{'allowed'} = '.server.allowed'
    if -s('.server.allowed');
  # This file contains aliases of clients (UUIDs, hostnames, and IP addresses).
  ${$self}{'file'}{'aliases'} = '.server.aliases'
    if -s('.server.aliases');
  my (@sortedFiles);
  local $_;
  while (@args)
    {
      my ($key) = shift(@args);
      die("MerDB::new given undef as a key name by " . join(' ', caller()))
        unless defined $key;
      if (ref($key) eq 'HASH')
        {
          unshift(@args, $_, $key->{$_})
            foreach keys(%{$key});
          next;
        }
      unshift(@args, @{$key}),
        next
        if ref($key) eq 'ARRAY';
      die("MersDB::new given an odd number of parameters by " . join(' ', caller()))
        unless @args;
      foreach ('createdTables', 'dbh', 'error', )
        {
          die("MersDB::new - '$key' cannot be given as it is used internally\n"
              . "  caller " . join(' ', caller()))
            if $key eq $_;
        }
      if ($key eq 'sortedFiles')
        {
          @sortedFiles = (),
            delete(${$self}{'sortedFiles'}),
            shift(@args),
            next
            unless defined $args[0];
          push(@sortedFiles, shift(@args)),
            next
            unless ref($args[0]);
          die("MersDB::new - 'sortedFiles' must be a single file name or an ARRAY or HASH ref of file names\n"
              . "  caller " . join(' ', caller()))
            if ref($args[0]) ne 'ARRAY' && ref($args[0]) ne 'HASH';
          push(@sortedFiles, @{$args[0]}),
            shift(@args),
            next
            if ref($args[0]) eq 'ARRAY';
          push(@sortedFiles, keys(%{$args[0]}));
          shift(@args);
          next;
        }
      ${$self}{$key} = shift(@args);
      delete(${$self}{$key}), # value of undef deletes a default or prior key
        next
        unless defined(${$self}{$key});
      die("MersDB::new given UUID with unexpected characters\n")
        if $key eq 'uuid' && ${$self}{$key} !~ m/$uuidPattern/;
      ${$self}{'uuid'} = $1,
        ${$self}{'uuid'} =~ s,[.]pl$,,
        if $key eq 'uuid';
      next unless exists $dbiAttrs{$key};
      die("MersDB::new - DBI::connect attribute $key given non-numeric value ${$self}{$key}\n")
        if $dbiAttrs{$key} =~ m,^[0-9]+$, && ${$self}{$key} !~ m,^([0-9]+)$,;
      ${$self}{'dbiAttrs'}{$key} = $1,
        delete ${$self}{$key}
        if $dbiAttrs{$key} =~ m,^[0-9]+$,;
    }
  local $_;
  my $tbl;
  foreach $tbl (keys(%tables))
    {
      die "MersDB::new - SQL table '$tbl' has no 'create' entry"
        unless exists $tables{$tbl}{'create'};
      die "MersDB::new - SQL table '$tbl' has no 'columns' entry"
        unless exists $tables{$tbl}{'columns'};
      die "MersDB::new - SQL table '$tbl' has no 'patterns' entry"
        unless exists $tables{$tbl}{'patterns'};
      die "MersDB::new - table '$tbl' key 'value' should be a HASH ref"
        if exists $tables{$tbl}{'value'} && ref($tables{$tbl}{'value'}) ne 'HASH';
      #!!add more safety checks - e.g., all 'index' values should be numbers <= $#{$tables{$tbl}{'columns'}}
      my ($key);
      foreach $key (grep { m/patterns$/i } keys(%{$tables{$tbl}}))
        {
          die "MersDB::new - table '$tbl' key '$key' should be an ARRAY ref"
            unless ref($tables{$tbl}{$key}) eq 'ARRAY';
        }
    }
  foreach (keys(%dbiAttrs))
    {
      ${$self}{'dbiAttrs'}{$_} = $dbiAttrs{$_} unless exists ${$self}{'dbiAttrs'}{$_};
    }
  foreach (@files)
    {
      next unless exists ${$self}{'file'}{$_};
      delete(${$self}{'file'}{$_}), next if ${$self}{'file'}{$_} =~ m,[./][.][./],;
      delete(${$self}{'file'}{$_}), next unless ${$self}{'file'}{$_} =~ m,^([/.A-Za-z0-9]+)$,;
      ${$self}{'file'}{$_} = $1;
      delete(${$self}{'file'}{$_}), next if -l ${$self}{'file'}{$_};
      delete(${$self}{'file'}{$_}), next unless (-f ${$self}{'file'}{$_} || -d ${$self}{'file'}{$_}) && -s ${$self}{'file'}{$_};
    }
  ${$self}{'dbi'} = 'dbi:mysql:mersdevel' if !exists ${$self}{'dbi'} && -f('mysql.devel');
  ${$self}{'dbi'} = 'dbi:mysql:mersdevel' if !exists ${$self}{'dbi'} && $VERSION =~ m,(-dirty|debug),;
  ${$self}{'dbi'} = 'dbi:mysql:mers' if !exists ${$self}{'dbi'};
  if (-s('mysql.db'))
    { # Get the name of the mysql db from this file (mostly for debugging and development; see the makefile).
      my ($dbfh);
      open($dbfh, '<', 'mysql.db')
        or die "could not read 'mysql.db' for name of mysql db to use, $!";
      my ($dbname) = <$dbfh>;
      $dbname =~ m,^(mers[a-zA-Z0-9]+)$,
        or die "unsupported characters in db name in file 'mysql.db' - should be alphanumeric starting with 'mers'\n";
      $dbname = $1;
      ${$self}{'dbi'} = 'dbi:mysql:' . "$dbname";
    }
  if (!exists(${$self}{'dbuser'}))
    {
      my ($user) = $1 if getlogin() =~ m/^([a-zA-Z0-9]+)$/;
      $user = $1 if !defined $user && $ENV{'LOGNAME'} =~ m/^([a-zA-Z0-9]+)$/;
      $user = $1 if !defined $user && getpwuid($<) =~ m/^([a-zA-Z0-9]+)$/;
      $user = $1 if !defined $user && $ENV{'USER'} =~ m/^([a-zA-Z0-9]+)$/;
      defined($user) or die('cannot determine the username to use to connect to the database' . "\n");
    }
  # Safety check - Perl complained 'keyIndex' or 'valueIndex' was undef in code where 'value' existed, so:
  foreach (keys(%tables))
    {
      next unless exists $tables{$_}{'value'};
      die "MersDB.pm logic error: table $_ has 'value' but not 'keyIndex'"
        unless exists $tables{$_}{'keyIndex'};
      die "MersDB.pm logic error: table $_ has 'value' but not 'valueIndex'"
        unless exists $tables{$_}{'valueIndex'};
    }
  # Another safety check
  foreach (keys(%{$tables{'missing'}{'m'}}))
    {
      die "MersDB.pm logic error: 'missing' 'm' '$_' '$tables{'missing'}{'m'}{$_}' is not a key in 'missing' 'value'"
        unless ($tables{'missing'}{'m'}{$_} eq 'all'
                || exists $tables{'missing'}{'value'}{$tables{'missing'}{'m'}{$_}});
    }
  bless $self, $class;
  $self->addSortedFiles(@sortedFiles)
    if @sortedFiles;
  return $self;
}

sub doneFile
{
  my $self = shift;
  return undef unless $self->{'file'}{'done'};
  return $self->{'file'}{'doneFile'};
}

sub primesFile
{
  my $self = shift;
  return undef unless $self->{'file'}{'primes'};
  return $self->{'file'}{'primes'};
}

sub gapsFile
{
  my $self = shift;
  return undef unless $self->{'file'}{'gaps'};
  return $self->{'file'}{'gaps'};
}

sub resultsDir
{
  my $self = shift;
  return undef unless $self->{'file'}{'results'};
  return $self->{'file'}{'results'};
}

sub addSortedFiles
{ # Add the given (non-empty) files to the list given to bsearch.
  #   Soft links and paths containing '..' as a component are ignored.
  my $self = shift;
  foreach (@_)
    { #!!should these safety checks be 'error's ?
      next if m,[./][.][./],;
      next unless m,^([/.A-Za-z0-9]+)$,;
      next unless -f $1 && -s $1 && !-l $1;
      $self->{'sortedFiles'}{"$1"} = 1;
    }
  return(keys(%{$self->{'sortedFiles'}}));
}

sub connect
{
  my $self = shift;
  my $uuid = shift;
  $self->{'error'} = "MersDB::connect - given UUID contains unexpected characters\n",
    return(undef)
    if $uuid && $uuid !~ m/$uuidPattern/;
  $uuid = "$1",
    $uuid =~ s,[.]pl$,,
    if $uuid;
  $self->{'error'} = "MersDB::connect given different UUID than MersDB::new or a prior connect\n",
    return(undef)
    if $self->{'uuid'} && $self->{'uuid'} ne $uuid;
  $self->{'uuid'} = $uuid
    unless $self->{'uuid'};
  $self->{'prepared'} = { };
  $self->{'dbh'} = DBI->connect($self->{'dbi'}, $self->{'dbuser'}, $self->{'dbpassword'},
                                $self->{'dbiAttrs'});
  $self->{'error'} = "MersDB::connect - DBI::connect failed\n",
    delete($self->{'dbh'}),
    return(undef)
    unless $self->{'dbh'};
  $self->{'error'} = "MersDB::createTables failed\n",
    delete($self->{'dbh'}),
    return(undef)
    unless $self->createTables();
  my (@rtn) = $self->queryRows('allowed', $uuid);
  delete($self->{'dbh'}),
    return(undef)
    if !@rtn && exists $self->{'error'};
  if (defined(specialExponent($uuid)))
    {
      $self->{'error'} = "MersDB::connect - UUID $uuid can only be used by $uuid.pl\n",
        delete($self->{'dbh'}),
        return(undef)
        if $0 !~ m,/$uuid[.]pl$, && $0 ne "$uuid" . '.pl';
    }
  else
    { # This probably only matters when clients on a different computer than the database
      #   server are supported.  Most such clients would use sendServer.pl and askServer.pl
      #   to send results and get work from mersServer.pl rather than use MersDB.pm or DBI.pm.
      shift(@rtn)
        while (@rtn && $rtn[0]->{'uuid'} ne $uuid);
      $self->{'error'} = "MersDB::connect - UUID $uuid not in allowed list\n",
        delete($self->{'dbh'}),
        return(undef)
        if $0 !~ m,/$uuid[.]pl$, && $0 ne "$uuid" . '.pl';
    }
  @rtn = $self->update('contact', $self->{'uuid'}, time());
  return($self->{'dbh'})
    if @rtn && $rtn[0];
  delete($self->{'dbh'});
  $self->{'error'} = '' unless $self->{'error'};
  $self->{'error'} .= "updating contact time for UUID '$self->{'uuid'}' failed\n";
  return(undef);
}

sub dbError
{ #!!drop this in favor of Perl's die()/eval() style?
  my ($self) = shift;
  $self->{'error'} .= "  " . $self->{'dbh'}->errstr() . "\n"
    if exists $self->{'dbh'} && $self->{'dbh'} && $self->{'dbh'}->err() && $self->{'error'};
  $self->{'error'} = $self->{'dbh'}->errstr() . "\n"
    if exists $self->{'dbh'} && $self->{'dbh'} && $self->{'dbh'}->err() && !$self->{'error'};
  return($self->{'error'})
    if exists $self->{'error'} && $self->{'error'};
  $self->{'error'} = "call MersDB::createTables() first\n",
    return($self->{'error'})
    unless exists($self->{'createdTables'}) && $self->{'createdTables'} eq $self->{'uuid'};
  return('');
}

sub dbTables()
{
  my ($self) = shift;
  return(sort(grep { !m,^[A-Za-z]$, } keys(%tables)));
}

sub columns()
{
  my ($self) = shift;
  my ($table) = shift;
  return(undef) unless exists $tables{$table};
  return(undef) unless exists $tables{$table}{'columns'};
  return($tables{$table}{'columns'});
}

sub createTables
{
  my ($self) = shift;
  my (@drop) = @_; # list of tables to be dropped first, or 'all'
  delete($self->{'createdTables'});
  $self->{'error'} = "not connected to a database - call MersDB::connect\n",
    return(undef)
    unless exists $self->{'dbh'} && $self->{'dbh'};
  my $dbh = $self->{'dbh'};
  $self->{'error'} = "not connected to a database - call MersDB::connect\n",
    return(undef)
    unless $dbh;
  $self->{'error'} = "not connected to a database - call MersDB::connect\n",
    return(undef)
    unless exists($self->{'uuid'});
  local $_;
  # Never drop tables unless a development db is being used.
  @drop = () unless $self->{'dbi'} =~ m/devel$/;
  @drop = keys(%tables) if @drop && $drop[0] eq 'all';
  foreach (@drop)
    {
      if (!$dbh->do("drop table if exists $_"))
        {
          $self->{'error'} = '' unless $self->{'error'};
          $self->{'error'} = "drop table $_ failed\n";
          $self->{'error'} .= " - " . $self->{'dbh'}->errstr() . "\n"
            if $self->{'dbh'}->errstr();
        }
    }
  foreach (keys(%tables))
    {
      if (!$dbh->do($tables{$_}{'create'}))
        {
          $self->{'error'} = '' unless $self->{'error'};
          $self->{'error'} .= "create table $_ failed\n";
          $self->{'error'} .= " - " . $self->{'dbh'}->errstr() . "\n"
            if $self->{'dbh'}->errstr();
        }
    }
  return(undef)
    if exists $self->{'error'};

  # This must be before the next loop or the calls to fetchValues() will fail.
  $self->{'createdTables'} = $self->{'uuid'};

  foreach (sort(grep { exists($tables{$_}{'value'})
                         && scalar(keys(%{$tables{$_}{'value'}})) } keys(%tables)))
    { # 'X', 'done', and 'missing' need to be initialized if they did not exist.
      #   The initial values have been chosen with that initialization order in mind.
      my ($initValues) = $tables{$_}{'value'};
      $tables{$_}{'value'} = { }; # empty hash ref
      my (@results) = $self->fetchValues($_);
      next
        if @results && $results[$#results];
      # nothing was fetched (or an error) so the table in the db needs to be initialized
      #!!this might be simpler by giving mers format input for those tables
      foreach my $key (sort(keys(%{$initValues})))
        { # The value has to be deleted or ignore() will prevent it from being saved in the db.
          my ($value) = $initValues->{$key};
          if (m/^[$validLetters]$/)
            {
              $self->update($_, $key, $_, $value)
                or die "could not update db with initial data $key $_ $value";
              next;
            }
          if ($_ eq 'done')
            { # 'done' entries are given to update() by the mers letter, not 'done'
              $self->update($value, $key, $value)
                or die "could not update db with initial data 'done' $key $value";
              next;
            }
          if ($_ eq 'missing')
            {
              my (@effects) = $self->update('missing', 'replace', $key, $value, $key);
              @effects && $effects[0]
                or die "could not update db with initial data 'missing' $key $value";
              # This is needed only for 'missing' since only it does not use update().
              $tables{'missing'}{'value'}{$key} = $value;
            }
        }
    }

  # Initialize the 'allowed' and 'aliases' tables if they are empty.
  my (@results) = $self->fetchValues('allowed');
  return(undef)
    unless @results && $results[0];
  @results = $self->keyList('allowed');
  if (!@results)
    {
      my (%defaults);
      #!!$maxValues here should be $maxTasks from mersServer.pl, but they have the same value
      #!!  move one or both to MersInfo.pm?
      $defaults{$_} = [ $Mersenne::workLetters, $maxValues, ]
        foreach ('127.0.0.1', @specialUUIDs);
      if (exists($self->{'file'}{'allowed'})
          && -f($self->{'file'}{'allowed'}) && -s($self->{'file'}{'allowed'}))
        { #!!require a checksum or signature of some kind?  the file to be read-only?
          my ($fh);
          $self->{'error'} = "could not open/read $self->{'file'}{'allowed'}, $!",
            return(undef)
            unless open($fh, '<', $self->{'file'}{'allowed'});
          while (<$fh>)
            { # Max task count, mers letters of allowed tasks, and IP or hostname or UUID.
              chomp;
              next if !$_ || m/^ *#/;
              $self->{'error'} = "$self->{'file'}{'allowed'}:$.: unsupported characters\n",
                return(undef)
                unless m/^([-. 0-9a-zA-Z]+)$/;
              my (@line) = split(/ /, $1);
              $self->{'error'} = "$self->{'file'}{'allowed'}:$.: should have exactly two spaces - '$_'\n",
                return(undef)
                unless $#line == 2;
              $self->{'error'} = "$self->{'file'}{'allowed'}:$.: non-numeric max tasks - '$_'\n",
                return(undef)
                unless $line[0] =~ m/^([0-9]+)$/;
              $line[0] = $1;
              $self->{'error'} = "$self->{'file'}{'allowed'}:$.: unexpected characters for allowed work types - '$_'\n",
                return(undef)
                unless $line[1] =~ m/^([$Mersenne::workLetters]+)$/;
              $line[1] = "$1";
              $self->{'error'} = "$self->{'file'}{'allowed'}:$.: unexpected characters in hostname, UUID, or IP address - '$_'\n",
                return(undef)
                unless $line[2] =~ m/^([-.0-9a-zA-Z]+)$/;
              $line[2] = "$1";
              # The columns in the SQL table are not in the same order as the file content.
              @results = $self->update('allowed', $line[2], $line[1], $line[0]);
              $self->{'error'} = ("$self->{'file'}{'allowed'}:$.: updating 'allowed' in db failed for '$_'\n  results - "
                                  . join(' ', @results) . "\n"),
                return(undef)
                unless @results && $results[0];
              delete($defaults{$line[2]});
            }
          close($fh) or die "could not close .server.allowed, $!\n";
        }
      foreach (keys(%defaults))
        {
          @results = $self->update('allowed', $_, @{$defaults{$_}});
          next
            if @results && $results[0];
          $self->{'error'} = '' unless $self->{'error'};
          $self->{'error'} .= "  while adding uuid '$_' to 'allowed' in MersDB::createTables\n";
          $self->{'error'} .= "  called by " . join(' ', caller());
          return(undef);
        }
    }
  @results = $self->fetchValues('aliases');
  return(undef)
    unless @results && $results[0];
  @results = $self->keyList('aliases');
  if (!@results)
    {
      my (%defaults) = ( '127.0.0.1' => 'localhost', );
      if (exists($self->{'file'}{'aliases'})
          && -f($self->{'file'}{'aliases'}) && -s($self->{'file'}{'aliases'}))
        {
          my ($fh);
          $self->{'error'} = "could not open/read $self->{'file'}{'aliases'}, $!",
            return(undef)
            unless open($fh, '<', $self->{'file'}{'aliases'});
          while (<$fh>)
            { # IP address ('alias') and hostname or UUID.
              #   A particular IP address will often have at least one UUID and one or more hostnames.
              chomp;
              next unless $_ && !m/^ *#/;
              my (@line) = split(/ /);
              $self->{'error'} = "$self->{'file'}{'aliases'}:$.: should have exactly one space - '$_'\n",
                return(undef)
                unless $#line == 1;
              $self->{'error'} = "$self->{'file'}{'aliases'}:$.: invalid IP address\n",
                return(undef)
                unless $line[0] =~ m/^([.0-9]+)$/;
              $line[0] = "$1";
              $self->{'error'} = "$self->{'file'}{'aliases'}:$.: unexpected characters in hostname or UUID - '$_'\n",
                return(undef)
                unless $line[1] =~ m/^([-.0-9a-zA-Z]+)$/;
              $line[1] = "$1";
              # The columns in the SQL table are not in the same order as the file content.
              @results = $self->update('aliases', $line[1], $line[0]);
              $self->{'error'} = ("$self->{'file'}{'aliases'}:$.: updating 'aliases' in db failed for\n '$_'\n results: "
                                  . join(' ', @results) . "\n"),
                return(undef)
                unless @results && $results[0];
              delete($defaults{$line[1]});
            }
          close($fh) or die "could not close .server.aliases, $!\n";
        }
      foreach (keys(%defaults))
        {
          @results = $self->update('aliases', $_, $defaults{$_});
          next if @results && $results[0];
          $self->{'error'} = '' unless $self->{'error'};
          $self->{'error'} .= "  while adding alias $_ for $defaults{$_}\n";
          return(undef);
        }
    }
  return($self->commit())
}

sub commit
{
  my $self = shift;
  $self->{'error'} = "not connected to a database - call MersDB::connect\n",
    return(undef)
    unless exists($self->{'createdTables'}) && $self->{'createdTables'} eq $self->{'uuid'};
  $self->{'error'} = "not connected to a database - call MersDB::connect\n",
    return(undef)
    unless exists $self->{'dbh'} && $self->{'dbh'};
  return(1)
    if $self->{'dbh'}->commit();
  $self->{'error'} = "database commit failed\n";
  return(undef);
}

sub disconnect
{
  my $self = shift;
  # If no db connected, success.
  return(1)
    unless exists $self->{'dbh'} && $self->{'dbh'};
  $self->{'prepared'} = {};
  my $result = $self->{'dbh'}->disconnect();
  delete $self->{'dbh'};
  delete $self->{'error'};
  delete $self->{'createdTables'};
  delete $self->{'uuid'};
  return($result);
}

sub execute
{
  my $self = shift;
  my $letter = shift;
  my $query = shift;
  my @fields = @_;
  my (@result) = (undef, 0);
  $self->{'error'} = "not connected to a database - call MersDB::connect\n",
    return(@result)
    unless exists($self->{'createdTables'}) && $self->{'createdTables'} eq $self->{'uuid'};
  $self->{'error'} = "not connected to a database - call MersDB::connect\n",
    return(@result)
    unless exists $self->{'dbh'} && $self->{'dbh'};
  $self->{'error'} = "not connected to a database - call MersDB::connect\n",
    return(@result)
    unless exists($tables{$letter});
  $self->{'error'} = 'SQL (database) table ' . "'$letter' has no query '$query'\n",
    return(@result)
    unless exists($tables{$letter}{$query});
  my (@values);
  my ($index) = 0;
  if ($query =~ m/^(delete|query)([0-9]*)$/)
    {
      $query = $1 . $2;
      $self->{'error'} = 'MersDB::execute table ' . "$letter $query needs one value but none (or several) were given\n",
        return(@result)
        unless $#fields == 0;
      $index = $tables{$letter}{$query . 'Index'} - 1;
      $self->{'error'} = 'MersDB::execute table ' . "$letter $query - value '$fields[0]' has unexpected characters\n",
        return(@result)
        unless $fields[0] =~ m/$tables{$letter}{'patterns'}[$index]/;
      push(@values, "$1");
    }
  elsif ($query =~ m/^(count)([0-9]*)$/)
    {
      $query = $1 . $2;
      $self->{'error'} = 'MersDB::execute - table ' . "$letter does not have query $query\n",
        return(@result)
        unless exists $tables{$letter}{$query};
      $self->{'error'} = ('MersDB::execute - table ' . "$letter $query needs "
                          . ($#{$tables{$letter}{$query . 'Patterns'}})
                          . " values but " . ($#fields + 1) . " were given\n"),
        return(@result)
        unless $#fields == $#{$tables{$letter}{$query . 'Patterns'}};
      $index = -1;
      while (++$index <= $#fields)
        {
          $self->{'error'} = 'MersDB::execute table ' . "$letter $query - value '$fields[$index]' has unexpected characters\n",
            return(@result)
            unless $fields[$index] =~ m/$tables{$letter}{$query . 'Patterns'}[$index]/;
          push(@values, "$1");
        }
    }
  elsif ($query eq 'fetchByPriority' || $query eq 'exists' || $query eq 'count')
    {
      my (@patterns);
      push(@patterns, @{$tables{$letter}{$query . 'Patterns'}})
        if exists($tables{$letter}{$query . 'Patterns'});
      my ($which) = -1;
      $self->{'error'} = ("MersDB::execute table '$letter' query '$query' has " . ($#patterns + 1)
                          . " fields but " . ($#fields + 1) . "arguments given\n"),
        return(@result)
        unless $#fields == $#patterns;
      while (++$which <= $#fields)
        {
          $self->{'error'} = "MersDB::execute table $letter - value '$fields[$which]' has unexpected characters\n",
            return(@result)
            unless $fields[$which] =~ m,$patterns[$which],;
          push(@values, $1);
        }
    }
  else
    {
      while (++$index && exists($tables{$letter}{$index}) && $tables{$letter}{$index} - 1 <= $#fields)
        { # This can happen for oq data read via bsearch with no stage 2 bound, e.g.
          $fields[$tables{$letter}{$index} - 1] = 0
            unless defined($fields[$tables{$letter}{$index} - 1]);
          # Got a taint failure in the call to execute just below for a number for no clear reason, so ...
          $self->{'error'} = 'MersDB::execute table ' . "$letter $query index $index unexpected data\n",
            return(@result)
            unless $fields[$tables{$letter}{$index} - 1] =~ m/$tables{$letter}{'patterns'}[$index - 1]/;
          push(@values, "$1");
        }
      # This next should only happen for oq stage 2 bounds and (missing) priorities but is also OK for E
      push(@values, '0'),
        ++$index
        while exists($tables{$letter}{$index}) && $index > 2;
    }
  # This should be the only call of DBI::prepare*() in MersDB.pm.
  $self->{'prepared'}{$letter}{$query} = $self->{'dbh'}->prepare($tables{$letter}{$query}, undef)
    unless exists $self->{'prepared'}{$letter}{$query} && $self->{'prepared'}{$letter}{$query};
  my ($sth) = $self->{'prepared'}{$letter}{$query};
  $self->{'error'} = 'MersDB::execute - preparing SQL for table ' . "$letter $query failed\n",
    return(@result)
    unless $sth;
  # This should be the only call of DBI::execute() in MersDB.pm.
  my ($ok) = $sth->execute(@values);
  $executions{"$sth"} = [ $ok, $letter, $query, @fields, @values, ];
  print("DBI::execute called - '" . join(' ', @{$executions{"$sth"}}) . "'\n"
        . "sth $sth active flag is " . ($sth->{'Active'} ? "$sth->{'Active'}" : '<false>') . "\n")
    if 0;
  delete($executions{"$sth"})
    unless $sth && $sth->{'Active'};
  @result = ($ok, $sth),
    return(@result)
    if $ok;
  $self->{'error'} = 'DBI::execute failed -' . $sth->{'Statement'};
  $self->{'error'} .= 'with ' . join(' ', @values)
    if @values;
  $self->{'error'} .= "\n";
  return(@result)
}

sub lowestMissingExponent
{
  my $self = shift;
  my $letter = shift;
  $self->{'error'} = "not connected to a database - call MersDB::connect\n",
    return(undef)
    unless exists($self->{'createdTables'}) && $self->{'createdTables'} eq $self->{'uuid'};
  $self->{'error'} = "table 'missing' does not exist\n",
    return(undef)
    unless exists $tables{'missing'};
  $self->{'error'} = "table 'missing' does not have 'm' entry\n",
    return(undef)
    unless exists $tables{'missing'}{'m'};
  $self->{'error'} = "table 'missing' 'm' does not have 'value' entry\n",
    return(undef)
    unless exists $tables{'missing'}{'value'};
  $letter = 'c'
    if $letter =~ m,^([lmLMSDPd]|all)$,; # special case - smallest Mersenne with no cofactor SPRP check
  $self->{'error'} = "table 'missing' 'm' does not have '$letter' entry\n",
    return(undef)
    unless exists $tables{'missing'}{'m'}{$letter};
  my ($m) = $tables{'missing'}{'m'}{$letter};
  $self->{'error'} = "table 'missing' 'm' 'value' does not have '$m' entry\n",
    return(undef)
    unless exists $tables{'missing'}{'value'}{$m};
  return($tables{'missing'}{'value'}{$m});
}

our (%validDoneTransitions) =
  ( # allowed changes including no-ops (value of 1, currently not needed)
   'SD' => 2, 'dS' => 2, 'SS' => 1, 'DD' => 1,
  );

sub markDone
{
  my $self = shift;
  my $newlyDone = shift;
  my (@results);
  $self->{'error'} = "call MersDB::createTables first\n",
    return(@results)
    unless exists($self->{'createdTables'}) && $self->{'createdTables'} eq $self->{'uuid'};
  $self->{'error'} = "MersDB::markDone given unexpected argument\n",
    return(@results)
    if !defined($newlyDone) || ref($newlyDone) ne 'HASH';
  local $_;
  my ($X);
  # Need the (largest) X exponent with value 0 in case the next exponent is now 'done'.
  $X = $_
    foreach (sort { $a <=> $b } grep { $tables{'X'}{'value'}{$_} == 0 } keys(%{$tables{'X'}{'value'}}));
  my ($exp);
  my (@changes);
  foreach $exp (sort { $a <=> $b } keys(%{$newlyDone}))
    {
      $self->{'error'} = "MersDB::markDone unexpected mers letter $newlyDone->{$exp}\n",
        @results = (),
        return(@results)
        unless $newlyDone->{$exp} =~ m/^$donePattern$/;
      $newlyDone->{$exp} = $1;
      if ($exp != $X + 1)
        {
          push(@results, $exp, 'unchanged'),
            next
            if (exists($tables{'done'}{'value'}{$exp}) && exists($newlyDone->{$exp})
                && $tables{'done'}{'value'}{$exp} eq $newlyDone->{$exp});
          $self->{'error'} = "MersDB::markDone M$exp already $tables{'done'}{'value'}{$exp} ignoring $newlyDone->{$exp}\n",
            @results = (),
            return(@results)
            if (exists($tables{'done'}{'value'}{$exp}) && exists($newlyDone->{$exp})
                && !exists($validDoneTransitions{$tables{'done'}{'value'}{$exp}
                                                 . $newlyDone->{$exp}}));
        }
      ($tables{'done'}{'value'}{$exp} = $newlyDone->{$exp})
        if $newlyDone->{$exp} =~ m,$donePattern,;
      push(@changes, $exp);
    }
  return(@results)
    unless $#changes >= 0;
  while (@changes)
    {
      $exp = shift @changes;
      if ($newlyDone->{$exp} =~ m/$donePattern/)
        {
          my (@effects) = $self->execute('done', 'replace', $exp, $tables{'done'}{'value'}{$exp});
          $self->{'error'} .= "MersDB::markDone M$exp replace $tables{'done'}{'value'}{$exp} failed\n",
            @results = (),
            return(@results)
            unless $effects[0];
          push(@results, $effects[0], 'done', 'replace', $exp, $tables{'done'}{'value'}{$exp});
        }
      my ($m) = $tables{'missing'}{'m'}{$newlyDone->{$exp}};
      $self->{'error'} .= "MersDB::markDone logic error - M$exp $newlyDone->{$exp} is not 'all' for 'missing' exponents\n",
        @results = (),
        return(@results)
        unless $m eq 'all';
      my (@list) = grep { m,^[$validLetters]$, && exists($tables{$_}{'value'}) } keys(%tables);
      foreach('sprp', @list)
        {
          next if $_ eq 'X'; 
          next if m,^(S|sprp)$, && $newlyDone->{$exp} eq 'S' && $exp % 8 != 4;
          next if m,^([LM]|sprp)$, && $newlyDone->{$exp} eq 'S' && $exp % 8 == 4;
          delete($tables{$_}{'value'}{$exp});
          my (@effects) = $self->execute($_, 'delete', $exp);
          $self->{'error'} .= "MersDB::markDone M$exp $_ delete failed\n",
            @results = (),
            return(@results)
            unless $effects[0];
          push(@results, $effects[0], $_, 'delete', $exp);
        }
      push(@list, 'U')
        unless exists $tables{'U'} && exists $tables{'U'}{'value'};
      foreach (@list)
        {
          delete($tables{$_}{'value'}{$exp})
            if (exists $tables{$_} && exists $tables{$_}{'value'} && exists $tables{$_}{'value'}{$exp});
          next unless exists $tables{'missing'}{'value'}{$_} && $exp == $tables{'missing'}{'value'}{$_};
          my (@aLME) = $self->advanceLME($exp, $_);
          push(@results, @aLME);
          unshift(@results, 0),
            return(@results)
            unless @aLME && $aLME[0];
        }
      next unless $m eq 'all' && $newlyDone->{$exp} =~ m,$donePattern, && defined $X && $exp == $X + 1;
      my (@effects) = $self->execute('X', 'delete', $X);
      $self->{'error'} .= "MersDB::markDone M$exp - M$X X delete failed\n",
        @results = (),
        return(@results)
        unless @effects && $effects[0];
      push(@results, $effects[0], 'delete', $X);
      ++$X;
      ++$X
        while (defined($self->value('done', $X))
               && exists($tables{'missing'}{'m'}{$self->value('done', $X)})
               && $tables{'missing'}{'m'}{$self->value('done', $X)} eq 'all');
      --$X; # now the largest exponent (in a row) that is 'done', not the smallest that isn't
      @effects = $self->update('X', $X, 'X', 0);
      $self->{'error'} .= "MersDB::markDone M$exp - M$X X update failed\n",
        @results = (),
        return(@results)
        unless @effects && $effects[0];
      push(@results, $effects[0], 'replace', $X, 'X', 0);
      delete($tables{'X'}{'value'}{$_})
        foreach (grep { $tables{'X'}{'value'}{$_} == 0 } keys(%{$tables{'X'}{'value'}}));
      $tables{'X'}{'value'}{$X} = 0;
    }
  $self->commit()
    or return(undef);
  return(@results);
}

sub ignore
{ # Compare the new data with what is in the db and return 1 to not store it if appropriate.
  #   Data not in %tables is not be checked - 0 will be returned in such cases.
  my $self = shift;
  my (@fields) = @_;
  $self->{'error'} = "call MersDB::createTables first\n",
    return(undef)
    unless exists($self->{'createdTables'}) && $self->{'createdTables'} eq $self->{'uuid'};
  shift(@fields)
    if $fields[0] eq 'M(';
  #!!the 1 and 2 subscripts to @fields here presume mers format
  return(1) unless exists $tables{$fields[1]};
  return(1) unless exists $tables{$fields[1]}{'value'};
  # No value after the mers letter is erroneous except for mers letters DPS, which should not get here.
  return(1) if $#fields < 2;
  if (!exists($tables{$fields[1]}{'value'}{$fields[0]})
      || !defined($tables{$fields[1]}{'value'}{$fields[0]}))
    { # only X data affects other data with a different exponent
      # trial factoring extent has to be at least 2*n + 1 where n is the exponent
      return(1) if $fields[1] =~ m,^[HU]$, && $fields[2] <= 2*$fields[0] + 1;
      # cofactor digits cannot be larger than the exponent (or even the same, but the same is used to create a new entry)
      return(1) if $fields[1] =~ m/^[clmLMS]$/ && $fields[0] < $fields[2];
      return(0) unless $fields[1] eq 'X';
      foreach my $entry (sort { $a <=> $b } keys(%{$tables{'X'}{'value'}}))
        {
          return(1) # New values that are smaller are "covered" by old data ...
            if (($fields[0] <= $entry && $fields[2] <= $tables{'X'}{'value'}{$entry})
                # ... except _values_ of 0 are only "covered" by larger exponents that also have a 0 value.
                && ($fields[2] > 0 || $tables{'X'}{'value'}{$entry} == 0));
        }
      return(0);
    }
  my ($cmp) = ($fields[2] <=> $tables{$fields[1]}{'value'}{$fields[0]});
  # ignore if new effort, stage 1 bound, or factoring extent is smaller, or is equal with no (max ECM or stage 2) bound
  if ($fields[1] =~ m/^[EoqUX]$/)
    {
      return(0) if $cmp > 0;
      return(1) if $cmp < 0 || $#fields == 2;
      return(0);
    }
  # Ignore if new digits in cofactor is not smaller _and_ 'sprp' is not 2
  #   (which indicates an incorrect too-small digit count).
  return(1) if $fields[1] =~ m/^[clmLMS]$/ && $cmp >= 0 && $self->value('sprp', $fields[0]) != 2;
  return(0);
}

# This splits the argument BigInt into 'limbs'; it is a utility function, not a member function of MersDB.
sub limbs($)
{
  my ($num) = shift;
  my (@limbs);
  $num = Math::BigInt->new("$num")
    unless ref($num);
  push(@limbs, $num->copy());
  my ($limb) = 0;
  while ($limbs[$limb] > $SQLmaxInt)
    {
      push(@limbs, $limbs[$limb]->copy()->brsft($SQLmaxBits));
      $limbs[$limb] %= Math::BigInt->bone()->blsft($SQLmaxBits);
      ++$limb;
    }
  return(@limbs);
}

sub update
{ # Update %tables and, if connected, the db
  my $self = shift;
  my $letter = shift; # actually the SQL table (key in %tables) in most cases
  my (@results);
  $self->{'error'} = "MersDB::update - db tables not created or UUID mismatch\n",
    return(@results)
    unless exists($self->{'createdTables'}) && $self->{'createdTables'} eq $self->{'uuid'};
  my $missing;
  $missing = $letter,
    $letter = shift
    if $letter eq 'missing'; # update lowest exponent missing data, nothing else
  my (@fields) = @_; # remaining arguments, which are often the fields of the mers format
  pop(@fields) while !defined($fields[$#fields]); # shouldn't happen
  shift(@fields) if $fields[0] eq 'M(';
  local $_;
  #!!if not connected, track the changes that need to be commited? or just push everything after a reconnect?
  if (exists($self->{'dbh'}) && $self->{'dbh'} && !$missing)
    {
      if ($letter =~ m,^(contact)|(allowed)|(aliases)$,)
        { # always replace (never insert)
          $self->{'error'} = "MersDB::update table $letter replace SQL is missing\n",
            return(@results)
            unless exists $tables{$letter}{'replace'}; #!!should also check 'maxIndex' against $#fields
          #!!only "local" UUIDs should be allowed to update these
          my ($ok, $sth) = $self->execute($letter, 'replace', @fields);
          $sth->finish()
            if $sth && $sth->{'Active'};
          $self->commit()
            or return(@results);
          push(@results, $ok, $letter, @fields),
            if $ok;
          return(@results);
        }
      elsif ($letter eq 'reserve')
        { # insert (never replace); no 'value' hash
          $self->{'error'} = "MersDB::update table reservations insert SQL is missing\n",
            return(@results)
            unless exists $tables{'reservations'}{'insert'}; #!!should also check 'maxIndex' against $#fields
          my ($ok, $sth) = $self->execute('reservations', 'insert', @fields);
          $self->commit()
            or return(@results);
          push(@results, $ok, $letter, @fields),
            if $ok;
          return(@results);
        }
      elsif ($letter eq 'release')
        { # 'delete'/release/free a reservation; no 'value' hash
          $self->{'error'} = "MersDB::update table reservations delete problem\n",
            return(@results)
            unless exists($tables{'reservations'}{'delete'}) && exists($tables{'reservations'}{'deleteIndex'});
          my ($ok, $sth) = $self->execute('reservations', 'delete', @fields);
          $self->commit()
            or return(@results);
          push(@results, $ok, $letter, @fields),
            if $ok;
          return(@results);
        }
      elsif ($letter eq 'C')
        { # A factor - there can be more than one per exponent and they are often more than 64 bits.
          # The caller provides the ord(inal) after the factor; the factor is split into 'limbs' here.
          my (@limbs) = limbs($fields[2]);
          foreach my $limb (0 .. $#limbs)
            {
              my ($ok, $sth) = $self->execute('C', 'insert', $fields[0], 'C', $fields[3], $limb, $limbs[$limb]);
              $self->{'error'} = "MersDB::update could not save M$fields[0] C $fields[3] $limb\n",
                @results = (),
                return(@results)
                unless $ok;
              push(@results, $ok, 'C', 'insert', $fields[0], 'C', $fields[3], $limb, $limbs[$limb]);
            }
        }
      elsif ($letter =~ m/^$donePattern$/)
        {
          push(@results, $self->markDone( { $fields[0] => $1, } ));
          $self->{'error'} .= "MersDB::update - markDone $fields[0] $1 failed\n",
            @results = (),
            return(@results)
            unless @results;
        }
      elsif (exists($tables{$letter}) && exists($tables{$letter}{'value'}))
        {
          return('no change')
            if $letter =~ m/^[$validLetters]$/ && $self->ignore(@fields);
          if (exists($tables{$letter}{'value'}))
            { # No table should have 'value' without 'keyIndex' and 'valueIndex' but Perl complained about undef as an array index here.
              if (!exists($tables{$letter}{'keyIndex'}) || !exists($tables{$letter}{'valueIndex'}))
                { # This has not happened for months as of 30 Nov 2021, so simplify it.
                  warn("table $letter has 'value' but not 'keyIndex' or not 'valueIndex'; deleting 'value'\n"
                       . "  for fields " . join(' ', @fields));
                  delete $tables{$letter}{'value'};
                }
              else
                {
                  if ($letter =~ m,^[Eoq]$,)
                    { # Calculate the priority.
                      push(@fields, 0)
                        while $#fields < 3;
                      my ($pri) = Mersenne::mersPriority(@fields);
                      return(@results)
                        unless $pri >= 0.0;
                      $fields[4] = $pri;
                    }
                  elsif ($letter eq 'X')
                    { # Any now "covered" data has to be deleted from the db first.
                      my ($X) = $fields[$tables{$letter}{'keyIndex'} - 1];
                      my ($extent) = $fields[$tables{$letter}{'valueIndex'} - 1];
                      foreach (keys(%{$tables{'X'}{'value'}}))
                        { # M(x)X: 0 is an exception - it can only be "covered" by another 0 value with a larger exponent.
                          next if ($_ > $X || $tables{'X'}{'value'}{$_} > $extent
                                   || ($tables{'X'}{'value'}{$_} < 1 && $extent > 0));
                          my ($ok, $sth) = $self->execute('X', 'delete', $_);
                          $self->{'error'} .= "MersDB::update - delete M$_ X failed\n",
                            @results = (),
                            return(@results)
                            unless $ok;
                          push(@results, $ok, 'X', 'delete', $_);
                          delete($tables{'X'}{'value'}{$_});
                        }
                    }
                  $tables{$letter}{'value'}{$fields[$tables{$letter}{'keyIndex'} - 1]} = $fields[$tables{$letter}{'valueIndex'} - 1];
                  my ($ok, $sth) = $self->execute($letter, 'replace', @fields);
                  $self->{'error'} .= "MersDB::update - $letter replace " . join(' ', @fields) . " failed\n",
                    @results = (),
                    return(@results)
                    unless $ok;
                  push(@results, $ok, $letter, 'replace', @fields);
                  if ($letter =~ m/^[clmLMS]$/ && !$self->value('sprp', $fields[0]))
                    { # "new" SPRP result - presume it needs to be verified against the known factors
                      ($ok, $sth) = $self->execute('sprp', 'replace', $fields[0], 1);
                      $self->{'error'} .= "MersDB::update - replace M$_ sprp 1 failed\n",
                        @results = (),
                        return(@results)
                        unless $ok;
                      push(@results, $ok, 'sprp', 'replace', $fields[0], 1);
                    }
                }
            }
        }
    }
  my (%changes);
  if (exists($tables{'missing'}{'m'}{$letter}))
    {
      my (@m) = ($tables{'missing'}{'m'}{$letter}, );
      @m = keys(%{$tables{'missing'}{'value'}})
        if $m[0] eq 'all';
      foreach my $m (@m)
        {
          if (!$missing && $fields[0] == $tables{'missing'}{'value'}{$m})
            {
              ++$tables{'missing'}{'value'}{$m};
              $tables{'missing'}{'value'}{'U'} = Math::Prime::Util::next_prime($tables{'missing'}{'value'}{'U'})
                if $m eq 'U';
              $changes{$m} = 1;
            }
          $tables{'missing'}{'value'}{$m} = $fields[0],
            $changes{$m} = 1
            if $missing;
          while (exists($tables{'done'}{'value'}{$tables{'missing'}{'value'}{$m}})
                 || (exists($tables{$m}{'value'}) && exists($tables{$m}{'value'}{$tables{'missing'}{'value'}{$m}})))
            {
              ++$tables{'missing'}{'value'}{$m};
              $tables{'missing'}{'value'}{'U'} = Math::Prime::Util::next_prime($tables{'missing'}{'value'}{'U'})
                if $m eq 'U';
              $changes{$m} = 1;
            }
        }
    }
  elsif ($letter eq 'X' && $fields[2] == 0)
    { # all Mersennes with exponents thru $fields[0] are completely factored; rarely changes
      foreach (keys(%{$tables{'missing'}{'value'}}))
        {
          $tables{'missing'}{'value'}{$_} = $fields[0] + 1,
            $changes{$_} = 1
            if $tables{'missing'}{'value'}{$_} <= $fields[0];
        }
    }
  foreach my $m (keys(%changes))
    {
      next unless exists($tables{'missing'}{'value'}{$m});
      my (@effects) = $self->execute('missing', 'replace', $m, $tables{'missing'}{'value'}{$m});
      $self->{'error'} .= "MersDB::update - replace M$tables{'missing'}{'value'}{$m} missing $m $tables{'missing'}{'value'} failed\n",
        @results = (),
        return(@results)
        unless @effects && $effects[0];
      push(@results, $effects[0], 'missing', 'replace', $m, $tables{'missing'}{'value'}{$m});
    }
  push(@results, '0E0')
    unless @results;
  return(@results)
    if $self->commit();
  @results = ();
  return(@results);
}

sub advanceLME($$)
{
  my ($self) = shift;
  my ($exp) = shift;
  my ($m) = shift;
  local $_;
  my (@results);
  $self->{'error'} = "MersDB::advanceLME - db tables not created or UUID mismatch\n",
    return(@results)
    unless exists($self->{'createdTables'}) && $self->{'createdTables'} eq $self->{'uuid'};
  $self->{'error'} = "MersDB::advanceLME - exponent $exp is not an integer\n",
    return(@results)
    unless $exp =~ m,^([1-9][0-9]*)$,;
  $exp = $1;
  $self->{'error'} = "MersDB::advanceLME - invalid mers letter $m\n",
    return(@results)
    unless exists $tables{'missing'}{'m'}{$m};
  $m = $tables{'missing'}{'m'}{$m};
  $self->{'error'} = "MersDB::advanceLME - exponent $exp is not the current lowest missing $m exponent\n",
    return(@results)
    unless $exp == $tables{'missing'}{'value'}{$m};
  ++$tables{'missing'}{'value'}{$m};
  # 'U' only makes sense for prime exponents
  $tables{'missing'}{'value'}{'U'} = Math::Prime::Util::next_prime($tables{'missing'}{'value'}{'U'})
    if $m eq 'U';
  my (@effects) = $self->execute('missing', 'replace', $m, $tables{'missing'}{'value'}{$m});
  $self->{'error'} .= "MersDB::advanceLME M$exp missing replace failed\n",
    @results = (),
    return(@results)
    unless $effects[0];
  push(@results, $effects[0], 'missing', 'replace', $m, $tables{'missing'}{'value'}{$m});
  return(@results)
    if $self->commit();
  @results = ();
  return(@results);
}

sub save
{
  my ($self) = shift;
  my ($m) = shift;
  my (@results);
  local $_;
  $self->{'error'} = "MersDB::save - db tables not created or UUID mismatch\n",
    return(@results)
    unless exists($self->{'createdTables'}) && $self->{'createdTables'} eq $self->{'uuid'};
  $self->{'error'} = "MersDB::save given non-Mersenne argument\n",
    return(@results)
    unless ref($m) eq 'Mersenne';
  $self->{'error'} = "MersDB::save - lowest missing c exponent is undef\n",
    return(@results)
    unless defined $self->lowestMissingExponent('c');
  $self->{'error'} = "MersDB::save - Mersenne errors:\n" . $m->errors() . "\n",
    return(@results)
    unless $m->check();
  $self->{'error'} = 'MersDB::save does not support "X" mers format data' . "\n",
    return(@results)
    if scalar($m->get('X'));
  my (@xexps) = $self->keyList('X');
  my (@upRtn);
  my (@f);
  my ($needsSPRP);
 SAVE_FACTORS_AGAIN:
  my ($oldFactors);
  # This would be 'if' except for needing to try again after an SQL insert
  #   of the factors fails due to a duplicate entry.
  while ($m->n() <= $self->lowestMissingExponent('c'))
    { # Only save factors in the db for "small" exponents.
      local $self->{'dbh'}{'RaiseError'}; # catch errors here rather than die()ing
      if (!defined($oldFactors))
        { # This only needs to be done the first time here - $m is not affected by a db failure later.
          $oldFactors = $self->find($m, 'C');
          $self->{'error'} = ("MersDB::save - Mersenne::parse failed for '$oldFactors' - "
                              . $m->errors() . "\n"),
            return(@results)
            unless $m->parse($oldFactors);
        }
      # Factors already in the db have to be wiped as they could be divisible by some in $m->f().
      my ($ok) = $self->remove('C', $m->n());
      $self->{'error'} .= "MersDB::save - could not remove existing factors of M" . $m->n() . "\n",
        return(@results)
        unless $ok;
      push(@results, 'remove', "M" . $m->n(), 'C', $ok);
      @f = $m->f();
      if (defined($self->value('sprp', $m->n())) && $self->value('sprp', $m->n()) == 3)
        { # Some factors might be composite - check and leave 'sprp' == 3 if so
          my ($allPrime) = 1;
          foreach (@f)
            {
              $allPrime = 0,
                last
                unless Math::Prime::Util::is_prime($_);
            }
          $needsSPRP = '3'
            unless $allPrime;
        }
      foreach (0 .. $#f)
        {
          @upRtn = $self->update('C', $m->n(), 'C', $f[$_], $_);
          push(@results, @upRtn),
            next
            if @upRtn && $upRtn[0];
          $self->{'error'} = "Mers::save - update M" . $m->n() . " C #$_ $f[$_] failed\n",
            unshift(@results, 0), # indicate failure
            return(@results)
            unless $self->{'error'} =~ m, Duplicate entry ,;
          warn("saving factors of M" . $m->n() . " in the db hit existing data - " . $self->dbError());
          delete $self->{'error'};
          next SAVE_FACTORS_AGAIN;
        }
      last;
    }
  foreach (grep { $m->get($_) } ('P', 'D'))
    { # These two imply that all the factors are known.
      @upRtn = $self->update($_, $m->n(), $_);
      $self->{'error'} .= "Mers::save - update M" . $m->n() . " $_ failed\n",
        unshift(@results, 0),
        return(@results)
        unless @upRtn;
      push(@results, $self->update($_, $m->n(), $_));
      $needsSPRP = 'done'
        unless $needsSPRP;
    }
  my (%values); # for ScLlMm
  foreach (split(//, 'ScLlMmEoqUHGgJ'))
    {
      next if m,^[LM]$,i && $m->n() % 8 != 4;
      next if m,^[Sc]$, && $m->n() % 8 == 4;
      if ($needsSPRP && $needsSPRP eq 'done')
        { # 'done' (DP) indicates that the Mersenne is completely factored - other data (except X) can be ignored.
          my (@rmRtn) = $self->remove($_, $m->n());
          $self->{'error'} = 'MersDB::save - remove ' . $m->n() . " $_ failed - " . join(' ', @rmRtn) . "\n",
            unshift(@results, 0),
            return(@results)
            unless @upRtn && $upRtn[0];
          push(@results, @upRtn);
          next
            unless (exists($tables{'missing'}{'m'}{$_})
                    && $m->n() == $self->lowestMissingExponent($_));
          my ($ok) = $self->advanceLME($m->n(), $tables{'missing'}{'m'}{$_});
          unshift(@results, 0, 'advancing lme failed'),
            return(@results)
            unless $ok;
          push(@results, 'lme', $ok);
          next;
        }
      my (@values) = $m->get($_);
      if (!@values)
        {
          my ($mers) = $self->find($m, $_);
          next unless $mers;
          $self->{'error'} = "MersDB::save - Mersenne::parse failed for '$mers' - " . $m->errors() . "\n",
            unshift(@results, 0),
            return(@results)
            unless $m->parse($mers);
          @values = $m->get($_);
          if ($needsSPRP && $needsSPRP eq 'S'
              && exists($tables{'missing'}{'m'}{$_})
              && $m->n() == $self->lowestMissingExponent($_))
            { # The Mersenne's cofactor(s) are SPRP and likely prime, so advance the lowest
              #   missing exponent for this kind despite having no relevant info.
              my ($ok) = $self->advanceLME($m->n(), $tables{'missing'}{'m'}{$_});
              unshift(@results, 0, 'advancing lme failed'),
                return(@results)
                unless $ok;
              push(@results, 'lme', $ok);
            }
          next unless @values;
        }
      # Ignore cofactor digit counts less than the cofactor length
      #   - a factor has been lost or the digit count is wrong.
      warn("missing factor - 'M" . $m->n(). "$_: $values{$_}' but cofactor has "
           . $m->cofactorLength($_) . " digits; ignoring $_ value"),
        next
        if m,^[SLMclm]$, && $values[0] < $m->cofactorLength($_);
      @upRtn = $self->update($_, $m->n(), $_, @values);
      $self->{'error'} = "MersDB::save - update M" . $m->n() . " $_ " . join(' ', @values) . " failed\n",
        unshift(@results, 0),
        return(@results)
        unless @upRtn;
      push(@results, @upRtn);
      if (exists($tables{'missing'}{'m'}{$_}) && $m->n() == $self->lowestMissingExponent($_))
        {
          my ($ok) = $self->advanceLME($m->n(), $tables{'missing'}{'m'}{$_});
          unshift(@results, 0, 'advancing lme failed'),
            return(@results)
            unless $ok;
          push(@results, 'lme', $ok);
        }
      next unless m,^[ScLMlm]$,;
      $self->{'error'} = "MersDB::save - Mersenne::get " . $m->n() . " $_ returned multiple values - " . join(' ', @values) . "\n",
        unshift(@results, 0),
        return(@results)
        unless $#values == 0;
      $values{$_} = $values[0];
      if ($values{$_} != $m->cofactorLength($_))
        { # The cofactor needs to be SPRP tested - it could be prime.
          $needsSPRP = '1'
            unless $needsSPRP && $needsSPRP eq '3';
          next;
        }
      $needsSPRP = 'S',
        next
        if $_ eq 'S' && (!$needsSPRP || $needsSPRP ne '3');
      $needsSPRP = '0',
        next
        if $_ eq 'c' && (!$needsSPRP || $needsSPRP ne 3);
      $needsSPRP = 'S',
        next
        if $needsSPRP && $needsSPRP =~ m,^l$,i && m,^m$,i && (!$needsSPRP || $needsSPRP ne 3);
      $needsSPRP = 'S',
        next
        if $needsSPRP && $needsSPRP =~ m,^m$,i && m,^l$,i && (!$needsSPRP || $needsSPRP ne 3);
      $needsSPRP = $_
        if m,^[lm]$,i && (!$needsSPRP || $needsSPRP ne 3);
    }
  # If factors were saved, presume a new SPRP check is needed unless enough clmLMS data
  #   is available (checked in the prior foreach()) or the Mersenne is completely factored.
  $needsSPRP = 'factors'
    if $needsSPRP && $needsSPRP =~ m,^[lm]$,i && $#f >= 0 && $m->n() > $xexps[0];
  if (!defined($needsSPRP) && $m->n() > $self->lowestMissingExponent('c'))
    {
      push(@results, 0E0)
        unless @results;
      return(@results);
    }
  if ($needsSPRP && $needsSPRP eq 'done')
    {
      my (@rmRtn) = $self->remove('sprp', $m->n());
      $self->{'error'} = "MersDB::save - remove sprp flag for " . $m->n() . " failed - " . join(' ', @rmRtn) . "\n",
        unshift(@results, 0),
        return(@results)
        unless @rmRtn && $rmRtn[0];
      push(@results, @rmRtn);
      return(@results);
    }
  $needsSPRP = '0'
    if $needsSPRP && $needsSPRP eq 'S';
  $needsSPRP = '1'
    if !defined $needsSPRP || $needsSPRP !~ m,^[0-9]+$,;
  @upRtn = $self->update('sprp', $m->n(), $needsSPRP);
  $self->{'error'} = "MersDB::save - update sprp flag to $needsSPRP for " . $m->n() . ' failed - ' . join(' ', @upRtn) . "\n",
    unshift(@results, 0),
    return(@results)
    unless @upRtn && $upRtn[0];
  push(@results, @upRtn);
  return(@results);
}

# Get the values in the DB and save them in the 'value' entries in %tables for later.
#   Should only be used for small tables (e.g., 'missing', 'X', 'done', 'allowed', 'aliases').
sub fetchValues
{
  my ($self) = shift;
  $self->{'error'} = "not connected to a database - call MersDB::connect\n",
    return(undef)
    unless exists($self->{'createdTables'}) && $self->{'createdTables'} eq $self->{'uuid'};
  $self->{'error'} = "not connected to a database - call MersDB::connect\n",
    return(undef)
    unless exists $self->{'dbh'} && $self->{'dbh'};
  local $_;
  my (@tables);
  while (@_)
    {
      my ($arg) = shift @_;
      next unless $arg;
      push(@tables, $arg), next if exists $tables{$arg};
      my ($tbl);
      while (length($arg))
        {
          $tbl = substr($arg, 0, 1);
          $arg = substr($arg, 1);
          while (length($arg) && !exists($tables{$tbl}))
            {
              $tbl .= substr($arg, 0, 1);
              $arg = substr($arg, 1);
            }
          $self->{'error'} = 'MersDB::fetchValues - no such table in db: ' . "$tbl\n",
            return(undef)
            unless exists $tables{$tbl};
          push(@tables, $tbl);
        }
      $self->{'error'} = 'MersDB::fetchValues - no such table in db: ' . "$tbl\n",
        return(undef)
        if $tbl && !exists $tables{$tbl};
    }
  push(@tables, keys(%tables))
    unless @tables;
  my ($table);
  foreach $table (@tables)
    { # Verify all the tables requested support fetching before querying the db for any table
      $self->{'error'} = "MersDB::fetchValues - table '$table' not supported\n",
        return(undef)
        unless (exists($tables{$table}{'value'}) && exists($tables{$table}{'fetch'})
                && exists($tables{$table}{'keyIndex'}) && exists($tables{$table}{'valueIndex'}));
    }
  my (@results);
  foreach $table (@tables)
    {
      my ($ok, $sth) = $self->execute($table, 'fetch');
      push(@results, 'fetch', $table, $ok, $self->dbError()),
        next
        unless $ok;
      my (%indices); # key is mers format field number, value is column number in query
      my ($column) = 0; # column number of database table - for 'X', 1 ('x') and 2 ('extent')
      while (exists($tables{$table}{++$column}))
        {
          $indices{$tables{$table}{$column}} = $column # %indices = { 1 => 1, 3 => 2, }; for 'X', e.g.
        }
      my ($count) = 0;
      while (my $row = $sth->fetchrow_arrayref)
        { # indices within a returned row array are one off from indices within a list of '?'s in SQL and one off from within a split() of mers format
          $tables{$table}{'value'}{$row->[$indices{$tables{$table}{'keyIndex'}} - 1]} = $row->[$indices{$tables{$table}{'valueIndex'}} - 1];
          ++$count;
        }
      print("executions missing $sth\n")
        unless exists $executions{"$sth"};
      delete $executions{"$sth"};
      $sth->finish() if $sth->{'Active'};
      $tables{$table}{'fetched'} = 1;
      push(@results, 'fetch', $table, $count);
    }
  return(@results);
}

sub fetchByPriority
{ # returns list of exponents in priority order
  my ($self) = shift;
  my ($letter) = shift;
  my ($low) = shift;
  my ($count) = shift;
  my (@results);
  $self->{'error'} = "call MersDB::createTables first\n",
    return(@results)
    unless exists($self->{'createdTables'}) && $self->{'createdTables'} eq $self->{'uuid'};
  $self->{'error'} = "MersDB::fetchByPriority - unexpected arguments\n",
    return(@results)
    unless defined($letter) && exists($tables{$letter}) && exists($tables{$letter}{'value'}) && exists($tables{$letter}{'fetchByPriority'}) && $count =~ m/^[1-9][0-9]*$/;
  $low = 0.0 unless $low;
  my ($ok, $sth) = $self->execute($letter, 'fetchByPriority', $low, $count);
  return undef
    unless $ok;
  my (%indices); # key is mers format field number, value is column number in query
  my ($column) = 0; # of query
  $indices{$tables{$letter}{$_}} = $_
    foreach (grep { m,^[1-9][0-9]*$, } keys(%{$tables{$letter}}));
  while (my $row = $sth->fetchrow_arrayref)
    { # indices in a returned row array are one off from indices in a list of '?'s in SQL and one off from a split() of mers format
      my ($exp) = $row->[$indices{$tables{$letter}{'keyIndex'}} - 1];
      next if $letter ne 'S' && exists $tables{'S'}{'value'}{$exp}; # skip S exponents without querying the db
      next if $letter !~ m,^[LMS]$, && exists $tables{'L'}{'value'}{$exp} && exists $tables{'M'}{'value'}{$exp}; # skip LM exponents
      my ($val) = $row->[$indices{$tables{$letter}{'valueIndex'}} - 1];
      $tables{$letter}{'value'}{$exp} = $val;
      push(@results, $exp);
    }
  print("executions missing $sth\n")
    unless exists $executions{"$sth"};
  delete $executions{"$sth"};
  $sth->finish() if $sth->{'Active'};
  return(@results);
}

sub fetchRowsByPriority
{ # returns list (ARRAY ref) of mers format arrays in priority order - presumes mers format and SQL columns are in the same order except the mers letter
  my ($self) = shift;
  my ($letter) = shift;
  my ($low) = shift;
  my ($count) = shift;
  my ($results);
  $self->{'error'} = "call MersDB::createTables first\n",
    return(undef)
    unless exists($self->{'createdTables'}) && $self->{'createdTables'} eq $self->{'uuid'};
  my ($table) = $letter;
  $table = 'done'
    if $letter =~ m,$donePattern,;
  $self->{'error'} = "MersDB::fetchRowsByPriority - unsupported table $table\n",
    return(undef)
    unless defined($table) && exists($tables{$table}) && exists($tables{$table}{'value'}) && exists($tables{$table}{'fetchByPriority'});
  $low = 0.0 unless $low;
  $self->{'error'} = "MersDB::fetchRowsByPriority - invalid priority $low\n",
    return(undef)
    unless $low =~ m/$floatPattern/;
  $low = $1;
  $count = 1 unless $count;
  $self->{'error'} = "MersDB::fetchRowsByPriority - invalid count $count\n",
    return(undef)
    unless $count =~ m/^([1-9][0-9]*)$/;
  $count = $1;
  my (@args);
  push(@args, $letter)
    if $table eq 'done';
  push(@args, $low, $count);
  my ($ok, $sth) = $self->execute($table, 'fetchByPriority', @args);
  return(undef)
    unless $ok;
  my (%indices); # key is mers format field number, value is column number in query
  my ($column) = 0; # of query
  $indices{$tables{$table}{$column}} = $column
    while exists $tables{$table}{++$column};
  $column = 0;
  while (my $row = $sth->fetchrow_arrayref)
    { # indices within a returned row array are one off from indices within a list of '?'s in SQL and one off from within a split() of mers format
      my (@mers) = ( @{$row}, );
      my ($exp) = $mers[$indices{$tables{$table}{'keyIndex'}} - 1];
      next if $letter ne 'S' && exists $tables{'S'}{'value'}{$exp}; # skip S exponents without querying the db
      next if $letter !~ m,^[LMS]$, && exists $tables{'L'}{'value'}{$exp} && exists $tables{'M'}{'value'}{$exp}; # skip LM exponents
      my ($val) = $mers[$indices{$tables{$table}{'valueIndex'}} - 1];
      $tables{$table}{'value'}{$exp} = $val;
      # Insert the mers letter; the caller wants it.
      splice(@mers, 1, 0, $letter);
      #!!Bug in DBI.pm or DBD.pm or ... is always returning '0' for BIG DOUBLEs, so calculate the priority.
      $val = Mersenne::mersPriority(@mers),
        $mers[$tables{$letter}{$tables{$table}{'pri'}} - 1] = $val
        if exists $tables{$table}{'pri'};
      push(@{$results}, [ @mers, ] );
    }
  print("executions missing $sth\n")
    unless exists $executions{"$sth"};
  delete $executions{"$sth"};
  $sth->finish() if $sth->{'Active'};
  $results = [ ] # empty ARRAY ref
    unless $results;
  return($results);
}

sub fetchRows
{ # Returns list of ARRAY refs in SQL's default order, usually sorted by the primary key(s).
  #   Does _not_ update $tables{$table}{'value'} HASH.
  #   Presumes the fields are one-to-one - 1 => 1, 2 => 2, etc.
  my ($self) = shift;
  my ($table) = shift;
  local $_;
  my (@results);
  $self->{'error'} = "call MersDB::createTables first\n",
    return(@results)
    unless exists($self->{'createdTables'}) && $self->{'createdTables'} eq $self->{'uuid'};
  $self->{'error'} = "MersDB::fetchRows - unsupported table $table\n",
    return(@results)
    unless defined($table) && exists($tables{$table}) && exists($tables{$table}{'fetch'});
  my ($ok, $sth) = $self->execute($table, 'fetch');
  return undef
    unless $ok;
  while (my $row = $sth->fetchrow_arrayref)
    { # Need a "deep" copy; $row is the same address every time
      my ($r) = [ @{$row}, ];
      push(@results, $r);
    }
  print("executions missing $sth\n")
    unless exists $executions{"$sth"};
  delete $executions{"$sth"};
  $sth->finish() if $sth->{'Active'};
  return(@results);
}

sub queryRows
{ # Returns list of HASH refs in SQL's default order, usually sorted by the primary key(s).
  #   Does _not_ update $tables{$table}{'value'} HASH.
  # The caller has to know what fields and values to pass in and what the HASH keys will be
  #   based on the SQL of the table.
  my ($self) = shift;
  my ($table) = shift;
  my (@fields) = @_;
  my (@results);
  local $_;
  $self->{'error'} = "call MersDB::createTables first\n",
    return(@results)
    unless exists($self->{'createdTables'}) && $self->{'createdTables'} eq $self->{'uuid'};
  $self->{'error'} = "MersDB::queryRows - unsupported table $table\n",
    return(@results)
    unless defined($table) && exists($tables{$table}) && exists($tables{$table}{'query'});
  my ($ok, $sth) = $self->execute($table, 'query', @fields);
  return(@results)
    unless defined($ok);
  while (my $row = $sth->fetchrow_hashref)
    { # Need a "deep" copy; $row is the same address every time
      my ($r) = { %{$row}, };
      push(@results, $r);
    }
  print("executions missing $sth\n")
    unless exists $executions{"$sth"};
  delete $executions{"$sth"};
  $sth->finish() if $sth->{'Active'};
  return(@results);
}

sub find
{
  my ($self) = shift;
  my ($exp) = shift;
  my ($letters) = shift;
  my ($recurse) = 1;
  $recurse = 0,
    $exp = $letters,
    $letters = shift
    if $exp eq 'value';
  my ($mers); # to "merge" multiple clmLMSDPEoq lines and check factors
  $mers = Mersenne->new('n' => $exp)
    if (!ref($exp) || ref($exp) eq 'Math::BigInt') && $exp <= $self->lowestMissingExponent('c');
  $mers = $exp,
    $exp = $mers->n()
    if ref($exp) eq 'Mersenne';
  $self->{'error'} = "call MersDB::createTables first\n",
    return(undef)
    unless exists($self->{'createdTables'}) && $self->{'createdTables'} eq $self->{'uuid'};
  $letters = $validLetters unless $letters;
  $self->{'error'} = "MersDB::find - invalid exponent $exp\n",
    return(undef)
    unless defined($exp) && $exp =~ m/^([1-9][0-9]*)$/;
  $exp = $1;
  $self->{'error'} = "MersDB::find - unknown mers letters given - '$letters'\n",
    return(undef)
    unless defined($letters) && $letters =~ m/^([$validLetters]+)$/;
  $letters = "$1";
  local $_;
  $self->{'error'} = "not connected to a database - call MersDB::connect\n",
    return(undef)
    unless exists $self->{'dbh'} && $self->{'dbh'};
  my ($dbh) = $self->{'dbh'};
  my ($otherLetters) = ''; # list of mers letters not in the database but requested; bsearch will be used for their data
  my ($found) = ''; # list of mers letters seen so far; needed to possibly add clm line(s) and not repeat data
  my ($left) = $letters;
  # Get factors before cofactor info since cofactor sizes are affected by known factors.
  # Get all cofactor related info if any is requested.
  $left = 'CPDSLMclm' . $left if $letters =~ m/[CPDSLMclm]/;
  $left = 'C' . $left, # factors should always be retrieved first as noted
    $otherLetters = 'C' # always do bsearch for factors since the db may not have them all
    if $letters =~ s/C//g;
  $left =~ s/[lm]//ig unless $exp % 8 == 4;
  $left =~ s/[cS]//g if $exp % 8 == 4;
  my (@data); # mers format data requested
  while (length($left))
    {
      my ($m) = substr($left, 0, 1);
      $left =~ s/$m//g;
      next if $m =~ m/^["$found"]$/;
      next if $m =~ m/$donePattern/ && $found =~ m/[$doneLetters]/;
      $otherLetters .= $m,
        next
        unless exists $tables{$m} || $m =~ m/$donePattern/;
      if ($recurse && $m =~ m,^[HUclmLMS]$, && exists($tables{$m}{'value'})
          && exists($tables{$m}{'value'}{$exp}) && $tables{$m}{'value'}{$exp})
        { # These mers letters only have one value and %tables has everything needed.
          $self->{'error'} = "MersDB::find - Mersenne::parse $exp $m failed - " . $mers->errors(),
            return(undef)
            if $mers && !$mers->parse("M( $exp )$m: " . $self->value($m, $exp) . "\n");
          push(@data, "M( $exp )$m: " . $self->value($m, $exp) . "\n")
            unless $mers;
          $found .= $m unless $found =~ /$m/;
          next;
        }
      my ($table) = $m;
      $table = 'done' if $m =~ m/$donePattern/;
      if ($recurse && $table eq 'done')
        {
          if (defined($self->value('done', $exp)))
            {
              $m = $self->value('done', $exp);
              $self->{'error'} = "MersDB::find - Mersenne::parse $exp $m failed - " . $mers->errors(),
                return(undef)
                if $mers && !$mers->parse("M( $exp )$m\n");
              push(@data, "M( $exp )$m\n")
                unless $mers;
              $found .= $m;
              last; # must be P or D and factors already fetched, so nothing more to get
            }
          $otherLetters .= $m unless $otherLetters =~ m,$m, && $found !~ m,$m,;
        }
      if ($dbh && exists($tables{$table}) && exists($tables{$table}{'query'}))
        { # The requested data should be in the db and %tables knows how to get it.
          my ($ok, $sth) = $self->execute($table, 'query', $exp);
          next unless $ok;
          # $ord and $value only needed for 'C' (factors) since those values might be > 2^64
          my ($ord) = 0;
          my ($value) = Math::BigInt->bzero();
          while (my $row = $sth->fetchrow_arrayref)
            {
              if ($m ne 'C')
                { # Copy the data or the next pass thru the loop will affect it.
                  my (@info) = ( @{$row}, );
                  # Drop the priority; it's not strictly part of the mers format.
                  #   !!... and fetchrow_arrayref() is always returning 0 for it anyway.
                  # The -1 is needed because 'pri' here is the SQL index, not the mers format index
                  #   or index into @indices.
                  splice(@info, $tables{$table}{'pri'} - 1, 1)
                    if (exists($tables{$table}{'pri'})
                        && $tables{$table}{'pri'} - 1 <= $#info);
                  # Drop the exponent.  Since the mers letter is not in the SQL table ...
                  splice(@info, $tables{$table}{'keyIndex'} - 1, 1)
                    if ($tables{$table}{'keyIndex'} - 1 <= $#info
                        && $info[$tables{$table}{'keyIndex'} - 1] == $exp);
                  # ... only the 'data' (after the mers letter in mers format) should remain.
                  next
                    if !@info || ($table =~ m,^[Eoq]$, && !$info[0]); # zero means no actual work/effort/progress
                  push(@data, "M( $exp )$m: " . join(' ', @info) . "\n");
                  $self->{'error'} = "MersDB::find - Mersenne::parse '$data[-1]' failed - " . $mers->errors(),
                    return(undef)
                    if $mers && !$mers->parse($data[-1]);
                  pop(@data);
                  $tables{$table}{'value'}{$exp} = $info[1]
                    if exists $tables{$table}{'value'};
                  $found .= $m
                    unless $found =~ /$m/;
                  next;
                }
              $value += Math::BigInt->new("$row->[3]")->blsft(($row->[2])*$SQLmaxBits),
                next
                if $ord == $row->[1];
              # factor #$ord is complete; add it to $mers.
              $mers->f($value)
                if $mers && $value;
              push(@data, "M( $exp )C: $value\n")
                if !$mers && $value;
              $ord = $row->[1];
              #!!verify that the limb, $row->[2], is 0?
              $value = Math::BigInt->new("$row->[3]");
            }
          print("executions missing $sth\n")
            unless exists $executions{"$sth"};
          delete $executions{"$sth"};
          $sth->finish()
            if $sth->{'Active'};
          $mers->f($value)
            if $mers && $m eq 'C' && $value;
          push(@data, "M( $exp )C: $value\n")
            if !$mers && $m eq 'C' && $value;
          $otherLetters .= $m
            if !$value && index($otherLetters, $m) < 0 && $found !~ m/$m/;
          next;
        }
    }
  if (-s($self->{'file'}{'bsearch'}) && -x($self->{'file'}{'bsearch'}) && ($otherLetters || !@data))
    { # Run bsearch and add to @data appropriately.  Get all cofactor info if any is needed.
      $otherLetters = 'PDSLMclm' . $otherLetters
        if $otherLetters =~ m/[PDSLMclm]/;
      # only one of HU will be present, so include either if either is requested
      $otherLetters .= 'H'
        if $otherLetters && $otherLetters =~ m/U/ && $otherLetters !~ m/H/;
      $otherLetters .= 'U'
        if $otherLetters && $otherLetters =~ m/H/ && $otherLetters !~ m/U/;
      # If the Mersenne is prime or completely factored, only factors (C) might be missing.
      $otherLetters = 'C'
        if $mers && ($mers->get('S') || $mers->get('D') || $mers->get('P') || ($mers->get('L') && $mers->get('M')));
      my ($cmd) = "$self->{'file'}{'bsearch'} -1n $exp";
      # This directory (or file) contains all non-gaps data for all incompletely factored Mersenne numbers.
      $cmd .= " $self->{'file'}{'results'}"
        if exists($self->{'file'}{'results'}) && (-s($self->{'file'}{'results'}) || -d($self->{'file'}{'results'}));
      # This file contains all known data for completely factored Mersenne numbers.
      $cmd .= " $self->{'file'}{'done'}"
        if $self->{'file'}{'done'} && $otherLetters =~ m/[CD]/;
      # This file contains the list of known Mersenne primes.
      $cmd .= " $self->{'file'}{'primes'}"
        if $self->{'file'}{'primes'} && $otherLetters =~ m/P/;
      # This file contains the known trial factoring gaps and joins for all incompletely factored prime exponent Mersenne numbers.
      $cmd .= " $self->{'file'}{'gaps'}"
        if $self->{'file'}{'gaps'} && $otherLetters =~ m/[GgJ]/;
      # These files were given to addSortedFiles().
      delete($self->{'sortedFiles'})
        if $self->{'sortedFiles'} && ref($self->{'sortedFiles'}) ne 'HASH';
      $cmd .= ' ' . join(' ', keys(%{$self->{'sortedFiles'}}))
        if exists $self->{'sortedFiles'};
      # Ignore these two signals while running bsearch
      my ($oldINT) = $SIG{'INT'};
      my ($oldTERM) = $SIG{'TERM'};
      $SIG{'INT'} = "IGNORE";
      $SIG{'TERM'} = "IGNORE";
      my ($fh);
      my ($pid) = open($fh, "$cmd |");
      $self->{'error'} = "MersDB::find could not run '$cmd', $!\n",
        return(undef)
        unless $pid;
      my (@newData);
      my ($failed);
      while (<$fh>)
        {
          next
            unless m/^M[(] $exp [)]/;
          next
            if !$otherLetters || !m/^M[(] $exp [)][$otherLetters]/;
          $found .= $1
            if (m/^M[(] $exp [)]([clmLMS$doneLetters]): /
                || m/^M[(] $exp [)]([LMS$doneLetters])$/); # older style LMS - no digit count
          chomp(),
            push(@newData, "$_\n"),
            next
            if m,[GgJ]:,; # Mersenne.pm does not support GgJ data.
          push(@newData, $_),
            next
            unless $mers;
          next
            if $mers->parse($_);
          $self->{'error'} = "MersDB::find - Mersenne::parse '$_' failed - " . $mers->errors();
          $failed = 1;
          last;
        }
      $SIG{'INT'} = $oldINT;
      $SIG{'TERM'} = $oldTERM;
      $self->{'error'} .= "MersDB::find could not close bseach child, $!\n",
        $failed = 1
        unless close($fh);
      return(undef)
        if $failed;
      while (@newData)
        {
          my ($line) = shift(@newData);
          push(@data, $line)
            unless scalar(grep { $_ eq $line } @data);
        }
    }
  # "Invent" the clm line(s) if they were requested and are not present.
  if ($exp % 8 == 4 && $letters =~ m/[lmLMSDP]/)
    {
      push(@data, "M( $exp )l: $exp\n") if $found !~ m/[LlS$doneLetters]/;
      push(@data, "M( $exp )m: $exp\n") if $found !~ m/[MmS$doneLetters]/;
    }
  elsif ($exp % 8 != 4 && $letters =~ m/[cSlmLM$doneLetters]/)
    {
      push(@data, "M( $exp )c: $exp\n") if $found !~ m/[cS$doneLetters]/;
    }
  return(join('', $mers->str(), @data))
    if $mers;
  return(join('', @data));
}

# Return the keys in the 'value' hash for the table.
sub keyList
{
  my ($self) = shift;
  my ($table) = shift;
  my ($lowest) = shift; # don't return smaller exponents
  my (@list);
  $self->{'error'} = "call MersDB::createTables first\n",
    return(@list)
    unless exists($self->{'createdTables'}) && $self->{'createdTables'} eq $self->{'uuid'};
  $self->{'error'} = "MersDB::keyList - unknown or unsupported table $table\n",
    return(@list)
    unless defined $table && exists $tables{$table}{'value'};
  $self->{'error'} = "MersDB::keyList for table '$table' called before fetchValues\n",
    return(@list)
    unless exists $tables{$table}{'fetched'};
  return(@list)
    unless scalar(keys(%{$tables{$table}{'value'}})) > 0;
  return(sort { ($a =~ m/^[0-9]+$/ && $b =~ m/^[0-9]+$/)
                  ? ($a <=> $b)
                  : ($a cmp $b) } keys(%{$tables{$table}{'value'}}))
    unless defined($lowest);
  return(sort { $a <=> $b } grep { $_ >= $lowest } keys(%{$tables{$table}{'value'}}));
}

sub value
{ # Return the value in the 'value' hash for the table and key.
  my ($self) = shift;
  my ($table) = shift;
  my ($exp) = shift;
  $self->{'error'} = "not connected to a database - call MersDB::connect\n",
    return(undef)
    unless exists($self->{'createdTables'}) && $self->{'createdTables'} eq $self->{'uuid'};
  $self->{'error'} = "MersDB::value - missing table or exponent or unsupported table\n",
    return(undef)
    unless defined $table && defined $exp && exists $tables{$table}{'value'};
  $self->{'error'} = "MersDB::value - invalid exponent\n",
    return(undef)
    if $table ne 'missing' && $exp !~ m/^([1-9][0-9]*)$/;
  $self->{'error'} = "MersDB::value - invalid mers letter\n",
    return(undef)
    if $table eq 'missing' && $exp !~ m/^([cEoqU])$/;
  $exp = $1;
  return($tables{$table}{'value'}{"$exp"})
    if exists $tables{$table}{'value'}{"$exp"};
  return($exp)
    if exists $tables{$table}{'fetched'} && $table =~ m/^[clm]$/;
  return(undef)
    if exists $tables{$table}{'fetched'} && $table =~ m,^(done)|(missing)|(sprp)|([CXSLM])$,;
  return(0)
    if exists $tables{$table}{'fetched'};
  $self->{'error'} = "MersDB::value - unsupported table $table\n",
    return(undef)
    unless exists $tables{$table}{'query'} && exists $tables{$table}{'queryIndex'} && exists $tables{$table}{'valueIndex'};
  # This table has not been read by fetchValues(), so the database or results/ directory or M/factors file has to be consulted.
  my ($val);
  if (length($table) > 1)
    { # Not mers format, so just a single value and not in results or M/factors - call execute()
      #   directly except 'reservations' has 3 columns and 3 different queries but not this one.
      my ($ok, $sth) = $self->execute($table, 'query', $exp);
      return(undef)
        unless $ok;
      my (%indices); # key is mers format field number, value is column number in query
      my ($column) = 0; # of query
      $indices{$tables{$table}{$_}} = $_
        foreach (grep { m,^[1-9][0-9]*$, } keys(%{$tables{$table}}));
      while (my $row = $sth->fetchrow_arrayref)
        { # see fetchByPriority's similar code
          my ($key) = $row->[$indices{$tables{$table}{'queryIndex'}} - 1];
          my ($value) = $row->[$indices{$tables{$table}{'valueIndex'}} - 1];
          $tables{$table}{'value'}{$key} = $value;
          $val = $value,
            next
            if $exp == $key; # should always be true, so:
          warn("MersDB::value '$table' '$exp' 'query' returned unexpected row - " . join(' ', @{$row}) . "\n");
        }
      print("executions missing $sth\n")
        unless exists $executions{"$sth"};
      delete $executions{"$sth"};
      $sth->finish() if $sth->{'Active'};
      $tables{$table}{'value'}{$exp} = $val
        if exists $tables{$table}{'value'};
      return($val);
    }
  # The first argument is to prevent find() from calling this recursively, forcing find() to query
  #   the db, the results directory, or the M/factors file.
  my ($mers) = $self->find('value', $exp, $table);
  chomp($mers); # if $table is 'c', 'l' or 'm', this could be two lines, one for 'l' and one for 'm', so ...
  my (@lines) = split(/[\n]/, $mers); # ... split the lines ...
  my (@line) = grep { m/[)]$table/ } @lines; # ... get the line needed ...
  my (@fields) = split(/ /, $line[0]) if $#line == 0; # ... split the line into fields ...
  @fields = split(/ /, $lines[0]) if !@fields && @lines == 1; # ... check for the single line case ...
  my ($i) = $tables{$table}{'valueIndex'};
  $val = $fields[$i] if $#fields >= $i;
  $tables{$table}{'value'}{$exp} = $val # ... and retain and return the single value requested.
    if defined $val && exists $tables{$table}{'value'};
  return($val);
}

sub next
{
  my ($self) = shift;
  my ($table) = shift;
  $self->{'error'} = "call MersDB::createTables first\n",
    return(undef)
    unless exists($self->{'createdTables'}) && $self->{'createdTables'} eq $self->{'uuid'};
  $self->{'error'} = "MersDB::next(): unsupported table $table\n",
    return(undef)
    unless defined $table && exists $tables{$table} && exists $tables{$table}{'next'};
  $self->{'error'} = "MersDB::next - table '$table' 'next' SQL is missing 'limit 1'\n",
    return(undef)
    unless $tables{$table}{'next'} =~ m/ limit 1 *$/;
  my ($ok, $sth) = $self->execute($table, 'next');
  return $ok unless $ok;
  my ($value);
  while (my $row = $sth->fetchrow_arrayref)
    {
      $value = $row->[0];
    }
  print("executions missing $sth\n")
    unless exists $executions{"$sth"};
  delete $executions{"$sth"};
  $sth->finish() if $sth->{'Active'};
  $value = $self->lowestMissingExponent($table)
    unless $value;
  return($value);
}

sub remove
{
  my ($self) = shift;
  my ($table) = shift;
  my ($exp) = shift;
  $self->{'error'} = "call MersDB::createTables first\n",
    return(undef)
    unless exists($self->{'createdTables'}) && $self->{'createdTables'} eq $self->{'uuid'};
  $self->{'error'} = "MersDB::remove - " . (defined($table) ? "unsupported table '$table'" : 'no table given') . "\n",
    return(undef)
    unless defined $table && exists $tables{$table} && exists $tables{$table}{'delete'} && exists $tables{$table}{'deleteIndex'};
  $self->{'error'} = "MersDB::remove - " . (defined($exp) ? "non-numeric exponent '$exp'" : ' no exponent given') . "\n",
    return(undef)
    unless defined $exp && $exp =~ m/^([1-9][0-9]*)$/;
  $exp = $1;
  delete $tables{$table}{'value'}{$exp} if exists $tables{$table}{'value'} && exists $tables{$table}{'value'}{$exp};
  my ($ok, $sth) = $self->execute($table, 'delete', $exp);
  return($ok);
}

sub reserve
{
  my ($self) = shift;
  my (@fields) = @_;
  $self->{'error'} = "MersDB::reserve requires three arguments - exp, UUID, kind\n",
    return(undef)
    unless $#fields == 2;
  $self->{'error'} = "MersDB::reserve - invalid Mersenne exponent - '$fields[0]'\n",
    return(undef)
    unless $fields[0] =~ m,^([0-9]+)$,;
  $fields[0] = $1;
  $self->{'error'} = "MersDB::reserve - Mersenne exponent is too large - '$fields[0]'\n",
    return(undef)
    unless $fields[0] <= $SQLmaxInt;
  $self->{'error'} = "MersDB::reserve - UUID empty or includes unexpected characters - '$fields[1]'\n",
    return(undef)
    unless $fields[1] =~ m,$uuidPattern,;
  $fields[1] = $1;
  $self->{'error'} = "MersDB::reserve - unexpected character for kind - '$fields[2]'\n",
    return(undef)
    if $fields[2] !~ m,^([?!c$Mersenne::workLetters])$,;
  $fields[2] = $1;
  # Don't raise an error here due to race condition.
  local $self->{'dbh'}{'RaiseError'};
  my (@rtn) = $self->update('reserve', @fields);
  return(@rtn);
}

sub release
{
  my ($self) = shift;
  my (@fields) = @_;
  my (@results);
  $self->{'error'} = "MersDB::release requires three arguments - exp, UUID, kind\n",
    return(@results)
    unless $#fields == 2;
  $self->{'error'} = "MersDB::release - invalid Mersenne exponent - '$fields[0]'\n",
    return(@results)
    unless $fields[0] =~ m,^([0-9]+)$,;
  $fields[0] = $1;
  $self->{'error'} = "MersDB::release - Mersenne exponent too large - '$fields[0]'\n",
    return(@results)
    unless $fields[0] <= $SQLmaxInt;
  $self->{'error'} = "MersDB::release - UUID empty or includes unexpected characters - '$fields[1]'\n",
    return(@results)
    unless $fields[1] =~ m,$uuidPattern,;
  $fields[1] = $1;
  $self->{'error'} = "MersDB::release - unexpected character for kind - '$fields[2]'\n",
    return(@results)
    if $fields[2] !~ m,^([?!c$Mersenne::workLetters])$,;
  $fields[2] = $1;
  my ($ok, $sth) = $self->execute('reservations', 'exists', @fields);
  return(@results)
    unless $ok;
  while (my $row = $sth->fetchrow_hashref)
    {
      push(@results, { %{$row}, } );
    }
  print("executions missing $sth\n")
    unless exists $executions{"$sth"};
  delete $executions{"$sth"};
  $sth->finish() if $sth->{'Active'};
  $self->{'error'} = "MersDB::release() given unreserved values - " . join(' ', @fields) . "\n",
    unshift(@results, 0),
    return(@results)
    unless ($#results == 0
            && $results[0]->{'exp'} == $fields[0]
            && $results[0]->{'uuid'} eq $fields[1]
            && $results[0]->{'kind'} eq $fields[2]);
  return($self->update('release', $fields[0]));
}

sub countReserved
{
  my ($self) = shift;
  my ($uuid) = shift;
  my ($kind) = shift;
  $self->{'error'} = "call MersDB::createTables first\n",
    return(undef)
    unless exists($self->{'createdTables'}) && $self->{'createdTables'} eq $self->{'uuid'};
  local $_;
  $self->{'error'} = "MersDB::countReserved called with unexpected UUID '$uuid'\n",
    return(undef)
    unless defined $uuid && $uuid =~ m/$uuidPattern/;
  $uuid = $1;
  $self->{'error'} = "MersDB::countReserved called with unexpected kind '$kind'\n",
    return(undef)
    unless defined $kind && $kind =~ m/^([?!A-Za-z])$/;
  $kind = $1;
  my ($ok, $sth) = $self->execute('reservations', 'count', $uuid, $kind);
  return undef unless $ok;
  $ok = 0;
  $ok += $_->[0] foreach ($sth->fetchrow_arrayref);
  print("executions missing $sth\n")
    unless exists $executions{"$sth"};
  delete $executions{"$sth"};
  $sth->finish() if $sth->{'Active'};
  return($ok);
}

sub reservations
{ # return the reservations with the given field
  my ($self) = shift;
  my ($value) = shift; # the exp, uuid, or kind
  my ($index);
  my (@results);
  $self->{'error'} = "MersDB::reservations - no value given\n",
    return(@results)
    unless $self && defined($value);
  $self->{'error'} = "call MersDB::createTables first\n",
    return(@results)
    unless exists($self->{'createdTables'}) && $self->{'createdTables'} eq $self->{'uuid'};
  local $_;
  if ($value =~ m/^([0-9]+)$/)
    { # exponent is the most common query; 0 is allowed to "lock" the db
      $index = 1;
    }
  elsif ($value =~ m/^([?!A-Za-z])$/)
    { # mers letter is the only column that can be a single character
      $index = 3;
    }
  elsif ($value =~ m/$uuidPattern/)
    { # UUID of the client/worker
      $index = 2;
    }
  else
    { # invalid data given
      $self->{'error'} = "MersDB::reservations - unsupported value '$value'\n";
      return @results;
    }
  $value = $1; # all three cases set $1
  my ($ok, $sth) = $self->execute('reservations', 'query' . $index, $value);
  return @results unless $ok;
  while (my $row = $sth->fetchrow_hashref)
    {
      push(@results, { %{$row}, } );
    }
  print("executions missing $sth\n")
    unless exists $executions{"$sth"};
  delete $executions{"$sth"};
  $sth->finish() if $sth->{'Active'};
  return(@results);
}

sub reduceMemory
{ # Drop "extra" 'value' data to reduce memory usage - only Eoqclm and sprp, most likely,
  #   but this code supports any entry in %tables with a "sub-key" of 'value' that is
  #   either 'sprp' or a letter in $validLetters.
  my ($self) = shift;
  my ($dropped) = 0;
  local $_;
  # The 'sprp' table is currently the only table with different logic for dropping data from
  #   $tables{'sprp'}{'value'} since next() doesn't need 'value' - simply wiping 'value' is best.
  if (scalar(keys(%{$tables{'sprp'}{'value'}})) > $maxValues)
    {
      delete $tables{'sprp'}{'fetched'};
      $tables{'sprp'}{'value'} = { };
      return(1);
    }
  foreach (split(//, $validLetters))
    {
      next if !exists($tables{$_}) || !exists($tables{$_}{'value'}) || scalar(keys(%{$tables{$_}{'value'}})) <= $maxValues;
      delete $tables{$_}{'fetched'};
      while (scalar(keys(%{$tables{$_}{'value'}})) > $maxValues)
        {
          my ($key) = each %{$tables{$_}{'value'}};
          delete $tables{$_}{'value'}{$key};
          ++$dropped;
        }
    }
  return($dropped);
}

# Use the given statement handle to print the rows of the query.
sub printRows
{
  my ($self) = shift;
  my ($h) = shift;
  my ($cnt);
  local $_;
  $self->{'error'} = "call MersDB::connect first\n",
    return
    unless exists $self->{'dbh'} && $self->{'dbh'} && $h && ref($h) eq 'HASH';
  while (my $row = $h->fetchrow_hashref)
    {
      $self->{'error'} = "MersDB::printRows could print header line, $!\n",
        return(undef)
        if !$cnt && !print(join(",", sort(keys(%{$row}))) . "\n");
      ++$cnt;
      my ($line);
      foreach (sort(keys(%{$row})))
        {
          $line .= ',' if defined $line;
          $line = '' if !defined $line;
          $line .= $row->{$_};
        }
      $self->{'error'} = "MersDB::printRows - print() failed, $!\n",
        return(undef)
        unless print("$line\n");
    }
  return(1);
}

1;
