#!/usr/bin/perl -Tw
# MersInfo.pm: info about the mers software package.

package MersInfo;

use strict;
use strict 'vars';

# String read from file 'version' with a default value.
#   The 'version' file will be read for the actual value.
our ($VERSION) = '2021-Nov-07';

my ($unread) = 1; # has the 'version' file been read? should contain git version info
our ($maxLogSize) = (1 << 26); # rotateLog() file size limit in bytes (64MB)

sub author
{
  return('Will Edgington <wedgingt@acm.org>');
}

sub url
{
  return('https://sourceforge.net/projects/mers/');
}

sub version
{ # This 'version' file is created by the makefile.
  if ($unread && open(my $vFH, '<', 'version'))
    {
      $VERSION = <$vFH>;
      $VERSION or die "could not read from 'version' after opening it, $!";
      chomp $VERSION;
      close($vFH) or die "could not close 'version' after reading it, $!";
      $unread = 0;
    }
  return($VERSION);
}

sub doWorkDir
{ # This subdir is used by doWork.pl and other scripts and programs it runs.
  return('working.doWork');
}

# List of common regular expression patterns to avoid typos.
#   Returned by pattern(), just below.
my (%patterns) = (
  # Pattern for arbitrary non-negative integers.
  'bigint' => q{^([0-9]+)$},
  # This pattern is from 'man perldata' to match a 'C float' but no leading '-'.
  #   $1 will be the entire float, $2 the + if present, $3 the integer part, etc.
  #   $1 is used often and $3 at least once.
  'float' => q{^(([+]?)(\d+)(\.\d*)?([Ee]([+-]?\d+))?)$},
  # Pattern for UUIDs.
  'uuid' => q{^([-0-9A-Za-z._]+)$},
  );

sub patterns($)
{
  my ($name) = shift;
  return(undef)
    unless exists $patterns{$name};
  return($patterns{$name});
}

# Given a number and a hash reference, return the next larger key in the hash.
#   If the number is a key, return it.  If the number is larger than all keys,
#   return the smallest key.
# The intent is to iterate over the keys in numeric order repeatedly and easily
#   without needing to ever sort the keys.
# Utility sub, not a member function.
sub nextKey($$)
{
  local $_;
  my ($key, $hash) = @_;
  my ($larger, $smallest);
  return($key) if exists $hash->{$key};
  foreach (keys(%{$hash}))
    {
      $smallest = $_ if !defined $smallest || $smallest > $_;
      next unless $_ > $key;
      $larger = $_ if !defined $larger || $larger > $_;
    }
  return($larger) if defined $larger;
  return($smallest) if defined $smallest;
  return(undef); # empty hash
}

# "Rotate" a log file.  Utility sub, not a member function.
#   Ignore ENOENT since there's a race when called for the same file at the same time.
sub rotateLog($$)
{
  my ($f, $cnt) = @_;
  return
    unless -e $f && -f $f && -s $f >= $maxLogSize;
  $cnt = 9
    if !$cnt || $cnt !~ m/^([1-9][0-9]*)$/;
  $cnt = $1
    if $cnt && $cnt =~ m/^([1-9][0-9]*)$/;
  $cnt = 9
    if $cnt < 9;
  my ($n) = 0;
  my (@b) = split(/[.]/, $f);
  my ($e) = '';
  $e = pop(@b)
    if $#b > 0;
  my ($b) = join('.', @b);
  ++$n
    while ($e && -s("$b-$n.$e") && $n < $cnt + 1);
  ++$n
    while (!$e && -s("$b.$n") && $n < $cnt + 1);
  # If this last file, one beyond $cnt, exists, another call is likely in progress, so pause briefly.
  sleep(1)
    if ($e && -s("$b-$n.$e")) || (!$e && -s("$b.$n"));
  while ($n > 0)
    {
      !$e
        || rename("$b-" . ($n - 1) . ".$e", "$b-$n.$e")
        or $!{'ENOENT'} or die "could not rename to $b-$n.$e, $!";
      $e
        || rename("$b." . ($n - 1), "$b.$n")
        or $!{'ENOENT'} or die "could not rename to $b.$n, $!";
      --$n;
    }
  !$e || rename("$b.$e", "$b-0.$e") or $!{'ENOENT'} or die "could not rename $b.$e to $b-0.$e, $!\n";
  $e || rename("$b", "$b.0") or $!{'ENOENT'} or die "could not rename $b to $b.0, $!\n";
  $n = $cnt + 1;
  # Remove the extra file if present to avoid stalling later calls now that we're done.
  unlink("$b-$n.$e")
    if $e && -f("$b-$n.$e");
  unlink("$b.$n")
    if !$e && -f("$b.$n");
}

# This is usually just '1;' but this verifies all the Perl scripts using this module have set this.
$ENV{'PATH'} eq '/usr/bin:/bin';
