#!/usr/bin/perl -Tw
# Mersenne.pm: Math::BigInt functions related to Mersenne numbers.

package Mersenne;

use strict;
use strict 'vars';
use warnings;

BEGIN
{ # This is currently needed on MSWin10 with Cygwin when Math::Prime::Util
  #   and/or Math::BigInt is installed using the cpan program in a user's
  #   Cygwin home directory.
  if ((!exists($ENV{'HOME'}) || $ENV{'HOME'} !~ m,^/home/[a-zA-Z]+/,i)
      && exists($ENV{'USER'}) && $ENV{'USER'} =~ m,^([a-zA-Z]+),i)
    {
      $ENV{'USER'} = $1;
      $ENV{'HOME'} = ('/home/' . $ENV{'USER'}) if -d('/home/' . $ENV{'USER'});
    }
  if (exists($ENV{'HOME'}) && $ENV{'HOME'} =~ m,^(/home/[a-zA-Z]+)/?$,i)
    {
      $ENV{'HOME'} = $1;
      push(@INC, $ENV{'HOME'} . '/perl5/lib/perl5/x86_64-cygwin-threads-multi/')
        if -d $ENV{'HOME'} . '/perl5/lib/perl5/x86_64-cygwin-threads-multi/';
      push(@INC, $ENV{'HOME'} . '/perl5/lib/perl5/x86_32-cygwin-threads-multi/')
        if -d $ENV{'HOME'} . '/perl5/lib/perl5/x86_32-cygwin-threads-multi/';
    }
  push(@INC, '.') if -s('MersInfo.pm');
}

use Math::BigInt try => 'GMP';
use Math::Prime::Util;

use MersInfo;

our ($VERSION) = MersInfo::version();

use parent 'Exporter';

our @EXPORT_OK = qw();

use Class::Tiny qw ( );

# Don't allow NaN or infinity in calculations.
our ($traps) = Math::BigInt->config( { trap_nan => 1, trap_inf => 1, } );

my ($floatPattern) = MersInfo::patterns('float');

# All the mers letters in mers format that might be work/tasks to complete:
#   'E' ECM effort and largest bound;
#   'G' trial factoring gap to close;
#   'g' trial factoring gap between powers of 2;
#   'H' trial factoring extent;
#   'o' P-1 bound(s);
#   'q' P+1 bound(s);
#   'S' SPRP cofactors to prove prime;
#   'U' trial factoring extent for a Mersenne that might be prime (prime exponent and
#        no known factors and no completed Lucas-Lehmer or SPRP test).
# The order is not arbitrary - work for a particular Mersenne should be done in this order -
#     check "known" factors ('C') first, then prove SPRP factors and cofactors are prime (all
#     at once after the cofactors are SPRP), then do SPRP checks of cofactors ('c', 'l', 'm';
#     work internal to mersServer.pl, so not listed here), then P+1, P-1, ECM, etc.
#   But checking factors, 'C', is done as they are read if they are for "small" Mersenne
#     numbers (exponents no larger than the first with no cofactor SPRP checks), also only by
#     the server, so 'C' is not mentioned here.
#   'X', 1strs work, is also only done via the server, by a child running $oneStrsExe.
#   'qoE' can be done in parallel/rotation/etc., so list them together, but the current
#     implemention only allows one task per exponent; see mersServer.pl's findWork().
# This is used in m// pattern matching and a default set of allowed work types in mersServer.pl
#   and in save.pl to validate the allowed work field of the 'allowed' SQL table.
our ($workLetters) = 'S' . 'qoE' . 'Gg' . 'HU';

# Maximum stage 1 bound for ECM, P-1, and P+1 based on a limitation of the ecm program from 'man ecm'
#   (though this value is slightly smaller, from an error from the program itself in 2019).
our ($maxStage1Bound) = Math::BigInt->new('9007199254740991');

# new(): create an object, usually by passing in 'n' and its value like:
#   my ($m) = Mersenne->new('n', 127);
# If 'n' is missing, the number will be used as the exponent anyway.
# Other supported keys are CDEGHJKLMPSUXcdeglmoq - the 'mers letters' of the 'mers format';
#     the value is saved in $self->{$letter} unless noted:
# C - a factor of M('n'); can be several, each factor is stored in @{$self->'factors'} rather than $self->{'C'}.
#     The cofactor(s) calculated from M('n') and the known factors are stored in @{$self->{'cofactors'}}.
# D - 1 if M('n') is completely factored, $self->{'D'} should not exist otherwise.
# E1 - the ECM effort (sum of stage 1 bound for all ECM curves done) ...
# E2 - ... and largest stage 1 bound for all ECM curves.
# H - the trial factor extent (except for gaps) if a factor is known or an LL or PRP test is complete (only for prime 'n').
# L - digit count of the first cofactor that passed an SPRP check.
# M - digit count of the second cofactor that passed an SPRP check.
# P - 1 if M('n') is prime - a Mersenne prime; $self->{'P'} should not exist otherwise.
# S - digit count of cofactor that passed an SPRP check.
# U - trial factoring extent (except for gaps) if no factor is known and no LL or PRP test is done (only for prime 'n').
# X - extent of 1strs effort, which finds the Mersenne exponent for each factor in factor order; see mergeX().
# c - digit count of composite cofactor when 'n' is not an odd multiple of 4.
# d - old letter for 'S'; changed to $self->{'S'}.
# e - stage 1 bound and count of curves done at that bound; merged into 'E' data.  $self->{'e'} should never exist.
# l - digit count of composite L half of cofactor when 'n' is an odd multiple of 4 (see sub y()).
# m - digit count of composite M half of cofactor when 'n' is an odd multiple of 4 (see sub z()).
# n - the exponent itself.
# o1 - the largest stage 1 bound of P-1 factoring ...
# o2 - ... and the largest stage 2 bound at that stage 1 bound.
# q1 - the largest stage 1 bound of P+1 factoring ...
# q2 - ... and the largest stage 2 bound at that stage 1 bound.
# These last three are not retained due to the amount of data and are only for prime 'n':
# G - the start and stop of a gap in trial factoring.
# J - the start and stop of a 'join' in trial factoring.
# g - gap in trial factoring from 2^(value1) to 2^(value2).
sub new
{
  my $class = shift;
  my $self = {};
  local $_;
  my ($key);
  while (@_)
    {
      $key = shift;
      last unless @_;
      return(undef) if ref($key); # do not allow references to be keys
      ${$self}{$key} = shift;
      $key = undef;
    }
  if (defined($key))
    { # odd number of parameters given
      if ($key =~ m/^([1-9][0-9]*)$/ && !exists($self->{'n'}))
        { # a number - the exponent
          $self->{'n'} = Math::BigInt->new("$1");
        }
      elsif ($key =~ m/^M[(] /)
        { # mers format - parse it later
          $self->{'mersfmt'} .= "\n" . $key;
        }
      else
        { # unknown - fail
          return(undef);
        }
    }
  delete($self->{'errors'})
    if exists $self->{'errors'};
  bless $self, $class;
  $self->parse()
    if exists $self->{'mersfmt'} && $self->{'mersfmt'};
  exists($self->{'n'})
    or die "Mersenne->new() called without value of the exponent, 'n'\n";
  if (ref($self->{'n'}) ne 'Math::BigInt')
    {
      $self->{'n'} =~ m/^([1-9][0-9]*)$/
        or die "Mersenne->new() called with non-numeric value for the exponent, 'n'\n";
      $self->{'n'} = Math::BigInt->new("$1");
    }
  my (@factors);
  foreach my $key ('factors', 'cofactors', )
    {
      next unless exists $self->{$key};
      delete($self->{$key}),
        next
        unless ref($self->{$key}) eq 'ARRAY';
      foreach (@{$self->{$key}})
        {
          push(@factors, Math::BigInt->new("$1")),
            next
            if !ref($_) && m/^([1-9][0-9]*)$/;
          next unless ref($_) eq 'Math::BigInt';
          push(@factors, $_);
        }
      delete($self->{$key});
    }
  $self->f(@factors)
    if (@factors);
  return $self if $self->check();
  return undef;
}

sub errors
{
  my ($self) = shift;
  return('')
    unless exists $self->{'errors'};
  delete $self->{'errors'},
    return('')
    unless (ref($self->{'errors'}) eq 'ARRAY'
            && $#{$self->{'errors'}} >= 0);
  return(join("\n", @{$self->{'errors'}}) . "\n")
}

# Check the data for problems.  Always returns $self unless no 'n'.
#   Adds problems to $self->{'errors'}.
sub check
{
  my ($self) = shift;
  push(@{$self->{'errors'}}, "no or invalid exponent 'n'"),
    return undef
    unless exists $self->{'n'} && $self->{'n'} =~ m/^[1-9][0-9]*$/;
  my (@factors) = $self->f();
  delete($self->{'P'})
    if exists($self->{'P'}) && @factors;
  $self->{'S'} = $self->{'d'},
    delete($self->{'d'})
    if $self->{'d'}; # 'M( n )d' is an older format for current 'M( n )S'
  if ($self->{'n'} % 8 == 4)
    { # lmLM cofactors, not Sc; see y() (lL) and z() (mM).
      delete($self->{'c'});
      delete($self->{'S'});
      # 'n' is obviously composite, so the following are not relevant
      delete($self->{'G'});
      delete($self->{'g'});
      delete($self->{'H'});
      delete($self->{'J'});
      delete($self->{'U'});
      delete($self->{'P'});
    }
  else
    { # S or c cofactor, not lmLM
      delete($self->{'c'})
        if $self->{'S'} && $self->{'c'} && $self->{'S'} < $self->{'c'};
      delete($self->{'S'})
        if $self->{'S'} && $self->{'c'} && $self->{'S'} >= $self->{'c'};
      delete($self->{'l'});
      delete($self->{'m'});
      delete($self->{'L'});
      delete($self->{'M'});
      # These last three are not retained.
      delete($self->{'G'});
      delete($self->{'g'});
      delete($self->{'J'});
    }
  foreach (grep { exists $self->{$_} } ('S', 'L', 'M', 'c', 'l', 'm'))
    {
      ($self->{$_} = $self->{'n'}),
        next
        if $self->{$_} < 2;
      delete($self->{$_})
        if $self->{$_} < $self->cofactorLength($_);
    }
  if ($self->{'P'})
    { # Prime - M(n) is a Mersenne prime; nothing else is needed and it being prime is too hard to confirm
      delete($self->{$_})
        foreach (grep { !m/^[nP]$/ } keys(%{$self}));
      return $self;
    }
  if ($self->{'D'} || $self->{'S'})
    { # Done (all factors proven prime) or SPRP (all factors have passed SPRP tests)
      #   - only the factors and cofactors might be needed.
      # In older version of this code, 'S' did not have a digit count, so treat 1 as the exponent.
      ($self->{'S'} = $self->{'n'})
        if $self->{'S'} && $self->{'S'} < 2;
      delete($self->{'S'})
        if $self->{'D'} || $self->{'S'} > $self->{'n'};
      if ($self->{'D'} || $self->{'S'})
        {
          delete($self->{$_})
            foreach (grep { !m/^[nDS]$/ && !m/^(co)?factors$/ } keys(%{$self}));
          return($self);
        }
    }
  if ($self->{'L'})
    { # digits in first cofactor, SPRP
      # Previously, 'L' did not have a digit count, so treat 1 as the exponent.
      ($self->{'L'} = $self->{'n'})
        if $self->{'L'} < 2;
      delete($self->{'L'})
        if (ref($self->{'L'}) ne 'Math::BigInt' || $self->{'L'} > $self->{'n'}
            || (exists($self->{'l'}) && $self->{'l'} < $self->{'L'} && $self->{'l'} > 1));
      delete($self->{'l'})
        if exists $self->{'L'};
    }
  if ($self->{'M'})
    { # digits in second cofactor, SPRP
      # Previously, 'M' did not have a digit count, so treat 1 as the exponent.
      ($self->{'M'} = $self->{'n'})
        if $self->{'M'} < 2;
      delete($self->{'M'})
        if (ref($self->{'M'}) ne 'Math::BigInt' || $self->{'M'} > $self->{'n'} || $self->{'M'} < 2
            || (exists($self->{'m'}) && $self->{'m'} < $self->{'M'} && $self->{'m'} > 1));
      delete($self->{'m'})
        if exists $self->{'M'};
    }
  if ($self->{'c'})
    { # digits in the cofactor, composite per SPRP test
      delete($self->{'c'})
        if ref($self->{'c'}) ne 'Math::BigInt' || $self->{'c'} > $self->{'n'};
    }
  if ($self->{'l'})
    { # digits in first cofactor, composite per SPRP test
      delete($self->{'l'})
        if ref($self->{'l'}) ne 'Math::BigInt' || $self->{'l'} > $self->{'n'};
    }
  if ($self->{'m'})
    { # digits in second cofactor composite per SPRP test
      delete($self->{'m'})
        if ref($self->{'m'}) ne 'Math::BigInt' || $self->{'m'} > $self->{'n'};
    }
  if (exists($self->{'E1'}))
    { # ECM effort and largest stage 1 bound
      delete($self->{'E1'})
        if ref($self->{'E1'}) ne 'Math::BigInt';
      delete($self->{'E2'})
        if exists($self->{'E2'}) && ref($self->{'E2'}) ne 'Math::BigInt';
    }
  if (exists($self->{'o1'}))
    { # P-1 largest stage 1 bound and any stage 2 bound at that stage 1 bound
      delete($self->{'o1'})
        if ref($self->{'o1'}) ne 'Math::BigInt';
      delete($self->{'o2'})
        if exists($self->{'o2'}) && ref($self->{'o2'}) ne 'Math::BigInt';
    }
  if (exists($self->{'q1'}))
    { # P+1 largest stage 1 bound and any stage 2 bound at that stage 1 bound
      delete($self->{'q1'})
        if ref($self->{'q1'}) ne 'Math::BigInt';
      delete($self->{'q2'})
        if exists($self->{'q2'}) && ref($self->{'q2'}) ne 'Math::BigInt';
    }
  if (exists($self->{'U'}))
    { # trial factoring extent, no known factors and no completed primality test
      delete($self->{'U'})
        if ref($self->{'U'}) ne 'Math::BigInt';
      $self->{'H'} = $self->{'U'},
        delete($self->{'U'})
        if (exists($self->{'U'})
            && ((exists($self->{'H'}) && $self->{'H'} < $self->{'U'})
                || (@factors && (!exists($self->{'H'}) || $self->{'H'} < $self->{'U'}))));
    }
  if (exists($self->{'H'}))
    { # trial factoring extent, M(n) known to be composite
      delete $self->{'H'}
        if ref($self->{'H'}) ne 'Math::BigInt';
    }
  return($self)
    unless exists($self->{'X'});
  # 1strs extent or, when the value is 0, the largest exponent in sequence that is completely factored
  delete($self->{'X'}),
    return($self)
    if ref($self->{'X'}) ne 'Math::BigInt';
  foreach my $k (grep { !m/^[nX]$/ } keys(%{$self}))
    {
      push(@{$self->{'errors'}}, "'X' data should not be mixed with other data in one Mersenne instance: $_\n");
      last;
    }
  return($self);
}

# Return the exponent.
sub n
{
  my ($self) = shift;
  return $self->{'n'};
}

# Return the info for the given mers letter, including 'n' for the exponent.
sub get
{
  my ($self) = shift;
  my ($letter) = shift;
  my (@values);
  return $self->{'n'} if $letter eq 'n' && !wantarray();
  return(($self->{'n'}, )) if $letter eq 'n' && wantarray();
  return undef unless $letter =~ m/^([CcDEGgHJLlMmoPqSUX])$/;
  return($self->f()) if $letter eq 'C';
  if ($letter =~ m/^[DLMPSclmUHX]$/)
    {
      return($self->{$letter}) if defined $self->{$letter} && !wantarray();
      push(@values, $self->{$letter}) if defined $self->{$letter};
    }
  elsif ($letter =~ m/^[EoqGJg]$/ && exists($self->{$letter . '1'}))
    {
      return($self->{$letter . '1'}) unless wantarray();
      push(@values, $self->{$letter . '1'});
      push(@values, $self->{$letter . '2'})
        if exists($self->{$letter . '2'});
    }
  return(@values);
}

# Returns M(n) (= 2^n - 1) for the 'n' given to new().
sub m # 'm'ersenne; M('n') for any 'n'
{
  my $self = shift;
  return(Math::BigInt->bone()->blsft($self->{'n'})->bsub(1));
}

# Returns (2^n) % p without using values larger than p*p.
#   If p divides M(n), this will return 1 and save p as a factor.
sub r # 'r'emainder; M('n') % p for any 'n' and p
{
  my ($self, $p) = @_;
  my ($r) = Math::BigInt->bone();
  my ($n) = $self->{'n'};
  my ($s) = 2*$r;
  while ($n > 0)
    {
      $r *= $s if $n % 2 == 1;
      $s *= $s;
      $n /= 2;
      $r %= $p if $r > $p;
      $s %= $p if $s > $p;
    }
  $self->f($p) if $r == 1 && exists $self->{'cofactors'};
  return($r);
}

# Returns the L part of M(n) when n = 8*a - 4.
sub y
{
  my ($self) = @_;
  return 0 if $self->{'n'} % 8 != 4;
  my ($a) = $self->{'n'} + 4;
  $a /= 8;
  return(Math::BigInt->bone()->blsft(2*$a - 1) - Math::BigInt->bone()->blsft($a) + 1);
}

# Returns the M part of M(n) when n = 8*a - 4.
sub z
{
  my ($self) = @_;
  return 0 if $self->{'n'} % 8 != 4;
  my ($a) = $self->{'n'} + 4;
  $a /= 8;
  return(Math::BigInt->bone()->blsft(2*$a - 1) + Math::BigInt->bone()->blsft($a) + 1);
}

# Returns the cofactor(s) of M(n) after algebraic and known factors have been pulled out.
#   Known factors are actually pulled out by calling f(), not this, but f() updates
#   the cofactors while determining whether the arguments given to f() are factors.
sub c # 'c'ofactor(s) - mers letters 'c', 'l', 'm', 'D', 'S', 'L', 'M'
{
  my ($self) = @_;
  local $_;
  if (!exists($self->{'cofactors'}))
    {
      if ($self->{'n'} % 8 != 4)
        {
          push(@{$self->{'cofactors'}}, $self->m());
        }
      else
        {
          push(@{$self->{'cofactors'}}, $self->y());
          push(@{$self->{'cofactors'}}, $self->z());
        }
      foreach my $p (Math::Prime::Util::divisors($self->{'n'}))
        { # If $p divides n evenly, then M($p) divides M(n) evenly.
          my ($i) = Math::BigInt::bgcd($self->{'cofactors'}[0], Math::BigInt->bone()->blsft($p) - 1);
          while ($i > Math::BigInt->bone() && $i < $self->{'cofactors'}[0])
            {
              $self->{'cofactors'}[0] /= $i;
              $i = Math::BigInt::bgcd($self->{'cofactors'}[0], Math::BigInt->bone()->blsft($p) - 1);
            }
          next if $#{$self->{'cofactors'}} == 0;
          $i = Math::BigInt::bgcd($self->{'cofactors'}[1], Math::BigInt->bone()->blsft($p) - 1);
          while ($i > Math::BigInt->bone() && $i < $self->{'cofactors'}[1])
            {
              $self->{'cofactors'}[1] /= $i;
              $i = Math::BigInt::bgcd($self->{'cofactors'}[1], Math::BigInt->bone()->blsft($p) - 1);
            }
        }
    }
  return(@{$self->{'cofactors'}});
}

# Return the number of digits in one of the cofactors.
#   If 'n' % 8 != 4, the digits of 'c' cofactor.
#   Otherwise, the digits of the 'l' or 'm' cofactor as specified by the only argument.
sub cofactorLength
{
  my ($self) = shift;
  my ($letter) = shift;
  $letter = 'c' if !$letter || $self->{'n'} % 8 != 4;
  $self->c() if !exists $self->{'cofactors'};
  my ($cnt) = $self->{'cofactors'}[$letter =~ m/^[mM]$/ ? 1 : 0]->length();
  return($cnt);
}

# Returns the known non-algebraic factors of M('n') ...
# ... given as arguments to this call or prior calls of this same function
# ... or found via GCD with cofactors or with factors given to prior calls.
sub f
{
  my ($self) = shift;
  local $_;
  my (@args) = sort { $a <=> $b } grep { $_ && m,^([1-9][0-9]*)$, && $self->{'n'} < $1 } @_;
  $self->c() if @args && !exists $self->{'cofactors'};
  my (@f) = sort { $a <=> $b } @{$self->{'factors'}}
    if exists $self->{'factors'};
  my (@newFactors);
  my ($removed) = 0; # count of composite factors in $self->{'factors'} removed
  while (@args || @newFactors)
    {
      push(@newFactors, @args),
        @args = (sort { $a <=> $b } @newFactors),
        @newFactors = ()
        if @newFactors;
      my $arg = shift(@args);
      $arg = Math::BigInt->new("$arg") unless ref($arg) eq 'Math::BigInt';
      next unless $arg % 2 == 1 && $arg > $self->{'n'} && $arg % $self->{'n'} == 1;
      foreach my $fac (@f)
        { # pull known factors out of $arg
          last if $arg < $fac;
          $arg /= $fac
            while $arg % $fac == 0;
        }
      next unless $arg > 1; # nothing left; try the next one
      my ($i) = 0;
      my ($gcd);
      while ($i <= $#f)
        { # GCD with each known factor
          $gcd = $arg->bgcd($f[$i]);
          ++$i, next unless $gcd > 1;
          # $f[$i] is a composite factor rather than a prime factor
          $self->{'cofactors'}[0] *= $f[$i]; # multiply it back into the (first) cofactor
          $f[$i] /= $gcd;
          push(@newFactors, $gcd, $f[$i]);
          splice(@f, $i, 1); # remove $f[$i] from the list of factors
          ++$removed;
          $arg /= $gcd;
        }
      next unless $arg > 1;
      while ($self->{'cofactors'}[0] % $arg == 0)
        {
          $arg = 1, last unless $self->{'cofactors'}[0] > $arg;
          $self->{'cofactors'}[0] /= $arg;
          if ($self->{'cofactors'}[0] > $arg)
            {
              @f = sort { $a <=> $b } (@f, $arg);
              next;
            }
          # The "factor" is larger; swap with the cofactor
          @f = sort { $a <=> $b } (@f, $self->{'cofactors'}[0]);
          $self->{'cofactors'}[0] = $arg;
          $arg = 1;
          last;
        }
      next unless $arg > 1;
      $gcd = $arg->bgcd($self->{'cofactors'}[0]);
      if ($gcd > 1)
        {
          $arg = 1, last unless $self->{'cofactors'}[0] > $gcd;
          @f = sort { $a <=> $b } (@f, $gcd);
          $self->{'cofactors'}[0] /= $gcd;
          $arg /= $gcd;
          if ($arg > $self->{'cofactors'}[0])
            { # New "factor" is larger than the "cofactor" - swap them.
              @f = sort { $a <=> $b } (@f, $self->{'cofactors'}[0]);
              $self->{'cofactors'}[0] = $arg;
              $arg = 1;
            }
        }
      next unless $arg > 1;
      next if $#{$self->{'cofactors'}} < 1;
      while ($self->{'cofactors'}[1] % $arg == 0)
        {
          $arg = 1, last unless $self->{'cofactors'}[1] > $arg;
          $self->{'cofactors'}[1] /= $arg;
          if ($self->{'cofactors'}[1] > $arg)
            {
              @f = sort { $a <=> $b } (@f, $arg);
              next;
            }
          # The "factor" is larger; swap with the cofactor
          @f = sort { $a <=> $b } (@f, $self->{'cofactors'}[1]);
          $self->{'cofactors'}[1] = $arg;
          $arg = 1;
          last;
        }
      next unless $arg > 1;
      $gcd = $arg->bgcd($self->{'cofactors'}[1]);
      next unless $gcd > 1;
      next unless $self->{'cofactors'}[1] > $gcd;
      @f = sort { $a <=> $b } (@f, $gcd);
      $self->{'cofactors'}[1] /= $gcd;
      $arg /= $gcd;
      if ($arg > $self->{'cofactors'}[1])
        { # New "factor" is larger than the "cofactor" - swap them.
          push(@newFactors, $self->{'cofactors'}[1]);
          $self->{'cofactors'}[1] = $arg;
          next;
        }
      next unless $arg > 1;
      push(@newFactors, $arg);
    }
  @{$self->{'factors'}} = @f
    if ((@f || $removed) && (!exists($self->{'factors'}) || $#{$self->{'factors'}} < $#f));
  if (exists($self->{'factors'}) && ref($self->{'factors'}) eq 'ARRAY')
    {
      @f = sort { $a <=> $b } @{$self->{'factors'}};
      @{$self->{'factors'}} = @f;
      return(@{$self->{'factors'}})
    }
  @f = ();
  return(@f)
    unless exists $self->{'factors'} && exists $self->{'cofactors'};
  push(@{$self->{'errors'}}, "internal problem - 'factors' or 'cofactors' not created");
  delete $self->{'factors'};
  delete $self->{'cofactors'};
  return(@f);
}

# Parse the string argument and add to the internal data of M('n') appropriately.
#   Currently only supports 'mers format'.
sub parse
{
  my ($self) = shift;
  my ($data) = shift;
  $data = $_ unless $data;
  local $_;
  push(@{$self->{'errors'}}, "Mersenne::parse does not support references: " . ref($data)),
    return(0)
    if ref($data) ne '';
  $data = '' unless defined $data;
  ($data .= "\n" . $self->{'mersfmt'})
    if exists $self->{'mersfmt'} && $self->{'mersfmt'};
  delete($self->{'mersfmt'})
    if exists $self->{'mersfmt'};
  return(1)
    unless $data;
  delete($self->{'errors'})
    if exists($self->{'errors'}) && ref($self->{'errors'}) ne 'ARRAY';
  my ($errCount) = 0;
  ($errCount = 1 + $#{$self->{'errors'}})
    if exists $self->{'errors'};
  $data =~ s,[\r]+,\n,g;
  my (@lines) = split(/[\n]+/, $data);
  foreach (@lines)
    {
      next
        if m/^$/ || m/^[#]/; # allow empty lines and comments
      push(@{$self->{'errors'}}, "unexpected data format: $_"),
        last
        unless m/^M[(] ([1-9][0-9]*) [)]([CcDEeGHJLlMmoPqSU]):? *([0-9]*) *([0-9]*)/;
      my ($one) = Math::BigInt->new("$1");
      my ($two) = "$2";
      my ($three) = Math::BigInt->new("$3")
        if defined($3) && $3 ne '';
      my ($four) = Math::BigInt->new("$4")
        if defined($4) && $4 ne '';
      $self->{'n'} = $one
        unless exists $self->{'n'};
      push(@{$self->{'errors'}}, "multiple exponents - $self->{'n'} and $one"),
        last
        unless $self->{'n'} == $one;
      next
        if $two =~ m/^[GJ]$/; # don't retain trial factoring gaps or joins
      $self->{$two} = 1,
        next
        if $two =~ m/^[DP]$/; # no digit counts - all (primitive) factors proven prime or Mersenne prime
      if ($two =~ m/^[LMS]$/ && !$three)
        { # allow lack of digit counts - old format
          ($self->{$two} = $self->{'n'})
            if !$self->{$two};
          next;
        }
      push(@{$self->{'errors'}}, "missing value after mers letter: $_"),
        last
        unless defined $three;
      $self->f($three),
        next
        if $two eq 'C';
      if (exists($self->{$two}) || exists($self->{$two . '1'}))
        { # Retain lower cofactor digit counts (except '1' or less).
          $self->{$two} = $three,
            next
            if $two =~ m/^[clmLMS]$/ && ($self->{$two} > $three || $self->{$two} < 2);
          # Retain larger trial factoring extents.
          $self->{$two} = $three,
            next
            if $two =~ m/^[UH]$/ && $self->{$two} < $three;
          # Done with those two kinds of data.
          next
            if $two =~ m/^[cDHlmLMSU]$/;
          if ($two =~ m/^[oq]$/)
            { # P-1 and P+1 stage one and stage two bounds - retain larger.
              if (!exists($self->{$two . '1'}) || $self->{$two . '1'} < $three)
                {
                  $self->{$two . '1'} = $three;
                  $self->{$two . '2'} = $four
                    if defined $four;
                  delete($self->{$two . '2'})
                    unless defined $four;
                  next;
                }
              next
                if $self->{$two . '1'} > $three || !defined($four);
              $self->{$two . '2'} = $four
                if !exists($self->{$two . '2'}) || $self->{$two . '2'} < $four;
              next;
            }
          if ($two eq 'E')
            { # Effort (sum of bound times curves) and largest bound.
              # Do not assume the efforts are independent - retain the largest.
              $self->{'E1'} = $three
                if $self->{'E1'} < $three;
              next
                unless defined $four;
              $self->{'E2'} = $four
                if !exists($self->{'E2'}) || $self->{'E2'} < $four;
              next;
            }
          push(@{$self->{'errors'}}, "logic error while parsing additional data: $_");
          next;
        }
      # Nothing prior matches this new data, so save it.
      $self->{$two} = $three,
        next
        if $two =~ m/^[cDHlmLMPSU]$/;
      if ($two eq 'e')
        { # Stage 1 bound and curves at that bound - convert to 'E'ffort and add it in.
          next
            unless $four;
          $self->{'E1'} = Math::BigInt->bzero()
            unless exists $self->{'E1'};
          $self->{'E2'} = Math::BigInt->bzero()
            unless exists $self->{'E2'};
          $self->{'E1'} += $three * $four;
          $self->{'E2'} = $three
            if $self->{'E2'} < $three;
          next;
        }
      $four = Math::BigInt->bzero()
        unless $four;
      $self->{$two . '1'} = $three,
        $self->{$two . '2'} = $four,
        next
        if $two =~ m/^[Eoq]$/;
      push(@{$self->{'errors'}}, "logic error while parsing data: $_");
    }
  $self->check();
  return(1)
    unless exists($self->{'errors'});
  return($errCount == 1 + $#{$self->{'errors'}});
}

# Print the data in mers format to a given file handle (default STDOUT).
sub print
{
  my ($self) = shift;
  my ($fh) = shift;
  $fh = *STDOUT
    unless defined $fh;
  my ($str) = $self->str();
  return undef
    unless $str;
  return(print($fh $str));
}

# Return a string in mers format of all the info.
#   The lines should be in the same order as this ASCII based sort:
#     sort -k 2,2n -k 3,3 -k 4,4n -k 5,5n -k 6,6n -k 1
#   The most important field in this context is the mers letter.
#   Also important is that capital letters sort before lower case letters.
#     The primary reason for that is that the factors (mers letter 'C')
#     come first, before any other data, so other data can be checked
#     against the known factors when appropriate.
sub str
{
  my ($self) = shift;
  my ($data) = '';
  local $_;
  my ($n) = $self->{'n'};
  my (@values) = $self->f();
  $data .= "M( $n )C: $_\n"
    foreach (@values);
  foreach ('P', 'D')
    {
      $data .= "M( $n )$_\n",
        return($data)
        if $self->{$_};
    }
  @values = $self->get('E');
  foreach ('E', 'H', 'L', 'M', 'S', 'U', 'c', 'l', 'm', 'o', 'q')
    {
      @values = $self->get($_);
      next unless @values;
      $data .= "M( $n )$_\n", next if m,^[cDlmLMPS]$, && $values[0] < 2;
      $data .= "M( $n )$_" . ": " . join(' ', @values) . "\n";
    }
  return($data);
}

# Merge self (must be X data) with other (argument data, any mers letter).
#   As for merge(), return $self if $other is "subsumed" by $self (perhaps after modifying $self); return 0 otherwise.
#   $self and $other might be modified.
sub mergeX
{ # M( 1206 )X: 0 # means all Mersennes smaller than M(1207) are completely factored (either a prime or all factors are SPRP)
  # M( 18446744073709551615 )X: 17752425177 # means all factors <= the value are known for Mersennes with that or a smaller exponent
  my ($self) = shift;
  my ($other) = shift;
  return(0) unless ref($self) eq ref($other);
  die "Mersenne::mergeX() called when self is not 'X' data"
    unless exists $self->{'X'};
  if (!exists($other->{'X'}))
    {
      $self->{'n'}++,
        return($self)
        if $self->{'n'} + 1 == $other->{'n'} && ($other->{'P'} || $other->{'D'} || $other->{'S'});
      if ($self->{'X'} == 0)
        {
          return($self) if $other->{'n'} <= $self->{'n'};
          return(0);
        }
      return(0) if $self->{'n'} < $other->{'n'};
      ($other->{'H'} = $self->{'X'})
        if exists($other->{'H'}) && $other->{'H'} < $self->{'X'};
      ($other->{'U'} = $self->{'X'})
        if exists($other->{'U'}) && $other->{'U'} < $self->{'X'};
      # Check $other GJg here if those are retained.
      return(0);
    }
  return($self) if $self->{'n'} >= $other->{'n'} && $self->{'X'} >= $other->{'X'};
  $self->{'n'} = $other->{'n'},
    $self->{'X'} = $other->{'X'},
    return($self)
    if $self->{'n'} < $other->{'n'} && $self->{'X'} < $other->{'X'};
  return(0);
}

# Merge the data in $other into $self.
#   Return 0 if $other is still useful - some or all of its data is not in $self.
#   Return $self if $other is no longer useful - all of its data is now in $self.
#   $self and $other might (both) be modified.
#   Do not call this if $self or $other has not been check()d or has 'errors'.
sub merge
{
  my ($self) = shift;
  my ($other) = shift;
  return(0) unless ref($self) eq ref($other);
  return($self->mergeX($other)) if exists $self->{'X'};
  die "Mersenne::merge() called with argument 'X' data when self is not 'X' data"
    if exists $other->{'X'};
  return(0) unless $self->{'n'} == $other->{'n'};
  $self->f(@{$other->{'factors'}}), delete($other->{'factors'}) if exists $other->{'factors'};
  delete($other->{'cofactors'}) if exists $other->{'cofactors'};
  local $_;
  foreach ('P', 'D', 'S', )
    {
      return($self) if exists($self->{$_});
      $self->{$_} = $other->{$_}, return($self->check()) if exists $other->{$_};
    }
  foreach ('c', 'l', 'm', 'L', 'M')
    {
      next unless exists $other->{$_};
      ($self->{$_} = $other->{$_}) if !exists($self->{$_}) || $self->{$_} > $other->{$_};
    }
  if (exists($other->{'E1'}))
    {
      $self->{'E1'} = $other->{'E1'} if !exists($self->{'E1'}) || $self->{'E1'} < $other->{'E1'};
      $self->{'E2'} = $other->{'E2'} if exists($other->{'E2'}) && (!exists($self->{'E2'}) || $self->{'E2'} < $other->{'E2'});
    }
  foreach ('o', 'q')
    {
      next unless exists $other->{$_ . '1'};
      next if exists $self->{$_ . '1'} && $self->{$_ . '1'} > $other->{$_ . '1'};
      $self->{$_ . '1'} = $other->{$_ . '1'};
      next unless exists $other->{$_ . '2'};
      next if exists($self->{$_ . '2'}) && $self->{$_ . '2'} >= $other->{$_ . '2'};
      $self->{$_ . '2'} = $other->{$_ . '2'};
    }
  if (exists($other->{'H'}))
    {
      ($self->{'H'} = $other->{'H'}) if !exists($self->{'H'}) || $self->{'H'} < $other->{'H'};
      ($self->{'H'} = $self->{'U'}) if exists($self->{'U'}) && $self->{'H'} < $self->{'U'};
    }
  if (exists($other->{'U'}))
    {
      ($self->{'U'} = $other->{'U'}) if !exists($self->{'H'}) && (!exists($self->{'U'}) || $self->{'U'} < $other->{'U'});
      ($self->{'H'} = $other->{'U'}) if exists($self->{'H'}) && $self->{'H'} < $other->{'U'};
    }
  return($self->check());
}

sub needsSPRP
{
  my ($self) = shift;
  my (@cofactors) = $self->c();
  my ($len) = $cofactors[0]->length();
  my (@values);
  my (%digits); # values is digits of cofactor; key is 'c', 'l', 'm', 'S', 'L', or 'M'
  if ($self->n() % 8 != 4)
    {
      $digits{'c'} = $self->get('c');
      $digits{'c'} = $self->n()
        unless $digits{'c'};
      $digits{'S'} = $self->get('S');
      delete($digits{'c'})
        if $digits{'c'} && $digits{'S'} && $digits{'c'} > $digits{'S'};
      delete($digits{'S'})
        if $digits{'c'} && $digits{'S'};
      return(1) if $digits{'c'} && $len != $digits{'c'};
      return(1) if $digits{'S'} && $len != $digits{'S'};
      return(0);
    }
  $digits{'l'} = $self->get('l');
  $digits{'l'} = $self->n()
    unless $digits{'l'};
  $digits{'m'} = $self->get('m');
  $digits{'m'} = $self->n()
    unless $digits{'m'};
  $digits{'L'} = $self->get('L');
  $digits{'M'} = $self->get('M');
  delete($digits{'l'})
    if $digits{'l'} && $digits{'L'} && $digits{'l'} > $digits{'L'};
  delete($digits{'m'})
    if $digits{'m'} && $digits{'M'} && $digits{'m'} > $digits{'M'};
  delete($digits{'L'})
    if $digits{'l'} && $digits{'L'};
  delete($digits{'M'})
    if $digits{'m'} && $digits{'M'};
  return(1) if $digits{'l'} && $len != $digits{'l'};
  return(1) if $digits{'L'} && $len != $digits{'L'};
  $len = $cofactors[1]->length();
  return(1) if $digits{'m'} && $len != $digits{'m'};
  return(1) if $digits{'M'} && $len != $digits{'M'};
  return(0);
}

# Calculate the priority of the work as an array of mers format.
#   Not a member function/sub - only looks at the arguments.
sub mersPriority(@)
{
  my (@mers) = @_;
  # @mers could be only the exponent and 'S', so:
  return(-1.0) unless $#mers >= 1;
  return(-6.0)
    unless $mers[0] =~ m,^([1-9][0-9]+)$,;
  $mers[0] = $1;
  # clm tasks (SPRP checks) are now done in exponent order as internal work in mersServer.pl.
  #   LMS primality proofs of the factor(s).
  return("$mers[0]" . '.0')
    if $mers[1] =~ m/^[)]?([clmLMS])[:]?$/;
  # Some callers probably include the close paren and colon in $mers[1], which are not expected here.
  #   L and M lines are OK if both are present for the same exponent - presume that here.
  return(-2.0)
    unless $mers[1] =~ m/^([)]?[$workLetters][:]?)$/;
  $mers[1] = $1;
  return(-4.0)
    if $mers[0] <= 1200; # completely factored
  my ($pri);
  if ($mers[1] =~ m/^[)]?[GgUH][:]?$/)
    { # For trial factoring gaps or extent, log(count of possible trial factors) and break ties.
      return(-3.0) unless $#mers >= 2 && $mers[2];
      my ($start) = ($mers[1] !~ m/g/) ? $mers[2] : Math::BigInt->bone()->blsft($mers[2]);
      # All factors are larger than the exponent.
      $start = $mers[0] + 1
        if $start <= $mers[0];
      # All factors of odd exponent Mersennes are larger than twice the exponent.
      $start += $mers[0]
        if $mers[0] % 2 && $start <= 2*$mers[0];
      # For UH, calculate the priority as if a gap to the next power of 2.
      $mers[3] = Math::BigInt->bone()->blsft(length($mers[2]->to_bin()))
        if $mers[1] =~ m/^[)]?[UH][:]?$/;
      my ($stop) = ($mers[1] !~ m/g/
                    ? (ref($mers[3]) eq 'Math::BigInt' ? $mers[3]->copy() : Math::BigInt->new("$mers[3]"))
                    : Math::BigInt->bone()->blsft("$mers[3]"));
      $#mers = 2 if $mers[1] =~ m/^[)]?[UH][:]?$/; # wipe the added entry
      $stop->bsub($start);
      return(-7.0) if $stop < 0; # no actual gap - start exceeds stop
      $stop += 2*$mers[0] if $stop < 1; # increase an "empty" gap before calculating the priority
      $pri = $stop . '.0'; # convert to floating point
      $pri /= 4000000.0; # constant to reduce size of final value
      $pri /= ($mers[0] . '.0');
      $pri = log($pri + 1.5) + 1.0 - 1000.0/($start . '.0') - 1000.0/($mers[0] . '.0'); # break ties, force floating point
      $pri *= log($mers[0] . '.0') if $mers[0] > Math::BigInt->bone()->blsft(29);
      $pri *= log($mers[0] . '.0') if $mers[0] > Math::BigInt->bone()->blsft(58);
      # For UH, multiply the priority by the exponent of 2 to prefer gaps.
      $pri *= length($mers[2]->to_bin())
        if $mers[1] =~ m/^[)]?[UH][:]?$/;
      return($1)
        if $pri =~ m,$floatPattern,;
      return(-8.0);
    }
  return(-10.0)
    unless $mers[1] =~ m/^[)]?[Eoq][:]?$/;
  # For Eoq, log(exponent + largest (effort or stage1 bound))
  my ($bnd) = $#mers >= 2 && $mers[2] ? $mers[2] : 0;
  $bnd = 0
    unless $bnd;
  $bnd = "$bnd" . '.5';
  $bnd -= 1000.0/($mers[0] . '.0'); # break ties and force floating point
  return(log(("$mers[0]" . '.0') + $bnd));
}

1;
