#!/usr/bin/perl -Tw
# Ask factordb.com for the known factors of a Mersenne number.

use strict;
use strict 'vars';
use warnings;
use HTTP::Tiny;
use Math::BigInt try => 'GMP';

BEGIN { push(@INC, '.'); $ENV{'PATH'} = '/usr/bin:/bin'; }
use Mersenne;
use MersInfo;

my ($usage) = ("Usage: askFactordb.pl [--check] [--version] [-h] [--help] [-d [debugLevel]] [-c] [--certs] <exponent>\n"
               . "  --check\n    check the configuration and command line\n"
               . "  --version\n    print the version and exit\n"
               . "  -h or --help\n    print this usage message and exit\n"
               . "  --debug [level]\n    print more info, mostly about progress\n"
               . "  -c or --certs\n    only retrieve primality certificates\n"
               . "  <exponent>\n    the exponent of the Mersenne number to ask factordb.com about\n"
               . "Ask factordb.com for factors and primality certificates of the Mersenne number.\n"
               . "If --certs is given, only primality certificates of cofactors will be requested.\n"
               . "Note that factordb.com only has primality certificates for numbers with at least 300 digits.\n"
  );

my ($baseURL) = 'http://www.factordb.com/';
my ($minCertified) = Math::BigInt->new('1e300');
my (%retrieved); # key is URL, value is $response->{'content'} so a given URL is not fetched twice
my ($http) = HTTP::Tiny->new();

# Given a list of HREFs (only the paths, not the site), retrieve the HTML.
sub fetchHREF($)
{
  my ($href) = @_;
  local $_;
  return if exists $retrieved{$href}; # already fetched
  my ($url) = $baseURL . $href;
  my ($response) = $http->put($url, { 'Referer' => $baseURL, } );
  my ($err) = "URL $url status $response->{'status'}, reason $response->{'reason'}; HTTP headers of response:\n";
  while (my ($k, $v) = each %{$response->{'headers'}})
    {
      $err .= ("$k $_\n") foreach (ref $v eq 'ARRAY' ? @$v : $v);
    }
  if (!($response->{'success'}))
    { #http://www.factordb.com/?query=M11715 failed, URL http://www.factordb.com/?query=M11715 status 599, reason Internal Exception; HTTP headers of response:
      if (exists($response->{'reason'}) && $response->{'reason'} eq 'Internal Exception')
        { #  reason Internal Exception
          if (exists($response->{'content'}) && $response->{'content'} =~ m/^Could not connect to /)
            { # content Could not connect to 'www.factordb.com:80': Name or service not known
              print("delay 300\n"); # seconds
              exit(0);
            }
        }
      die "$url failed, $err\n  " . join(' ', %{$response});
    }
  $retrieved{$href} = $response->{'content'};
}

# Given the HTML fetched from a URL, return an array of HREFs to fetch.
#   If factordb.com's HTML format changes, this will likely need changes to match.
sub parseHREFs($)
{
  my ($content) = @_;
  my (@hrefs);
  local $_;
  $_ = $content;
  s/[\r\n]+/ /g;
  while (s/^[^<]*[<]//)
    {
      next unless s/^a +//i;
      if (m/^href[=]"/i)
        {
          s/^href[=]"//i;
          push(@hrefs, $1), next if s/^((cert)|(index)[.]php[?](show)?id[=][0-9]+)"[>]//i;
        }
      if (m/^onclick="getdata[(]'prime','/i)
        {
          s/^onclick="getdata[(]'prime','//i;
          push(@hrefs, $1) if s/^(frame_prime[.]php[?](show)?id[=][0-9]+)'[)]"[>]//i;
        }
    }
  return(@hrefs);
}

sub parseNumber($)
{
  my ($numHREF) = @_;
  my ($html) = $retrieved{$numHREF};
  my ($number) = '';
  $html =~ s/[\r\n]+/ /g;
  return unless $html =~ s,.*?Number[<]/td[>] *[<]td.*?[>],,;
  $number .= "$1"
    while ($html =~ s,([0-9]+)[<]br[>] *,,);
  return(Math::BigInt->new("$number"));
}

my ($checkConfig) = scalar(grep { $_ eq '--check' } @ARGV) if @ARGV;
my ($version) = MersInfo::version();
print(STDOUT "$0 version $version\n") if @ARGV && ($checkConfig || grep { $_ eq '--version' } @ARGV);
exit(0)
  if scalar(grep { $_ eq '--version' } @ARGV);
print("$usage"),
  exit(0)
  if scalar(grep { m,^(-h|--help)$, } @ARGV);

my ($exp);
my ($debug);
my ($certsOnly);
while (@ARGV)
{
  my ($opt) = shift;
  exit(0) if $opt eq '--check' && !@ARGV;
  $debug = 1, next if $opt eq '-d';
  $certsOnly = 1, next if $opt eq '--certs' or $opt eq '-c';
  $opt =~ m/^-/ and die "unexpected option: $opt\n";
  $exp = $1, last if $opt =~ m/^([1-9][0-9]+)$/;
  die "unexpected argument '$opt'; only -c, --certs, --check, --version, and a numeric exponent are supported\n";
}
die("unexpected argument '$ARGV[0]'; only -c, --certs, --check, --version, and a numeric exponent are supported\n")
  if @ARGV;
exit(0) if $checkConfig;
die("missing argument; a Mersenne exponent is required\n")
  unless defined $exp;

my ($url) = $baseURL;
$url .= '?query=M' . "$exp";
my ($firstURL) = $url;

my ($response) = $http->put($url, { 'Referer' => $baseURL, } );

my ($err) = "URL $url status $response->{'status'}, reason $response->{'reason'}; HTTP headers of response:\n";
while (my ($k, $v) = each %{$response->{'headers'}})
{
  $err .= ("$k $_\n") foreach (ref $v eq 'ARRAY' ? @$v : $v);
}

if (!($response->{'success'}))
{ #http://www.factordb.com/?query=M11715 failed, URL http://www.factordb.com/?query=M11715 status 599, reason Internal Exception; HTTP headers of response:
  if (exists($response->{'reason'}) && $response->{'reason'} eq 'Internal Exception')
    { #  reason Internal Exception
      if (exists($response->{'content'}) && $response->{'content'} =~ m/^Could not connect to /)
        { # content Could not connect to 'www.factordb.com:80': Name or service not known
          print("delay 300\n"); # seconds
          exit(0);
        }
    }
  die "$url failed, $err\n  " . join(' ', %{$response});
}

length($response->{'content'}) or die "no HTML in response, $err";
$retrieved{'query=M' . "$exp"} = $response->{'content'};

print("# $url content is:\n$response->{'content'}\n")
  if $debug;

my (@hrefs) = parseHREFs($response->{'content'});

print("# hrefs are:\n" . join("\n", @hrefs) . "\n")
  if $debug;

if (!@hrefs)
{ # I think this should never happen currently unless a server-side CPU limit has been exceeded.
  die "$url did not include itself or any other relevant HREFs:\n$response->{'content'}\n"
    unless ($response->{'content'} =~ m/You have reached your hourly limit of ([1-9][0-9]+) seconds CPU-time[.]/);
  # Tell caller to not try again for an hour.
  print("delay 3600\n");
  exit(0);
}

# The first two should be the "same" - the current page, for M$exp.
# The 'id' page has the query result, a link to any primality cert, etc.
shift(@hrefs),
  shift(@hrefs)
  if $#hrefs > 0 && $hrefs[0] =~ s/showid/id/ && $hrefs[0] eq $hrefs[1];

# Change to index?showid= to get the numbers (factors)
my (@numHrefs) = @hrefs;
s/[?]id=/?showid=/
  foreach (@numHrefs);

# Fetch the numbers and parse them from the HTML
my ($m) = Mersenne->new('n', $exp);
my ($cnt) = 0;
my (%factorHrefs);
foreach (@numHrefs)
{
  fetchHREF($_);
  my ($factor) = parseNumber($_);
  my (@factors) = $m->f($factor);
  next unless $factor > $minCertified;
  $factorHrefs{$cnt} = $_,
    ++$cnt,
    next
    if $#factors >= 0 && $factors[$#factors] == $factor;
  my (@cofactors) = $m->c();
  $factorHrefs{'S'} = $_,
    next
    if $cofactors[0] == $factor && $exp % 8 != 4;
  if ($exp % 8 == 4)
    {
      $factorHrefs{'L'} = $_,
        next
        if $cofactors[0] == $factor;
      $factorHrefs{'M'} = $_,
        next
        if $cofactors[1] == $factor;
    }
  # Factor from factordb.com is unexpected but may be a multiple of a known factor or cofactor
  my ($changed) = 1;
  while ($factor > 1 && $changed)
    {
      $changed = 0;
      foreach my $f (@factors, @cofactors, )
        {
          next unless $factor >= $f;
          $factor = 1,
            last
            if $factor == $f;
          $factor /= $f,
            $changed = 1,
            next
            if $factor % $f == 0;
        }
    }
  next if $factor <= 1;
  $m = Mersenne->new('n', $exp);
  my (@list) = sort { $a <=> $b } (@factors, $factor, @cofactors, );
  $m->f(@list);
}

if (!$certsOnly)
{ # Presume all factors (including cofactors) are too small unless a larger one is found.
  my ($certsNeeded) = 0;
  my (@factors) = $m->f();
  if (@factors)
    {
      print("M( $exp )C: ", join("\nM( $exp )C: ", @factors) . "\n");
    }
  foreach (@factors, $m->c(), )
    {
      ++$certsNeeded,
        last
        if $_ > $minCertified;
    }
  print("small factors\n")
    unless $certsNeeded;
}

# There are a lot of L and M cert files from factordb.com with no matching M or L file,
#   so put them all in a subdir if it exists.  doProof.pl will look there first.
my ($subdir) = '.';
$subdir = './certs' if -s './certs';
# Sort the factors (in numeric order) before the cofactors.
foreach my $f (sort { ($a =~ m/[0-9]/) ? ($b =~ m/[0-9]/ ? ($a <=> $b) : -1) : ($b =~ m/[0-9]/ ? 1 : $a cmp $b) } keys(%factorHrefs))
{ # Change to cert.php?id= to get any primality certificates
  warn("$subdir/cert.M$exp.$f.out already exists\n"),
    print(STDOUT "cert.M$exp.$f.out\n"),
    next
    if -s "$subdir/cert.M$exp.$f.out";
  $_ = $factorHrefs{$f};
  next unless s/^index/cert/;
  next unless s/[?]showid=/?id=/;
  fetchHREF($_);
  next unless exists $retrieved{$_} && $retrieved{$_} ne '';
  my ($fh);
  open($fh, '>', "$subdir/cert.M$exp.$f.out") or die "could not create $subdir/cert.M$exp.$f.out, $!\n";
  print($fh $retrieved{$_});
  close($fh) or die "could not close new file $subdir/cert.M$exp.$f.out, $!\n";
  print(STDOUT "cert.M$exp.$f.out\n");
}

exit(0);

=pod

Here's a response for M15409 on 12 July 2019.
The part of the response with the needed info is all on one line.

<input type="text" size=100 name="query" value="M15409">
<input type="submit" value="Factorize!"> &nbsp; (<a href="help.php?page=0">?</a>)
</center>
<br /><br />
</form><table border=0 width="98%"><tr><td align="center" colspan=3 bgcolor="#BBBBBB"><b>Result:</b></td>
</tr><tr><td bgcolor="#DDDDDD">status <a href="status.html" target="_blank">(?)</a></td>
<td bgcolor="#DDDDDD">digits</td>
<td bgcolor="#DDDDDD">number</td>
</tr><tr><td>FF</td>
<td>4639 <a href="index.php?showid=1000000000000015409">(show)</a></td>
<td><a href="index.php?id=1000000000000015409"><font color="#002099">2^15409-1</font></a><sub>&lt;4639&gt;</sub> = <a href="index.php?id=326023"><font color="#000000">326023</font></a> &middot; <a href="index.php?id=524287"><font color="#000000">524287</font></a> &middot; <a href="index.php?id=6163601"><font color="#000000">6163601</font></a> &middot; <a href="index.php?id=16303307543"><font color="#000000">16303307543</font></a><sub>&lt;11&gt;</sub> &middot; <a href="index.php?id=375470362146497753"><font color="#000000">375470362146497753</font></a><sub>&lt;18&gt;</sub> &middot; <a href="index.php?id=1100000000008707918"><font color="#000000">1205618453...21</font></a><sub>&lt;66&gt;</sub> &middot; <a href="index.php?id=1100000000213136430"><font color="#000000">2311981238...63</font></a><sub>&lt;83&gt;</sub> &middot; <a href="index.php?id=1100000000213136429"><font color="#000000">1502740709...43</font></a><sub>&lt;92&gt;</sub> &middot; <a href="index.php?id=1100000001159047800"><font color="#000000">1379182423...81</font></a><sub>&lt;4355&gt;</sub></td>
</tr></table><br><br><div id="moreinfo"><table border=0 width="98%"><tr><td align="center" bgcolor="#BBBBBB"><b>More information</b> <a onclick="getdata('moreinfo','frame_moreinfo.php?id=1000000000000015409')"><img src="expand.png" border=0></a></td>
</tr></table></div><br /><div id="ecm"><table border=0 width="98%"><tr><td align="center" bgcolor="#BBBBBB"><b>ECM</b> <a onclick="getdata('ecm','frame_ecm.php?id=1000000000000015409')"><img src="expand.png" border=0></a></td>
</tr></table></div><br><br>
<center>
    <font size="-1">
    factordb.com - 57 queries to generate this page (0.18 seconds) (<a href="res.php">limits</a>) (<a href="imp.html">Imprint</a>) (<a href="datenschutz.html">Privacy Policy</a>)
    </font>
</center>
</body>
</html>

What should be in @hrefs for M15409:
index.php?showid=1000000000000015409
index.php?id=1000000000000015409
index.php?id=326023
index.php?id=524287
index.php?id=6163601
index.php?id=16303307543
index.php?id=375470362146497753
index.php?id=1100000000008707918
index.php?id=1100000000213136430
index.php?id=1100000000213136429
index.php?id=1100000001159047800

URL to fetch a cert file:
http://www.factordb.com/cert.php?id=1100000001159047800

URL to get the value of a number:
http://www.factordb.com/index.php?showid=1100000000213136429

HTML for that URL:
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN" 
  "http://www.w3.org/TR/html4/loose.dtd">
<html>
    <head>
        <title>factordb.com</title>
	<script type="text/javascript" src="ajax.js"></script>
    </head>
    <body><table border=0 width="98%"><tr><td width=14% align="center" bgcolor="#BBBBBB"><a href="index.php">Search</a></td>
<td width=14% align="center" bgcolor="#BBBBBB"><a href="sequences.php">Sequences</a></td>
<td width=14% align="center" bgcolor="#BBBBBB"><a href="result.php">Report results</a></td>
<td width=14% align="center" bgcolor="#BBBBBB"><a href="tables.php">Factor tables</a></td>
<td width=14% align="center" bgcolor="#BBBBBB"><a href="status.php">Status</a></td>
<td width=14% align="center" bgcolor="#BBBBBB"><a href="downloads.php">Downloads</a></td>
<td width=14% align="center" bgcolor="#BBBBBB"><a href="login.php">Login</a></td>
</tr></table><br><form action="index.php" method="GET"><input type="hidden" name="showid" value="1100000000213136429"><table border=0 width="98%"><tr><td align="center" colspan=2 bgcolor="#BBBBBB">Additional information (Internal ID <a href="index.php?id=1100000000213136429">1100000000213136429</a>)</td>
</tr><tr><td align="center" bgcolor="#DDDDDD">Digits (Base <select name="base" size=1" onchange="submit()"><option value="2">2<option value="3">3<option value="4">4<option value="5">5<option value="6">6<option value="7">7<option value="8">8<option value="9">9<option value="10" selected>10<option value="11">11<option value="12">12<option value="13">13<option value="14">14<option value="15">15<option value="16">16<option value="17">17<option value="18">18<option value="19">19<option value="20">20<option value="21">21<option value="22">22<option value="23">23<option value="24">24<option value="25">25<option value="26">26<option value="27">27<option value="28">28<option value="29">29<option value="30">30<option value="31">31<option value="32">32<option value="33">33<option value="34">34<option value="35">35<option value="36">36</select>)</td>
<td align="center" bgcolor="#DDDDDD">92</td>
</tr><tr><td align="center">Number</td>
<td align="center">15027407098633642094023724924629416240581289482791222070473057361777251946987551028766864143<br>
</td>
</tr></table></form><br><br>
<center>
    <font size="-1">
    factordb.com - 6 queries to generate this page (0.00 seconds) (<a href="res.php">limits</a>) (<a href="imp.html">Imprint</a>) (<a href="datenschutz.html">Privacy Policy</a>)
    </font>
</center>
</body>
</html>

http://www.factordb.com/index.php?id=1100000001159047800 content is:
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN" 
  "http://www.w3.org/TR/html4/loose.dtd">
<html>
    <head>
        <title>factordb.com</title>
	<script type="text/javascript" src="ajax.js"></script>
    </head>
    <body><table border=0 width="98%"><tr><td width=14% align="center" bgcolor="#BBBBBB"><a href="index.php">Search</a></td>
<td width=14% align="center" bgcolor="#BBBBBB"><a href="sequences.php">Sequences</a></td>
<td width=14% align="center" bgcolor="#BBBBBB"><a href="result.php">Report results</a></td>
<td width=14% align="center" bgcolor="#BBBBBB"><a href="tables.php">Factor tables</a></td>
<td width=14% align="center" bgcolor="#BBBBBB"><a href="status.php">Status</a></td>
<td width=14% align="center" bgcolor="#BBBBBB"><a href="downloads.php">Downloads</a></td>
<td width=14% align="center" bgcolor="#BBBBBB"><a href="login.php">Login</a></td>
</tr></table><br><form action="index.php" method="get">
<center>
<input type="text" size=100 name="query" value="(2^15409-1)/(2^811-1)/19781307260753411033320600524459524461073">
<input type="submit" value="Factorize!"> &nbsp; (<a href="help.php?page=0">?</a>)
</center>
<br /><br />
</form><table border=0 width="98%"><tr><td align="center" colspan=3 bgcolor="#BBBBBB"><b>Result:</b></td>
</tr><tr><td bgcolor="#DDDDDD">status <a href="status.html" target="_blank">(?)</a></td>
<td bgcolor="#DDDDDD">digits</td>
<td bgcolor="#DDDDDD">number</td>
</tr><tr><td>P</td>
<td>4355 <a href="index.php?showid=1100000001159047800">(show)</a></td>
<td><a href="index.php?id=1100000001159047800"><font color="#000000">1379182423...81</font></a><sub>&lt;4355&gt;</sub> = <a href="index.php?id=1100000001159047800"><font color="#000000">1379182423...81</font></a><sub>&lt;4355&gt;</sub></td>
</tr></table><br><br><div id="moreinfo"><table border=0 width="98%"><tr><td align="center" bgcolor="#BBBBBB"><b>More information</b> <a onclick="getdata('moreinfo','frame_moreinfo.php?id=1100000001159047800')"><img src="expand.png" border=0></a></td>
</tr></table></div><br /><div id="ecm"><table border=0 width="98%"><tr><td align="center" bgcolor="#BBBBBB"><b>ECM</b> <a onclick="getdata('ecm','frame_ecm.php?id=1100000001159047800')"><img src="expand.png" border=0></a></td>
</tr></table></div><br /><div id="prime"><table border=0 width="98%"><tr><td align="center" bgcolor="#BBBBBB"><b>Primality</b> <a onclick="getdata('prime','frame_prime.php?id=1100000001159047800')"><img src="expand.png" border=0></a></td>
</tr></table></div><br><br>
<center>
    <font size="-1">
    factordb.com - 6 queries to generate this page (0.00 seconds) (<a href="res.php">limits</a>) (<a href="imp.html">Imprint</a>) (<a href="datenschutz.html">Privacy Policy</a>)
    </font>
</center>
</body>
</html>

http://www.factordb.com/index.php?showid=1100000001159047800 content is:
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN" 
  "http://www.w3.org/TR/html4/loose.dtd">
<html>
    <head>
        <title>factordb.com</title>
	<script type="text/javascript" src="ajax.js"></script>
    </head>
    <body><table border=0 width="98%"><tr><td width=14% align="center" bgcolor="#BBBBBB"><a href="index.php">Search</a></td>
<td width=14% align="center" bgcolor="#BBBBBB"><a href="sequences.php">Sequences</a></td>
<td width=14% align="center" bgcolor="#BBBBBB"><a href="result.php">Report results</a></td>
<td width=14% align="center" bgcolor="#BBBBBB"><a href="tables.php">Factor tables</a></td>
<td width=14% align="center" bgcolor="#BBBBBB"><a href="status.php">Status</a></td>
<td width=14% align="center" bgcolor="#BBBBBB"><a href="downloads.php">Downloads</a></td>
<td width=14% align="center" bgcolor="#BBBBBB"><a href="login.php">Login</a></td>
</tr></table><br><form action="index.php" method="GET"><input type="hidden" name="showid" value="1100000001159047800"><table border=0 width="98%"><tr><td align="center" colspan=2 bgcolor="#BBBBBB">Additional information (Internal ID <a href="index.php?id=1100000001159047800">1100000001159047800</a>)</td>
</tr><tr><td align="center" bgcolor="#DDDDDD">Digits (Base <select name="base" size=1" onchange="submit()"><option value="2">2<option value="3">3<option value="4">4<option value="5">5<option value="6">6<option value="7">7<option value="8">8<option value="9">9<option value="10" selected>10<option value="11">11<option value="12">12<option value="13">13<option value="14">14<option value="15">15<option value="16">16<option value="17">17<option value="18">18<option value="19">19<option value="20">20<option value="21">21<option value="22">22<option value="23">23<option value="24">24<option value="25">25<option value="26">26<option value="27">27<option value="28">28<option value="29">29<option value="30">30<option value="31">31<option value="32">32<option value="33">33<option value="34">34<option value="35">35<option value="36">36</select>)</td>
<td align="center" bgcolor="#DDDDDD">4355</td>
</tr><tr><td align="center">Number</td>
<td align="center">137918242396364593907932389361640954993976795769293167281994565924492916788441072039000993068573786446282088895572147649<br>
004940723177033751040221849385631302140510214195801776763323195529669233120566991449131734057581025108883131874312452114<br>
607334143768886581043779772140982785226342182586833930332409145896331305809054812347829225232667390211008278394670877131<br>
815312281258827532768928730419744618097257664286354533291685712769576173230287263854998362714409786340091375913526389932<br>
425088631040935291985945358566215771021565799009892058345566636399487105631330662834110398305790224837264699042969967615<br>
789331615391573881573828143001901846028556546287879588711352372481390016002790407299110397655965267075142879008000339784<br>
995555424063655673401844401512914339713919023141525138780075497686383965828307957491721658649518973720003507615340727980<br>
470957631604045203678166284984198715869370646701995329428394046391059175598480306730865258129388864226590497370614773100<br>
788186598398332698826912239437807010062096011372306131686106589322964272318267665466973508677071782892038682864894037195<br>
942269088435665018579866470888547772227425486091576439716499611232472264630929976502049416503231916557448241259502541166<br>
637897094547896013459414997509555782314929387407875163157173141051365962440022232343945476697454377835441521433058873564<br>
304383940217850159919946326395792370406796200782800804740424683364518551413452239852326998038697377545204764353375547379<br>
188297746657058584803159529647569839816451649048072794661335633422878027793460232256802695137609597287686448870653789691<br>
566057454254616521439408001111047273091060623023139913871063066234816687670368679347950171776003980871002760806940347720<br>
618264436289035024937178740300193659319865281983065991075492585701639670973695406204612763836296679766507462294251105018<br>
017770270621394301151375697019597736631011705520657793186371548703161287057171293415320821080527910753851353781810914456<br>
419572213354212166239152234194246092537924720522139123563368799957129484656423437090961905690135710266890845971682012376<br>
359125131265268284002970849700649068840001639899211610110391013815680706349578593791625961698365121871906666856634998876<br>
498433741638999926402163189334109612319714791878975552221388960317207435183446574454487577284804405364640098615654883024<br>
443144786380880740041585074108832545364867080032287721207206969230837848059733618948078249677968402334235747799665177800<br>
403571363163558869796036463380527878357340170758033796671634983596928819023805256944143559789846095113943279143711972317<br>
173975500726811412960815780630590119923419841094987075354370339887441269102212575482238604320755671663942017072899904484<br>
260285377520062010663293383201681810122441093040940148863434578948045846384565592283486330248804488211071992180290128343<br>
993993735730447096182979521473533113580853180079475170366587412337891609530253623557729988851681246447818199349391271754<br>
635607263926890904792899960209142521656115235289100811391178299145602605603737493261073952174693724406006360196479431129<br>
348144894587683372845642581249135306167182372844784218840731272562505927145650576959898259457976031001060900195864824087<br>
524404473391040086559044654358017131230753798838880608333859188701153313327810105107162607334812322970922773104057622439<br>
492421315184116419306511234308975632886316965759013853570672091495400194160271705511759418874703846983599087878818798526<br>
640258535134340189023440973594863946112673124957645744905154281202969850458299633323351370327806666160536300067862298391<br>
444183517275848766251233404105289428734947625095816413871926660801539143002927982191136357775753638885450613127896234617<br>
114765345096962179209576622630197796327806582008256574154279438222769980098876912424764644501264606327174820099289156512<br>
614266805148313023085657555962150523753722647330468463199142862337221133445937554136787039842795468234777230956105782269<br>
578878199364786779289158711915539207573930832982708163747315252574153970698683067424527552387781399637910679007968861839<br>
256238800407690102462187817789012670802324373363588995248797411429542704988396327899040706380622570954561487453556247848<br>
167385782193017863525317270201387412486600239317549983660785957851611014405706738587747800636557566627724426288013333679<br>
784791611421291393281707298937397253947742188654412247645808942041378753284581346551268812794909387447059236873869239601<br>
76760956519028698023017301121277681<br>
</td>
</tr></table></form><br><br>
<center>
    <font size="-1">
    factordb.com - 6 queries to generate this page (0.02 seconds) (<a href="res.php">limits</a>) (<a href="imp.html">Imprint</a>) (<a href="datenschutz.html">Privacy Policy</a>)
    </font>
</center>
</body>
</html>

http://www.factordb.com/frame_prime.php?id=1100000001159047800 content:
<center><table border=0 width="600"><tr><td align="center" colspan=2 bgcolor="#BBBBBB"><b>Primality proving</b></td>
</tr><tr><td align="center" colspan=2 bgcolor="#DDDDDD">Proven by certificate</td>
</tr><tr><td align="center">Size</td>
<td align="center">1,155,764 bytes</td>
</tr><tr><td align="center" bgcolor="#DDDDDD">Status</td>
<td align="center" bgcolor="#DDDDDD">Verified</td>
</tr><tr><td align="center">Uploaded by</td>
<td align="center">DukeBG</td>
</tr><tr><td align="center" bgcolor="#DDDDDD">Date</td>
<td align="center" bgcolor="#DDDDDD">September 19, 2018, 10:31 pm</td>
</tr><tr><td align="center">Download</td>
<td align="center"><a href="cert.php?id=1100000001159047800">Download</a></td>
</tr></table></center><center><form action="index.php" method="GET"><input type="hidden" name="id" value="1100000001159047800"><input type="hidden" name="open" value="prime"><table border=0 width="600"><tr><td align="center" colspan=2 bgcolor="#BBBBBB">Proof by N-1/N+1</td>
</tr><tr><td align="center" bgcolor="#DDDDDD"><a href="index.php?id=1100000001159052038">N-1</a></td>
<td align="center" bgcolor="#DDDDDD">(0.483% factored)</td>
</tr><tr><td align="center"><a href="index.php?id=1100000001159052039">N+1</a></td>
<td align="center">(0.165% factored)</td>
</tr><tr><td align="center" bgcolor="#DDDDDD">Combined N-1/N+1-test</td>
<td align="center" bgcolor="#DDDDDD">(0.648% factored)</td>
</tr><tr><td align="center" colspan=2><a href="helper.php?id=1100000001159047800">Download helper file</a></td>
</tr></table></form></center><center><form action="index.php" method="GET"><input type="hidden" name="id" value="1100000001159047800"><input type="hidden" name="open" value="prime"><table border=0 width="600"><tr><td align="center" colspan=2 bgcolor="#BBBBBB"><b>Probable prime</b></td>
</tr><tr><td align="center" bgcolor="#DDDDDD">Bases checked</td>
<td align="center" bgcolor="#DDDDDD">3</td>
</tr><tr><td align="center">Check base (2-255)</td>
<td align="center"><input type="text" name="basetocheck" size=4 maxlength=3></td>
</tr><tr><td align="center" colspan=2 bgcolor="#DDDDDD"><input type="submit" value="Check"></td>

http://www.factordb.com/frame_prime.php?id=1100000001156405082 content is:
<center><table border=0 width="600"><tr><td align="center" colspan=2 bgcolor="#BBBBBB"><b>Primality proving</b></td>
</tr><tr><td align="center" colspan=2 bgcolor="#DDDDDD">Proven by certificate</td>
</tr><tr><td align="center">Size</td>
<td align="center">1,597,185 bytes</td>
</tr><tr><td align="center" bgcolor="#DDDDDD">Status</td>
<td align="center" bgcolor="#DDDDDD">Verified</td>
</tr><tr><td align="center">Uploaded by</td>
<td align="center">DukeBG</td>
</tr><tr><td align="center" bgcolor="#DDDDDD">Date</td>
<td align="center" bgcolor="#DDDDDD">October 26, 2018, 7:01 pm</td>
</tr><tr><td align="center">Download</td>
<td align="center"><a href="cert.php?id=1100000001156405082">Download</a></td>
</tr></table></center><center><form action="index.php" method="GET"><input type="hidden" name="id" value="1100000001156405082"><input type="hidden" name="open" value="prime"><table border=0 width="600"><tr><td align="center" colspan=2 bgcolor="#BBBBBB">Proof by N-1/N+1</td>
</tr><tr><td align="center" bgcolor="#DDDDDD"><a href="index.php?id=1100000001156414101">N-1</a></td>
<td align="center" bgcolor="#DDDDDD">(0.326% factored)</td>
</tr><tr><td align="center"><a href="index.php?id=1100000001156414102">N+1</a></td>
<td align="center">(0.154% factored)</td>
</tr><tr><td align="center" bgcolor="#DDDDDD">Combined N-1/N+1-test</td>
<td align="center" bgcolor="#DDDDDD">(0.48% factored)</td>
</tr><tr><td align="center" colspan=2><a href="helper.php?id=1100000001156405082">Download helper file</a></td>
</tr></table></form></center><center><form action="index.php" method="GET"><input type="hidden" name="id" value="1100000001156405082"><input type="hidden" name="open" value="prime"><table border=0 width="600"><tr><td align="center" colspan=2 bgcolor="#BBBBBB"><b>Probable prime</b></td>
</tr><tr><td align="center" bgcolor="#DDDDDD">Bases checked</td>
<td align="center" bgcolor="#DDDDDD">3</td>
</tr><tr><td align="center">Check base (2-255)</td>
<td align="center"><input type="text" name="basetocheck" size=4 maxlength=3></td>
</tr><tr><td align="center" colspan=2 bgcolor="#DDDDDD"><input type="submit" value="Check"></td>
</tr></table></form></center>

=cut
