#!/usr/bin/perl -Tw
# Ask GIMPS for tasks based on the command line.

use strict;
use strict 'vars';
use HTTP::Tiny;
use File::stat;

BEGIN { push(@INC, '.'); $ENV{'PATH'} = '/usr/bin:/bin'; }
use MersInfo;

my ($usage)
  = "./askGIMPS.pl [-c<GPUs>] [-o<tasksFile>] [-t<taskCount>] [-w<taskType>] [-l<lowExp>] [-h<highExp>]\n"
  . "      [-f<passwordFile>] [--getTaskTypes] [--check] [--version] [--help]\n"
  . "  The passwordFile should contain two lines like:\n"
  . "    -u<GIMPSuser>\n"
  . "    -p<GIMPSpassword>\n"
  . "  ... which could instead be given on the command line (but that can make them visible\n"
  . "  ... to other users on the system).\n";
#   The contents of passwordFile could include any option(s) and their argument(s) and will
#      all be parsed and acted on when the file is read, before later options and arguments.
#   The argument to -c is the number of GPU cores to be used, default 1.
#   The tasksFile will be created to contain the newly reserved tasks.
#   The taskCount is the number of tasks to reserve for each GPU core (-c), default 1.
#   The taskType is the numeric GIMPS task type ("option value" in the HTML) to reserve,
#     default 2 (trial factoring).  All the task types GIMPS has are supported.
#   The -l and -h options limit the reservations to that range of (Mersenne) exponents
#     but it is better to let GIMPS decide what to reserve unless the software that will
#     do the work does not support some exponents.
#   Giving --getTaskTypes will get the GIMPS task types and print them if different from
#     what this script uses.  Only -u and -p can be used with this option.

my ($checkConfig) = scalar(grep { $_ eq '--check' } @ARGV) if @ARGV;
my ($version) = MersInfo::version();
print(STDOUT "$0 version $version\n") if @ARGV && ($checkConfig || grep { $_ eq '--version' } @ARGV);
print(STDOUT "$usage")
  if scalar(grep { $_ eq '--help' } @ARGV);
exit(0) if @ARGV && scalar(grep { $_ eq '--version' || $_ eq '--help' } @ARGV);

open(STDERR, ">&STDOUT")
  or die("could not redirect STDERR to STDOUT, $!\n");

my ($baseURL) = 'https://www.mersenne.org/manual_assignment/';
my ($gimpsUser, $gimpsPassword); # no defaults; given with -u and -p on command line (perhaps inside a file given with -f)
my ($cores, $tasksPerCore, $workType) = (1, 1, 2); # 1 core, 1 task, Trial Factoring
my ($lowestExp, $highestExp) = ('', ''); # default: any exponent(s) OK
my ($outfile);
my ($getTaskTypes) = scalar(grep { $_ eq '--getTaskTypes' } @ARGV);
while (@ARGV)
{
  my ($opt) = shift();
  next
    if $opt =~ m,^(--check|--getTaskTypes)$,;
  $opt =~ m/^-(.)(.*)$/ or die "unexpected option or argument: $opt\n";
  ($gimpsUser = ("$2" ne '') ? $2 : shift()), next if $1 eq 'u';
  ($gimpsPassword = ("$2" ne '') ? $2 : shift()), next if $1 eq 'p';
  if ($1 eq 'f') # read options from a file
    {
      my ($fn) = ($2 ne '') ? $2 : shift();
      $fn ne '' or die "no file name given with -f\n";
      -f($fn) && -s($fn) or die "non-file or empty file given with -f\n";
      my ($fh);
      open($fh, '<', "$fn") or die "could not open $fn, $!\n";
      my(@newArgs);
      push(@newArgs, <$fh>);
      chomp(@newArgs);
      unshift(@ARGV, @newArgs);
      next;
    }
  die "unexpected option '$1'; --getTaskTypes can only be used with -u, -p, and -f"
    if $getTaskTypes;
  ($cores = ("$2" ne '') ? $2 : shift()), next if $1 eq 'c';
  ($outfile = ("$2" ne '') ? $2 : shift()), next if $1 eq 'o';
  ($tasksPerCore = ("$2" ne '') ? $2 : shift()), next if $1 eq 't';
  ($workType = ("$2" ne '') ? $2 : shift()), next if $1 eq 'w'; # (preferred) work type
  ($lowestExp = ("$2" ne '') ? $2 : shift()), next if $1 eq 'l';
  if ($1 eq 'h')
    {
      $highestExp = ("$2" ne '') ? $2 : shift();
      print($usage),
        exit(0)
        unless defined $highestExp;
      next;
    }
  die "unexpected option '$1'; only -[cfhloptw], --check, --version, --help, and --getTaskTypes are supported\n";
}
exit(0) if $checkConfig;

if ($getTaskTypes)
{
  print("--getTaskTypes ignored; only needed by developer while testing\n"),
    exit(0)
    if !-f('mysql.devel');
  my ($info);
  if (-f('.askedGIMPStaskTypes'))
    {
      $info = stat('.askedGIMPStaskTypes');
      print("last asked GIMPS for the task types less than 30 days ago at "
            . localtime($info->mtime())),
        exit(0)
        if $info->mtime() + 30*24*60*60 > time();
    }
  unlink('.askedGIMPStaskTypes');
  open(my $aGgT, '>', '.askedGIMPStaskTypes')
    or die("could not create .askedGIMPStaskTypes, $!");
  print($aGgT localtime(time()))
    or die("could not print localtime into .askedGIMPStaskTypes, $!");
  close($aGgT)
    or die("could not close .askedGIMPStaskTypes, $!");
  $info = stat('.askedGIMPStaskTypes');
  die("could not stat .askdGIMPStaskTypes or modify time did not update, $!")
    unless $info && $info->mtime() + 60 > time();
}

# From $baseURL source, 26 Dec 2015:
# <p>Manual assignments are <span class="bold">good for only six months</span>. Assignments can be extended manually.</p>
# <form method="get" action="/manual_assignment/">
#  <span class="bold">Number of CPUs (cores) you need assignments for (maximum 12):</span>
#  <input size="2" type="text" name="cores" value="1" style="text-align: right;"><br>
#  <span class="bold">Number of assignments you want for each core:</span>
#  <input size="2" type="text" name="num_to_get" value="1" style="text-align: right;"><br>
#  <span class="bold">Preferred work type:</span>
#   <select name="pref">
#    <option value="102">World record tests</option>
#    <option value="100">Smallest available first-time tests</option>
#    <option value="101">Double-check tests</option>
#    <option value="4">P-1 factoring</option>
#    <option value="2">Trial factoring</option>
#    <option value="5">ECM factoring</option>
#    <option value="104">100 million digit tests</option>
#   </select><br>
#  Optional exponent range: <input size="10" type="text" name="exp_lo" value="" style="text-align: right;">
#   to <input size="10" type="text" name="exp_hi" value="" style="text-align: right;"><br>
#  <button type="submit">Get Assignments</button>
# </form>
# As of 2021 Aug 31, the entire list of options is all on the same line as 'Preferred work type:'.
# This should parse these from the HTML, but that would require retrieving it first, doubling the interactions with the server.
my (%workTypes) =
  (
   2 => 'Trial factoring',
   4 => 'P-1 factoring',
   5 => 'ECM factoring',
   8 => 'ECM on Mersenne cofactors',
   101 => 'Double-check LL tests',
   150 => 'Smallest available first-time PRP tests',
   151 => 'Double-check PRP tests',
   152 => 'World record PRP tests',
   153 => '100 million digit PRP tests',
   155 => 'Double-check using PRP with proof',
   160 => 'First PRP tests on Mersenne cofactors',
   161 => 'Double-check PRP tests on Mersenne cofactors',
  );

my ($fh) = *STDOUT;
if ($outfile)
{
  die("unexpected file name with -o - must be in working.doWork/ subdir and only '.', letters, and digits\n")
    unless $outfile =~ m:^(working[.]doWork/[.A-Za-z0-9]*)$:;
  $outfile = "$1";
  die("file with -o '$outfile' already exists\n")
    if -e $outfile;
  die("file with -o '$outfile' cannot be a soft link\n")
    if -l $outfile;
  open($fh, '>', $outfile)
    or die("cannot create '$outfile', $!\n");
}

my ($url) = $baseURL;
if (!$getTaskTypes)
{
  defined($gimpsUser) && $gimpsUser ne '' && defined($gimpsPassword) && $gimpsPassword ne ''
    or die "-u (mersenne.org user) and -p (mersenne.org password) are required to get work\n";
  if (!exists($workTypes{$workType}))
    {
      print($fh "unexpected work type (-w) $workType; choose from these numbers:\n");
      print($fh "$_ $workTypes{$_}\n") foreach (keys(%workTypes));
      die "See $baseURL for details\n";
    }
  $cores > 0 && $cores < 13
    or die "-c (CPU cores) must be from 1 to 12; default 1\n";
  $tasksPerCore > 0
    or die "-t (tasks per core) must be at least 1; default 1\n";
  $lowestExp eq '' || $lowestExp > 1206
    or die "-l (lowest exponent) must be at least 1207; smaller Mersenne numbers have been completely factored\n";
  $lowestExp = 1007 if $lowestExp eq '' && $highestExp ne '';
  $highestExp eq '' || $lowestExp <= $highestExp
    or die "-h (highest exponent) must be at least as large as $lowestExp\n";
  $url .= "?cores=$cores";
  $url .= "&num_to_get=$tasksPerCore";
  $url .= "&pref=$workType";
  $url .= "&exp_lo=$lowestExp&exp_hi=$highestExp";
}

my ($http) = HTTP::Tiny->new();
my ($tags) = { 'Referer' => $baseURL, };
$tags->{'user_login'} = $gimpsUser,
  $tags->{'user_password'} = $gimpsPassword
  unless $getTaskTypes;
my ($response) = $http->post_form($url, $tags);

my ($err) = "status $response->{'status'}, reason $response->{'reason'}; HTTP headers of response:\n";
while (my ($k, $v) = each %{$response->{'headers'}})
{
  $err .= ("    $k $_\n") for (ref $v eq 'ARRAY' ? @$v : $v);
}
if (!$response->{'success'})
{
  die "$url failed, $err\n  " . join(' ', %{$response} . "  ")
    unless $getTaskTypes;
  # Failing to get the task types just delays updating them - don't fail during 'make'.
  warn("$url failed");
  print("$err");
  print('  ' . join(' ', %{$response}));
  exit(0);
}
length($response->{'content'})
  or die "no HTML in response, $err  ";

# Unless getting the task types, $response->{'content'} should include something like these three lines:
#[ ]*<a href="/account/"><button id="button2">wedgingt<br>logged in</button></a><p style="text-align:right;"><a href="/manual_assignment/?logout=u">Logout</a></p>    </div>
#<!--BEGIN_ASSIGNMENTS_BLOCK-->Factor=EF99B4ECC09A61CBA622729CA63D042B,132365237,71,72
#<!--END_ASSIGNMENTS_BLOCK--></pre></p>
my ($foundLoggedIn) = 1 if $getTaskTypes; # getting the task types does not require logging in
my (@assignments);
my ($options);
my ($changes);
my (%newWorkTypes);
my ($printed) = 1 if $getTaskTypes; # there's no work to be printed if getting the task types
foreach (split(/\n/, $response->{'content'}))
{
  $foundLoggedIn = 1,
    next
    if $gimpsUser && m,a href="/account/"[>][<][^>]*[>]$gimpsUser[<]br[>]logged in[<],o;
  while ($getTaskTypes && s,[<]option value[=]\"([0-9]+)\"[>]([- 0-9A-Za-z]+)[<]/option[>],,)
    {
      $options = 1;
      my ($value) = $1;
      my ($name) = $2;
      $newWorkTypes{$value} = $name;
      $changes = 1
        unless exists($workTypes{$value}) && $workTypes{$value} eq $name;
    }
  next
    if $getTaskTypes;
  print($fh join("\n", @assignments) . "\n"),
    @assignments = (),
    $printed = 1
    if m,[<][!]--END_ASSIGNMENTS_BLOCK--[>],;
  push(@assignments, $_),
    next
    if @assignments;
  push(@assignments, $_),
    next
    if s,^[<][!]--BEGIN_ASSIGNMENTS_BLOCK--[>],,;
}
$foundLoggedIn
  or die 'invalid user or password; not "logged in", ' . "$err  ";
$printed
  or die "Could not find any assignments in response, $err"  ;

#!!This started showing up sometime in Jan. 2022; figure out why
#print("The GIMPS web site no longer has 'options' that list the task types.\n"),
  exit(0)
  unless $options;

print("The GIMPS work types have not changed.\n"),
  exit(0)
  unless $changes;

print("The GIMPS work types have changed; update askGIMPS.pl to match.\n");
print(" $_ => '$newWorkTypes{$_}'\n")
  foreach (sort { $a <=> $b } keys(%newWorkTypes));
exit(1);
