#!/usr/bin/perl -Tw
# Ask a "mers format" server for one or more tasks.

use strict;
use strict 'vars';
use warnings;
use Socket qw(:crlf);
use IO::Socket::INET;

BEGIN { push(@INC, '.') if -s('MersInfo.pm'); $ENV{'PATH'} = "/usr/bin:/bin"; }

use MersInfo;

my ($usage) = ("Usage: askServer.pl [-{?w}] [-h <host>] [-p <port>] [-b <min_bits> <max_bits>] [-m <max_bits>] [-o <outFile>]\n"
               . "  [-r <min_exp> <max_exp>] [-u <UUID>] [<cmd>] [<count>] [<host>] [<port>] [--help|-h] [--verison] [--check]\n");

# Split the arguments at spaces as the shell would.
my (@args);
push(@args, split(/ /, $_)) foreach (@ARGV);
@ARGV = @args;

my ($errors) = '';
my ($checkConfig) = scalar(grep { $_ eq '--check' } @ARGV) if @ARGV;
my ($version) = MersInfo::version();
print(STDOUT "$0 version $version\n") if @ARGV && ($checkConfig || grep { $_ eq '--version' } @ARGV);
print(STDOUT "$usage")
  if grep { $_ eq '--help' } @ARGV;
exit(0) if grep { $_ eq '--version' || $_ eq '--help'} @ARGV;

open(STDERR, ">&STDOUT")
 or die("could not redirect STDERR to STDOUT, $!\n");

# The values here are used as a pattern to determine whether a line of data from the server is "work" of the particular type (key) or not.
my (%letter) =
  ( # Elliptic curve method (factoring); any exponent.
    'ECM' => " [)][Ee]: ", 'ECMFactorer' => " [)][Ee]: ",
    # Gaps in or to extend trial factoring; prime exponents only.
    'Joiner' => " [)][GgHU]: ", 'JoiningFactorer' => " [)][GgHU]: ", 'KJoiner' => " [)][GgHU]: ",
    # P-1 method factoring; any exponent.
    'OP-1' => " [)]o: ", 'OMinus1Factorer' => " [)]o: ",
    # P+1 method factoring; any exponent.
    'QP+1' => " [)]q: ", 'QPlus1Factorer' => " [)]q: ",
    # Gaps in trial factoring from one power of two to another; prime exponents only.
    'PowerOf2Joiner' => " [)][Gg]: ", 'PowerOf2JoiningFactorer' => " [)][Gg]: ", 'powerOf2Joiner' => " [)][Gg]: ", 'powerOf2JoiningFactorer' => " [)][Gg]: ",
    # Same type of work with different exponent and potential factor limits; prime exponents only.
    'GpuFactorer' => " [)][gG]: ", 'gpuFactorer' => " [)][gG]: ",
    # Prove that the (SPRP) factors are prime; any exponent.  LM should have been merged into S (d until Jan 2019) by the server.
    'DPrimalityProver' => " [)][dS]", 'dPrimalityProver' => " [)][dS]",
  );

my ($cnt, $host, $port, $ipaddr, $saddr, $proto, $uuid, $wait);
my ($onlyPrimeExps, $minBits, $maxBits, $minExp, $maxExp);
my ($outfile);
my ($cmd) = shift() || 0;
for ( ; $cmd =~ m/^-[A-Za-z?]/; $cmd = shift())
{
  next if $cmd eq '--check'; # already recognized above
  $wait = !$wait, next if $cmd eq '-w';
  $onlyPrimeExps = !$onlyPrimeExps, next if $cmd eq '-O';
  next if ($cmd eq '-b' && ($minBits = shift()) =~ m/^[0-9]+$/ && ($maxBits = shift()) =~ m/^[0-9]+$/ && ($minBits < $maxBits || $maxBits == 0));
  next if ($cmd eq '-c' && ($cnt = shift()) =~ m/^[1-9][0-9]*$/);
  next if ($cmd eq '-m' && ($maxBits = shift()) =~ m/^[0-9]+$/);
  next if ($cmd eq '-o' && ($outfile = shift()));
  next if ($cmd eq '-p' && defined($port = shift()));
  next if ($cmd eq '-r' && ($minExp = shift()) =~ m/^[0-9]+$/ && ($maxExp = shift()) =~ m/^[0-9]+$/ && $minExp <= $maxExp);
  next if ($cmd eq '-u' && ($uuid = shift()));
  if ($cmd eq '-h')
    {
      $host = shift();
      next if defined($host);
      print($usage);
      exit(0);
    }
  $cmd eq '-b' && ($errors .= "-b expects arguments specifying minimum and maximum number of bits in 'factors'\n");
  $cmd eq '-m' && ($errors .= "-m expects an argument specifying maximum number of bits in 'factors'\n");
  $cmd eq '-r' && ($errors .= "-r expects arguments specifying minimum and maximum Mersenne exponent\n");
  $cmd eq '-u' && ($errors .= "-u expects an argument giving a UUID or other string that is unique to this client of the mers server\n");
  $errors =~ m/Usage:/ || ($errors .= $usage);
}
if (!$checkConfig && (!defined($cmd) || !exists($letter{$cmd}) || $letter{$cmd} eq ''))
{
  $errors .= "Supported commands (first non-flag argument): ECM OP-1 QP+1 {,K}Joiner {G,g}puFactorer {P,p}owerOf2Joiner {D,d}PrimalityProver\n";
  $errors =~ m/Usage:/ || ($errors .= $usage);
}

$cnt = shift() || 1 if !defined($cnt);
if ($cnt =~ m/^([1-9][0-9]*)$/)
{
  $cnt = $1;
}
else
{
  $errors .= "count of tasks is not digits but '$cnt'; '-?' for help\n";
}
die "$errors" if $errors;
exit 0 if $checkConfig;

$cnt < 2 or $cmd ne 'Joiner' && $cmd !~ m/^PowerOf2Join/ && $cmd !~ m/^GpuFactorer/
  or die "cannot get more than one task with disconnecting client type\n";

my ($fh) = *STDOUT;
if ($outfile)
{
  die("unexpected file name with -o - must be in working.doWork/ subdir and only '.', letters, and digits\n")
    unless $outfile =~ m:^(working[.]doWork/[.A-Za-z0-9]*)$:;
  $outfile = "$1";
  die("file with -o '$outfile' already exists\n")
    if -e $outfile;
  open($fh, '>', $outfile)
    or die("cannot create '$outfile', $!\n");
}

$host = shift() || 'localhost' if !defined($host);
$host =~ m,^([-.A-Za-z0-9]+)$,
  or die "unexpected host '$host'\n";
$host = $1;
$port = shift() || 'slucas' if !defined($port);
$port = $1 if $port =~ m/(\d+)/;
$port .= '(2203)' if $port !~ m/(\d+)/;
my ($sock) = IO::Socket::INET->new('PeerAddr' => "$host" . ':' . "$port",
                                   'Proto' => 'tcp',
                                   'Type' => SOCK_STREAM,
  );
$sock
  or die "connecting to $host port $port failed, $!";

my ($prefix) = '';
$prefix .= " primes" if $onlyPrimeExps;

$prefix .= " min_bits $minBits" if defined($minBits) && $minBits != 0;
$prefix .= " max_bits $maxBits" if defined($maxBits) && $maxBits != 0;

$prefix .= " range_start $minExp" if $minExp;
$prefix .= " range_stop $maxExp" if $maxExp;
$prefix =~ s/^ //;

print($sock "UUID $uuid$CRLF") if $uuid;
print($sock "multitalented$CRLF" . "work $cmd $prefix$CRLF") if $prefix ne '';
print($sock "$cmd$CRLF") if $prefix eq '';

my ($ask) = 0;
my ($gotOne) = 0;
my ($mers);
my ($saveFile);
my ($sFcontent) = '';
while (defined($mers = <$sock>))
{
  chomp($mers);
  chop($mers) while $mers =~ m/[\n\r]$/;
  $mers =~ s/\r//g;
  next unless $mers;
  $saveFile = $1, next if ($mers =~ m/^save file begin (ecm[.]([1-9][0-9]+)[.]p[mp]1([.]M)?)$/);
  if ($mers =~ m/^save file end ([a-z][-._a-zA-Z0-9]+)$/)
    {
      die "save file end gives different file name, '$1', not expected '$saveFile'\n"
        unless $1 eq $saveFile;
      die "save file '$saveFile' has unexpected content\n"
        unless $sFcontent =~ m/^([-+=;:.@ A-Za-z0-9$CRLF]+)$/;
      print($fh "save file begin $saveFile$CRLF", "$sFcontent", "save file end $saveFile$CRLF")
        or die "could not print save file $saveFile to " . ($outfile ? $outfile : 'STDOUT') . ", $!";
      $saveFile = undef;
      $sFcontent = '';
      next;
    }
  $sFcontent .= "$mers$CRLF", next if ($saveFile);
  die "unexpected line or save file name: '$mers'\n"
    if $mers =~ m/^save file /;
  if ($mers !~ m/^M[(] [1-9][0-9]* [)]/o)
    {
      last if $mers =~ m/^done$/;
      print($fh "$mers\n");
      next if ($wait && $mers =~ m/ [Ww]ait[, ]/o); # wait for the server to find some relevant work
      last; # disconnect rather than wait
    }
  print($fh "$mers\n");
  if ($mers =~ m/$letter{$cmd}/o)
    {
      --$cnt;
      print($sock "done$CRLF"), last if $cnt < 1;
      $ask = 1;
      ++$gotOne;
      next;
    }
} continue {
  print($sock "work $cmd $prefix$CRLF") if $prefix ne '' && $ask;
  print($sock "work $cmd$CRLF") if $prefix eq '' && $ask;
  $ask = 0;
}
close($sock) or die "close of socket failed, $!";
print(STDOUT "could not get all work requested; $cnt missing\n") if $cnt != 0;
$gotOne or die "no work from the server";
exit(0);
