#!/usr/bin/perl -Tw
# Check ECM save files for a particular Mersenne so mersServer.pl doesn't have to.

use strict;
use strict 'vars';
use warnings;
use POSIX;
use Math::BigInt try => 'GMP';

BEGIN { push(@INC, '.') if -s('Mersenne.pm'); $ENV{'PATH'} = "/usr/bin:/bin"; }
use Mersenne;
use MersInfo;
use ECMSave;

my ($usage) = ("Usage: checkECMfiles.pl [--warn] [--delete] [-q|--quiet] <exponent>\n"
               . "    [-d [debugLevel]] [--check] [--help|-h] [--version]\n"
               . "  Check ECM save files of the Mersenne with the given exponent.\n"
               . "  -q or --quiet are toggles to print the 'best' save file content\n"
               . "    or just the file names of the 'best' save files.\n"
               . "  --warn and --delete are toggles for the last two arguments of\n"
               . "    of ECMSave::findBestSaveFile(), default off.\n"
               . "  There can be up to four 'best' save files for a single Mersenne number.\n"
  );

my ($debug) = 0; # from -d if given
my ($errors) = '';
my ($checkConfig) = scalar(grep { m/^--check$/ } @ARGV) if @ARGV;
my ($version) = MersInfo::version();
print(STDOUT "$0 version $version\n") if @ARGV && ($checkConfig || scalar(grep { $_ eq '--version' } @ARGV));
print(STDOUT "$usage")
  if scalar(grep { m,^(-h|--help)$, } @ARGV);
exit(0) if scalar(grep { m,^(--version|--help|-h), } @ARGV);

# Ignore STDIN ...
close(STDIN)
  or die("could not close STDIN, $!");
# ... but perl -w complains if STDIN is opened only for writing.
open(STDIN, '<', '/dev/null')
  or die("could not open /dev/null for STDIN, $!");

my ($exp, $delete, $quiet, $warn);
while (@ARGV)
{
  $_ = $ARGV[0];
  shift, next if m/^--check$/;
  shift, $debug = shift, next if m,^-d$, && $#ARGV > 0;
  $debug = $1, shift, next if m,^-d([1-9][0-9]*)$,;
  $quiet = !$quiet, shift, next if m,^-(-quiet|q)$,;
  $warn = !$warn, shift, next if m,^--warn$,;
  $delete = !$delete, shift, next if m,^--delete$,;
  $errors .= "only one argument is supported - '$exp' and '$_' given\n",
    last
    if $exp;
  $exp = $1,
    shift,
    next
    if m/^([1-9][0-9]+)$/;
  $errors .= "unknown option or argument '$_'\n";
}
$errors .= "no exponent given\n"
  unless $exp || $checkConfig;
die("$0 - problems:\n$errors$usage")
  if $errors;
exit(0) if $checkConfig;

# Do not delete files modified after this time(),
#   either a day ago (normally) or a minute ago (during testing).
$delete = time() - (-f('mysql.devel') ? 60 : 24*60*60)
  if $delete;

my ($quit);
$SIG{'HUP'} = "IGNORE";
foreach ('INT', 'TERM')
{
  $SIG{$_} = sub {
    my ($sig) = @_;
    warn "caught SIG$sig\n";
    $quit = 1;
  };
}

eval
{ # setpriority() is not always implemented
  setpriority(0, 0, 20);
};

my ($mers) = Mersenne->new('n' => $exp);
my ($best) = ECMSave::findBestSaveFile($exp, 'pm1', '', $mers, $warn, $delete);
($quiet ? print($best->bound() . ' ' . $best->file() . "\n") : $best->print())
  if $best && $best->bound() && -s $best->file();
$best = ECMSave::findBestSaveFile($exp, 'pp1', '', $mers, $warn, $delete);
($quiet ? print($best->bound() . ' ' . $best->file() . "\n") : $best->print())
  if $best && $best->bound() && -s $best->file();
exit(0)
  unless $exp % 8 == 4;
$best = ECMSave::findBestSaveFile($exp, 'pm1', '.M', $mers, $warn, $delete);
($quiet ? print($best->bound() . ' ' . $best->file() . "\n") : $best->print())
  if $best && $best->bound() && -s $best->file();
$best = ECMSave::findBestSaveFile($exp, 'pp1', '.M', $mers, $warn, $delete);
($quiet ? print($best->bound() . ' ' . $best->file() . "\n") : $best->print())
  if $best && $best->bound() && -s $best->file();
exit(0);
