static const char program_name[] = "contract";
/*
	Creates a file DATABASE (George Woltman's format)
	  from a file in Will Edgington's "mers" format.

	Rounds the trial factoring extent down to the nearest power of 2.

	08/05/97: Jean-Charles MEYRIGNAC
	16 Aug 1997 & later, Will Edgington
*/

#include "version.h"

#ifdef macintosh
# include <console.h>
# define ENOENT -43
# ifndef isascii
#  define isascii(x) (((x) | 0177) == 0177)
# endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#if defined(sun) || defined(__alpha) || defined(linux) || defined(__sgi__) || defined(__NetBSD__) || defined(macintosh) || defined(__CYGWIN__) || defined(__APPLE__)
# include <errno.h>
#endif

#include <math.h>
/* math.h is only needed for log() to take base 2 logarithms; if you don't have */
/*  log(), you'll have to find the largest power of two that's less than a given */
/*  integer in some other way */

/* the natural logarithm of two */
#define LN2 (0.693147180559945309417232121458176568075500134360255254120680009493393621969694715605863326996418687542001481020570685733)
#ifndef log2
#define log2(x) (log(x)/LN2)
#endif

const char Usage[] = "Usage: contract [-fhkuv] [-i inputfile] [-o outputfile]\n"
  "    Purpose: create DATABASE format files containing Mersenne trial\n"
  "        factoring extent data to the nearest power of 2.\n"
  "    Note: each option must have its own '-'; do not group them.\n"
  "    -f: toggle: store power of 2 just less than the smallest factor in\n"
  "        the output file rather than factoring extents (default: false).\n"
  "    -h: print this usage information\n"
  "    -i <inputfile>: specify filename to read from\n"
  "    -k: toggle: store power of 2 just less than the k value (from 2*k*p+1)\n"
  "        for factors rather than factors themselves (default: false)\n"
  "    -o <outputfile>: specify filename to write into (cannot exist)\n"
  "    -r<min.exp> <max.exp>: restrict the exponents to be recorded\n"
  "        (default: 7 to (2^32 - 1))\n"
  "    -u: toggle: only store factoring extents for exponents that might\n"
  "        be Mersenne primes (i.e., no known factor and no completed\n"
  "        Lucas-Lehmer test) (default: false)\n"
  "    -v: verbosity level: if not given, no output except to -o's file;\n"
  "        if given once, print counts of exponents at each extent;\n"
  "        if twice, also print 'exponent,extent' for skipped data\n"
  "            (w/too large an extent to be recorded correctly);\n"
  "        if thrice, also print 'exponent,extent' for each entry;\n"
  "        if four times, back to default (as if no -v was given)\n"
  "    No exponent will appear more than once in the output file\n"
  "        if input is sorted by exponent.\n"
  "    If the input is not sorted by exponent, a warning will be printed.\n"
  "    The primary use of the -f and -k options is probably to get\n"
  "        the counts of exponents rather than the output file.\n"
  "    See mersfmt.txt for input formats.\n"
  "--wedgingt@acm.org "
  ;

const char *infile = NULL;
const char *outfile = NULL;
FILE *in, *out;

#if defined(linux) && defined(__GNUC__) && defined(__i386__) && !defined(DO_NOT_USE_LONG_LONG_EXPONENT)
typedef unsigned long long UL;
# define FMT_UL "%Lu"
#else
typedef unsigned long UL;
# define FMT_UL "%lu"
#endif

/* Maximums inherent in the format of DATABASE files */
/* First, the largest exponent that fits in 3 bytes: */
#define MAX_SIMPLE_EXPONENT ((UL)0xFFFFFF)
/* Second, the largest exponent that, in effect, fits in 4 bytes: */
#define MAX_0xFF_EXPONENT (0xFF + 0x100*((UL)0xFFFFFF))
/* Third, the flag that indicates whether the corresponding Mersenne number is */
/*   known to be composite or not: */
#define COMPOSITE_FLAG (0x80)
/* Fourth, the constant in the fourth byte that is used to indicate */
/*   exponents in the range 2^24 thru 2^32 - 1 inclusive. */
#define BIG_EXPONENT (0xFF)
/* Fifth, the constant in the fourth byte that is used to indicate */
/*   exponents that cannot fit in three bytes: */
#define LARGE_EXPONENT (0x7F)
/* Sixth, the largest factoring extent; it's basicly what's left of */
/*   the four bytes in each DATABASE entry after the above. */
#define MAX_POWER_OF_2 (0x7E)

/* To count how many exponents there are for each (power of 2) factoring extent. */
/*   [MAX_POWER_OF_2 + 1] entry counts how many exponents exceeded MAX_POWER_OF_2. */
int stats[MAX_POWER_OF_2 + 2];

int main(int argc, char **argv)
 {
  UL exponent = 0, old = 0; /* exponent from this line of input and from prior line */
  UL first = (UL)7, last = ~(UL)0; /* smallest and largest exponents, from -r */
  UL addition = 0; /* 0xFF or 0x7F style addition to the exponent */
  char M; /* letter of lowM.txt format: 'U', 'I', 'H', 'C', 'D', 'd', 'P', etc. */
  unsigned char p; /* power of 2 extent of trial division */
  int n = 0; /* several uses */
  int cnt = 0; /* count of distinct exponents read */
  int skipped = 0; /* count of apparent factors seen in 'exponent,extent' format: ignored */
  double f; /* extent of trial factoring */
  char verbose = 0; /* how much to print to user/stdout */
  /* rest are flags/booleans/toggles: */
  char bare_exponent = 0; /* just the exponent? or with a trial factoring extent or other data? */
  char store_k = 0; /* -k/-K: store the lengths of the first factors divided by 2*exponent */
  char store_factors = 0; /* -f/-F: store the lengths of the first factors in bits, not factoring extent */
  char only_U_in_db = 0; /* -u/-U: only put unknown primality Mersennes into DATABASE */
  char warned = 0; /* warn user the first time exponents read decrease */

#ifdef macintosh
  argc = ccommand(&argv);
#endif
  for (n = 1; n < argc; n++)
   {
    if (argv[n][0] != '-' || argv[n][1] == '\0')
     {
      fprintf(stderr, "%s: unknown argument: '%s'\n%s", program_name, argv[n], Usage);
      return(1);
     }
    switch (argv[n][1])
     {
      case 'f':
      case 'F':
        store_factors = !store_factors;
        break;
      case 'h':
      case 'H':
      case '?':
        fprintf(stderr, "%s\n", Usage);
        return(0);
      case 'i':
      case 'I':
        if (infile != NULL)
          fprintf(stderr, "%s: warning: previous input file '%s' ignored\n",
                  program_name, infile);
        if (argv[n][2] != '\0')
          infile = argv[n] + 2;
        else
          if (n < argc - 1)
            infile = argv[++n];
          else
            infile = NULL;
        if (n >= argc || infile == NULL || infile[0] == '\0')
         {
          fprintf(stderr, "%s: no input file given after -i\n%s\n",
                  program_name, Usage);
          return(1);
         }
        break;
      case 'k':
      case 'K':
        store_k = !store_k;
        break;
      case 'o':
      case 'O':
        if (outfile != NULL)
          fprintf(stderr, "%s: warning: previous output file '%s' ignored\n",
                  program_name, outfile);
        if (argv[n][2] != '\0')
          outfile = argv[n] + 2;
        else
          if (n < argc - 1)
            outfile = argv[++n];
          else
            outfile = NULL;
        if (outfile == NULL || outfile[0] == '\0')
         {
          fprintf(stderr, "%s: no output file given after -o\n%s\n",
                  program_name, Usage);
          return(1);
         }
        break;
      case 'r':
      case 'R':
        {
         char *next = NULL;
         errno = 0;
         if (isascii(argv[n][2]) && isdigit(argv[n][2]))
           first = strtoul(argv[n] + 2, &next, 10);
         else
           if (n < argc - 1 && argv[++n] != NULL && isascii(argv[n][0]) && isdigit(argv[n][0]))
             first = strtoul(argv[n], &next, 10);
           else
             first = 0L;
         if (n < argc - 1 && argv[++n] != NULL && isascii(argv[n][0]) && isdigit(argv[n][0]))
           last = strtoul(argv[n], &next, 10);
         else
           last = 0L;
         if (first > last || first < 7L || errno != 0 || next == NULL || next[0] != '\0')
          {
           fprintf(stderr, "%s: smallest exponent must be at least 7 and %s\n%s",
                   program_name, "less than largest exponent\n", Usage);
           return(1);
          }
        }
        break;
      case 'u':
      case 'U':
        only_U_in_db = !only_U_in_db;
        break;
      case 'v':
      case 'V':
        verbose++;
        if (verbose > 3)
          verbose = 0;
        break;
      default:
        fprintf(stderr, "%s: unknown option '%s'\n%s",
                program_name, argv[n], Usage);
        return(1);
     }
   }
  errno = 0;
  if (infile == NULL)
    in = stdin;
  else
    if ((in = fopen(infile, "r")) == NULL)
     {
      perror(infile);
      return(errno);
     }
  if (outfile == NULL)
    fprintf(stderr, "%s: writing to %s\n", program_name, outfile = "DATABASE");
  if ((out = fopen(outfile, "rb")) != NULL)
   {
    fprintf(stderr, "%s: output file %s already exists\n", program_name, outfile);
    exit(1);
   }
  if (errno == ENOENT)
    errno = 0;
  if ((out = fopen(outfile, "wb")) == NULL)
    perror(outfile);
  if (errno != 0)
    return(errno);
  for (n = 0; n < sizeof(stats)/sizeof(stats[0]); n++)
    stats[n] = 0;
  while (!feof(in) && !ferror(in) && !ferror(out) && (n = fgetc(in)) != EOF)
   {
    while (n != EOF && isascii(n) && isspace(n))
      n = fgetc(in); /* some fscanf()s require white space in input if present at start of format string */
    if ((bare_exponent = !!(isascii(n) && isdigit(n))) != 0)
      ungetc(n, in);
    if ((!bare_exponent && fscanf(in, "( " FMT_UL " )%c", &exponent, &M) < 2) ||
        (bare_exponent && (M = 'U', fscanf(in, FMT_UL, &exponent)) < 1))
     {
      fprintf(stderr, "unrecognized input near or after 'M( " FMT_UL " )' in %s\n", exponent, infile);
      break;
     }
    if (exponent < old && !warned)
     {
      warned = 1;
      fprintf(stderr, "%s: warning: exponents not in numerically increasing order near " FMT_UL "\n",
              program_name, exponent);
     }
    if (exponent < first || exponent > last || exponent == old)
      /* arrange to set old, read the rest of the line, and continue */
      M = 'D';
    switch (M)
     {
      case 'D': case 'd': case 'P': /* all factors known, cofactor is pseudo-prime, or prime */
      case 'l': case 'm': case 'L': case 'M': /* # of digits or pseudo-prime halves info */
        old = exponent;
        /* fall thru */
      case 'C': /* factor */
        if (only_U_in_db)
          old = exponent;
        if (old != exponent && (store_factors || store_k))
          break;
        /* fall thru */
      case 'c': case 'e': case 'E': case 'o': case 'q': /* # of digits, ECM, P-1, or P+1 data */
      case 'X': /* 1strs data: somewhat relevant, but no way to record in this format */
        /*!!Well, it could be used to set minimal factoring extents for each exponent */
        while ((n = fgetc(in)) != EOF && n != '\n')
          ; /* skip rest of line */
        continue;
      case 'G': case 'H': case 'I': case 'J': case 'U':
        /* gap, highest attempted, interrupted, or unknown primality w/ extent of search */
        if ((only_U_in_db && M == 'H') || store_factors || store_k)
         {
          old = exponent;
          while ((n = fgetc(in)) != EOF && n != '\n')
            ; /* skip rest of line */
          continue;
         }
        break; /* handled outside switch() in case we need to break out of the loop */
      default:
        fprintf(stderr, "%s: unrecognized mersfmt flag letter after 'M( " FMT_UL " )' of '%c' in %s\n",
                program_name, exponent, M, infile);
        return(1);
     }
    old = exponent; /* so the output (DATABASE style) file gets at most one entry for any exponent */
    if (!bare_exponent && ((n = fgetc(in)) != ':' || (n = fgetc(in)) != ' '))
     {
      fprintf(stderr, "unrecognized input after 'M( " FMT_UL " )' in %s\n", exponent, infile);
      return(1);
     }
    if (M == 'J') /* 'J'oined gap data, so the extent of factoring is the second number, not the first */
     {
      while (n != EOF && isascii(n) && isspace(n))
        n = fgetc(in);
      if (!isdigit(n))
       {
        (void)fprintf(stderr, "%s: invalid mersfmt 'J' line for exponent " FMT_UL " in %s\n",
                      program_name, exponent, infile);
        return(1);
       }
      while (n != EOF && isascii(n) && isdigit(n))
        n = fgetc(in);
      if (n != ' ')
       {
        (void)fprintf(stderr, "%s: invalid mersfmt 'J' line for exponent " FMT_UL " in %s\n",
                      program_name, exponent, infile);
        return(1);
       }
     }
    if (bare_exponent)
      if ((n = fgetc(in)) != ',')
       {
        if (n != EOF && n != '\n')
         {
          (void)fprintf(stderr, "%s: unknown format (exponent '%c') near 'exponent' " FMT_UL "\n",
                        program_name, n, exponent);
          return(1);
         }
        p = 0; /* completely bare exponent, on a line by itself */
        /*!!If results.x were available, could get implied extent from it */
       }
      else /* looking like exponent','extent format */
        if (fscanf(in, "%d", &n) != 1)
         {
          (void)fprintf(stderr, "%s: unknown format (exponent ',' non-digits) near 'exponent' " FMT_UL "\n",
                        program_name, exponent);
          return(1);
         }
        else
          if (n > MAX_POWER_OF_2 || n < 0)
           { /* n < 0 means it overflowed 32 bits (or was preceded by an (unexpected) '-'); */
            skipped++; /* more likely to be a factor than to be the factoring extent */
            continue;
           }
          else /* finally: looks OK to store */
            p = (unsigned char)n;
    else /* this line is in mersfmt: get the trial factoring extent as a double */
      if (fscanf(in, "%lf", &f) != 1)
       {
        fprintf(stderr, "%s: unrecognized (mersfmt) input after 'M( " FMT_UL " )%c: ' in %s\n",
                program_name, exponent, M, infile);
        return(1);
       }
      else /* check for '2^56' format */
        if (fabs(f - 2.0) < 0.1 && (n = fgetc(in)) == '^' && fscanf(in, "%d", &n) == 1)
         {
          if (n > MAX_POWER_OF_2) /* too big to store */
            p = MAX_POWER_OF_2 + 1;
          else
            p = (unsigned char)n;
          if (p == 0)
            f = 1.5;
          else
            if (M == 'I' /* && p > 0 */ )
              f = pow(2.0, --p + 0.5);
            else
              f = pow(2.0, (double)p);
         }
        else
         {
          if (M == 'I') /* 'I'nterrupted; likely obsolete now, but ... */
            f -= 2*(double)exponent;
          if (!store_factors && !store_k && (f - 1.0)/2.0 <= (double)exponent)
            p = 0, f = 1.5; /* extent is too small to matter */
          else
           {
            if (store_k)
              f /= 2*(double)exponent; /* factor = 2*k*exponent + 1, so f is now k */
            if (f >= ldexp(1.0, MAX_POWER_OF_2))
              p = MAX_POWER_OF_2 + 1; /* too large an extent, factor, or k to store */
            else
             { /* Get p from f, watching out for truncation when close to next power of 2 */
              p = (unsigned char)(log2(f));
              if (!store_factors && !store_k && fabs(f - pow(2.0, p + 1.0)) < 0.1)
                p++;
             }
           }
         }
    /* note that certain strange input, like 'M( 11 )C: 2^\nM( ', might be accepted, harmlessly */
    while ((n = fgetc(in)) != EOF && n != '\n')
      ; /* skip rest of line */
    ++stats[p];
    ++cnt;
    if (p > MAX_POWER_OF_2 || !(bare_exponent || store_factors || store_k ||
                                (f >= pow(2.0, (double)p) - 0.1 && f < pow(2.0, p + 1.0))))
     { /* factoring extent is too large to record or something similar */
      if (verbose > 1)
        printf("# " FMT_UL "," FMT_UL "\n", exponent, (UL)log2(f));
      continue;
     }
    if (verbose > 2)
      printf(FMT_UL ",%d\n", exponent, p);
    if (M == 'H' && !store_factors && !store_k)
      p |= COMPOSITE_FLAG; /* at least one LL test or otherwise known to be a composite Mersenne */

    /* Finally: put something into the output file */
    if (exponent > MAX_0xFF_EXPONENT && addition < exponent >> 24)
     {
      if (warned)
       { /* The 'additions' in the format cannot be 'undone', so: */
        fprintf(stderr, "%s: cannot store out-of-order exponents larger than " FMT_UL " %s",
                program_name, (UL)MAX_0xFF_EXPONENT, "in DATABASE format\n");
        return(1);
       }
      /* Greater than 32 bits and haven't put enough additions in */
      while (addition + 0xFFFFFF < exponent >> 24)
       { /* add another 0xFFFFFF000000 to later exponents */
        fputc(0xFF, out);
        fputc(0xFF, out);
        fputc(0xFF, out);
        fputc(LARGE_EXPONENT, out); /* mark as not-an-entry, but an addition */
        addition += 0xFFFFFF;
       }
      if (addition < exponent >> 24)
       { /* store higher bits of the exponent that aren't added in already */
        fputc((int)(((exponent >> 24) - addition) & 0xFF), out);
        fputc((int)((((exponent >> 24) - addition) >> 8) & 0xFF), out);
        fputc((int)((((exponent >> 24) - addition) >> 16) & 0xFF), out);
        fputc(LARGE_EXPONENT, out); /* mark as high bits of a large exponent */
        addition = exponent >> 24;
       }
     }
    else
      if (exponent > MAX_SIMPLE_EXPONENT && addition < exponent >> 24)
       { /* store the high 8 bits of an exponent between 2^24 and 2^31 - 1 inclusive */
        addition = exponent >> 24;
        fputc(0, out);
        fputc(0, out);
        fputc((int)addition & 0xFF, out);
        fputc(BIG_EXPONENT, out); /* mark as high 8 bits of a 32 bit exponent */
       }
    fputc((int)(exponent & 0xFF), out); /*!!these three lines depend on value of MAX_SIMPLE_EXPONENT */
    fputc((int)((exponent >> 8) & 0xFF), out);
    fputc((int)((exponent >> 16) & 0xFF), out);
    fputc(((int)p) & 0xFF, out); /*!!this line depends on value of both COMPOSITE_FLAG and MAX_POWER_OF_2 */
   }
  if (ferror(in))
    fprintf(stderr, "%s: stdio error(s) occured reading %s\n", program_name, infile);
  if (ferror(out))
    fprintf(stderr, "%s: stdio error(s) occured writing %s\n", program_name, outfile);
  if (errno == 0 && ferror(out))
    errno = ferror(out);
  if (errno == 0 && ferror(in))
    errno = ferror(in);
  if (errno == 0 && !feof(in))
    errno = feof(in);
  (void)fclose(in);
  (void)fclose(out);
  if (cnt > 0 && verbose > 0)
   {
    printf("# %d exponents/Mersenne numbers\n# %6s  Extent of factoring\n", cnt, "Count");
    for (n = 0; n < sizeof(stats)/sizeof(stats[0]); ++n)
      if (stats[n] > 0)
        printf("# %8d  2^ %3d\n", stats[n], n);
   }
  if (skipped > 0)
    printf("# skipped %d apparent factors in 'exponent,factor' format\n", skipped);
  return(errno);
 }
