#!/usr/bin/perl -Tw
# Use the Linux ecm program to run one ECM task from the server.

use strict;
use strict 'vars';
use warnings;
use POSIX;
use Math::BigFloat;
use Math::BigInt try => 'GMP';

BEGIN { push(@INC, '.') if -s('Mersenne.pm'); $ENV{'PATH'} = "/usr/bin:/bin"; }
use Mersenne;
use MersInfo;

my ($usage) = ("Usage: doECM.pl [-maxmem <amount>] [--<duration|time> <seconds>] [files]\n"
               . "  [--stop <directory>] [--skip <directory>] [--tmp <directory>]\n"
               . "  [--stop=<directory>] [--skip=<directory>] [--tmp=<directory>]\n"
               . "  [-d [debugLevel]] [--check] [--help|-h] [--version]\n"
               . "Use ECM to try to find new factors of the Mersenne number.\n"
               . "  STDIN should be mers format information about the Mersenne.\n"
               . "The value given with -maxmem is passed on to the Linux 'ecm' program.\n"
               . "Once a given duration or time has passed, the script will exit the next time the ECM program exits.\n"
               . "  The default duration is 10 minutes (600 seconds).\n"
               . "If a new factor is found, it will be printed in mers format and this script will exit.\n"
               . "The ECM bound and count of curves at that bound will also be printed in mers format.\n"
               . "  The bound will be increased appropriately after each run of the ECM program.\n"
               . "Note that all files given should contain data for the same Mersenne number.\n"
               . "  STDIN will be read if no files are given.\n"
  );

# Stop if any of these directories are created, which should match the ones used by doWork.pl.
my (@stopDirs) = ( 'working.doWork/stop', 'working.doWork/skipecm' );
# Also stop if a factor is found or this many seconds have passed:
my ($duration) = 10*60; # default, in seconds - ten minutes
$duration = 30
  if -f 'mysql.devel'; # 30 seconds if testing - 'make tests'

my ($subdir) = 'ecm-work'; # subdirectory for related data files

my ($debug) = 0; # from -d if given

my ($errors) = '';
my ($checkConfig) = scalar(grep { m/^--check$/ } @ARGV) if @ARGV;
my ($version) = MersInfo::version();
print(STDOUT "$0 version $version\n") if @ARGV && ($checkConfig || scalar(grep { $_ eq '--version' } @ARGV));
print(STDOUT "$usage")
  if scalar(grep { m,^(-h|--help)$, } @ARGV);
exit(0) if scalar(grep { m,^(--version|--help|-h), } @ARGV);

my ($maxmem);
my (@files);
while (@ARGV)
{
  $_ = $ARGV[0];
  shift(), next if m/^--check$/;
  shift(), $debug = shift(), next if m,^-d$, && $#ARGV > 0;
  $debug = $1, shift(), next if m,^-d([1-9][0-9]*)$,;
  shift(), $maxmem = shift(), next if m,^-maxmem$, && $#ARGV > 0;
  shift(), $stopDirs[0] = shift(), next if m,^--stop$, && $#ARGV > 0;
  shift(), $stopDirs[1] = shift(), next if m,^--skip$, && $#ARGV > 0;
  # The tmp directory is not needed.
  shift(), shift(), next if m,^--tmp$, && $#ARGV > 0;
  $stopDirs[0] = $1, shift(), next if m,^--stop=(.+)$,;
  $stopDirs[1] = $1, shift(), next if m,^--skip=(.+)$,;
  shift(), next if m,^--tmp=(.+)$,;
  shift(), $duration = shift(), next if m,^--(duration|time)$, && $#ARGV > 0;
  $duration = $2, shift(), next if m,^--(duration|time)=([1-9][0-9]+)$,;
  push(@files, shift()),
    next
    if -f && -s && ! -l;
  $errors .= "unknown option or argument '" . shift() . "'\n";
}
$errors .= "debug level, -d, should be a number - '$debug' given\n"
  unless !defined($debug) || $debug =~ m,^([0-9]+)$,;
$debug = "$1"
  if defined $debug;
$errors .= "-maxmem should be a number greater than 9 - '$maxmem' given\n"
  unless !defined($maxmem) || $maxmem =~ m,^([1-9][0-9]*)[km]?$,;
$maxmem = $1
  if defined $maxmem;
$errors .= "duration, --duration or --time, should be a positive number - '$duration' given\n"
  unless !defined($duration) || $duration =~ m,^([1-9][0-9]*)$,;
$duration = $1
  if defined $duration;
$errors .= "unexpected characters in stop directory - '$stopDirs[0]'\n"
  unless $stopDirs[0] =~ m,^([-.+/[:alnum:]]+)$,;
$stopDirs[0] = $1;
$errors .= "unexpected characters in skip directory - '$stopDirs[1]'\n"
  unless $stopDirs[1] =~ m,^([-.+/[:alnum:]]+)$,;
$stopDirs[1] = $1;

die("$0 - problems:\n$errors$usage")
  if $errors && !$checkConfig;
warn("$errors")
  if $errors;

my ($minECMBound) = Math::BigInt->new(120);

$SIG{'HUP'} = "IGNORE";
foreach ('INT', 'TERM')
{
  $SIG{$_} = sub {
    my ($sig) = @_;
    warn "caught SIG$sig\n";
  };
}

# Based on prior effort (sum of all the stage 1 bounds used), calculate the bound to use next.
sub boundFromEffort($)
{
  my ($effort) = @_;
  return($minECMBound) if !$effort || $effort <= $minECMBound;
  my ($bFlt) = Math::BigFloat->new($effort);
  $bFlt->blog()->bmul(3)->bdiv(4)->bexp();
  my ($bInt) = $bFlt->as_int();
  return($minECMBound) if $bInt <= $minECMBound;
  return($bInt) if $bInt < $Mersenne::maxStage1Bound;
  return($Mersenne::maxStage1Bound);
}

-d("$subdir") || mkdir("$subdir") or $errors .= "could not mkdir $subdir, errno $!\n";
chdir("$subdir") or $errors .= "could not chdir to $subdir, errno $!\n";
s,^,../, foreach @stopDirs;
s,^,../, foreach grep { ! m,^/, } @files;

MersInfo::rotateLog('log.txt', 99);
my ($logFH);
open($logFH, '>>', 'log.txt') or $errors .= "cannot append to log.txt, $!\n";
print($logFH 'Start at ' . localtime() . "\n");
die $errors if $errors;

exit(0) if $checkConfig;

# Program and arguments for factoring Mersenne numbers.
# The 'ecm' program is available for most Linux distributions
#   and can do ECM, P-1, and P+1 work, including resuming from
#   a checkpoint file.
my ($ecm);
foreach (split(/:/, $ENV{'PATH'}))
{
  next unless -d;
  $ecm = "$_" . '/ecm', last if -x "$_" . '/ecm' && ! -d "$_" . '/ecm';
}
defined($ecm) or die 'cannot find ecm program in $PATH' . "\n";

$ecm .= ' -v -timestamp';
!defined($maxmem) or $maxmem =~ m/^([0-9]+[km]?)$/ or die "-maxmem given with unrecognized value: '$maxmem'\n";
$ecm .= " -maxmem $1" if defined $maxmem;
$ecm = "echo debug $debug $ecm" if $debug;

eval
{ # setpriority() is not always implemented
  setpriority(0, 0, 20);
};

# Retain intended output in $mers to avoid printing anything until just before
#   exiting so the caller isn't blocked waiting for the rest of the output.
my ($mers) = '';

@ARGV = @files;
my ($e, $m, $bnd, $effort, $LorM);
my (%digits);
my ($cnt) = 0;
 INPUT:
while (<<>>)
{
  print($logFH "$_");
  chomp;
  next unless $_;
  m/^M[(] ([1-9][0-9]*) [)](.): ([0-9][0-9]*)/o
    or die "$ARGV:$.: unrecognized input: $_\n";
  $e = Math::BigInt->new("$1"),
    $m = Mersenne->new('n', $e)
    if !defined $e;
  "$e" eq "$1"
    or die "$ARGV:$.: input contains data for M($e) and M($1)\n";
  my ($l) = $2;
  warn("$ARGV:$.: M($1)$l - already completely factored or prime\n"),
    $m->print(),
    print("$_\n"),
    exit(0)
    if $l =~ m,^[SDP],;
  $digits{$l} = $1,
    next
    if m/^M[(] [1-9][0-9]* [)][clmLM]: ([1-9][0-9]*)/o;
  next
    if $l =~ m/^[oqGHJULM]$/; # ignore P-1, P+1, gaps, and trial factoring extents
  my ($f) = Math::BigInt->new("$3");
  if ($l eq 'e')
    { # indicates a single curve to be performed; deprecated - use 'E' lines
      $bnd = $f;
      $effort = ($bnd < $minECMBound) ? 0 : $bnd;
      next;
    }
  if ($l eq 'E')
    { # indicates the ECM 'effort' (sum over curves done of the bound of each curve) previously done
      $m->parse($_)
        or die("$ARGV:$.: unexpected - Mersenne::parse() failed");
      $effort = $f;
      $bnd = boundFromEffort($effort);
      next;
    }
  if ($l eq 'C')
    {
      my (@factors) = $m->f($f);
      while (@factors)
        {
          next INPUT
            if $f == $factors[0];
          $f /= $factors[0]
            while $f % $factors[0] == 0;
          next INPUT
            if $f < 2;
          shift(@factors);
        }
      $mers .= "$_ # is not a factor\n"
        unless @factors;
      next;
    }
  die "$ARGV:$.: unexpected input - $_\n";
}

warn("$ARGV:$.: M($e)LM - already completely factored\n"),
  $m->print(),
  exit(0)
  if exists($digits{'L'}) && exists($digits{'M'});

$effort = $minECMBound if (!defined($effort));

# Always do at least one curve.
my ($curves) = 1;
$bnd = $minECMBound
  unless defined $bnd;
while ($bnd < $minECMBound)
{
  $effort += $minECMBound;
  ++$curves;
  $bnd = boundFromEffort($effort);
}

$digits{'l'} = $e if !$digits{'l'} && $e % 8 == 4 && !$digits{'L'};
$digits{'m'} = $e if !$digits{'m'} && $e % 8 == 4 && !$digits{'M'};
$digits{'c'} = $e if !$digits{'c'} && $e % 8 != 4;
delete($digits{'l'}), delete($digits{'m'}), delete($digits{'L'}), delete($digits{'M'}) unless $e % 8 == 4;
delete($digits{'c'}) if $e % 8 == 4;

my (@cofactors) = $m->c();

# The ecm program can try to factor more than one number in a single
#   run so give it both cofactors when there are two.
my ($numFH);
open($numFH, '>', "ecm.$e.ECM.txt") or die "could not create ecm.$e.ECM.txt, $!\n";
print($numFH "$cofactors[0]\n") if $digits{'c'};
print($numFH "$cofactors[0]\n") if $digits{'l'};
print($numFH "$cofactors[1]\n") if $digits{'m'};
close($numFH) or die "close of ecm.$e.ECM.txt (cofactor input to ecm) failed, $!\n";

END
{
  unlink("ecm.$e.ECM.txt")
    if $e;
}

my ($stopTime) = time() + $duration;
my (@newFactors);
while (time() < $stopTime)
{
  my ($runFH);
  open($runFH, "$ecm -c $curves -one -inp ecm.$e.ECM.txt $bnd 2>&1 |") or die "could not run ecm, $!\n";
  my (%lines);
  while (<$runFH>)
    {
      print($logFH "$_");
      chomp;
      $lines{$1} = $_,
        next
        if m,^(GMP-ECM) ,;
      $lines{'input'} = $1,
        next
        if m,^Input number is ([1-9][.0-9]+) ,;
      push(@newFactors, Math::BigInt->new("$1")),
        $mers .= 'M( ' . "$e" . ' )C: ' . "$1\n",
        next
        if (m/^Found .* factor of .* digits: ([1-9][0-9]*)$/);
      $lines{'B1'} = $1,
        $lines{'B2'} = $2,
        next
        if (m/^Using B1=([1-9][0-9]*), B2=([1-9][0-9]*)/);
      $lines{'B1'} = $1
        if (m/^Using B1=[1-9][0-9]*-([1-9][0-9]*)/);
    }
  if (!close($runFH))
    { # The ecm program exits non-zero if a factor found, so this is messy.
      my ($status) = $? >> 8; # see 'man perlvar'
      # See 'man ecm' for why these exit statuses are OK when a factor is found.
      @newFactors && !($status & ~(2|4|8))
        or die "ecm program failed, exiting $?";
    }
  last if $debug; # output won't include the expected %lines
  exists($lines{'GMP-ECM'})
    or die "ecm program output format has changed - 'GMP-ECM' with version not printed\n";
  exists($lines{'input'})
    or die "ecm program output format has changed - 'Input number' not printed\n";
  exists($lines{'B1'}) && exists($lines{'B2'})
    or die "ecm program output format has changed - 'Using B1=' not printed or did not include 'B2='\n";
  $mers .= 'M( ' . "$e" . ' )e: ' . "$lines{'B1'} $curves\n";
  $effort += $bnd*$curves;
  $bnd = boundFromEffort($effort);
  $curves = 1;
  last if @newFactors || -d $stopDirs[0] || -d $stopDirs[1];
}

$mers =~ s/\\\n//g; # join folded lines
print(STDOUT "$mers");
exit(0);
