#!/usr/bin/perl -Tw
# Run one of the GPU trial factorers, to either close a trial factoring
#   gap (mers format 'G' or 'g') or to do a Trial Factoring (2) task
#   from GIMPS, either read from STDIN.
# If running the GPU trial factoring program on MSWindows with Cygwin
#   fails with errors that cudart64_12.dll cannot be found, copy it
#   from the NVIDIA tools bin directory to the same directory as the
#   GPU trial factoring program.  The Cygwin ldd program should print
#   the library (*.dll) files that any program uses.
use strict;
use strict 'vars';
use Math::BigInt try => 'GMP';
use Scalar::Util; # tainted()

BEGIN { push(@INC, '.') if -s('Mersenne.pm'); $ENV{'PATH'} = "/usr/bin:/bin"; }
use Mersenne;
use MersInfo;

# If no GPU is found or the mfakt* program prints an unexpected error (starting with 'ERROR:'), this is set and the script will fail.
my ($noGPU) = 0;

# Names of GPU trial factoring programs.
#   Leave off '.exe'; it will be checked as well.
my (@gpuPgms) = ('mfaktc', 'mfakto', );

my ($usage) = ("Usage: doGPU.pl [-m <trials>] [highBits]\n"
               . "  [--stop <directory>] [--skip <directory>] [--tmp <directory>]\n"
               . "  [--stop=<directory>] [--skip=<directory>] [--tmp=<directory>]\n"
               . "  [--check] [--debug] [--help|-h] [--version]\n"
               . "Use a GPU trial factoring program to try to find new factors of the Mersenne number.\n"
               . "  STDIN should be mers format information about the Mersenne, usually 'G' or 'g' (gap) lines.\n"
               . "The programs searched for are " . join(' ', @gpuPgms) . "\n"
  );

# Largest Mersenne exponent that is the smallest supported by known versions of mfakt*.
#   May be reduced inside getGPUlimits().
my ($smallestExp) = 1e5;

my ($errors) = '';
my ($checkConfig) = scalar(grep { m/^--check$/ } @ARGV) if @ARGV;
my ($version) = MersInfo::version();
print(STDOUT "$0 version $version\n")
  if ($checkConfig || grep { $_ eq '--version' } @ARGV);
print(STDOUT "$usage")
  if grep { m,^(--help|-h)$, } @ARGV;
exit(0)
  if @ARGV && grep { m,^(--version|--help|-h)$, } @ARGV;

my ($debug) = scalar(grep { m/^--debug$/ } @ARGV) if @ARGV;

# Do not print warnings immediately; save them for just prior to exiting
#   to avoid stalling doWork.pl while the work is being done.
my ($warnings) = '';

$SIG{'HUP'} = "IGNORE";
foreach ('INT', 'TERM')
{
  $SIG{$_} = sub
    {
      my ($sig) = @_;
      $warnings .= (localtime() . " caught SIG$sig\n");
    };
}

my ($hiBits);
my (@options);
my ($stopDir) = "working.doWork/stop";
my ($skipDir) = $stopDir;
my ($dashM);
while (@ARGV)
{
  shift, next if $ARGV[0] eq '--check';
  shift, shift, next if $ARGV[0] =~ m,^--tmp$, && $#ARGV > 0;
  shift, $stopDir = shift, next if $ARGV[0] =~ m,^--stop$, && $#ARGV > 0;
  shift, $skipDir = shift, next if $ARGV[0] =~ m,^--skip$, && $#ARGV > 0;
  $stopDir = $1, shift, next if $ARGV[0] =~ m,^--stop=(.+)$,;
  $skipDir = $1, shift, next if $ARGV[0] =~ m,^--skip=(.+)$,;
  shift, next if $ARGV[0] =~ m,^--tmp=(.+)$,;
  shift, $dashM = Math::BigInt->new("$1"), shift, next if $ARGV[0] =~ m,^-m$, && $#ARGV > 0 && $ARGV[1] =~ m,^([1-9][0-9]*)$,;
  $dashM = Math::BigInt->new("$1"), shift, next if $ARGV[0] =~ m,^--?m=([1-9][0-9]*)$,;
  $hiBits = $1, shift, next if $ARGV[0] =~ m,^([1-9][0-9]+)$,;
  push(@options, split(/=/, shift())), next if $ARGV[0] =~ m,=,;
  $errors .= "only a single numeric argument is supported; '$_' was given\n";
  shift;
}
$hiBits = 0 if !defined $hiBits;
$errors .= "unexpected characters in stop directory - '$stopDir'\n"
  unless $stopDir =~ m,^([-.+/[:alnum:]]+)$,;
$stopDir = $1;
$errors .= "unexpected characters in skip directory - '$skipDir'\n"
  unless $skipDir =~ m,^([-.+/[:alnum:]]+)$,;
$skipDir = $1;

die("$0 - problems:\n$errors")
  if $errors;

# The key is largest exponent that stops factoring at 2^value as of 6 May 2017.
# Further, the value is limited to $hiBits (the only supported command line
#   argument) except that the task read from STDIN can start and/or end past
#   $hiBits or the value from this hash.
my (%GIMPSTrialFactorLimit) = ( 23390000 => 66,
                                29690000 => 67,
                                37800000 => 68,
                                47450000 => 69,
                                58520000 => 70,
                                75670000 => 71,
                                96830000 => 72,
                                115300000 => 73,
                                147500000 => 74,
                                186400000 => 75,
                                227300000 => 76,
                                264600000 => 77,
                                337400000 => 78,
                                420400000 => 79,
                                516000000 => 80,
  );

my ($exe);
my (@exeDir); # array of path components; use join('/', @exeDir) to get the path.
my (@likelyPlaces) = ('cuda', 'github', 'git', 'mfaktc', 'mfakto', 'gpuowl', );
my (@try) = @likelyPlaces;
push(@try, "../$_")
  foreach (@likelyPlaces);
@exeDir = ('.', );
 FINDEXE:
while (@try && @exeDir && !$exe)
{
  foreach my $findExe (@gpuPgms)
    {
      $exe = join('/', @exeDir, $findExe), last FINDEXE if -x(join('/', @exeDir, $findExe)) && ! -d(join('/', @exeDir, $findExe));
      $exe = join('/', @exeDir, $findExe) . '.exe', last FINDEXE if -x(join('/', @exeDir, $findExe) . '.exe') && ! -d(join('/', @exeDir, $findExe) . '.exe');
    }
  if (-d(join('/', @exeDir, $try[0])))
    {
      push(@exeDir, $try[0]);
      @try = @likelyPlaces;
      next;
    }
  shift(@try);
  next if @try;
  last if !@exeDir;
  @try = @likelyPlaces;
  while (@exeDir)
    {
      shift(@try) while @try && $try[0] ne $exeDir[$#exeDir];
      pop(@exeDir);
      shift(@try) if @try;
      last if @try;
      @try = @likelyPlaces;
    }
  foreach my $findExe (@gpuPgms)
    {
      $exe = join('/', @exeDir, $findExe), last FINDEXE if -x join('/', @exeDir, $findExe) && ! -d join('/', @exeDir, $findExe);
      $exe = join('/', @exeDir, $findExe) . '.exe', last FINDEXE if -x(join('/', @exeDir, $findExe) . '.exe') && ! -d(join('/', @exeDir, $findExe) . '.exe');
    }
}
die("exe filepath is tainted")
  if Scalar::Util::tainted($exe);
if (!defined($exe) || ! -x($exe))
{
  $checkConfig or
    die "cannot find a GPU factoring program (mfaktc or mfakto)\n";
  print(STDOUT "cannot find a GPU factoring program (mfaktc or mfakto)\n");
  exit(0);
}
my ($exeDir) = join('/', @exeDir);

eval
{ # setpriority() is not always implemented
  setpriority(0, 0, 20);
};

my (%minExp);
# Read data relating the minimum Mersenne exponent to the count of GPU sieve primes from gpulimits.txt if it exists.
sub getGPUlimits();
getGPUlimits();

# Return the number of primes to use to sieve the factor candidates on the GPU
#   based on the Mersenne exponent.
sub findGPUSievePrimes($);

print(STDOUT " found $exe\n") if $checkConfig;

my ($exp, $letter, $start, $stop, $hex, $args);
my ($useMFG) = '';
my ($noFactor) = 1; # 1 is no factor found; 0 if a factor has been found
my ($mersAll, $resultsAll);

# Reads the given file and returns two strings, the first for mers
#   results and the second for GIMPS results.
sub readResults($);

# readResults prints to this handle, so it has to be opened first and might as well stay open.
# ... but an existing file likely has errors in it that may not have been printed, so:
my ($logName) = "dogpu.$$.log";
$logName = ('working.doWork/' . $logName) if -d 'working.doWork' && ! -d 'working.doWork/logs';
$logName = ('working.doWork/logs/' . $logName) if -d 'working.doWork/logs';
my ($logFH);
MersInfo::rotateLog($logName, 99);
if (!$checkConfig)
{
  open($logFH, '>>', $logName) or $errors .= "could not append to $logName, errno $!\n";
  select $logFH;
  $| = 1;
  ($mersAll, $resultsAll) = readResults('results.txt');
}

eval # so the many calls to die can be caught and results printed before exiting
{
  my ($lowK, $hiK); # for -tfk runs to close G gaps
  while (! $checkConfig && $noFactor)
    {
      last if !$debug && grep { -d } ($stopDir, $skipDir, );
      $exp = $args = undef;
      # Read new work from STDIN
      while (<STDIN>)
        { # Factor=38F015B7DD0F758A1DC7D6E5D429396F,132973657,71,72
          s/[\r\n]+$//;
          s/ *#.*//; # ignore comments ...
          next unless $_; # ... and empty lines.
          if (m/^Factor=([0-9A-F]+),([1-9][0-9]*),([1-9][0-9]*),([1-9][0-9]*)/)
            { # GIMPS format; the hex value is a checksum internal to GIMPS
              $hex = $1;
              $exp = Math::BigInt->new("$2");
              $start = Math::BigInt->new("$3");
              $stop = Math::BigInt->new("$4");
              $start < $stop or die "stdin GIMPS factoring task starts at or after it ends: $_\n";
              $letter = 'g'; # implied
              $useMFG = "M( $exp )g: 2^$start 2^$stop\n" if $exp < $smallestExp;
              last;
            }
          if (m/^M[(] ([1-9][0-9]*) [)]([Gg]): ([1-9][0-9]*) ([1-9][0-9]*)/o)
            { # mers format
              $exp = Math::BigInt->new("$1");
              $letter = $2;
              $start = Math::BigInt->new("$3");
              $stop = Math::BigInt->new("$4");
              $start < $stop or die "stdin mers format gap starts at or after it ends: $_\n";
              $hex = $exp; # not part of this (mers) input format; use the exponent, partly to track that this is not a GIMPS task
              if ($exp < $smallestExp)
                {
                  $useMFG = "M( $exp )$letter" . ": $start $stop\n";
                }
              elsif ($dashM)
                {
                  my ($stop2) = 2*$exp*$dashM + $start;
                  $stop = $stop2 if $stop2 < $stop;
                }
              last;
            }
          die "Unrecognized input, '$_'";
        }
      last unless defined $exp; # STDIN was empty?
      if ($letter ne 'G')
        { # Do the work one bit at a time; most kernels in mfaktc and mfakto only support one bit per task.
          $args = " -tf $exp $start " . ($start + 1);
        }
      else
        { # Close the G gap thru the first power of 2 within the gap or try $dashM potential factors.
          my ($bedbl) = $exp;
          $bedbl->bmul(2);
          $lowK = Math::BigInt->new("$start") if !defined($lowK);
          $hiK = Math::BigInt->new("$stop") if !defined($hiK);
          my ($lowLog) = $lowK->copy()->blog(2)->binc();
          my ($hiLog) = $hiK->copy()->blog(2);
          my ($nextHi) = $hiK;
          $nextHi = Math::BigInt->new(1)->blsft($lowLog) if ($lowLog < $hiLog);
          $lowK->bdiv($bedbl);
          $nextHi->bdiv($bedbl)->binc();
          $args = " -tfk $exp $lowK $nextHi";
          # next time, start where this run leaves off
          $lowK = $nextHi;
          # leave $hiK for the next run, just like 'g' gaps are closed one power of 2 at a time
        }

      if ($useMFG eq '')
        {
          my ($bmin, $bmax);
          my ($knownError) = 0;
          my ($cudaErrors) = '';
          my ($prior) = '';
          my ($repeats) = 0;
          while ($args && (! -d($stopDir) || $debug))
            {
              my ($cmd) = "$exe";
              $cmd .= " '" . join("' '", @options) . "'" if @options;
              die("cmd tainted by options")
                if Scalar::Util::tainted($cmd);
              my ($gpuSP) = findGPUSievePrimes($exp) if scalar keys(%minExp) > 1;
              $cmd .= " --GPUSievePrimes $gpuSP" if defined $gpuSP;
              die("cmd tainted by gpuSP")
                if Scalar::Util::tainted($cmd);
              $cmd .= $args if $args;
              die("cmd tainted by args")
                if Scalar::Util::tainted($cmd);
              my ($exeFH);
              open($exeFH, "$cmd < /dev/null 2>&1 |") or die "could not run $cmd in $exeDir, errno $!\n";
              while (<$exeFH>)
                {
                  if ($_ eq $prior)
                    {
                      ++$repeats;
                    }
                  else
                    {
                      print($logFH "exe $exe <repeated prior line $repeats times>\n") if $repeats;
                      print($logFH "exe $exe $_");
                      $repeats = 0;
                      $prior = $_;
                    }
                  ++$knownError, $cudaErrors .= "  $_", next if m/^Error -5 [(]Out of resources[)]: Enqueuing kernel[(]clEnqueueNDRangeKernel[)]/;
                  ++$knownError, $cudaErrors .= "  $_", next if m/^ERROR: cudaStreamCreate[(][)] failed for stream /;
                  ++$knownError, $cudaErrors .= "  $_", next if m/^ *cudaGetLastError[(][)] returned [0-9]+: all CUDA-capable devices are busy or unavailable/;
                  ++$knownError, $cudaErrors .= "  $_", next if m/^ *cudaGetLastError[(][)] after using kernel [-_a-zA-Z0-9]+ returned [0-9]+: unspecified launch failure/;
                  ++$knownError, $cudaErrors .= "  $_", next if m/^WARNING: exponents [<] [1-9][0-9]* are not supported/;
                  ++$knownError, $cudaErrors .= "  $_", next if m/^WARNING: ignoring line /;

                  # This one complaint will always be present after the work has been done, so ignore it.
                  ++$knownError, $cudaErrors .= "  $_", next if m/^ERROR: get_next_assignment/;

                  # This is not really an error, but indicates the configuration of $exe is set to stop immediately when a factor is found;
                  #  the configuration should be changed to continue in case there is another factor before the next power of 2.
                  ++$noGPU, next if m/^found [1-9][0-9]* factors? for M[1-9][0-9]* from 2\^[1-9][0-9]* to 2^[1-9][0-9]* [(]partially tested[)] [[]/;

                  $exp = $1, $bmin = $2, $bmax = $3, next if m/^got assignment: exp=([1-9][0-9]*) bit_min=([0-9]+) bit_max=([0-9]+) /;
                  $exp = $1, $bmin = $2, $bmax = $3, next if m/^Starting trial factoring M([1-9][0-9]*) from 2\^([0-9]+) to 2\^([0-9]+) /;

                  if (m/^ERROR: cannot parse k_min for option "-tfk": : Numerical result out of range/
                      || m/^ERROR: cannot parse k_min for option "-tfk": : Result too large/)
                    { # This is always followed by an "Ignoring" line, so adding to $useMFG here would be redundant
                      $cudaErrors .= "  $_";
                      ! -f("M$exp" . '.ckp') || unlink("M$exp" . '.ckp') or die "could not unlink M$exp.ckp after error, $!\n";
                      next;
                    }
                  if (m/^ERROR: cudaGetLastError[()]* returned 9: invalid configuration argument/)
                    { # This seems to indicate none or little of the work is done and repeats with the same input, so give the work to mersfacgmp
                      $useMFG .= ("M( $exp )G: 2^$bmin 2^$bmax\n");
                      $cudaErrors .= "  $_";
                      ! -f("M$exp" . '.ckp') || unlink("M$exp" . '.ckp') or die "could not unlink M$exp.ckp after error, $!\n";
                      next;
                    }
                  if (m/^ERROR: GPU sieve requested but current settings don't allow exponents below [0-9]+[.]/)
                    { # this also happens if the exponent is simply too small for mfakt* no matter what sieving was requested
                      $args = undef;
                      $useMFG .= ("M( $exp )g: 2^$start 2^" . ($start + 1) . "\n") if $letter eq 'g';
                      $useMFG .= ("M( $exp )G: $start $stop\n") if $letter eq 'G';
                    }
                  if (m/^ *Ignoring TF M([0-9]+) from 2\^([0-9]+) to 2\^([0-9]+)[!]/)
                    {
                      $useMFG .= ("M( $1 )G: 2^$2 2^$3\n"),
                        $args = undef,
                        next
                        if index($args, ' -tf ') >= 0;
                      $args = undef;
                      $lowK->bmul(2)->bmul($exp)->bdec();
                      $hiK->bmul(2)->bmul($exp)->binc();
                      $useMFG .= ("M( $exp )G: $lowK $hiK\n");
                    }
                  if (m/^ *Ignoring TF M([0-9]+) from k_min=([0-9]+) to k_max=([0-9]+)[!]/)
                    {
                      $args = undef;
                      my ($mink) = Math::BigInt->new("$2");
                      my ($maxk) = Math::BigInt->new("$3");
                      $mink->bmul(2)->bmuladd($exp, -1);
                      $maxk->bmul(2)->bmuladd($exp, 1);
                      $useMFG .= ("M( $exp )G: $start $stop\n");
                    }

                  # The following may not cover all the work and presumes $exe does not support bit_min, but it's better than erroring out.
                  $useMFG .= ("M( $exp )G: 2^$1 2^" . ($1 + 1) . "\n"), next if m/^ERROR: No suitable kernel found for bit_min=([1-9][0-9]*), bit_max=([1-9][0-9]*)[.]*/;
                  
                  if (m/^ERROR: clear_assignment[()] \/ modify_assignment[()]: can't rename workfiles/)
                    { # This left behind an empty file named __worktodo__.tmp and caused both gpu work types to be skipped
                      ++$knownError;
                      $cudaErrors .= "  $_";
                      ! -f("M$exp" . '.ckp') || unlink("M$exp" . '.ckp') or die "could not unlink M$exp.ckp after error, $!\n";
                      -f('__worktodo__.tmp') && unlink('__worktodo__.tmp') &&
                        next;
                      # fall thru to the catch-all noGPU case
                    }

                  # These are known to be "fatal" - the GPU software is missing, out of date, or similar.
                  ++$noGPU, next if m/^ERROR: init_CL(3, 0) failed$/;
                  ++$noGPU, next if m/^ERROR: current CUDA driver version is lower than the CUDA toolkit version used during compile/;
                  ++$noGPU, next if m/^Select device - Error -1001 .Unknown errorcode .not an OpenCL error..: clGetPlatformIDs/;
                  ++$noGPU, next if m/^ *CUDA runtime version  *0[.]0/;
                  ++$noGPU, next if m/^ *CUDA driver version  *0[.]0/;
                  ++$noGPU, next if m/^Please update your graphics driver/;
                  # Catch-all - these might need to be handled as a "knownError" instead.
                  ++$noGPU, next if m/^ERROR: /;
                }
              $warnings .= "CUDA errors:\n$cudaErrors\ndone\n" if $debug;
              ++$knownError, $warnings .= (localtime() . " close of pipe from $exe failed, exited $?\n")
                if !close($exeFH);
              $noGPU == 0 or die "$exe seems unable to find a GPU to use; see $logName\n";
              last if $useMFG ne '' || -s 'results.txt';
              $knownError <= 20 or die "$exe failed many times in a row; see $logName\n";
              sleep(5) if $knownError && ! $debug;
            }
        }
      if ($useMFG ne '')
        { # Instead of running doJoiner.pl here, tell the caller to do it.
          $warnings .= "using doJoiner.pl for a GIMPS task\n" if $hiBits;
          print(STDOUT "$useMFG");
          last;
        }
      elsif (-s 'results.txt')
        {
          my ($mers, $results) = readResults('results.txt');
          $resultsAll .= $results;
          $mers ne '' or die "no mers info in results.txt\n";
          $mersAll .= $mers;
        }
      elsif ($hiBits > 0 && -d($stopDir) && $resultsAll ne '')
        { # Presume that the user killed $exe in order to stop the caller (doWork.pl):
          #   save the current results in results.txt so they will be reported next
          #   time in case there is more work to be done for the same exponent.
          my ($resultsFH);
          open($resultsFH, '>', 'results.txt') or die "could not save partial results to results.txt, errno $!\n";
          print($resultsFH $resultsAll);
          close($resultsFH) or die "could not close results.txt to save partial results, errno $!\n";
          $resultsAll = '';
          print("checkpoint file M$exp.ckp\n")
            if -s "M$exp.ckp";
          # Printing $mersAll is fine; that does not get sent to GIMPS.
          die "presuming user killed $exe since stop directory exists and results.txt is empty\n";
        }
      last if !$noFactor;
      my ($more);
      ++$start, $more = ($start < $stop) if $args =~ m/ -tf /; # GIMPS or 'g' style gap between powers of 2
      $more = ($lowK->bcmp($hiK) < 0) if $args =~ m/ -tfk /; # 'G' style gap between arbitrary potential factors
      $args = undef if !$more;
      last unless $hiBits > 0 || $more;
      next if $args;
      if ($hiBits > 1)
        {
          my ($limit);
          for my $key (sort { $a <=> $b } keys(%GIMPSTrialFactorLimit))
            {
              $limit = $GIMPSTrialFactorLimit{$key}, last if $exp < $key;
            }
          next if ! defined($limit) || $stop >= $limit || $stop >= $hiBits;
        }
    } # while ! $checkConfig && $noFactor && (! -d "$stopDir" || $debug)

  die "$errors" if $errors;
}; # eval

my ($err) = $@; # die message from eval
print(STDOUT "$warnings") if $warnings;
$err eq '' or print(STDOUT "$err") or warn("could not print err message, errno $!\n");
!$errors || $errors eq $err || print(STDOUT "$errors");
if (!$checkConfig)
{
  $mersAll eq '' or print(STDOUT "$mersAll\n") or warn("could not print mers info, errno $!\n");
  $resultsAll eq '' or print(STDOUT "$resultsAll") or warn("could not print results.txt, errno $!\n");
  close($logFH) or warn "could not close $logName, errno $!\n";
  $err ne '' or -d($stopDir) or unlink($logName);
  $useMFG ne '' or $err ne '' or -d($stopDir) or "$mersAll" ne '' or "$resultsAll" ne '' or die "no input\n";
}
exit($err ? 1 : 0);

sub readResults($)
{
  my ($file) = @_;
  local $_;
  my ($mers, $results) = ('', '', );
  return(($mers, $results, )) unless -s "$file";
  my ($rh);
  open($rh, '<', "$file") or die "could not read $file, errno $!\n";
  while (<$rh>)
    {
      chomp;
      s,[\r\n]+$,,; # chomp doesn't chop \r, at least on MSWin10 with Cygwin's perl.
      s,[\r]+, ,g; # change others to spaces just in case
      $results .= "$_\n";
      print($logFH "$_\n");
      next
        if (m/^found ([1-9][0-9]* factors?) for M([1-9][0-9]+) from 2\^([1-9][0-9]+) to 2\^([1-9][0-9]+) .*partially tested\b/);
      my ($user, $computer) = ($1, $2) if s.UID: ([A-Za-z0-9]+)/([A-Za-z0-9]+), ..o;
      my ($n, $rest, $lowK, $hiK);
      $n = "$1", $rest = "$2"
        if (m/^M[(] *([1-9][0-9]*) *([)].*)$/);
      $n = "$1", $rest = ")J: 2^$2 2^$3 # $4"
        if (m/^no factor for M([1-9][0-9]+) from 2\^([1-9][0-9]+) to 2\^([1-9][0-9]+) (.*)$/);
      $n = "$1", $lowK = Math::BigInt->new("$2"), $hiK = Math::BigInt->new("$3"), $rest = $4
        if (m/^no factor for M([1-9][0-9]+) from k_min=([1-9][0-9]+) to k_max=([1-9][0-9]+) (.*)$/);
      $n = "$1", $lowK = Math::BigInt->new("$2"), $hiK = Math::BigInt->new("$3"), $rest = $4
        if (m/^M[(] *([1-9][0-9]*) *[)]K: ([0-9]+) ([1-9][0-9]+)(.*)$/);
      if (m/^M([1-9][0-9]+) has a factor: ([1-9][0-9]+) (.*)$/)
        {
          $noFactor = 0;
          $n = "$1";
          my ($factor) = "$2";
          $rest = ")C: $factor # $3";
          if ($rest =~ m/^ *[[]TF[:]([1-9][0-9]+)[:]([1-9][0-9]+)[*][:]mfakt/)
            { # the '*' here indicates factoring stopped at or near the factor
              $mers .= "M( $n $rest\n"; # add the mers format factor line
              my ($low) = Math::BigInt::bone()->blsft("$1");
              $rest = ")J: $low $factor # $rest\n";
              $mers .= "M( $n $rest\n"; # add the reduced join mers format line
              next;
            }
        }
      $n = "$2", $rest = ")J: 2^$3 2^$4 # $1 $5"
        if (m/^found ([1-9][0-9]* factors?) for M([1-9][0-9]+) from 2\^([1-9][0-9]+) to 2\^([1-9][0-9]+) (.*)$/);
      next unless defined($rest);
      $lowK *= 2*$n, $hiK *= 2*$n, ++$lowK, ++$hiK, $rest = ")J: $lowK $hiK $rest"
        if defined $lowK;
      $rest .= " # GIMPS user $user computer $computer" if (defined($user));
      $mers .= "M( $n $rest\n";
    }
  close($rh);
  unlink($file) or die "unable to remove $file, errno $!\n";
  return(($mers, $results, ));
}

# mfaktc-0.21 has required configuration data in mfaktc.ini that restricts
#   the supported Mersenne exponents so I've added a program gpulimits.exe
#   that produces the relevant data in gpulimits.txt.
sub getGPUlimits()
{
  my ($limitsfn) = 'gpulimits.txt';
  $limitsfn = "$exeDir/$limitsfn" unless -s "$limitsfn";
  return unless -s $limitsfn;
  my ($limitsFH);
  warn("could not read $limitsfn, $!\n"),
    return
    unless open($limitsFH, '<', "$limitsfn");
  while (<$limitsFH>)
    {
      my (@flds) = split(/,/);
      if (m/SievePrimes/)
        {
          my ($i);
          for $i (0 .. $#flds)
            {
              $minExp{'GPUSievePrimes'} = $i if $flds[$i] eq 'GPUSievePrimes';
              $minExp{'GPUSievePrimes'} = $i if $flds[$i] eq 'SievePrimes';
              $minExp{'MinExponent'} = $i if $flds[$i] eq 'MinExponent';
            }
          next;
        }
      next
        unless exists $minExp{'GPUSievePrimes'} && exists $minExp{'MinExponent'};
      next
        unless m/^([1-9][0-9]*),([1-9][0-9]*),([1-9][0-9]*)*$/;
      @flds = ($1, $2, $3);
      $minExp{$flds[$minExp{'MinExponent'}]} = $flds[$minExp{'GPUSievePrimes'}];
      $smallestExp = $flds[$minExp{'MinExponent'}] if $smallestExp > $flds[$minExp{'MinExponent'}];
    }
  close($limitsFH);
  %minExp = ()
    unless exists $minExp{'GPUSievePrimes'} && exists $minExp{'MinExponent'} && scalar keys(%minExp) > 1;
  warn("no recognized data in $limitsfn\n")
    unless scalar keys(%minExp) > 1;
  return;
}

sub findGPUSievePrimes($)
{
  my ($exp) = @_;
  local $_;
  return(undef)
    unless scalar keys(%minExp) > 1 && $exp > 0;
  my ($which) = 0;
  foreach my $key (grep { m/^[0-9]+$/ && $_ < $exp } keys(%minExp))
    {
      $which = $key
        if $which < $key;
    }
  return(undef)
    if $which == 0;
  return($minExp{$which});
}
