#!/usr/bin/perl -Tw
# Use mersfacgmp to try to find new factors of (prime exponent) Mersenne numbers.

use strict;
use strict 'vars';

use Math::BigInt try => 'GMP';

BEGIN { push(@INC, '.') if -s('MersInfo.pm'); $ENV{'PATH'} = "/usr/bin:/bin"; }

use MersInfo;

my ($usage) = ("Usage: doJoiner.pl [options for 'mersfacmp' or other program]\n"
               . "  [--stop <directory>] [--skip <directory>] [--tmp <directory>]\n"
               . "  [--stop=<directory>] [--skip=<directory>] [--tmp=<directory>]\n"
               . "  [--check] [--help|-h] [--version]\n"
               . "Use 'mersfacgmp' or other trial factoring program to try to find new factors of the Mersenne number.\n"
               . "  STDIN should be mers format information about the Mersenne, usually 'G' or 'g' (gap) lines.\n"
  );

if (grep { m/^(--check|--version|--help|-h)$/ } @ARGV)
{
  my ($version) = MersInfo::version();
  print("$0 version $version\n");
  print($usage)
    if grep { m/^(--help|-h)$/ } @ARGV;
  exit(0);
}

my ($quit);
$SIG{'HUP'} = "IGNORE";
foreach ('INT', 'TERM')
{
  $SIG{$_} = sub {
    my ($sig) = @_;
    $quit = 1;
    # this could be added to $output, but mersfacgmp exits quickly, unlike sprpgmp, when interrupted.
    warn "caught SIG$sig\n";
  };
}

# Using mersfacgmp in $PWD is preferable; the one in mers/ may be a test version or similar.
my ($joiner) = './mers/mersfacgmp' if -x './mers/mersfacgmp';
$joiner = './mersfacgmp' if -x './mersfacgmp';
defined($joiner) or die "cannot find mersfacgmp executable in '.' or './mers/'\n";

my (@args);
my ($topDir) = 'working.doWork';
-d $topDir
  || mkdir($topDir)
  or $topDir = '/var/tmp';
-d $topDir
  or $topDir = '/usr/tmp';
-d $topDir
  or $topDir = '/tmp';
my ($stopDir) = "$topDir/stop"
  if -d $topDir;
my ($skipDir) = "$topDir/skipJoiner"
  if -d $topDir;
my ($tmpDir) = "$topDir/tmp/joiner"
  if -d "$topDir";
!$tmpDir || -d $tmpDir
  or mkdir("$topDir/tmp") && mkdir($tmpDir)
  or $tmpDir = "$topDir/joiner";
!$tmpDir || -d $tmpDir
  or mkdir($tmpDir)
  or $tmpDir = "$topDir";
foreach (@ARGV)
{
  next unless $_;
  $stopDir = $1, next if $stopDir && $stopDir eq '--stop' && m,^([-._/[:alnum:]]+)$,;
  $skipDir = $1, next if $skipDir && $skipDir eq '--skip' && m,^([-._/[:alnum:]]+)$,;
  $tmpDir = $1, next if $tmpDir && $tmpDir eq '--tmp' && m,^([-._/[:alnum:]]+)$,;
  $stopDir = '--stop', next if $_ eq '--stop';
  $skipDir = '--skip', next if $_ eq '--skip';
  $tmpDir = '--tmp', next if $_ eq '--tmp';
  $stopDir = $1, next if m,^--stop=([-._/[:alnum:]]+)$,;
  $skipDir = $1, next if m,^--skip=([-._/[:alnum:]]+)$,;
  $tmpDir = $1, next if m,^--tmp=([-._/[:alnum:]]+)$,;
  die "$0: unsupported option or argument '$_'\n"
    unless m,^(-[a-zA-Z0-9]), || m,^([0-9]+)$,;
  push(@args, "$1");
}

if (!@args)
{ # Default arguments for $joiner; ignore if other arguments are given.
  @args = ('-g', '-a', '-e1', );
}

my ($output) = '';
if (!grep { m,^-m, } @args)
{ # Let it have our stdin; anything else is messy and overkill.
  my ($fh);
  open($fh, '-|', "$joiner", @args)
    or die "could not run $joiner, $!\n";
  ($output .= $_)
    while (<$fh>);
  $output =~ s/\\\n//g; # rejoin lines split with backslashes
  print("$output");
  warn("no output from $joiner\n")
    unless $output;
  close($fh)
    or die "closing $joiner failed, $!\n";
  exit(0);
}

# When -m is given, this needs to give an H line to mersfacmp, not a G line, and
#   print the remaining (or original) gap as well as the J and HU lines.
# Also, doWork.pl now adjusts the value given with -m and figures out how much
#   of the gap remains after this completes.
-d($tmpDir)
  or die("cannot find a directory for temporary files\n");
while (<STDIN>)
{
  chomp;
  next unless $_;
  m,^M[(] *([1-9][0-9]*) *[)]([A-Za-z]):? *([1-9][0-9]*)? *([1-9][0-9]*)?, or die "unexpected input - '$_'\n";
  my ($n) = "$1";
  my ($l) = "$2";
  my ($low) = "$3" if $3;
  my ($hi) = "$4" if $4;
  next unless $l =~ m,^G$,i;
  $output .= "$_\n"; # send the input back in case the gap is not covered
  if (!$low)
    {
      print("$output") if $output;
      die "missing gap start - '$_'\n";
    }
  if (!$hi)
    {
      print("$output") if $output;
      die "missing gap end - '$_'\n";
    }
  $low = ($l eq 'G' ? Math::BigInt->new("$low") : Math::BigInt->bone()->blsft("$low"));
  $hi = ($l eq 'G' ? Math::BigInt->new("$hi") : Math::BigInt->bone()->blsft("$hi"));
  my ($mfgFH);
  print("$output"),
    warn("could not create tmp file $tmpDir/$$, $!"),
    exit(0)
    unless open($mfgFH, '>', "$tmpDir/$$");
  print($mfgFH "M( $n )U: $low\n");
  print("$output"),
    warn("could not close tmp file $tmpDir/$$, $!"),
    exit(0)
    unless close($mfgFH);
  open($mfgFH, '-|', "$joiner", '-a', @args, "$tmpDir/$$") or die "could not run $joiner, $!\n";
  while (<$mfgFH>)
    { # Usually a 'J'oin line and a 'U' or 'H' (extent) line.
      $output .= "$_";
      next
        unless m,^M[(] *([1-9][0-9]*) *[)]([A-Za-z]):? *([0-9]*)? *([0-9]*)?,;
      next unless $n eq "$1" && "$2" eq 'J' && defined($3) && defined($4);
      my ($joinLow) = Math::BigInt->new("$3");
      my ($joinHi) = Math::BigInt->new("$4");
      next unless $joinLow <= $low;
      # Reduce the original gap at the start by this join:
      $low = $joinHi;
    }
  unlink("$tmpDir/$$") if -e "$tmpDir/$$";
  print("$output"),
    warn("could not close $joiner run from $tmpDir/$$, $!"),
    exit(0)
    unless close($mfgFH);
  last;
}
print("$output");
# There should only be one 'G' or 'g' line on STDIN but don't fail just because of that.
print()
  while (<STDIN>);
exit(0);
