#!/usr/bin/perl -Tw
# Do the "next" task based on fetch.pl's "next".

use strict;
use strict 'vars';

BEGIN { push(@INC, '.'); $ENV{'PATH'} = '/usr/bin:/bin'; }

use MersInfo;
use Mersenne;

my ($choices) = "$Mersenne::workLetters";
# Drop letters that do not make sense here - factors, gaps, joins,
#   trial factoring extents, and primality proving.  The last simply
#   takes a lot longer than the other kinds of work and is easy to
#   do with 'fetch.pl S' and doProof.pl.
$choices =~ s/[CGgHJULMS]//g;
$choices .= 'clm'; # not in $Mersenne::workLetters - SPRP check possibly needed

# Note that this is modified later to add the options supported for the work scripts.
my ($usage) = "$0 --version\n  print version and exit\n"
  . "$0 <--check>\n  check the configuration and arguments\n"
  . "$0 --help (or -h)\n  print this usage message and exit\n"
  . "$0 <letter>\n"
  . "  Do the 'next' task for the given mers letter - one of $choices\n"
  . "  The 'next' exponent is determined by './fetch.pl next'\n"
  . "$0\n"
  . "  Do the 'next' task of any mers letter - $choices\n"
  . "  The 'next' exponent and kind is determined by './fetch.pl next'\n";

# These are checked from least preferable to most preferable.
my (@sprp) = ('../mers/sprpgmp') if -x '../mers/sprpgmp' || -x '../mers/sprpgmp.exe';
$sprp[0] = './mers/sprpgmp' if -x './mers/sprpgmp' || -x './mers/sprpgmp.exe';
$sprp[0] = './sprpgmp' if -x 'sprpgmp' || -x 'sprpgmp.exe';
defined($sprp[0]) or die "cannot find any SPRP-capable executable (sprpgmp)\n";
push(@sprp, '-C');

my (%workScript) = ('E' => [ './doECM.pl', ], 'o' => ['./doP+-1.pl', ], 'q' => ['./doP+-1.pl', ],
                    'c' => [ @sprp, ], );

foreach ( [ './fetch.pl', ], values(%workScript))
{
  -f($_->[0]) && -x(_) && !-l($_->[0])
    or die("script '$_' is not in the current directory, is not exectable, or is a soft link\n");
}

# These options to the %workScripts are the only ones supported here.
my (%scriptOptions) = (
  'E' => { '-d' => '([0-9]+)', '-maxmem' => '([1-9][0-9]+)',
           '--time' => '([1-9][0-9]+)', '--duration' => '([1-9][0-9]+)', },
  'o' => { '-d' => '([0-9]+)', '-maxmem' => '([1-9][0-9]+)',
           '-i' => '([1-9][0-9]+)', '--increment' => '([1-9][0-9]+)', },
  'q' => { '-d' => '([0-9]+)', '-maxmem' => '([1-9][0-9]+)',
           '-i' => '([1-9][0-9]+)', '--increment' => '([1-9][0-9]+)', },
  'c' => { },
  );
my (%allOptions);
my ($letter, $option);
foreach $letter (keys(%scriptOptions))
{
  $allOptions{$_} = ($scriptOptions{$letter}{$_})
    foreach keys(%{$scriptOptions{$letter}});
}

my (%optionUsages) = ( '-d' => '[-d <debugLevel>]',
                       '-maxmem' => '[-maxmem <memoryMaximum>]',
                       '--duration' => '[--duration <seconds>]',
                       '--time' => '[--time <seconds>]',
                       '-i' => '[-i <stage1increase>]',
                       '--increment' => '[--increment <stage1increase>]',
  );

my (%added);
foreach $letter (sort(keys(%scriptOptions)))
{
  foreach (sort(keys(%{$scriptOptions{$letter}})))
    {
      next if exists $added{$_};
      $added{$_} = 1;
      $usage .= "$0 $optionUsages{$_}\n  as for doP+-1.pl (the stage 1 bound increment)\n",
        next
        if m,-i,;
      $usage .= "$0 $optionUsages{$_}\n  as for doECM.pl or doP+-1.pl\n";
    }
}
$usage .= "Repeating an option is not supported.\n";
$usage .= "Note that --duration and --time have the same effect - do not give both.\n";

# This is needed to match part of a string, not an entire string.
my ($uuidPiece) = MersInfo::patterns('uuid');
$uuidPiece =~ s,^\^,,;
$uuidPiece =~ s,\$$,,;

my ($checkConfig);
$checkConfig = 1
  if grep { $_ eq '--check' } @ARGV;
print("$0 version " . MersInfo::version() . "\n"),
  exit(0)
  if $checkConfig || grep { $_ eq '--version' } @ARGV;
print("$usage"),
  exit(0)
  if grep { $_ eq '--help' || m,^-[hH?]$, } @ARGV;

$letter = undef;
my (%options);
while (@ARGV)
{
  shift(),
    next
    if $ARGV[0] eq '--check';
  if ($ARGV[0] !~ m,^(--?[a-z]+)$,)
    {
      die("$0: two distinct mers letters given, '$letter' and '$ARGV[0]' - only one (or none) is supported\n$usage")
        if defined $letter && $letter ne $ARGV[0];
      $letter = shift();
      next;
    }
  $option = "$1";
  shift();
  die("$0: unknown option - '$option'\n$usage")
    unless exists $optionUsages{$option};
  die("$0: no argument given with option '$option'\n$usage")
    unless $#ARGV >= 0;
  die("$0: unexpected argument given with option '$option' - '$ARGV[0]'\n$usage")
    unless $ARGV[0] =~ m,^$allOptions{$option}$,;
  die("$0: option '$option' given with '$options{$option}' and with '$1'\n")
    if exists $options{$option} && $options{$option} ne "$1";
  die("$0: option '--time' given with '$options{'--time'}' and option '--duration' given with '$1'\n")
    if exists $options{'--time'} && $option eq '--duration' && $options{'--time'} ne "$1";
  die("$0: option '--duration' given with '$options{'--duration'}' and option '--time' given with '$1'\n")
    if exists $options{'--duration'} && $option eq '--time' && $options{'--duration'} ne "$1";
  $options{$option} = "$1";
  shift();
}

die("unexpected argument '$letter'\n$usage")
  if defined $letter && $letter !~ m,^([$choices])$,;
$letter = $1
  if $1;
$letter = 'c'
  if $letter && ($letter eq 'l' || $letter eq 'm');
my ($letterARGV) = $letter;

exit(0)
  if $checkConfig;

my ($exp, $reserved);
my ($unexpected) = '';
while (!$reserved)
{ # Reserving the exponent might fail, so this is a bit messy.
  my (@fetchArgs) = ('./fetch.pl', 'next');
  push(@fetchArgs, $letterARGV)
    if $letterARGV;
  my ($childFH);
  print("# " . join(' ', @fetchArgs) . "\n");
  open($childFH, '-|', @fetchArgs)
    or die("could not run '" . join(' ', @fetchArgs) . "' - $?");
  print("# output from " . join(' ', @fetchArgs) . ":\n");
  $letter = undef;
  while (<$childFH>)
    {
      chomp;
      print(STDERR "# $_ # ignored\n"),
        next
        if $. > 1;
      print("# $_\n");
      m,^([1-9][0-9]*) ([$choices])$,
        or (m,^([1-9][0-9]*)$, && $letterARGV)
        or die("unexpected output from fetch.pl\n");
      !defined($exp) || !defined($letter) || ($1 != $exp && $2 ne $letter)
        or die("same exponent and letter as prior time\n");
      $exp = $1;
      $letter = $2
        if $2;
    }
  close($childFH)
    or die("closing '" . join(' ', @fetchArgs) . "' failed, $?");

  $letter = $letterARGV
    if !$letter && $letterARGV;
  $letter = 'c'
    if $letter && ($letter eq 'l' || $letter eq 'm');
  exists($workScript{$letter})
    or die("internal error - work script for letter '$letter' is not known\n");

  open($childFH, '-|', './reserve.pl', $exp, 'localhost', $letter)
    or die("could not run './reserve.pl $exp localhost $letter', $?");
  while (<$childFH>)
    {
      next if m,reserve.*version,;
      print("# $_");
      $reserved = '1',
        next
        if m/^reserved $exp localhost $letter$/;
      $reserved = { 'exp' => $exp, 'uuid' => $1, 'kind' => $2, },
        next
        if m/^already reserved - $exp $uuidPiece ([?!c$Mersenne::workLetters])$/;
      $unexpected .= ("unexpected output from './reserve.pl $exp localhost $letter' - $_");
    }
  close($childFH) || $unexpected
    || (defined($reserved) && ref($reserved) eq 'HASH')
    or die("could not close './reserve.pl $exp localhost $letter', $?");

  last
    if $reserved && !$unexpected && !($?);
  $unexpected .= "'./reserve.pl $exp localhost $letter' exited $?\n"
    if defined $reserved && ref($reserved) ne 'HASH' && $?;
  $unexpected .= "... but reserved $exp localhost $letter\n",
    last
    if defined $reserved && $reserved eq '1';
  # If M($exp) needs an SPRP check, ignore reservations other than c (SPRP) or S (primality proof).
  exit(1)
    if (!defined($reserved) || ref($reserved) ne 'HASH'
        || $reserved->{'kind'} =~ m,^[cS]$, || $letter ne 'c');
  last;
}
print(STDERR " STDERR from './reserve.pl $exp localhost $letter' - \n$unexpected\n STDERR done\n")
  if $unexpected;

my ($mers);
END
{
  $mers->print()
    if $mers;
}

my ($fetchFH);
print("# ./fetch.pl $exp\n");
open($fetchFH, '-|', './fetch.pl', $exp)
  or die("could not run './fetch.pl $exp', $?");
while (<$fetchFH>)
{
  print("# $_");
  $mers = Mersenne->new('n' => $exp)
    unless $mers;
  $mers->parse($_)
    or die("Mersenne::parse of '$_' failed - " . $mers->errors());
}
die("no output from './fetch.pl $exp\n")
  unless $mers;

my ($tmpFile) = "/tmp/doNext.$$";
my ($random) = int(rand(1e9));
$random = int(rand(1e9))
  while -e "$tmpFile.$random";
$tmpFile .= ".$random";
END
{
  warn("see $tmpFile for related data\n")
    if $tmpFile && -s "$tmpFile";
}

my ($workFH);
open($workFH, '>', "$tmpFile")
  or die("could not create $tmpFile, $!");
$mers->print($workFH)
  or die("could not write to $tmpFile, $!");
close($workFH)
  or die("could not close $tmpFile, $!");

foreach $option (keys(%options))
{ # Drop given options that the selected script does not support.
  delete($options{$option})
    unless exists $scriptOptions{$letter}{$option};
}

my (%saveFile);
my ($data) = ''; # more data to give to save.pl later - ecm save file contents
my (@options);
push(@options, $_, $options{$_})
  foreach (keys(%options));
print("# " . join(' ', @{$workScript{$letter}}, @options) . " $tmpFile\n");
open($workFH, '-|', @{$workScript{$letter}}, @options, $tmpFile)
  or die("could not run '" . join(' ', @{$workScript{$letter}}, @options, $tmpFile) . "', $?");
while (<$workFH>)
{ # This support for ecm program save files was copied from spreadWork.pl 1 July 2022
  print("# $_");
  if (m/^save file begin (ecm[.]([1-9][0-9]+)[.](p[mp]1)([.]M)?)$/)
    {
      %saveFile = ( 'file' => "$1", 'exp' => "$2", 'method' => "$3", 'M' => ($4 ? "$4" : ''), 'content' => '', );
      next;
    }
  if (m/^save file end ([a-z][-._a-zA-Z0-9]+)$/)
    {
      die "save file end gives different file name, '$1', not the expected '$saveFile{'file'}'"
        unless $1 eq $saveFile{'file'};
      die "save file '$1' has unexpected content"
        unless $saveFile{'content'} =~ m/^([-+=;:.@ A-Za-z0-9\n\r]+)$/;
      $data .= "save file begin $saveFile{'file'}\n$saveFile{'content'}save file end $saveFile{'file'}\n";
      %saveFile = ();
      next;
    }
  $saveFile{'content'} .= "$_",
    next
    if exists $saveFile{'content'};
  next
    if m,^$, || m,^ *#,;
  $mers->parse($_)
    or die("Mersenne::parse of '$_' failed - " . $mers->errors());
}
close($workFH)
  or die("could not close '" . join(' ', @{$workScript{$letter}}, @options, $tmpFile) . "', $?");

if ($mers->needsSPRP() || $letter eq 'c')
{
  open($workFH, '>', "$tmpFile")
    or die("could not create $tmpFile, $!");
  $mers->print($workFH)
    or die("could not write to $tmpFile, $!");
  close($workFH)
    or die("could not close $tmpFile, $!");
  print("# " . join(' ', @{$workScript{'c'}}) . " $tmpFile\n");
  open($workFH, '-|', @{$workScript{'c'}}, $tmpFile)
    or die("could not run '" . join(' ', @{$workScript{'c'}}, $tmpFile) . "', $?");
  while (<$workFH>)
    {
      print("# $_");
      $mers->parse($_)
        or die("Mersenne::parse of '$_' failed - " . $mers->errors());
    }
  close($workFH)
    or die("could not close '" . join(' ', @{$workScript{'c'}}, $tmpFile) . "', $?");
}

open($workFH, '>', "$tmpFile")
  or die("could not create $tmpFile, $!");
$mers->print($workFH)
  or die("could not write to $tmpFile, $!");
print($workFH $data)
  if $data;
close($workFH)
  or die("could not close $tmpFile, $!");
print("# ./save.pl $tmpFile\n");
open($workFH, '-|', './save.pl', $tmpFile)
  or die ("could not run './save.pl $tmpFile', $?");
print($_)
  while (<$workFH>);
close($workFH)
  or die ("could not close './save.pl $tmpFile', $?");

my ($released);
print("# ./release.pl $exp localhost $letter\n");
open($workFH, '-|', './release.pl', $exp, 'localhost', $letter)
  or die("could not run './release.pl $exp localhost $letter', $?");
while (<$workFH>)
{
  next if m,release.*version,;
  print("# $_");
  $released = '1', next if m/^released $exp localhost $letter$/;
  $unexpected .= ("unexpected output from './release.pl $exp localhost $letter' - $_");
}
close($workFH) || $unexpected || defined($released)
  or die("could not close './release.pl $exp localhost $letter', $?");

unlink($tmpFile)
  or warn("could not unlink $tmpFile, $!");
exit(0);
