#!/usr/bin/perl -Tw
# Use the (Linux) ecm (GMP-ECM) program's -pm1 or -pp1 option to run one P-1 or P+1 task from the server.

use strict;
use strict 'vars';
use Math::BigInt try => 'GMP';

BEGIN { push(@INC, '.') if -s('Mersenne.pm'); $ENV{'PATH'} = "/usr/bin:/bin"; }

use MersInfo;
use Mersenne;
use ECMSave;

my ($usage) = ("Usage: doP+-1.pl [maxB1bound] [-maxmem <amount>] [<-i|--increment> <boundIncrement>] [maxB1bound] [files]\n"
               . "  [--stop <directory>] [--skip <directory>] [--tmp <directory>]\n"
               . "  [--stop=<directory>] [--skip=<directory>] [--tmp=<directory>]\n"
               . "  [-d [debugLevel]] [--check] [--help|-h] [--version]\n"
               . "Use P-1 or P+1 to try to find new factors of the Mersenne number.\n"
               . "  STDIN should be mers format information about the Mersenne.\n"
               . "The maxB1bound limits the stage 1 bound used.\n"
               . "The amount given with -maxmem is passed on to the Linux 'ecm' program.\n"
               . "The P-1 or P+1 stage 1 and stage 2 bounds will also be printed in mers format.\n"
               . "  The stage 1 bound to use is based on the mers format input, any existing 'ecm' save files\n"
               . "    and the value given with any -i or --increment option.\n"
               . "If a new factor is found, it will be printed in mers format.\n"
               . "  Note that factors found might be composite.\n"
  );

my ($checkConfig) = scalar(grep { m/^--check$/ } @ARGV);
my ($version) = MersInfo::version();
print(STDOUT "$0 version $version\n")
  if $checkConfig || grep { $_ eq '--version' } @ARGV;
print(STDOUT "$usage")
  if grep { $_ eq '--help' || $_ eq '-h' } @ARGV;
exit(0)
  if grep { $_ eq '--version' || $_ eq '--help' || $_ eq '-h' } @ARGV;

$SIG{'HUP'} = "IGNORE";
foreach ('INT', 'TERM')
{
  $SIG{$_} = sub {
    my ($sig) = @_;
    warn "caught SIG$sig\n";
  };
}

# Maximum stage one bound used for P+/-1 factoring by the 'ecm' program.
my ($maxB1) = $Mersenne::maxStage1Bound;

# Maximum memory ecm should use (-maxmem option).
#   This default is mostly to avoid failing an assert in the current
#   (Linux) ecm program; any maxmem value seems to avoid the problem,
#   but also note that the default stage 2 bound for larger stage 1
#   bounds uses too much RAM for most PCs, especially running ecm in
#   parallel on multiple cores or CPUs that share memory.
my ($maxmem) = '20';

# Default amount to increase the stage 1 bound (-i).
my ($bndIncr) = 1e7;

# Is this a debug run?
my ($debug);

my (@files);
while (@ARGV)
{
  shift, next if $ARGV[0] =~ m/^--check$/;
  shift, shift, next if $ARGV[0] eq '--skip' && $#ARGV > 0 && $ARGV[1] =~ m,^([-._/[:alnum:]]+)$,;
  shift, shift, next if $ARGV[0] eq '--stop' && $#ARGV > 0 && $ARGV[1] =~ m,^([-._/[:alnum:]]+)$,;
  shift, shift, next if $ARGV[0] eq '--tmp' && $#ARGV > 0 && $ARGV[1] =~ m,^([-._/[:alnum:]]+)$,;
  shift, next if $ARGV[0] =~ m,^--(skip|stop|tmp)=([-._/[:alnum:]]+)$,;
  if ($ARGV[0] eq '-d')
    {
      $debug = 1;
      shift();
      $debug = shift();
      next;
    }
  if ($ARGV[0] eq '-i' || $ARGV[0] eq '--increment')
    {
      shift();
      $bndIncr = shift();
      next;
    }
  if ($ARGV[0] eq '-maxmem')
    {
      shift();
      $maxmem = shift();
      next;
    }
  $maxB1 = shift(),
    next
    if $ARGV[0] =~ m,^([1-9][0-9]*)$,;
  push(@files, shift()),
    next
    if -f($ARGV[0]) && -s($ARGV[0]) && ! -l($ARGV[0]);
  die("unexpected argument or option - '$ARGV[0]'\n");
}
$maxB1 =~ m/^([1-9][0-9]*)$/
  or die "P+/-1 stage one bound must be numeric\n";
$maxB1 = Math::BigInt->new("$1");
$maxB1 = $Mersenne::maxStage1Bound
  if $maxB1 > $Mersenne::maxStage1Bound;
$maxmem =~ m/^([1-9][0-9])[km]?$/
  or die "-maxmem given with unrecognized value '$maxmem'\n";
$maxmem = "$1";
!defined($debug) || ($debug =~ m/^([0-9]+)$/ && $debug < 1024)
  or die "-d (debug) requires a number from 0 to 1023\n";
$debug = $1
  if defined $debug;
$bndIncr = 1000
  unless $bndIncr =~ m/^([0-9]+)$/ && $bndIncr > 1e3;
$bndIncr = $1
  if $1 && $1 >= 1000;
exit(0)
  if $checkConfig;

eval
{ # setpriority() is not always implemented
  setpriority(0, 0, 20);
};

my ($ecm);
foreach (split(/:/, $ENV{'PATH'}))
{
  next unless -d;
  $ecm = "$_" . '/ecm',
    last
    if -x "$_" . '/ecm' && ! -d "$_" . '/ecm';
}
defined($ecm)
  or die 'cannot find ecm program in PATH' . " $ENV{'PATH'}\n";

# Program name and options for both possible runs (LM halves of exponents == 4 % 8).
#   Do not run stage 2 unless memory limits are given.
my ($args) = "$ecm -v -timestamp";
$args .= " -maxmem $maxmem"
  if defined $maxmem;
$args = "echo debug $debug $args"
  if $debug;

my ($uuid) = "$0";
$uuid =~ s,^.*/(.*)$,,;
my ($subdir) = ECMSave::subdir($uuid = $1);
$subdir = 'p+-1.sav'
  unless $subdir && -d $subdir && ! -l $subdir;
-d($subdir) || mkdir($subdir)
  or die "could not mkdir $subdir, $!";
my ($logName) = $subdir . '/log.txt';
sleep(1)
  if -f($logName) && ! -s($logName);
MersInfo::rotateLog($logName, 99);
open(my $logFH, '>>', $logName)
  or die "could not append to $logName $!\n";
print($logFH 'Start at ' . localtime() . "\n");

my ($e); # Mersenne exponent
my ($mers); # Mersenne object/instance
my ($cnt) = 0;
@ARGV = @files;
while (<<>>)
{
  chomp;
  next unless $_;
  m/^M[(] ([1-9][0-9]*) [)](.): ([0-9][0-9]*).*$/
    or die "$ARGV:$.: unrecognized input: '$_'";
  $e = Math::BigInt->new("$1")
    if !defined($e);
  "$e" eq "$1"
    or die("$ARGV:$.: input contains data for more than one exponent: $e and $1\n");
  $mers = Mersenne->new('n' => $e)
    if !defined($mers);
  die("$ARGV:$.: parsing input '$_' failed - " . $mers->errors())
    unless $mers->parse($_);
  print($logFH "$_\n");
}

# Do P-1 or P+1 depending on the input - never do both.  Do P+1 if the prior bounds are the same.
my ($kind);
my ($obnd) = $mers->get('o');
my ($qbnd) = $mers->get('q');
$kind = 'pm1' if !defined $obnd || (defined $qbnd && $obnd < $qbnd);
$kind = 'pp1' if !defined $qbnd || (defined $obnd && $qbnd <= $obnd);

# Do P+1 if neither has been done for this Mersenne.
$qbnd = 0, $kind = 'pp1' unless defined($obnd) || defined($qbnd);

$args .= " -$kind";

# Append all output to this and only print it just before exiting.
my ($output) = '';
END
{
  print($output) if $output;
}

$subdir = ECMSave::subdir($uuid, $e);
$subdir = 'p+-1.sav'
  unless $subdir && -d $subdir && ! -l $subdir;
-d("$subdir") || mkdir("$subdir")
  or die "could not mkdir $subdir, $!";

my (%need) = ( 'S' => ($e % 8 != 4), 'L' => ($e % 8 == 4), 'M' => ($e % 8 == 4), );
foreach my $part ('S', 'L', 'M', )
{
  next unless $need{$part};
  $need{$part} = 0,
    next
    if defined $mers->get($part) && $mers->get($part) == $mers->cofactorLength($part);
  # A 'c' cofactor save file is named the same as an 'l' cofactor save file.
  #   Save files for 'm' cofactors include a trailing '.M'.
  my ($which) = $part eq 'M' ? 1 : 0;
  my ($save) = ECMSave::findBestSaveFile($e, $kind, $which ? '.M' : '', $mers);
  die($save ? $save : "ECMSave::findBestSaveFile M$e $kind $which returned undef")
    unless $save && ref($save);
  my (@cofactors) = $mers->c();
  my ($savedBound) = $save ? $save->bound() : 0;
  # If there is no save file, only go to $bndIncr.  Otherwise, a large bound
  #   without a save file could take days or weeks; avoid that.
  my ($bnd) = 0;
  $bnd = $savedBound
    if $savedBound && $bnd < $savedBound;
  my ($savefn) = $save->file()
    if $save;
  $savefn = ($subdir . "/ecm.$e.$kind" . ($which ? '.M' : ''))
    unless $savefn;
  my ($newSavefn) = "$savefn.new";
  die("new save file $newSavefn is a directory; cannot continue at")
    if -d $newSavefn;
  if (-l($newSavefn))
    {
      print($logFH "new save file is a soft link; deleting it\n");
      unlink($newSavefn)
        or die("could not unlink $newSavefn, $!\n");
    }
  elsif (-s($newSavefn))
    {
      print($logFH "new save file $newSavefn already exists; 'rotating' it\n");
      MersInfo::rotateLog($newSavefn, 99);
    }
  elsif (-f($newSavefn))
    {
      print($logFH "new save file is empty; deleting it\n");
      unlink($newSavefn)
        or die("could not unlink $newSavefn, $!\n");
    }
  my ($cffn) = "$savefn.cofactor";
  my ($resume) = "-go " . (2*$e) . " -inp $cffn -savea $newSavefn";
  $resume .= " -resume $savefn"
    if $savedBound && -s "$savefn";
  open(my $numFH, '>', "$cffn") or die "could not create $cffn, $!\n";
  print($numFH "$cofactors[$which]\n");
  close($numFH) or die "could not close $cffn, $!\n";
  my ($newB1) = $bnd + $bndIncr;
  $newB1 = $maxB1 if $newB1 > $maxB1;
  $newB1 = $Mersenne::maxStage1Bound if $newB1 > $Mersenne::maxStage1Bound;
  $output .= "M( $e )" . ($kind eq 'pm1' ? 'o' : 'q') . ": $bnd\n",
    next
    if $bnd >= $newB1;
  my ($logfn) = $newSavefn;
  $logfn =~ s,[.]new$,.log,;
  open(my $ecmLogFH, '>', "$logfn") or die "could not create $logfn, $!\n";
  # Always do at least some stage 2 work; see comment with $maxmem above.
  my ($bnd2) = $newB1 + ($maxmem ? $bndIncr : 10000);
  # Doing stage2 for $bndIncr is too expensive for large cofactors - "disable" it for now.
  $bnd2 = $newB1 + 10000;
  print($ecmLogFH "$args $resume $newB1 $bnd2\n");
  print($logFH "$args $resume $newB1 $bnd2\n");
  open(my $runFH, "$args $resume $newB1 $bnd2 2>&1 |")
    or die "cannot run $ecm executable, $!\n";
  my ($prpcofactor);
  my ($newFactors); # or 1 if some are found
  my (%lines);
  while (<$runFH>)
    {
      print($logFH "$_");
      print($ecmLogFH "$_");
      chomp;
      $lines{$1} = $_,
        next
        if m,^(GMP-ECM) ,;
      $lines{'input'} = $1,
        next
        if m,^Input number is ([1-9][0-9]+) ,;
      if (m/^Found .* factor of .* digits: ([1-9][0-9]*)$/)
        {
          my ($f) = Math::BigInt->new("$1");
          $mers->f($f);
          $newFactors = 1;
          $output .= 'M( ' . "$e" . ' )C: ' . "$f\n";
          next;
        }
      $prpcofactor = 1
        if m/prime cofactor [1-9][0-9]* has [1-9][0-9]* digits/i;
      $lines{'B1'} = $1,
        $lines{'B2'} = $2
        if (m/^Using B1=([1-9][0-9]*), B2=([1-9][0-9]*)/);
      $lines{'B1'} = $1
        if (m/^Using B1=[1-9][0-9]*-([1-9][0-9]*)/);
    }
  if (!close($runFH))
    { # The ecm program exits non-zero if a factor found, so this is messy.
      my ($status) = $? >> 8; # see 'man perlvar'
      # See 'man ecm' for why these exit statuses are OK when a factor is found.
      $newFactors && !($status & ~(2|4|8))
        or die "ecm program failed, exiting $?";
    }
  close($ecmLogFH) || die "could not close $logfn, $!\n";
  last if $debug;
  if ($prpcofactor)
    {
      ! -s("$newSavefn")
        or die "PRP cofactor but $newSavefn is not empty\n";
      unlink("$newSavefn")
        or die "could not remove $newSavefn, $!\n";
      # could print M()[SLM] line but unsure how many PRP tests were done
      next;
    }
  my ($save2) = ECMSave->new('file' => $newSavefn, 'exp' => $e, 'M' => $which);
  die($save2 ? $save2 : "ECMSave::new $newSavefn M$e $which returned undef")
    unless $save2 && ref($save2);
  $save2->str()
    or die "problems saving or printing new save file '$newSavefn', $! - " . $save2->errors();
  $output .= $save2->str();
  $savedBound = $save2->bound();
  $savedBound
    or die "new ecm save file $newSavefn had unexpected content\n";
  # Do not move the new save file into a different top level directory.
  #   This should only matter when the old save file is under serv.sav, which
  #   mersServer.pl may access at any time.  In contrast, only this script
  #   modifies files under p+-1.sav/ - the server may read and/or copy them,
  #   but doesn't remove or modify them.
  if (substr($savefn, 0, index($savefn, '/')) ne substr($newSavefn, 0, index($newSavefn, '/')))
    {
      $savefn = $newSavefn;
      $savefn =~ s,[.]new,,;
    }
  rename("$newSavefn", "$savefn")
    or die "could not rename $newSavefn to $savefn, $!\n";
  $savefn = substr($savefn, 0, rindex($savefn, '/') + 1)
    if $savefn =~ m,/,;
  $savefn = ''
    unless $savefn =~ m,/,;
  $savefn .= substr($logfn, rindex($logfn, '/') + 1);
  $logfn eq $savefn
    || rename("$logfn", "$savefn")
    or die "could not rename $logfn to $savefn, $!\n";
  $output .= "M( $e )" . ($kind eq 'pm1' ? 'o' : 'q') . ": $savedBound\n";
}
exit(0);
