#!/usr/bin/perl -Tw
# Use primo/latest/primo or primo/primo to prove the primality of SPRP factors of Mersenne numbers.
#   I've started creating 'latest' as a soft link to, e.g., '4.3.3' where that is
#     the latest version of primo from http://www.ellipsa.eu/index.html
# This script will complain if factors appear after cofactors, etc., but not verify that
#   they divide the Mersenne number.
# The primo program currently has no command-line interface, so the user has to load
#   the primo/$exp.$n.in files to be "certified" (all at once or one at a time) and confirm the result.

use strict;
use strict 'vars';
use warnings;

BEGIN { push(@INC, '.') if -s('Mersenne.pm'); $ENV{'PATH'} = "/usr/bin:/bin"; }
use Mersenne;
use MersInfo;

my ($usage) = ("Usage: doPrimo.pl [-d [debugLevel]] [--check] [--version] [-h|--help]\n"
               . "  [--stop <directory>] [--skip <directory>] [--tmp <directory>]\n"
               . "  [--stop=<directory>] [--skip=<directory>] [--tmp=<directory>]\n"
  );

# factordb.com only retains primality certificates for 300+ digits
my ($minFDBretains) = Math::BigInt->new('1e300');
# count of factors (including cofactors) that are "big enough" for factordb.com to have and want certs
my ($bigEnough) = 0;

my ($checkConfig) = scalar(grep { m/^--check$/ } @ARGV)
  if @ARGV;
my ($version) = MersInfo::version();
print(STDOUT "$0 version $version\n")
  if $checkConfig || grep { $_ eq '--version' } @ARGV;
print(STDOUT "$usage")
  if grep { $_ eq '--help' || $_ eq '-h' } @ARGV;
exit(0)
  if grep { $_ eq '--version' || $_ eq '--help' || $_ eq '-h' } @ARGV;

my ($debug) = 0;
while (@ARGV)
{
  shift, next if $ARGV[0] =~ m/^--check$/;
  shift, shift, next if $ARGV[0] eq '--skip' && $#ARGV > 0 && $ARGV[1] =~ m,^([-._/[:alnum:]]+)$,;
  shift, shift, next if $ARGV[0] eq '--stop' && $#ARGV > 0 && $ARGV[1] =~ m,^([-._/[:alnum:]]+)$,;
  shift, shift, next if $ARGV[0] eq '--tmp' && $#ARGV > 0 && $ARGV[1] =~ m,^([-._/[:alnum:]]+)$,;
  shift, next if $ARGV[0] =~ m,^--(skip|stop|tmp)=([-._/[:alnum:]]+)$,;
  $ARGV[0] eq '-d' or die "unexpected argument '$ARGV[0]'; only --check, --version, and -d <debugLevel> are supported\n";
  $debug = 1;
  shift;
  $debug = $1, shift if @ARGV && $debug && $ARGV[0] =~ m/^([0-9]+)$/;
}

my ($errors) = '';
my ($prove) = './primo';
my ($subdir);
# Find the directory where the executable lives.
foreach ('.', './primo', './primo/latest')
{
  $subdir = $_
    if -x "$_/primo" && ! -d "$_/primo";
}
$checkConfig || defined($subdir)
  or $errors .= "primo executable is not in ., ./primo, or ./primo/latest\n";
!$checkConfig || defined($subdir)
  or print("primo executable is not in ., ./primo, or ./primo/latest subdirs\n");
-d("primo/done") || mkdir("primo/done")
  or $errors .= "could not mkdir primo/done, $!\n";

exit(0)
  if $checkConfig;

mkdir('primo/lock') or $errors .= "could not mkdir primo/lock, $!\n";
die "$errors"
  if $errors;

$SIG{'HUP'} = "IGNORE";
foreach ('INT', 'TERM')
{
  $SIG{$_} = sub {
    my ($sig) = @_;
    warn "caught SIG$sig\n";
  };
}

eval
{ # setpriority() is not always implemented
  setpriority(0, 0, 20);
};

# These go in each file primo has to be given to do the proof of the number (between them).
my ($prefix) = "[Candidate]\nN=";
my ($suffix) = '';
my ($commentChar) = ';'; # primo treats the rest of a line as a comment
my ($minDigits) = 6; # primo doesn't support "candidates" less than 65,537.

my ($cnt) = 0;
my ($data) = '';
my ($line, $e, $m);
my ($mers);
my ($fh);
while (defined($line = <STDIN>))
{
  chomp($line);
  $line =~ /^M[(] *([1-9][0-9]*) *[)]([CLMSd]): *([1-9][0-9]*) *([0-9]*)/o
    or $line =~ /^M[(] *([1-9][0-9]*) *[)]([LMSd])/o
    or die "unexpected input:\n$data$line\n";
  defined($3) || $2 ne 'C'
    or die "missing factor: $data$line\n";
  $e = $1,
    $mers = Mersenne->new('n', "$e")
    unless defined $e;
  $e eq $1 or die "data for multiple exponents in input:\n$data$line\n";
  !defined($m) or $m eq 'C' or ($m eq 'L' && $e % 8 == 4)
    or die "Unexpected mers data: $data$line\n";
  defined($m = $2)
    or die "bug: missing mers letter: $data$line\n";
  my ($one, $two);
  $one = Math::BigInt->new("$3")
    if defined $3;
  $two = "$4"
    if defined $4;
  $data eq '' || $m eq 'C' || index($data, ')' . $m . ':') == -1
    or die "Multiple $m lines in mers data: $data$line\n";
  next
    unless $m eq 'C';
  defined($one)
    or die "no factor given: $line\n";
  $data .= "$line\n";
  $mers->f($one);
  my ($digits) = $one->length();
  next
    if $digits < $minDigits;
  # only factors are dealt with here; cofactors in the input are ignored
  ++$cnt;
  ++$bigEnough
    if $digits > $minFDBretains;
  ! -f("$subdir/$e.$cnt.in") && ! -s("$subdir/$e.$cnt.in")
    or die "$subdir/$e.$cnt.in already exists\n";
  open($fh, '>', "$subdir/$e.$cnt.in")
    or die "cannot create $subdir/$e.$cnt.in, $!\n";
  print $fh "$prefix$one$suffix\n";
  print $fh "$commentChar $line\n"; # record the input line in the same file
  close($fh)
    or die "cannot close $subdir/$e.$cnt.in, $!\n";
}
defined($e)
  or die "no input, $!";
defined($m) && ($m =~ m/^[CSd]$/ || ($m eq 'M' && $e % 8 == 4))
  or die "Unexpected mers data: $data";

my (@cofactors) = $mers->c();
my ($letter) = ($e % 8 == 4) ? 'L' : 'S';
foreach (@cofactors)
{
  ! -f("$subdir/$e.$letter.in") && ! -s("$subdir/$e.$letter.in")
    or die "$subdir/$e.$letter.in already exists\n";
  open($fh, '>', "$subdir/$e.$letter.in")
    or die "cannot create $subdir/$e.$letter.in, $!\n";
  print $fh "$prefix$_$suffix\n";
  print $fh "$commentChar M( $e )$letter\n";
  close($fh)
    or die "cannot close $subdir/$e.$letter.in, $!\n";
  $letter = 'M';
  ++$bigEnough
    if $_ > $minFDBretains;
}

my (@list) = (1 .. $cnt);
push(@list, 'S')
  if $e % 8 != 4;
push(@list, 'L', 'M')
  if $e % 8 == 4;

# Find existing primality certificates and move them into the $subdir directory
#   so the user will see them in the primo GUI.
my (@dirs) = ('./certs', '.', './primo/done/', './primo/done/factordb/', );
my (%certs);
foreach my $d (@dirs)
{
  foreach (@list)
    {
      next
        if exists $certs{$_};
      $certs{$_} = "cert.M$e.$_.out",
        next
        if -s "$subdir/cert.M$e.$_.out";
      ! -s("$subdir/cert.M$e.$_.out.primo")
        || rename("$subdir/cert.M$e.$_.out.primo", "$subdir/cert.M$e.$_.out")
        or die "could not move $subdir/cert.M$e.$_.out.primo to $subdir/cert.M$e.$_.out, $!\n";
      $certs{$_} = "cert.M$e.$_.out",
        next
        if -s "$subdir/cert.M$e.$_.out";
      ! -s("$d/cert.M$e.$_.out")
        || rename("$d/cert.M$e.$_.out", "$subdir/cert.M$e.$_.out")
        or die "could not move $d/cert.M$e.$_.out to $subdir/cert.M$e.$_.out, $!\n";
      $certs{$_} = "cert.M$e.$_.out",
        next
        if -s "$subdir/cert.M$e.$_.out";
      ! -s("$d/cert.M$e.$_.out.primo")
        || rename("$d/cert.M$e.$_.out.primo", "$subdir/cert.M$e.$_.out")
        or die "could not move $d/cert.M$e.$_.out.primo to $subdir/cert.M$e.$_.out, $!\n";
      $certs{$_} = "cert.M$e.$_.out"
        if -s "$subdir/cert.M$e.$_.out";
    }
}

if (scalar(keys(%certs)) < $bigEnough && !-f('mysql.devel'))
{ # ask factordb.com for primality certificates
  my ($pid) = open(FACTORDB, '-|', './askFactordb.pl', '--certs', "$e");
  $pid
    or die "could not run ./askFactordb.pl --certs $e, $!\n";
  while (<FACTORDB>)
    {
      chomp;
      -f && -s
        or die "./askFactordb.pl printed $_ but that file does not exist or is empty\n";
      if (!m/^cert[.]M([1-9][0-9]*)[.]([LMS]|[0-9A-F]+)[.]out$/ || $e != $1)
        {
          print(STDERR "# factordb.com cert file name created by askFactordb.pl has unexpected format: $_\n")
            if $debug;
          next;
        }
      rename("$_", "$subdir/$_")
        or die "could not rename cert file $_ to $subdir subdir, $!\n";
      $certs{$2} = "$_";
      unlink("$subdir/$e.$2.in")
        or die "could not remove $subdir/$e.$2.in after creating it, $!\n";
    }
  close(FACTORDB)
    or die "could not close ./askFactordb.pl, $!\n";
  print(STDERR "# cert files from factordb.com: $subdir/" . join(' $subdir/', sort(values(%certs))) . "\n")
    if $debug > 2 && scalar(keys(%certs));
}
elsif (scalar(keys(%certs)))
{
  print(STDERR "# needed cert files already exist: $subdir/" . join(' $subdir/', sort(values(%certs))) . "\n")
    if $debug > 2;
  foreach (keys(%certs))
    {
      -s("$subdir/$certs{$_}")
        or die "cannot find cert file $subdir/$certs{$_}, $!\n";
      next unless -s("$subdir/$e.$_.in");
      unlink("$subdir/$e.$_.in")
        or die "could not remove $subdir/$e.$_.in after creating it, $!\n";
    }
}

my ($cwd) = $ENV{'PWD'};
$cwd =~ m,^([-:./a-zA-Z0-9]+)$,
  or die "PWD in environment has unexpected characters\n";
$cwd = "$1";
-d $cwd
  or die "PWD in environment is not a directory - $cwd, $!";
$subdir eq '.' || chdir("$subdir")
  or die "could not change directory to $subdir, $!";
system("$prove")
  && die "$prove failed, " . ($! ? $! : $?) . "\n";
$subdir eq '.' || chdir("$cwd")
  or die "could not change directory back to $cwd, $!";

# Check the "work report files" to verify all the factors were "certified prime".
#   More than one number (Input=<number> in one *.in file) can be "certified" in one "work report" file (*.wr).
my ($dh);
opendir($dh, "$subdir")
  or die "could not opendir $subdir, $!\n";
my (@reports, %toRename);
foreach (readdir($dh))
{
  push(@reports, $_) if m/^primo-[0-9A-F]+[.]wr$/ && -f "$subdir/$_" && -s "$subdir/$_";
}
closedir($dh)
  or die "could not closedir $subdir, $!\n";

$toRename{$_} = 1
  foreach (values(%certs));
print(STDERR "# report files in $subdir subdir: " . join(' ', @reports) . "\n"
      . "# files to rename in $subdir subdir: " . join(' ', keys(%toRename)) . "\n")
  if $debug > 2;

#!!Checking the report files like this is fine, but verifying that the factors
#!!  and cofactors were tested can only be done by reading the certificate files.
my (%verified);
my ($certified);
foreach my $file (@reports)
{
  my ($inputE, $inputI);
  open($fh, '<', "$subdir/$file") or die "could not read $subdir/$file, $!\n";
  while (<$fh>)
    {
      chomp;
      $certified = 1
        if m/^Status=Candidate certified prime$/;
      $certified = 1
        if m/^Status=Valid (certificate|signature)$/;
      $certified = 1
        if m/^Certificate( [1-9][0-9]*)?=Valid, [0-9]+[.]?[0-9]*s$/;
      $inputE = $2, $inputI = $3
        if m/^(File|Input)=([1-9][0-9]+)[.]([LMS]|[0-9A-F]+)[.]in$/ && $2 == $e;
      $inputE = $2, $inputI = $3
        if m/^(File|Input)=cert[.]M([1-9][0-9]+)[.]([LMS]|[0-9A-F]+)[.]out$/ && $2 == $e;
      $toRename{$1} = 1
        if m/^Report=(primo-[0-9A-F]+[-][0-9A-F]+[.]cr)$/ && -s "$subdir/$1";
      $toRename{$1} = 1
        if m/^Output=(primo-[0-9A-F]+([-][0-9A-F]+)?[.]out)$/ && -s "$subdir/$1";
      next
        unless defined($inputE) && defined($inputI) && $certified;
      $verified{$inputE . '.' . $inputI . '.in'} = $file;
      print(STDERR "# marked $inputE $inputI verified in $file\n")
        if $debug;
      $certified = $inputE = $inputI = undef;
    }
  
  print(STDERR "# found a certified line without a File or Input line in $file\n")
    if defined $certified;
  print(STDERR "# found a File or Input line without a certified line in $file\n")
    if defined $inputE or defined $inputI;
  close($fh)
    or die  "could not close $subdir/$file, $!\n";
}

foreach (@list)
{
  exists($verified{"$e.$_.in"})
    or die "M( $e )C #$_ was not certified\n";
  -f("$subdir/$verified{$e . '.' . $_ . '.in'}") && -s("$subdir/$verified{$e . '.' . $_ . '.in'}")
    or die "M( $e )C #$_ was not certified\n";
  $toRename{"$e.$_.in"} = 1
    if -f "$subdir/$e.$_.in" && -s "$subdir/$e.$_.in"; # could have been unlink()ed if matching cert dl'd from factordb.com
  $toRename{$verified{"$e.$_.in"}} = 1;
}
foreach (keys(%toRename))
{
  ! -e("primo/done/$_")
    or die "primo/done/$_ already exists\n";
}
foreach (keys(%toRename))
{
  if (-s "$_")
    {
      my ($final) = $_;
      $final =~ s,^.*/,,;
      rename("$_", "primo/done/$final")
        or die "could not rename $_ to primo/done/$final, $!\n";
      next;
    }
  if (-s "$subdir/$_")
    {
      my ($final) = $_;
      $final =~ s,^.*/,,;
      rename("$subdir/$_", "primo/done/$final")
        or die "could not rename $subdir/$_ to primo/done/$final, $!\n";
      next;
    }
  die "cannot rename $_ - no such file or empty file\n";
}
print "$data" . "M( $e )D\n";
rmdir('primo/lock')
  or die "could not rmdir primo/lock, $!\n";
exit(0);

# 'Work report' file created by primo v4.3.3 :
=pod
[PRIMO - Work Report]
Version=4.3.3 - LX64
WebSite=https://www.ellipsa.eu/
Work=Verification
ID=B433801A0E39A
Created=Nov-10-2021 07:35:21 AM

[Common]
Path=/home/wedgingt/src/math/primes/Mersenne/lucas/primo/4.3.3/
Selected=1
Processed=1
Valid=1
Certificate=Valid, 0.04s

[Certificate]
Input=cert.M6485.S.out
Status=Valid signature


# 'Work report' file created by primo v4.1.1 :
% more primo/primo-B3AC9040BE2AC.wr 
[PRIMO - Work Report]
Version=4.1.1 - LX64
WebSite=http://www.ellipsa.eu/
Work=Verification
ID=B3AC9040BE2AC
Created=12-13-2015 06:51:27 PM

[Common]
Path=/mnt/jupiter/home/wedgingt/src/math/primes/Mersenne/lucas/primo/
Selected=1
Processed=1
Certified=1
Candidate 1=Certified, 374s

[Candidate 1]
Input=7754.3.in
Report=primo-B3AC9040BE2AC-001.cr
Output=primo-B3AC9040BE2AC-001.out
Status=Candidate certified prime

% 
=pod
[PRIMO - Work Report]
Version=4.3.3 - LX64
WebSite=https://www.ellipsa.eu/
Work=Verification
ID=B433A0205821B
Created=Nov-12-2021 09:25:15 AM

[Common]
Path=/home/wedgingt/src/math/primes/Mersenne/lucas/primo/4.3.3/
Selected=2
Processed=2
Valid=2
Certificate 1=Valid, 0.02s
Certificate 2=Valid, 0.01s

[Certificate 1]
Input=cert.M13700.L.out
Status=Valid signature

[Certificate 2]
Input=cert.M13700.M.out
Status=Valid signature


