#!/usr/bin/perl -Tw
# Create or verify primality certificates for all factors of a Mersenne number.

use strict;
use strict 'vars';
use warnings;

BEGIN { push(@INC, '.') if -s('Mersenne.pm'); $ENV{'PATH'} = "/usr/bin:/bin"; }
use Mersenne;
use MersInfo;

my ($usage) = ("Usage: doPrimo.pl [-d [debugLevel]] [--check] [--version] [-h|--help]\n"
               . "  [--stop <directory>] [--skip <directory>] [--tmp <directory>]\n"
               . "  [--stop=<directory>] [--skip=<directory>] [--tmp=<directory>]\n"
  );

my (%certHeaders) =
  ( # key is certificate kind/type/style; value is pattern that matches the first line in such a file
    # 'MPU' certificates are created and verified by sub()s in Math::Prime::Util.
   'MPU' => '^[[]MPU - Primality Certificate[]]$',
    # 'PRIMO' certificatres are created and verified by the primo program;
    #   see http://www.ellipsa.eu/index.html
   'PRIMO' => '^[[]PRIMO - Primality Certificate[]]$',
  );

# Run the primo program to create or verify a single PRIMO format certificate.
#   The first three arguments are the exponent of the Mersenne number
#     and the factor "index" (1 for the first (smallest) factor, e.g.)
#     or S, L, or M (for those cofactors), and the factor to be proven prime.
sub createPRIMOcert($$$);
sub verifyPRIMOcert($$$$); # fourth argument is the file containing the certificate to verify

my ($primoDir); # directory containing the primo executable
foreach ('./primo', './primo/latest')
{
  $primoDir = $_
    if -x "$_/primo" && ! -d "$_/primo";
}

# factordb.com only wants primality certificates for 300+ digits and only accepts PRIMO certificates.
my ($minFDBretains) = Math::BigInt->new('1e300');
# Count of factors (including cofactors) that are "big enough" for factordb.com to have and want certs.
my ($bigEnough) = 0;

my ($checkConfig) = scalar(grep { m/^--check$/ } @ARGV)
  if @ARGV;
my ($version) = MersInfo::version();
print(STDOUT "$0 version $version\n")
  if $checkConfig || grep { $_ eq '--version' } @ARGV;
print(STDOUT "$usage")
  if grep { $_ eq '--help' || $_ eq '-h' } @ARGV;
exit(0)
  if grep { $_ eq '--version' || $_ eq '--help' || $_ eq '-h' } @ARGV;

my ($debug) = 0;
while (@ARGV)
{
  shift, next if $ARGV[0] =~ m/^--check$/;
  shift, shift, next if $ARGV[0] eq '--skip' && $#ARGV > 0 && $ARGV[1] =~ m,^([-._/[:alnum:]]+)$,;
  shift, shift, next if $ARGV[0] eq '--stop' && $#ARGV > 0 && $ARGV[1] =~ m,^([-._/[:alnum:]]+)$,;
  shift, shift, next if $ARGV[0] eq '--tmp' && $#ARGV > 0 && $ARGV[1] =~ m,^([-._/[:alnum:]]+)$,;
  shift, next if $ARGV[0] =~ m,^--(skip|stop|tmp)=([-._/[:alnum:]]+)$,;
  $ARGV[0] eq '-d' or die "unexpected argument '$ARGV[0]'; only --check, --version, and -d <debugLevel> are supported\n";
  $debug = 1;
  shift;
  $debug = $1, shift if @ARGV && $debug && $ARGV[0] =~ m/^([0-9]+)$/;
}

Math::Prime::Util::prime_set_config('verbose' => ($debug ? $debug : 1));

my ($errors) = '';
my ($subdir) = 'certs';
my ($lock) = "$subdir/lock";
-d($subdir) || mkdir($subdir) or $errors .= "could not mkdir $subdir, $!";
-d("$subdir/done") || mkdir("$subdir/done") or $errors .= "could not mkdir $subdir/done, $!";
$checkConfig || mkdir($lock) or $errors .= "could not mkdir $lock, $!";
defined($primoDir)
  or print(STDERR "primo executable is not in ./primo or ./primo/latest subdirs\n");

exit(0) if $checkConfig;

$SIG{'HUP'} = "IGNORE";
foreach ('INT', 'TERM')
{
  $SIG{$_} = sub {
    my ($sig) = @_;
    warn "caught SIG$sig\n";
  };
}

eval
{ # setpriority() is not always implemented
  setpriority(0, 0, 20);
};

my ($cnt) = 0;
my ($data) = '';
my ($e, $m);
my ($mers);
my (%toProve);
while (<STDIN>)
{
  chomp;
  m/^M[(] *([1-9][0-9]*) *[)]([CLMSd]): *([1-9][0-9]*) *([0-9]*)/o
    or m/^M[(] *([1-9][0-9]*) *[)]([LMSd])/o
    or die "unexpected input:\n$data$_\n";
  defined($3) || $2 ne 'C' or die "missing factor: $data$_\n";
  $e = $1,
    $mers = Mersenne->new('n', "$e")
    if (!defined($e));
  $e eq $1 or die "data for multiple exponents in input:\n$data$_\n";
  !defined($m) or $m eq 'C' or ($m eq 'L' && $e % 8 == 4) or die "Unexpected mers data: $data$_\n";
  defined($m = $2) or die "bug: missing mers letter: $data$_\n";
  my ($one, $two);
  $one = Math::BigInt->new("$3") if defined $3;
  $two = "$4" if defined $4;
  $data eq '' || $m eq 'C' || index($data, ')' . $m . ':') == -1 or die "Multiple $m lines in mers data: $data$_\n";
  next if $m ne 'C';
  defined($one) or die "missing factor: $_\n";
  $data .= "$_\n";
  $mers->f($one);
  # only factors are dealt with here; cofactors in the input are ignored
  ++$bigEnough if $one >= $minFDBretains;
  $toProve{++$cnt} = $one;
}
defined($e) or die "no input, $!";
defined($m) && ($m =~ m/^[CSd]$/ || ($m eq 'M' && $e % 8 == 4)) or die "Unexpected mers data: $data";

my (@cofactors) = $mers->c();
my ($letter) = 'S';
$letter = 'L' if $e % 8 == 4;
foreach (@cofactors)
{
  ++$bigEnough if $_ >= $minFDBretains;
  $toProve{$letter} = $_;
  $letter = 'M';
}

# Find existing primality certificates and move them into the $subdir/ subdirectory.
my (@dirs) = ($subdir, '.', $subdir . '/done/', $subdir . '/done/factordb/', );
my (%certs);
my (@list) = (1 ... $cnt);
push(@list, 'S') if $e % 8 != 4;
push(@list, 'L', 'M') if $e % 8 == 4;
foreach my $d (@dirs)
{
  foreach my $n (@list)
    { # Prefer PRIMO certs since factordb.com does not accept others.
      next if exists $certs{$n};
      foreach my $kind ('.PRIMO', '.MPU', '.') # last is unknown - get kind from content
        {
          next unless -s "$d/cert.M$e.$n$kind.out";
          $certs{$n} = "cert.M$e.$n$kind.out",
            last
            if -s "$subdir/cert.M$e.$n$kind.out";
          rename("$d/cert.M$e.$n$kind.out", "$subdir/cert.M$e.$n$kind.out")
            or die "could not move cert.M$e.$n$kind.out into $subdir/, $!";
          $certs{$n} = "cert.M$e.$n$kind.out";
          last;
        }
    }
}

if (scalar(keys(%certs)) < $bigEnough && !-f('mysql.devel'))
{ # Ask factordb.com for primality certificates (if production install of mers package - mysql.devel).
  my ($fdbfh);
  my ($pid) = open($fdbfh, '-|', './askFactordb.pl', '--certs', "$e");
  $pid or die "could not run ./askFactordb.pl --certs $e, $!";
  while (<$fdbfh>)
    {
      chomp;
      (-f("$subdir/$_") && -s(_) && !-l("$subdir/$_")) # call stat() once, not twice
        || (-f($_) && -s(_) && !-l($_)) # again
        or die "./askFactordb.pl printed $_ but that file does not exist or is empty\n";
      if (!m/^cert[.]M([1-9][0-9]*)[.]([LMS]|[0-9A-F]+)[.]out$/ || $e != $1)
        {
          print(STDERR "# factordb.com cert file name created by askFactordb.pl has unexpected format: $_\n")
            if $debug;
          next;
        }
      $certs{$2} = "$_";
      -s("$subdir/$_") or rename("$_", "$subdir/$_")
        or die "could not move cert file $_ into $subdir, $!";
    }
  close($fdbfh) or die "could not close ./askFactordb.pl, $!";
  print(STDERR "# cert files from factordb.com: $subdir/" . join(" $subdir/", sort(values(%certs))) . "\n")
    if $debug > 2 && scalar(keys(%certs));
}
elsif (scalar(keys(%certs)))
{
  print(STDERR "# needed cert files already exist: $subdir/" . join(" $subdir/", sort(values(%certs))) . "\n")
    if $debug > 2;
  foreach (keys(%certs))
    {
      -s("$subdir/$certs{$_}") or die "cannot find cert file $subdir/$certs{$_}, $!";
    }
}

# M($n) might have no factors, just cofactors, but it has to have an S or L cofactor
my ($n) = 1;
$n = 'S' unless exists $toProve{$n};
$n = 'L' unless exists $toProve{$n};
while (exists($toProve{$n}))
{
  if (exists($certs{$n}))
    {
      my ($certfh);
      open($certfh, '<', "$subdir/$certs{$n}") or die "could not read $subdir/$certs{$n}, $!";
      my ($certKind);
      my ($cert) = ''; # content of cert file, only needed for MPU certificates
      my ($line) = -1;
      while (<$certfh>)
        {
          $cert .= $_
            if defined $cert;
          chomp;
          s,[\r\n]+$,,;
          if ($. == 1)
            {
              foreach (keys(%certHeaders))
                {
                  $certKind = $_,
                    last
                    if m,$certHeaders{$_},;
                }
              if (!$certKind)
                {
                  print(STDERR "ignoring $subdir/$certs{$n} certificate file - unknown format\n")
                    if $debug;
                  rename("$subdir/$certs{$n}", "$subdir/$certs{$n}.unknown")
                    or die "could not rename to $subdir/$certs{$n}.unknown, $!";
                  print(STDERR "renamed to $subdir/$certs{$n}.unknown\n")
                    if $debug;
                  last;
                }
              $cert = undef # the primo program reads the file directly
                if $certKind eq 'PRIMO';
              next
                if $certs{$n} =~ m,[.]$certKind[.]out$,;
              my ($withKind) = $certs{$n};
              $withKind =~ s,[.]out$,.$certKind.out,
                or die("logic error - $certs{$n} does not end with '.out'");
              rename("$subdir/$certs{$n}", "$subdir/$withKind")
                or die("could not move $subdir/$certs{$n} to $subdir/$withKind");
            }
          $line = $. + 1, next if m/^Proof for:$/;
          $line = $. + 2, next if m/^[[]Candidate[]]$/;
          next unless $line == $.;
          my ($number);
          if (m/^N ([1-9][0-9]*)$/)
            { # 'MPU' (Perl Math::Prime::Util::verify_prime()) format number proven
              die("file $subdir/$certs{$n} starts with non-MPU line but has MPU format for number to prove prime\n")
                unless $certKind && $certKind eq 'MPU';
              $number = Math::BigInt->new("$1");
            }
          elsif (m/^N[=]\$([1-9A-F][0-9A-F]*)$/)
            { # PRIMO format number proven
              die("file $subdir/certs{$n} starts with non-PRIMO line but has PRIMO format for number to prove prime\n")
                unless $certKind && $certKind eq 'PRIMO';
              $number = Math::BigInt->from_hex("$1");
            }
          else
            {
              print(STDERR "ignoring $subdir/$certs{$n} certificate file - unknown format\n")
                if $debug;
              rename("$subdir/$certs{$n}", "$subdir/$certs{$n}.unknown")
                or die "could not rename to $subdir/$certs{$n}.unknown, $!";
              print(STDERR "renamed to $subdir/$certs{$n}.unknown\n")
                if $debug;
              $cert = undef;
              last;
            }
          next if $number == $toProve{$n};
          print(STDERR "cert in $subdir/$certs{$n} is for a different number, not factor $n\n")
            if $debug;
          rename("$subdir/$certs{$n}", "$subdir/$certs{$n}.wrong")
            or die "could not rename to $subdir/$certs{$n}.wrong, $!";
          print(STDERR "renamed to  $subdir/$certs{$n}.wrong\n")
            if $debug;
          $cert = undef;
          last;
        }
      if ($cert && $certKind && $certKind eq 'MPU')
        {
          my ($result) = Math::Prime::Util::verify_prime($cert);
          next if $result;
          die "MPU cert in $subdir/$certs{$n} for factor $n failed to verify\n";
        }
      if ($certKind && $certKind eq 'PRIMO')
        {
          die "PRIMO cert in $subdir/$certs{$n} cannot be verified without 'primo' executable\n"
            unless $primoDir;
          my ($result) = verifyPRIMOcert($e, $n, $toProve{$n}, $subdir/$certs{$n});
          next if $result;
          die "PRIMO cert in $subdir/$certs{$n} for factor $n failed to verify\n";
        }
      die "unknown certificate file $subdir/$certs{$n} type or content\n";
    }
  if ($toProve{$n} >= $minFDBretains && !-f('mysql.devel'))
    { # Use primo for "large" numbers unless testing.
      die "PRIMO certificates cannot be verified without 'primo' executable\n"
        unless $primoDir;
      my ($result) = createPRIMOcert($e, $n, $toProve{$n});
      die("PRIMO failed to prove M$e factor $n prime\n")
        unless $result;
      next;
    }
  my ($result, $proof) = Math::Prime::Util::is_provable_prime_with_cert($toProve{$n});
  die("factor $n is composite\n")
    if $result == 0;
  die("factor $n could not be proved prime\n")
    if $result == 1;
  die("proof $n is prime is empty\n")
    unless $proof;
  my ($certfh);
  open($certfh, '>', "$subdir/cert.M$e.$n.out")
    or die("could not create $subdir/cert.M$e.$n.out to save proof of factor $n, $!");
  print($certfh $proof)
    or die("could not print proof of factor $n into $subdir/cert.M$e.$n.out, $!");
  close($certfh)
    or die("could not close $subdir/cert.M$e.$n.out to save proof of factor $n, $!");
}
continue
{
  if ($n =~ m,^[SM]$,)
    {
      $n = 'done';
    }
  elsif ($n eq 'L')
    {
      $n = 'M';
    }
  else
    {
      ++$n;
      $n = 'L' if !exists($toProve{$n}) && $e % 8 == 4;
      $n = 'S' if !exists($toProve{$n}) && $e % 8 != 4;
    }
}

print "$data" . "M( $e )D\n";
rmdir($lock) or die "could not rmdir $lock, $!";
exit(0);

# The rest of this file is related to using the primo program.

my (%primo) =
  ( # These go in each file primo has to be given to create the primality proof of a number (between them).
   'prefix' => "[Candidate]\nN=",
   'suffix' => '',
   'commentChar' => ';', # primo treats the rest of a line as a comment
   'minDigits' => 6, # primo doesn't support "candidates" less than 65,537.
  );

sub createPRIMOcert($$$)
{ # exponent, factor "index", and factor
  my ($e, $n, $f) = @_;
  ! -f("$subdir/$e.$n.in") && ! -s("$subdir/$e.$n.in")
    or die "$subdir/$e.$n.in already exists\n";
  open(my $fh, '>', "$subdir/$e.$n.in")
    or die "cannot create $subdir/$e.$n.in, $!\n";
  print $fh "$primo{'prefix'}$f$primo{'suffix'}\n";
  print $fh "$primo{'commentChar'} M$e factor $n\n"
    if $n =~ m,^[1-9][0-9]*$,;
  print $fh "$primo{'commentChar'} M$e cofactor $n\n"
    unless $n =~ m,^[1-9][0-9]*$,;
  close($fh)
    or die "cannot close $subdir/$e.$n.in, $!\n";
  #!!incomplete - pull code from doPrimo.pl but only creating one certificate file here;
  #!!incomplete - find and rename the cert file;
  return(0); #!!default to failing until code is written
}

sub verifyPRIMOcert($$$$)
{ # exponent, factor "index", factor, and cert file
  my ($e, $n, $f, $certFile) = @_;
  #!!incomplete - pull code from doPrimo.pl but only checking one certificate file here;
  #!!incomplete - find and rename the cert file;
  return(0); #!!default to failing until code is written
}
