#!/usr/bin/perl -Tw
# Do ECM, P-1, P+1, gap closing, and primality proving, alternately, until something goes wrong.
#   Try to give an equal amount of time to each type of work that uses the same resource (CPUs or GPUs).
#   Get work from the server: ECM, P-1, P+1, closing trial factoring gaps, primality proving, etc.
# Also supports trial factoring work from GIMPS at www.mersenne.org on GPUs with mfaktc or mfakto.
# To not do a particular kind of work, create a, e.g., skipecm directory in $topDir.
# To resume doing skipped work, remove the skipecm directory and create a reset directory.
# A reset directory can contain a times file with entries like the $topDir/times file that tracks
#   how many tasks of each type should be queued and how much time has been spent on each work type
#   in excess of the time spent on other work types.

#!!Add a test mode to verify that children are run correctly, create expected save files, etc.
#!!  Include test variants of actual children that simply print output to test code paths here
#!!Print verbose output to a log file instead of STDOUT?
#!!  And rotate it when it becomes large even during a run?

use strict;
use strict 'vars';
use warnings;
use re '/aa'; # regular expressions emphasize ASCII
use Fcntl; # only needed to set non-blocking on file descriptors, but that's essential to the logic
use File::stat;
use Scalar::Util;

use Math::BigInt try => 'GMP';
use Math::BigFloat try => 'GMP';

BEGIN { push(@INC, '.') if -s('Mersenne.pm'); $ENV{'PATH'} = "/usr/bin:/bin"; }

use MersInfo;
use Mersenne;

my ($usage) = ("Usage: doWork.pl [-h][ ][mersServer] [-q[ ]<queueTime>] [-v[verbosity]]\n"
               . "  [--check] [--check-all] [--help] [--version] [--versions]\n"
               . "The 'mersServer' argument is the host where the mers format server\n"
               . "  is running; it is given to the askServer.pl and sendServer.pl scripts.\n"
               . "The 'queueTime' argument is the amount of work of each type to try to\n"
               . "  have on hand; the default is 30 days (30*24*60*60 seconds).\n"
               . "The 'verbosity' value prints information about what is happening;\n"
               . "  higher values (up to 8) print more information.\n"
               . "--check and --check-all check various configuration data and exit\n"
               . "  without creating worker children.  --check-all also runs all the\n"
               . "  scripts this script calls with --check.\n"
  );

my ($startTime) = time();
my ($lastCheckOldFiles) = $startTime;

my (@args);
push(@args, split(/ /, $_)) foreach (@ARGV);
@ARGV = @args;

# When not zero/false, check that work can be done rather than actually doing any work.
my ($checkConfig) = 0;

# The list of scripts that do the work, get more work, send results, etc.
my (@scripts) = ( './askGIMPS.pl', './askServer.pl', './sendServer.pl', './spreadWork.pl',
                  './doECM.pl', './doGPU.pl', './doJoiner.pl', './doP+-1.pl', );

my ($err) = 0;

# This is often used in arithmetic rather than Math::BigInt->new() of a constant like "5".
my ($zero) = Math::BigInt->bzero();

# Defaults in case they are not all mentioned in $topDir/times or to create it.
my ($nextSend) = $zero;
my ($maxTime) = $zero + 1024;
$maxTime->bpow(6); # now 2^60
my ($sendInterval) = $zero + 3600; # increased when connect fails, reduced some when connect succeeds
my ($minInterval) = $sendInterval; # an hour
my ($maxInterval) = $minInterval * 24*7; # a week

#!!See the !! comment just before WriteRenewals().
my ($nextReadWorkFiles) = $sendInterval + time();

# Print version info for all the scripts this uses that are both in the current directory and executable.
$err ||= (0 != system("$_ --version")) foreach (grep { -x "$_" } @scripts);

print($usage)
  if grep { m/^--help$/ } @ARGV;

# Note the plural in the pattern here; --versions is distinct, limited to this script.
exit($err) if @ARGV && grep { m/^--(versions?|help)$/ } @ARGV;

# The server to get non-GIMPS work from by IP address or hostname.
#   The only server implemented presently is mine; email wedgingt@acm.org to use it.
#   But only work from and reported to GIMPS will be sent to GIMPS, for which this
#     server is not needed, so the default is to not use this server.
my ($mersServer);
my ($warned) = 0;

# When not zero, print what is happening.  Higher values will print more information.
#   Partly for debugging this script, but some output is the incoming work and results.
my ($verbose);

# Amount of work to queue for each processor, in seconds.
my ($queueTime);

# UUID to uniquely identify this script on this computer to the mers server.
#   The IP address is not enough when this computer is behind a masq'ing
#   firewall from the mers server.
# The mers server (not GIMPS) has to know this UUID before it will
#   reserve and send work.
my ($uuid);

# This is needed but sometimes for strings including other data, so ...
my ($floatWhole) = MersInfo::patterns('float');
my ($floatPattern) = $floatWhole;
$floatPattern =~ s,^\^,,;
$floatPattern =~ s,\$$,,;

my ($errors) = '';

my (@arguments) = @ARGV;
while (@ARGV)
{
  $verbose = ($1 ne '' ? $1 : 1), shift, next if $ARGV[0] =~ m,^-v([0-9]*),;
  $checkConfig = 1, shift, next if $ARGV[0] =~ m,^--check$,;
  $checkConfig = 2, shift, next if $ARGV[0] =~ m,^--check-all$,;
  $mersServer = shift, next if $ARGV[0] =~ m,^[^-],;
  shift, $queueTime = $zero + shift, next if $ARGV[0] =~ m,^-q$,;
  $queueTime = Math::BigInt->new("$1"), shift, next if $ARGV[0] =~ m,^-q(.*),;
  if ($ARGV[0] =~ m,^-h$,)
    { # -h with an argument specifies the server for mers format tasks
      shift();
      $mersServer = shift();
      next
        if defined($mersServer);
      # -h alone is help
      print($usage);
      exit(0);
    }
  $mersServer = $1, shift, next if $ARGV[0] =~ m,^-h(.+),;
  $errors .= "unknown option '" . shift() . "'\n";
}

# These two scripts are needed for all work types.
$errors .= "cannot execute sendServer.pl, $!\n" unless -x './sendServer.pl';
$errors .= "cannot execute spreadWork.pl, $!\n" unless -x './spreadWork.pl';

die $errors if $errors && !$checkConfig;
warn $errors if $errors;

my ($topDir) = 'working.doWork';
my ($stopDir) = "$topDir/stop";
my ($lockDir) = "$topDir/lock";
my ($locked);

# Check for and deal with old files, modified before $start.
sub CheckOldFiles();
my (@oldFileDirs) = ("$topDir/logs", "$topDir/done", "$topDir/tmp", "$topDir/spread", );

foreach ("$topDir", "$topDir/work", @oldFileDirs)
{
  !-l($_) or die "$_ must be a directory, not a soft link\n";
  -d($_) or mkdir($_) or die "could not create $_ directory, $!\n";
}

!-l($lockDir) or die "$lockDir must be a directory, not a soft link\n";
mkdir($lockDir) or die "could not create lock directory '$lockDir', $!\n";
$locked = 1;
END
{ # Only remove the lock directory if this run created it.
  !$locked || !defined $lockDir || ! -d $lockDir || rmdir($lockDir)
    or warn(localtime() . " could not rmdir $lockDir, $!\n");
}

if (! -t || $checkConfig > 1)
{
  $err ||= (0 != system("$_ --check")) foreach (grep { -x "$_" } @scripts);
}

# We do not want input ...
close(STDIN) or die "could not close STDIN, $!\n";
# ... but perl -w complains if STDIN is opened only for writing.
open(STDIN, '<', '/dev/null') or die "could not open /dev/null for STDIN, $!\n";

$SIG{'HUP'} = "IGNORE";
foreach ('INT', 'TERM')
{
  $SIG{$_} = sub {
    my ($sig) = @_;
    local ($!, $?);
    warn "caught SIG$sig\n";
    mkdir("$stopDir") or warn "in SIG$sig handler could not mkdir $stopDir, $!\n";
  };
}

select STDERR; $| = 1; # flush STDERR after every print
select STDOUT; $| = 1; # flush STDOUT after every print

my ($wait) = 15; # seconds to wait in select in many cases; also used in other places to detect hung children

# The number of CPUs and GPUs available for work.
#   These values can also be set in $topDir/times, which overrides /proc/cpuinfo for $cpus.
my (%processors) = ( 'cpu' => 1, 'gpu' => 0 );
if (-f('/proc/cpuinfo') && open(my $cpuFH, '<', '/proc/cpuinfo'))
{
  while (<$cpuFH>)
    {
      $processors{'cpu'} = 1 + $1 if m/^processor\s*:\s*([0-9]+)/;
    }
  print(STDOUT "$processors{'cpu'} CPUs listed in /proc/cpuinfo\n");
}

# Checked now and then.  If the first is false (0) right after checking, this script will call die.
my ($canDoSomeWork) = 0;
my (%canDoSome);
$canDoSome{$_} = 0 foreach keys %processors;

# Do not start new (CPU) work if the load is higher than this.
# Can be set in the $topDir/times file, which overrides this default.
# Should be about 0.5 plus the number of independent CPUs available for use for work.
# Note that the GPU factorers do not affect the load enough to notice since they (mostly) use a GPU, not any CPU.
#   This is presumed to be true for other types of processors - anything other than CPUs.
my ($loadLimit) = 1.5;

# These are default values for entries in the $workConfig hash, just below.
my (%defaults) = ( 'proc' => 'cpu', 'time' => $zero + 1, 'longest' => $zero + 1,
                   'localTasks' => 10, 'taskLimit' => 5000, 'workAvailable' => 1,
                   'nextAsk' => $zero + 1, 'ask' => './askServer.pl', 'askArgs' => '',
                   'ratio' => $zero + 1, 'server' => 'mers',
                   # This limits the count of local tasks of each work type and
                   #   is saved in and read from the times file.
                   'taskLimit' => 5000,
  );

# Initial configuration of work tasks, using the work type as the key.
#   'name' is the argument to give to askServer.pl or askGIMPS.pl to get more work of the kind.
#   'cmd' is the command line to perform a single task.
#   'cmdList' is a list of possible commands, from worst to best (fastest, better I/O for select() here, etc.).
#   'cmdArgs' contains additional arguments to pass the command and is stored in the times file.
#   'skipDir' can be created to skip new tasks of the work type.
#   'skip' is 0 or 1, 1 indicating no new tasks of the work type should be started.
#   'askArgs' will be added to the askServer.pl or askGIMPS.pl command line.
#   'workAvailable' - did the server have suitable work last time we asked ?
#   'time' is seconds spent on tasks but the smallest 'time' value is subtracted from all time values and
#     the type with the smallest 'time' will be the type of the next task created in most cases.  That
#     causes task types to be done in round-robin fashion.  'time' of CPU types is not affected by 'time'
#     of GPU types and vice versa since CPU usage and GPU usage is (mostly) independent.
#   'longest' is the greatest number of (wall clock, real) seconds used by a single task.
#   'onlyOne' only exists when the command can only be run one at a time, never in parallel with itself.
#     This could be dealt with by the relevant perl scripts by creating a new working directory, but that is
#     not really needed for most of these work types since they take quite a while to do one task.
#     The value is the "pretend" number of that work type in progress, which can be more than one if it is
#     faster than other work types.
#     !!Consider multiplying the ratio of onlyOne types by the number of CPUs.
#   'partial' only exists if the command can be interrupted or otherwise complete only part of a task
#     and would finish that task the next time it is called rather than do new work, so the output needs
#     to be checked here to see whether the work given to it was done or not.
#   'localTasks' is the current number of tasks desired to have locally to avoid having nothing to do.
#   'taskLimit' is the maximum number of tasks to ever have locally (minimum 10).
#   'proc' is either 'cpu' or 'gpu' depending on the processor type the program (mostly) uses.
#     Support for other kinds or processors should be fairly easy (but somewhat involved) to add.
#   'server' is either 'GIMPS', 'mers', or the mers server's IP address or hostname.
#   'ask' is the script to run to get more work.
#   'more' indicates that more work can be created from the current work, which is only done if the
#     server cannot be contacted or isn't giving us as much work as we're doing ...
#     ... or, for trial factoring gaps, the child did not cover the entire gap with a new join.
#   'nextAsk' is the next time() to ask the server for this type of work.
#   'others' is a list of other kinds of work that this client can do.
#   'tuneTime' is the number of seconds a child "should" take; 'tuneOpt' and 'tuneValue' must also be given.
#   'tuneOpt' is the option to give the child process with the 'tuneValue'.
#   'tuneValue' is a number that determines how much the child should do, which is adjusted based on the
#     run time of each child to get future runs closer to 'tuneTime'.
#   'tuneAdjust' is a number to adjust 'tuneValue' by when it is "close" to the desired value.
#   'tuneValueMin' is the minimum value for 'tuneValue'.  When 'tuneValue' would be reduced below this,
#     'tuneTime' is incremented when 'tuneValue' is reset to this.
#!!This conflates the client list with the list of work types, which is wrong for mfaktc now, requiring the
#!!  new 'others' field to deal with that.  Separating them would also allow simpler code when reading from
#!!  spreadWork.pl.
my ($workConfig) = {
  # Elliptic Curve Method - factoring
  'ecm' => # do not create 'more' work if a (new) factor ('C') was found
    { 'name' => 'ECM', 'cmd' => './doECM.pl', 'more' => 'C', 'letter' => 'E',
      'tuneTime' => 600, 'tuneOpt' => '--time', 'tuneValue' => 600, 'tuneAdjust' => 15,
      'tuneValueMin' => 120, # two minutes
    },
  # closes gaps in prior trial factoring effort; this script should support any CPU and close any gap
  'joiner' =>
    { 'name' => 'KJoiner', 'cmdList' => [ './doJoiner.pl' ], 
      'letter' => 'G', 'more' => 'G', # the rest of the gap can be closed even if a factor was found
      'tuneTime' => 300, 'tuneOpt' => '-m', 'tuneValue' => 1024, 'tuneAdjust' => 1024,
      'tuneValueMin' => 120, # two minutes
      'others' => [ 'joinNotGPU', 'gpu', ], },
  # same but only for gpu tasks that mfaktc or mfakto could not do; never requested from the server
  'joinNotGPU' =>
    { 'cmdList' => [ './doJoiner.pl' ],
      'letter' => 'G', 'more' => 'G',
      'tuneTime' => 300, 'tuneOpt' => '-m', 'tuneValue' => 1024, 'tuneAdjust' => 1024,
      'tuneValueMin' => 120, # two minutes
    },
  # uses mfaktc or mfakto trial factorer, either of which requires a GPU but is much faster than mersfacgmp
  'gpu' =>
    { 'name' => 'gpuFactorer', 'cmd' => './doGPU.pl', 'proc' => 'gpu', 'onlyOne' => 0,
      'letter' => 'G',
      'tuneTime' => 3600, 'tuneOpt' => '-m', 'tuneValue' => 1024*1024, 'tuneAdjust' => 1024,
      'tuneValueMin' => 120, # two minutes
      'others' => [ 'joiner', ], },
  # P-1 factoring using Linux's ecm program
  'pminus1' =>
    { 'name' => 'OP-1', 'cmd' => './doP+-1.pl', 'more' => 'C', 'letter' => 'o',
      'tuneTime' => 600, 'tuneOpt' => '-i', 'tuneValue' => 1e6, 'tuneAdjust' => 1024,
      'tuneValueMin' => 1000, # stage 1 bound increment - hardcoded limit in doP+-1.pl
    },
  # P+1 factoring using Linux's ecm program
  'pplus1' =>
    { 'name' => 'QP+1', 'cmd' => './doP+-1.pl', 'more' => 'C', 'letter' => 'q',
      'tuneTime' => 600, 'tuneOpt' => '-i', 'tuneValue' => 1e6, 'tuneAdjust' => 1024,
      'tuneValueMin' => 1000, # stage 1 bound increment - hardcoded limit in doP+-1.pl
    },
  # GIMPS trial factoring using mfaktc or mfakto, requires a GPU; the 'name' of '2' is the GIMPS task type, passed to askGIMPS.pl.
  #   'cmdArgs' should always have an integer value > 0, partly so doGPU.pl will know the work is for GIMPS rather than 'gpu' work.
  'gpuGimpsFactor' =>
    { 'name' => '2', 'cmd' => './doGPU.pl', 'cmdArgs' => '60', 'ask' => './askGIMPS.pl', 'server' => 'GIMPS', 'proc' => 'gpu',
      'letter' => 'G',
      'partial' => 1, 'onlyOne' => 0, 'taskLimit' => 100,
    },
  #!!doProof.pl, doPrimo.pl, doGpuOwl.pl, ...
  #!!gpuowl (.exe) should probably have more than one entry here - PRP, GimpsJoiner, P-1, etc.
  #!!  Note that it has a minimum FFT size and thus a minimum exponent.
  #!!  Also, I don't see support in it for known factors.
 };

CheckOldFiles();
foreach my $kind (keys(%{$workConfig}))
{
  foreach my $key (keys(%defaults))
    {
      next if exists $workConfig->{$kind}{$key};
      next if $key eq 'nextAsk' && !exists $workConfig->{$kind}{'name'};
      $workConfig->{$kind}{$key} = $defaults{$key};
      print(STDOUT "copied default $key to $kind, $workConfig->{$kind}{$key}\n") if defined($verbose) && $verbose > 7;
    }
  $workConfig->{$kind}{'skipDir'} = "$topDir/skip$kind";
  $workConfig->{$kind}{'skip'} = -d($workConfig->{$kind}{'skipDir'}) ? 1 : 0;
  foreach my $dir ('done', 'tmp', )
    {
      $workConfig->{$kind}{$dir} = "$topDir/$dir/$kind";
      -d("$workConfig->{$kind}{$dir}") or mkdir("$workConfig->{$kind}{$dir}") or die "could not create $workConfig->{$kind}{$dir} directory, $!\n";
      my (@files) = ($workConfig->{$kind}{$dir}, );
      my ($i) = 0;
      while ($i <= $#files)
        { # Add the files in any subdir and drop the subdirs.
          ++$i,
            next
            if ! -d $files[$i] || -l $files[$i];
          opendir(my $dh, $files[$i]) or die "could not read directory $files[$i], $!\n";
          my (@entries) = grep { !m,^[.][.]?$, } readdir($dh);
          closedir($dh) or die "could not close directory $files[$i], $!\n";
          foreach my $f (@entries)
            {
              warn("unexpected file name '$f' found in '$files[$i]'"),
                next
                unless $f =~ m,^([-/.A-Za-z0-9]+)$,;
              push(@files, "$files[$i]/$1");
            }
          splice(@files, $i, 1);
        }
      next unless @files;
      print("existing files under $workConfig->{$kind}{$dir} -\n"
            . join("\n  ", @files) . "\n");
      if ($dir eq 'done')
        { # Files in 'done' might contain results from before a crash.
          open(my $sfh, '>>', "$topDir/sending")
            or die "could not append to $topDir/sending, $!\n";
          foreach my $file (@files)
            {
              next unless -s("$file") && -f(_) && ! -l("$file");
              open(my $dfh, '<', "$file")
                or die("could not open $file, $!\n");
              print("copied from $file:\n");
              print(),
                print($sfh $_)
                while (<$dfh>);
              print("done with $file\n");
              close($dfh)
                or die("could not close $file, $!\n");
            }
          close($sfh)
            or die "could not close $topDir/sending, $!\n";
        }
      foreach my $file (@files)
        {
          unlink("$file")
            or print("could not remove $file, $!\n");
        }
    }
  if (!defined($mersServer) && $workConfig->{$kind}{'server'} ne 'GIMPS' && !$workConfig->{$kind}{'skip'})
    {
      warn("no non-GIMPS server was given (with -h), so there is nowhere to ask for non-GIMPS tasks\n") unless $warned;
      $warned = 1;
      $workConfig->{$kind}{'skip'} = 1;
    }
  if ($workConfig->{$kind}{'server'} eq 'GIMPS' && !$workConfig->{$kind}{'skip'}
      && !(-f('.gimps') && -s('.gimps') or -f('gimps.txt') && -s('gimps.txt')))
    {
      warn("cannot ask mersenne.org for $kind work without user and password in a file named '.gimps' or 'gimps.txt'\n");
      $workConfig->{$kind}{'skip'} = 1;
    }
  my ($prog, $args);
  if (!exists($workConfig->{$kind}{'cmd'}))
    {
      warn("Work type $kind has neither a commmand nor a list of possible commands; skipping it\n"),
        $workConfig->{$kind}{'skip'} = 1,
        next
        if (!exists($workConfig->{$kind}{'cmdList'}));
      foreach my $cmd (@{$workConfig->{$kind}{'cmdList'}})
        {
          ($prog, $args) = split(/ /, $cmd, 2);
          $workConfig->{$kind}{'cmd'} = $cmd if -x $prog;
        }
      warn("None of the commands for work type $kind are executable; skipping it\n"),
      $workConfig->{$kind}{'skip'} = 1,
      next
        unless exists $workConfig->{$kind}{'cmd'};
    }
  if (exists($workConfig->{$kind}{'tuneTime'}))
    {
      exists($workConfig->{$kind}{'tuneOpt'})
        or die "kind $kind has 'tuneTime' but not 'tuneOpt'\n";
      exists($workConfig->{$kind}{'tuneAdjust'})
        or die "kind $kind has 'tuneTime' but not 'tuneAdjust'\n";
      $workConfig->{$kind}{'tuneValue'} = Math::BigInt->new("$1"),
        next
        if (exists $workConfig->{$kind}{'tuneValue'}
            && $workConfig->{$kind}{'tuneValue'} =~ m,^([1-9][0-9]*)$,);
      $workConfig->{$kind}{'tuneValue'} = 1024;
      $workConfig->{$kind}{'tuneValueMin'} = 1024
        unless $workConfig->{$kind}{'tuneValueMin'};
    }
  # Without this, an arbitrary number of them could be started:
  $workConfig->{$kind}->{'onlyOne'} = 0 if exists($workConfig->{$kind}{'ignoreCount'});
  ($prog, $args) = split(/ /, $workConfig->{$kind}->{'cmd'}, 2);
  next if -x "$prog";
  -d "$workConfig->{$kind}{'skipDir'}" || mkdir "$workConfig->{$kind}{'skipDir'}"
    or die "could not create $workConfig->{$kind}{'skipDir'}, $!\n";
  $workConfig->{$kind}{'skip'} = 1;
}

foreach (keys(%{$workConfig}))
{
  next if -d "$topDir/work/$_" || mkdir "$topDir/work/$_";
  warn("could not create $topDir/work/$_, $!\n");
  $workConfig->{$_}{'skip'} = 1;
}

foreach (keys(%{$workConfig}))
{
  next if $workConfig->{$_}{'skip'}
    || -x $workConfig->{$_}{'ask'};
  warn("cannot execute $workConfig->{$_}{'ask'} so skipping $_ work\n");
  $workConfig->{$_}{'skip'} = 1;
}

# Return the number of tasks of the given type to ask the server for.
sub AskCount($);

# Something was creating keys in %{$workConfig} with no actual value, so check for such.
# More recently, BigInt values are sometimes becoming NaN; detect, warn, and stop.
my ($badKey);
sub CheckWorkConfig();

# Create more work if sensible to do so.  See also description of 'more' key in %workConfig.
# Needs the kind of task, the just-completed task's input file, and the output from the just-completed task.
sub CreateMoreWork($$@);

# Find the next work of the given type (a key in %workNames).
#   Also updates $tasks{$type} with a hash of all the task files from their priorities.
# If the second argument is true, return quickly rather than trying to read the entire work directory.
sub FindNextWork($$);

# Get the Mersenne exponent from the work file
sub GetFileExp($);

# Get the current system load average.
sub GetLoad();

# Get the mers format lines from the given file.
sub GetMersLines($);

# Save the appropriate workConfig data into the given file.
sub SaveConfig($);

# Start a work task, returing a reference to the file handle to read from.
sub StartTask($$$$$);

# Use fcntl() to mark the file handle non-blocking.
sub Unblock($$);

# Read a file to update the workConfig data.
sub UpdateConfig($$$);

# Put appropriate 'renew <exp>' lines in the first argument (work.doWork/renewals).
#   Free least urgent - or non-numeric priority - task of each type with max count.
sub WriteRenewals($);

UpdateConfig("$topDir/times", defined($verbose), defined($queueTime))
  if -s "$topDir/times";
$verbose = 0 unless defined $verbose;
$queueTime = $zero + 30*24*60*60 unless defined $queueTime;
$uuid = `uuidgen` if !defined $uuid;
chomp($uuid) if defined $uuid;

print(STDOUT "Using");
print(STDOUT " $processors{$_} $_") foreach sort(keys(%processors));
print(STDOUT "\n verbose is $verbose, queue time is $queueTime\n\n");

my ($asking) = ''; # the kind of work we are asking for from the server, if any
my ($askingPID) = 0; # the process ID of a currently running askServer.pl or askGIMPS.pl
my ($askingFH); # Perl file handle for askServer.pl or askGIMPS.pl
my (%askFor); # the count of tasks being requested from the server
my ($askingFile); # temporary file containing the output from askServer.pl or askGIMPS.pl, the tasks the server gave us
my ($sendingPID) = 0; # the process ID of a currently running sendServer.pl (which also sends to GIMPS)
my ($sendQ); # Perl file handle for copying results to $topDir/sending
my ($sendingFH); # Perl file handle for sendServer.pl
my ($fdbitsIn) = ''; # the vec(tor) of bits of open file descriptors to read from
my ($fdbitsOut) = ''; # the vec(tor) of bits of open file descriptors that select() says are ready to read from
my ($lastSendWorked) = 1; # in case we have no work and have to ask for some before we can send any results
my ($lastSentTo) = ''; # 'gimps' or 'mers'; alternate if we have data for both
my ($workJustCompleted) = 0; # count of tasks completed since one was started or select() was called to wait for activity
my (%tasks); # hash by kind of work of hashes by priority of each task of that kind
my (%exponents); # hash by Mersenne exponent to (partial) file path containing the work for the exponent
my (%dropList); # hash listing the exponents (keys) the mers server has indicated are reserved by some other client

# %workInProgress contains info related to each task in progress keyed by the file descriptor.
my (%workInProgress); # = ( 0 => { 'type' => (a key from %{$workConfig}), 'FH' => (open($fh, '...'), $fh), 'work' => (input file name), 'start' => time(), etc. }, );

# Processor type (cpu or gpu, currently) of the last work started.
my ($lastProcStarted) = '';

eval
{ # setpriority() is not always implemented
  setpriority(0, 0, 20);
};

PRIMARY:
while (-d $lockDir)
{
  CheckWorkConfig();
  if (-d "$topDir/reset")
    {
      $warned = 0;
      $nextSend = $zero + time() + 50*(keys(%{$workConfig}) + 2);
      foreach (keys(%{$workConfig}))
        {
          $workConfig->{$_}{'skip'} = -d($workConfig->{$_}{'skipDir'}) ? 1 : 0;
          $workConfig->{$_}{'time'} = $zero + 100 if $workConfig->{$_}{'time'} >= $maxTime && ! $workConfig->{$_}{'skip'};
          $workConfig->{$_}{'longest'} = $workConfig->{$_}{'time'} + 1 if $workConfig->{$_}{'longest'} >= $maxTime;
          $workConfig->{$_}{'nextAsk'} = ($nextSend -= 50) if ! $workConfig->{$_}{'skip'} && exists $workConfig->{$_}{'name'};
          if (!defined($mersServer) && $workConfig->{$_}{'server'} ne 'GIMPS' && !$workConfig->{$_}{'skip'})
            {
              warn("no non-GIMPS server was given (with -h), so there is nowhere to ask for non-GIMPS tasks\n") unless $warned;
              $warned = 1;
              $workConfig->{$_}{'skip'} = 1;
            }
        }
      $nextSend -= 50;
      if (-f ("$topDir/reset/times") && -s("$topDir/reset/times"))
        {
          UpdateConfig("$topDir/reset/times", 0, 0);
          unlink("$topDir/reset/times") or $! == 2 or die "cannot remove $topDir/reset/times, $!\n";
        }
      if (!rmdir("$topDir/reset"))
        {
          warn("ignoring and removing other files in $topDir/reset:\n");
          system("ls -l $topDir/reset && rm -rf $topDir/reset");
        }
    }
  else
    {
      foreach (keys(%{$workConfig}))
        {
          $workConfig->{$_}{'skip'} = 1 if -d "$workConfig->{$_}{'skipDir'}";
          $workConfig->{$_}{'localTasks'} = $workConfig->{$_}{'taskLimit'} if $workConfig->{$_}{'localTasks'} > $workConfig->{$_}{'taskLimit'};
          $workConfig->{$_}{'localTasks'} = 10 if $workConfig->{$_}{'localTasks'} < 10;
        }
    }
  $sendInterval = $minInterval if ($sendInterval < $minInterval);
  $maxTime = $sendInterval*5*keys(%{$workConfig}) if $maxTime < $sendInterval*5*keys(%{$workConfig}); # should never happen
  $sendInterval = $maxInterval if ($sendInterval > $maxInterval);
  $sendInterval = $queueTime/5 if ($sendInterval > $queueTime/5);
  $sendInterval = $minInterval if ($sendInterval < $minInterval);

  my (%minTime);
  CheckWorkConfig();
  $canDoSomeWork = 0;
  $canDoSome{$_} = 0 foreach keys %processors;
  my (@order) = sort { ($workConfig->{$a}{'skip'} <=> $workConfig->{$b}{'skip'})
                         || ($workConfig->{$a}{'time'} <=> $workConfig->{$b}{'time'})
                         || (exists($workConfig->{$a}{'onlyOne'}) <=> exists($workConfig->{$b}{'onlyOne'}))
    } keys(%${workConfig});
  foreach (@order)
    {
      $canDoSomeWork = 1,
        (++$canDoSome{$workConfig->{$_}{'proc'}})
        if ! $workConfig->{$_}{'skip'} && $processors{$workConfig->{$_}{'proc'}} > 0;
      $workConfig->{$_}{'time'} = $maxTime + 1 if $workConfig->{$_}{'skip'};
      $workConfig->{$_}{'time'} -= 60 if !$workConfig->{$_}{'skip'} && $workConfig->{$_}{'time'} > 7*24*60*60;
      $workConfig->{$_}{'time'} = Math::BigInt->new(31*24*60*60) if !$workConfig->{$_}{'skip'} && $workConfig->{$_}{'time'} > 365*24*60*60;
      print(STDOUT "In times loop, $_ time is $workConfig->{$_}{'time'}; skip is $workConfig->{$_}{'skip'}\n") if $verbose > 5 + $workConfig->{$_}{'skip'};
      $minTime{$workConfig->{$_}{'proc'}} = [] if !exists $minTime{$workConfig->{$_}{'proc'}} || ref($minTime{$workConfig->{$_}{'proc'}}) ne 'ARRAY';
      push(@{$minTime{$workConfig->{$_}{'proc'}}}, $workConfig->{$_}{'time'}) if $workConfig->{$_}{'time'} && !$workConfig->{$_}{'skip'};
    }
  warn("cannot do any supported work - all types are being skipped or have no processors\n"),
    last
    if $checkConfig && !$canDoSomeWork;
  $canDoSomeWork or die "cannot do any supported work - all types are being skipped or have no processors\n";
  CheckWorkConfig();
  foreach my $proc (keys(%minTime))
    {
      next if ! $canDoSome{$proc};
      @{$minTime{$proc}} = sort { $a <=> $b } @{$minTime{$proc}}[0 .. $#{$minTime{$proc}}];
      push(@{$minTime{$proc}}, 0) if $#{$minTime{$proc}} < 0; # ensure the array has a value
      push(@{$minTime{$proc}}, $minTime{$proc}[0]) if $#{$minTime{$proc}} == 0; # duplicate a single value
      if (!$checkConfig && $minTime{$proc}[0] < $zero)
        {
          $workConfig->{$_}{'longest'} -= 2 foreach grep { $workConfig->{$_}{'proc'} eq $proc && $workConfig->{$_}{'time'} <= 0; } keys(%{$workConfig});
          $workConfig->{$_}{'time'} -= $minTime{$proc}[0] - 2 foreach grep { $workConfig->{$_}{'proc'} eq $proc; } keys(%{$workConfig});
          print(STDOUT "minTime($proc) was $minTime{$proc}[0]\n") if $verbose > 4 && $processors{$proc} > $zero;
          next PRIMARY;
        }
      print(STDOUT localtime() . " minTime($proc) is now $minTime{$proc}[0]\n") if $verbose;
    }
  SaveConfig("$topDir/times"),
  SaveConfig("$topDir/times.good")
    unless $checkConfig;
  CheckWorkConfig();
  foreach (keys(%{$workConfig}))
    {
      next if $workConfig->{$_}{'skip'} || $checkConfig;
      $workConfig->{$_}{'time'} -= $minTime{$workConfig->{$_}{'proc'}}[0];
    }

  my ($load) = $canDoSome{'cpu'} ? GetLoad() : $loadLimit + $workJustCompleted + 2.0;
  @order = sort { ($workConfig->{$a}{'skip'} <=> $workConfig->{$b}{'skip'})
                    || ($workConfig->{$a}{'time'} <=> $workConfig->{$b}{'time'})
                    || (exists($workConfig->{$a}{'onlyOne'}) <=> exists($workConfig->{$b}{'onlyOne'}))
    } keys(%${workConfig});
  CheckWorkConfig();
  my ($start, $lackOfWork);
  foreach my $kind (@order)
    {
      if ($workConfig->{$kind}{'proc'} eq $lastProcStarted && $workConfig->{$kind}{'longest'} > 60)
        {
          print(STDOUT "skipping $kind on $workConfig->{$kind}{'proc'} with 'longest' $workConfig->{$kind}{'longest'} > 60\n")
            if $verbose > 5;
          next;
        }
      print(STDOUT "$kind time is now $workConfig->{$kind}{'time'}; skip is $workConfig->{$kind}{'skip'}; " . scalar(keys(%workInProgress)) . " in progress; load $load\n") if $verbose > 4;
      last if (-d($stopDir) || $workConfig->{$kind}{'skip'} == 1);
      print(STDOUT "$kind time is now $workConfig->{$kind}{'time'}; " . scalar(keys(%workInProgress)) . " in progress\n") if $verbose == 3 && $workConfig->{$kind}{'time'} < 2*$maxTime;
      CheckWorkConfig();
      # Limit CPU work that is not the lowest time based on the CPU load average.
      if ($workConfig->{$kind}{'proc'} eq 'cpu' && ((!exists($workConfig->{$kind}{'ignoreCount'}) && $kind eq $order[0]) ? 0.0 : 1.0) + $load >= $loadLimit + $workJustCompleted)
        {
          print(STDOUT "skipping $kind based on load and ignoreCount\n")
            if $verbose > 5;
          next;
        }
      print(STDOUT "load or ignoreCount allows more $kind work\n") if $verbose > 4;
      CheckWorkConfig();
      my ($active) = scalar(grep { $workConfig->{$workInProgress{$_}{'type'}}{'proc'} eq $workConfig->{$kind}{'proc'} } keys(%workInProgress));
      next unless $active < $processors{$workConfig->{$kind}{'proc'}};
      print(STDOUT "$active tasks in progress on $processors{$workConfig->{$kind}{'proc'}} $workConfig->{$kind}{'proc'}s allows more $kind work\n") if $verbose > 4;
      CheckWorkConfig();
      if (exists($workConfig->{$kind}{'onlyOne'}) && $workConfig->{$kind}{'onlyOne'} > 0)
        { # Pretend more were started so the times can be updated to find the next work type to do.
          $workConfig->{$kind}{'time'} += $workConfig->{$kind}{'longest'}/2, next if exists($workConfig->{$kind}{'ignoreCount'});
          my ($pretend) = $minTime{$workConfig->{$kind}{'proc'}}[1] - $workConfig->{$kind}{'time'};
          $pretend->bdiv($workConfig->{$kind}{'longest'});
          $pretend->binc();
          $pretend = $zero + 1 if $pretend < 1;
          $workConfig->{$kind}{'time'} += $minTime{$workConfig->{$kind}{'proc'}}[1] + $workConfig->{$kind}{'longest'}*$pretend;
          $workConfig->{$kind}{'onlyOne'} += $pretend;
          print(STDOUT localtime() . " pretending start $kind onlyOne $workConfig->{$kind}{'onlyOne'} pretend $pretend\n") if $verbose > 2;
          CheckWorkConfig();
          next;
        }
      CheckWorkConfig();
      my ($workFile, $pri);
      my (@kinds) = ($kind, );
      push(@kinds, @{$workConfig->{$kind}{'others'}}) if exists $workConfig->{$kind}{'others'};
    SEARCH:
      foreach my $chkKind (@kinds)
        {
          next unless exists($tasks{$chkKind}) && ref($tasks{$chkKind}) eq 'HASH' && scalar(%{$tasks{$chkKind}});
        ENTRIES:
          foreach my $entry (sort { $a <=> $b } keys(%{$tasks{$chkKind}}))
            { # Would not have to sort all the keys if it was a C++ std::map. !!Change from hash to array?
              $workFile = "$tasks{$chkKind}{$entry}";
              print(STDOUT " may start $workFile\n") if $verbose > 4;
              foreach my $worker (keys(%workInProgress))
                { # Don't repeat a task that is in progress.
                  undef $workFile, next ENTRIES if $workInProgress{$worker}{'work'} eq $workFile;
                }
              $pri = "$entry", last SEARCH if -f "$workFile" && -s "$workFile";
              print(STDOUT " tasks list had bad entry $workFile\n") if $verbose > 2;
              delete $tasks{$chkKind}{$entry};
              undef $workFile;
            }
        }
      CheckWorkConfig();
      if (!defined($workFile))
        { # This can take as long as a gpu task on MSWin10/Cygwin, so account for it
          $start = time();
          $workFile = FindNextWork($kind, 1);
          # Wipe $start if there's no task to start so $lastProcStarted is wiped later
          #   unless something else is started.
          $start = undef
            unless defined $workFile;
          $pri = substr($workFile, 1 + rindex($workFile, '/'))
            if defined($workFile) && $workFile =~ m,/,;
        }
      CheckWorkConfig();
      if (!defined($workFile))
        { # We have no work of this type; try not to check again for a while.
          $lackOfWork = $kind;
          $workConfig->{$kind}{'time'} += int(1 + rand($sendInterval->numify()));
          $workConfig->{$kind}{'time'} += $minTime{$workConfig->{$kind}{'proc'}}[1] + (1 + int(rand($sendInterval->numify()))) if $workConfig->{$kind}{'time'} < $minTime{$workConfig->{$kind}{'proc'}}[1];
          next if !exists $workConfig->{$kind}{'name'};
          # Ask for more sooner _unless_ there's a restriction (usually on the bits or exponent range).
          $workConfig->{$kind}{'nextAsk'} -= int(rand($sendInterval->numify()/10))
            if ($lastSendWorked && $workConfig->{$kind}{'nextAsk'} > $nextSend
                && ($workConfig->{$kind}{'askArgs'} eq '' || (exists($askFor{$kind}) && $askFor{$kind} > 120)));
          ++$workConfig->{$kind}{'localTasks'} if ($workConfig->{$kind}{'localTasks'} < $workConfig->{$kind}{'taskLimit'} && ($workConfig->{$kind}{'workAvailable'} || $workConfig->{$kind}{'askArgs'} eq '') && !exists($workConfig->{$kind}{'ignoreCount'}));
          print(STDOUT localtime() . " no $kind work; nextAsk reduced to " . localtime($workConfig->{$kind}{'nextAsk'}->numify()) . "\n") if $verbose > 2;
          next;
        }
      CheckWorkConfig();
      my ($exp) = GetFileExp($workFile);
      delete $tasks{$kind}{$pri},
        next
        if !defined($exp);
      # Disable this for now; it should check whether it is work in progress first
      if (exists($dropList{$exp}) && 0)
        { # Should have already been dealt with, but easy to check.
          delete $tasks{$kind}{$pri};
          delete $exponents{$exp};
          delete $dropList{$exp};
          unlink("$workFile") or $! == 2 or die "could not remove $workFile, to-be-dropped for $exp, $!\n";
          next;
        }
      CheckWorkConfig();
      $lackOfWork = undef;
      my ($workExtra) = '';
      $workExtra .= " $workConfig->{$kind}{'cmdArgs'}"
        if exists $workConfig->{$kind}{'cmdArgs'};
      $workExtra .= " $workConfig->{$kind}{'tuneOpt'} $workConfig->{$kind}{'tuneValue'}"
        if exists $workConfig->{$kind}{'tuneOpt'} && exists $workConfig->{$kind}{'tuneValue'};
      $workExtra .= " --tmp=$workConfig->{$kind}{'tmp'}"
        if exists $workConfig->{$kind}{'tmp'};
      $workExtra .= " --skip=$workConfig->{$kind}{'skipDir'}"
        if exists $workConfig->{$kind}{'skipDir'};
      $workExtra .= " --stop=$stopDir";
      print(STDOUT "checkConfig: would start $workConfig->{$kind}{'cmd'} $workExtra\n"),
        next
        if $checkConfig;
      $start = time()
        unless defined $start;
      my ($taint);
      print(STDOUT "kind is tainted\n"),
        $taint = 1
        if Scalar::Util::tainted($kind);
      print(STDOUT "cmd is tainted\n"),
        $taint = 1
        if Scalar::Util::tainted($workConfig->{$kind}{'cmd'});
      print(STDOUT "workExtra is tainted\n"),
        $taint = 1
        if Scalar::Util::tainted($workExtra);
      print(STDOUT "workFile is tainted\n"),
        $taint = 1
        if Scalar::Util::tainted($workFile);
      print(STDOUT "pri is tainted\n"),
        $taint = 1
        if Scalar::Util::tainted($pri);
      next
        if $taint;
      my ($fh, $outfile, $errfile) = StartTask($kind, $workConfig->{$kind}{'cmd'}, $workExtra, $workFile, $pri);
      $workInProgress{'new'} = { 'FH' => $fh, 'output' => $outfile, 'errout' => $errfile,
                                 'type' => $kind, 'work' => $workFile, 'priority' => $pri,
                                 'exp' => $exp, 'start' => $start, };
      # Pretend it will take the longest to tend to start other kinds of work next.
      $workConfig->{$kind}{'time'} += $workConfig->{$kind}{'longest'}*$workConfig->{$kind}{'ratio'};
      if (!defined($workInProgress{'new'}{'FH'}))
        {
          $start = undef;
          print(STDOUT "$0 could not start $kind work\n") if $verbose > 2;
          delete($workInProgress{'new'});
          next;
        }
      my ($fd) = fileno($workInProgress{'new'}{'FH'});
      $fd >= 0 or die "fileno() returned $fd for pipe from new work\n";
      $workInProgress{$fd} = $workInProgress{'new'};
      vec($fdbitsIn, $fd, 1) = 1;
      delete $workInProgress{'new'};
      warn("workConfig $kind does not exist but should have just been started\n"),
      next
        if !exists($workConfig->{$kind});
      CheckWorkConfig();
      $workConfig->{$kind}{'onlyOne'} = 1 if exists $workConfig->{$kind}{'onlyOne'};
      ++$load if $workConfig->{$kind}{'proc'} eq 'cpu';
      print(STDOUT localtime() . " started $workConfig->{$kind}{'cmd'}$workExtra $workFile " . (exists($workConfig->{$kind}{'onlyOne'}) ? "onlyOne $workConfig->{$kind}{'onlyOne'}" : "") . "\n") if $verbose > 2;
      $lastProcStarted = $workConfig->{$kind}{'proc'};
      # Do not start work of two different types without checking the time values again
      # ... unless the longest time taken is very short.
      last unless $workConfig->{$kind}{'longest'} <= $wait;
    }
  last if $checkConfig;

  $lastProcStarted = '' unless defined $start;
  my ($time) = time();
  CheckWorkConfig();
  @order = ();
  @order = grep { ! $workConfig->{$_}{'skip'} && exists($workConfig->{$_}{'nextAsk'}) && $workConfig->{$_}{'nextAsk'} < $time } keys(%{$workConfig})
    unless -d("$stopDir") || $askingPID != 0 || ! $lastSendWorked;
  foreach (@order)
    {
      $askFor{$_} = AskCount($_);
      $askFor{$_} = 127 if $askFor{$_} > 127;
      $workConfig->{$_}{'nextAsk'} = $zero + $time + 120 + int(rand($askFor{$_} < 127 ? int(rand($sendInterval->numify())) : 60));
      $workConfig->{$_}{'nextAsk'} += int(rand($sendInterval->numify())) if $askFor{$_} < 1;
      $workConfig->{$_}{'nextAsk'} += $maxTime if -d $workConfig->{$_}{'skipDir'} && (!exists($tasks{$_}) || ref($tasks{$_}) ne 'HASH' || !scalar(%{$tasks{$_}}));
      print(STDOUT localtime($time) . " asking $workConfig->{$_}{'server'} for $askFor{$_} $_ tasks of $workConfig->{$_}{'localTasks'} desired; next check at " . localtime($workConfig->{$_}{'nextAsk'}->numify()) . "\n") if $verbose > 2;
      next unless $askFor{$_} > 0;
      $askingFile = "$topDir/asking.$_.$$";
      CheckWorkConfig();
      if ($workConfig->{$_}{'server'} ne 'GIMPS')
        { # Ask $mersServer for these kinds of work.
          if (!defined($mersServer))
            {
              warn("no non-GIMPS server was given (with -h), so there is nowhere to ask for non-GIMPS tasks\n") unless $warned;
              $warned = 1;
              next;
            }
          my (@args) = ("$workConfig->{$_}{'ask'}", split(' ', "$workConfig->{$_}{'askArgs'}"), '-h', "$mersServer");
          push(@args, '-u', "$uuid") if $uuid;
          push(@args, '-o', "$askingFile", "$workConfig->{$_}{'name'}", "$askFor{$_}");
          $askingPID = open($askingFH, '-|', @args);
          if ($askingPID > 0)
            {
              $asking = $_;
              Unblock($askingFH, "$workConfig->{$_}{'ask'} $uuid $askFor{$_}");
              vec($fdbitsIn, fileno($askingFH), 1) = 1;
              last;
            }
          warn("could not ask $mersServer for $workConfig->{$_}{'name'} work, $!\n");
          $askingPID = 0;
          next;
        }
      CheckWorkConfig();
      # Ask GIMPS (mersnne.org) rather than $mersServer.
      $workConfig->{$_}{'nextAsk'} += $maxTime,
      warn("cannot ask mersenne.org for $_ work without user and password in file named '.gimps' or 'gimps.txt'\n"),
        next
        unless -f('.gimps') && -s('.gimps') or -f('gimps.txt') && -s('gimps.txt');
      $workConfig->{$_}{'nextAsk'} += $maxTime,
      warn("unknown GIMPS work type for $_ work; cannot ask for more\n"),
        next
        unless exists($workConfig->{$_}{'name'}) && $workConfig->{$_}{'name'} =~ m/^[0-9]+$/;
      my ($fn) = -s('.gimps') ? '.gimps' : 'gimps.txt';
      $askingPID = open($askingFH, '-|', "$workConfig->{$_}{'ask'}", "-f$fn", "-c1",
                        "-o$askingFile", "-t$askFor{$_}", "-w$workConfig->{$_}{'name'}");
      if ($askingPID > 0)
        {
          $asking = $_;
          Unblock($askingFH, "$workConfig->{$_}{'ask'} GIMPS $askFor{$_}");
          vec($fdbitsIn, fileno($askingFH), 1) = 1;
          last;
        }
      warn("could not ask mersenne.org for $_ work, $!\n");
      $askingPID = 0;
      next;
    }

  CheckWorkConfig();
  #!!This may be reducing $nextSend too quickly but $lackOfWork should be rare unless the server is down or unreachable
  $nextSend -= $wait if $lackOfWork;
  if ($sendingPID == 0 && $nextSend < ($time = time()))
    {
      ! -s("$topDir/queue") || -s("$topDir/sending")
        || rename("$topDir/queue", "$topDir/sending")
        or die "could not rename $topDir/queue to $topDir/sending, $!\n";
      if (-s("$topDir/queue"))
        { # sending must already have something in it; append newer results from queue
          open($sendQ, '<', "$topDir/queue") or die "could not read $topDir/queue, $!\n";
          open($sendingFH, '>>', "$topDir/sending") or die "could not append to $topDir/sending, $!\n";
          while (<$sendQ>)
            {
              chomp;
              warn("$topDir/queue:$." . ': contained non-printable characters')
                if s,[[:^print:]],,g;
              print($sendingFH "$_\n");
              print(STDOUT " queue to sending $_\n") if $verbose > 5;
            }
          close($sendingFH) or die "could not close sending, $!\n";
          close($sendQ) or die "could not close queue, $!\n";
          unlink("$topDir/queue") or $! == 2 or die "could not unlink queue, $!\n";
        }
      ! -s("$topDir/queueGimps") || -s("$topDir/sendingGimps")
        || rename("$topDir/queueGimps", "$topDir/sendingGimps")
        or die "could not rename $topDir/queueGimps to $topDir/sendingGimps, $!\n";
      if (-s("$topDir/queueGimps"))
        {
          open($sendQ, '<', "$topDir/queueGimps") or die "could not read $topDir/queueGimps, $!\n";
          open($sendingFH, '>>', "$topDir/sendingGimps") or die "could not append to $topDir/sendingGimps, $!\n";
          while (<$sendQ>)
            {
              print($sendingFH $_);
              print(STDOUT " queueGimps to sendingGimps $_") if $verbose > 5;
            }
          close($sendingFH) or die "could not close sendingGimps, $!\n";
          close($sendQ) or die "could not close queueGimps, $!\n";
          unlink("$topDir/queueGimps") or $! == 2 or die "could not unlink queueGimps, $!\n";
        }
      $nextSend = $sendInterval + $time; # do not try again sooner no matter what happens
      my (@toSend);
      # Only send to GIMPS or to the mers server; don't send to both at once.
      #   Otherwise, if sending to one is failing, the other data will be sent repeatedly.
      if (-s("$topDir/sendingGimps") && ($lastSentTo ne 'gimps' || !defined($mersServer)))
        {
          $lastSentTo = 'gimps';
          push(@toSend, '-g', '1', "$topDir/sendingGimps");
        }
      else
        {
          $lastSentTo = 'mers';
          WriteRenewals("$topDir/renewals");
          push(@toSend, '-h', "$mersServer") if -s("$topDir/renewals") || -s("$topDir/sending");
          push(@toSend, '-u', "$uuid") if $uuid;
          push(@toSend, "$topDir/renewals") if -s("$topDir/renewals");
          push(@toSend, "$topDir/sending") if -s("$topDir/sending");
        }
      if (@toSend)
        {
          $sendingPID = open($sendingFH, '-|', './sendServer.pl', @toSend);
          if (defined($sendingPID) && $sendingPID > 0)
            {
              Unblock($sendingFH, "sendServer.pl " . join(' ', @toSend));
              vec($fdbitsIn, fileno($sendingFH), 1) = 1;
              print(STDOUT localtime($time) . " sending to " . ($lastSentTo ne 'gimps' ? "$mersServer" : 'mersenne.org') . "\n") if $verbose > 2;
            }
          else
            {
              warn(localtime($time) . " could not start sendServer.pl to " . ($lastSentTo ne 'gimps' ? "$mersServer" : "mersenne.org" ) . ", $!\n");
              $sendingPID = $lastSendWorked = 0;
            }
        }
    }

  if ((scalar(keys(%workInProgress)) . $askingPID . $sendingPID) eq '000' || $lackOfWork)
    { # Have to pause (in the following select()) to avoid a busy loop when no work is being done because the load is too high.
      print(STDOUT localtime() . " nothing executing; minTime(cpu) is $minTime{'cpu'}[0]\n") if $verbose > 2 && $processors{'cpu'} > 0 && !$lackOfWork;
      print(STDOUT localtime() . " lack of $lackOfWork work; minTime($workConfig->{$lackOfWork}{'proc'}) is $minTime{$workConfig->{$lackOfWork}{'proc'}}[0]\n") if $verbose > 2 && $lackOfWork;
      $workJustCompleted = 0,
        last
        if -d $stopDir;
    }

  # Wait until a server should be contacted unless ...
  my ($next) = $nextSend;
  CheckWorkConfig();
  foreach my $key (grep { exists $workConfig->{$_}{'name'} } keys(%{$workConfig}))
    {
      $next = $workConfig->{$key}{'nextAsk'} if $next > $workConfig->{$key}{'nextAsk'};
    }
  $time = time();
  $next -= $time;
  # ... nothing is in progress or more could be started once the load is low enough or ...
  $next = $wait
    if $next < $wait
    || (! -d($stopDir)
        && (!scalar(%workInProgress)
            || 1.0 + scalar(grep { $workConfig->{$workInProgress{$_}{'type'}}{'proc'} eq 'cpu'; } keys(%workInProgress)) < $loadLimit
            || scalar(grep { $workConfig->{$workInProgress{$_}{'type'}}{'proc'} eq 'gpu'; } keys(%workInProgress)) < $processors{'gpu'}));
  # ... wait only a second if the load is low enough to start another task if there is (likely) work to start.
  $next = 1 if ! -d($stopDir) && $next <= $wait && (1.0 + $load < $loadLimit + $workJustCompleted) && !$lackOfWork;
  CheckOldFiles(),
    $next -= $wait
    if $next > 2*$wait && $time > $lastCheckOldFiles + 5*60;
  print(STDOUT localtime($time) . " calling select for $next seconds, load $load, completed $workJustCompleted\n") if $verbose > 1;
  my ($ready) = select($fdbitsOut = $fdbitsIn, undef, undef, $next);
  $time = time();
  print(STDOUT localtime($time) . " select returned with " . $ready . " fds active" . (($ready < 0) ? ", errno $!" : '') . "\n") if $verbose > 1 || $ready < 0;
  $workJustCompleted = 0;
  if ($ready < 1)
    {
      if ($askingPID != 0 && $workConfig->{$asking}{'nextAsk'} + 5*($askFor{$asking} + $minInterval) < $time)
        { # Request to server stalled.
          vec($fdbitsIn, fileno($askingFH), 1) = 0;
          kill('TERM', $askingPID) or kill('KILL', $askingPID) or kill(9, $askingPID) or warn(localtime() . " could not kill asking $askingPID, $!\n");
          close($askingFH) or warn(localtime() . " could not close asking $asking, $!\n");
          $workConfig->{$asking}{'nextAsk'} = $zero + $time + 50 + 10*$workConfig->{$asking}{'time'};
          $asking = '';
          $askingPID = 0;
          next;
        }
      if ($sendingPID != 0 && $nextSend + $minInterval < $time)
        { # Sending to server stalled.
          vec($fdbitsIn, fileno($sendingFH), 1) = 0;
          kill('TERM', $sendingPID) or kill('KILL', $sendingPID) or kill(9, $sendingPID) or warn(localtime() . " could not kill sending $sendingPID, $!\n");
          close($sendingFH) or warn(localtime() . " could not close sending $sendingPID, $!\n");
          $sendingPID = 0;
          warn(localtime() . " sending to the server stalled; increasing interval\n");
          $sendInterval *= 102 + int(rand(5));
          $sendInterval /= 100;
          $nextSend = $sendInterval + $time;
          $lastSendWorked = 0;
          next;
        }
    }

  CheckWorkConfig();
  if ($sendingPID != 0 && vec($fdbitsOut, fileno($sendingFH), 1) == 1)
    { # Sending to the server is done.
      --$ready;
      print(STDOUT localtime() . " SENDING ready\n") if $verbose > 1;
      while (<$sendingFH>)
        {
          print(STDOUT "sending stdout/stderr, $_");
          next unless m/^drop ([1-9][0-9]*)$/;
          # The server says work for exponent $1 is no longer relevant or is reserved by some other computer; drop it.
          $dropList{$1} = 1, next if !exists($exponents{$1});
          # Should always get to here - the server only replies this way if we asked to renew an existing reservation.
          #   But $mersServer may not know about GIMPS reservations.
          #!!This dependency on the file path matching this pattern should be avoided.
          next if $exponents{$1} =~ m,gimps,i;
          ! -f("$topDir/work/$exponents{$1}") or unlink("$topDir/work/$exponents{$1}") or $! == 2
            or die "could not remove $topDir/work/$exponents{$1}, to-be-dropped for $1, $!\n";
          delete $exponents{$1};
        }
      my ($retryErr) = 1 if $!{'EWOULDBLOCK'} || $!{'EAGAIN'} || $!{'EINTR'};
      next if !defined($_) && $retryErr;
      vec($fdbitsIn, fileno($sendingFH), 1) = 0;
      if (close($sendingFH))
        {
          $lastSentTo eq 'gimps' || ! -f("$topDir/sending") or unlink("$topDir/sending") or $! == 2 or die "could not unlink sending, $!\n";
          $lastSentTo ne 'gimps' || ! -f("$topDir/sendingGimps") or unlink("$topDir/sendingGimps") or $! == 2 or die "could not unlink sendingGimps, $!\n";
          $sendInterval /= 2 if ($sendInterval > $minInterval && $lastSendWorked);
          $lastSendWorked = 1;
        }
      else
        {
          warn(localtime() . " could not close sending stdout/err, $!\n");
          $sendInterval *= 103 + int(rand(7));
          $sendInterval /= 100;
          $lastSendWorked = 0;
        }
      !kill('ZERO', $sendingPID) or kill('TERM', $sendingPID) or kill('KILL', $sendingPID) or kill(9, $sendingPID) or warn(localtime() . " could not kill sending $sendingPID, $!\n");
      $sendingPID = 0;
    }

  CheckWorkConfig();
  if ($askingPID != 0 && vec($fdbitsOut, fileno($askingFH), 1) == 1)
    { # Asking for work from the server is done.
      --$ready;
      print(STDOUT localtime() . " ASKING $asking ready\n") if $verbose > 1;
      my ($missing) = 0;
      while (<$askingFH>)
        { # Do not try again for a while since something went wrong.
          chomp;
          s/[\r\n]+//g;
          $workConfig->{$asking}{'nextAsk'} = $zero + time() + 130 + int(rand(50));
          $workConfig->{$asking}{'workAvailable'} = 0;
          next if m/^ *$/ || m/^server exiting$/ || m/^no work from the server/ || m/^the server has no suitable work$/;
          # But try sooner if the server simply didn't have _some_ of the work requested
          $missing = $1, $workConfig->{$asking}{'nextAsk'} -= 80 if m/^Could not get all work requested; ([1-9][0-9]*) missing$/ && $1 < $askFor{$asking};
          print(STDOUT "$workConfig->{$asking}{'ask'} $asking replied $_\n") if $verbose > 1;
        }
      my ($retryErr) = 1 if $!{'EWOULDBLOCK'} || $!{'EAGAIN'} || $!{'EINTR'};
      next if !defined($_) && $retryErr;
      vec($fdbitsIn, fileno($askingFH), 1) = 0;
      if (!close($askingFH) || $? != 0)
        {
          warn(localtime() . " could not close $workConfig->{$asking}{'ask'} $asking" . (($! == 0) ? "" : ", $!") . "\n");
          $workConfig->{$asking}{'nextAsk'} = $nextSend + $sendInterval/10 if ! -s $askingFile || $missing == $askFor{$asking};
        }
      if (! -s $askingFile || $missing == $askFor{$asking})
        { # Something very wrong; don't try again for a while.
          $workConfig->{$asking}{'nextAsk'} = $nextSend + 250 + int(rand(50));
          $workConfig->{$asking}{'nextAsk'} += 1200 if $!{'ECONNABORTED'} || $!{'ECONNREFUSED'} || $!{'ECONNRESET'};
          warn(localtime() . " $workConfig->{$asking}{'ask'} $asking created empty file" . (($! == 0) ? "" : ", $!") . "\n");
        }
      if (-s $askingFile)
        {
          my ($spreadFH);
          open($spreadFH, '-|', './spreadWork.pl ' . "$askingFile") or die "could not start spreadWork.pl, $!\n";
          while (<$spreadFH>)
            { # The spreadWork.pl script can't tell whether mersfmt 'G:' work should be joiner or gpu; deal with it here.
              #!!'g:' work should be in one list that both joiner and gpu workers draw from
              #!!   - but gpu workers are faster than joiner, so joiner should only be used if gpu can't be or there's
              #!!   spare CPU time for joiner.  See also the 'others' entries in $workConfig.
              chomp;
              my ($kind, $exp, $pri, $junk) = split;
              print(STDOUT "$0 $_\n"), next if $kind =~ m,spreadWork,; # error or complaint or version
              print(STDOUT "$0 spreadWork missing priority: $_\n"), next if !defined($pri);
              print(STDOUT "$0 spreadWork invalid priority: $_\n"), next if $pri !~ m/$floatWhole/;
              $pri = $1;
              print(STDOUT "$0 spreadWork invalid exponent: $_\n"), next unless $exp =~ m,^([1-9][0-9]+)$,;
              $exp = $1;
              print(STDOUT "$0 spreadWork unexpected characters in kind: $_\n"), next unless $kind =~ m,^([A-Za-z0-9]+)$,;
              $kind = $1;
              print(STDOUT "$0 spreadWork unknown kind: $_\n"), next unless exists($workConfig->{$kind});
              rename("$topDir/work/$kind/$pri", "$topDir/work/$asking/$pri")
                ? $kind = $asking
                : warn(localtime() . " $0 spreadWork for $asking spread $kind but $topDir/work/$kind/$pri could not be renamed, $!\n")
                if ($kind eq 'joiner' && $asking eq 'gpu' && -s("$topDir/work/$kind/$pri"));
              $tasks{$kind}{"$pri"} = "$topDir/work/$kind/$pri",
                $workConfig->{$kind}{'workAvailable'} = 1
                if !$junk;
              $exponents{$exp} = "$kind/$pri",
                print(STDOUT " spreadWork, $_\n"),
                next
                unless exists($exponents{$exp});
              # It seems GIMPS allows two Factor= tasks for the same exponent at the same time, or nearly so.
              #   The mers server should never allow two tasks on the same exponent at any time.
              #   !!... but somehow does; at least $exponents{$exp} can exist here when the server sends work for that exponent.
              my ($rd, $ap);
              warn(localtime() . " $0 spreadWork for $asking spread $kind but $tasks{$kind}{$pri} could not be read\n"),
              delete $tasks{$kind}{"$pri"},
                next
                unless open($rd, '<', $tasks{$kind}{"$pri"});
              print(STDOUT " exponent $exp work in $kind/$pri could not be appended to from new spreadWork $tasks{$kind}{$pri}\n"),
                $exponents{$exp} = "$kind/$pri",
                next
                unless open($ap, '>>', "$topDir/work/$exponents{$exp}");
              print($ap $_), print(STDOUT " spreadWork exponent $exp $_") while (<$rd>);
              close($rd) or warn("could not close $tasks{$kind}{$pri}, appending it to $topDir/work/$exponents{$exp}\n");
              close($ap) or warn("could not close $topDir/work/$exponents{$exp} after appending $tasks{$kind}{$pri}\n");
              delete($tasks{$kind}{"$pri"}),
                next
                if unlink($tasks{$kind}{"$pri"});
              warn("could not unlink $tasks{$kind}{$pri} after appending it to $topDir/work/$exponents{$exp}\n");
            }
          close($spreadFH) or warn("could not close spreadWork.pl, errno $? exited $!\n");
        }
      ! -e($askingFile) or unlink("$askingFile") or $! == 2 or warn("could not remove $asking task file, $!");
      ! -e($askingFile . '2') || unlink($askingFile . '2') or warn("could not remove $askingFile" . "2, $!");
      ! kill('ZERO', $askingPID) or kill('TERM', $askingPID) or kill('KILL', $askingPID) or kill(9, $askingPID) or warn(localtime() . " could not kill asking $askingPID, $!\n");
      $askingPID = 0;
      $asking = '';
    }

  CheckWorkConfig();
  # Check the work in progress.
  print(STDOUT localtime() . " some work is done\n") if $ready > 0 && $verbose > 2;
  foreach my $fd (keys(%workInProgress))
    {
      next unless vec($fdbitsOut, $fd, 1);
      my ($working) = $workInProgress{$fd}{'type'};
      ++$workJustCompleted if exists($workConfig->{$working}) && exists($workConfig->{$working}{'proc'}) && $workConfig->{$working}{'proc'} eq 'cpu';
      my ($fh) = $workInProgress{$fd}{'FH'};
      while (<$fh>)
        { # This should be rare; the child's STDOUT and STDERR are redirected to (distinct) files.
          print(STDOUT "working $working $fd read $_");
        }
      my ($retryErr) = 1 if $!{'EWOULDBLOCK'} || $!{'EAGAIN'} || $!{'EINTR'};
      next if !defined($_) && $retryErr;
      vec($fdbitsIn, $fd, 1) = 0;
      --$ready;
      if (!close($workInProgress{$fd}{'FH'}))
        {
          warn(localtime() . " could not close work $working $fd $workInProgress{$fd}{'work'}, $!\n");
          if (!exists($workConfig->{$working}))
            {
              warn("working type $working unknown");
            }
          else
            { # Only skip this work if not stopping; user may kill worker early after creating the stop directory.
              $workConfig->{$working}{'skip'} = 1 unless -d $stopDir;
            }
        }
      my ($took) = $zero + time() - $workInProgress{$fd}{'start'};
      my ($trials);
      if (-s("$workInProgress{$fd}{'output'}"))
        {
          my ($doneFH, $foundExp);
          open($sendQ, '>>', "$topDir/queue") or die "could not append to $topDir/queue, $!\n";
          open($doneFH, '<', "$workInProgress{$fd}{'output'}") or die "could not read done $working $fd $workInProgress{$fd}{'output'}, $!\n";
          my ($line) = '';
          my ($gimpsResults) = '';
          my (@mersResults); # only needed for call to CreateMoreWork()
          my ($work) = ''; #!!hack: if doGPU.pl says the GPU factoring program can't do the task, give it to doJoiner.pl
          my ($saveFile);
          my ($sFcontent) = '';
          while (<$doneFH>)
            {
              chomp;
              # If the output includes a save file, the content could match blank lines, etc.
              $saveFile = "$1",
                next
                if (m/^save file begin (ecm[.]([1-9][0-9]+)[.]p[mp]1([.]M)?)$/);
              if (m/^save file end ([a-z][-._a-zA-Z0-9]+)$/)
                {
                  print(STDOUT " end of save file '$1' but no prior 'save file begin'\n"),
                    next
                    unless defined $saveFile;
                  print(STDOUT "save file end gives different file name, '$1', not expected '$saveFile'\n"),
                    $saveFile = undef,
                    next
                    unless $1 eq $saveFile;
                  print(STDOUT "save file '$1' has unexpected content\n"),
                    $saveFile = undef,
                    next
                    unless $sFcontent =~ m/^([-+=;:.@ A-Za-z0-9\n]+)$/;
                  chomp $sFcontent;
                  print($sendQ "save file begin $saveFile\n$sFcontent\nsave file end $saveFile\n");
                  $saveFile = undef;
                  $sFcontent = '';
                  next;
                }
              $sFcontent .= "$_\n", next if defined $saveFile;
              next if m/^ *$/;
              my ($ch) = chop;
              $line .= $_;
              next if $ch eq "\\";
              $line .= $ch;
              # Save files can have (nearly) arbitrary content but all other content should be printable ASCII.
              warn("$workInProgress{$fd}{'output'}" . ":$." . ': contained non-printable characters')
                if $line =~ s,[[:^print:]],,g;
              my ($e);
              if ($line =~ m/^M[(] ([1-9][0-9]*) [)](.)/)
                { # cannot require more without first looking for DdLMSP
                  $e = $1;
                  my ($l) = $2;
                  push(@mersResults, "$line");
                  $line = '', next if $l =~ m/^g$/i;
                  delete $exponents{$e};
                  print($sendQ "$line\n");
                  $foundExp = 1 if $workInProgress{$fd}{'exp'} == $e;
                  if ($line =~ m/^M[(] [1-9][0-9]* [)]J: ([0-9]+) ([1-9][0-9]+)/)
                    {
                      $trials = Math::BigInt->new("$2") - Math::BigInt->new("$1");
                      $trials /= 2*$e;
                      $trials = undef unless $trials > 0;
                    }
                }
              elsif ($line =~ m/^checkpoint file ([.a-zA-Z0-9]+)/)
                { # The script left a checkpoint file; retain it and the task for later.
                  $workInProgress{$fd}{'checkpoint'} = $1;
                }
              print(STDOUT "done $working, $line\n")
                if $verbose > 1 || !defined($e);
              $gimpsResults .= "$line\n"
                if !defined($e) && m, for M[1-9][0-9]+ , && $workConfig->{$working}{'server'} =~ m/^GIMPS$/i;
              $gimpsResults .= "$line\n"
                if !defined($e) && m,M[1-9][0-9]+ has a factor: [1-9][0-9]+, && $workConfig->{$working}{'server'} =~ m/^GIMPS$/i;
              $line = '';
            }
          if (length($gimpsResults))
            {
              my ($gimpsQ);
              open($gimpsQ, '>>', "$topDir/queueGimps") or die "could not append to $topDir/queueGimps, $!\n";
              print($gimpsQ $gimpsResults);
              close($gimpsQ) or die "could not close $topDir/queueGimps, $!\n";
            }
          close($doneFH) or die "could not close done $working, $!\n";
          close($sendQ) or die "could not close queue $working, $!\n";
          exists($workInProgress{$fd}{'work'}) or die "logic bug: workInProgress $fd 'work' not present but work $working just performed\n";
          # Do not create more work unless the work can be created - trial factoring gaps, e.g., can't be created
          # ... but could be split or reduced by the actual work done.
          CreateMoreWork($working, $workInProgress{$fd}{'work'}, @mersResults)
            if $workConfig->{$working}{'more'};
          if (($foundExp || !exists($workConfig->{$working}{'partial'}))
              && !exists($workInProgress{$fd}{'checkpoint'}))
            {
              ! -f($workInProgress{$fd}{'work'})
                or unlink($workInProgress{$fd}{'work'}) or $! == 2
                or die "could not unlink completed $workInProgress{$fd}{'work'}, $!\n";
              delete($tasks{$working}{"$workInProgress{$fd}{'priority'}"});
            }
        }
      else
        {
          my (@lines) = GetMersLines($workInProgress{$fd}{'work'});
          die("logic problem - GetMersLines() returned scalar, not array\n")
            if $lines[0] =~ m,^[0-9]+$,;
          warn(localtime() . " empty output $workInProgress{$fd}{'output'} from trying to perform $workInProgress{$fd}{'work'}, $!\n  "
               . join("\n  ", @lines) . "\n");
          $workConfig->{$working}{'skip'} = 1 if exists $workConfig->{$working};
        }
      if (-s("$workInProgress{$fd}{'errout'}"))
        {
          my ($errFH);
          open($errFH, '<', "$workInProgress{$fd}{'errout'}")
            or die "could not open $workInProgress{$fd}{'errout'} from child on fd $fd\n";
          print(STDOUT " fd $fd error - '$_'")
            while (<$errFH>);
          close($errFH)
            or die "could not close $workInProgress{$fd}{'errout'} from child on fd $fd\n";
        }
      if (!exists($workConfig->{$working}))
        {
          warn("working type $working unknown");
        }
      else
        {
          $took = 10 if $took <= 10;
          $took *= $workConfig->{$working}{'ratio'};
          $took = 10 if $took <= 10;
          $workConfig->{$working}{'time'} -= $workConfig->{$working}{'longest'}*$workConfig->{$working}{'onlyOne'}, $workConfig->{$working}{'onlyOne'} = 0 if exists($workConfig->{$working}{'onlyOne'});
          $workConfig->{$working}{'time'} -= $workConfig->{$working}{'longest'} unless exists($workConfig->{$working}{'onlyOne'});
          # Account for the actual time the task took rather than the 'predicted' time used to encourage other kinds of work.
          $workConfig->{$working}{'time'} += $took;
          $workConfig->{$working}{'time'} = $zero + 10*$took if $workConfig->{$working}{'time'} >= $maxTime;
          $workConfig->{$working}{'longest'} = $took if ($workConfig->{$working}{'longest'} < $took);
          # Only adjust $localTasks{$kind} if we have performed at least one task and only based on longest task of this type.
          #!!Ought to be based on (rolling) average of durations, saved from run to run like %{$workConfig}{$working}{'localTasks'} itself, but
          #!!  tasks are usually done in duration order so ...
          my ($temp) = $queueTime/$workConfig->{$working}{'longest'};
          # Divide by the number of work types sharing the same processors, if any (perhaps skip directories were created after starting this work).
          $temp /= $canDoSome{$workConfig->{$working}{'proc'}} if $canDoSome{$workConfig->{$working}{'proc'}} > 1;
          if (exists($workConfig->{$working}{'tuneTime'}))
            { # This is pretty crude but likely sufficient for now.
              if ($trials)
                {
                  my ($ratio) = Math::BigFloat->new($trials);
                  $ratio /= Math::BigFloat->new($workConfig->{$working}{'tuneValue'});
                  $ratio *= $took;
                  $took = $ratio->as_int();
                  #!!incomplete?
                }
              if (3*$took < $workConfig->{$working}{'tuneTime'} && $workConfig->{$working}{'tuneValue'} < 1e10)
                {
                  $workConfig->{$working}{'tuneValue'} *= 2;
                }
              elsif ($took < $workConfig->{$working}{'tuneTime'})
                {
                  $workConfig->{$working}{'tuneValue'} += $workConfig->{$working}{'tuneAdjust'};
                }
              elsif (9*$took > 10*$workConfig->{$working}{'tuneTime'})
                {
                  $workConfig->{$working}{'tuneValue'} /= 2;
                }
              elsif ($took > $workConfig->{$working}{'tuneTime'} && $workConfig->{$working}{'tuneTime'} > 2*$workConfig->{$working}{'tuneAdjust'})
                {
                  $workConfig->{$working}{'tuneValue'} -= $workConfig->{$working}{'tuneAdjust'};
                }
              $workConfig->{$working}{'tuneValue'} = $workConfig->{$working}{'tuneValueMin'},
                ++$workConfig->{$working}{'tuneTime'}
                if (exists($workConfig->{$working}{'tuneValueMin'})
                    && $workConfig->{$working}{'tuneValue'} < $workConfig->{$working}{'tuneValueMin'});
            }

          # Multiply by the number of processors for this type.
          #  !!Should subtract a constant load for cpu work (e.g., Prime95 starting at boot), but knowing that value is difficult.
          #  !!My only half-way reasonable thought right now is to subtract (count of active cpu tasks + this one) from $load.
          $temp *= ($processors{$workConfig->{$working}{'proc'}} - 0); # !! -0 for unknown constant load.

          if ($workConfig->{$working}{'localTasks'} > $workConfig->{$working}{'taskLimit'})
            {
              $workConfig->{$working}{'localTasks'} = $workConfig->{$working}{'taskLimit'};
            }
          elsif ($workConfig->{$working}{'localTasks'} < 10)
            {
              $workConfig->{$working}{'localTasks'} = 10;
            }
          elsif ($temp > $workConfig->{$working}{'localTasks'} + 2)
            {
              ++$workConfig->{$working}{'localTasks'};
              print(STDOUT "increasing max $working tasks to $workConfig->{$working}{'localTasks'}; estimated tasks at $workConfig->{$working}{'longest'} seconds is $temp\n") if $verbose > 2;
            }
          elsif ($temp < $workConfig->{$working}{'localTasks'} - 2)
            {
              --$workConfig->{$working}{'localTasks'};
              print(STDOUT "decreasing max $working tasks to $workConfig->{$working}{'localTasks'}; estimated tasks at $workConfig->{$working}{'longest'} seconds is $temp\n") if $verbose > 2;
            }
          # Reduce 'longest' some if it is more than twice the time the task took.
          $workConfig->{$working}{'longest'} *= 9,
            $workConfig->{$working}{'longest'} /= 10
            if $workConfig->{$working}{'longest'} > 2*$took;
        }
      unlink("$workInProgress{$fd}{'output'}")
        or $! == 2
        or die "could not unlink done $working $fd $workInProgress{$fd}{'output'}, $!\n";
      unlink("$workInProgress{$fd}{'errout'}")
        or $! == 2
        or die "could not unlink done $working $fd $workInProgress{$fd}{'errout'}, $!\n";
      delete $workInProgress{$fd};
    }
  warn(localtime() . " $ready ready per select() beyond known active descriptors; vector $fdbitsIn\n")
    if $ready > 0;
} # while (-d $lockDir)

$checkConfig || ! -d $stopDir || -d "$stopDir/make" or rmdir($stopDir) or warn(localtime() . " could not rmdir $stopDir, $!\n");
exit($err);

# Get the Mersenne exponent from the work file
sub GetFileExp($)
{
  my ($file) = @_;
  local ($_);
  warn("GetFileExp() could not open $file, $!\n"),
    return(undef)
    unless open(my $fh, '<', "$file");
  my ($line) = <$fh>;
  close($fh) or die "GetFileExp() could not close $file, $!\n";
  # All mersfmt lines should match this pattern
  return($1) if defined($line) && $line =~ m/^M[(] ([1-9][0-9]*) [)]./;
  # GIMPS Trial Factoring lines - Factor=EF99B4ECC09A61CBA622729CA63D042B,132365237,71,72
  return($1) if defined($line) && $line =~ m/^[A-Za-z]+=[0-9A-F]+,([1-9][0-9]+),[1-9][0-9]+,[1-9][0-9]+$/;
  warn(localtime() . " bad work line in $file, $line")
    if defined($line);
  !defined($file) || ! -f("$file") || unlink("$file") or $! == 2 or die "GetFileExp() could not remove $file, $!\n";
  return(undef);
}

#!!The current run, 2020 Mar 15, has no renewal lines for PPlus1 ('q') work which does not match
#!!  the count of files in $topDir/work/pplus1/, 1264.  So add some debugging and read the
#!!  directories once a day or so.

# Put appropriate 'renew <exp> <kind>' lines in the first argument (work.doWork/renewals).
#   Free least urgent - or non-numeric priority - task of each type for which we have the desired max count or is being skipped.
sub WriteRenewals($)
{
  my ($renewals) = @_;
  local ($_);
  my ($dir) = "$topDir/work";
  my ($file, $line, $dirh, $high, $highExp, $fileExp);
  my (%workCounts); #!! count of tasks in %exponents of each kind (key of HASH).
  $workCounts{$_} = 0
    foreach (grep { !$workConfig->{$_}{'skip'} } keys(%{$workConfig}));
  open(my $renewFH, '>', "$renewals.tmp") or die "WriteRenewals() cannot open $renewals.tmp, $!\n";
  if (scalar(keys(%exponents)) && time() < $nextReadWorkFiles)
    {
      foreach (keys(%dropList))
        { # $mersServer may not know about GIMPS reservations.
          # Disable this for now; as is, it could remove a task in progress.
          last;
          #!!This dependency on the file path matching this pattern should be avoided.
          next if !exists($exponents{$_}) || $exponents{$_} =~ m,gimps,i;
          unlink("$topDir/work/$exponents{$_}") if -f "$topDir/work/$exponents{$_}";
          delete $exponents{$_};
        }
      %dropList = ();
      foreach (keys(%exponents))
        { # The server wants the mers letter as well to know what kind of work it is.
          my (@parts) = split(/\//, $exponents{$_});
          print($renewFH "renew $_\n"), next unless $#parts > 0 && exists $workConfig->{$parts[0]};
          print($renewFH "renew $_\n"), next unless exists $workConfig->{$parts[0]}{'letter'};
          ++$workCounts{$parts[0]};
          print($renewFH "renew $_ $workConfig->{$parts[0]}{'letter'}\n");
        }
      print(STDOUT "Count of tasks of each kind in memory\n");
      print(STDOUT "$_\t$workCounts{$_}\n")
        foreach (keys(%workCounts));
    RENEWALS:
      foreach (keys(%tasks))
        { #!!This dependency on the key in %tasks matching 'gimps' should be avoided.
          next if m,gimps,i;
          next if !exists($tasks{$_}) || !exists($workConfig->{$_}) || ref($tasks{$_}) ne 'HASH';
          next if scalar(keys(%{$tasks{$_}})) < $workConfig->{$_}{'localTasks'} && $workConfig->{$_}{'skip'} == 0;
          my ($toFree, $nextToLast) = sort { $b <=> $a } keys(%{$tasks{$_}});
          next unless defined $toFree;
          my ($workFile) = "$topDir/work/$_/$toFree";
          # Don't remove work that is in progress
          warn("may remove $workFile as lowest priority work\n") if $verbose > 6;
          foreach my $worker (keys(%workInProgress))
            { # Don't remove a task that is in progress.
              warn("  $workInProgress{$worker}{'work'} in progress\n") if $verbose > 6;
              next RENEWALS if $workInProgress{$worker}{'work'} eq "$workFile";
            }
          next unless -f $workFile && -s $workFile;
          $fileExp = GetFileExp($workFile);
          delete $tasks{$_}{"$toFree"};
          unlink($workFile) or $! == 2 or die "WriteRenewals() could not remove $workFile, least urgent work, $!\n";
          next if !defined $fileExp;
          print($renewFH "free $fileExp\n");
          delete $exponents{$fileExp};
        }
    }
  else
    { # Read the work files every 24 hours.
      $nextReadWorkFiles = $minInterval*24 + time();
      print(STDOUT localtime() . " reading work files to create renewals; may take a while\n") if $verbose > 1;
      opendir(my $workDH, "$dir") or die "WriteRenewals() could not opendir $dir, $!\n";
      my (%workPaths);
      $workPaths{$_} = 1
        foreach (grep { !m,^[.][.]?$, } readdir($workDH));
      closedir($workDH) or die "WriteRenewals() could not closedir $dir, $!\n";
    DIR2:
      foreach my $dir2 (keys(%{$workConfig}))
        {
          delete $workPaths{$dir2};
          undef $high;
          opendir(my $worktypeDH, "$dir/$dir2") or die "WriteRenewals() could not opendir $dir/$dir2, $!\n";
          my (@entries) = (grep { ! -d("$dir/$dir2/$_") && -f(_) && -s(_) && !-l("$dir/$dir2/$_") } readdir($worktypeDH));
          closedir($worktypeDH) or die "WriteRenewals() could not closedir $dir/$dir2, $!\n";
          my ($fileCnt) = 0;
          foreach my $f (@entries)
            { # GetFileExp() and FindNextWork() remove empty work files.
              warn("ignoring unexpected file $f in $dir/$dir2"),
                next
                unless $f =~ m,$floatWhole,;
              $f = "$1";
              $fileExp = GetFileExp("$dir/$dir2/$f");
              #!!This dependency on the file path matching this pattern should be avoided.
              # Disable this for now; it should check that it is not in progress first
              if (0 && defined($fileExp) && exists($dropList{$fileExp}) && $dir2 =~ m,gimps,i)
                { # Should have already been dealt with, but easy to do.
                  warn("removing $dir/$dir2/$f per drop request from server") if $verbose > 6;
                  unlink("$dir/$dir2/$f") or $! == 2 or die "WriteRenewals() could not remove $dir/$dir2/$f, to-be-dropped $fileExp, $!\n";
                  delete $exponents{$fileExp};
                  delete $dropList{$fileExp};
                  next;
                }
              if (defined($fileExp))
                {{
                  print($renewFH "renew $fileExp\n"), last unless exists $workConfig->{$dir2};
                  print($renewFH "renew $fileExp\n"), last unless exists $workConfig->{$dir2}{'letter'};
                  ++$workCounts{$dir2};
                }}
              ++$fileCnt;
              $exponents{$fileExp} = "$dir2/$f";
              next if $dir2 =~ m,gimps,i || (defined($high) && $f =~ m/$floatWhole/ && ($high !~ m/$floatWhole/ || $f < $high));
              $high = $f;
              $highExp = $fileExp;
            }
          #!!This dependency on the file path matching this pattern should be avoided.
          next if (($fileCnt < $workConfig->{$dir2}{'localTasks'} && $workConfig->{$dir2}{'skip'} == 0) || !defined($high) || $dir2 =~ m,gimps,i);
          # Don't remove work that is in progress
          warn("may remove $dir/$dir2/$high as lowest priority work\n") if $verbose > 6;
          foreach my $worker (keys(%workInProgress))
            { # Don't remove a task that is in progress.
              warn("  $workInProgress{$worker}{'work'} in progress\n") if $verbose > 6;
              next DIR2 if $workInProgress{$worker}{'work'} eq "$dir/$dir2/$high";
            }
          warn("high $high tainted"),
            next DIR2
            if Scalar::Util::tainted($high);
          unlink("$dir/$dir2/$high")
            or $! == 2
            or die "WriteRenewals() could not remove $dir/$dir2/$high, least urgent work, $!\n";
          print($renewFH "free $highExp\n");
          delete $exponents{$highExp};
        }
      print(STDOUT "Count of tasks of each kind on disk\n");
      print(STDOUT "$_\t$workCounts{$_}\n")
        foreach (keys(%workCounts));
      if (scalar(keys(%workPaths)))
        {
          warn("extra files in $dir");
          print("  $_")
            foreach (sort(keys(%workPaths)));
          print("\n");
        }
    }
  close($renewFH) or die "WriteRenewals() could not close $renewals.tmp, $!\n";
  rename("$renewals.tmp", "$renewals") or die "WriteRenewals() could not rename $renewals.tmp to $renewals, $!\n";
}

sub AskCount($)
{
  my ($kind) = @_;
  local ($_);
  return(0) if (!exists($workConfig->{$kind}) || $workConfig->{$kind}{'skip'});
  FindNextWork($kind, 0)
    if !exists($tasks{$kind}) || ref($tasks{$kind}) ne 'HASH' || scalar(keys(%{$tasks{$kind}})) < 10;
  return($workConfig->{$kind}{'localTasks'}) if !exists($tasks{$kind}) || ref($tasks{$kind}) ne 'HASH';
  my ($cnt) = scalar(keys(%{$tasks{$kind}}));
  return($cnt >= $workConfig->{$kind}{'localTasks'} ? 0 : $workConfig->{$kind}{'localTasks'} - $cnt);
}

# When a GPU factorer cannot do a task, call this with the exponent and mersfmt G or g line
#   to create the implied task.  This is also used by CreateMoreWork() when doJoiner.pl does
#   not completely cover a gap.
sub CreateJoinTask($$$)
{ # Kind of work to create, exponent, and gap in mers format.
  my ($kind, $e, $task) = @_; # Mersenne exponent and mers format line (G or g)
  my ($pri) = Mersenne::mersPriority(split(' ', $task));
  return unless $pri >= 0.0;
  $pri += 0.01 while (-s "$topDir/work/$kind/$pri");
  my ($cjt);
  open($cjt, '>', "$topDir/work/$kind/$pri") or die "could not create $topDir/work/$kind/$pri, $!\n";
  print($cjt $task . "\n");
  close($cjt) or die "could not close $topDir/work/$kind/$pri, $!\n";
  $tasks{$kind}{$pri} = "$topDir/work/$kind/$pri";
  $workConfig->{$kind}{'workAvailable'} = 1;
  $exponents{$e} = "$kind/$pri";
}

# Create more work when appropriate.
# Arguments are the kind of work just completed, the input file for it, and the mers output from it.
sub CreateMoreWork($$@)
{
  my ($kind, $inFile, @newMers) = @_;
  local $_;
  # Do not create more work if the mers server has been contacted and we have more than a third of the task limit
  # ... unless a trial factoring gap was not closed.
  return if ($lastSendWorked && 3*($workConfig->{$kind}{'localTasks'}) > $workConfig->{$kind}{'taskLimit'}
             && $workConfig->{$kind}{'letter'} ne 'G');
  my (@oldMers) = GetMersLines($inFile);
  return unless $#oldMers >= 0;
  die("logic problem - GetMersLines() returned scalar, not array\n")
    if $oldMers[0] =~ m,^[0-9]+$,;
  my ($pri); # priority of the new task
  my ($n); # the Mersenne exponent
  # Retain all the (unique) factors.
  my (@factors);
  my ($prior) = '';

  # This presumes the old factors are distinct, prime, etc., but the server should have verified them.
  my ($oldFactorCount) = 0;
  ++$oldFactorCount foreach (grep { m/C:/ } @oldMers);

  foreach (sort { length($a) <=> length($b) || $a cmp $b } grep { m/C:/ } (@oldMers, @newMers, ))
    { # skip duplicate and obviously composite or false factors
      next if $prior eq $_;
      $prior = $_;
      m/^M[(] ([1-9][0-9]*) [)]C: ([1-9][0-9]*)/ or return; # not mers format
      $n = Math::BigInt->new("$1") unless $n;
      my ($f) = Math::BigInt->new("$2");
      next if $f % $n != 1; # prime factors of composite exponent Mersennes are not always 1 mod 2*n but must be 1 mod n
      foreach my $prior (@factors)
        {
          $f = 0, last if $f == $prior;
          $f /= $prior if $f % $prior == 0;
        }
      next if $f < 1;
      #!!This is systematically low for the 'c' case, which is fine in this context.
      #!!It is almost always too high for the 'lm' case, which is somewhat problematic.
      #!!Correct would be to use Mersenne.pm to pull out the factors and count the digits remaining in the clm cofactors.
      #!!Almost as accurate would be using BigFloat to calculate base 2 logs of the factors and subtract those from the exponent.
      $pri -= $f->length();
      push(@factors, $f);
    }
  # Add cofactor lines; new output could have new SPRP data since the ecm program checks a new cofactor.
  my (@cofacLines);
  my ($list) = ''; # of mers letters seen
  # Sorting by length and then by cmp is enough since LM lines will be shorter.
  foreach (sort { length($a) <=> length($b) || $a cmp $b }
           grep { m/[)][clmLM]/ } (@oldMers, @newMers, ))
    { # ignore l (m) lines after an L (M) line has been seen
      m/[)](.)/;
      next if $list && $list =~ m/$1/i;
      $list .= $1;
      push(@cofacLines, $_);
    }
  my ($newFactorCount) = ($#factors + 1) - $oldFactorCount;
  # If a new factor was found, doing more should wait for the SPRP check of the new cofactor(s).
  return
    if $newFactorCount && $workConfig->{$kind}{'more'} eq 'C';
  # Don't create new work of the same type if it is being skipped.
  return if $workConfig->{$kind}{'skip'};
  # The new task is always the same $kind as the old unless it was gpu, which becomes joinNotGPU.
  my ($mers) = '';
  $mers = ("M( $n )C: " . join("\nM( $n )C: ", @factors) . "\n")
    if @factors && $workConfig->{$kind}{'more'} eq 'C';
  $mers .= join("\n", @cofacLines) . "\n";
  my ($newPri);
  # Only ECM, P-1, and P+1 work can be created like this; gaps can't
  #   ... except when the child did not cover the entire gap.
  if ($kind eq 'ecm')
    {
      my ($effort) = Math::BigInt->bzero();
      my ($hiBnd) = Math::BigInt->bzero();
      foreach (@oldMers, @newMers, )
        { # retain largest effort and bound
          next unless m/^M[(] ([1-9][0-9]*) [)]E: ([1-9][0-9]*) ([1-9][0-9]*)/;
          $n = Math::BigInt->new("$1") unless $n;
          my ($newEffort) = Math::BigInt->new("$2");
          my ($newBound) = Math::BigInt->new("$3");
          $effort = $newEffort if $effort < $newEffort;
          $hiBnd = $newBound if $hiBnd < $newBound;
        }
      foreach (@newMers)
        { # add in new curves; !!should only be done if @newMers has no (later?) E: lines
          next unless m/^M[(] ([1-9][0-9]*) [)]e: ([1-9][0-9]*) *([0-9]*)/;
          $n = Math::BigInt->new("$1") unless $n;
          my ($bound) = Math::BigInt->new("$2");
          my ($curves) = ($3 ? Math::BigInt->new("$3") : 1);
          $effort += $bound*$curves;
          $hiBnd = $bound if $hiBnd < $bound;
        }
      return unless $n && $effort && $hiBnd;
      $mers .= "M( $n )E: $effort $hiBnd\n";
      $newPri = log("$n.0" + "$effort.0");
    }
  elsif ($kind eq 'pminus1' || $kind eq 'pplus1')
    {
      my ($stage1Bound) = Math::BigInt->bzero();
      my ($letter);
      foreach (@newMers)
        {
          next unless m/^M[(] ([1-9][0-9]*) [)]([oq]): ([1-9][0-9]*) *([0-9]*)/;
          $n = Math::BigInt->new("$1") unless $n;
          $letter = "$2";
          next unless ($kind eq 'pminus1' && $letter eq 'o')
            || ($kind eq 'pplus1' && $letter eq 'q');
          $stage1Bound = Math::BigInt->new("$3") + 1000000;
        }
      return unless $n && $stage1Bound;
      $letter = 'o' if !$letter && $kind eq 'pminus1';
      $letter = 'q' if !$letter;
      $mers .= "M( $n )$letter" . ": $stage1Bound\n";
      $newPri = log("$n.0" + "$stage1Bound.0");
    }
  elsif ($workConfig->{$kind}{'letter'} eq 'G')
    {
      my (@gapsLeft);
      foreach my $newMers (@newMers)
        { # Only new joins and old gaps need to be "compared".
          next unless $newMers =~ m/^M[(] ([1-9][0-9]*) [)]J: ([1-9][0-9]*) ([1-9][0-9]*)/;
          $n = Math::BigInt->new("$1") unless $n;
          my ($jLow) = Math::BigInt->new("$2");
          my ($jHi) = Math::BigInt->new("$3");
          next unless $jLow <= $jHi; # ignore invalid joins
          while (@oldMers)
            {
              my ($oldGap) = shift(@oldMers);
              next unless $oldGap =~ m/^M[(] ([1-9][0-9]*) [)]G: ([1-9][0-9]*) ([1-9][0-9]*)/;
              next unless $n == Math::BigInt->new("$1");
              my ($gLow) = Math::BigInt->new("$2");
              my ($gHi) = Math::BigInt->new("$3");
              next unless $gLow <= $gHi; # ignore invalid gaps
              next if $jLow <= $gLow && $gHi <= $jHi; # covered gap
              my (@lines) = GetMersLines($inFile);
              die("logic problem - GetMersLines() returned scalar, not array\n")
                if $lines[0] =~ m,^[0-9]+$,;
              print("no overlap of gap and join - skipping $kind:\n"
                    . $oldGap . " # gap\n" . $newMers . " # join\n"
                    . "work $inFile content:\n"
                    . join("\n", @lines) . "\n"),
                $workConfig->{$kind}{'skip'} = 1,
                next
                if $jHi < $gLow || $gHi < $jLow;
              push(@gapsLeft, "M( $n )G: $jHi $gHi\n")
                if $jHi < $gHi;
              push(@gapsLeft, "M( $n )G: $gLow $jLow\n")
                if $gLow < $jLow;
            }
          push(@oldMers, @gapsLeft);
          @gapsLeft = ();
        }
      CreateJoinTask($kind eq 'gpu' ? 'joinNotGPU' : $kind, $n, $_)
        foreach (@oldMers);
      return;
    }      
  return unless $n && $newPri;
  warn("CreateMoreWork - new priority $newPri of M$n is not floating point"),
    return
    unless $newPri =~ m/$floatWhole/ && int($newPri) != $newPri;
  my ($newFN) = "$1" . "$n";
  $newFN .= '0'
    while -s "$topDir/work/$kind/$newFN" || exists $tasks{$kind}{$newFN};
  my ($fh);
  return unless open($fh, '>', "$topDir/work/$kind/$newFN");
  print($fh "$mers");
  close($fh) or return;
  $tasks{$kind}{$newFN} = "$topDir/work/$kind/$newFN";
  warn("CreateMoreWork - $tasks{$kind}{$newFN}\n");
}

# This can take a while, at least on MSWindows, due to all the disk accesses,
#   so retain the list of work files in a hash, $tasks{$kind}, by their priority
#   and return quickly if the second argument is true.
sub FindNextWork($$)
{
  my ($kind, $quick) = @_;
  my ($file, $result);
  local ($_);
  # The "quick" method results in small issues elsewhere, so only use it on MSWindows for now.
  #!!Should do a bit of this rather than stalling in select(), similar to internal work in the mers server.
  #!!Disable $quick for now; it seems to confuse things so that reservations are lost
  $quick = 0; #!!unless $^O =~ m/MSWin32/i;
  print(STDOUT localtime() . " finding $kind work; may take a while\n") if $verbose > 2;
  delete $tasks{$kind};
  opendir(my $workDH, "$topDir/work/$kind")
    or die "FindNextWork() could not opendir $topDir/work/$kind, $!\n";
  my (@entries) = sort { $a <=> $b } grep { m,$floatWhole, } readdir($workDH);
  closedir($workDH)
    or die "FindNextWork() could not closedir $topDir/work/$kind, $!\n";
 TASK:
  foreach (@entries)
    {
      -s "$topDir/work/$kind/$_"
        or unlink "$topDir/work/$kind/$_"
        or $! == 2
        or die "FindNextWork() could not unlink empty $topDir/work/$kind/$_, $!\n";
      next
        unless -f "$topDir/work/$kind/$_" && ! -l "$topDir/work/$kind/$_";
      m,$floatWhole,
        or die "FindNextWork() - file '$_' under $topDir/work/$kind was a number but is not a number now";
      my ($f) = "$1";
      $tasks{$kind}{"$f"} = "$topDir/work/$kind/$f";
      foreach my $fd (keys(%workInProgress))
        {
          next TASK
            if $workInProgress{$fd}{'work'} eq "$topDir/work/$kind/$f";
        }
      $result = $f
        if !defined($result) || $f < $result;
      last
        if $quick;
    }
  return(undef)
   unless defined($result);
  my ($taint);
  print(STDOUT "FindNextWork topDir is tainted\n"),
    $taint = 1
    if Scalar::Util::tainted($topDir);
  print(STDOUT "FindNextWork kind is tainted\n"),
    $taint = 1
    if Scalar::Util::tainted($kind);
  print(STDOUT "FindNextWork result is tainted\n"),
    $taint = 1
    if Scalar::Util::tainted($result);
  return(undef)
    if $taint;
  return("$topDir/work/$kind/$result");
}

# Get the mers format lines from the given file.
sub GetMersLines($)
{
  my ($file) = @_;
  local $_;
  my (@mers);
  return(@mers) unless -s $file;
  my ($infh);
  return(@mers) unless open($infh, '<', "$file");
  my (@gapLines); # indices of @mers that are 'G'ap lines
  while (<$infh>)
    {
      chomp;
      next
        unless m/^M[(] [1-9][0-9]* [)](.)/;
      my ($l) = "$1";
      push(@mers, $_);
      push(@gapLines, $#mers)
        if $l eq 'G' && m/ [)]G: /;
    }
  if ($#gapLines > 0)
    { # Somehow, some gap task files have more than one gap, usually overlapping.  Combine them if so.
      print("GetMersLines - logic error - $file line $gapLines[0] is not a gap:\n$mers[$gapLines[0]]\n"),
        @mers = (),
        return(@mers)
        unless $mers[$gapLines[0]] =~ m/^M[(] ([1-9][0-9]*) [)]G: +([1-9][0-9]*) +([1-9][0-9]*)/;
      my ($exp) = Math::BigInt->new("$1");
      my ($low) = Math::BigInt->new("$2");
      my ($high) = Math::BigInt->new("$3");
      foreach my $i (1 .. $#gapLines)
        {
          print("GetMersLines - logic error - $file line $gapLines[$i] is not a gap:\n$mers[$gapLines[$i]]\n"),
            @mers = (),
            return(@mers)
            unless $mers[$gapLines[$i]] =~ m/^M[(] ([1-9][0-9]*) [)]G: +([1-9][0-9]*) +([1-9][0-9]*)/;
          my ($exp2) = Math::BigInt->new("$1");
          my ($low2) = Math::BigInt->new("$2");
          my ($high2) = Math::BigInt->new("$3");
          print("GetMersLines - logic error - $file contains lines for more than one Mersenne - M$exp and M$exp2\n"),
            @mers = (),
            return(@mers)
            unless $exp == $exp2;
          print("GetMersLines - $file contains gaps that do not overlap:\nM( $exp )G: $low $high\nM( $exp )G: $low2 $high2\n"),
            @mers = (),
            return(@mers)
            if $low > $high2 || $high < $low2;
          # Two gaps overlap - retain the lower low value and the higher high value.
          $low = $low2
            if $low > $low2;
          $high = $high2
            if $high < $high2;
        }
      while (@gapLines)
        { # Delete all the pre-existing gaps ...
          my ($i) = pop(@gapLines);
          splice(@mers, $i, 1);
        }
      # ... and add the merged gap.
      push(@mers, "M( $exp )G: $low $high");
    }
  return(@mers);
}

sub GetLoad()
{ # For various reasons, never return less than the count of CPU tasks.
  my ($loadFH);
  local ($_);
  my ($load) = $1 if open($loadFH, '<', '/proc/loadavg') && <$loadFH> =~ m/$floatPattern/ && close($loadFH) && $1 > 0.0;
  $load = 0.1 unless $load;
  my ($min) = 0.0 + scalar(grep { $workConfig->{$workInProgress{$_}{'type'}}{'proc'} eq 'cpu' } keys(%workInProgress));
  return($load) if $load >= $min;
  return($min);
}

sub StartTask($$$$$)
{ # Note that 'exec' cannot be used; MSWindows/Cygwin does not implement it
  #   correctly in that the select() is never notified when the child exits,
  #   (which may be a bug in the shell related to the redirections).
  my ($kind, $cmd, $extra, $file, $pri) = @_;
  warn("StartTask given unexpected kind '$kind'\n"),
    return
    unless exists $workConfig->{$kind};
  my ($outfile) = "$workConfig->{$kind}{'done'}/$pri";
  if (-e($outfile))
    { # $outfile, ending with $pri, should be unique, but ...
      my ($suffix) = 0;
      ++$suffix while -e "$outfile.$suffix";
      $outfile .= ".$suffix";
    }
  my ($errfile) = "$outfile.err";
  my ($pid) = open my $fh, '-|', "$cmd $extra < $file > $outfile 2> $errfile";
  my (@result) = ($fh, $outfile, $errfile);
  Unblock($fh, "$cmd $extra into $outfile");
  return(@result) if $pid;
  warn(localtime() . " could not start work $kind, $!\n");
  $workConfig->{$kind}{'skip'} = 1 if exists $workConfig->{$kind};
  return(undef);
}

sub SaveConfig($)
{
  my ($file) = @_;
  local ($_);
  open(my $saveFH, '>', "$file") or die "could not rewrite $file, $!\n";
  print($saveFH "$_" . " count $processors{$_}\n") foreach keys(%processors);
  print($saveFH "verbose $verbose\nmaxTime $maxTime\n" . "nextSend $nextSend\n" . "UUID $uuid\n"
        . "queue $queueTime\nsendInterval $sendInterval\nloadLimit $loadLimit\n");
  print($saveFH "taskLimit $defaults{'taskLimit'}\n");
  foreach (keys(%{$workConfig}))
    {
      $workConfig->{$_}{'time'} = $maxTime if ! exists $workConfig->{$_}{'time'};
      foreach my $key ('localTasks', 'taskLimit', 'askArgs', 'cmdArgs', 'time', 'skip', 'longest', 'ratio',
                       'tuneOpt', 'tuneTime', 'tuneValue', 'tuneAdjust', 'tuneValueMin', )
        {
          print($saveFH "$key $_" . '=' . "$workConfig->{$_}{$key}\n") if exists($workConfig->{$_}{$key}) && $workConfig->{$_}{$key} ne '';
        }
    }
  close($saveFH) or die "could not close $file, $!\n";
}

sub UpdateConfig($$$)
{
  my ($file, $ignoreVerbose, $ignoreQueue) = @_;
  local ($_);
  open(my $timesFH, '<', $file) or warn "cannot read $file, $!\n";
  print(STDOUT "reading new $file file:\n") if defined($verbose) && $verbose > 4;
 LINE :
  while (<$timesFH>)
    {
      print(STDOUT "  $_") if defined($verbose) && $verbose > 4;
      chomp;
      foreach my $procType (keys(%processors))
        {
          $processors{$procType} = $1, next LINE if (m/^$procType count ([0-9]+)$/i);
        }
      $verbose = $1 if (m/^verbose[= ]([0-9]+)$/i && (!defined($verbose) || !$ignoreVerbose));
      next if (m/^verbose[= ]([0-9]+)$/i);
      $queueTime = $zero + $1 if (m/^queue[= ]([1-9][0-9]*)$/i && (!defined($queueTime) || !$ignoreQueue));
      next if (m/^queue[= ]([1-9][0-9]*)$/i);
      $uuid = $1 if (m/^uuid[= ]([-_.a-z0-9]+)$/i && !defined($uuid));
      next if (m/^uuid[= ]([-_.a-z0-9]+)$/i);

      # $3 here is the integer portion of the float pattern
      $sendInterval = $zero + $3, next if (m/^sendInterval[= ]$floatPattern$/i);

      $loadLimit = $1, next if (m/^loadLimit[= ]$floatPattern$/i);
      $maxTime = $zero + $1, next if (m/^maxTime[= ]([1-9][0-9]*)$/i);
      $nextSend = $zero + $3, next if (m/^nextSend[= ]$floatPattern$/i);
      $defaults{'taskLimit'} = $1, next if (m/^taskLimit[= ]([1-9][0-9]+)$/i);
      # These would be in the next loop but the values do not have to be digits, unlike for the other entries.
      $workConfig->{$1}{'askArgs'} = $2, next if (m/^askArgs ([a-zA-Z0-9]+)[= ]([[:print:]]*)$/i) && exists($workConfig->{$1});
      warn("unknown work type '$1'; ignoring '$_'\n"), next if m/^askArgs ([a-zA-Z0-9]+)[= ]([[:print:]]*)$/i;
      $workConfig->{$1}{'cmdArgs'} = $2, next if (m/^cmdArgs ([a-zA-Z0-9]+)[= ]([[:print:]]*)$/i) && exists($workConfig->{$1});
      warn("unknown work type '$1'; ignoring '$_'\n"), next if m/^cmdArgs ([a-zA-Z0-9]+)[= ]([[:print:]]*)$/i;
      $workConfig->{$1}{'tuneOpt'} = $2, next if (m/^tuneOpt ([a-zA-Z0-9]+)[= ]([[:print:]]*)$/i) && exists($workConfig->{$1}) && exists($workConfig->{$1}{'tuneTime'});
      warn("unknown or unexpected work type '$1' with 'tuneOpt'; ignoring '$_'\n"), next if m/^tuneOpt ([a-zA-Z0-9]+)[= ]([[:print:]]*)$/i;
      for my $key ('localTasks', 'taskLimit', 'time', 'skip', 'longest', 'ratio', 'tuneTime', 'tuneValue', 'tuneAdjust', 'tuneValueMin', )
        {
          if (m/^$key ([a-zA-Z0-9]+)[= ](-?[0-9]+|NaN)$/i)
            {
              my (@matches) = ($1, $2, );
              print("ignoring '$key' for work type '$matches[0]' since 'tuneTime' not supported\n"),
                next LINE
                if $key =~ m,^tune, && !exists $workConfig->{$matches[0]}{'tuneTime'};
              if (exists($workConfig->{$matches[0]}) && exists($workConfig->{$matches[0]}{$key})
                  # Only 'time' can be negative.  !!Most others should not be 0 but that is not verified.
                  && ($matches[1] eq 'NaN' || $matches[1] >= 0 || $key eq 'time'))
                {
                  $workConfig->{$matches[0]}{$key} = Math::BigInt->new($matches[1] eq 'NaN' ? '1000' : "$matches[1]");
                  next LINE;
                }
              warn("unknown work type '$matches[0]'; ignoring '$_'\n");
              next LINE;
            }
        }
      warn "$file contains unexpected data: $_\n";
      unlink("$file" . ".bad") or $! == 2 # ENOENT, no such file or directory
        or warn "could not remove old $file.bad, $!\n";
      rename("$file", "$file" . ".bad");
      last;
    }
  close($timesFH) or warn "could not close $file, $!\n";
  $defaults{'taskLimit'} = 10 if $defaults{'taskLimit'} < 10;
  $defaults{'taskLimit'} = 20000 if $defaults{'taskLimit'} > 20000;
  foreach (keys(%{$workConfig}))
    {
      print(STDOUT "checking $_ in workConfig\n") if defined($verbose) && $verbose > 3;
      $workConfig->{$_}{'skip'} = ($processors{$workConfig->{$_}{'proc'}}) < 1 ? 1 : 0;
      $workConfig->{$_}{'nextAsk'} = $nextSend + int(rand(120)) if exists $workConfig->{$_}{'nextAsk'};
      $workConfig->{$_}{'taskLimit'} = $defaults{'taskLimit'} if $workConfig->{$_}{'taskLimit'} > $defaults{'taskLimit'};
      $workConfig->{$_}{'taskLimit'} = 10 if $workConfig->{$_}{'taskLimit'} < 10;
      $workConfig->{$_}{'localTasks'} = $workConfig->{$_}{'taskLimit'} if $workConfig->{$_}{'localTasks'} > $workConfig->{$_}{'taskLimit'};
      $workConfig->{$_}{'localTasks'} = 10 if $workConfig->{$_}{'localTasks'} < 10;
    }
  $workConfig->{'gpuGimpsFactor'}{'taskLimit'} = 500 if $workConfig->{'gpuGimpsFactor'}{'taskLimit'} > 500;
}

# Check for and deal with old files, last changed before this started.
#   Retain the directories and files being checked so the next call can
#     check the next file or directory instead of starting over every time.
my (@checkDirList, @checkFileList);
sub CheckOldFiles()
{
  local $_;
  @checkDirList = @oldFileDirs
    unless @checkDirList;
  while (@checkDirList)
    {
      my ($dir) = $checkDirList[0];
      my ($dh);
      if (!@checkFileList)
        {
          opendir($dh, $dir) or die "could not readdir $dir, $!";
          foreach (grep { !m,^[.][.]?$, } readdir($dh))
            {
              push(@checkFileList, "$dir/$1"),
                next
                if m,^([.0-9A-Za-z]+)$,;
              warn("unexpected file name '$_' in $dir\n");
            }
          closedir($dh) or die "could not  closedir $dir, $!";
        }
      while (@checkFileList)
        {
          my ($f) = shift(@checkFileList);
          warn("unexpected file name '$f' found while checking old files"),
            next
            unless $f =~ m,^([-/.A-Za-z0-9]+)$,;
          $f = $1;
          print("  removing soft link $f\n"),
            unlink($f),
            return
            if -l $f;
          if (-d($f))
            {
              opendir($dh, "$f") or die "could not readdir $f, $!";
              unshift(@checkFileList, "$f/$_")
                foreach grep { !m,^[.][.]?$, } readdir($dh);
              closedir($dh) or die "could not  closedir $f, $!";
              next;
            }
          my ($info) = stat($f);
          print("  $f no longer exists or stat() failed, $!\n"),
            next
            unless $info && ref($info) eq 'File::stat';
          print("  $f is not a file, directory, or soft link - " . $info->mode() . "\n"),
            next
            unless -f($f);
          # Only files changed before this script started are examined here.
          next if !$info || $startTime < $info->mtime();
          print("  removing empty file $f\n"),
            unlink("$f"),
            return
            unless -s "$f";
          my ($fh);
          print("  cannot open $f, $!\n"),
            next
            unless open($fh, '<', "$f");
          my ($results) = '';
          my ($remove, $exp, $method);
          while (<$fh>)
            { # doGPU log files should end with a mers format K line:
              $results .= $_,
                $remove = 1,
                last # this pattern is purposely broad
                if m,^M[(] +[1-9][0-9]+ +[)][A-Za-z],;
              # ecm.*.p[mp]1.log files should have the ecm command line first:
              $remove = 1
                if $. == 1 && $f =~ m,/ecm[.]([0-9]+)[.]p[mp]1[.]log$, && m,/ecm .* $1 ,;
              if (m,Using special division for factor of 2^([1-9][0-9]+)([-+])1,)
                {
                  $exp = Math::BigInt->("$1");
                  $exp *= 2 if $2 eq '+';
                }
              # Example when the ecm program finds a factor:
              #   Found prime factor of 10 digits: 1045741327
              if (m,Found .* factor of .* digits: ([1-9][0-9]*),)
                { # The exponent should be part of the log file's name.
                  my ($factor) = Math::BigInt->("$1");
                  $exp = Math::BigInt->("$1")
                    if $f =~ m,^ecm[.]([1-9][0-9]+)[.]p[mp]1[.]log$,;
                  print("  likely factor from unexpectedly named log file $f - $factor\n"),
                    last
                    unless $exp;
                  $results .= "M( $exp )C: $factor\n",
                    $remove = 1,
                    next
                    if $factor % $exp == 1;
                  # not actually a factor; ignore it
                }
              # Example ecm program output with P+/-1 bounds:
              #   Using B1=10000000, B2=24793476, polynomial x^1, x0=2755937080
              if (m/Using B1=([1-9][0-9]+), B2=([1-9][0-9]+)/)
                {
                  my ($stage1) = "$1";
                  my ($stage2) = "$2";
                  if ($f =~ m,^ecm[.]([1-9][0-9]+)[.]p([mp])1[.]log$,)
                    {
                      $exp = Math::BigInt->("$1") unless $exp;
                      $method = ($2 eq 'm' ? 'o:' : 'q:');
                    } # note the colon ':' in $method's value
                  $results .= "M( $exp )$method $stage1 $stage2\n";
                  $remove = 1;
                  next;
                }
              #!!incomplete - other possible needed data?
            }
          print("  cannot close $f, $!\n"),
            next
            unless close($fh);
          if ($results)
            { # Add the results to $topDir/sending
              my ($sfh);
              open($sfh, '>>', "$topDir/sending") or die "could not append to $topDir/sending, $!";
              print($sfh $results);
              close($sfh) or die "could not close to $topDir/sending, $!";
              unlink("$f") if $remove;
              next;
            }
          #!!incomplete - other conditions to remove the file?;
        }
      shift(@checkDirList);
    }
}

# Check some things and complain appropriately.
sub CheckWorkConfig()
{
  my ($problem);
  local ($_);
  $problem = "logic error: nextSend is NaN\n" if $nextSend->is_nan();
  $problem .= "logic error: maxTime is NaN\n" if $maxTime->is_nan();
  $problem .= "logic error: sendInterval is NaN\n" if $sendInterval->is_nan();
  $problem .= "logic error: sendInterval is NaN\n" if $sendInterval->is_nan();
  $problem .= "logic error: queueTime is NaN\n" if $queueTime->is_nan();
  foreach my $localKey (keys(%{$workConfig}))
    {
      foreach ('time', 'longest', 'nextAsk', 'proc', 'localTasks', 'taskLimit',
               'workAvailable', 'ask', 'askArgs', 'server', 'taskLimit',
               'tuneTime', 'tuneOpt', 'tuneValue', 'tuneAdjust', 'tuneValueMin', )
        {
          next if $_ eq 'nextAsk' && !exists $workConfig->{$localKey}{'name'};
          next if m,^tune, && !exists($workConfig->{$localKey}{'tuneTime'});
          $problem .= "logic error: $localKey does not have $_ entry\n",
            $workConfig->{$localKey}{'skip'} = 1,
            next
            unless ((exists($workConfig->{$localKey}{$_}) && defined($workConfig->{$localKey}{$_}))
                    || (m,^tune, && !exists($workConfig->{$localKey}{'tuneTime'})));
          next unless m/^(time|longest|nextAsk|ratio)$/;
          $workConfig->{$localKey}{$_} = Math::BigInt->new("$workConfig->{$localKey}{$_}")
            unless ref($workConfig->{$localKey}{$_}) eq 'Math::BigInt';
          $problem .= "logic error: $localKey $_ is NaN\n",
            $workConfig->{$localKey}{'skip'} = 1
            if $workConfig->{$localKey}{$_}->is_nan();
        }
      $workConfig->{$localKey}{'ratio'} = Math::BigInt->new("$workConfig->{$localKey}{'ratio'}")
        unless ref($workConfig->{$localKey}{'ratio'}) eq 'Math::BigInt';
      $workConfig->{$localKey}{'ratio'} = $zero + 1
        if ($workConfig->{$localKey}{'ratio'} <= $zero || $workConfig->{$localKey}{'ratio'}->is_nan()
            || $workConfig->{$localKey}{'ratio'} >= $maxTime);
    }
  return unless $problem;
  warn($problem);
  -d $stopDir || mkdir($stopDir)
    or die "could not mkdir $stopDir, $!\n";
  return;
}

sub Unblock($$)
{
  my ($fh, $msg) = @_;
  my ($flags) = fcntl($fh, F_GETFL, 0);
  return(1) if $flags && fcntl($fh, F_SETFL, $flags | O_NONBLOCK);
  warn("fcntl() for $msg, fd " . fileno($fh) . ", could not get flags or could not set non-blocking, $!");
  return(undef);
}
