#!/usr/bin/perl -Tw
# Use the Math::Prime::Util package to factor composite factors of Mersennes.

use strict;
use strict 'vars';
use warnings;
use POSIX;

# Setting $PATH here is only needed with Perl's -d.
BEGIN { push(@INC, '.') if -s 'Mersenne.pm'; $ENV{'PATH'} = "/usr/bin:/bin" }

# This is also only needed when Perl's -d is in effect.
use Term::ReadLine;

use Mersenne;
use MersInfo;

my ($usage) = "$0 [files]\n  Factor composite factors in the (mers format) input (files or STDIN).\n"
  . "$0 -d[ ]<debugLevel>\n  The debugLevel is given to Math::Prime::Util::prime_set_config as 'verbose'.\n"
  . "$0 --version\n  print version and exit\n"
  . "$0 --check\n  check the configuration and arguments\n"
  . "$0 --help (or -h)\n  print this usage message and exit\n";

my ($debug) = 0; # from -d if given
my ($errors) = ''; # during command line parsing
my ($quit) = 0; # quit early due to signal
my ($checkConfig) = scalar(grep { m/^--check$/ } @ARGV);
my ($version) = MersInfo::version();
print(STDOUT "$0 version $version\n")
  if $checkConfig || scalar(grep { m/^--(help|version)$/ || m/^-h$/ } @ARGV);
exit(0)
  if scalar(grep { m/^--(help|version)$/ || m/^-h$/ } @ARGV);

my (@files);
while (@ARGV)
{
  $_ = $ARGV[0];
  shift, next if m/^--check$/;
  shift, $debug = shift, next if m,^-d$,;
  $debug = $1, shift, next if m,^-d([1-9][0-9]*)$,;
  push(@files, $1), shift, next if m,^([-_.a-zA-Z0-9]+)$, && -s $1 && -f $1 && ! -l $1;
  $errors .= "unknown option or argument '" . shift() . "'\n";
}
$errors .= "debug value must be numeric\n"
  unless $debug =~ m/^([0-9]+)$/;
die $errors
  if $errors && !$checkConfig;
$debug = $1;

Math::Prime::Util::prime_set_config('verbose' => ($debug ? $debug : 0));

warn $errors if $errors;
$SIG{'HUP'} = "IGNORE";

@ARGV = @files;

foreach ('INT', 'TERM')
{
  $SIG{$_} = sub {
    my ($sig) = @_;
    warn "caught SIG$sig\n";
    $quit = 1;
  };
}
exit(0) if $checkConfig;

eval
{ # setpriority() is not always implemented
  setpriority(0, 0, 20);
};

sub factors($$)
{
  my ($exp, $factor) = @_;
  local $_;
  # For "small" factors, callling MPU::factor() is enough.
  return(Math::Prime::Util::factor($factor))
    if $exp < Math::BigInt->bone()->blsft(96);
  # For larger factors, try certain methods to limit the CPU time
  #   and perhaps not completely factor the composite factor.
  # Since $factor is a factor of M($n), $factor % $n == 1 and all
  #   its factors must also be 1 % $n.
  my (@factors) = Math::Prime::Util::pbrent_factor($factor);
  return(@factors) if $#factors > 0;
  @factors = Math::Prime::Util::prho_factor($factor);
  return(@factors) if $#factors > 0;
  @factors = Math::Prime::Util::pminus1_factor($factor);
  return(@factors) if $#factors > 0;
  @factors = Math::Prime::Util::pplus1_factor($factor);
  return(@factors) if $#factors > 0;
  my ($bound) = 128;
  while ($bound <= 2048*2048)
    {
      @factors = Math::Prime::Util::ecm_factor($factor, $bound, 2*$bound, 8);
      return(@factors) if $#factors > 0;
      $bound *= 2;
    }
  return(@factors);
}

my ($mersennes, $n);
while (<<>>)
{
  m/^M[(] ([1-9][0-9]*) [)]/
    or die "unexpected input: $_\n";
  $n = Math::BigInt->new("$1");
  $mersennes->{$n} = Mersenne->new('n' => "$n")
    unless exists $mersennes->{$n};
  warn("problems in Mersenne::parse -\n" . $mersennes->{$n}->errors()),
    next
    unless $mersennes->{$n}->parse();
  last
    if $quit;
}
for ($n = MersInfo::nextKey(0, $mersennes); defined $n; $n = MersInfo::nextKey($n, $mersennes))
{
  my (@f) = grep { !Math::Prime::Util::is_prime($_) } $mersennes->{$n}->f();
  while (@f && !$quit)
    {
      my (@f2) = factors($n, $f[0]);
      shift(@f),
        next
        unless $#f2 > 0;
      @f = grep { !Math::Prime::Util::is_prime($_) } $mersennes->{$n}->f(@f2);
    }
  $mersennes->{$n}->print();
  delete $mersennes->{$n};
  last
    if $quit;
}
exit(0);
