#!/usr/bin/perl -Tw
# Fetch mers format data from the mers database, the M/factors, and/or the results directory.

use strict;
use strict 'vars';

BEGIN { push(@INC, '.'); $ENV{'PATH'} = '/usr/bin:/bin'; }

use Mersenne;
use MersDB;
use MersInfo;

my ($MDBvL) = "$MersDB::validLetters";
# Drop letters that do not make sense here - factors, gaps, joins, and trial factoring extents.
$MDBvL =~ s/[CGgHJU]//g;

my ($usage) = "$0 --version\n  print version and exit\n"
  . "$0 <-[hH?]>\n$0 --help\n  print this usage message and exit\n"
  . "$0 --check\n  check the configuration (mostly for more complex scripts)\n"
  . "$0 <exponent> [<letters>]\n  print info for a single Mersenne number\n"
  . "$0 <letter> [count] [minPriority]\n"
  . "$0 '-e' <letter> [count] [minPriority]\n"
  . "$0 '--exponent' <letter> [count] [minPriority]\n"
  . "$0 '--valueOnly' <letter> [count] [minPriority]\n"
  . "  print info for single letter of mers format (-e, --exponent, or --valueOnly prints only the exponent(s))\n"
  . "$0 '--csv' <letter> [count] [minPriority]\n"
  . "  same but in 'comma separated values' format\n"
  . "$0 'sprp' [count] [minExponent]\n  print exponents with cofactors that need an SPRP check\n"
  . "$0 'missing' [letters]\n"
  . "  print the smallest exponents with no data for the relevant mers letters - EUcoq\n"
  . "$0 '--valueOnly' 'missing' [letters]\n"
  . "$0 '--exponent' 'missing' [letters]\n"
  . "$0 '-e' 'missing' [letters]\n"
  . "  same but only prints the exponent for the relevant mers letters (sorted by letter - EUcoq)\n"
  . "$0 '--csv' 'missing' [letters]\n"
  . "  same but in 'comma separated values' format\n"
  . "$0 'next' <letter>\n"
  . "  print the exponent of the Mersenne that is to be 'worked on' 'next' for the given\n"
  . "  mers letter - one of EoqclmLMS (ordered by the priority or exponent lower values first)\n"
  . "$0 'next'\n"
  . "  print the exponent and mers letter of the Mersenne that is to be 'worked on' 'next'\n"
  . "  among mers letters cEoq (Eoq ordered by the priority, c by smaller exponents first).\n"
  . "  Eoq work will be printed unless that exponent needs an SPRP check (c 'work').\n"
  . "$0 'reservations'\n"
  . "  print all the reservations in the db\n"
  . "$0 'reservations' <letters>\n"
  . "  print the reservations of the given mers letter(s)\n"
  . "$0 'reservations' <exponent> [exponents]\n"
  . "  print any reservation for the given exponent(s) (should be only one)\n"
  . "$0 'reservations' <UUID> [UUIDs]\n"
  . "  print the reservations by the given UUID(s) and/or hostname(s)\n"
  . "$0 'listTables'\n  lists the tables present in the database that can be printed\n"
  . "$0 <dbTableName>\n  prints the contents of the given table in the database\n"
  . "Supported mers format letters are '$MDBvL'; factors (C) and trial factoring data (GgHJU)\n"
  . "  are not in the database or do not make sense in this context (except for 'missing' U).\n"
  . "Count defaults to 1 unless the mers format letter is 'X', which ignores the count.\n"
  . "Minimum priority or exponent is a non-negative number, floating point;\n"
  . "  output will be sorted by the same with no smaller values.\n"
  . "  Priority for ECM (E), P-1 (o), and P+1 (q); exponent otherwise (clmSLMDP).\n"
  . "  Default is 0.\n"
  . "'reservations' also supports --valueOnly and --csv but always prints the exponent, UUID, and kind.\n";

my ($bigintPattern) = MersInfo::patterns('bigint');
my ($floatPattern) = MersInfo::patterns('float');
my ($uuidPattern) = MersInfo::patterns('uuid');

print("$0 version " . MersInfo::version() . "\n"),
  exit(0)
  if grep { $_ eq '--version' } @ARGV;
print("$usage"),
  exit(0)
  if grep { $_ eq '--help' || m,^-[hH?]$, } @ARGV;
my ($checkConfig) = scalar(grep { $_ eq '--check' } @ARGV);
if ($checkConfig)
{
  my ($ac) = $#ARGV;
  while ($ac >= 0)
    {
      next unless $ARGV[$ac] eq '--check';
      splice(@ARGV, $ac, 1);
    }
  continue
    {
      --$ac;
    }
}

# The first argument is usually an exponent indicating the Mersenne to print data for.
my ($n) = shift();
exit(0)
  if $checkConfig && !defined $n;
defined($n)
  or die "no arguments given\n$usage";
exit(0)
  if $checkConfig;

my ($mersDB) = MersDB->new();
my ($dbh) = $mersDB->connect('fetch');
$dbh
  or die "connect to database failed, " . $mersDB->dbError();
my (@results) = $mersDB->fetchValues('missing', 'done', 'X');

END
{
  my ($rc) = $dbh ? $dbh->disconnect() : 1;
  $rc or die("disconnect from database failed: " . $dbh->errstr . "\n");
}

if ($n eq 'listTables')
{
  print("# Tables in the database (other than single letters of the mers format)\n");
  print("#   $_\n")
    foreach ($mersDB->dbTables());
  exit(0);
}

# The second argument is usually one or more letters of the mersfmt to print data for ...
my ($letters) = shift; # ... but can be a count of tasks, exponent, mers letters, or UUID ...
                       # ... if $n is not an exponent.
my ($valueOnly) = 0;
$valueOnly = 1,
  $n = $letters,
  $letters = shift
  if $n =~ m,^(--valueOnly)|(-e)|(--exponent)$,;
$valueOnly = 'csv',
  $n = $letters,
  $letters = shift
  if $n eq '--csv';

if ($n =~ m,$bigintPattern,)
{ # Mersenne exponent and mersfmt letters.
  $n = $1;
  $letters = '' unless defined $letters;
  my ($mers) = $mersDB->find($n, $letters);
  $mers = "# no matching data found\n"
    unless $mers;
  print($mers);
  exit(0);
}

if ($n eq 'missing')
{
  $letters = "U$MDBvL" unless $letters;
  die "unsupported mersfmt letter(s) given\n$usage"
    unless $letters =~ m/^([U$MDBvL]+)$/;
  $letters = $1;
  print("# smallest exponent with no data for each mers letter:\n")
    if !$valueOnly;
  print('missing.' . join(',missing.', @{$mersDB->columns($n)}) . "\n")
    if $valueOnly eq 'csv';
  foreach (sort(split(//, $letters)))
    {
      next
        unless $mersDB->value('missing', $_);
      print('# ' . $_ . ' ' . $mersDB->value('missing', $_) . "\n"),
        next
        unless $valueOnly;
      print(join(',', $_, $mersDB->value('missing', $_)) . "\n"),
        next
        if $valueOnly eq 'csv';
      print($mersDB->value('missing', $_) . "\n");
    }
  exit(0);
}

if ($n eq 'X')
{
  my (@xexps) = $mersDB->keyList('X');
  print("M( $_ )X: " . $mersDB->value('X', $_) . "\n")
    foreach (@xexps);
  exit(0);
}

# The last supported argument is a minimum priority for work queries ($n is 'sprp' or a mers letter).
my ($pri) = shift;
if (defined($pri))
{
  die("priority '$pri' given but not supported with '$n'\n")
    unless $n eq 'sprp' || $n =~ m/^([$MDBvL]+)$/;
  die("priority must be numeric and non-negative; '$pri' given\n")
    unless $pri =~ /^$floatPattern$/;
  $pri = $1 + 0.0; # avoid taint failure and ensure floating point
}

if ($n =~ m,^(D|P|sprp)$,)
{ # 'D'one (completely factored) or 'P'rime or SPRP info
  $n = $1; # avoid taint failure
  print("# Mersenne exponents with cofactors that need an SPRP check\n")
    if $n eq 'sprp' && !$valueOnly;
  print("$n.exponent\n")
    if $valueOnly eq 'csv';
  my ($rows) = $mersDB->fetchRowsByPriority($n, $pri, $letters);
  die("MersDB::fetchRowsByPriority failed - " . $mersDB->dbError())
    unless defined $rows;
  if ($n eq 'sprp' && $#{$rows} < 0)
    {
      print($mersDB->lowestMissingExponent('c') . "\n");
      exit(0);
    }
  while (@{$rows})
    {
      my ($row) = shift(@{$rows});
      print("$row->[0]\n"),
        next
        if $valueOnly eq '1' || $n eq 'sprp';
      print("M( $row->[0] )$n\n");
    }
  exit(0);
}

if ($n eq 'reservations' && defined($letters))
{
  my (@cols) = @{$mersDB->columns($n)};
  print("# Table $n - columns '" . join("' '", @cols) . "'\n")
    unless $valueOnly;
  print("$n." . join(",$n.", @cols) . "\n")
    if $valueOnly eq 'csv';
  while ($letters =~ m,$bigintPattern,)
    {
      $letters = $1;
      my (@rows) = $mersDB->reservations($letters);
      print(($valueOnly ? '' : '# ') . join((!$valueOnly || $valueOnly ne 'csv') ? ' ' : ',', $_->{'exp'}, $_->{'uuid'}, $_->{'kind'}) . "\n")
        foreach (@rows);
      $letters = shift();
      exit(0)
        unless defined $letters;
      $letters =~ m,$bigintPattern,
        or die "invalid exponent given after 'reservations'\n$usage";
    }
  while ($letters =~ m,^([$MDBvL]+)$,)
    {
      $letters = $1;
      my (@rows) = $mersDB->reservations(substr($letters, 0, 1));
      print(($valueOnly ? '' : '# ') . join((!$valueOnly || $valueOnly ne 'csv') ? ' ' : ',', $_->{'exp'}, $_->{'uuid'}, $_->{'kind'}) . "\n")
        foreach (@rows);
      $letters = substr($letters, 1);
      $letters = shift()
        if !$letters;
      exit(0)
        unless defined $letters;
      $letters =~ m,^([$MDBvL]+)$,
        or die "invalid mers letters(s) given after 'reservations'\n$usage";
    }
  while ($letters =~ m,$uuidPattern,)
    {
      $letters = $1;
      my (@rows) = $mersDB->reservations($letters);
      print(($valueOnly ? '' : '# ') . join((!$valueOnly || $valueOnly ne 'csv') ? ' ' : ',', $_->{'exp'}, $_->{'uuid'}, $_->{'kind'}) . "\n")
        foreach (@rows);
      $letters = shift();
      exit(0)
        unless defined $letters;
      $letters =~ m,$uuidPattern,
        or die "invalid UUID given after 'reservations'\n$usage";
    }
  die "unexpected argument '$letters' after 'reservations'\n$usage";
}

if ($n eq 'next')
{
  my ($best, %choice);
  # This next string is in "priority" order except 'c' work last since it compares the
  #   exponent not the (calculated) priority.
  $letters = 'qoEc',
    $best = 1
    unless $letters;
  $letters =~ m/^([EoqclmLMS]+)$/
    or die "unsupported mersfmt letter(s) given with 'next'\n$usage";
  $letters = $1;
  my ($printLetter) = length($letters) > 1;
  foreach (split(//, $letters))
    {
      my ($exp) = $mersDB->next(m/^[Eoq]$/ ? $_ : 'sprp');
      $_ = 'c'
        if m,^[lm]$, && $exp % 8 != 4;
      $_ = 'S'
        if m,^[LM]$, && $exp % 8 != 4;
      $_ = 'l'
        if $_ eq 'c' && $exp % 8 == 4;
      $_ = 'L'
        if $_ eq 'S' && $exp % 8 == 4;
      my ($value) = $mersDB->value($_, $exp);
      $value = 0
        unless defined $value;
      $_ = 'c', 
        $value = $mersDB->value($_, $exp)
        if m,^[lmLMS]$, && ($value >= $exp || (!$value && $_ eq 'S' && $exp % 8 != 4));
      $_ = 'L',
        $value = $mersDB->value($_, $exp)
        if !$value && $_ eq 'S' && $exp % 8 == 4;
      $_ = 'm',
        $value = $mersDB->value($_, $exp)
        if !$value && $_ eq 'l';
      $_ = 'M',
        $value = $mersDB->value($_, $exp)
        if !$value && $_ eq 'L';
      $value = 0 # effort (E) or stage 1 bound (oq)
        if !$value && m,^[Eoq]$,;
      my ($pri) = m,^[clmSLM]$, ? $exp : Mersenne::mersPriority($exp, $_, $value);
      my ($lmec) = $mersDB->lowestMissingExponent($_);
      if ($exp != $lmec && !$mersDB->reservations($lmec))
        {
          my ($lmecPri) = Mersenne::mersPriority($lmec, $_, 0);
          $exp = $lmec,
            $pri = $lmecPri
            if $pri <= 0.0 || ($pri > $lmecPri && $lmecPri >= 0.0);
        }
      print("$exp $_\n"),
        next
        unless $best;
      %choice = ('exp' => $exp, 'letter' => $_, 'pri' => $pri),
        next
        unless exists $choice{'exp'};
      if (m,^[clmSLM]$, || $choice{'letter'} =~ m,^[clmSLM]$,)
        { # Work on smaller Mersenne first.
          next if $exp > $choice{'exp'};
          %choice = ('exp' => $exp, 'letter' => $_, 'pri' => $pri),
            next
            if $exp < $choice{'exp'};
          # Same Mersenne - prefer clmSLM work over Eoq work.
          next if $choice{'letter'} =~ m,^[clmLMS]$,;
          %choice = ('exp' => $exp, 'letter' => $_, 'pri' => $pri);
          next;
        }
      # Both are Eoq - choose based on priority (lower values first).
      next
        if $pri > $choice{'pri'};
      %choice = ('exp' => $exp, 'letter' => $_, 'pri' => $pri),
        next
        if $pri < $choice{'pri'};
      # Same priority - prefer q over o (since doP+-1.pl does)
      #   and o over E (since qo only needs to be done once at a given bound).
      %choice = ('exp' => $exp, 'letter' => $_, 'pri' => $pri),
        next
        if $_ eq 'q';
      %choice = ('exp' => $exp, 'letter' => $_, 'pri' => $pri),
        next
        if $_ eq 'o' && $choice{'letter'} ne 'q';
    }
  print("$choice{'exp'} $choice{'letter'}\n")
    if exists $choice{'exp'};
  exit(0);
}

if (length($n) > 1 && !defined($letters))
{
  die "first argument must be numeric or a mersfmt letter with a few exceptions\n$usage"
    unless scalar(grep { m,^$n$, } $mersDB->dbTables());
  # Currently 'done', 'reservations', 'aliases', 'allowed', and 'contact'.
  #   Only 'done' contains Mersenne data.
  my (@rows) = $mersDB->fetchRows($n);
  if ($n eq 'done')
    {
      if (!$valueOnly)
        {
          print("M( $_->[0] )$_->[1]\n")
            foreach (@rows);
          exit(0);
        }
      print(join(($valueOnly eq 'csv' ? ',' : ' '), $_->[0], $_->[1]) . "\n")
        foreach (@rows);
    }
  elsif (!$valueOnly)
    {
      print("# Table $n - columns '" . join("' '", @{$mersDB->columns($n)}) . "'\n");
      print("# " . join(' ', @{$_}) . "\n")
        foreach (@rows);
    }
  elsif ($valueOnly eq 'csv')
    {
      print("$n." . join(",$n.", @{$mersDB->columns($n)}) . "\n");
      print(join(',', @{$_}) . "\n")
        foreach (@rows);
    }
  else
    {
      print(join(' ', @{$_}) . "\n")
        foreach (@rows);
    }
  exit(0);
}

!defined($letters)
  || ($n =~ m/$bigintPattern/ && $letters =~ m/^([$MDBvL]+)$/)
  || ($n eq 'missing' && $letters =~ m/^([EUcoq]+)$/)
  || ($n !~ m/$bigintPattern/ && $letters =~ m/$bigintPattern/)
  or die "unexpected arguments\n$usage";
$letters = $1
  if defined $letters;

if ($n =~ m,^([EoqclmSLM])$,)
{ # ECM, P-1, and P+1 data have priorities ...
  $n = $1;
  $pri = 0.0 unless defined($pri);
  $letters = 1 unless defined($letters);
  #   ... and clmSLM data uses the exponent as the priority.
  my (@cols) = @{$mersDB->columns($n)};
  my (@zeros);
  $zeros[$_] = 0
    foreach (0 .. ($#cols - 1));
  my ($low) = $mersDB->lowestMissingExponent($n);
  my (@reservations) = $mersDB->reservations($low);
  $low = undef
    if @reservations;
  my ($lowData) = $mersDB->find($low, $n)
    if $low;
  my ($rows) = $mersDB->fetchRowsByPriority($n, $pri, $letters);
  die("MersDB::fetchRowsByPriority failed - " . $mersDB->dbError())
    unless defined $rows;
  my ($lowPri); # priority of $lowData per Mersenne::mersPriority()
  if ($low && $lowData && $lowData =~ m,[)]"$n": ,)
    { # The Mersenne has $n (Eoq, usually) data but not in the db
      #   (or there's a bug related to updating the db's missing table)
      #   so only print the data from fetchRowsByPriority().
      $lowData = '';
    }
  elsif ($low && $lowData)
    { # This can only happen for lmSLM - MersDB::find() will return c data
      #   in that case - "M( 4401 )c: 4401\n".
      $lowPri = $low;
    }
  elsif ($low && $n =~ m,^[Eoq]$,)
    {
      $lowPri = Mersenne::mersPriority($low, $n, 0, 0);
      $lowData = "M( $low )$n: 0 0 $lowPri\n";
    }
  elsif ($low && $n =~ m,^[clmSLM]$, && $low % 8 != 4)
    {
      $lowData = "M( $low )c: $low\n";
      $lowPri = Mersenne::mersPriority($low, 'c', 0, 0);
    }
  elsif ($low && $n =~ m,^[clmSLM]$, && $low % 8 == 4)
    {
      $lowData = "M( $low )l: $low\nM( $low )m: $low\n";
      $lowPri = Mersenne::mersPriority($low, 'l', 0, 0);
    }
  if ($low && $lowData && !@{$rows})
    {
      print("$n." . join(",$n.", @cols) . "\n"
            . "$low," . join(',', @zeros) . "\n"),
        exit(0)
        if $valueOnly eq 'csv';
      print("$low\n"),
        exit(0)
        if $valueOnly;
      print("$lowData");
      exit(0);
    }
  while (@{$rows})
    {
      my ($row) = shift(@{$rows});
      # Mersennes with larger exponents may not have any data.
      next
        if $#{$row} < 1 && $row->[0] > $mersDB->lowestMissingExponent('c');
      print("$n." . join(",$n.", @cols) . "\n"),
        @cols = ()
        if @cols && $valueOnly eq 'csv';
      if ($lowData && $lowPri < $row->[-1])
        {
          print("$lowData")
            if !$valueOnly;
          print("$low" . ($valueOnly eq 'csv' ? (',' . join(',', @zeros)) : '') . "\n")
            if $valueOnly;
          $lowData = '';
          last
            unless @{$rows};
          pop(@{$rows});
        }
      if (!$valueOnly)
        {
          print("M( " . shift(@{$row}) . " )$n: ");
          shift(@{$row})
            if $row->[0] eq $n;
          print(join(' ', @{$row}) . "\n");
          next;
        }
      print(join(',', @{$row}) . "\n"),
        next
        if $valueOnly eq 'csv';
      print("$row->[0]\n");
    }
  exit(0);
}

die "unexpected arguments\n$usage";
