#include "version.h"

#define DO_NOT_USE_LONG_DOUBLE
#include "isprime.h"

const char *program_name = NULL;

/* Return 0 iff n is composite, setting the second argument to any factor that is found. */
/*!!Should have more arguments so that "witnesses" can be returned. */
int isPrime(FACTOR n, FACTOR *factor)
 {
  FACTOR a, b, k, q, r, one, tmp, nMinusK, aNmK, binNKN;
  FACTOR *polyXNA, *polyXAN, *polyTmp;
  double logN; /* logBase2(n) */
  mpf_t sqFourLogN; /* square of (4*logBase2(n)) */
  mpf_t rSqFourLogN; /* r times prior */
  mpf_t minR; /* minimum r for which q might be suitable */
  mpf_t tmpDbl; /* temporary, usually converted from double or mpz_t */
  unsigned long polyIndex; /* index into polynomials in x */
  size_t intLogN;
  int qIsPrime; /* is the current value of q prime? 1 is probable prime; 2 is definitely prime */
  unsigned long intR; /* number of coeff. in polynomials mod (x^r - 1) */
  unsigned long nModR; /* n % r */

  assert(factor != NULL);
  fac_set_int((UL)0, factor);
  if (fac_less_than_int(n, (UL)2))
    return(COMPOSITE);
  if (fac_less_than_int(n, (UL)4))
    return(PRIME);
  /* Step 2 */
  fac_init_var(r);
  fac_set_int((UL)2, &r);
  /* Steps 3-10 for r == 2, replacing GCD with modulo: */
  fac_init_var(b);
  mpz_mod(b, n, r);
  if (fac_less_than_int(b, (UL)1))
   {
    fac_set(r, factor);
    return(COMPOSITE);
   }
  /* Step 1 */
  if (mpz_perfect_power_p(n))
    return(COMPOSITE);
  fac_set_int((UL)3, &r);
  /* Initialization */
  fac_init_var(a);
  fac_init_var(q);
  fac_init_var(tmp);
  /* In step 7: q must be at least 4*sqrt(r)*logBase2(n) && ... */
  /*    So calculate sqFourLogN = square(4*logBase2(n)) for later use. */
  /*!!Yikes: no direct way to compute floating point logs nor floating point square roots ... */
  intLogN = mpz_sizeinbase(n, 2) - 1;
  fac_set(n, &tmp);
  while (!mpz_fits_ulong_p(tmp))
    mpz_tdiv_q_2exp(tmp, tmp, 1); /*!!1 is arbitrary but most precise later and most expensive here */
  logN = log2((double)mpz_get_ui(tmp));
  logN = logN - trunc(logN) + (double)intLogN;
  mpf_init2(sqFourLogN, 100); /*!!this much precision is probably overkill */
  mpf_set_d(sqFourLogN, logN);
  mpf_pow_ui(sqFourLogN, sqFourLogN, 2);
  mpf_mul_ui(sqFourLogN, sqFourLogN, 16);
  /* In step 7: q must be at least 4*sqrt(r)*logBase2(n) && ... */
  /* So (r - 1)/2 >= q >= 4*sqrt(r)*logBase2(n) */
  /*    r - 1 >= 8*sqrt(r)*logBase2(n) */
  /*    r^2 - 2*r + 1 >= 64*r*(logBase2(n))^2 */
  /*    r - 2 + 1/r >= 64*(logBase2(n))^2 */
  /*    r - 2 + 1 >= 64*(logBase2(n))^2 */
  /*    r > 64*(logBase2(n))^2 */
  mpf_init2(minR, 100);
  mpf_init2(tmpDbl, 100);
  mpf_set_d(minR, 64.0);
  mpf_set_d(tmpDbl, logN);
  mpf_mul(minR, minR, tmpDbl);
  mpf_mul(minR, minR, tmpDbl);
  mpf_set_z(tmpDbl, n);
  mpf_sqrt(tmpDbl, tmpDbl);
  if (mpf_cmp(minR, tmpDbl) > 0) /* if minR > sqrt(n) ... */
    mpf_set(minR, tmpDbl); /* ... then minR = sqrt(n) */
  /* Step 3: while (r < n), r too small for q to matter */
  mpf_set_z(tmpDbl, r);
  while (mpf_cmp(tmpDbl, minR) <= 0)
   { /* Step 4: if GCD(n, r) != 1, output COMPOSITE; replace GCD with modulo */
    mpz_mod(b, n, r);
    if (fac_less_than_int(b, (UL)1))
     {
      fac_set(r, factor);
      return(COMPOSITE);
     }
    mpz_nextprime(r, r);
    mpf_set_z(tmpDbl, r);
   }
  mpz_pow_ui(a, r, 2);
  if (fac_less_than(n, a))
    return(PRIME);
  mpf_init2(rSqFourLogN, 100);
  fac_init_var(one);
  fac_set_int((UL)1, &one);
  /* Step 3 while (r < n), r big enough that q might exist */
  for ( ; fac_less_than(r, n); mpz_nextprime(r, r))
   {
    /* Step 4: replace GCD with modulo */
    mpz_mod(b, n, r);
    if (fac_less_than_int(b, (UL)1))
     {
      fac_set(r, factor);
      return(COMPOSITE);
     }
    /* Step 5: if (r is prime) */
    if (!isPrime(r, &tmp))
     {
      fac_add_int(r, (UL)1, &r);
      continue;
     }
    /* Step 6: q becomes the largest prime factor of (r - 1) */
    fac_sub_int(r, (UL)1, &b);
    fac_set_int((UL)2, &a);
    /* If (r - 1)/2 is (a Sophie-Germain) prime, we can skip trying to factor it. */
    qIsPrime = (mpz_remove(q, b, a) == 1);
    if (qIsPrime != 0 && (qIsPrime = mpz_probab_prime_p(q, 5)) == 1)
      /* merely "probably prime": make certain by calling this recursively */
      if (!(qIsPrime = 2*isPrime(q, &tmp)) && !fac_less_than_int(tmp, (UL)1))
       { /* Found a factor; pull its powers out */
        assert(mpz_remove(q, q, tmp) > 0);
        /* can't pull it out completely, as it _might_ be the largest prime factor */
        mpz_mul(q, q, tmp);
       }
    /* Step 7: q must be at least 4*sqrt(r)*logBase2(n) && ... */
    /*    So q^2 must be at least r*square(4*logBase2(n)) to be of use. */
    mpf_set_z(tmpDbl, r);
    mpf_mul(rSqFourLogN, tmpDbl, sqFourLogN);
    mpf_set_z(tmpDbl, q);
    mpf_pow_ui(tmpDbl, tmpDbl, 2);
    fac_add_int(a, (UL)1, &a);
    if (!qIsPrime && mpf_cmp(tmpDbl, rSqFourLogN) > 0)
     { /* q is composite but big enough, so factor q */
      for (mpz_pow_ui(b, a, 2);
           !qIsPrime && fac_less_than(b, q) && mpf_cmp(tmpDbl, rSqFourLogN) > 0;
           mpz_pow_ui(b, a, 2))
       {
        (void)mpz_remove(b, q, a);
        if (fac_less_than_int(b, (UL)2))
         {
          fac_set(a, &q);
          qIsPrime = 1;
         }
        else
         {
          fac_set(b, &q);
          if ((qIsPrime = mpz_probab_prime_p(q, 5)) == 1)
            if (!(qIsPrime = 2*isPrime(q, &tmp)) && !fac_less_than_int(tmp, (UL)1))
             { /* Found a factor; pull its powers out */
              assert(mpz_remove(q, q, tmp) > 0);
              mpz_mul(q, q, tmp);
             }
         }
        mpf_set_z(tmpDbl, q);
        mpf_pow_ui(tmpDbl, tmpDbl, 2);
        mpz_nextprime(a, a);
       }
      if (qIsPrime != 2)
        qIsPrime = 2*!fac_less_than(b, q);
     }
    /* Step 7, second half: if n^((r-1)/q) != 1 mod r */
    if (!qIsPrime)
      continue;
    mpz_fdiv_qr(tmp, b, r, q); /* tmp = r/q; b = r % q */
    assert(mpz_cmp(b, one) == 0);
    mpz_powm(b, n, tmp, r); /* b = (n^tmp) % r */
    if (mpz_cmp(b, one) != 0)
      /* Step 8 */
      break;
    /* Step 9: increment r, replaced by r becomes next largest prime */
   }
  /* Step 10: end of loop incrementing r */
  /* Step 11: for a = 1 to 2*sqrt(r)*logBase2(n) */
  fac_set_int((UL)1, &a);
  /* Use minR to store a's high limit */
  mpf_set_z(minR, r);
  mpf_sqrt(minR, minR);
  mpf_mul_ui(minR, minR, 2);
  mpf_set_d(tmpDbl, logN);
  mpf_mul(minR, minR, tmpDbl);
  /* Need several temporaries */
  fac_init_var(k);
  fac_init_var(nMinusK);
  fac_init_var(aNmK);
  fac_init_var(binNKN);
  if (!mpz_fits_ulong_p(r))
   {
    fprintf(stderr, "%s: r >= 2^32; no memory for polynomials in step 12\n",
            program_name);
    exit(2);
   }
  intR = mpz_get_ui(r);
  polyXNA = (FACTOR *)calloc(intR, sizeof(FACTOR));
  polyXAN = (FACTOR *)calloc(intR, sizeof(FACTOR));
  polyTmp = (FACTOR *)calloc(intR, sizeof(FACTOR));
  if (polyXNA == NULL || polyXAN == NULL || polyTmp == NULL)
   {
    fprintf(stderr, "%s: calloc() returned NULL; no memory for polynomials in step 12\n",
            program_name);
    exit(2);
   }
  for (polyIndex = intR; polyIndex-- > 0; )
   {
    fac_init_var(polyXNA[polyIndex]);
    fac_init_var(polyXAN[polyIndex]);
    fac_init_var(polyTmp[polyIndex]);
   }
  nModR = mpz_fdiv_ui(n, intR); /* = n % r */
  for (mpf_set_z(tmpDbl, a);
       mpf_cmp(tmpDbl, minR);
       fac_add_int(a, (UL)1, &a), mpf_set_z(tmpDbl, a))
   {
    /* Step 12: if ((x - a)^n != (x^n - a)) mod (x^r - 1, n), n is COMPOSITE */
    mpz_neg(q, a); /* q = -a */
    mpz_fdiv_r(q, q, n); /* q %= n */
    /* (x - a) % (x^r - 1) == (x + q), so starting value of polyXAN is easy: */
    fac_set(q, polyXAN + 0);
    fac_set_int((UL)1, polyXAN + 1);
    /* Now, raise polyXAN to the n'th power, mod'ing by (x^r - 1) and n along the way */
    for (intLogN++; intLogN > 0; ) /* originally, the highest bit set in n */
     {
      if (mpz_tstbit(n, --intLogN))
       {
        polySqModMod(polyXAN, intR, n, polyTmp);
        polyMultiplyModMod(polyTmp, intR, n, polyXAN, polyXNA);
       }
      else
        polySqModMod(polyXAN, intR, n, polyXNA);
      /* copy polyXNA (result) back into polyXAN */
      for (polyIndex = intR; polyIndex-- > 0; )
        fac_set(polyXNA[polyIndex], polyXAN + polyIndex);
     }
    /* (x^n - a) % (x^r - 1) == (x^(n % r) + q), so polyXNA is trivial but must ... */
    /* ... be init'd from scratch since it was used as a temporary above. */
    fac_set(q, polyXNA + 0);
    for (polyIndex = intR; polyIndex-- > 1; )
      fac_set_int((UL)0, polyXNA + polyIndex);
    fac_set_int((UL)1, polyXNA + nModR);
    /* Finally, compare polyXNA with polyXAN */
    for (polyIndex = intR; polyIndex-- > 0; )
      if (mpz_cmp(polyXAN[polyIndex], polyXNA[polyIndex]) != 0)
        return(COMPOSITE);
   }
  return(PRIME);
 }

/* Square polyInX modulo (x^r - 1) modulo n, storing result in poly2 */
void polySqModMod(FACTOR *polyInX, unsigned long r, FACTOR n, FACTOR *poly2)
 { /*!!Could be faster, but I don't know that algorithm off the top of my head */
  assert(r % 2 == 1);
  polyMultiplyModMod(polyInX, r, n, polyInX, poly2);
 }

/* Multiply polyInX by poly2 modulo (x^r - 1) modulo n, storing result in poly3 */
void polyMultiplyModMod(FACTOR *polyInX, unsigned long r, FACTOR n, FACTOR *poly2, FACTOR *poly3)
 {
  unsigned long y, z, a;
  FACTOR tmp;

  assert(r % 2 == 1);
  fac_init_var(tmp);
  /* This first loop also init's all of poly3, since r must be odd */
  for (y = a = 0; y < r; y++, a = (2*y) % r)
    fac_mul(polyInX[y], poly2[y], poly3 + a);
  for (y = 0; y < r - 1; y++)
    for (z = y + 1; z < r; z++)
     {
      a = (y + z) % r;
      fac_mul(polyInX[y], poly2[z], &tmp);
      fac_mul_int(tmp, (UL)2, &tmp);
      fac_add(tmp, poly3[a], poly3 + a);
     }
  for (y = 0; y < r; y++)
    mpz_fdiv_r(poly3[y], poly3[y], n);
 }

#ifdef ISPRIMEMAIN

# ifdef macintosh
#  include <console.h>
# endif

FACTOR savef, numer;

int main(int argc, char **argv)
 {
  FACTOR n, factor;

# ifdef macintosh
  argc = ccomand(&argv);
# endif  
  if (argv == NULL || argv[0] == NULL)
    program_name = "isprimegmp";
  else
    for (program_name = argv[0] + strlen(argv[0]); program_name > argv[0] && *program_name != '/'; )
      --program_name;
  if (*program_name == '/')
    ++program_name;
  setup();
  /* init the math package and variables */
  fac_init();
  fac_init_var(n);
  fac_init_var(factor);
  fac_init_var(savef); /* these two are not actually used */
  fac_init_var(numer);
  for (shouldTerminate = !fscan_fac(stdin, &n);
       !shouldTerminate;
       shouldTerminate = shouldTerminate || !fscan_fac(stdin, &n))
    if (isPrime(n, &factor))
     {
      fprintf(stdout, "Prime ");
      fprint_fac(stdout, n);
      fprintf(stdout, "\n");
     }
    else
     {
      fprintf(stdout, "Composite ");
      fprint_fac(stdout, n);
      if (!fac_less_than_int(factor, (UL)1))
       {
        fprintf(stdout, "\nhas factor ");
        fprint_fac(stdout, factor);
       }
      fprintf(stdout, "\n");
     }
  return(0);
 }
#endif
