#!/usr/bin/perl -Tw
# Reserve an exponent in the db based on the command line.
#   Mostly for the make that updates the db and, the results directory,
#   and other flat files - that makefile is not in this package.

use strict;
use strict 'vars';
use warnings;

BEGIN { push(@INC, '.'); $ENV{'PATH'} = '/usr/bin:/bin'; }

my ($usage) = "reserve.pl [-e] <exponent> [-u] <UUID> [-k] <kind>\n"
  . "reserve.pl <UUID> <kind> [file(s) or directory(ies)]\n"
  . "  <exponent> is a non-negative integer indicating the Mersenne number to reserve.\n"
  . "  <UUID> is the UUID or program or script name making the reservation.\n"
  . "  <kind> is a 'mers letter' (or a single digit).\n"
  . "  Any argument that exists as a file or directory is treated as one\n"
  . "    unless preceded by '-e', '-u', or '-k'.\n"
  . "Only exponents for which work is going to be done should be reserved.\n"
  . "An exponent of 0 is special and should only be used by mersServer.pl, runSPRP.pl,\n"
  . "  and the Makefile that updates the db and the flat files.\n"
  . "  Those UUIDs are 'mersServer', 'runSPRP', and 'make'.\n"
  . "  <kind> is '!' for those but does not have to be; only the exponent and UUID matter.\n"
  . "Trying to reserve exponents that are already reserved will fail.\n"
  . "  Failing to reserve an exponent part way thru a file will exit then,\n"
  . "    after reserving exponents earlier in the file as printed.\n"
  ;

use Mersenne;
use MersDB;

my ($version) = MersInfo::version();
print("$0 version $version\n"),
  exit(0)
  if scalar(grep { m/^--(version|help|check)$/ || $_ eq '-h' } @ARGV);

print("$usage"),
  exit(0)
  if scalar(grep { m/^-(h|-help)$/ } @ARGV);

print("$0 version $version\n"),
  exit(0)
  if scalar(grep { m/^--check$/ } @ARGV);

my ($bigintPattern) = MersInfo::patterns('bigint');
my ($uuidPattern) = MersInfo::patterns('uuid');

# This script is used by doNext.pl, which might do SPRP checks, so 'c' needs to be added here.
my ($okLetters) = "!?c$Mersenne::workLetters";

my ($exp, $uuid, $kind, @files, $checkConfig);
while (@ARGV)
{
  if ($ARGV[0] eq '-e')
    {
      shift();
      die("more than one exponent on the command line is not supported - '$exp' and '$ARGV[0]' given\n")
        if $exp;
      $exp = shift();
      die("exponent must be an integer > 9 - '$exp'")
        unless $exp =~ m,^([0-9]+)$,;
      next;
    }
  if ($ARGV[0] eq '-k')
    {
      shift();
      $kind = shift();
      $kind = $1,
        next
        if $kind =~ m,^([0-9])$,;
      die("kind must be an mers format work letter or '!' or '?' - '$kind'")
        unless $kind =~ m,^([$okLetters])$,;
      next;
    }
  if ($ARGV[0] eq '-u')
    {
      shift();
      $uuid = shift();
      die("unexpected characters in UUID - '$uuid'")
        unless $uuid =~ m,$uuidPattern,;
      next;
    }
  if ($#ARGV == 2)
    {
      die("only one exponent is supported except in files")
        if $exp;
      die("only one UUID is supported")
        if $uuid;
      die("only one kind (mers format work letter) is supported")
        if $kind;
      die("exponent must be integer - '$ARGV[0]'")
        unless $ARGV[0] =~ m,$bigintPattern,;
      $exp = $1;
      die("UUID includes unexpected characters - '$ARGV[1]'")
        unless $ARGV[1] =~ m,$uuidPattern,;
      $uuid = $1;
      die("unexpected kind - '$ARGV[2]'")
        unless $ARGV[2] =~ m,^([$okLetters])$,;
      $kind = $1;
      shift(); shift(); shift();
      last;
    }
  if ($ARGV[0] =~ m,^([0-9]+)$,)
    {
      die("only one exponent is supported except in files")
        if $exp;
      $exp = $1;
      shift();
      next;
    }
  if ($ARGV[0] =~ m,$uuidPattern,)
    {
      die("only one UUID is supported")
        if $uuid;
      $uuid = "$1";
      shift();
      next;
    }
  if ($ARGV[0] =~ m,^([$okLetters])$,)
    {
      die("only one kind (mers format work letter) is supported")
        if $kind;
      $kind = "$1";
      shift();
      next;
    }
  $checkConfig = 1,
    shift(),
    next
    if $ARGV[0] eq '--check';
  push(@files, shift()),
    next
    if -f $ARGV[0] && -s $ARGV[0];
  die("unexpected argument or option: '$ARGV[0]'");
}

die("missing exponent or file(s) or directory(ies)")
  unless defined($exp) || @files;
die("missing UUID")
  unless defined($uuid);
die("missing kind (mers format letter)")
  unless defined($kind);

exit(0)
  if $checkConfig;

my ($mersDB) = MersDB->new();
my ($dbh) = $mersDB->connect('reserve');

END
{
  my ($rc) = $dbh ? $dbh->disconnect() : 1;
  $rc or die("disconnect from database failed: " . $dbh->errstr . "\n");
}

my (@r) = $mersDB->update('contact', 'reserve', time());
@r && $r[0] or die("could not update 'contact' 'reserve', " . $mersDB->dbError() . "\n");

@r = $mersDB->update('contact', $uuid, time());
@r && $r[0] or die("could not update 'contact' '$uuid', " . $mersDB->dbError() . "\n");

if (defined($exp))
{
  my (@reservations) = $mersDB->reservations($exp);
  if (@reservations)
    {
      print("already reserved - $_->{'exp'} $_->{'uuid'} $_->{'kind'}\n")
        foreach (@reservations);
      exit(0);
    }
  @r = $mersDB->reserve($exp, $uuid, $kind);
  print("reserved $exp $uuid $kind\n"),
    exit(0)
    if @r && $r[0];
  print($mersDB->dbError()),
    exit(1)
    unless $mersDB->dbError() =~ m,Duplicate entry '$exp' for key ,;
}
while (@files)
{
  my ($file) = shift(@files);
  die("file names can only include letters, digits, and -/.,_ - '$file'")
    unless $file =~ m'^([-/._0-9a-zA-Z])$';
  $file = $1;
  die("soft links are not supported - '$file'")
    if -l $file;
  die("path with '/../' or starting '../' not supported - '$file'")
    if $file =~ m,^[.][.]/, || $file =~ m,/[.][.]/,;
  if (-d($file))
    {
      opendir(my $dh, "$file")
        or die "could not read directory '$file' - $!";
      my (@entries) = grep { !m,^[.][.]?$, } readdir($dh);
      closedir($dh)
        or die "could not close directory '$file' - $!";
      # Reverse sort so the list prepended to @files is sorted.
      unshift(@files, "$file/$_")
        foreach sort { $b <=> $a } (@entries);
      next;
    }
  die("path '$file' is not a regular file nor a directory\n")
    unless -f($file);
  my ($fh);
  open($fh, '<', $file)
    or die("could not read '$file' - $!");
  while(<$fh>)
    {
      m,^([0-9]+)$,
        or die "unexpected format in '$file' - '$_'";
      $exp = $1;
      $kind = '!' unless $kind;
      @r = $mersDB->reserve($exp, $uuid, $kind);
      print("reserved $exp $uuid $kind\n"),
        next
        if @r && $r[0];
      print($mersDB->dbError());
      exit(1);
    }
  close($fh)
    or die("could not close '$file' - $!");
}
exit(0);
