#!/usr/bin/perl -Tw
# Run SPRP checks on Mersenne cofactors based on the SQL 'sprp' table's needed column.
#   Usually run from mersServer.pl (as "internal" work).

use strict;
use strict 'vars';

BEGIN { push(@INC, '.'); $ENV{'PATH'} = "/usr/bin:/bin"; }

use Mersenne;
use MersDB;

my ($usage) = ("Usage: runSPRP.pl [-v] [--version] [--check] [-h] [--help]\n"
               . "  Run SPRP tests of the cofactors of the smallest Mersenne that\n"
               . "    has cofactors that might be prime according to the database.\n"
               . "  With -v, progress 'comments' will be printed.\n"
  );

my ($UUID) = 'runSPRP'; # for the 'reservations' and 'contact' tables in the db

# We do not read STDIN ...
close(STDIN)
  or die("could not close STDIN, $!\n");
# ... but perl -w complains if STDIN is opened only for writing.
open(STDIN, '<', '/dev/null')
  or die("could not open /dev/null for STDIN, $!\n");

print("$0 version " . MersInfo::version() . "\n"),
  exit(0)
  if scalar(grep { m/^--version$/ } @ARGV);

print($usage),
  exit(0)
  if scalar(grep { m/^-(h|-help)$/ } @ARGV);

# Find the SPRP executable, checking from least preferable to most preferable.
my (@sprp) = ('../mers/sprpgmp', '-C') if -x '../mers/sprpgmp' || -x '../mers/sprpgmp.exe';
@sprp = ('./mers/sprpgmp', '-C') if -x './mers/sprpgmp' || -x './mers/sprpgmp.exe';
@sprp = ('./sprpgmp', '-C') if -x 'sprpgmp' || -x 'sprpgmp.exe';
@sprp && -x $sprp[0]
  or die "cannot find any SPRP-capable executable (sprpgmp)\n";

-d('runSPRP.tmp') || mkdir('runSPRP.tmp', 0700)
  or die "could not create runSPRP.tmp subdir for temporary files, $!";

exit(0) # nothing more to check - no other options or any arguments are supported
  if scalar(grep { m/^--check$/ } @ARGV);

my ($verbose) = shift();
$verbose = '1'
  if defined($verbose) && $verbose eq '-v';

die("extra arguments or options given\n" . $usage)
  if @ARGV || (defined($verbose) && $verbose ne '1');

my ($output) = '';
END
{
  print($output) if $output;
}

$SIG{'HUP'} = "IGNORE";
foreach ('INT', 'TERM')
{
  $SIG{$_} = sub {
    my ($sig) = @_;
    $output .= "\n\n# caught SIG$sig\n\n";
  };
}

eval
{ # setpriority() is not always implemented
  setpriority(0, 0, 20);
};

my ($mersDB) = MersDB->new();
my ($dbh) = $mersDB->connect($UUID);
die("$0: could connect to the database\n")
  unless $dbh;

END
{
  my ($rc) = $dbh ? $dbh->disconnect() : 1;
  $rc or die("disconnect from database failed: " . $dbh->errstr . "\n");
}

# This is copied from mersServer.pl::init().
my (@dbResults) = $mersDB->fetchValues('missing', 'done', 'X');
die("unexpected results from MersDB::fetchValues() of missing, done, and X:\n"
    . '  ' . join(' ', @dbResults))
  if ($#dbResults != 3*3 - 1 # 'fetch', $table, $cnt for each of 'missing', 'done', 'X'
      || $dbResults[0] ne 'fetch' || $dbResults[1] ne 'missing'
      || $dbResults[3] ne 'fetch' || $dbResults[4] ne 'done'
      || $dbResults[6] ne 'fetch' || $dbResults[7] ne 'X'
      # missing cEoqU, done thru at least 4, and at least one X line (with value 0)
      || $dbResults[2] < 5 || $dbResults[5] < 3 || $dbResults[8] < 1);

print("# db results - " . join(' ', @dbResults) . "\n")
  if $verbose;

my ($specExp) = MersDB::specialExponent($UUID);
defined($specExp)
  or die "MersDB::specialExponent $UUID returned undef\n";

my (@r) = $mersDB->reservations($specExp);
print("# UUID $_->{'uuid'} has reserved exponent " . $specExp . "\n")
  foreach (@r);
die("... so cannot continue\n")
  if @r;

@r = $mersDB->reservations($UUID);
if (@r)
{ # If the last time this was run was prior to the last reboot, ignore the reservation.
  print("# runSPRP.pl already has reservations in the db; checking them.\n");
  print("# " . $_->{'exp'} . ' ' . $_->{'uuid'} . ' ' . $_->{'kind'} . "\n")
    foreach (@r);
  my ($fh);
  print("# /proc/uptime does not exist or could not be opened for reading, $!\n")
    unless -f('/proc/uptime') && open($fh, '<', '/proc/uptime');
  my ($line) = <$fh>
    if $fh;
  print("# /proc/uptime had unexpected data: '$line'\n")
    if $fh && $line !~ m,^([0-9]+),;
  # Presume no single SPRP check takes more than a day (so an lm check could take 2 days).
  my ($delay) = $1
    if defined $1;
  $delay = 2*24*60*60
    if !$delay or $delay > 2*24*60*60;
  !$fh
    or close($fh)
    or die("could not close /proc/uptime, $!");
  my ($reboot) = time() - $delay;
  my (@contacts) = $mersDB->queryRows('contact', $UUID);
  foreach (@contacts)
    {
      next unless $_->{'uuid'} eq $UUID; # safety check - should always be true
      next if $_->{'time'} < $reboot;
      die("at least one reservation was made since the last reboot or in the last 2 days\n");
    }
  # No 'contact' times or it/they are prior to the last reboot, so wipe any old reservations.
  foreach (@r)
    {
      my (@results) = $mersDB->release($_->{'exp'}, $_->{'uuid'}, $_->{'kind'});
      die("# could not release reservation M" . $_->{'exp'} . " UUID " . $_->{'uuid'}
          . " kind " . $_->{'kind'} . "\n"
          . $mersDB->dbError() . "\n")
        unless @results && $results[0];
    }
  @r = $mersDB->reservations($UUID);
  print("# existing reservations:\n")
    if @r;
  print("# " . $_->{'exp'} . ' ' . $_->{'uuid'} . ' ' . $_->{'kind'} . "\n")
    foreach (@r);
  die("there is still at least one reservation by '$UUID'\n")
    if @r;
}

@r = $mersDB->update('contact', $UUID, time());
die("updating db table contact for '$UUID' failed\n"
    . "  " . $mersDB->dbError() . "\n")
  unless @r && $r[0];
my ($reserved);
@r = $mersDB->reserve(1, $UUID, 'c');
die("could not reserve exponent 1 for 'c' (SPRP) work\n  " . $mersDB->dbError() . "\n")
  unless @r && $r[0];
$reserved = 1;

END
{
  if ($mersDB && $reserved)
    {
      @r = $mersDB->release(1, $UUID, 'c');
      die("could not release 1 '$UUID' 'c'\n"
          . "  " . $mersDB->dbError() . "\n")
        unless @r && $r[0];
    }
}

sub needsSPRP($)
{
  my ($m) = @_;
  my ($data) = $mersDB->find($m->n(), 'CSLMclm');
  die("$0: MersDB::find failed for M" . $m->n() . " CSLMclm\n"
      . "  " . $mersDB->dbError() . "\n")
    unless defined $data;
  die("parsing data from MersDB::find failed:\n" . $m->errors() . "\n"
      . $data)
    unless $m->parse($data);
  local $_;
  my ($need) = $mersDB->value('sprp', $m->n());
  if ($need && $need == 3)
    { # some factors might be composite - factor any that are
      my (@f) = $m->f();
      foreach (@f)
        {
          next if Math::Prime::Util::is_prime($_);
          print("# found composite factor M( " . $m->n() . ")C: $_\n")
            if $verbose;
          my ($newFactor);
          foreach $newFactor (Math::Prime::Util::factors($_))
            {
              print"M( " . m->n() . " )C: $newFactor\n";
              $m->f($newFactor);
            }
        }
    }
  $need = $m->needsSPRP();
  return(1)
    if $need;
  my (@r) = $mersDB->update('sprp', $m->n(), 0);
  die("could not update M" . $m->n() . " 'sprp' to 0\n"
      . "  " . $mersDB->dbError() . "\n")
    unless @r && $r[0];
  print("# updated M" . $m->n() . " sprp to 0\n")
    if $verbose;
  return(0);
}

my ($e, $m); # exponent, Mersenne instance
# Loop until an SPRP check is needed.
while (defined($e = $mersDB->next('sprp')))
{ # 'D'one, 'S'PRP, or 'P'rime implies no SPRP check is needed.
  $e =~ m/^([1-9][0-9]*)$/
    or die "MersDB::next() for 'sprp' returned non-numeric value '$e'\n";
  $e = Math::BigInt->new("$1");
  $e = $mersDB->lowestMissingExponent('c')
    if $e > $mersDB->lowestMissingExponent('c');
  $mersDB->remove('sprp', $e),
    next
    if defined $mersDB->value('done', $e);
  $m = Mersenne->new('n', $e);
  last
    if needsSPRP($m);
}

# Found a Mersenne cofactor (or two) that needs an SPRP check ...
#   ... or there is no such Mersenne in the db.
# If no such Mersenne, check the lowest Mersenne with no SPRP data in the db.
#   And, if the Mersenne from the db is larger than that one, check that one instead.
if (!defined($e) || $e > $mersDB->lowestMissingExponent('c'))
{
  $e = $mersDB->lowestMissingExponent('c');
  defined($e)
    or die "MersDB::lowestMissingExponent(c) returned undef\n";
  $e =~ m/^([1-9][0-9]*)$/
    or die "MersDB::lowestMissingExponent(c) returned non-numeric value '$e'\n";
  $e = Math::BigInt->new("$1");
  # But don't check it if it is way ahead of the other lowest missing exponents.
  print("# lowest missing 'c' exponent is far ahead of other missing exponents\n"),
    exit(0)
    unless scalar(grep { $e <= 10*$mersDB->lowestMissingExponent($_) } ('E', 'o', 'q'));
  $m = Mersenne->new('n', $e)
    if !defined $m || $m->n() != $e;
  needsSPRP($m)
    or exit(0);
}

# Print any factors "discovered" by Mersenne::f().
print("M( $e )C: $_\n")
  foreach ($m->f());

# Run the SPRP check(s) of M($e)/$m after reserving it.
@r = $mersDB->reservations($e);
foreach (@r)
{ # All other work depends on the SPRP check being done,
  #   so ignore all other reservations.
  next unless $_->{'uuid'} eq $UUID;
  print("# M $e already reserved for SPRP check\n");
  exit(0);
}
if (!@r)
{ # Can't reserve if already reserved.
  @r = $mersDB->reserve($e, $UUID, 'c');
  print("# could not reserve exponent $e - " . join(' ', @r) . "\n"),
    exit(0)
    if $#r < 0 || !$r[0];
  $reserved = $e;
}

END
{
  if ($mersDB && ref($mersDB) eq 'MersDB' && defined($e) && $reserved == $e)
    {
      print("# could not save M$e to the db: " . $mersDB->dbError() . "\n")
        if $m && !$mersDB->save($m);
      my (@r) = $mersDB->release($e, $UUID, 'c');
      print("# could not release exponent $e\n  " . $mersDB->dbError() . "\n")
        unless @r && $#r >= 0;
    }
  ! -f("runSPRP.tmp/$$.in")
    || unlink("runSPRP.tmp/$$.in")
    or die "could not unlink runSPRP.tmp/$$.in, $!";
}

my ($fh);
open($fh, '>', "runSPRP.tmp/$$.in")
  or die "could not create runSPRP.tmp/$$.in, $!\n";
$m->print($fh)
  or die "could not print to runSPRP.tmp/$$.in, $!\n";
close($fh)
  or die "could not close runSPRP.tmp/$$.in, $!\n";
print("# running SPRP check for\n"),
  $m->print()
  if $verbose;

open($fh, '-|', @sprp, 'runSPRP.tmp/' . "$$" . '.in')
  or die "could not run " . join(' ', @sprp) . " for $e from runSPRP.tmp/$$.in, $!\n";
while (<$fh>)
{
  print;
  chomp;
  if (m/^M[(] ([1-9][0-9]*) [)]C: ([1-9][0-9]*)/)
    { # Any "factor" printed by the child needs to be checked since one case is:
      # M( 46991 )C: 130848429180565527354835801 # composite factor given
      # ... and the prime factors of this composite factor were not recorded.
      my ($n) = Math::BigInt->new("$1");
      $n == $m->n()
        or die join(' ', @sprp) . " runSPRP.tmp/$$.in for " . $m->n() . " printed results for $n: $_\n";
      my ($composite) = Math::BigInt->new("$2");
      my (@parts) = Math::Prime::Util::factor($composite);
      my (@factors) = $m->f(@parts);
      foreach (@factors)
        {
          print("M( $n )C: $_\n");
          shift(@parts) if $_ == $parts[0];
        }
      # Anything left should not be a factor, but ...
      print("M( $n )C: $_\n")
        foreach (@parts);
      next;
    }
  m/^M[(] ([1-9][0-9]*) [)]([clmSLM]): ([1-9][0-9]*)/
    or die "unexpected output from $sprp[0] for $e: $_\n";
  my ($n) = Math::BigInt->new("$1");
  my ($l) = $2;
  my ($len) = $3;
  $e == $n
    or die "$sprp[0] for $e printed results for $n: $_\n";
  next unless $l =~ m/^[clmSLM]$/;
  defined($len)
    or die "unexpected output from $sprp[0] for $e: $_\n";
  @r = $mersDB->update($l, $e, $l, $len);
  die("updating M$e $l $len failed, " . $mersDB->dbError() . "\n")
    unless @r && $r[0];
}
close($fh)
  or die join(' ', @sprp) . " runSPRP.tmp/$$.in for $e failed, errno $!, exited $?\n";

# An END block above releases the reservation and updates 'contact' ...
# ... but updating 'sprp' should only be done if the close() of $sprp succeeds.
my ($needed) = needsSPRP($m);
@r = $mersDB->update('sprp', $e, $needed);
print("# could not update M$e 'sprp' flag in the db: " . $mersDB->dbError())
  unless @r && $r[0];
print("# after running $sprp[0] for M$e, still needs SPRP check\n")
  if $needed;
exit($needed);
