#!/usr/bin/perl -Tw
# Save input mers format data to the mers database.
#   All (or no) factors for any particular Mersenne should be given together
#   in numeric order, but this fetches factors from the db as well when appropriate.
# Also initializes empty aliases and allowed tables in the db from flat files.

use strict;
use strict 'vars';

BEGIN { push(@INC, '.'); $ENV{'PATH'} = '/usr/bin:/bin'; }

use Mersenne;
use MersDB;
use ECMSave;

my ($usage) = ("save.pl [-f] [-q] [--version] [--check] [-h] [--help] [files]\n"
               . "  Save mers format data to the database from the given files or STDIN.\n"
               . "    Also, if the allowed or aliases SQL tables are empty, fill them from\n"
               . "      files names .server.allowed and .server.aliases.\n"
               . "  -f clears the allowed and aliases SQL tables and fills them from the\n"
               . "      .server.allowed and .server.aliases files.\n"
               . "  -q turns off the default output describing interactions with the database.\n");

print("$0 version " . MersInfo::version() . "\n"),
  exit(0)
  if scalar(grep { m/^--version$/ } @ARGV);

print($usage),
  exit(0)
  if scalar(grep { m/^-(h|-help)$/ } @ARGV);

my ($fill, $quiet, $checkConfig);
my (@files);
while (@ARGV)
{
  $fill = 1,
    shift,
    next
    if ($ARGV[0] eq '-f');
  $quiet = !$quiet,
    shift,
    next
    if ($ARGV[0] eq '-q');
  $checkConfig = 1,
    shift,
    next
    if $ARGV[0] eq '--check';
  die("$0: unknown option '$ARGV[0]' given\n$usage")
    if $ARGV[0] =~ m,^-,;
  die("$0: '$ARGV[0]' does not exist or is not a file")
    unless -e($ARGV[0]) && -f($ARGV[0]) && !-l($ARGV[0]);
  push(@files, shift());
}

exit(0)
  if $checkConfig;

my ($mersDB) = MersDB->new();
my ($dbh) = $mersDB->connect('save');
die("MersDB::connect failed to return a valid database handle")
  unless $dbh;

END
{
  my ($rc) = $dbh ? $dbh->disconnect() : 1;
  $rc or die("disconnect from database failed: " . $dbh->errstr . "\n");
}

if ($fill)
{ # MersDB doesn't support deleting all rows from a table to preclude accidental deletions.
  my ($ok) = $dbh->do('delete from aliases');
  die("deleting all from aliases SQL table failed - " . $mersDB->dbError())
    unless $ok;
  $ok = $dbh->do('delete from allowed');
  die("deleting all from allowed SQL table failed - " . $mersDB->dbError())
    unless $ok;
}

my ($fh);
@ARGV = @files;
my (@results) = $mersDB->fetchValues('missing', 'done', 'aliases', 'allowed', );
die("MersDB::fetchValues failed - " . $mersDB->dbError())
  unless @results && $results[0];
print(STDOUT '# ' . join(' ', $0, @ARGV, @results) . "\n")
  if @results && !$quiet;

@results = $mersDB->keyList('aliases');
print(STDOUT "# aliases - " . join(' ', $0, @ARGV, @results) . "\n")
  if @results && !$quiet;
if ($#results < 0 && -s('.server.aliases') && -f('.server.aliases') && !-l('.server.aliases'))
{ # Initialize the SQL 'aliases' table from the flat file .server.aliases.
  #!!require a checksum or signature of some kind?  the file to be read-only?
  open($fh, '<', '.server.aliases') or die "could not read .server.aliases, $!\n";
  while (<$fh>)
    { # IP address ('alias') and hostname or UUID.
      #   A particular IP address will often have both one UUID and at least one hostname.
      chomp;
      next unless $_ && !m/^ *#/;
      my (@line) = split(/ /);
      $#line == 1 or die ".server.aliases:$.: should have exactly one space - '$_'\n";
      $line[0] =~ m/^([.0-9]+)$/ or die ".server.aliases:$.: invalid IP address\n";
      $line[0] = "$1";
      $line[1] =~ m/^([-.0-9a-zA-Z]+)$/
        or die ".server.allowed:$.: unexpected characters in hostname or UUID - '$_'\n";
      $line[1] = "$1";
      # The columns in the SQL table are not in the same order as the file content.
      @results = $mersDB->update('aliases', $line[1], $line[0]);
      @results
        or die(".server.aliases:$.: updating 'aliases' in db failed for\n '$_'\n results: "
               . join(' ', @results) . "\n dbError " . $mersDB->dbError() . "\n");
      print(STDOUT "# update 'aliases' results " . join(' ', @results) . "\n")
        if @results && !$quiet;
    }
  close($fh) or die "could not close .server.aliases, $!\n";
}

@results = $mersDB->keyList('allowed');
print(STDOUT "# allowed - " . join(' ', $0, @ARGV, @results) . "\n")
  if @results && !$quiet;
if ($#results < 0 && -s('.server.allowed') && -f('.server.allowed') && !-l('.server.allowed'))
{ # Initialize the SQL 'allowed' table from the flat file .server.allowed.
  #!!require a checksum or signature of some kind?  the file to be read-only?
  open($fh, '<', '.server.allowed') or die "could not read .server.allowed, $!\n";
  while (<$fh>)
    { # Max task count, mers letters of allowed tasks, and IP or hostname or UUID.
      chomp;
      next unless $_ && !m/^ *#/;
      m/^[-. 0-9a-zA-Z]+$/ or die ".server.allowed:$.: unsupported characters\n";
      my (@line) = split(/ /);
      $#line == 2 or die ".server.allowed:$.: should have exactly two spaces - '$_'\n";
      $line[0] =~ m/^([1-9][0-9]*)$/ or die ".server.allowed:$.: non-numeric max tasks - '$_'\n";
      $line[0] = $1;
      $line[1] =~ m/^([$Mersenne::workLetters]+)$/
        or die ".server.allowed:$.: unexpected characters for allowed work types - '$_'\n";
      $line[1] = "$1";
      $line[2] =~ m/^([-.0-9a-zA-Z]+)$/
        or die ".server.allowed:$.: unexpected characters in hostname, UUID, or IP address - '$_'\n";
      $line[2] = "$1";
      # The columns in the SQL table are not in the same order as the file content.
      @results = $mersDB->update('allowed', $line[2], $line[1], $line[0]);
      @results
        or die(".server.allowed:$.: updating 'allowed' in db failed for\n '$_'\n results:"
               . join(' ', @results) . "\n dbError " . $mersDB->dbError() . "\n");
      print(STDOUT "# update 'allowed' results " . join(' ', @results) . "\n")
        if @results && !$quiet;
    }
  close($fh) or die "could not close .server.allowed, $!\n";
}

@results = ();

my ($m); # Mersenne instance - M(n)
my ($saveFile, $sFcontent); # file name and content of (Linux) ecm program save file
my ($saveExp); # Mersenne exponent related to the ecm program save file
while (<<>>)
{
  chomp;
  s/#.*//; # ignore comments starting with '#'
  s/[\t ][\t ]*$//; # drop trailing spaces and tabs
  next unless $_;
  $saveFile = $1,
    $saveExp = $2,
    next
    if (m/^save file begin (ecm[.]([1-9][0-9]+)[.]p[mp]1([.]M)?)$/);
  if (m/^save file end ([a-z][-._a-zA-Z0-9]+)$/)
    {
      die("'save file end $1' read but no prior 'save file begin'\n")
        unless defined $saveFile;
      die("save file end gives different file name, '$1', not expected '$saveFile'\n")
        unless $1 eq $saveFile;
      die("save file '$saveFile' has unexpected content\n")
        unless $sFcontent =~ m/^([-+=;:.@ A-Za-z0-9\r\n]+)$/;
      my ($fh);
      foreach my $topDir (grep { $_ ne '.' } (ECMSave::subdirs()))
        {
          next if ! -d $topDir || -l $topDir;
          $topDir .= '/';
          $topDir .= ($saveExp % $ECMSave::expMod) . "/"
            if (-d("$topDir" . ($saveExp % $ECMSave::expMod))
                || mkdir("$topDir" . ($saveExp % $ECMSave::expMod)));
          if (-s("$topDir$saveFile"))
            {
              my ($cnt) = 0;
              ++$cnt while -s "$topDir$saveFile.$cnt";
              $saveFile .= ".$cnt";
            }
          open($fh, '>', "$topDir$saveFile")
            or die("could not open save file $topDir$saveFile, $!");
          print($fh "$sFcontent")
            or die("could not write save file $topDir$saveFile, $!");
          close($fh)
            or die("could not close new save file $topDir$saveFile, $!");
          $saveFile = $saveExp = undef;
          $sFcontent = '';
          last;
        }
      next;
    }
  $sFcontent .= "$_\n", next if $saveFile;
  die("unexpected line or save file name: '$_'\n")
    if m/^save file /;
  my (@fields) = split(' ', $_);
  $fields[0] eq 'M(' or die 'bad input: ' . $_;
  $fields[1] =~ m/^([1-9][0-9]*)$/ or die 'invalid Mersenne exponent: ' . $_;
  $fields[1] = Math::BigInt->new("$1");
  $fields[2] =~ m'^[)]([cCDEeHlLmMoPqSUXJGg]):?$' or die 'unknown format or mers letter: ' . $_;
  next if $1 =~ /^[GJg]$/; # gaps and joins aren't stored in the db
  $fields[2] = $1;
  $fields[3] = Math::BigInt->new("$1") if $#fields > 2 && $fields[3] =~ m/^([0-9]+)$/;
  $fields[4] = Math::BigInt->new("$1") if $#fields > 3 && $fields[4] =~ m/^([0-9]+)$/;
  # values less than 1 for these mers letters is nothing - no work has been done
  next if $fields[2] =~ m/^[EoqHUclmeC]$/ && $fields[3] < 1;
  # factors must be larger than the exponent
  next if $fields[2] eq 'C' && $fields[3] <= $fields[1];
  # $fields[5] for ECM effort is optional (priority); the db calculates it
  pop(@fields) if $fields[2] eq 'E' && $#fields > 4;
  $#fields <= 4 or die 'too many columns: ' . $_;
  if (!defined($m) || $m->n() != $fields[1])
    {
      if ($m)
        {
          my (@saved) = $mersDB->save($m);
          die("$0: MersDB::save failed - " . $mersDB->dbError() . "\n"
            . "  after results " . join(' ', @results) . "\n")
            unless @saved && $saved[0];
          push(@results, @saved);
          print(STDOUT '# ' . join(' ', $0, $ARGV, @results) . "\n")
            if @results && !$quiet;
        }
      @results = ();
      $m = undef;
      if ($fields[1] <= $mersDB->lowestMissingExponent('c'))
        {
          $m = Mersenne->new('n', $fields[1]);
          my ($mers) = $mersDB->find($fields[1], 'C');
          $m->parse($mers)
            unless $mers =~ m/^[ \n\r]*$/;
        }
    }
  if ($fields[2] eq 'C')
    { # only save factors if the cofactor will be checked
      $m->f($fields[3]) if defined $m;
      next;
    }
  # 'e' needs prior 'E' data to update the db
  $m->parse($mersDB->find($m->get('n'), 'E'))
    if defined($m) && $fields[2] eq 'e' && !$m->get('E');
  $m->parse($_),
    next
    if defined($m);
  push(@results, $mersDB->update($fields[2], @fields));
  print(STDOUT '# ' . join(' ', $0, $ARGV, @results) . "\n"),
    @results = ()
    if @results && !$quiet;
}
exit(0)
  unless defined $m || @results;
push(@results, $mersDB->save($m))
  if defined $m;
die("MersDB::save failed - " . $mersDB->dbError() . "\n"
    . "  after results " . join(' ', @results) . "\n")
  unless @results && $results[0];
print(STDOUT '# ' . join(' ', $0, @results) . "\n")
  if !$quiet;
exit(0);
