#!/usr/bin/perl -Tw
# Send mers or GIMPS results from the argument file(s) to a server.

use strict;
use strict 'vars';
use Socket qw(:crlf);
use IO::Socket::INET;

my (@args);
push(@args, split(/ /, $_)) foreach (@ARGV);
@ARGV = @args;

BEGIN { push(@INC, '.') if -s('MersInfo.pm'); $ENV{'PATH'} = "/usr/bin:/bin"; }
use MersInfo;

my ($usage) = ("Usage: sendServer.pl [-h <host>] [-p <port>] [-[fFg][<value>] [-u <UUID>]\n"
               . "  [<host>] [<port>] [files] [--help|-h] [--verison] [--check]\n"
               . "Read the files (or STDIN) and send those results to the mers server or to GIMPS.\n"
               . "  -f and -F tell the mers server to free reservations of Mersennes with results unless <value> is 0.\n"
               . "  -g only sends results to GIMPS unless <value> is 0.\n"
               . "  -h specifies the host (or IP address) of the mers server.\n"
               . "  -p specifies the (TCP) port of the mers server.\n"
               . "All options except --help, --version, and --check must be before any argument without an option.\n"
  );

my ($check) = 0;
if (grep { m/^--(check|version|help)$/ } @ARGV)
{
  my ($version) = MersInfo::version();
  print(STDOUT "$0 version $version\n");
  print(STDOUT "$usage")
    if grep { m/^--help$/ } @ARGV;
  exit(0)
    if grep { m/^--(version|help)$/ } @ARGV;
  $check = 1;
}

open(STDERR, ">&STDOUT")
  or die "could not redirect STDERR to STDOUT, $!";

my ($socket, $port, $ipaddr, $saddr, $proto, $e, $e2, $host, $opt, $from, $uuid);
my ($free) = 'free';
my ($gimpsOnly) = 0;
my ($verbose) = 0;
while ($#ARGV >= 0 && $ARGV[0] =~ m/^-/) {
  ($opt = shift()) =~ s/^-(.)(.*)//;
  next
    if "$1$2" eq '-check';
  ($free = ("$2" ne '') ? "$2" : 'free'), next if $1 eq 'f';
  ($free = ("$2" ne '') ? "$2" : 'Free'), next if $1 eq 'F';
  ($gimpsOnly = ("$2" ne '') ? "$2" : shift), next if $1 eq 'g';
  ($port = ("$2" ne '') ? "$2" : shift), next if $1 eq 'p';
  ($uuid = ("$2" ne '') ? "$2" : shift), next if $1 eq 'u';
  ($verbose = ("$2" ne '') ? "$2" : shift), next if $1 eq 'v';
  if ($1 eq 'h')
    {
      ($host = ("$2" ne '') ? "$2" : shift);
      print(STDOUT "$usage"),
        exit(0)
        unless $host;
      next;
    }
  die("unknown option '-$1'\n$usage");
}

exit(0) if $check;

if (!$gimpsOnly)
{
  $host = 'localhost' if !$host;
  $host =~ m,^([-.A-Za-z0-9]+)$,
    or die "unexpected host '$host'\n";
  $host = "$1";
  $port = 'slucas' if !$port;
  $port = $1 if $port =~ m/(\d+)/;
  $port .= '(2203)' if $port !~ m/(\d+)/;
  $socket = IO::Socket::INET->new('PeerAddr' => "$host" . ':' . "$port",
                                  'Proto' => 'tcp',
                                  'Type' => SOCK_STREAM,
    );
  $socket
    or die "connecting to $host port $port failed, $!";
  !$uuid || print($socket 'UUID ' . $uuid . "\n")
    or die "printing UUID $uuid to $host port $port failed, $!";
  !$verbose || print(STDOUT "verbose connected to $host port $port\n")
    or die "printing verbose output failed, $!";
}

my ($gimpsResults) = '';
my ($saveFile);
my ($sFcontent) = '';
while (<<>>)
{ # Support results.txt format of Prime95/mprime and mfaktc (GPU trial factorer).
  chomp;
  $gimpsResults .= "$_$CRLF"
    if $ARGV =~ m/gimps/i || $gimpsOnly;
  next
    if $gimpsOnly;
  if (m/^(([Ff]ree)|([Rr]enew)) ([1-9][0-9]*)$/)
    {
      $e2 = undef;
      print($socket "$1 $4$CRLF") or die "printing '$1 $4' to $host port $port failed, $!";
      print(STDOUT "verbose sent $1 $4\n") if $verbose > 5 && ! $gimpsOnly;
      next;
    }
  $saveFile = $1,
    next
    if (m/^save file begin (ecm[.]([1-9][0-9]+)[.]p[mp]1([.]M)?)$/);
  if (m/^save file end ([a-z][-._a-zA-Z0-9]+)$/)
    {
      die("'save file end $1' read but no prior 'save file begin'\n")
        unless defined $saveFile;
      die("save file end gives different file name, '$1', not expected '$saveFile'\n")
        unless $1 eq $saveFile;
      die("save file '$saveFile' has unexpected content\n")
        unless $sFcontent =~ m/^([-+=;:.@ A-Za-z0-9$CRLF]+)$/;
      print($socket "save file begin $saveFile$CRLF", "$sFcontent", "save file end $saveFile$CRLF")
        or die("could not print save file $saveFile to $host port $port, $!");
      $saveFile = undef;
      $sFcontent = '';
      next;
    }
  $sFcontent .= "$_$CRLF",
    next
    if $saveFile;
  die("unexpected line or save file name: '$_'\n")
    if m/^save file /;
  my ($user, $computer) = ($1, $2)
    if s.UID: ([A-Za-z0-9]+)/([A-Za-z0-9]+), ..o;
  my ($rest);
  $e = $1, $rest = $2
    if m/^M[(] *([1-9][0-9]*) *([)].*)$/o;
  $e = $1, $rest = " )J: 2^$2 2^$3 # $4"
    if m/^no factor for M([1-9][0-9]+) from 2\^([1-9][0-9]+) to 2\^([1-9][0-9]+) (.*)$/o;
  $e = $1, $rest = " )C: $2 # $3"
    if m/^M([1-9][0-9]+) has a factor: ([1-9][0-9]+) (.*)$/o;
  $e = $2, $rest = " )J: 2^$3 2^$4 # $1 $5"
    if m/^found ([1-9][0-9]+ factors?) for M([1-9][0-9]+) from 2\^([1-9][0-9]+) to 2\^([1-9][0-9]+) (.*)$/o;
  next
    unless defined($rest);
  $rest .= " # GIMPS user $user computer $computer"
    if defined($user);
  if (defined($e2) && $e != $e2 && $free =~ m/^[Ff]ree$/)
    {
      print($socket "$free $e2$CRLF")
        or die "could not print '$free $e2' to $host port $port, $!";
    }
  $e2 = $e;
  print($socket "M( $e $rest$CRLF")
    or die "could not print 'M( $e $rest' to $host port $port, $!";
  print(STDOUT "verbose sent M( $e $rest\n")
    if $verbose > 4;
}

my ($problem);
if (!$gimpsOnly)
{
  $problem = 1;
  if (defined($e) && $free =~ m/^[Ff]ree$/)
    {
      print($socket "$free $e$CRLF")
        or die "could not print '$free $e' to $host port $port, $!";
    }
  $| = 1;
  print($socket "done$CRLF")
    or die "could not print 'done' to $host port $port, $!";
  my ($fdBits) = '';
  vec($fdBits, fileno($socket), 1) = 1;
  my ($fdCnt) = select($fdBits, undef, undef, 30);
  while ($fdCnt >= 0)
    {
      foreach (<$socket>)
        {
          $problem = 0,
            last
            if m/^done\b/i;
          print(STDOUT "$_")
            or die "could not print '$_' to STDOUT, $!";
        }
      # Special case: the server does not respond with 'done' if the client UUID starts with 'GIMPS-' or 'factordb-'.
      $problem = 0
        if !$! && $uuid && $uuid =~ m/^(factordb|GIMPS)-/;
      last
        if $problem == 0 || ($! && !($!{'EWOULDBLOCK'} || $!{'EAGAIN'} || $!{'EINTR'}));
      vec($fdBits, fileno($socket), 1) = 1;
      $fdCnt = select($fdBits, undef, undef, 30);
    }
  close($socket)
    or die "close of socket failed, $!";
}
print(STDOUT "no 'done' line from server\n")
  if $problem;
exit($problem)
  unless $gimpsResults ne '';

# Send $gimpsResults to mersenne.org.
use HTTP::Tiny;
my ($baseURL) = 'https://www.mersenne.org/manual_result/';
my ($gimpsUser, $gimpsPassword); # read from .gimps or gimps.txt
my ($fn) = -s('.gimps') ? '.gimps' : 'gimps.txt';
if (-s("$fn"))
{
  my ($fh);
  open($fh, '<', "$fn")
    or die "could not open $fn, $!\n";
  while (<$fh>)
    {
      chomp;
      $gimpsUser = $1 if m/^-u(.*)$/;
      $gimpsPassword = $1 if m/^-p(.*)$/;
    }
}
warn("cannot send results to mersenne.org without username and password from .gimps or gimps.txt\n"),
  exit(0)
  unless defined($gimpsUser) && defined($gimpsPassword);

# From $baseURL source, 27 Dec 2015:
# <p class="size4">Paste your results in the box below, or upload your results file (max 2MB).</p>
# <form method="post" enctype="multipart/form-data" action="/manual_result/" onSubmit="[...<submit condition>...]">
#  <input type="file" name="data_file" style="margin-right:20px;">
#  <button id="button1" style="width:200px;margin-bottom:10px;" type="submit">Submit</button><br>
#  <input type="hidden" name="was_logged_in_as" value="362">
#  <textarea autofocus placeholder="Paste your results here" name="data" style="width: 800px; height: 400px; background-color: lightyellow;" wrap="off"></textarea>
# </form>

# Since the hidden value for was_logged_in_as is needed, two HTTP interactions are required, the first solely to get that value.
my ($url) = $baseURL;
my ($http) = HTTP::Tiny->new();
my ($response) = $http->post_form($url, { 'Referer' => $baseURL, 'user_login' => $gimpsUser, 'user_password' => $gimpsPassword } );

my ($err) = "status $response->{'status'}, reason $response->{'reason'}; HTTP headers of response:\n";
print(STDOUT "verbose first response headers\n") if $verbose > 3 && $response->{'success'};
while (my ($k, $v) = each %{$response->{'headers'}})
{
  $err .= ("$k $_\n") for (ref $v eq 'ARRAY' ? @$v : $v);
  next unless $verbose > 3;
  print(STDOUT "verbose $k " . (ref $v eq 'ARRAY' ? @$v : $v) . "\n") if $response->{'success'};
}
if ($verbose > 3)
{
  print(STDOUT "verbose done with first response headers\n") if $response->{'success'};
  print(STDOUT "verbose $_ $response->{$_}\n") foreach grep { $_ ne 'content' } sort(keys(%{$response}));
  print(STDOUT "verbose done with first response\n");
}
print(STDOUT "verbose first response success is $response->{'success'}\n") if $verbose == 1;

$response->{'success'} or die "$url failed, $err";
length($response->{'content'}) or die "no HTML in response, $err";

# Was splitting this at \n but \n is in the middle of one of these patterns now.
my ($content) = $response->{'content'};
$content =~ s,\n, ,g; # change \n to space
$content =~ s,   *, ,g; # change multiple spaces to one
my ($loginValue, $resultsName);
foreach ($content)
{
  $loginValue = $1 if m,[<]input type[=]"hidden" name[=]"was_logged_in_as" value[=]"([0-9a-zA-Z]+)"[>],o;
  $resultsName = $1 if m,[<]textarea.*? name[=]"([a-z_]+)" ,;
  last if defined($loginValue) && defined($resultsName);
}
defined($loginValue) && defined($resultsName)
  or die "$baseURL implementation has changed; cannot find 'was_logged_in_as' and/or 'textarea name='\n";

my ($submitted) = $http->post_form($url, { 'Referer' => $baseURL, 'user_login' => $gimpsUser, 'user_password' => $gimpsPassword,
                                           'was_logged_in_as' => $loginValue, $resultsName => $gimpsResults } );
$err = "status $submitted->{'status'}, reason $submitted->{'reason'}; HTTP headers of second response:\n";
print(STDOUT "verbose second response headers\n") if $verbose > 2 && $response->{'success'};
while (my ($k, $v) = each %{$submitted->{'headers'}})
{
  $err .= ("$k $_\n") for (ref $v eq 'ARRAY' ? @$v : $v);
  next unless $verbose > 2;
  print(STDOUT "verbose $k " . (ref $v eq 'ARRAY' ? @$v : $v) . "\n") if $response->{'success'};
}
if ($verbose > 2)
{
  print(STDOUT "verbose done with second response headers\n") if $response->{'success'};
  print(STDOUT "verbose $_ $response->{$_}\n") foreach grep { $_ ne 'content' } sort(keys(%{$response}));
  print(STDOUT "verbose done with second response\n");
}
print(STDOUT "verbose second response success is $submitted->{'success'}\n") if $verbose == 1;

$submitted->{'success'} or die "$url failed, $err";
length($submitted->{'content'}) or die "no HTML in second response, $err";

exit(0);
