#!/usr/bin/perl -Tw
# Spread work, usually from askServer.pl, into separate work files for doWork.pl.

use strict;
use strict 'vars';
use Scalar::Util;

BEGIN { push(@INC, '.') if -s('MersInfo.pm'); $ENV{'PATH'} = "/usr/bin:/bin"; }
use MersInfo;
use Mersenne;
use ECMSave;

my ($usage) = ("Usage: spreadWork.pl [files] [--help|-h] [--verison] [--check]\n"
               . "  Read the files (or STDIN) and create 'work files' for doWork.pl.\n"
               . "  Work files under working.doWork/work/ are given to child workers by doWork.pl.\n"
               . "  This script is normally run by doWork.pl rather than a user (person).\n"
  );

my ($check);
if (grep { m/^--(check|version|help)$/ || $_ eq '-h' } @ARGV)
{
  my ($version) = MersInfo::version();
  print(STDOUT "$0 version $version\n");
  print(STDOUT "$usage")
    if grep { m/^--help$/ || $_ eq '-h' } @ARGV;
  exit(0)
    if grep { m/^--(version|help)$/ || $_ eq '-h' } @ARGV;
  $check = 1;
}

foreach (grep { !m/^-/ && ! -f($_) } @ARGV)
{
  warn("file '$_' does not exist, $!");
  $check = 2;
}
foreach (grep { !m/^-/ && -l($_) } @ARGV)
{
  warn("'$_' is a soft link\n");
  $check = 2;
}
foreach (grep { m/^-/ && !m/^--(check|version|help)$/ && $_ ne '-h' } @ARGV)
{
  warn("unknown option '$_'\n");
  $check = 2;
}
exit(2)
  if $check && $check > 1;

my ($errors) = '';

sub FreeExp($);
sub MoveToWork($$$$$);

my (%workOrder, %workMers, %workCnt, %workPri, @pieces, $asking, $exp);

# %workOrder is constant after this.  Value is field of (often mers) format to use to sort work ...
#   ... but the priority of all mers format tasks is now from Mersenne::mersPriority().
$workOrder{'ecm'} = -1;
$workOrder{'joiner'} = -1;
$workOrder{'gpu'} = -1;
$workOrder{'pminus1'} = -1;
$workOrder{'pplus1'} = -1;
# GIMPS - use time() for the priority - mersenne.org wants tasks done in the order they were reserved
$workOrder{'gpuGimpsFactor'} = 0;

# %workMers is constant after this.  Key is work type/kind and value is a pattern which matches a line
#   that should be used with %workOrder to get the task's priority.
$workMers{'ecm'} = '^M[(] [1-9][0-9]* [)][Ee]: [0-9]';
$workMers{'joiner'} = '^M[(] [1-9][0-9]* [)]G: [0-9]';
$workMers{'gpu'} = '^M[(] [1-9][0-9]* [)]g: [0-9]';
$workMers{'pminus1'} = '^M[(] [1-9][0-9]* [)]o: [0-9]';
$workMers{'pplus1'} = '^M[(] [1-9][0-9]* [)]q: [0-9]';
$workMers{'gpuGimpsFactor'} = '^Factor=[0-9A-F]+,([1-9][0-9]+),[1-9][0-9]+,[1-9][0-9]+$';

for my $type (keys(%workOrder)) {
  defined($workMers{$type}) or $errors .= "%workOrder/%workMers values do not match\n";
}

for my $type (keys(%workMers)) {
  defined($workOrder{$type}) or $errors .= "%workOrder/%workMers values do not match\n";
  $workCnt{$type} = 0;
  $workPri{$type} = 0;
}

my ($floatPattern) = MersInfo::patterns('float');

die("$errors") if $errors;
exit(0) if $check;

my ($taskFH);
my (%saveFile);
while (<<>>)
{ # Save files must be dealt with first since they could contain, e.g., empty lines.
  s,\015\012$,,; # mersServer.pl appends these two characters to every line, even save file content
  if (m/^save file begin (ecm[.]([1-9][0-9]+)[.](p[mp]1)([.]M)?)$/)
    {
      %saveFile = ( 'file' => "$1", 'exp' => "$2", 'method' => "$3", 'M' => ($4 ? "$4" : ''), 'content' => '', );
      next;
    }
  if (m/^save file end ([a-z][-._a-zA-Z0-9]+)$/)
    {
      die "save file end gives different file name, '$1', not expected '$saveFile{'file'}'"
        unless $1 eq $saveFile{'file'};
      die "save file '$1' has unexpected content"
        unless $saveFile{'content'} =~ m/^([-+=;:.@ A-Za-z0-9\n\r]+)$/;
      my ($err) = ECMSave::createSaveFile(%saveFile);
      print(STDOUT "$0 - $err\n")
        if $err && ref($err) ne 'ECMSave';
      %saveFile = ();
      next;
    }
  $saveFile{'content'} .= "$_",
    next
    if exists $saveFile{'content'};
  s/[ \t][ \t]+/ /g;
  s/[ \n\r]$//g;
  s/^ +//g;
  next unless $_;
  my ($kind);
  next if (m/^could not get all work requested; /);
  next if (m/^the server has no suitable work/);
  next if (m/^connect failed, /);
  next if (m/^server exiting/);
  if (m/$workMers{'gpuGimpsFactor'}/o)
    { # From mersenne.org, which wants tasks done in the order they were reserved so ...
      $exp = $1;
      $kind = 'gpuGimpsFactor';
      # ... set the priority to the current time in seconds; the exponent will be appended to break ties.
      my ($pri) = time();
      $pri =~ m,^([1-9][0-9]+)$, or die "time not numeric, $!";
      $pri = $1;
      open($taskFH, '>', "working.doWork/spread/$$") or die "could not create new task file, $!";
      print($taskFH "$_\n");
      close($taskFH) or die "could not close new task file, $!";
      MoveToWork($exp, $pri, $kind, "working.doWork/spread/$$", 1);
      $exp = 0;
      next;
    }
  if (m/^M[(] ([1-9][0-9]*) [)].: +([0-9]+)/)
    {
      if (!defined($exp) || $exp != $1)
        {
          if ($exp)
            {
              close($taskFH) or die "could not close new task file, $!";
              MoveToWork($exp, $workPri{$asking}, $asking, "working.doWork/spread/$$", $workCnt{$asking});
            }
          open($taskFH, '>', "working.doWork/spread/$$") or die "could not create new task file, $!";
          $exp = $1;
          local $_;
          $workCnt{$_} = $workPri{$_} = 0
            foreach (keys(%workMers));
        }
      print($taskFH "$_\n");
      for my $type (keys(%workOrder))
        {
          next unless m/$workMers{$type}/;
          $asking = $type;
          @pieces = split(' ', $_);
          shift(@pieces);
          scalar(@pieces) > 2 or die "bad pattern or other bug while getting priority for line $_";
          ++$workCnt{$asking};
          next unless scalar(@pieces) >= $workOrder{$asking};
          $workPri{$asking} = Mersenne::mersPriority(@pieces),
            next
            if $workOrder{$asking} < 0;
          my ($increment) = $1
            if (scalar(@pieces) > $workOrder{$asking} &&
                $pieces[$workOrder{$asking}] =~ m,$floatPattern,);
          $increment = 0.0
            unless $increment;
          $workPri{$asking} += $increment + 0.5 - 1.0/("$exp.0");
          last;
        }
      next;
    }
  next if m,max tasks,; # server thinks this client has enough work (all mers letters)
  warn("unexpected/unknown input format: '$_'");
  $exp = 0;
} # end of while (<<>>)
!$exp or close($taskFH) or die "could not close new task file, $!";
!$exp or MoveToWork($exp, $workPri{$asking}, $asking, "working.doWork/spread/$$", $workCnt{$asking});
exit(0);

sub MoveToWork($$$$$)
{
  my ($exp, $pri, $type, $file, $cnt) = @_;
  warn("got $type task with no task lines; ignoring it\n"),
    FreeExp($exp),
    return(0) # failed to get new work
    if $cnt <= 0;
  # Use the average $pri rather than the sum.
  $pri /= $cnt;
  my ($workFile) = "working.doWork/work/$type/";
  -d($workFile)
    or mkdir($workFile)
    or die("could not mkdir $workFile, $!");
  $pri = "0." if ($pri < 0.1);
  $pri .= '.' if ($pri !~ m/[.]/o);
  $workFile .= "$pri$exp"; # append the exponent to break ties
  die("file '$file' in MoveToWork is tainted")
    if Scalar::Util::tainted($file);
  die("workFile '$workFile' in MoveToWork is tainted")
    if Scalar::Util::tainted($file);
  print(STDOUT "$type $exp $pri$exp\n"),
    return(1) # have new work
    if (rename("$file", "$workFile"));
  warn "could not rename new task file $file to $workFile, $!\n";
  return(0);
}

sub FreeExp($)
{
  defined($_[0]) or die "FreeExp() called without an exponent";
  open(my $sendQ, '>>', 'working.doWork/queue') or die "could not append to working.doWork/queue, $!";
  print($sendQ "free $_[0]\n");
  close($sendQ) or die "could not close working.doWork/queue, $!";
}
