#include "version.h"
/* Read exponents, known factors, and other info and check the
   cofactors' primality with a strong pseudo-prime test.  The only
   supported input format is in the usual mers package format:

M( 1213 )C: 327511
M( 1213 )C: 715079841864486572798550389025242773233223013545064445043124471
M( 1213 )E: 9528550800 3162277
M( 1213 )H: 33685965952003082153441520768783740767778887180162088784223559881
M( 1213 )c: 297
M( 1213 )o: 4294950000
M( 1213 )q: 4294950000

   All the info for a particular exponent should be together, but the
   order of lines for a single exponent should not matter.  'C' lines
   denote factors, 'H' lines the highest trial factor attempted, 'G'
   lines denote gaps in trial factoring, and 'E' lines list the ECM
   "work" (the sum over all prior curves of the bound, presently) and
   largest bound.  Note that the primary output of sprp will always be
   of the form:

M( 1213 )C: 327511
M( 1213 )c: 297
M( 17029 )S

   The first line is an example of a factor being discovered, which
   will sometimes occur during SPRP testing.  The second says that the
   cofactor is composite and has 297 digits.  The third says that the
   remaining cofactor is "probably" prime, as it has passed ten
   pseudo-prime tests.  These formats are also valid input to sprp,
   though 'S' and 'D' (for 'SPRP' and 'done') lines will mean that
   exponent will be skipped.  (If -i is given a value larger than 0,
   other lines will be printed but those are _not_ valid input to the
   program).

   Please send any questions, comments, improvements, fixes, bugs, and
   all output of the forms above to me via email:

   wedgingt@acm.org

   so I can update the data about Mersenne numbers that I maintain.

   For a lengthier description of the primary input/output format, see
   mersfmt.txt.

*/

#include "setup.h"
#include "factor.h"
#include <time.h>
#include <sys/times.h>
#include <math.h>
#include <unistd.h>
#include <stdlib.h>

/* only for prototypes; rw.c is not linked with this program. */
#include "rw.h"

#ifdef macintosh
# include <console.h>
#endif

/* Number of times to test a new factor's or cofactor's primality */
#ifndef FERMAT_TESTS
# define FERMAT_TESTS (-10)
#endif

#if !defined(MAX_SPLITS)
# define MAX_SPLITS (2)
/* Mersennes that can be algebraicly factored by the program can be split into at most this many factors */
#endif

#if MAX_SPLITS < 2
# undef MAX_SPLITS
# define MAX_SPLITS (2)
#endif

#ifdef MAX_EXPONENTS
# undef MAX_EXPONENTS
#endif
#define MAX_EXPONENTS (10)

#if MAX_EXPONENTS < 5*MAX_SPLITS
# undef MAX_EXPONENTS
# define MAX_EXPONENTS (5*MAX_SPLITS)
#endif

#define ARRAY_SLOTS (MAX_EXPONENTS + MAX_SPLITS + 1)

#ifdef M_LN10l
# define LN10 (M_LN10l)
#else
# ifdef M_LN10
#  define LN10 (M_LN10)
# else
#  define LN10 (2.302585092994045684017991454684364207601101488628772976033327)
# endif
#endif

/* flags used in the global done[] array */
#define DONE_FLAG_DONE (0x1)
  /* bit 0: done, as in completely factored? */
#define DONE_FLAG_CHECK_PRIMALITY (0x2)
  /* bit 1: check (original or new) cofactor's primality? */
#define DONE_FLAG_SPLIT (0x4)
  /* bit 2: split exponent due to complicated algebraic factorization?  see split_LM() */
#define DONE_FLAG_L_OR_M (0x8)
  /* bit 3: for split exponents, L (bit off) or M (bit on) */

/* macros used to test done[] vs. DONE_FLAG*'s */
#define DONE_YET(entry) ((done[entry] & DONE_FLAG_DONE) == DONE_FLAG_DONE)
#define CHECK_PRIMALITY(entry) ((done[entry] & DONE_FLAG_CHECK_PRIMALITY) == DONE_FLAG_CHECK_PRIMALITY)
#define IS_SPLIT(entry) ((done[entry] & DONE_FLAG_SPLIT) == DONE_FLAG_SPLIT)
#define IS_SPLIT_L(entry) ((done[entry] & (DONE_FLAG_SPLIT|DONE_FLAG_L_OR_M)) == DONE_FLAG_SPLIT)
#define IS_SPLIT_M(entry) ((done[entry] & (DONE_FLAG_SPLIT|DONE_FLAG_L_OR_M)) == (DONE_FLAG_SPLIT|DONE_FLAG_L_OR_M))

/* macros to set or unset flags in done[] */
#define DONE_NOW(entry) (done[entry] &= ~(int)DONE_FLAG_CHECK_PRIMALITY, done[entry] |= DONE_FLAG_DONE)
#define DONE_CLEAR(entry) (done[entry] = 0)
#define DONE_DO_CHECK(entry) (done[entry] |= DONE_FLAG_CHECK_PRIMALITY)
#define DONE_DONT_CHECK(entry) (DONE_NOW(entry))
#define DONE_SPLITTING_L(entry) ((done[entry] |= DONE_FLAG_SPLIT), (done[entry] &= ~(int)DONE_FLAG_L_OR_M))
#define DONE_SPLITTING_M(entry) (done[entry] |= (DONE_FLAG_SPLIT|DONE_FLAG_L_OR_M))

const char *program_name = NULL;
FACTOR zero, one, two; /* constants after first set near start of main() */
UL y[ARRAY_SLOTS] = { 0L }; /* Mersenne exponents */
FACTOR z[ARRAY_SLOTS]; /* remaining cofactor */
char done[ARRAY_SLOTS] = { 0L }; /* DONE_FLAG* flags */

extern FACTOR savef, numer;

#ifdef fac_init_random_start_var_dec
fac_init_random_start_var_dec;
#endif

/* Duplicate from rw.c; referenced in functions in factor.c unused by this program. */
const char *archive_name PROTO((void))
 {
  return("mersarch.txt");
 }

/* Duplicate from rw.c; referenced in functions in factor.c unused by this program. */
FILE *open_archive PROTO((void))
 {
  return(fopen(archive_name(), "a"));
 }

/* If y[which] is composite, pull out the factors of z[which] that are due to y[which]'s factors. */
#ifdef __STDC__
static void pull_factors(int which)
#else
static void pull_factors(which)
int which;
#endif
 {
  UL yFactor = 2L, yRemaining;
  static FACTOR M_y_Cofactor, zFactor, zRemainder;
  static char doneInit = '\0';

  if (doneInit == '\0')
   {
    fac_init_var(M_y_Cofactor);
    fac_init_var(zFactor);
    fac_init_var(zRemainder);
    doneInit = 'd';
   }
  fac_set_int(y[which], &zFactor);
  fac_gcd(z[which], zFactor, &zRemainder);
  while (!fac_equal(zRemainder, one))
   { /* should only occur if y[which] is not square-free */
    fac_div(z[which], zRemainder, z + which, &zFactor);
    fac_set_int(y[which], &zFactor);
    fac_gcd(z[which], zFactor, &zRemainder);
   }
  for (yRemaining = y[which]; yFactor <= yRemaining && 2L*yFactor <= y[which]; yFactor++, yFactor |= 1)
    if (yRemaining % yFactor == 0L)
     {
      for (yRemaining /= yFactor; yRemaining % yFactor == 0L; )
        yRemaining /= yFactor;
      fac_left_shift(one, y[which]/yFactor, &M_y_Cofactor);
      fac_sub(M_y_Cofactor, one, &M_y_Cofactor); /* now M(y[which])/yFactor) */
      /* GCD(M(a), M(b)) == M(GCD(a, b)), so divide out M(y[which])/yFactor) */
      fac_gcd(z[which], M_y_Cofactor, &zFactor);
      while (!fac_equal(zFactor, one))
       {
        fac_div(z[which], zFactor, z + which, &zRemainder);
        fac_gcd(z[which], M_y_Cofactor, &zFactor);
       }
     }
 }

/* if y[which] % 8 == 4, split z[which] into the L (z[which]) and M (z[which + 1]) halves based on */
/*  algebraic factorization: M(8*n - 4) = M(4*n - 2)*(M(4*n - 2) + 2) (difference of squares); */
/*  M(4*n - 2) will have already been pulled out by pull_factors(), but M(4*n - 2) + 2 == */
/*  2^(4*n - 2) + 1 == (2^(2*n - 1) - 2^n + 1)*(2^(2*n - 1) + 2^n + 1) == L*M.  But note that */
/*  pull_factors() may have pulled factors of L and/or M as well as M(4*n - 2), so we have to do */
/*  more fac_gcd()'s. */
#ifdef __STDC__
static int split_LM(int which)
#else
static int split_LM(which)
int which;
#endif
 {
  UL n = (y[which] + 4)/8;
  static FACTOR L, M, temp;
  static char doneInit = '\0';

  if (doneInit == '\0')
   {
    fac_init_var(L);
    fac_init_var(M);
    fac_init_var(temp);
    doneInit = 'd';
   }
  if (8*n - 4 != y[which] || n < 4)
    return(0); /* should not have been called; either exponent is not the right form or is very small */
  fac_left_shift(one, n - 1, &temp); /* 2^(n - 1) */
  fac_sub(temp, one, &L); /* 2^(n - 1) - 1 == M(n - 1) */
  fac_add(temp, one, &M); /* 2^*(n - 1) + 1 */
  fac_left_shift(L, n, &temp); /* 2^(2*n - 1) - 2^n */
  fac_add(temp, one, &L);
  fac_left_shift(M, n, &temp); /* 2^(2*n - 1) + 2^n */
  fac_add(temp, one, &M);
  /* in case we return early: */
  DONE_CLEAR(which);
  DONE_NOW(which);
  DONE_CLEAR(which + 1);
  DONE_NOW(which + 1);
  fac_gcd(z[which], M, z + which + 1);
  if (fac_equal(z[which + 1], one))
    return(0); /* something is wrong */
  fac_gcd(z[which], L, &temp);
  if (fac_equal(temp, one))
    return(0); /* ditto */
  fac_div(z[which], temp, &M, &L);
  if (!fac_equal(L, zero))
    return(0); /* implies fac_gcd() bug */
  if (!fac_equal(M, z[which + 1]))
    return(0); /* something is wrong in that I don't understand algebra/situation */
  fac_set(temp, z + which);
  y[which + 1] = y[which];
  DONE_CLEAR(which);
  DONE_CLEAR(which + 1);
  DONE_DO_CHECK(which);
  DONE_DO_CHECK(which + 1);
  DONE_SPLITTING_L(which);
  DONE_SPLITTING_M(which + 1);
  return(1);
 }

#ifdef __STDC__
int main(int argc, char **argv)
#else
int main(argc, argv)
int argc;
char **argv;
#endif
 {
  FACTOR a; /* lots of uses */
  FACTOR b; /* factor from input, other uses */
  FACTOR d; /* remainder of division of Mersenne by 'known' factor */
  FACTOR f; /* lots of uses */
  FACTOR ztmp; /* just for swapping within z[] */
  unsigned long p; /* Mersenne exponent read */
  unsigned long first = 32L, last = ~(unsigned long)0; /* ignore exponents outside this range */
  unsigned long limit = 0L; /* maximum size of cofactors to check */
  int which = 0; /* index of exponent that is being worked on */
  int max; /* count of incompletely factored Mersennes (split ones count as more than one, however) */
  int info = 0; /* how much extra info to print */
  int compos = 0; /* return value from fac_fermat() */
  char *outfn = NULL; /* output file name; replaces stdout */
  int split = 0; /* after a call to split_LM(), this is sometimes used to modify which or max */
  int splits_warned = MAX_SPLITS; /* warn if split gets too high */
  double startTime, stopTime;
  struct tms tms0, tms1;
  const double ticksPerSecond = (double)sysconf(_SC_CLK_TCK);
  char M; /* 'C', 'G', 'H', 'U', 'E', 'e', etc. */
  char flag = 0; /* -D means done: print cofactor in decimal when it is thought to be prime: 0x8 bit */
#define PRINT_COFACTORS ((flag & 0x8) == 0x8)
                 /* -C means check input factors to verify they are prime */
#define CHECK_FACTORS ((flag & 0x10) == 0x10)
                 /* -v means verify: do one pseudo-prime test of each cofactor marked 'S', 'D', 'L', or 'M' */
#define VERIFY_SPRP_COFACTORS ((flag & 0x20) == 0x20)
                 /* -w means print cofactors (in mers format) that need to be checked but are greater than limit (above) */
#define PRINT_WORK ((flag & 0x40) == 0x40)
  int ch = '\0'; /* for reading a char at a time and scanf() return value */

#ifdef macintosh
  argc = ccommand(&argv);
#endif
  /* setup */
  if (argv == NULL || argv[0] == NULL)
    program_name = "sprp";
  else
    for (program_name = argv[0] + strlen(argv[0]); program_name > argv[0] && *program_name != '/'; )
      --program_name;
  if (*program_name == '/')
    ++program_name;
  setup();
  fac_init();
  fac_init_var(zero);
  fac_init_var(one);
  fac_init_var(two);
  fac_set_int((UL)0, &zero);
  fac_set_int((UL)1, &one);
  fac_set_int((UL)2, &two);
  fac_init_var(a);
  fac_init_var(b);
  fac_init_var(d);
  fac_init_var(f);
  fac_init_var(ztmp);
  fac_init_random((UL)time(NULL));

  /* parse the command line */
  for (argc = 1; // to index argv
       !shouldTerminate && argv != NULL && argv[argc] != NULL && argv[argc][0] == '-';
       argc += (shouldTerminate ? 0 : 1))
    switch (argv[argc][1])
     { /*!!create UL get_arg()? in setup.c? or rw.c? */
      case 'C':
        flag ^= 0x10; /* CHECK_FACTORS */
        break;
      case 'D':
        flag ^= 0x8; /* PRINT_COFACTORS */
        break;
      case 'i': case 'I':
        if (isascii(argv[argc][2]) && isdigit(argv[argc][2]))
          info = atoi(argv[argc] + 2);
        else
          if (argv[++argc] != NULL && isascii(argv[argc][0]) && isdigit(argv[argc][0]))
            info = atoi(argv[argc]);
          else
            info = -1, shouldTerminate = 1;
        break;
      case 'l':
        if (isascii(argv[argc][2]) && isdigit(argv[argc][2]))
          limit = strtoul(argv[argc] + 2, NULL, 10);
        else
          if (argv[++argc] != NULL && isascii(argv[argc][0]) && isdigit(argv[argc][0]))
            limit = strtoul(argv[argc], NULL, 10);
          else
            limit = 1, shouldTerminate = 1;
        break;
      case 'o': case 'O':
        if (argv[argc][2] != '\0')
          outfn = argv[argc] + 2;
        else
          if (argv[++argc] != NULL && argv[argc][0] != '\0')
            outfn = argv[argc];
          else
            outfn = NULL, shouldTerminate = 1;
        break;
      case 'r': case 'R':
        if (isascii(argv[argc][2]) && isdigit(argv[argc][2]))
          first = strtoul(argv[argc] + 2, NULL, 10);
        else
          if (argv[++argc] != NULL && isascii(argv[argc][0]) && isdigit(argv[argc][0]))
            first = strtoul(argv[argc], NULL, 10);
          else
            first = 0L, shouldTerminate = 1;
        if (!shouldTerminate && argv[++argc] != NULL && isascii(argv[argc][0]) && isdigit(argv[argc][0]))
          last = strtoul(argv[argc], NULL, 10);
        else
         {
          if (!shouldTerminate)
            --argc; // undo the ++argc if this if's condition
          last = 0L;
          shouldTerminate = 1;
         }
        if (first > last || first < 7L)
          shouldTerminate = 1;
        break;
      case 'v': case 'V':
        flag ^= (char)0x20; /* VERIFY_SPRP_COFACTORS with a single new SPRP test */
        break;
      case 'w': case 'W':
        flag ^= (char)0x40; /* PRINT_WORK cofactors that should be checked larger than limit */
        if (limit < 2)
          limit = 2; /* don't do any SPRP tests here when -w is given unless -l also given */
        break;
      default: /* help option or unknown option or any arg */
        shouldTerminate = 1; /* get out of this loop and into the next if */
        break;
     }

  if (shouldTerminate)
   {
    if (argv[argc][0] != '-' ||
        (argv[argc][1] != 'h' && argv[argc][1] != 'H' && argv[argc][1] != '?'))
      fprintf(stderr, "%s: unknown flag or argument: '%s'\n", argv[0], argv[argc]);
    fprintf(stderr, "  Usage: %s [-{hH?CDv}]\n", argv[0]);
    fprintf(stderr, "         [-i<info>] [-l<limit>] [-o<outputfilename>] [<inputfilename> ...]\n");
    fprintf(stderr, "  Performs strong pseudo-prime checks of current cofactors of Mersenne numbers.\n");
    fprintf(stderr, "    -C: toggle: check primality of input factors\n");
    fprintf(stderr, "    -D: toggle: print, in decimal, any cofactors that are pseudo-prime\n");
    fprintf(stderr, "    -h, -H, -?: print this usage information and exit\n");
    fprintf(stderr, "    -i<info>: how much extra info to print, boolean per bit:\n");
    fprintf(stderr, "      1: print each exponent as it is first read;\n");
    fprintf(stderr, "      2: print exponent just before SPRP checking its cofactor;\n");
    fprintf(stderr, "      4: print first exponent in array when it changes;\n");
    fprintf(stderr, "      8: print exponent and position in array when latter changes;\n");
    fprintf(stderr, "    -l<limit>: limit on the size of cofactors to check;\n");
    fprintf(stderr, "    -o<outputfilename>: file to write non-error output to (default stdout)\n");
    fprintf(stderr, "    -r<first> <last>: skip exponents less than <first> or more than <last>\n");
    fprintf(stderr, "     (<first> must be > 6; Mersennes with exponents under 1206 are completely factored)\n");
    fprintf(stderr, "    -v: toggle: do one new SPRP test of each cofactor marked 'S', 'D', 'L', or 'M'.\n");
    fprintf(stderr, "    Any input files should be given after all options\n");
    fprintf(stderr, "  All toggles default to off.\n");
    fprintf(stderr, "  Reads from stdin if no input file names given.\n");
    fprintf(stderr, "  Note that some input lines will override the command line values per exponent.\n");
    fprintf(stderr, "  Maintained by <a href=\"mailto:wedgingt@acm.org\">Will Edgington</a> as part of\n");
    fprintf(stderr, "  the <a href=\"https://sourceforge.net/projects/mers/\">mers package</a>.\n");
    return(1);
   }

  /* check arguments given */
  if (info < 0)
    fprintf(stderr, "%s: info value (-i) must be non-negative\n\t(\'%s -h\' will print help info)\n",
            argv[0], program_name);
  if (limit == 1)
    fprintf(stderr, "%s: limit (-l) must be integer >= 2\n\t(\'%s -h\' will print help info)\n",
            argv[0], program_name);
  if (info < 0 || limit == 1)
    return(2);

  if (argv[argc] != NULL && freopen(argv[argc], "r", stdin) == NULL)
   {
    fprintf(stderr, "%s: could not open input file '%s'\n\t(\'%s -h\' will print help info)\n",
            argv[0], argv[argc], program_name);
    perror(argv[argc]);
    return(errno);
   }
  if (argv[argc] != NULL)
    ++argc;
  if (outfn != NULL && freopen(outfn, "a", stdout) == NULL)
   {
    fprintf(stderr, "%s: could not open output file '%s' for appending\n\t(\'%s -h\' will print help info)\n",
            argv[0], outfn, program_name);
    perror(outfn);
    return(errno);
   }
  /* try to prevent data from being lost in the stdio buffers */
  fflush(stdout);
  setbuf(stdout, NULL);
  fflush(stderr);
  setbuf(stderr, NULL);

  for (max = 0; max < ARRAY_SLOTS; max++)
    fac_init_var(z[max]), DONE_CLEAR(max), DONE_NOW(max);

  /* primary loop; ends just before EOF */
  for (y[max = 0] = 0L; !shouldTerminate; )
   { /* if no more in this input file, done or open the next one */
    if (feof(stdin))
     {
      if (argv[argc] == NULL && max < 0)
        break;
      if (argv[argc] != NULL && freopen(argv[argc], "r", stdin) == NULL)
       {
        fprintf(stderr, "%s: could not open input file '%s'\n\t(\'%s -h\' will print help info)\n",
                argv[0], argv[argc], program_name);
        perror(argv[argc]);
        return(errno);
       }
      if (argv[argc] != NULL)
        ++argc;
     }
    /* if we have room, read and parse a line of input */
    if (max < MAX_EXPONENTS && !feof(stdin) && (ch = fgetc(stdin)) != EOF)
     { /* this is a mess because some scanf's _require_ whitespace if a blank starts the format string */
      while (isascii(ch) && isspace(ch))
        ch = fgetc(stdin);
      /* Only support mersfmt, but the expected colon cannot end the scanf format */
      /*   string because 'S', 'D', 'L', and 'M' lines do not always have one. */
      if (ch != 'M' || 2 != scanf("( %lu )%c", &p, &M))
       {
        fprintf(stderr, "%s: input format not recognized\n", argv[0]);
        if (y[max] != 0L || (max > 0 && y[--max] != 0L))
          fprintf(stderr, "  (after some data for exponent " PRINTF_FMT_UL " read correctly)\n", y[max]);
        return(3);
       }
      if (p < first || p > last)
       {
        while ((ch = fgetc(stdin)) != EOF && ch != '\n')
          ;
        continue;
       }
      fac_set(one, &b);
      if ((ch = fgetc(stdin)) == ':')
       {
        fac_set(zero, &b);
        for (ch = fgetc(stdin); isascii(ch) && isspace(ch); )
          ch = fgetc(stdin);
        for ( ; isascii(ch) && isdigit(ch); ch = fgetc(stdin))
         {
          fac_mul_int(b, (UL)10, &b);
          fac_add_int(b, (UL)(ch - '0'), &b);
         }
       }
      if (fac_equal(b, zero))
       {
        if (presence("DPEeoq", M) == NULL)
         {
          fprintf(stderr, "%s: input format not recognized\n", argv[0]);
          if (y[max] != 0L || (max > 0 && y[--max] != 0L))
            fprintf(stderr, "  (after some data for exponent " PRINTF_FMT_UL " read correctly)\n", y[max]);
          return(4);
         }
       }
      if (presence("SLMclm", M) != NULL && (fac_equal(b, zero) || !fac_less_than_int(b, p + 1)))
        fac_set_int(p, &b);
      if (y[max] != p)
       {
        if (y[max] != 0L && !DONE_YET(max))
          max++;
        y[max] = p;
        DONE_CLEAR(max);
        DONE_DO_CHECK(max);
        fac_left_shift(one, y[max], z + max);
        fac_sub(z[max], one, z + max);
        pull_factors(max);
        if ((info & 1) != 0)
          printf("first (mers) input for M%lu\n", p);
        if (y[max] % 8 == 4 && y[max] > 20)
         {
          if (split_LM(max) != 0)
            max++;
          else
            fprintf(stderr, "%s: split_LM() returned 0 for M" PRINTF_FMT_UL "\n", program_name, y[max]);
         }
       }
      while (ch != EOF && ch != '\n' && isspace(ch))
        ch = fgetc(stdin);
      if (ch != EOF && ch != '\n')
        ungetc(ch, stdin);
      while (ch != EOF && ch != '\n')
        ch = fgetc(stdin); /* skip the rest of the line */
      split = 0;
      switch (M)
       {
        case 'P': /* Mersenne prime; don't try to confirm or deny, since LL test is as fast as 1 SPRP test */
          DONE_NOW(max);
          continue;
        case 'o': /* b is P-1 stage one bound; no help to SPRP */
        case 'q': /* b is P+1 stage one bound; no help to SPRP */
        case 'e': /* ECM curves; no help to SPRP */
        case 'E': /* ECM "work"; no help to SPRP */
        case 'G': /* trial factoring 'G'ap data; no help to SPRP */
        case 'H': /* b is the trial factoring extent; no help to SPRP */
        case 'J': /* trial factoring 'J'oin data; no help to SPRP */
        case 'X': /* b is 1strs extent for exponents up to y[max]; no help for SPRP */
          break;
        case 'l': /* digits of composite cofactor L half of a split exponent */
          if (max > 0 && y[max - 1] == y[max] && IS_SPLIT_L(max - 1))
            split = 1;
          /* fall thru */
        case 'c': /* digits of composite cofactor of non-split exponent */
        case 'm': /* digits of composite cofactor M half of a split exponent */
          if ((IS_SPLIT(max - split) || M != 'c') && (!IS_SPLIT_L(max - split) || M != 'l') &&
              (!IS_SPLIT_M(max - split) || M != 'm'))
            fprintf(stderr, "%s: split %d disagrees with mers letter %c for M" PRINTF_FMT_UL "\n",
                    program_name, split, M, y[max - split]);
          if (fac_less_than_int(b, fac_digits(z[max - split])))
           {
            fprintf(stderr, "%s: size of cofactor has increased for M" PRINTF_FMT_UL "; missing factor?\n",
                    program_name, y[max - split]);
            DONE_DONT_CHECK(max - split);
           }
          else
            if (fac_less_than_int(b, 1 + fac_digits(z[max - split])))
              DONE_DONT_CHECK(max - split);
            else
             {
              DONE_DO_CHECK(max - split);
              if ((info & 2) != 0)
                printf("new cofactor to check for M" PRINTF_FMT_UL "%s (%d of %d)\n", y[max - split],
                       IS_SPLIT(max - split) ? (IS_SPLIT_L(max - split) ? "L" : "M") : "",
                       max - split, max - split);
              if (PRINT_WORK && 0 < limit && limit <= fac_digits(z[max - split]))
               {
                printf("M( " PRINTF_FMT_UL " )%c: ", y[max - split], M);
                fprint_fac(stdout, b);
                printf(" # needs SPRP test; cofactor has %zu digits\n", fac_digits(z[max - split]));
               }
             }
          break;
        case 'C':
          if (CHECK_FACTORS)
           {
            fac_set(b, &d);
            if ((compos = fac_fermat(&d, -2, 3)) != 0)
             {
              printf("M( " PRINTF_FMT_UL " )C: ", y[max]);
              fprint_fac(stdout, d);
              if (compos > 0)
                printf(" # composite factor given\n");
              else
                printf(" # factor of given factor; cofactor not checked\n");
             }
           }
          split = 0;
          fac_gcd(z[max], b, &d);
          while (!fac_equal(b, d) && max - split > 0 && y[max - split - 1] == y[max - split])
           {
            if (!fac_less_than(d, two))
             {
              printf("M( " PRINTF_FMT_UL " )C: ", y[max - split]);
              fprint_fac(stdout, d);
              printf(" # factor of given factor\n");
              fac_div(z[max - split], d, &a, &f);
              if (fac_equal(f, zero))
                fac_set(a, z + max - split);
              else
                fprintf(stderr, "%s: bug in fac_gcd() or fac_div() in main() for M" PRINTF_FMT_UL "\n",
                        program_name, y[max - split]);
             }
            fac_gcd(z[max - ++split], b, &d);
           }
          if (split > splits_warned)
            fprintf(stderr, "%s: MAX_SPLITS exceeded; recompile with at least %d\n", program_name,
                    splits_warned = split);
          if (!fac_equal(b, d))
           {
            printf("M( " PRINTF_FMT_UL " )C: ", y[max - split]);
            fprint_fac(stdout, b);
            printf(" # %s: false factor or factor of smaller Mersenne\n", program_name);
            if (fac_less_than(d, two))
              break;
            printf("M( " PRINTF_FMT_UL " )C: ", y[max - split]);
            fprint_fac(stdout, d);
            printf(" # factor of given factor\n");
           }
          fac_div(z[max - split], d, &a, &b);
          if (!fac_equal(b, zero))
            fprintf(stderr, "%s: bug in fac_gcd() or fac_div() in main() for M(" PRINTF_FMT_UL ")\n",
                    program_name, y[max - split]);
          else
            if (fac_equal(a, one))
             {
              printf("M( " PRINTF_FMT_UL " )C: ", y[max - split]);
              fprint_fac(stdout, d);
              printf(" # given factor is cofactor\n");
              DONE_DO_CHECK(max - split);
              break;
             }
            else
              if (CHECK_FACTORS && fac_less_than(a, d))
               {
                printf("M( " PRINTF_FMT_UL " )C: ", y[max - split]);
                fprint_fac(stdout, d);
                printf(" # given factor is larger than cofactor\n");
                printf("M( " PRINTF_FMT_UL " )C: ", y[max - split]);
                fprint_fac(stdout, a);
                printf(" # cofactor\n");
                fac_set(d, z + max - split);
               }
              else
                fac_set(a, z + max - split);
          fac_gcd(z[max - split], d, &b);
          if (fac_equal(b, d))
           { /* two known exceptions: M1755 % (3511*3511) == 0; other is even exponent */
            if (!PRINT_WORK || y[max - split] > 1755)
             { // if -w, don't print this unless it's a new discovery
              printf("%s: possible contradiction of square-free conjecture\nM( " PRINTF_FMT_UL " )C: ",
                     program_name, y[max - split]);
              fprint_fac(stdout, d);
              printf(" # %s: square is a factor as well\n", program_name);
             }
            fac_div(z[max - split], d, &a, &b);
            if (!fac_equal(b, zero))
              fprintf(stderr, "%s: bug in fac_gcd() or fac_div() in main() for M(" PRINTF_FMT_UL ")\n",
                      program_name, y[max - split]);
            else
              if (fac_equal(a, one))
               {
                printf("M( " PRINTF_FMT_UL " )C: ", y[max - split]);
                fprint_fac(stdout, d);
                printf(" # given factor is cofactor\n");
                DONE_DO_CHECK(max - split);
                break;
               }
              else
                fac_set(a, z + max - split);
           }
          break;
        case 'd': /* changed to S in 2019 January */
          M = 'S';
          /* fall thru */
        case 'D': /* cofactor has been proven to be prime */
        case 'S': /* cofactor is prime or SPRP to at least 10 bases */
        case 'L': /* L half cofactor is prime or SPRP to at least 10 bases */
          split = (max > 0 && y[max - 1] == y[max] && IS_SPLIT_L(max - 1));
          if (M == 'L' && !IS_SPLIT_L(max - split))
            continue; /* invalid data or already marked done */
          /* fall thru */
        case 'M': /* M half cofactor is prime or SPRP to at least 10 bases */
          if (M == 'M' && !IS_SPLIT_M(max))
            continue;
          /* If -v given, check them all */
          if (VERIFY_SPRP_COFACTORS)
           {
            fac_set(z[max - split], &b);
            compos = fac_fermat(&b, 1, 3);
            switch (compos)
             {
              case 0: /* could not prove composite, which is expected */
                printf("M( " PRINTF_FMT_UL " )%c # verified\n", y[max - split], M);
                DONE_NOW(max - split);
                break;
              case 1: /* composite, which is unexpected, so "complain": */
                printf("M( " PRINTF_FMT_UL " )%c: %zu # %s: cofactor was marked SPRP or prime!\n",
                       y[max - split], M == 'L' ? 'l' : (M == 'M' ? 'm' : 'c'),
                       fac_digits(z[max - split]), program_name);
                DONE_DONT_CHECK(max - split);
                continue;
              default: /* undocumented return value, so "complain": */
                fprintf(stderr, "%s: fac_fermat(b, 1, 3) returned undocumented value (%d)\n",
                        program_name, compos);
                fprintf(stderr, "\twhen verifying SPRP cofactor of M(" PRINTF_FMT_UL ")%c\n", y[max - split], M);
                return(5);
             }
           }
          if (M == 'D')
            DONE_NOW(max);
          else if (presence("SLMd", M))
           {
            if (fac_less_than_int(b, fac_digits(z[max - split])))
             {
              fprintf(stderr, "%s: size of cofactor has increased for M" PRINTF_FMT_UL "; missing factor?\n",
                      program_name, y[max - split]);
              DONE_DONT_CHECK(max - split);
             }
            else if (fac_less_than_int(b, 1 + fac_digits(z[max - split])))
              DONE_DONT_CHECK(max - split);
            else
             {
              DONE_DO_CHECK(max - split);
              if ((info & 2) != 0)
                printf("new cofactor to check for M" PRINTF_FMT_UL "%s (%d of %d)\n", y[max - split],
                       IS_SPLIT(max - split) ? (IS_SPLIT_L(max - split) ? "L" : "M") : "",
                       max - split, max - split);
              if (PRINT_WORK && 0 < limit && limit <= fac_digits(z[max - split]))
               {
                printf("M( " PRINTF_FMT_UL " )%c: ", y[max - split], M);
                fprint_fac(stdout, b);
                printf(" # needs SPRP test; cofactor has %zu digits\n", fac_digits(z[max - split]));
               }
             }
           }
          if (!PRINT_COFACTORS)
            continue;
          printf("M( " PRINTF_FMT_UL " )%c: ", y[max - split], M);
          fprint_fac(stdout, z[max - split]);
          putchar('\n');
          if (M != 'D')
            continue;
          split++;
          while (max - split >= 0 && y[max - split] == y[max - split + 1])
           { /* assume 'D' applies to all parts of a split exponent */
            DONE_NOW(max - split);
            printf("M( " PRINTF_FMT_UL " )%c: ", y[max - split], M);
            fprint_fac(stdout, z[max - split]);
            putchar('\n');
            split++;
           }
          continue;
        default:
          fprintf(stderr, "%s: unexpected mersfmt 'key' letter (%c) for exponent " PRINTF_FMT_UL "\n",
                  program_name, M, y[max]);
          return(6);
       } /* switch (M): mersfmt letter */
     } /* if (max < MAX_EXPONENTS && (ch = fgetc(stdin)) != EOF) */

    /* We need to read all of the lines for the first exponent before we can do anything. */
    if (max < 0 || (!feof(stdin) && y[0] == y[max]))
      continue; /* if max < 0, we have no data at all and will exit */
    which = 0;
    if ((info & 8) != 0)
      printf(":" PRINTF_FMT_UL ": (%d of %d) is current exponent\n", y[which], which, max);
    fac_set(zero, &b);

    while (CHECK_PRIMALITY(which))
     { /* check primality of new factor or of old or new cofactor */
      if (fac_equal(b, zero)) /* check cofactor's primality, but don't try to factor it here */
        fac_set(z[which], &b);
      /* if -l<limit>  is given, only check cofactor's with up to <limit> digits */
      if (0 < limit && limit <= fac_digits(z[which]))
       { /* we cannot be checking a factor, since we have not called fac_fermat() for this exponent yet */
        DONE_NOW(which);
        break;
       }
      if ((info & 4) != 0 && fac_equal(z[which], b))
        printf("checking cofactor of M" PRINTF_FMT_UL " (%d of %d); %zu digits\n",
               y[which], which + 1, max + 1, fac_digits(z[which]));
      fac_set(b, &f);
      startTime = (double)times(&tms0);
      compos = fac_fermat(&b, FERMAT_TESTS, 3);
      stopTime = (double)times(&tms1);
      switch (compos)
       {
        case 1: /* composite cofactor (all composite factors get factored just above) */
          printf("M( " PRINTF_FMT_UL " )%c: %zu # %g secs %g CPUsecs\n", y[which],
                 IS_SPLIT(which) ? (IS_SPLIT_L(which) ? 'l' : 'm') : 'c',
                 fac_digits(z[which]),
                 (stopTime - startTime)/ticksPerSecond,
                 (((double)tms1.tms_utime) - (double)tms0.tms_utime)/ticksPerSecond);
          DONE_DONT_CHECK(which);
          continue;
        case 0: /* new factor or the cofactor is prime or pseudo prime */
          if (fac_equal(z[which], b))
           { /* factorization of M(y[which]) (this part, if split) thought to be complete */
            printf("M( " PRINTF_FMT_UL " )%c: " PRINTF_FMT_UL, y[which], IS_SPLIT(which) ? (IS_SPLIT_L(which) ? 'L' : 'M') : 'S', fac_digits(z[which]));
            if (PRINT_COFACTORS)
             {
              printf(" ");
              fprint_fac(stdout, z[which]);
             }
            printf(" # %g secs %g CPUsecs\n", (stopTime - startTime)/ticksPerSecond,
                   (((double)tms1.tms_utime) - (double)tms0.tms_utime)/ticksPerSecond);
            DONE_NOW(which);
            continue;
           }
          printf("M( " PRINTF_FMT_UL " )C: ", y[which]);
          fprint_fac(stdout, a);
          fac_set(zero, &b);
          continue; /* check primality of the new cofactor */
        case -1: /* composite, and new value of b may give new factor of old b, still in f - not libgmp */
        case -2: /* composite, and new value of b is a factor of old b, still in f - not libgmp */
          fac_div(f, b, &a, &d);
          if (compos == -1 && !fac_equal(d, zero))
           { /*!!skip for now, but print what we know in case it happens */
            printf("%s: Please send this section (below) to wedgingt@acm.org; it may lead to a new factor\n",
                   program_name);
            printf("%s: fac_fermat(&b, %d, 3) returned -1 for M" PRINTF_FMT_UL "\n", program_name, FERMAT_TESTS, y[which]);
            if (!fac_equal(z[which], f))
             {
              printf("%s: cofactor (z[which]) is ", program_name);
              fprint_fac(stdout, z[which]);
              putchar('\n');
             }
            printf("%s: old b is ", program_name);
            fprint_fac(stdout, f);
            printf("\n%s: new b is ", program_name);
            fprint_fac(stdout, b);
            printf("\n%s\n", "$Id: sprp.c,v 8.10 2017/06/19 01:34:34 wedgingt Exp $");
            printf("Please send this section (above) to wedgingt@acm.org\n");
            DONE_NOW(which); /*!!pretend we've completed it just in case */
            continue;
           }
          if (!fac_equal(d, zero))
            printf("%s: warning: non-zero remainder via fac_fermat() return of -2 for M" PRINTF_FMT_UL "\n",
                   program_name, y[which]);
          if (fac_less_than(a, b))
            fac_set(a, &d),
              fac_set(b, &a),
              fac_set(d, &b);
          if (fac_equal(z[which], f)) /* b is factor of (now old) cofactor */
           {
            fac_set(a, z + which);
            fac_set(b, &a);
            continue; /* check primality of new factor */
           }
          printf("M( " PRINTF_FMT_UL " )C: ", y[which]);
          fprint_fac(stdout, b);
          printf(" # %s: fac_fermat() == -2\n", program_name);
          fac_set(a, &b);
          continue; /* check primality of b, cofactor of factor */
        default: /* undocumented return value */
          fprintf(stderr, "%s: fac_fermat() returned undocumented value (%d)\n",
                  program_name, compos);
          fprintf(stderr, "\twhen called for M(" PRINTF_FMT_UL ")\n", y[which]);
          return(9);
       } /* switch (compos) */
     } /* while (CHECK_PRIMALITY(which)) */

    /* If we're at the front of the list or this one is done and we're full, */
    /*   shift later entries down over any that are done. */
    if (which == 0 || DONE_YET(which))
     { /* borrow ch and compos for subscripts */
      for (ch = 0; ch < max; ch++)
        if (DONE_YET(ch) && (max >= MAX_EXPONENTS - MAX_SPLITS - 1 || ch == 0 || ch == max - 1))
          break; /* Shift later entries, but less often than before so as ... */
      for (compos = ch + 1; compos <= max; compos++)
       { /* Look for next one that is not done */
        if (DONE_YET(compos))
          continue;
        if (DONE_YET(ch)) /* Should always be true, but might as well test it. */
         { /* Found one that isn't done; move it to earliest slot that was done. */
          y[ch] = y[compos];
          fac_set(z[ch], &ztmp); /* Trade z's rather than calling zfree(). */
          fac_set(z[compos], z + ch);
          fac_set(ztmp, z + compos);
          done[ch] = done[compos];
          DONE_CLEAR(compos);
          DONE_NOW(compos);
          if (ch++ == 0 && (info & 4) != 0)
            printf("first exponent is now M" PRINTF_FMT_UL "\n", y[0]);
         }
       } /* for (compos = ch + 1; compos <= max; compos++) */
      while (max > 0 && DONE_YET(max))
        max--;
      if (max == 0 && DONE_YET(0))
       {
        if (feof(stdin) || ferror(stdin))
          max--;
        else
          y[0] = 0L;
       }
     } /* if (which == 0 || DONE_YET(which)) */
   } /* for (y[max = 0] = 0L; !shouldTerminate && (max >= 0 || !feof(stdin)); ) */
  return(0);
 }
