#!/usr/bin/perl -Tw
# Send tasks and receive results from clients doing work related to Mersenne numbers,
#   updating the database with results and asking the database for tasks to perform.
# See askServer.pl and sendServer.pl for the matching client-side code.

#!!one server port for receiving results and another for sending work?
#!!  one server port for sending each type of work?
#!!one server (process) for each server port?
#!!  the G (trial factoring gaps) process would simply read from results.gaps.sort
#!!    while checking reservations
#!!start each process from a single script as children?
#!!  would make it easy to support restarting them after problems
#!!or a single server process for clients to connect to on a single port
#!!    that then creates a child to deal with each client?
#!!  and the server decides which child script to run based on what the client asks for?

#!!add internal work to update files in results/ directory?
#!!  a merge sort of an existing file with stdin containing relevant data would work
#!!  '.lock' directory that bsearch blocks on?
#!!  while this internal work is running, simply skip running bsearch?

use strict;
use strict 'vars';
use warnings;
use re '/aa'; # regular expressions emphasize ASCII; no other encodings are used or recognized
use Fcntl; # only needed to set non-blocking on file descriptors (sockets), but that's essential to the logic
use Socket qw(:DEFAULT :crlf);

#!!This has tainted(), which returns true if the argument is tainted ala Perl's -T; see 'man perlsec'.
#!!  It was needed to debug a taint issue related to the save file names read from clients.
#use Scalar::Util;

BEGIN { push(@INC, '.') if -s 'Mersenne.pm'; $ENV{'PATH'} = "/usr/bin:/bin"; }

# This is certainly needed when Perl's -d is in effect.
use Term::ReadLine;

$SIG{__WARN__} = sub { my ($msg) = @_; warn("warning: " . localtime() . " $msg"); $DB::trace = 0; $DB::single = 1; } ;
sub useTrace()
{ # Avoid the perl -w warning about a possible typo for these three; this sub is never called.
  print($DB::trace, $DB::single, $DB::signal);
}

use Mersenne;
use MersDB;
use ECMSave;
use MersInfo;

sub init(); # initialize some "global" variables and set up the server socket
sub addData(@); # add the given array (exponent, letter, fields) to %data.
sub addGap($@); # create a trial factoring gap based on factor found by ECM, P-1, P+1, etc.
sub addWork; # add the given arguments (exponent, letter, fields) as work to %work.
sub checkSaveFile($$); # check a save file returned by a client against a prior one
sub closeClient($$); # close the client on the given file descriptor; do not send 'done' if 2nd arg true
sub newClient($); # accept a new client given the server's file handle
sub readClient($); # read data from client on given file descriptor
sub release($$$); # check args, call MersDB::release(), and adjust 'lastPri' in %clientTypes as needed
sub saveMers($); # save the given Mersenne instance to $mersDB
sub sendWork($); # send work to the client on given file descriptor

# These are set in init().
my (%allowed); # map from UUID to mers letters allowed to work on and max tasks outstanding
my (%UUIDaliases); # map from an alias (host name, IP address, UUID, etc.) to the UUID
my (%lastContact); # map from UUID to the time() a client using it last connected

# The UUID for this server itself.
my ($myUUID) = 'mersServer';

my ($usage) = "$0 [-[hH?]] [--help] [--check] [--version] [-p <port>] [--debug <value>]\n"
  . "[--saveFileRetainTime <days>]\n"
  . "[-s <fileByPriority>] [-S <fileByExponent>] [-P <fileToPrint>] [<fileToRead>]\n";
my ($checkConfig) = scalar(grep { m/^--check$/ } @ARGV)
  if @ARGV;
my ($version) = MersInfo::version();
print("$0 version $version\n");
exit(0) if scalar(grep { m/^--version$/ } @ARGV);
print(" starting at " . localtime() . " with command line:\n"
      . "  $0 " . join(' ', @ARGV) . "\n");

my ($servDir) = 'serv.sav';
my ($stopDir) = "$servDir/stop"; # exit if this exists while no clients or children are active
my ($ecmSaveDir) = ECMSave::subdir($myUUID); # directory for the (Linux) ecm program's save files
my ($saveFileRetainTime);

my ($bigintPattern) = MersInfo::patterns('bigint');
my ($bigintPartial) = $bigintPattern;
$bigintPartial =~ s,^\^,,;
$bigintPartial =~ s,\$$,,;
my ($uuidPattern) = MersInfo::patterns('uuid');
my ($uuidPiece) = $uuidPattern;
$uuidPiece =~ s,^\^,,;
$uuidPiece =~ s,\$$,,;

my (%inputToClientType); # key is mers letter; value is key in %clientTypes; set in init() from %clientTypes
my (%clientTypes) = # key is client type, value is related info
  ( # %clientTypes has to be set before parsing the command line since that can modify parts of this.
    'OneStrs' => { # 1strs child that increases mers X limits; only one will exist at a time.
      'child' => 1,
      # 'script' is the executable to run (for 'OneStrs', to extend the X data).
      #    This executable finds the exponent for (prime) factors in factor order, which is
      #    "backwards"/"inverted" compared to typical trial factoring programs, which search
      #    for factors of a particular Mersenne based on the exponent.
      'script' => '1strs8gmp',
      # This 'incr' is the difference between the initial prime (the last one from a prior run)
      #    to the ending prime; 1strs8gmp will usually go a bit farther.
      # 'incrFile' is a file that might contain the 'incr' from a prior run.
      'incr' => 1024, 'incrFile' => "$servDir/tmp/1strs-incr.txt",
    },
    'SPRPChecker' => { # Runs Pseudo-prime (PRP) tests of cofactors.
      # This is started in doSPRP() as internal work rather than given to clients.
      #   The database's 'sprp' table is queried for the smallest Mersenne that needs SPRP check(s).
      #   Also see the checkSPRP() internal work sub which adds missing exponents to the 'sprp' table.
      #   runSPRP.pl is used rather than sprpgmp (or similar) directly because runSPRP.pl updates
      #     the db rather than it's caller having to do that.  This includes when the digit count
      #     already matches the cofactor length, implying that only the 'sprp' 'needed' value in the
      #     database needs to be updated (i.e., the implied SPRP check has already been done).
      'child' => 2, 'script' => 'runSPRP.pl',
    },
    'CheckComposites' => { # Check composite factors, mostly by trying to factor them.
      'child' => 3, 'script' => 'factor.pl',
    },
    'CheckEcmSaveFiles' => { # Check a Mersenne's ecm program save files.
      # Unlike other children, this one is disk I/O bound not CPU bound, so allow it to run in parallel with other children.
      'child' => 4, 'script' => 'checkECMfiles.pl',
    },
    'DataFile' => { # a file on disk (or some other kind of client we can only read from)
      'input' => 1, # input, no need to print, could be unsorted
    },
    'DataToPrint' => { # DataFile, but always print what's read from it
      'input' => 0, # input but printed, likely unsorted
    },
    'Pipe' => { # UNIX FIFO/pipe named Work::pipe; otherwise, just like DataFile except it also allows commands to modify server behavior
      'pipe' => '.server.pipe', # must be a file despite the name; non-blocking I/O on a pipe doesn't seem to work
      'contact' => time(), # set on open() attempt and not tried again for 5 minutes after a failure
    },
    'Reconnect' => { # results from a prior disconnecting client: print them and release mentioned exponent(s)
      # presence is needed, but fields aren't
    },
    'SortedByExponent' => { # a file on disk that has work sorted by exponent
      'opt' => '-S', 'input' => 1, 'sort' => 'exp', # 'files' => [], # -S <file>
    },
    'SortedByPriority' => { # a file on disk that has work sorted by the priority scheme of Mersenne::mersPriority() (per mers letter)
      'opt' => '-s', 'input' => 1, 'sort' => 'pri', # 'files' => [], # -s <file>
    },
    'ECMFactorer' => { # Elliptic Curve Method factorer.  Run by doWork.pl rather than by this server.
      'worker' => 'E', 'pattern' => ' [)]([Ee]): ', 'other' => 'ECM', 'script' => 'doECM.pl',
    },
    'GPUFactorer' => { # Trial factoring (typically using a GPU) often from 2^x thru 2^y but perhaps based on k's (from 2*k*p+1)
      #   Also run by doWork.pl rather than by this server.
      'worker' => 'G', 'pattern' => ' [)]([Gg]): ', 'other' => 'GPU', 'script' => 'doGPU.pl',
    },
    'JoiningFactorer' => { # the mersfac* programs that can "join" gaps in trial factoring and do trial factoring itself (HU)
      #   Also run by doWork.pl rather than by this server.
      'worker' => 'G', 'pattern' => ' [)]([GgHU]): ', 'other' => 'Joiner', 'script' => 'doJoiner.pl',
    },
    'PowerOf2JoiningFactorer' => { # Prime95/mprime via AdvancedFactor lines in worktodo.ini, presently via batch.mers, sendServer.pl, etc.
      # This 'worker' would be 'g' but that would complicate some code here and the 'g' clients support the 'G' mers format.
      #   Also run by doWork.pl rather than by this server.
      #!!Prime95/mprime are effectively obsoleted by the GPU factorers mfaktc and mfakto; delete this in favor of GPUFactorer ?
      'worker' => 'G', 'pattern' => ' [)]([Gg]): ', 'other' => '(PowerOf2Joiner|KJoiner)',
      'script' => 'doGPU.pl',
    },
    'NFSFactorer' => { # Number Field Sieve factorer; !!incomplete; should have its own mers letter; similar to Eoq work
    },
    'PMinus1Factorer' => { # P-1 factorer.  Also run by doWork.pl rather than by this server.
      'worker' => 'o', 'pattern' => ' [)](o): ', 'other' => '(P-1|OP-1|OMinus1Factorer)',
      'script' => 'doP+-1.pl',
    },
    'PPlus1Factorer' => { # P+1 factorer.  Also run by doWork.pl rather than by this server.
      'worker' => 'q', 'pattern' => ' [)](q): ', 'other' => '(P[+]1|QP[+]1|QPlus1Factorer)',
      'script' => 'doP+-1.pl',
    },
    'PrimalityProver' => { # General primality prover, for other numbers besides Mersennes (including cofactors of Mersennes).
      # No longer run doWork.pl - uncommon enough to be run "by hand" when desired.
      'worker' => 'S', 'pattern' => ' [)]([LMSd])', 'other' => 'DPrimalityProver',
      'script' => 'doProof.pl', # or doPrimo.pl, which is much faster
    },
    'TrialFactorer' => { # attempts to divide Mersenne numbers by smaller numbers, always in order; support incomplete and not needed while gaps exist
      # Also run by doWork.pl rather than by this server.
      'worker' => 'U', 'pattern' => ' [)]([HU]): ', 'script' => 'doJoiner.pl',
    },
    'MultiTalented' => { # Can do - or request - more than one of the other types during a single connection
      # usually a client's askServer.pl; entry here needed, but no fields
    },
    'NewClient' => { # just connected and hasn't yet given a type and perhaps not a UUID
      # These values are defaults for clients too old to specify their limits and aren't needed here;
      #   see findWork().  All such old clients have been obsolete for a decade or more.
      #'minExp' => 31, 'maxExp' => $MersDB::SQLmaxInt/2, 'minBits' => 0, 'maxBits' => 0,
    },
    'ECMData' =>  { # db query; see getTasksEoqS()
      'db' => 'E', 'pattern' => ' [)](E): ', 'lastPri' => 0.0, 'sort' => 'pri',
    },
    'PMinus1Data' => { # db query; see getTasksEoqS()
      'db' => 'o', 'pattern' => ' [)](o): ', 'lastPri' => 0.0, 'sort' => 'pri',
    },
    'PPlus1Data' => { # db query; see getTasksEoqS()
      'db' => 'q', 'pattern' => ' [)](q): ', 'lastPri' => 0.0, 'sort' => 'pri',
    },
    'ProofData' => { # db query; see getTasksEoqS()
      'db' => 'S', 'pattern' => ' [)](S): ', 'lastPri' => 0.0, 'sort' => 'pri',
    },
  );

my ($debug) = 0; # toggleDebug() toggles the $debugBit of this
my ($debugBit) = 0; # incremented by shiftToggle()
my ($oldDebug) = 0;
my ($oldDebugBit) = 0;
my (%debug) = ( 'misc'  => 0, # miscellany
                'read'  => 1, # readClient() and related
                'pri'   => 2, # priority of work and related
                'rsrv'  => 3, # reservations and related
                'data'  => 4, # checkData(), checkMersData(), and related
                'need'  => 5, # getNeededData() and related
                'work'  => 6, # getTasksEoqS(), checkWork() and related
                'add'   => 7, # addWork() and related
                'internal' => 8, # internal work
                'save'  => 9, # save files of the (Linux) ecm program
  );
my ($maxDebugBit) = 0;
foreach (keys(%debug))
{
  $maxDebugBit = $debug{$_} if $maxDebugBit < $debug{$_};
}

# Save the last debug message's info by function and Mersenne to avoid repeating it.
my (%lastMsg);

my ($mersDB) = MersDB->new();
# This will be the DBI database handle - needed to avoid race conditions
#   when calling MersDB::reserve().
my ($dbh);

my ($reserved0);

END
{ # Release ($myExp, $myUUID, '!') before exiting so other db clients know the server is not running.
  if (defined($mersDB) && $reserved0)
    {
      my (@r) = $mersDB->release(MersDB::specialExponent($myUUID), $myUUID, '!');
      print("could not release(" . MersDB::specialExponent($myUUID) . ", $myUUID, '!') - " . join(' ', @r) . "\n")
        unless @r;
    }
  my ($rc) = defined($mersDB) ? $mersDB->disconnect() : 1;
  $rc or die("disconnect from database failed: " . $mersDB->dbError() . "\n");
}

# Tracks how many times each signal has been seen; see catchSIG() and init().
#   Once is all that matters currently.
my (%seenSIG);
# SIGIO is not always supported/delivered correctly, so init() tries to check
#   that and sets this if SIGIO appears to not work.
my ($sigIObroken);

my ($serverfh); # set in init() using $port
my ($proto) = scalar(getprotobyname('tcp'));
my ($okChars) = '-^@#$%()_+=\[\]\{\}:;.,/ A-Za-z0-9'; # only these are allowed from clients

# All the mers letters in mers format that might be work/tasks to complete; see
#   the comment in MersDB.pm.
my ($workLetters) = $Mersenne::workLetters;
# Comparative priority of the work types.
my (%letterPri) = ( '!' => 0, 'S' => 1, 'L' => 1, 'M' => 1, 'c' => 2, 'q' => 3, 'o' => 4, 'E' => 5, 'G' => 6, 'g' => 6, 'U' => 7, 'H' => 8, '?' => 100, );

# The most tasks a single UUID can have reserved at once for each work type.
my ($maxTasks) = 5000; # Arbitrary; matches old C/C++ server and doWork.pl.

# Check internal work and do some of it unless clients are waiting.
#   Return true if there is something to do.
#   Argument should be false if a client might be waiting.
sub checkInternalWork($);

# Check the (connected) clients, including perhaps sending more work.
#   Unlike other "internal work", this is called every time thru the primary loop.
sub checkClients();

# Data read from clients and not checked yet.  The key is a Mersenne exponent; the values are
#   array refs with the exponent (BigInt), mers letter, and later numeric values as BigInts.
# The internal work sub that checks this is checkMersData().
my (%mersData);

# Get work for clients to do.  Adds to %work, just below, based on %need.
sub getWork();
# Work (tasks) for clients to do.  Key is mers letter, value is array ref of hashes with keys 'exp', 'one', 'two',
#   and 'pri', each array being sorted by the priority of the tasks.
my (%work);
# For most work, there is only one task per exponent.  This tracks the exponents with work,
#   keyed by the exponent with the value being the kinds of available work as a string of mers letters.
my (%workExponents);
# The minimum amount of work (of each kind, by mers letter) to have in hand if feasible.
my ($minTasks) = 10; # arbitrary, but should be fairly small; 255 in old C/C++ server
# Need more work - key is mers letter, like %work; value is count of tasks "needed".
my (%need);

# Need the data for these Mersenne numbers.  Key is exponent; value is a string of the mers letters
#   and the key is deleted once the data has been read.
# Some of the data needed is retrieved in getNeededData(), one of the 'internal work' subs.
my (%needData);

# Need ecm program save files for these Mersenne numbers.  Key is exponent; value is always 1
#   since checkECMfiles.pl will print the save file names for both P-1 and P+1 in the same run.
my ($needSaveFiles) = { }; # empty HASH ref

# List of "best" ecm program save files by Mersenne exponent.
#!!incomplete - this likely needs to be trimmed as internal work
my (%saveFileNames); # key is exponent; value is complicated - see readCheckECMSaveFiles()

# A set of next values for various purposes.  Each purpose uses a different key of this hash.
my (%next);
# $next{'checkComposites'} The next Mersenne exponent that might have composite factors to deal with.
# $next{'checkData'} The next index and mers letter to check in %data
# $next{'checkEcmSaveExp'} The next Mersenne exponent to check save files for
# $next{'checkSPRP'} The next exponent to check in the MersDB 'sprp' table in checkSPRP().
# $next{'checkWork'} The next index and mers letter to check in %work
# $next{'trimM'} The next exponent in %mersennes to possibly "trim" (delete) in trimMersennes().
# $next{'xIndex'} The next index of @xexps (exponents with 'X' values) to "extend".

# The number of seconds a typical child should take ('all')
#   or the time() to next start a specific child (by key of %clientTypes).
my (%childDelay) = ( 'all' => 5*60, ); # 5 minutes

# This contains the connected clients, keyed by file descriptor.
my (%clients);
# Keys within $clients{$fd}:
# 'connect' - the time() the client connected or was created.
# 'exp' - the Mersenne exponent (only some internal work clients).
# 'fh' - file handle (fileno('fh') should always be $fd, the file descriptor).
# 'flag' - the flag (key of %clientTypes) of the client (e.g., SPRPChecker, ECMFactorer, etc.).
# 'free' - list (HASH ref) of Mersenne exponents the client is done with.
# 'input' - present if the client is input only, for the server to read (e.g., results.gaps.sort).
# 'kind' - the kind (type) of work the client wants to do (a key of %clientTypes).
# 'lastExponent' - Mersenne exponent last read from the client.
# 'lastLetter' - mers latter of the last work or task read from the client.
# 'lastPri' - priority of the last work or task read from the client (input only).
# 'limits' - restrictions the client has specified about work it wants or can perform.
# 'mersenne' - Mersenne object/instance being worked on by the client (only SPRPChecker clients presently).
# 'name' - the name of the client (file name, host name, UUID, etc.).
# 'partialLine' - last input read from the client with no newline or carriage return.
# 'renew' - list (HASH ref) of Mersenne exponents for which the client wants to renew its reservation.
# 'reservations' - count of the client's reserved Mersennes.
# 'save' - boolean: should the Mersenne instance be saved to the database? (SPRPChecker only).
# 'savefile' - info (HASH ref) about a savefile from the client (ECM/gmp-ecm only presently).
# 'tasks' - ARRAY ref of the tasks reserved for the client but not yet sent to it.
# 'uuid' - UUID of the client (usually a copy of 'name' if a 'input' client (file, e.g.)).

# Mersenne instances keyed by the exponent, partly for interacting with clients.
#   Only for smaller exponents, up to the lowest exponent with no 'c' (SPRP) data.
#   Data for larger exponents is not retained except in %data, just below.
my (%mersennes);

# Data for Mersennes that doesn't belong in %mersennes, mostly related to trial factoring gaps.
#   Key is mers letter (as for %work) but the values are hash refs, keyed by exponent.
#   Since a particular Mersenne number can have more than one join, that value is an
#   array of hash refs like the ones in %work:
#     $data{'J'}{$exp}[0]{ 'exp' => $exp, 'one' => $start, 'two' => $stop, }
my (%data);
# The list of exponents for which GgJHU (gap, join, extent) data has been read via MersDB::find().
#   The value is actually the mers letters (among GgJHU) that have been read for the exponent.
my (%joinExps);

# A list of Mersenne exponents with composite factors.
my (%composites); # key is exponent; value is 1 or 0 (or not present, same as 0)

# The vector of bits of the open file descriptors select() should check.
my ($fdbitsIn) = '';

init();

exit(0)
  if $checkConfig;

my ($delay) = !checkClients();
while (scalar(grep { !exists($clients{$_}{'input'}) } keys(%clients)) > 0
       || (!$seenSIG{'INT'} && !-d("$stopDir")))
{
  print(" debug changed from $oldDebug to $debug\n")
    if $oldDebug != $debug;
  $oldDebug = $debug;
  print(" debug bit changed from $oldDebugBit to $debugBit\n")
    if $oldDebugBit != $debugBit;
  $oldDebugBit = $debugBit;
  my ($fdbitsOut) = ''; # if still set after select(), active file descriptors, ready for read() or write()
  $seenSIG{'IO'} = $sigIObroken; # 0 or 1, depending on a test in init() and another two lines below
  my ($count) = select($fdbitsOut = $fdbitsIn, undef, undef, $delay);
  $sigIObroken = 1 if $count && !$seenSIG{'IO'};
  print("select() returned $count, errno $!\n")
    if $count < 0 && $!;
  newClient($serverfh)
    if $count > 0 && vec($fdbitsOut, fileno($serverfh), 1);
  if ($count > vec($fdbitsOut, fileno($serverfh), 1))
    {
      foreach (grep { vec($fdbitsOut, $_, 1) } keys(%clients))
        {
          print("calling readClient fd $_\n") if $debug & (1 << $debug{'read'});
          my ($rc) = readClient($_);
          $rc = '<undef>' unless defined $rc;
          print("readClient fd $_ returned $rc\n") if $debug & (1 << $debug{'read'});
        }
    }
  $delay = 15;
  $delay = 1 if checkInternalWork($count == 0) || -d("$stopDir");
  $delay = 0 if checkClients();
}
print(" exiting at " . localtime() . "; seen SIGINT is " . (defined($seenSIG{'INT'}) ? "$seenSIG{'INT'}" : '<undef>') . "\n");
rmdir("$stopDir") if -d("$stopDir");
exit(0);

sub newClient($)
{
  my ($serverfh) = shift;
  my ($clientfh, $clientAddr);
  #!!if $!{'EMFILE'} (too many open files), close the oldest (data?) client?
  print("accept() of new client failed, $!"),
    return
    unless defined($clientAddr = accept($clientfh, $serverfh));
  my ($port, $iaddr) = sockaddr_in($clientAddr);
  my ($name) = gethostbyaddr($iaddr, AF_INET);
  $name = inet_ntoa($iaddr) if !$name;
  print(" connection from $name [" . inet_ntoa($iaddr) . "] on port $port fd " . fileno($clientfh) . ' at ' . localtime() . "\n");
  return
    unless unblock($clientfh, "$name");
  $clientfh->autoflush(1);
  vec($fdbitsIn, fileno($clientfh), 1) = 1;
  $clients{fileno($clientfh)} = { 'fh' => $clientfh, 'name' => $name, 'flag' => 'NewClient', 'connect' => time(), 'tasks' => [], };
  #!!new recvClient child that doesn't print anything back to the server until the client is done?
  #!!  it would slurp everything from the client before sending anything back to the client or interacting with the db or this server
  #!!an alternative would be to acknowledge and update the db after each line from the client readClient() and sendServer.pl would
  #!!  have to read that acknowledgement or complaint (e.g., not reserved) before sending the next line
  #!!another possibility would be a recvClient.pl that listen()s on the server port, slurps the data from the client, and saves that
  #!!  data in a file for this server to read later at leisure; clients would ask for work on a different TCP port
  return;
}

sub readSPRPClient($)
{ # Only for SPRPChecker clients, which are created by doSPRP() (internal work).
  #   The runSPRP.pl script now has nearly all the logic, making much of this redundant.
  my ($fd) = @_;
  return(0) unless $clients{$fd}{'flag'} eq 'SPRPChecker';
  local $_;
  my ($script) = $clientTypes{'SPRPChecker'}{'script'};
  my ($fh) = $clients{$fd}{'fh'};
  my ($m);
  while (<$fh>)
    {
      print;
      chomp;
      next if !($_) || m,^[#],; # skip comments and empty lines
      if (m/^M[(] ([1-9][0-9]*) [)]C: ([1-9][0-9]*)(.*)/)
        {
          my ($n) = Math::BigInt->new("$1");
          my ($factor) = Math::BigInt->new("$2");
          my ($rest) = $3;
          next unless $factor % $n == 1 && $factor > $n;
          $clients{$fd}{'mersenne'} = Mersenne->new('n', $n)
            unless $clients{$fd}{'mersenne'};
          $m = $clients{$fd}{'mersenne'};
          print("$script printed info for more than one Mersenne\n"),
            next
            if $m && $m->n() != $n;
          $m = Mersenne->new($n)
            unless $m;
          $m->f($factor);
          $clients{$fd}{'save'} = 1;
          $needData{$n} .= 'C' if $needData{$n} && $needData{$n} !~ m,C,;
          $needData{$n} = 'C' if !$needData{$n};
          # M( 21950 )C: 18537362569772451225870001 # factor of given factor
          next if !$rest || $rest =~ m/ *# factor of given factor\b/;
          if ($rest =~ m/ *# composite factor given\b/)
            { # M( 46991 )C: 130848429180565527354835801 # composite factor given and the prime
              # ... factors of this composite factor were not recorded, so track the exponent so
              # ... checkComposites() can do that; factoring it here might stall things too much.
              $composites{$n} = 1;
              next;
            }
        }
      print("unexpected output from $script: $_\n"),
        next
        unless m/^M[(] ([1-9][0-9]*) [)]([clmLMS]):? ([1-9][0-9]*)? */;
      my ($n) = Math::BigInt->new("$1");
      my ($l) = $2;
      my ($len) = $3;
      $clients{$fd}{'mersenne'} = Mersenne->new('n', $n)
        unless $clients{$fd}{'mersenne'};
      $m = $clients{$fd}{'mersenne'};
      print("$script printed info for more than one Mersenne\n"),
        next
        if $m && $m->n() != $n;
      print("unexpected output from $script for $n: $_\n"),
        next
        unless defined $len;
      $m->f($mersennes{$n}->f())
        if exists $mersennes{$n};
      $m->parse("M( $n )$l" . ': ' . "$len\n")
        or print("parsing output of $script failed:\n" . $m->errors());
      $clients{$fd}{'save'} = 1;
    }
  my ($err) = $!;
  my ($retryErr) = 1 if $!{'EWOULDBLOCK'} || $!{'EAGAIN'} || $!{'EINTR'};
  print("reading from $clients{$fd}{'name'} fd $fd failed, $!\n")
    if $err && !$retryErr && !($!{'ECONNRESET'}) && !($!{'EPIPE'}); # unexpected error; ECONNRESET and EPIPE are common but imply the client is gone
  my ($more) = defined($_);
  if ($m)
    {
      saveMers($m)
        if $clients{$fd}{'save'};
      my ($v) = $mersDB->value('sprp', $m->n());
      if (!defined($v))
        {
          print("getting M" . $m->n() . " sprp from MersDB::value failed\n"
                . "  " . $mersDB->dbError());
        }
      elsif (!exists $mersennes{$m->n()} || $v == 2)
        {
          $mersennes{$m->n()} = $m;
        }
      elsif (exists($mersennes{$m->n()}))
        {
          $mersennes{$m->n()}->merge($m);
        }
    }
  return(0)
    if ($more && !$err) || ($err && $retryErr);
  my ($t) = time();
  my ($diff) = $t - $clients{$fd}{'connect'};
  $childDelay{$clients{$fd}{'flag'}} = $t + $diff;
  closeClient($fd, 1);
  return(undef);
}

sub readXClient($)
{ # Only for 'OneStrs' clients, which should only print mers format
  #   factors and 'X' lines and not read anything.
  my ($fd) = @_;
  my ($fh) = $clients{$fd}{'fh'};
  local $_;
  while (<$fh>)
    { # M( <exp> )[CX]: <value> # comment
      print("unexpected data from Child1strs on fd $fd: '$_'\n"),
        closeClient($fd, 1),
        return(undef)
        unless m/^M[(] ([1-9][0-9]+) [)]([CX]): ([0-9]+)\b/;
      my ($exp) = Math::BigInt->new("$1");
      my ($l) = "$2";
      my ($rest) = Math::BigInt->new("$3");
      print;
      if ($l eq 'X')
        {
          my (@r) = $mersDB->update('X', $exp, 'X', $rest);
          print("updating db with M$exp X $rest failed\n"
                . "  " . $mersDB->dbError())
            unless @r && $r[0];
          next;
        }
      # Factors of larger Mersennes are not saved in the db.
      # Factors of smaller Mersennes need to be saved to the db in order
      #   all at once which is only feasible via %mersennes.
      my ($lmec) = $mersDB->lowestMissingExponent('c');
      if (!defined($lmec) || $exp > $lmec)
        {
          return(0) if $seenSIG{'IO'} && !$sigIObroken;
          next;
        }
      my ($v) = $mersDB->value('done', $exp);
      next
        if defined($v) && $v !~ m,^[LMS]$,;
      $mersennes{$exp} = Mersenne->new('n' => $exp)
        unless exists $mersennes{$exp};
      $needData{$exp} .= 'C' if $needData{$exp} && $needData{$exp} !~ m,C,;
      $needData{$exp} = 'C' if !$needData{$exp};
      $mersennes{$exp}->f($rest);
      saveMers($mersennes{$exp});
      print("update sprp $exp 1 failed, " . $mersDB->dbError())
        unless $mersDB->update('sprp', $exp, 1);
      return(0) if $seenSIG{'IO'} && !$sigIObroken;
    }
  return(0) if !defined($_) && ($!{'EWOULDBLOCK'} || $!{'EAGAIN'} || $!{'EINTR'});
  print("reading from $clients{$fd}{'name'} UUID $clients{$fd}{'uuid'} fd $fd failed, $!\n")
    if $!;
  closeClient($fd, 1);
  return(undef);
}

sub readCheckEcmSaveFilesClient($)
{ # Only for CheckEcmSaveFiles clients, which are created by checkEcmSaveFiles() (internal work).
  my ($fd) = @_;
  return(0) unless $clients{$fd}{'flag'} eq 'CheckEcmSaveFiles';
  local $_;
  my ($script) = $clientTypes{'CheckEcmSaveFiles'}{'script'};
  my ($exp) = $clients{$fd}{'exp'};
  my ($expDir) = $exp % $ECMSave::expMod;
  my ($fh) = $clients{$fd}{'fh'};
  my ($pattern) = $ECMSave::filePatterns[-1];
  while (<$fh>)
    {
      chomp;
      next unless $_;
      print("$_\n"), # Most comments are warnings from ECMSave::new() and findBestSaveFile().
        next
        if m,^[#],;
      print("# ignored line without leading P-1 or P+1 bound from $clients{$fd}{'name'} fd $fd - '$_'\n"),
        next
        unless s,^ *$bigintPartial *,,;
      my ($bound) = Math::BigInt->new("$1");
      print("# ignored unexpected characters in ecm save file name from $clients{$fd}{'name'} fd $fd - '$_'\n"),
        next
        unless s,/($pattern)$,,;
      my ($file) = $1;
      print("# exponent $exp not in file name from child $clients{$fd}{'name'} fd $fd - '$file'\n"
            . "#   original file $_\n"),
        next
        unless $exp == $2;
      my ($method) = "$3"; # 'pm1' (P-1) or 'pp1' (P+1)
      my ($M) = "$4"; # '' (c or l cofactor) or '.M' (m cofactor)
      my ($subdir) = $_;
      my (@dir) = grep { (substr($subdir, 0, index($subdir, '/') + 1)) eq "$_/" } ECMSave::subdirs();
      push(@dir, '.')
        if !@dir && substr($_, 0, 1) ne '/' && grep { $_ eq '.' } ECMSave::subdirs();
      print("# logic error or unexpected directory from child $clients{$fd}{'name'} fd $fd - '$_/$file'\n"
            . "#   directories - " . join(' ', @dir) . "\n"),
        next
        unless ($#dir == 0 && -d($dir[0])
                && ((substr($_, 0, length($dir[0])) . '/' eq "$dir[0]" . '/')
                    || ($dir[0] eq '.' && substr($_, 0, 1) ne '/')));
      my ($final) = $_ . '/';
      $final = substr($_, length($dir[0]) + 1) . '/'
        if $dir[0] ne '.';
      $dir[0] .= "/$expDir"
        if -d "$dir[0]/$expDir" && $final =~ s,^$expDir/,,;
      print("# logic error or unexpected subdir from child $clients{$fd}{'name'} fd $fd - '$dir[0]/$file'\n"
            . "#   original file $subdir/$file\n"),
        next
        unless -s "$dir[0]/$file";
      my ($oq) = $method eq 'pm1' ? 'o' : 'q';
      my ($clm) = ($exp % 8 == 4) ? ($M ? 'm' : 'l') : 'c';
      $final = "$dir[0]/$final";
      if (exists($saveFileNames{$exp}{$oq}{$clm}))
        {
          my ($ref) = $saveFileNames{$exp}{$oq}{$clm};
          delete $saveFileNames{$exp}{$oq};
          $final = $ref->{'file'},
            $bound = ($ref->{'bound'})
            if $ref->{'file'} ne $final && $ref->{'bound'} > $bound;
        }
      $saveFileNames{$exp}{$oq}{$clm} = { 'file' => "$final", 'bound' => $bound, };
      delete $needSaveFiles->{$exp};
    }
  my ($err) = $!;
  my ($retryErr) = 1 if $!{'EWOULDBLOCK'} || $!{'EAGAIN'} || $!{'EINTR'};
  print("reading from $clients{$fd}{'name'} fd $fd failed, $!\n")
    if $err && !$retryErr && !($!{'ECONNRESET'}) && !($!{'EPIPE'}); # unexpected error; ECONNRESET and EPIPE are common but imply the client is gone
  my ($more) = defined($_);
  return(0)
    if ($more && !$err) || ($err && $retryErr);
  my ($t) = time();
  my ($diff) = $t - $clients{$fd}{'connect'};
  $childDelay{$clients{$fd}{'flag'}} = $t + $diff;
  my (@r) = release($clients{$fd}{'exp'}, $myUUID, '!');
  warn("checkEcmSaveFile - could not release $clients{$fd}{'exp'} $myUUID '!' - " . $mersDB->dbError())
    if $#r < 0 || !$r[0];
  closeClient($fd, 1);
  return(undef);
}

sub readClient($)
{ # return undef if the client is no longer or "bad"; 0 if nothing read (data clients only); anything 'true' otherwise
  my ($fd) = shift;
  local $_;
  print("starting readClient fd $fd\n") if $debug & (1 << $debug{'read'});
  print("readClient called with fd $fd, not in clients HASH\n"),
    return(undef)
    unless exists $clients{$fd};
  print("missing file handle for client on fd $fd\n"),
    closeClient($fd, 1),
    return(undef)
    unless exists($clients{$fd}{'fh'});
  print("missing 'flag'/kind for client on fd $fd\n"),
    closeClient($fd, 1),
    return(undef)
    unless exists($clients{$fd}{'flag'});
  print("client kind '$clients{$fd}{'flag'}' not in clientTypes on fd $fd\n"),
    closeClient($fd, 1),
    return(undef)
    unless exists($clientTypes{$clients{$fd}{'flag'}});
  return(readSPRPClient($fd)) if $clients{$fd}{'flag'} eq 'SPRPChecker';
  return(readXClient($fd)) if $clients{$fd}{'flag'} eq 'OneStrs';
  return(readCheckEcmSaveFilesClient($fd)) if $clients{$fd}{'flag'} eq 'CheckEcmSaveFiles';
  my ($inputOnly) = (exists($clients{$fd}{'input'}) || exists($clientTypes{$clients{$fd}{'flag'}}{'input'})
                     || exists($clientTypes{$clients{$fd}{'flag'}}{'db'}));
  $inputOnly ||= exists($clients{$fd}{'uuid'}) && $clients{$fd}{'uuid'} =~ m/^((GIMPS|factordb)-$uuidPiece)$/;
  my ($printInput) =  (!exists($clients{$fd}{'input'}) || !$clients{$fd}{'input'}
                       || !exists($clientTypes{$clients{$fd}{'flag'}}{'input'})
                       || !$clientTypes{$clients{$fd}{'flag'}}{'input'});
  return(0)
    if (exists($clientTypes{$clients{$fd}{'flag'}}{'sort'}) && $clientTypes{$clients{$fd}{'flag'}}{'sort'} eq 'pri'
        && exists($clients{$fd}{'lastLetter'}) && exists($work{$clients{$fd}{'lastLetter'}})
        && $need{$clients{$fd}{'lastLetter'}} < 1
        && $inputOnly && !$printInput
        && (($#{$work{$clients{$fd}{'lastLetter'}}} >= $minTasks
             && $clients{$fd}{'lastPri'} > $work{$clients{$fd}{'lastLetter'}}[$minTasks]->{'pri'})
            # Nearly all gaps come from an input file and aren't given to addWork() directly,
            #   so don't read more if %mersData is "large".
            || ($clients{$fd}{'lastLetter'} eq 'G'
                && (scalar(keys(%mersData)) > 10*$minTasks
                    || scalar(keys(%{$data{'G'}})) > 2*$maxTasks))));
  my ($lmec) = $mersDB->lowestMissingExponent('c');
  warn("readClient - MersDB::lowestMissingExponent 'c' returned '" . (defined($lmec) ? $lmec : 'undef')),
    return(0)
    unless $lmec;
  return(0) # this data client is sorted by exponent
    if (exists($clients{$fd}{'sort'}) && $clients{$fd}{'sort'} eq 'exp'
        && $clients{$fd}{'lastExp'} > $lmec);
  # This ($done) is used to return early if new work similar to the prior two
  #   conditions is read.  Returning early will happen if $done > $minTasks
  #   so that $sigIObroken can be dealt with without throttling all clients to
  #   one line per call.
  my ($done) = 0;
  my ($fh) = $clients{$fd}{'fh'};
  my ($mers); # Mersenne instance
  my ($exp); # last Mersenne exponent read in a mers format line
  my ($partialExp, $partialEndJoin);
  print("readClient about to start reading fd $fd\n") if $debug & (1 << $debug{'read'});
  $clients{$fd}{'partialLine'} = ''
    unless $clients{$fd}{'partialLine'};

  while (<$fh>)
    { # If no linefeed or carriage return at the end, read more.
      $clients{$fd}{'partialLine'} .= $_,
        next
        unless m/[\n\r$CRLF]+$/;
      # If already have part of a line, add to it.
      $_ = $clients{$fd}{'partialLine'} . $_,
        $clients{$fd}{'partialLine'} = ''
        if $clients{$fd}{'partialLine'};
      if (exists($clients{$fd}{'savefile'}) && ! m,^save file end ,)
        { # The client is part way thru sending the contents of an ecm save file;
          #   don't modify the content at all here but restrict it for safety.
          $clients{$fd}{'savefile'}{'content'} = '' unless exists $clients{$fd}{'savefile'}{'content'};
          my ($str) = $_; # save exact string - the save file format is unknown but ...
          s/[\n\r$CRLF]+//;
          s/[-+=;:.@ A-Za-z0-9]+//g; # ... don't allow other characters in it.
          if ($_)
            {
              saveMers($mers)
                if $mers;
              warn("save file '$clients{$fd}{'savefile'}{'file'}' from fd $fd has unexpected content");
              closeClient($fd, 1);
              return(undef);
            }
          $clients{$fd}{'savefile'}{'content'} .= "$str";
          next;
        }
      # The rest of the save file support is these three if() statements.
      $clients{$fd}{'savefile'} = { 'file' => $1, 'exp' => $2, 'method' => $3, 'M' => $4, },
        next
        if m/^save file begin (ecm[.]([1-9][0-9]+)[.](p[mp]1)([.]M|))[\n\r$CRLF]+$/;
      if (m/^save file end ([a-z][-._a-zA-Z0-9]+)[\n\r$CRLF]$/)
        {
          my ($sfe) = "$1";
          if (!exists($clients{$fd}{'savefile'}))
            {
              saveMers($mers)
                if $mers;
              print("'save file end' from fd $fd with no prior 'save file begin'\n");
              closeClient($fd, 1);
              return(undef);
            }
          if ($sfe ne $clients{$fd}{'savefile'}{'file'})
            {
              saveMers($mers)
                if $mers;
              print("save file end from fd $fd gives different file name, '$sfe', not the expected '$clients{$fd}{'savefile'}{'file'}'\n");
              closeClient($fd, 1);
              return(undef);
            }
          # Compare any existing old save file with the new one from the client.
          #   All the problems with save files except one are the server's fault and the one could be.
          checkSaveFile($fd, $mers && $mers->n() == $clients{$fd}{'savefile'}{'exp'} ? $mers : undef);
          delete $clients{$fd}{'savefile'};
          next;
        }
      if (m/^save file /)
        {
          s,\n,\\n,g;
          s,\r,\\r,g;
          warn("fd $fd name $clients{$fd}{'name'} unexpected 'save file' line - '$_'");
          saveMers($mers)
            if $mers;
          closeClient($fd, 1);
          return(undef);
        }

      s/\t+/ /g; # change tabs to a space
      s/  +/ /g; # drop redundant spaces
      s/^ //;
      s/[ \n\r$CRLF]+$//;
      next unless $_; # ignore blank lines

      #!!Join lines that end with a single backslash?  But never within save files, which should never be modified.
      # The GPU factoring programs (mfaktc and mkfakto) can be configured to stop after finding a factor,
      #   in which case they indicate not checking until the next power of two with a '*' ... which happens
      #   to be a good character to filter from untrusted input, so check for that '*' first.
      $partialExp = "$1",
        $partialEndJoin = "$2"
        if m/^M[(] ([1-9][0-9]*) [)]C: ([1-9][0-9]*) [#].*[*][:]mfakt/; # literal '*' implies mfakt* stopped on finding the factor
      if (!m/^[$okChars]*$/ && !$partialExp)
        {
          saveMers($mers)
            if $mers;
          print("unexpected characters from $clients{$fd}{'name'} UUID $clients{$fd}{'uuid'} fd $fd\n");
          closeClient($fd, 1);
          return(undef);
        }
      # askServer.pl and sendServer.pl end with 'done', so call closeClient() (which sends 'done')
      #!!should send 'drop' commands before or in closeClient()
      $done = $minTasks,
        $inputOnly = 1,
        $_ = undef,
        $! = 0,
        last
        if m/^done$/i;
      if (m/^M[(] ([1-9][0-9]*) [)](.)(.*)$/)
        { # Mers format - print _new_ data except 'X' (and thus nothing from input files).
          next if $2 eq 'X'; # ignore 'X' data except from Child1strs or the database
          my ($exp) = Math::BigInt->new("$1");
          my ($l) = "$2";
          my ($rest) = $3;
          # Trial factoring, including gaps and joins, only makes sense for Mersennes with a prime exponent.
          #!!warn the first time this happens for each UUID?
          next unless $l !~ m/^[GgHJKU]$/ || Math::Prime::Util::is_prime($exp);
          if ($partialExp && $exp == $partialExp)
            { # need text after the '#' starting a comment:
              # M( 206155927 )C: 10164264862472986173257 # [TF:73:74*:mfakto 0.15pre7 cl_barrett15_74_gs_2]
              if (m/[#] *[[]TF:([1-9][0-9]*)[:]([1-9][0-9]*)[*][:]mfakt.*$/)
                { # Trial factoring was stopped at or near the factor, not at the (larger) power of two
                  my ($comment) = "$0";
                  my ($low) = Math::BigInt->bone()->blsft("$1");
                  my ($hi) = Math::BigInt->bone()->blsft("$2");
                  print("unknown format/unexpected values - $_\n"),
                    $partialExp = $partialEndJoin = undef,
                    next
                    unless $low < $hi && $low < $partialEndJoin;
                  print("M( $exp )J: $low $partialEndJoin $comment\n");
                  $partialExp = $partialEndJoin = undef;
                  next;
                }
            }
          else
            {
              $partialExp = $partialEndJoin = undef;
            }
          $rest =~ s/#.*//; # drop comments before trying to parse
          $rest =~ s/  *$//; # drop trailing spaces
          $rest =~ s/^  *//; # drop leading spaces
          my (@rest) = split(/ /, $rest);
          # J lines can also be in a power of two format like:
          # M( 201580037 )J: 2^71 2^72 # [mfaktc 0.21 barrett76_mul32_gs]
          if ($l eq 'J' && $rest[0] eq ':' && $rest[1] =~ m/^2\^([1-9][0-9]*)$/)
            {
              $rest[1] = Math::BigInt->bone()->blsft("$1");
              print("unknown format/unexpected data - $_\n"),
                next
                unless $rest[2] =~ m/^2\^([1-9][0-9]*)$/;
              $rest[2] = Math::BigInt->bone()->blsft("$1");
              # Need to print in the normal format in case whatever reads the output file to
              #   update the flat files does not understand the power of two style.
              print("M( $exp )J: $rest[1] $rest[2]");
              foreach my $i (3 .. $#rest)
                { # these should all be comment text, not mers data per se
                  print(" $rest[$i]") if $rest[$i];
                }
              print("\n");
            }
          # Cannot print the input until here due to the format check inside prior if() block.
          #   Printing $_ could be skipped here in the power of two style case but it's better
          #   to always print exactly what was read just in case.
          print("$_ # fd $fd $clients{$fd}{'name'} $clients{$fd}{'uuid'} " . scalar(localtime()) . "\n")
            if $printInput;
          # K is like J but k-values from 2*k*p+1 restriction on possible factors rather than trial factors per se.
          #   K input is almost always accompanied by matching J input, but ...
          my ($might2) = ($l =~ m/^[EeGgJKoq]$/); # might have two values
          my ($need2) = ($l =~ m/^[EGgJK]$/); # eoq might have two but don't require two
          if (@rest && ($rest[0] ne ':' || $rest[1] !~ m/^[0-9]+$/ || ($need2 && $#rest < 2)
                        || ($#rest > 1 && $might2 && $rest[2] !~ m/^[0-9]+$/)))
            {
              print({$clients{$fd}{'fh'}} "unexpected characters after the mers letter - '$_'\n")
                if exists $clientTypes{$clients{$fd}{'flag'}}{'worker'};
              print("unexpected characters after the mers letter - '$_'\n");
              saveMers($mers)
                if $mers;
              closeClient($fd, 1);
              return(undef);
            }
          shift(@rest); # drop the ':'
          $rest[0] = Math::BigInt->new("$rest[0]") if @rest;
          if ($#rest > 0 && $might2)
            {
              $rest[1] = Math::BigInt->new("$rest[1]");
              $#rest = 1; # G data sometimes has a priority value after the stop value
              # Convert K (k) format to J (2*k*exp+1)
              $rest[0]->bmul($exp)->bmul(2)->binc(),
                $rest[1]->bmul($exp)->bmul(2)->binc(),
                $l = 'J',
                print("M( $exp )J: $rest[0] $rest[1] # K to J\n")
                if $l eq 'K';
            }
          else
            { # ensure @rest has at most one value
              $#rest = 0 if @rest;
            }
          $clients{$fd}{'free'}{$exp} = 1 unless $l eq 'C' || $inputOnly;
          if ($clients{$fd}{'input'} || $clientTypes{$clients{$fd}{'flag'}}{'input'})
            {{ # This is only needed for data clients, that get (usually sorted) tasks from a local file.
              $done = $minTasks
                if ($clients{$fd}{'flag'} eq 'DataToPrint' # Read data to print one line at a time to avoid "starving" other clients.
                    # Data from these clients is sorted by exponent and can wait until SPRP results catch up:
                    || (exists($clientTypes{$clients{$fd}{'flag'}}{'sort'}) && $clientTypes{$clients{$fd}{'flag'}}{'sort'} eq 'exp'
                        && $exp > $mersDB->lowestMissingExponent('c')));
              my ($pri) = Mersenne::mersPriority($exp, $l, @rest);
              $done = $minTasks,
                last
                if $pri < 0.0;
              $clients{$fd}{'lastExp'} = $exp;
              $clients{$fd}{'lastLetter'} = $l;
              $clients{$fd}{'lastPri'} = $pri;
              print(" client $clients{$fd}{'name'} flag $clients{$fd}{'flag'} fd $fd setting lastPri $pri\n")
                if $debug & (1 << $debug{'pri'});
              my ($min) = $minTasks;
              ($min += $need{$l})
                if $need{$l} > 0;
              $done = $min # data from these clients is sorted by priority
                if (exists($clientTypes{$clients{$fd}{'flag'}}{'sort'}) && $clientTypes{$clients{$fd}{'flag'}}{'sort'} eq 'pri'
                    && exists($work{$l}) && $#{$work{$l}} >= $min && $pri > $work{$l}[$min]->{'pri'});
              ++$done
                if ($clients{$fd}{'lastLetter'} eq 'G'
                    && (scalar(keys(%mersData)) > 2*$min || scalar(keys(%{$data{'G'}})) > $maxTasks));
              if ($done >= $min && (($debug & (1 << $debug{'read'}))
                                    || ($debug & (1 << $debug{'work'}))))
                {
                  print(" no more needed; " . (1 + $#{$work{$l}}) . " '$l' tasks in memory");
                  print(", " . scalar(keys(%{$data{$l}})) . " in $l data")
                    if exists $data{$l};
                  print(", " . scalar(keys(%mersData)) . " in mersData\n");
                }
            }}
          # ECM results (not effort like 'E') is also dealt with as data; the db only supports 'E' for ECM.
          push(@{$mersData{$exp}}, [ $exp, $l, @rest, ]),
            next
            if $l eq 'e';
          if ($l =~ m/^[GgHJU]$/)
            { # Trial factoring gaps and joins are never retained in %mersennes; they are added to %data instead.
              #   But gaps from a flat file (results.gaps.sort, usually) can be given directly to addData().
              print("readClient calling addData " . join(' ', $exp, $l, @rest) . "\n")
                if $debug && (1 << $debug{'data'});
              addData($exp, $l, @rest);
              next;
            }
          if ($l =~ m/^[SEoq]$/)
            { #!!other mers letters?
              my ($aw) = addWork($exp, $l, @rest);
              print("readClient - addWork '$aw' updating db with " . join(' ', $exp, $l, @rest) . "\n")
                if $aw !~ m/^(added|highExp)$/ || $debug & (1 << $debug{'add'});
            }
          # Add the data to the Mersenne instance.
          warn("parsing data for M$exp failed - " . $mersennes{$exp}->errors()),
            delete($mersennes{$exp})
            if exists($mersennes{$exp}) && !$mersennes{$exp}->parse($_);
          saveMers($mers),
            $mers = undef
            if defined($mers) && $mers->n() != $exp;
          ($mers = $mersennes{$exp})
            if !$mers && exists $mersennes{$exp};
          $mers = Mersenne->new('n' => $exp),
            $mersennes{$exp} = $mers
            if !defined $mers && $exp <= $mersDB->lowestMissingExponent('c');
          warn("parsing data for M$exp failed - " . $mers->errors()),
            $mers = undef
            if defined $mers && !$mers->parse($_);
          if ($l eq 'C')
            { # Create a trial factoring gap in case the next line is not a join, 'S', or 'D' for M($exp).
              next unless $rest[0] % $exp == 1;
              $needData{$exp} .= 'C' if $needData{$exp} && $needData{$exp} !~ m,C,;
              $needData{$exp} = 'C' if !$needData{$exp};
              addGap($exp, @rest);
              # Record the exponent if the factor is composite.
              $composites{$exp} = 1
                if $exp <= $mersDB->lowestMissingExponent('c') && !Math::Prime::Util::is_prime($rest[0]);
              next;
            }
          my ($info) = [ $exp, $l, ];
          push(@{$info}, Math::BigInt->new("$_"))
            foreach (@rest);
          push(@{$mersData{$exp}}, $info);
          next;
        }
      elsif ($partialExp)
        { # $partialExp only applies to mers format, just prior
          print("unexpected characters from $clients{$fd}{'name'} UUID $clients{$fd}{'uuid'} fd $fd\n");
          saveMers($mers)
            if $mers;
          closeClient($fd, 1);
          return(undef);
        }
      elsif ($inputOnly)
        { # only allow mers format from 'input' clients
          print("non-mers format data from input only client $clients{$fd}{'name'} fd $fd - '$_'\n");
          saveMers($mers)
            if $mers;
          closeClient($fd, 1);
          return(undef);
        }
      if (!exists($clients{$fd}{'uuid'}))
        {
          if (!m/^UUID /)
            {
              saveMers($mers)
                if $mers;
              print({$clients{$fd}{'fh'}} "send UUID or alias first$CRLF");
              print("client $clients{$fd}{'name'} sent '$_' before UUID or alias\n");
              closeClient($fd, 1);
              return(undef);
            }
          if (!m/^UUID $uuidPiece$/)
            {
              saveMers($mers)
                if $mers;
              print({$clients{$fd}{'fh'}} "invalid UUID line$CRLF");
              print("unexpected characters in UUID line from $clients{$fd}{'name'} on fd $fd\n");
              closeClient($fd, 1);
              return(undef);
            }
          my ($alias) = $1;
          if ($clients{$fd}{'name'} eq 'localhost' && $alias =~ m/^(GIMPS-$uuidPiece)$/)
            { # data from GIMPS via './sendServer.pl -u GIMPS-*' (www.mersenne.org)
              $inputOnly = 1;
              $clients{$fd}{'uuid'} = $1;
              $lastContact{$1} = time();
              $clients{$fd}{'flag'} = 'DataToPrint';
              next;
            }
          if ($clients{$fd}{'name'} eq 'localhost' && $alias =~ m/^(factordb-$uuidPiece)$/)
            { # data from factordb.com via './sendServer.pl -u factordb-*'
              $inputOnly = 1;
              $clients{$fd}{'uuid'} = $1;
              $lastContact{$1} = time();
              $clients{$fd}{'flag'} = 'DataToPrint';
              next;
            }
          if (!exists($UUIDaliases{$alias}) || $UUIDaliases{$alias} !~ m/$uuidPattern/)
            {
              saveMers($mers)
                if $mers;
              print({$clients{$fd}{'fh'}} "unknown UUID or alias$CRLF");
              print("client $clients{$fd}{'name'} sent unknown UUID or alias '$alias'\n");
              closeClient($fd, 1);
              return(undef);
            }
          $clients{$fd}{'uuid'} = $1;
          $lastContact{$clients{$fd}{'uuid'}} = time();
          my (@up) = $mersDB->update('contact', $clients{$fd}{'uuid'}, $lastContact{$clients{$fd}{'uuid'}});
          @up && $up[0]
            or print("updating db table contact for fd $fd UUID $clients{$fd}{'uuid'} failed\n");
          print(" start $clients{$fd}{'name'} UUID $clients{$fd}{'uuid'} alias $alias fd $fd at " . localtime() . "\n");
          next;
        }
      $clients{$fd}{'free'}{$1} = 1,
        next
        if m/^free ([1-9][0-9]+)$/i;
      if (m/^renew ([1-9][0-9]+) *([a-zA-Z])?$/i)
        { # $2 here is the mers letter
          my ($e) = $1;
          my ($l) = $2 ? $2 : '?'; # '?' is a placeholder when the kind is unknown          
          $clients{$fd}{'renew'}{$e} = "$l";
          my (@list) = $mersDB->reservations($e);
          #!!the client should be informed about most of these issues
          if ($#list < 0)
            {
              print("$clients{$fd}{'name'} renewed $e without reserving it\n");
              my (@r) = $mersDB->reserve($e, $clients{$fd}{'uuid'}, $l);
              ++($clients{$fd}{'reservations'}{$l}),
                next
                if @r && $r[0];
              warn("$clients{$fd}{'name'} could not reserve M$e for $l, " . $mersDB->dbError());
              next;
            }
          warn("M$e reserved more than once, by $list[0]->{'uuid'} and $list[1]->{'uuid'}")
            if $#list > 0;
          my ($ref) = pop(@list);
          warn("M$e in reservations has different exponent: M$ref->{'exp'} uuid $ref->{'uuid'} kind $ref->{'kind'}"),
            next
            if $ref->{'exp'} != $e;
          warn("M$e in reservations has no UUID"),
            next
            unless exists $ref->{'uuid'};
          print("$clients{$fd}{'name'} tried to renew $e, reserved by $ref->{'uuid'}\n"),
            next
            if $ref->{'uuid'} ne $clients{$fd}{'uuid'};
          #!!compare $l with $ref->{'kind'} ?
          next;
        }
      if ($clients{$fd}{'flag'} eq 'NewClient')
        { # Should be the kind of work requested.
          foreach my $kind (keys(%clientTypes))
            {
              $clients{$fd}{'flag'} = $kind,
                last
                if (m/^$kind /i || m/^$kind$/i
                    || (exists($clientTypes{$kind}{'other'})
                        && m/^($clientTypes{$kind}{'other'})$/));
            }
          if ($clients{$fd}{'flag'} eq 'NewClient')
            {
              saveMers($mers)
                if $mers;
              print({$clients{$fd}{'fh'}} "unknown client type$CRLF");
              warn("unknown client type on fd $fd - '$_'");
              closeClient($fd, 1);
              return(undef);
            }
          print(" fd $fd now $clients{$fd}{'flag'}\n"),
            next
            if exists $allowed{$clients{$fd}{'uuid'}};
          saveMers($mers)
            if $mers;
          print({$clients{$fd}{'fh'}} "unknown UUID$CRLF");
          warn("$clients{$fd}{'name'} fd $fd gave unknown UUID $clients{$fd}{'uuid'}");
          closeClient($fd, 1);
          return(undef);
        }
      # The client might be asking for a different kind of work or more work.
      my ($kind);
      if (s/^work //)
        { # Next word is the kind of work being requested, then limits/restrictions.
          my (@limits) = split(/ /);
          $kind = shift(@limits);
          while (@limits)
            {
              my ($limit) = shift(@limits);
              $clients{$fd}{'limits'}{'prime'} = 1,
                next # client will only do work related to prime exponents
                if $limit =~ m/^primes?$/i;
              my ($val);
              $val = $1,
                shift(@limits)
                if $limits[0] =~ m/^([0-9]+)$/;
              $clients{$fd}{'limits'}{'min_bits'} = $val,
                next
                if $limit =~ m/^(bits_min|min_bits)$/ && defined $val;
              $clients{$fd}{'limits'}{'max_bits'} = $val,
                next
                if $limit =~ m/^(bits_max|max_bits)$/ && defined $val;
              $clients{$fd}{'limits'}{'min_exp'} = $val,
                next
                if $limit =~ m/^(exp_min|min_exp|start_range|range_start)$/ && defined $val;
              $clients{$fd}{'limits'}{'max_exp'} = $val,
                next
                if $limit =~ m/^(exp_max|max_exp|stop_range|range_stop)$/ && defined $val;
              warn("unknown limit on work by $clients{$fd}{'name'} fd $fd - 'work $_'");
              saveMers($mers)
                if $mers;
              closeClient($fd, 1);
              return(undef);
            }
          s/ .*$//;
          # fall thru - what's left should be a single word - the kind of work requested
        }
      if (!defined($kind))
        {
          foreach my $type (keys(%clientTypes))
            {
              next unless m/^($type)$/i || (exists($clientTypes{$type}{'other'})
                                            && m/^($clientTypes{$type}{'other'})$/i);
              $kind = $type;
              last;
            }
          if (!defined($kind))
            {
              print({$clients{$fd}{'fh'}} "unknown work type$CRLF");
              warn("unknown work type from $clients{$fd}{'name'} UUID $clients{$fd}{'uuid'} fd $fd - '$_'");
              saveMers($mers)
                if $mers;
              closeClient($fd, 1);
              return(undef);
            }
        }
      if (defined($kind))
        {
          $clients{$fd}{'kind'} = $kind;
          $clients{$fd}{'flag'} = $kind if $clients{$fd}{'flag'} eq 'NewClient';
          next;
        }
      print({$clients{$fd}{'fh'}} "unexpected data - '$_'$CRLF");
      warn("unexpected data from fd $fd: '$_'");
      saveMers($mers)
        if $mers;
      closeClient($fd, 1);
      return(undef);
    }
  continue
    { # Read at least one line but don't allow this client to "freeze" out other clients.
      #   But, if $sigIObroken, then read up to $minTasks lines per call.
      $! = 0; # in case some system call other than read() has set it during the loop
      # Always try to read another line after a factor.
      last if ($done >= $minTasks || ($seenSIG{'IO'} && !$sigIObroken)) && (!defined($_) || !m/C:/);
      ++$done if $seenSIG{'IO'};
    }

  my ($err) = $!;
  my ($retryErr) = 1 if $!{'EWOULDBLOCK'} || $!{'EAGAIN'} || $!{'EINTR'};
  print("reading from $clients{$fd}{'name'} UUID "
        . (exists($clients{$fd}{'uuid'}) ? $clients{$fd}{'uuid'} : '<undef>')
        . " fd $fd failed, $!\n")
    if $err && !$retryErr && !($!{'ECONNRESET'}) && !($!{'EPIPE'}); # unexpected error
  my ($more) = defined($_);
  vec($fdbitsIn, $fd, 1) = 0 if $done || $inputOnly; # only read more via checkClients()
  closeClient($fd, 1)
    if $inputOnly && ((!$err && !$more) || ($err && !$retryErr)); # EOF or unexpected error
  saveMers($mers)
    if $mers;
  return(0) if $inputOnly;

  print("readClient returning, err " . (defined($err) ? "$err" : '<undef>')
        . ", more " . (defined($more) ? "$more" : '<undef>')
        . ", retryErr " . (defined($retryErr) ? "$retryErr" : '<undef>')
        . "\n")
    if $debug & (1 << $debug{'read'});
  closeClient($fd, 1), return(undef) if $err && !$retryErr; # unexpected error
  closeClient($fd, 0), return(0) if !$err && !$more; # EOF
  return(0) unless exists $clients{$fd}{'uuid'};
  return(1);
}

sub saveMers($)
{ #!!save to db or retain? both? this actually does both, to some extent, inside $mersDB
  my ($mers) = shift;
  #!!debug info
  warn("package %s called saveMers with undef or invalid Mersenne ref - " . join(' ', caller()) . "\n"),
    return(undef)
    unless $mers && ref($mers) eq 'Mersenne';
  my ($err) = $mers->errors();
  warn("Mersenne for M" . $mers->n() . " errors:\n$err"),
    return(undef)
    if $err;
  my (@saveRtn) = $mersDB->save($mers);
  return(1)
    if @saveRtn && $saveRtn[0];
  warn("could not save info for M" . $mers->n() . " into db: " . $mersDB->dbError());
  return(undef);
}

sub checkSaveFile($$)
{ # All errors but one here are server side and the one - the client trying to
  #   factor the wrong number - could be, so don't blame the client, so no
  #   point in returning error values or OK vs. not - all returns should be undef.
  my ($fd, $mers) = @_;
  warn("createSaveFile called but fd $fd has no 'savefile' info - " . join(' ', caller())),
    return(undef)
    unless exists $clients{$fd} && exists $clients{$fd}{'savefile'};
  my ($info) = $clients{$fd}{'savefile'};
  warn("createSaveFile called but fd $fd missing 'exp', 'method', 'M', or 'content' - " . join(' ', caller())),
    return(undef)
    unless $info->{'exp'} && $info->{'method'} && exists $info->{'M'} && $info->{'content'};
  my ($saveFile) = "$ECMSave::filePrefix" . "$info->{'exp'}.$info->{'method'}$info->{'M'}";
  warn("createSaveFile called but fd $fd 'file' $info->{'file'} does not match 'exp' $info->{'exp'}, 'method' $info->{'method'}, and/or 'M' $info->{'M'} - " . join(' ', caller())),
    return(undef)
    unless $saveFile eq $info->{'file'};
  $info->{'content'} =~ s/[\n\r$CRLF]+$//;
  my ($subdir) = ECMSave::subdir($myUUID, $info->{'exp'});
  $subdir = $ecmSaveDir
    unless -d($subdir) || mkdir($subdir);
  my ($newName) = "$subdir/$saveFile";
  !-l($newName)
    or unlink($newName)
    or die("could not remove soft link $newName, $!\n");
  if (-s("$newName"))
    { # Save the content elsewhere.
      my ($cnt) = 0;
      ++$cnt
        while -s "$newName.$cnt";
      $newName .= ".$cnt";
    }
  print("new ecm save file '$newName'\n")
    if ($debug & (1 << $debug{'save'}));
  my ($newfh);
  warn("could not create $newName, $!"),
    print("save file begin $saveFile\n$info->{'content'}\nsave file end $saveFile\n"),
    return(undef)
    unless open($newfh, '>', $newName);
  warn("could not print into $newName, $!"),
    print("save file begin $saveFile\n$info->{'content'}\nsave file end $saveFile\n"),
    unlink("$newName"),
    return(undef)
    unless print($newfh "$info->{'content'}\n");
  warn("could not close $newName, $!"),
    print("save file begin $saveFile\n$info->{'content'}\nsave file end $saveFile\n"),
    unlink("$newName"),
    return(undef)
    unless close($newfh);
  # Read it back and compare.
  my ($newSave) = ECMSave->new('file' => "$newName", 'exp' => $info->{'exp'}, 'method' => $info->{'method'}, 'M' => $info->{'M'});
  warn("ECMSave::new returned undef for $newName"),
    print("save file begin $saveFile\n$info->{'content'}\nsave file end $saveFile\n"),
    return(undef)
    unless $newSave;
  warn($newSave),
    $newSave = undef,
    return(undef)
    unless ref($newSave) eq 'ECMSave';
  warn("ECMSave::check for $saveFile failed - " . $newSave->errors() . "\n"),
    print("save file begin $saveFile\n$info->{'content'}\nsave file end $saveFile\n"),
    return(undef)
    unless $newSave->check($mers);
  my ($newContent) = $newSave->content();
  warn("ECMSave::content failed for $newName - " . $newSave->errors() . "\n"),
    print("save file begin $saveFile\n$info->{'content'}\nsave file end $saveFile\n"),
    return(undef)
    unless $newContent;
  $info->{'content'} .= ' '
    if (substr($newContent, -1) eq ' ' # should always be true
        && substr($info->{'content'}, -1) eq ';'); # likely always true; see readClient()
  warn("logic error - content of $newName not what was read from fd $fd; saved"),
    print("save file begin $saveFile\n$newContent\nsave file end $saveFile\n"),
    return(undef)
    unless $info->{'content'} eq $newContent;
  warn("new save file '$newName' bound is undef or 0 - " . $newSave->errors() . "\n"),
    print("save file begin $saveFile\n$newContent\nsave file end $saveFile\n"),
    return(undef)
    unless $newSave->bound();
  my ($oq) = $info->{'method'} eq 'pm1' ? 'o' : 'q';
  my ($clm) = ($info->{'exp'} % 8 == 4) ? ($info->{'M'} ? 'm' : 'l') : 'c';
  my ($bound) = $newSave->bound();
  if (exists($saveFileNames{$info->{'exp'}}{$oq}{$clm}))
    { # Keep the old info if it has a higher (stage 1) bound.
      my ($ref) = $saveFileNames{$info->{'exp'}}{$oq}{$clm};
      delete $saveFileNames{$info->{'exp'}}{$oq}{$clm};
      $newName = $ref->{'file'},
        $bound = ($ref->{'bound'})
        if $ref->{'file'} ne $newName && $ref->{'bound'} > $newSave->bound();
    }
  $saveFileNames{$info->{'exp'}}{$oq}{$clm} = { 'file' => $newName, 'bound' => $bound, };
  return(undef);
}

sub closeClient($$)
{
  my ($fd, $failed) = shift;
  vec($fdbitsIn, $fd, 1) = 0;
  print(" closing fd $fd at " . localtime() . "\n");
  warn("fd $fd not in list of clients"),
    return
    unless exists $clients{$fd};
  # Copy the data for use here and delete it from %clients no matter what happens.
  my ($client) = $clients{$fd};
  delete $clients{$fd};
  my ($task);
  delete($client->{'tasks'})
    if (exists($client->{'tasks'}) && ref($client->{'tasks'}) ne 'ARRAY');
  my ($saveFile) = ($client->{'savefile'}{'file'})
    if exists $client->{'savefile'} && exists $client->{'savefile'}{'file'};
  while (exists($client->{'tasks'}) && @{$client->{'tasks'}})
    {
      $task = shift(@{$client->{'tasks'}});
      next unless ref($task) eq 'HASH' && exists $task->{'exp'};
      --($client->{'reservations'}{$task->{'l'}})
        if release($task->{'exp'}, $client->{'uuid'}, $task->{'l'}); #!!could check UUID but reserved while connected, so ...
      delete($client->{'free'}{$task->{'exp'}});
      my (@lines) = split(/[\n\r$CRLF]+/, $task->{'mers'});
      # Only the last line (or two, if lmLM) contain the work; see findWork().
      my ($mers) = shift(@lines);
      $saveFile = $1,
        next
        if $mers =~ m/^save file begin (ecm[.]([1-9][0-9]+)[.](p[mp]1)([.]M)?)$/;
      $saveFile = undef,
        next
        if $mers =~ m/^save file end ([a-z][-._a-zA-Z0-9]+)$/ && $saveFile && $saveFile eq "$1";
      warn("fd $fd had unexpected save file data - '$mers' - for '$saveFile'"),
        $saveFile = undef,
        next
        if $mers =~ m/^save / && $saveFile;
      warn("fd $fd had unexpected save file data - '$mers' - after no 'save file begin'"),
        $saveFile = undef,
        next
        if $mers =~ m/^save /;
      next
        if $saveFile;
      warn(" fd $fd had task not in mers format - '$mers'"),
        next
        unless $mers =~ m/^M[(] ([1-9][0-9]*) [)](.)(.*)/;
      my ($exp) = Math::BigInt->new("$1");
      my ($letter) = "$2";
      next if $letter eq 'C'; # factors should not be in %mersData, only %mersennes
      my ($rest) = "$3";
      #!!Should first verify that the mers letter matches the count of values after the colon; see readClient()
      #!!  Put those safety checks in a sub so they can be used elsewhere.
      my ($info) = [ $exp, $letter, ];
      if ($rest =~ m/^:? *([0-9]+) *([0-9]+)?/)
        {
          push(@{$info}, Math::BigInt->new("$1")) if defined($1) && $1 ne '';
          push(@{$info}, Math::BigInt->new("$2")) if defined($2) && $2 ne '';
          push(@{$mersData{$exp}}, $info);
        }
    }

  if (!exists($client->{'input'}) && !exists($clientTypes{$client->{'flag'}}{'input'}))
    {
      delete($client->{'renew'}),
        ($client->{'renew'} = { })
        if exists $client->{'renew'} && ref($client->{'renew'}) ne 'HASH';
      delete($client->{'free'}),
        ($client->{'free'} = { })
        if exists $client->{'free'} && ref($client->{'free'}) ne 'HASH';
      my (@list) = $mersDB->reservations($client->{'uuid'});
      print("client $client->{'name'} UUID $client->{'uuid'} has " . ($#list + 1) . " reservations in db at close,\n"
            . "  " . keys(%{$client->{'free'}}) . " exponents in free list before adding reservations to it,\n")
        if $debug & (1 << $debug{'rsrv'});
      $client->{'free'}{$_->{'exp'}} = 1
        foreach (@list);
      print("  " . keys(%{$client->{'free'}}) . " exponents in free list after adding reservations to it\n"
            . " and " . keys(%{$client->{'renew'}}) . " exponents in the renew list.\n")
        if $debug & (1 << $debug{'rsrv'});
      foreach (keys(%{$client->{'free'}}))
        {
          next if exists $client->{'renew'}{$_};
          @list = $mersDB->reservations($_);
          if ($#list < 0)
            {
              print("$client->{'name'} tried to free $_ without reserving it\n")
                if $debug & (1 << $debug{'rsrv'});
              next;
            }
          warn("M$_ reserved more than once, by $list[0]->{'uuid'} and $list[1]->{'uuid'}")
            if $#list > 0; #!!implies a bug in MersDB.pm - 'reservations' primary key is the exponent
          my ($ref) = pop(@list);
          print("$client->{'name'} tried to release $_, reserved by $ref->{'uuid'}\n"),
            next
            if ($ref->{'uuid'} ne $client->{'uuid'}
                && $ref->{'uuid'} ne $client->{'name'}
                && (!exists($UUIDaliases{$client->{'uuid'}})
                    || $ref->{'uuid'} ne $UUIDaliases{$client->{'uuid'}})
                && (!exists($UUIDaliases{$ref->{'uuid'}})
                    || $UUIDaliases{$ref->{'uuid'}} ne $client->{'uuid'})
                && (!exists($UUIDaliases{$client->{'uuid'}})
                    || !exists($UUIDaliases{$ref->{'uuid'}})
                    || $UUIDaliases{$ref->{'uuid'}} ne $UUIDaliases{$client->{'uuid'}}));
          --($client->{'reservations'}{$ref->{'kind'}}),
            next
            if release($ref->{'exp'}, $ref->{'uuid'}, $ref->{'kind'});
          warn("$client->{'name'} could not release M$_, " . $mersDB->dbError());
        }
      if (exists($client->{'fh'})
          && !exists($clientTypes{$client->{'flag'}}{'db'})
          && !exists($clientTypes{$client->{'flag'}}{'pipe'})
          && !exists($clientTypes{$client->{'flag'}}{'child'})
          && !$failed
          && !print({$client->{'fh'}} "done$CRLF"))
        {
          warn("print done to $client->{'name'} $client->{'flag'} fd $fd failed, $!")
            unless $!{'ECONNRESET'} || $!{'EPIPE'} || $!{'EAGAIN'};
        }
    }
  !exists($client->{'fh'})
    || close($client->{'fh'})
    || $!{'ECONNRESET'} || $!{'EPIPE'} || $!{'EAGAIN'}
    or warn "closing $client->{'name'} fd $fd socket failed, errno $!, child exited $?";
  my ($type) = $client->{'flag'};
  return
    unless $type && exists $clientTypes{$type}{'incr'};
  # Modify 'incr' based on how long the child took this time
  #   and set the delay before starting another of this child.
  my ($t) = time();
  my ($diff) = $t - $client->{'connect'};
  $childDelay{$client->{'flag'}} = $t + $diff;
  $clientTypes{$type}{'incr'} /= 2,
    return
    if $diff > 2*$childDelay{'all'};
  $clientTypes{$type}{'incr'} -= 1024,
    return
    if $diff > $childDelay{'all'};
  $clientTypes{$type}{'incr'} *= 2,
    return
    if (3 + $debug)*$diff < $childDelay{'all'};
  $clientTypes{$type}{'incr'} += 1024
    if $diff < $childDelay{'all'};
  return
    unless exists $clientTypes{$type}{'incrFile'};
  # Save the new 'incr' in the file.
  my ($fh);
  !-l($clientTypes{$type}{'incrFile'})
    or unlink($clientTypes{$type}{'incrFile'})
    or die("could not unlink softlink $clientTypes{$type}{'incrFile'}, $!");
  warn("could not create $clientTypes{$type}{'incrFile'}, $!"),
    delete($clientTypes{$type}{'incrFile'}),
    return
    unless open($fh, '>', $clientTypes{$type}{'incrFile'});
  warn("could not write to clientTypes{$type}{'incrFile'}, $!"),
    delete($clientTypes{$type}{'incrFile'}),
    return
    unless print($fh "$clientTypes{$type}{'incr'}\n");
  warn("could not close clientTypes{$type}{'incrFile'}, $!"),
    delete($clientTypes{$type}{'incrFile'})
    unless close($fh);
}

# Helper - return a mers format string of the known factors.
#   Or the result will be 'D', 'P', 'sprp', or undef if the data
#   is not in %mersennes or the cofactor SPRP checks are not done.
sub getFactorString($$)
{ # Only EoqS tasks need the known factors.
  my ($exp, $l) = @_;
  return('')
    unless $l =~ m,^[EoqS]$,;
  my ($result) = '';
  my ($sprp) = $mersDB->value('sprp', $exp);
  return('sprp')
    if !defined($sprp) || $sprp;
  if (!exists($mersennes{$exp}))
    {
      $needData{$exp} .= 'C' if $needData{$exp} && $needData{$exp} !~ m,C,;
      $needData{$exp} = 'C' if !$needData{$exp};
      return undef;
    }
  my (@values) = $mersennes{$exp}->f();
  $result .= "M( $exp )C: $_$CRLF"
    foreach @values;
  return($result);
}  

sub findWork($$)
{ # Return the count of tasks (perhaps previously) reserved for the client but not yet sent.
  my ($fd) = shift;
  my ($quick) = shift; # if true, only get one task because the client is waiting
  delete $clients{$fd}{'tasks'} if exists($clients{$fd}{'tasks'}) && ref($clients{$fd}{'tasks'}) ne 'ARRAY';
  return(1 + $#{$clients{$fd}{'tasks'}}) if exists($clients{$fd}{'tasks'}) && $#{$clients{$fd}{'tasks'}} >= 0;
  warn("findWork - fd $fd has no 'flag'"),
    return(undef)
    unless exists $clients{$fd}{'flag'};
  warn("findWork - fd $fd 'flag' not in client types"),
    return(undef)
    unless exists $clientTypes{$clients{$fd}{'flag'}};
  warn("findWork - fd $fd client type is not a worker - $clientTypes{$clients{$fd}{'flag'}}"),
    return(undef)
    unless exists $clientTypes{$clients{$fd}{'flag'}}{'worker'};
  warn("findWork - fd $fd client 'worker' is not in work hash - $clientTypes{$clients{$fd}{'flag'}}{'worker'}"),
    return(undef)
    unless exists $work{$clientTypes{$clients{$fd}{'flag'}}{'worker'}};
  warn("findWork - fd $fd client 'uuid' does not exist"),
    return(undef)
    unless exists $clients{$fd}{'uuid'};
  warn("findWork - fd $fd client UUID $clients{$fd}{'uuid'} not in allowed list"),
    return(undef)
    unless exists $allowed{$clients{$fd}{'uuid'}};
  $quick = 1 if -d("$stopDir") || $seenSIG{'INT'};
  my ($l) = $clientTypes{$clients{$fd}{'flag'}}{'worker'};
  my ($onlyLetters) = $allowed{$clients{$fd}{'uuid'}}{'letters'};
  warn("client $clients{$fd}{'name'} UUID $clients{$fd}{'uuid'} has no allowed mers letters"),
    return(undef)
    unless $onlyLetters;
  warn("client $clients{$fd}{'name'} UUID $clients{$fd}{'uuid'} has no or invalid max tasks"),
    return(undef)
    unless exists $allowed{$clients{$fd}{'uuid'}}{'tasks'} && $allowed{$clients{$fd}{'uuid'}}{'tasks'} =~ m,$bigintPattern,;
  warn("client $clients{$fd}{'name'} UUID $clients{$fd}{'uuid'} is asking for $l work which is not allowed - '$onlyLetters'"),
    return(undef)
    unless $l =~ m/^[$onlyLetters]$/;
  if (!exists($clients{$fd}{'reservations'}{$l}) || $clients{$fd}{'reservations'}{$l} !~ m,$bigintPattern,)
    {
      my ($count) = $mersDB->countReserved($clients{$fd}{'uuid'}, $l);
      warn("could not count reservations for fd $fd UUID $clients{$fd}{'uuid'} - " . $mersDB->dbError()),
        $count = $maxTasks
        unless defined $count;
      $clients{$fd}{'reservations'}{$l} = $count;
    }
  print("# client $clients{$fd}{'name'} $clients{$fd}{'flag'} UUID $clients{$fd}{'uuid'} at max tasks $allowed{$clients{$fd}{'uuid'}}{'tasks'}, $clients{$fd}{'reservations'}{$l} reserved\n"),
    print({$clients{$fd}{'fh'}} "at max tasks, $allowed{$clients{$fd}{'uuid'}}{'tasks'}$CRLF"),
    return(0)
    if $allowed{$clients{$fd}{'uuid'}}{'tasks'} <= $clients{$fd}{'reservations'}{$l};
  local $_;
  # These defaults are roughly those of older clients that did not send their limits to the (C/C++) server.
  my ($minExp) = 0;
  my ($maxExp) = $MersDB::SQLmaxInt;
  my ($minValue) = 0;
  my ($maxValue) = Math::BigInt->bone()->blsft(1024); #!!this is "smaller" than the limits ...
     # ... of older trial factorers but well past the limits of GIMPS (www.mersenne.org)
  my ($onlyPrimeExps) = 0;
  if (exists($clients{$fd}{'limits'}))
    {
      $minExp = ($clients{$fd}{'limits'}{'min_exp'})
        if exists($clients{$fd}{'limits'}{'min_exp'});
      $maxExp = ($clients{$fd}{'limits'}{'max_exp'})
        if exists($clients{$fd}{'limits'}{'max_exp'});
      $minValue = Math::BigInt->bone()->blsft($clients{$fd}{'limits'}{'min_bits'})
        if exists($clients{$fd}{'limits'}{'min_bits'});
      $maxValue = Math::BigInt->bone()->blsft($clients{$fd}{'limits'}{'max_bits'})
        if exists($clients{$fd}{'limits'}{'max_bits'});
      $onlyPrimeExps = 1
        if exists $clients{$fd}{'limits'} && $clients{$fd}{'limits'}{'prime'};
      print({$clients{$fd}{'fh'}} "min_exp must be less than max_exp$CRLF"),
        warn("client $clients{$fd}{'name'} UUID $clients{$fd}{'uuid'} min exp $minExp conflicts with max exp $maxExp"),
        return(undef)
        unless $minExp < $maxExp;
      print({$clients{$fd}{'fh'}} "min_bits must be less than max_bits$CRLF"),
        warn("client $clients{$fd}{'name'} UUID $clients{$fd}{'uuid'} min bits $clients{$fd}{'limits'}{'min_bits'} conflicts with max bits $clients{$fd}{'limits'}{'max_exp'}"),
        return(undef)
        unless $minValue < $maxValue;
    }
  my ($i); # index in work list to start at
  my ($reason) = 'empty work list'; # ... that $work{$l}[0] was not a match
  while ($#{$clients{$fd}{'tasks'}} < $minTasks)
    { #!!$minTasks just above is arbitrary; allow the client to ask for a count?
      my ($match); # will be the index of the next work within the client's limits
      my ($exp, $one, $two); # fields of mers format line to send to client
      $i = 0 if !$i || $i > $#{$work{$l}};
      $reason = 'empty work list' if $i == 0; # ... that $work{$l}[0] was not a match
      for ( ; $i <= $#{$work{$l}}; ++$i)
        {
          $reason = 'SIGIO', last if $i > $minTasks && $seenSIG{'IO'} && !$sigIObroken;
          $exp = $work{$l}[$i]{'exp'};
          $reason = "$exp needs $needData{$exp} data",
            next
            if exists $needData{$exp}; # data has been requested but not read
          $reason = "$exp data not saved to db" if !$match;
          next if exists $mersData{$exp} && $l !~ m,^[GgUH]$,; # new data has not been saved in the db
          $reason = "$exp sprp 1" if !$match;
          my ($sprp) = $mersDB->value('sprp', $exp);
          next if $l =~ m,^[Eoq]$, && defined($sprp) && $sprp;
          $reason = "$exp not prime exponent" if !$match;
          next if $onlyPrimeExps && !Math::Prime::Util::is_prime($exp);
          $reason = "$exp not between $minExp and $maxExp" if !$match;
          next unless $minExp <= $exp && $exp <= $maxExp;
          $match = $i, last unless exists $work{$l}[$i]{'one'};
          $one = $work{$l}[$i]{'one'};
          $reason = "$exp first value not between $minValue and $maxValue" if !$match;
          next unless $minValue <= $one && $one <= $maxValue;
          $match = $i, last unless exists $work{$l}[$i]{'two'};
          $two = $work{$l}[$i]{'two'};
          $match = $i, last unless defined $two;
          $reason = "$exp second value not between $minValue and $maxValue" if !$match;
          next unless $minValue <= $two && $two <= $maxValue;
          $match = $i;
          last;
        }
      if (!defined($match))
        {
          $need{$l} = 0
            unless exists $need{$l};
          $debug |= (1 << $debug{'rsrv'}) | (1 << $debug{'pri'}) | (1 << $debug{'need'}) | (1 << $debug{'work'})
            if ++$need{$l} > 100;
          ($need{$l} = $minTasks + $#{$work{$l}})
            if $need{$l} < $#{$work{$l}};
          if ($#{$clients{$fd}{'tasks'}} < 0)
            {
              print("findWork - fd $fd no suitable work; need $l now $need{$l}\n");
              print("  work $l has " . ($#{$work{$l}} + 1) . " entries; first is $work{$l}[0]{'exp'}"
                    . (exists($work{$l}[0]{'one'}) ? " $work{$l}[0]{'one'}" : '') 
                    . (exists($work{$l}[0]{'two'}) ? " $work{$l}[0]{'two'}" : '')
                    . " - '"
                    . ($reason ? $reason : 'reason is unknown - logic error') . "'\n")
                if $#{$work{$l}} >= 0;
            }
          return(1 + $#{$clients{$fd}{'tasks'}});
        }
      my ($task);
      if ($l =~ m/^[EoqS]$/)
        { #!!are there any other cases in which the client needs more data about the Mersenne?
          $task = getFactorString($exp, $l);
          print("findWork - getFactorString $exp $l returned <undef>\n"),
            next
            unless defined $task;
          print("findWork - getFactorString $exp $l returned '$task'\n"),
            splice(@{$work{$l}}, $match, 1),
            next
            if ($task eq 'D' || $task eq 'P' || $task eq 'sprp');
          warn("findWork - getFactorString $exp $l returned unexpected value '" . join(' ', split(/[\n\r]+/, $task)) . "'"),
            next
            unless !$task || $task =~ m,^M[(] $exp [)],;
        }
      $task .= "M( $exp )" . $l;
      warn("findWork - fd $fd - task M$exp $l had no first value"),
        splice(@{$work{$l}}, $match, 1),
        next
        unless defined $one;
      $task .= ": $one";
      $task .= " $two" if defined $two;
      my ($pri);
      if ($l =~ m/^[EGg]$/)
        {
          $pri = Mersenne::mersPriority($exp, $l, $one, $two);
          warn("findWork - task M$exp $l priority $pri"),
            splice(@{$work{$l}}, $match, 1),
            next
            if $pri < 0.0;
        }
      $task .= "$CRLF";
      my (@r) = $mersDB->reservations($exp);
      foreach my $row (@r) # more than one row implies a bug in the SQL database or table creation
        { # already reserved but perhaps for a "lesser" work letter/type
          print("work $l # $match M$exp for fd $fd already reserved by $row->{'uuid'} kind $row->{'kind'}\n")
            if $row->{'uuid'} ne $clients{$fd}{'uuid'} || ($debug & (1 << $debug{'rsrv'}));
          last
            if ($row->{'uuid'} eq $clients{$fd}{'uuid'} # reserved by the same client ...
                || $letterPri{$row->{'kind'}} <= $letterPri{$l}); # ... or for "more important" work
          # reserved but this work is more "important" than the reserved work
          my (@released) = release($exp, $row->{'uuid'}, $row->{'kind'});
          --($clients{$fd}{'reservations'}{$row->{'kind'}})
            if $#released >= 0;
          print(" ... which is less 'urgent', so releasing it - " . join(' ', @released) . "\n");
          @r = () if @released;
          last;
        }
      splice(@{$work{$l}}, $match, 1),
        next
        if @r; # reserved - should not be in %work
      @r = $mersDB->reserve($exp, $clients{$fd}{'uuid'}, $l);
      if ($#r < 0 || !$r[0])
        { # error (including already reserved - SQL insert was used)
          print("findWork - could not reserve $exp, entry $match in work $l list\n");
          $need{$l} = 0
            unless exists $need{$l};
          ++$need{$l};
          $need{$l}++
            if $need{$l} < $minTasks;
          splice(@{$work{$l}}, $match, 1), next if @r;
          next;
        }
      ++$clients{$fd}{'reservations'}{$l};
      print("findWork - reserved $exp, entry $match of " . (1 + $#{$work{$l}})
            . " in work $l list, for fd $fd\n")
        if $debug & (1 << $debug{'rsrv'});
      my ($save);
      if ($l =~ m,^[oq]$, && exists($saveFileNames{$exp}) && exists($saveFileNames{$exp}{$l}))
        {
          my ($cl) = ($exp % 8 == 4 ? 'l' : 'c');
          if (exists $saveFileNames{$exp}{$l}{$cl}
              && exists $saveFileNames{$exp}{$l}{$cl}{'file'})
            {
              $save = ECMSave->new('file' => $saveFileNames{$exp}{$l}{$cl}{'file'}, 'exp' => $exp,
                                   'method' => ($l eq 'q' ? 'pp1' : 'pm1'), 'M' => '');
              warn("findWork - ECMSave::new failed for $saveFileNames{$exp}{$l}{$cl}{'file'} - $save\n"),
                $save = undef
                if $save && ref($save) ne 'ECMSave';
            }
        }
      $save = ECMSave::findBestSaveFile($exp, $l, '', undef, 1)
        if $l =~ m,^[oq]$, && !$save;
      warn("findWork - ECMSave::findBestSaveFile failed for $exp $l - $save\n"),
        $save = undef
        if $save && ref($save) ne 'ECMSave';
      if ($save)
        {
          my ($final) = $save->file();
          my ($content) = $save->content();
          if ($final && $content)
            {
              $final =~ s,^.*/,,;
              $task .= "save file begin $final$CRLF" . "$content" . "$CRLF"
                . "save file end $final$CRLF";
            }
        }
      if ($exp % 8 == 4 && $l =~ m,^[oq]$,)
        { # The 'M' cofactor of such Mersennes might have its own save file.
          if (exists($saveFileNames{$exp}) && exists($saveFileNames{$exp}{$l}))
            {
              if (exists $saveFileNames{$exp}{$l}{'m'}
                  && exists $saveFileNames{$exp}{$l}{'m'}{'file'})
                {
                  $save = ECMSave->new('file' => $saveFileNames{$exp}{$l}{'m'}{'file'}, 'exp' => $exp,
                                       'method' => ($l eq 'q' ? 'pp1' : 'pm1'), 'M' => '.M');
                  warn("findWork - ECMSave::new failed for $saveFileNames{$exp}{$l}{'m'}{'file'} - $save\n"),
                    $save = undef
                    if $save && ref($save) ne 'ECMSave';
                }
            }
          $save = ECMSave::findBestSaveFile($exp, $l, '.M')
            unless $save;
          warn("findWork - ECMSave::findBestSaveFile failed for $exp $l - $save\n"),
            $save = undef
            if $save && ref($save) ne 'ECMSave';
          if ($save)
            {
              my ($final) = $save->file();
              $final =~ s,^.*/,,;
              my ($contentM) = $save->content();
              $task .= "save file begin $final$CRLF" . "$contentM" . "$CRLF"
                . "save file end $final$CRLF"
                if $final && $contentM;
            }
        }
      my ($ref) = { 'mers' => $task, 'exp' => $exp, 'l' => $l, };
      $ref->{'pri'} = $pri if defined $pri;
      push(@{$clients{$fd}{'tasks'}}, $ref);
      # Delete the task from %work.  Gg work is still in %data.
      splice(@{$work{$l}}, $match, 1);
      warn("no entry for M$exp in workExponents hash despite having $l work"),
        $workExponents{$exp} = ''
        unless exists $workExponents{$exp};
      print(" workExponents for M$exp did not include $l work\n")
        unless $workExponents{$exp} =~ s/$l//g;
      return(1 + $#{$clients{$fd}{'tasks'}}) if $quick || ($seenSIG{'IO'} && !$sigIObroken);
      $i = $match; # next task that might be within the limits - entry $match just deleted
    }
  warn("findWork - fd $fd - no work found, need $l now $need{$l}")
    unless $#{$clients{$fd}{'tasks'}} >= 0;
  print("findWork - fd $fd already had $minTasks tasks\n")
    unless defined $i;
  return(1 + $#{$clients{$fd}{'tasks'}});
}

sub sendWork($)
{ # Returns undef if no %clients entry or the client is closed here, 1 otherwise.
  my ($fd) = shift;
  warn("sendWork() called with fd $fd, not in clients HASH"),
    closeClient($fd, 1),
    return(undef)
    unless exists $clients{$fd};
  warn("sendWork() called with fd $fd, which has no 'fh', no 'flag', or no 'uuid'"),
    closeClient($fd, 1),
    return(undef)
    unless exists $clients{$fd}{'fh'} && exists $clients{$fd}{'flag'} && exists $clients{$fd}{'uuid'};
  warn("sendWork() called with fd $fd UUID $clients{$fd}{'uuid'} not in allowed list"),
    closeClient($fd, 1),
    return(undef)
    unless exists $allowed{$clients{$fd}{'uuid'}};
  my ($fh) = $clients{$fd}{'fh'};
  my ($fw) = findWork($fd, 1);
  closeClient($fd, 1),
    return(undef)
    unless defined $fw;
  if ($fw)
    {
      my ($task) = $clients{$fd}{'tasks'}[0];
      if (exists($task->{'pri'}))
        { # sendServer.pl needs the priority for gaps and does not have the formula to calculate it
          my ($info) = $task->{'mers'};
          chomp($info);
          chop($info)
            while ($info =~ m/[$CRLF]$/);
          shift(@{$clients{$fd}{'tasks'}}),
            return(1)
            if print($fh $info . ' ' . $task->{'pri'} . "$CRLF");
        }
      else
        {
          shift(@{$clients{$fd}{'tasks'}}),
            return(1)
            if print($fh $task->{'mers'});
        }
    }
  elsif (-d("$stopDir") || $seenSIG{'INT'})
    {
      print($fh "server exiting$CRLF");
      return(undef);
    }
  elsif (print($fh "the server has no suitable work$CRLF"))
    {
      return(1);
    }
  return(1)
    if !defined($!) || $!{'EWOULDBLOCK'} || $!{'EAGAIN'} || $!{'EINTR'};
  print(" sending to $clients{$fd}{'name'} fd $fd failed, $!\n");
  closeClient($fd, 1);
  return(undef);
}

# Checking the clients is done every time thru the primary loop, unlike other "internal work".
#   Like other internal work, return 0 if no more needs to be done right now and anything true (1) otherwise,
#   but, unlike other internal work, should never return a negative value.
sub checkClients()
{ # Internal work - mostly, check that the (connected) clients are OK.
  local $_;
  my ($maxConnect) = 30;
  my ($moreToDo);
  foreach my $fd (keys(%clients))
    {
      print("checkClients calling readClient fd $fd\n") if $debug & (1 << $debug{'misc'});
      my ($rc) = readClient($fd); # all clients support this but may have been close()d if undef returned
      print("checkClients - readClient fd $fd returned " . (defined($rc) ? "$rc" : '<undef>') . "\n") if $debug & (1 << $debug{'misc'});
      next unless $rc;
      print("checkClients - readClient $fd returned '$rc' but $fd not in clients hash\n"),
        next
        unless exists $clients{$fd};
      print("missing 'flag'/kind for client on fd $fd\n"),
        closeClient($fd, 1),
        next
        unless exists $clients{$fd}{'flag'};
      next if exists $clients{$fd}{'input'};
      next unless exists $clients{$fd}{'uuid'};
      next unless exists $clientTypes{$clients{$fd}{'flag'}}{'worker'};
      next unless exists $work{$clientTypes{$clients{$fd}{'flag'}}{'worker'}};
      $moreToDo = 1; # client is still there, so there might be more to do
      print({$clients{$fd}{'fh'}} "connected more than $maxConnect minutes$CRLF" . "done$CRLF"),
        closeClient($fd, 1),
        next
        if ($lastContact{$clients{$fd}{'uuid'}} + 60*$maxConnect < time()
            && (!exists($clients{$fd}{'tasks'}) || $#{$clients{$fd}{'tasks'}} < 0));
      $rc = sendWork($fd);
      next unless $rc;
      #!!safety checks of 'fh', 'name', 'flag', 'uuid', 'connect', 'tasks'@{'mers','exp'} ?
      next;
    }
  continue
    {
      return(1) if $seenSIG{'IO'} && !$sigIObroken;
    }
  print("checkClients() fd loop done\n") if $debug & (1 << $debug{'misc'});
  foreach (grep { exists($clientTypes{$_}{'files'} ) } keys(%clientTypes))
    {
      delete $clientTypes{$_}{'files'}, next unless ref($clientTypes{$_}{'files'}) eq 'ARRAY';
      delete $clientTypes{$_}{'files'}, next unless $#{$clientTypes{$_}{'files'}} >= 0;
      while (@{$clientTypes{$_}{'files'}})
        {
          my ($file) = shift(@{$clientTypes{$_}{'files'}});
          warn("ignoring path '$file' containing '$1'"),
            next
            if $file =~ m,([/.][.][/.]),;
          warn("soft link '$file' ignored"),
            next
            if -l($file);
          warn("not a file or empty file '$file' ignored"),
            next
            unless -f($file) && -s($file);
          my ($fh);
          warn("could not read $file given with " . $clientTypes{$_}{'opt'} . ", $!"),
            next
            unless open($fh, '<', $file);
          my ($fd) = fileno($fh);
          # Most 'input' data is not printed in readClient() and most 'input' clients data is sorted
          #    (by exponent or by priority), so their last exponent, mers letter, and priority are
          #    retained in readClient() in 'lastExp', 'lastLetter', and 'lastPri'.
          $clients{$fd} = { 'fh' => $fh, 'input' => 1, 'name' => $file, 'uuid' => $file, 'flag' => $_, 'connect' => time(), };
          print(" opened $file on fd $fd as a $_ client at " . localtime() . "\n");
          $moreToDo = 1; # new (input only) client
        }
    }
  print("checkClients() files loop done\n") if $debug & (1 << $debug{'misc'});
  if (exists($clientTypes{'Pipe'}{'fh'})
      || (exists($clientTypes{'Pipe'}) && exists($clientTypes{'Pipe'}{'pipe'}) && $clientTypes{'Pipe'}{'pipe'}
          && -f($clientTypes{'Pipe'}{'pipe'}) && -s($clientTypes{'Pipe'}{'pipe'})))
    {{
      my ($fh);
      if (!exists($clientTypes{'Pipe'}{'fh'}))
        {
          last # if open() or unblock() fails, don't try again for a while
            if $clientTypes{'Pipe'}{'contact'} + $childDelay{'all'} >= time();
          !-l($clientTypes{'Pipe'}{'pipe'})
            or unlink($clientTypes{'Pipe'}{'pipe'})
            or die("could not unlink softlink $clientTypes{'Pipe'}{'pipe'}, $!");
          warn("could not open pipe $clientTypes{'Pipe'}{'pipe'} for reading, $!"),
            last
            unless  open($fh, '<', $clientTypes{'Pipe'}{'pipe'});
          close($fh),
            last
            unless unblock($fh, 'pipe');
          warn("could not remove $clientTypes{'Pipe'}{'pipe'}, $!"),
            close($fh),
            last
            unless unlink($clientTypes{'Pipe'}{'pipe'});
          $clientTypes{'Pipe'}{'fh'} = $fh;
          print("opened $clientTypes{'Pipe'}{'pipe'} fd " . fileno($fh) . "\n");
        }
      $fh = $clientTypes{'Pipe'}{'fh'} unless defined $fh;
      $clientTypes{'Pipe'}{'contact'} = time();
      while (<$fh>)
        {
          chomp;
          warn("unexpected data from pipe, '$_'"),
            $_ = undef,
            last
            unless m/^([-. a-zA-Z0-9]+)$/;
          refetchAllowed(), next if $_ eq 'allowed';
          refetchAliases(), next if $_ eq 'aliases';
          refetchContact(), next if $_ eq 'contact';
          $debug = $1, next if m/^debug ([0-9]+)$/;
          $debug ^= $debug{$1}, next if m/^debug ([a-z]+)$/ && exists $debug{$1};
          warn("non-digits for new debug level, $_"),
            $_ = undef,
            last
            if m/^debug /;
          push(@{$clientTypes{'DataFile'}{'files'}}, $1),
            next
            if m/^file (.*)$/ && -s $1 && ! -l $1;
          $mersDB->addSortedFiles("$1"),
            next
            if m/^byExp (.*)$/ && -s $1 && ! -l $1;
          push(@{$clientTypes{'SortedByPriority'}{'files'}}, $1),
            next
            if m/^byPri (.*)$/ && -s $1 && ! -l $1;
          if (m/^free ([1-9][0-9]*)$/)
            {
              my ($exp) = $1;
              my (@r) = $mersDB->reservations($exp);
              warn("exponent $exp was not reserved"),
                next
                unless @r;
              foreach my $row (@r)
                { # Should only be one; the exponent is the primary key of the SQL tables.
                  next unless $row && ref($row) eq 'HASH';
                  release($exp, $row->{'uuid'}, $row->{'kind'});
                }
              next;
            }
          if (m/^free $uuidPiece/)
            {
              my ($uuid) = $1;
              my (@r) = $mersDB->reservations($uuid);
              warn("nothing reserved for UUID $uuid"),
                next
                unless @r;
              foreach my $row (@r)
                {
                  next unless $row && ref($row) eq 'HASH';
                  release($row->{'exp'}, $uuid, $row->{'kind'});
                }
              next;
            }
          if (m/^print ([1-9][0-9]*)$/)
            { # Print all data and work for the exponent.
              my ($exp) = $1;
              print(" printing M$exp\n");
              $mersennes{$exp}->print()
                if exists $mersennes{$exp};
              foreach (sort(keys(%work)))
                {
                  foreach my $entry (@{$work{$_}})
                    {
                      next unless $entry->{'exp'} == $exp;
                      print("M( $exp )$_");
                      print(": $entry->{'one'}") if exists $entry->{'one'};
                      print(" $entry->{'two'}") if exists $entry->{'two'};
                      print("\n");
                    }
                }
              foreach (sort(keys(%data)))
                {
                  next unless exists $data{$_}{$exp};
                  foreach my $entry (@{$data{$_}{$exp}})
                    {
                      print("M( $exp )$_");
                      print(": $entry->{'one'}") if exists $entry->{'one'};
                      print(" $entry->{'two'}") if exists $entry->{'two'};
                      print("\n");
                    }
                }
              next;
            }
          #!!incomplete?
          print("unexpected/unsupported data from pipe, '$_'\n");
          $_ = undef;
          last;
        }
      continue
        {
          $moreToDo = 1, last if $seenSIG{'IO'};
        }
      last if $seenSIG{'IO'} || (!defined($_) && ($!{'EWOULDBLOCK'} || $!{'EAGAIN'} || $!{'EINTR'}));
      if (!defined($_))
        { # Unexpected data or no (more) data to be read
          warn("reading from $clientTypes{'Pipe'}{'pipe'} failed, $!")
            if $!;
          last;
        }
      close($clientTypes{'Pipe'}{'fh'})
        or warn("could not close pipe, $!");
      delete $clientTypes{'Pipe'}{'fh'};
      unlink($clientTypes{'Pipe'}{'pipe'})
        or warn("could not unlink $clientTypes{'Pipe'}{'pipe'}, $!");
    }}
  print("checkClients() returning " . (defined($moreToDo) ? "$moreToDo" : '<undef>') . "\n") if $debug & (1 << $debug{'misc'});
  return($moreToDo);
}

sub checkMersData()
{ # Internal work - update the db or call addData() with an entry from %mersData as appropriate.
  return(0) unless scalar(keys(%mersData));
  local $_;
  my (@exps) = sort { $a <=> $b } grep { !exists($needData{$_}) || !$needData{$_} } keys(%mersData);
  return(getNeededData()) if $#exps < 0;
  my (@xexps) = $mersDB->keyList('X');
  return(-1) unless @xexps; #!!should never happen; see init()
  $#exps = 2*$minTasks if $#exps > 2*$minTasks && $exps[$#exps] > $xexps[0] && $sigIObroken;
  $#exps = 10*$minTasks if $#exps > 10*$minTasks;
  print("checkMersData() " . scalar(keys(%mersData)) . " exponents - " . join(' ', @exps) . "\n") if $debug & (1 << $debug{'data'});
 EXP: # only needed for 'e' data that can't be dealt with yet
  foreach my $exp (@exps)
    {
      print("mersData key $exp is empty, not an ARRAY ref, or an empty array ref\n"),
        delete($mersData{$exp}),
        next
        unless $mersData{$exp} && ref($mersData{$exp}) eq 'ARRAY' && $#{$mersData{$exp}} > -1;
      while ($#{$mersData{$exp}} > -1)
        {
          my ($mersRef) = shift(@{$mersData{$exp}});
          next unless $mersRef && ref($mersRef) eq 'ARRAY';
          print("checkMersData() - " . join(' ', @{$mersRef}) . "\n") if $debug & (1 << $debug{'data'});
          my ($l) = $mersRef->[1];
          # Trial factoring gaps, joins, and extents are never saved in the db but are perhaps added to %data
          if (exists($data{$l}))
            {
              if (!exists($joinExps{$exp}) || $joinExps{$exp} !~ m/J/)
                { # Gaps need to be checked against existing joins.
                  $needData{$exp} = ''
                    unless exists $needData{$exp};
                  $needData{$exp} .= 'J',
                    next
                    if $l =~ m/^G$/i && $needData{$exp} !~ m/J/;
                }
              print("checkMersData() adding data " . join(' ', @{$mersRef}) . "\n") if $debug & (1 << $debug{'data'});
              addData(@{$mersRef});
              next;
            }
          if ($l eq 'C')
            { # Factors of smaller Mersennes need to be saved to the db all at once for each Mersenne,
              #   which should have already been done.
              my ($lmec) = $mersDB->lowestMissingExponent('c');
              warn("checkMersData - lowest missing exponent 'c' is undef\n" . '  ' . $mersDB->dbError()),
                next
                unless $lmec;
              warn("checkMersData - factors should not be in mersData - " . join(' ', @{$mersRef}))
                if $exp <= $lmec;
              # Factors of larger Mersennes are not saved in the db.
              next;
            }
          if ($l eq 'e')
            { # 'e' results need to be added into 'E' effort - work, not data.
              while ($#{$mersData{$exp}} > -1 && ref($mersData{$exp}) eq 'ARRAY' && $mersData{$exp}[0][1] eq 'e')
                { # Next entry is also 'e'; merge them.
                  $mersRef->[2] += $mersData{$exp}[0][2];
                  $mersRef->[3] = $mersData{$exp}[0][2]
                    if $mersRef->[3] < $mersData{$exp}[0][2];
                  shift(@{$mersData{$exp}});
                }
              push(@{$mersData{$exp}}, $mersRef),
                $needData{$exp} .= 'E',
                next EXP
                if exists $needData{$exp} && $needData{$exp} !~ m,E,;
              push(@{$mersData{$exp}}, $mersRef),
                $needData{$exp} = 'E',
                next EXP
                unless exists $mersennes{$exp} && $mersennes{$exp}->get('E');
              my ($line) = "M( $exp )e: " . $mersRef->[2];
              $line .= ' ' . $mersRef->[3] if $#{$mersRef} >= 3;
              $mersennes{$exp}->parse($line)
                or print("parsing ECM data failed:\n" . $mersennes{$exp}->errors());
              my (@ecm) = $mersennes{$exp}->get('E');
              my ($aw) = addWork($exp, 'E', @ecm);
              print("checkMersData 1 addWork '$aw' updating db with " . join(' ', $exp, 'E', @ecm) . "\n")
                if $aw !~ m/^(added|highExp)$/ || $debug & (1 << $debug{'add'});
              next;
            }
          print("checkMersData() updating db with " . join(' ', @{$mersRef}) . "\n")
            if $debug & (1 << $debug{'data'});
          my (@r) = $mersDB->update($l, @{$mersRef});
          print("could not update M$mersRef->[0] $l - " . $mersDB->dbError())
            unless @r && $r[0];
          @xexps = $mersDB->keyList('X')
            if $l =~ m/^[DPX]$/;
          if ($l eq 'D' || $l eq 'P' || $exp <= $xexps[0])
            { # 'D'one or 'P'rime
              delete($data{$_}{$exp}) foreach (keys(%data));
              next unless exists $workExponents{$exp};
              foreach (split(//, $workExponents{$exp}))
                { # First, a safety check - this 'next' should never happen.
                  next unless exists $work{$_};
                  my ($i) = 0;
                  while ($i <= $#{$work{$_}})
                    {
                      ++$i, next unless $work{$_}[$i]{'exp'} == $exp;
                      splice(@{$work{$_}}, $i, 1); # delete the task
                      # Only one task per exponent for all mers letters except Gg (gaps)
                      #   ... but look for more in case of problems elsewhere
                      #   as this is quite rare.
                    }
                }
              delete $workExponents{$exp};
            }
          if (exists($work{$l}))
            {
              my ($aw) = addWork(@{$mersRef});
              print("checkMersData 2 addWork '$aw' for " . join(' ', @{$mersRef}) . "\n")
                if $aw !~ m/^(added|highExp)$/ || $debug & (1 << $debug{'add'});
            }
          if ($seenSIG{'IO'} && !$sigIObroken)
            { # Return early to avoid stalling clients.
              delete $mersData{$exp} if $#{$mersData{$exp}} < 0;
              return(1);
            }
        }
      delete($mersData{$exp});
    }
  return(1);
}

sub checkData()
{ # Internal work - verify an entry within %data.
  local $_;
  $next{'checkData'}{'next'} = 0
    if $next{'checkData'}{'next'} >= length($next{'checkData'}{'letters'});
  my ($l) = substr($next{'checkData'}{'letters'}, $next{'checkData'}{'next'}, 1);
  $next{'checkData'}{$l} = 0,
    ++$next{'checkData'}{'next'},
    return(0)
    unless exists $data{$l} && scalar(keys(%{$data{$l}}));
  print("checkData - data $l has " . scalar(keys(%{$data{$l}})) . " entries\n")
    if $debug & (1 << $debug{'data'});
  my ($e) = $next{'checkData'}{$l};
  $e = MersInfo::nextKey($e, $data{$l})
    if !exists $data{$l}{$e};
  $next{'checkData'}{$l} = $e + 1;
  ++$next{'checkData'}{'next'};
  my (@xexps) = $mersDB->keyList('X');
  print("checkData - no X data\n"),
    return(-1)
    unless @xexps;
  if ($e <= $xexps[0] || defined($mersDB->value('done', $e)))
    { # M($e) is completely factored (D) or prime (P).
      delete($data{$_}{$e})
        foreach (keys(%data));
      delete($joinExps{$e});
      delete($work{$_}{$e})
        foreach (keys(%work));
      return(1);
    }
  #!!likely incomplete - check $data{$l}{$e}
  return(0)
    unless scalar(keys(%{$data{$l}})) > 2*$maxTasks;
  if ($l =~ m,^[Gg]$,)
    { #!!checkData data G has 717614 entries 8 Oct 2021
      my ($ref) = $data{$l}{$e}[0];
      my ($x) = 1;
      ++$x while ($x <= $#xexps && $xexps[$x] < $e);
      if ($x > $#xexps # If the gaps' Mersenne's exponent is larger than all the X (1strs) data ...
          || $ref->{'one'} <= $mersDB->value('X', $xexps[$x]))
        { # ... or it starts at or before the 1strs "boundary", simply drop the gap.
          shift(@{$data{$l}{$e}});
          return(1);
        }
      my ($aw) = addWork($e, $l, $ref->{'one'}, $ref->{'two'});
      print("checkData - addWork '$aw' for '$e $l $ref->{'one'} $ref->{'two'}"
            . (exists($ref->{'created'}) ? " # created $ref->{'created'}" : ' # created missing')
            . (exists($ref->{'mod1'}) ? " mod1 $ref->{'mod1'}" : '')
            . (exists($ref->{'mod2'}) ? " mod2 $ref->{'mod2'}" : '')
            . "\n")
        if $aw !~ m/^(added|highExp)$/ || $debug & (1 << $debug{'add'});
      if ($aw =~ m,^(added)|(invalid)|(notWork)|(negPri)|([DPS])|(done)|(lesser2?)$,)
        { # Moved to %work or not needed or invalid or ...
          shift(@{$data{$l}{$e}});
          return(1);
        }
      # Finding a new factor of a "small" Mersenne using trial factoring is highly unlikely
      #   but don't drop them when low on work.
      return(1)
        if $#{$work{$l}} < $maxTasks;
      if (2*$e < $mersDB->lowestMissingExponent('c'))
        {
          delete($data{$l}{$e});
          delete($data{'J'}{$e});
          delete($joinExps{$e});
        }
      return(1);
    }
  if ($l eq 'J')
    { # If there are no gaps for the Mersenne, drop the joins.
      delete($data{$l}{$e}),
        delete($joinExps{$e})
        if !exists($data{'G'}{$e});
      return(1);
    }
  print("checkData - $l list has " . scalar(keys(%{$data{$l}})) . " entries\n");
  return(0);
}

sub checkWork()
{ # Internal work - verify an entry within %work.
  local $_;
  $next{'checkWork'}{'next'} = 0
    if $next{'checkWork'}{'next'} >= length($next{'checkWork'}{'letters'});
  my ($l) = substr($next{'checkWork'}{'letters'}, $next{'checkWork'}{'next'}, 1);
  $next{'checkWork'}{$l} = 0
    if $next{'checkWork'}{$l} > $#{$work{$l}};
  ++$next{'checkWork'}{'next'};
  return(0)
    if $next{'checkWork'}{$l} > $#{$work{$l}};
  my ($i) = $next{'checkWork'}{$l}++;
  if (ref($work{$l}[$i]) ne 'HASH')
    { # bad entry; delete it
      print("checkWork - bad work entry - $l # $i\n");
      splice(@{$work{$l}}, $i, 1);
      return(1);
    }
  my ($e) = $work{$l}[$i]{'exp'};
  print("checkWork debug - $l work has " . (1 + $#{$work{$l}}) . " entries, checking M$e\n")
    if $debug & (1 << $debug{'work'});
  my (@xexps) = $mersDB->keyList('X');
  print("checkWork - no X data - " . $mersDB->dbError()),
    return(-1)
    unless @xexps;
  print("checkWork - no SPRP data - " . $mersDB->dbError()),
    return(-1)
    unless $mersDB->lowestMissingExponent('c');
  if ($e <= $xexps[0])
    { # M($e) is completely factored (including prime).
      delete $workExponents{$e};
      splice(@{$work{$l}}, $i, 1);
      return(1);
    }
  if (defined($mersDB->value('done', $e)))
    { # All factors are proven prime - D or P - so no work for this Mersenne remains.
      splice(@{$work{$l}}, $i, 1);
      delete($workExponents{$e});
      return(1);
    }
  splice(@{$work{$l}}, $i, 1),
    return(1)
    if $l =~ m/^[EoqS]$/ && $e > $mersDB->lowestMissingExponent('c');
  my (@list) = $mersDB->reservations($e);
  while (@list)
    { # reserved; delete the task unless the reservation(s) is(are) less 'urgent'
      shift(@list),
        next
        if (exists($letterPri{$l}) && exists($list[0]{'kind'}) && exists($letterPri{$list[0]{'kind'}})
            && $letterPri{$l} < $letterPri{$list[0]{'kind'}});
      print("checkWork - work $l M$e already reserved - $list[0]{'uuid'} $list[0]{'kind'}\n");
      splice(@{$work{$l}}, $i, 1);
      return(1);
    }
  if ($#{$work{$l}} > 3*$maxTasks)
    { # Plenty of work on hand - drop the "last" one.
      my ($i) = $#{$work{$l}};
      my ($e) = $work{$l}[$i]{'exp'};
      $workExponents{$e} =~ s,$l,, if exists $workExponents{$e};
      delete($workExponents{$e}) unless $workExponents{$e};
      delete($joinExps{$work{$l}[$i]{'exp'}})
        if $l =~ m,^G$,i;
      splice(@{$work{$l}}, $i, 1);
      return(1);
    }
  $workExponents{$e} = '' unless exists $workExponents{$e};
  $workExponents{$e} .= $l unless $workExponents{$e} =~ m/$l/;
  #!!likely incomplete
  #!!unless more checks are performed, $e should likely be limited to 'done' exponents
  return(0);
}

sub getNeededData(@)
{ # Internal work - call MersDB::find() to get needed data.
  my (@exps) = @_;
  local $_;
  print("getNeededData called with $exps[0]\n")
    if $#exps >= 0 && ($debug & (1 << $debug{'need'}));
  foreach (keys(%needData))
    {
      delete $needData{$_}, next unless $needData{$_};
      unshift(@exps, $_) if !@exps || $_ < $exps[0];
      last if $#exps > $minTasks*$minTasks || ($seenSIG{'IO'} && !$sigIObroken);
    }
  return(0) unless @exps;
  $#exps = $minTasks if $#exps > $minTasks && $minTasks > 20;
  print("getNeededData exps are " . join(' ', @exps) . " out of " .
        scalar(keys(%needData)) . " exponents\n")
    if $debug & (1 << $debug{'need'});
  print("getNeededData - no SPRP data - " . $mersDB->dbError()),
    return(-1)
    unless $mersDB->lowestMissingExponent('c');
  foreach (@exps)
    { # Don't look for GgJHU data for an exponent unless it's needed for checking gaps.
      $needData{$_} =~ s/[$joinExps{$_}]//g if exists $joinExps{$_};
      # Get Cc data as well unless only getting trial factoring data.
      #   Note that MersDB::find() adds PDSLMclm if any of those are asked for.
      ($needData{$_} = 'Cc' . $needData{$_})
        unless $needData{$_} =~ m/^([GgJUH]+)$/;
      # Don't look for CclmLMEoq data for "large" Mersennes - trial factoring doesn't need it.
      $needData{$_} =~ s/[CclmLMEoq]//g if $_ > 100 + $mersDB->lowestMissingExponent('c');
      delete $needData{$_}, next unless $needData{$_};
      my ($info) = '';
      print(localtime() . " getNeededData - $needData{$_} for M$_\n")
        if $debug & (1 << $debug{'need'});
      my ($wanted) = $needData{$_};
      my ($notFound) = $wanted;
      if ($needData{$_} && $needData{$_} =~ s/C//g)
        {
          my ($found) = $mersDB->find($_, 'C');
          warn("MersDB::find '$_' 'C' returned undef - " . $mersDB->dbError())
            unless defined $found;
          $info .= $found;
          $notFound =~ s,C,,g;
        }
      if ($needData{$_})
        {
          if ($needData{$_} =~ m/[GgJUH]/)
            {
              my ($letters) = $needData{$_};
              $joinExps{$_} = '' unless exists $joinExps{$_};
              while ($letters =~ s/([GgJUH])//)
                {
                  $joinExps{$_} .= $1 unless $joinExps{$_} =~ m/$1/;
                }
            }
          my ($found) = $mersDB->find($_, $needData{$_});
          warn("MersDB::find '$_' '$needData{$_}' returned undef - " . $mersDB->dbError()),
            next
            unless defined $found;
          $info .= $found;
        }
      if (!$info)
        {
          print("getNeededData - no info for M$_\n")
            if $debug & (1 << $debug{'need'});
          next;
        }
      my ($m);
      if ($_ <= $mersDB->lowestMissingExponent('c') && $needData{$_} =~ m/[clmLMS]/)
        {
          $m = Mersenne->new('n' => $_, 'mersfmt' => $info);
          warn("problems parsing data for M$_:\n  " . $m->errors()),
            next
            if $m->errors();
          $m->merge($mersennes{$_}) if exists $mersennes{$_};
          $mersennes{$_} = $m;
        }
      delete $needData{$_};
      my (@lines) = split(/[\n\r$CRLF]/, $info);
      while (@lines)
        {
          my (@mers) = split(/ /, shift(@lines));
          # J data is being lost here somehow, so print debug info.
          print("getNeededData - " . join(' ', @mers) . "\n")
            if $debug & (1 << $debug{'need'});
          next unless @mers;
          shift(@mers) if $mers[0] eq 'M(';
          next unless @mers;
          next unless $mers[0] == $_;
          $mers[0] = $_; # avoid taint failure
          next unless $mers[1] =~ s/^[)]?([$MersDB::validLetters]):?$/$1/;
          $notFound =~ s,$mers[1],,g;
          $notFound =~ s,[cS],,g
            if $mers[1] =~ m,^[clmSLM]$,;
          $notFound =~ s,l,,gi
            if $mers[1] =~ m,^l$,i;
          $notFound =~ s,m,,gi
            if $mers[1] =~ m,^m$,i;
          $notFound =~ s,g,,gi
            if $mers[1] =~ m,^g$,i;
          if ($#mers < 2)
            {
              $m->parse("M( $mers[0] )$mers[1]")
                if $m;
            }
          else
            {
              next unless $mers[2] =~ m/^([0-9]+)$/;
              $mers[2] = Math::BigInt->new("$1");
            }
          if ($#mers > 2)
            {
              next unless $mers[1] =~ m/^[EoqGgJ]$/ && $mers[3] =~ m/^([0-9]+)$/;
              $mers[3] = Math::BigInt->new("$1");
              $#mers = 3;
            }
          if ($m && $#mers > 1)
            {
              my (@v) = $m->get($mers[1]);
              if (@v)
                {
                  $notFound =~ s,$mers[1],,g;
                  $notFound =~ s,[cS],,g
                    if $mers[1] =~ m,^[clmSLM]$,;
                  $notFound =~ s,[lL],,g
                    if $mers[1] =~ m,^[lL]$,;
                  $notFound =~ s,[mM],,g
                    if $mers[1] =~ m,^[mM]$,;
                }
              if ($mers[1] =~ m/^[clmSLM]$/)
                { # cofactor digit counts are smaller with more known factors
                  $v[0] = $mers[0] unless @v && $v[0];
                  $mers[2] = $v[0] if $mers[2] > $v[0];
                }
              elsif ($mers[1] =~ m/^[Eoq]$/)
                { # E summed effort and oq stage 1 bounds are larger after more work
                  $v[0] = 0 unless @v && $v[0];
                  $v[1] = 0 unless @v && $v[1];
                  $mers[2] = $v[0] if $mers[2] < $v[0];
                  $mers[3] = $v[1] if $mers[2] <= $v[0] && ($#mers < 3 || $mers[3] < $v[1]);
                }
              # Other mers letters have no values in mers format (PD) or aren't in %mersennes (GgJKHU)
            }
          print("getNeededData calling addData " . join(' ', @mers) . "\n")
            if $debug && (1 << $debug{'need'});
          addData(@mers);
        }
      last if $seenSIG{'IO'} && !$sigIObroken;
      $wanted = $notFound;
      $notFound = '';
      while ($wanted)
        { # Something was requested but not found by MersDB::find()
          my ($none) = substr($wanted, 0, 1);
          $wanted = substr($wanted, 1);
          if ($none =~ m,^[Eoq]$,)
            { # Add "empty" task to %work as if from db (since there is nothing in the db)
              my ($aw) = addWork('skip', $_, $none, 0, 0);
              print("getNeededData - addWork '$aw' with " . join(' ', $_, $none, 0, 0) . "\n")
                if $debug & (1 << $debug{'add'});
              next;
            }
          elsif ($m)
            {
              my (@values) = $m->get($none);
              next if @values;
            }
          $notFound .= $none;
        }
      $info .= "# mers data:\n" . $m->str()
        if $notFound && $m && $m->str();
      $info .= "# Mersenne::str is empty\n"
        if $notFound && $m && !$m->str();
      my ($errs) = $m->errors()
        if $m;
      $errs =~ s/\n(.)/\n# $1/g
        if $errs;
      $info .= "# Mersenne::errors " . $errs
        if $notFound && $m && $errs;
      warn("getNeededData asked for $_ $notFound but MersDB::find found none of those\n"
           . '  ' . join(' ', caller()) . "\n# found:\n" . $info)
        if $notFound && exists($needData{$_}) && $needData{$_} =~ m/[^J]/;
    }
  return(1);
}

# Eoq work tries to factor the composite cofactors.
# S work is proving the factors and (SPRP) cofactors are all prime.
#   S work is therefore implied by the SPRP/clmLMS data that determined
#   the cofactor(s) is/are SPRP (mers letters LMS).
sub getTasksEoqS($)
{ # Argument is mers letter to get more work for.
  my ($l) = shift;
  # Other kinds of work are read from flat files (mostly trial factoring gaps from results.gaps.sort).
  delete($need{$l}),
    return(0)
    unless $l =~ m/^([EoqSG])$/;
  return(0)
    if $l =~ m/^g$/i;
  $l = "$1";
  local $_;
  # This should never die since %inputToClientType is set in init() from %clientTypes
  die("logic error: could not find client type that provides $l tasks/input\n")
    unless exists $inputToClientType{$l};
  my ($lmec) = $mersDB->lowestMissingExponent('c');
  print("getTasksEoqS - no SPRP data - " . $mersDB->dbError()),
    return(-1)
    unless $lmec;
  my ($clientType) = $inputToClientType{$l};
  my ($lastPri) = 0.0;
  my ($count) = $minTasks*$minTasks;
  $count = 100 if $sigIObroken && $minTasks > 10; # annoying but unavoidable
  my ($rows) = $mersDB->fetchRowsByPriority($l, $clientTypes{$clientType}{'lastPri'}, $count);
  warn("MersDB::fetchRowsByPriority '$l' failed - " . $mersDB->dbError()),
    return(-1)
    unless defined $rows;
  print("rows " . ($#{$rows} + 1) . " fetched with count $count from db for $l at priority $clientTypes{$clientType}{'lastPri'}\n")
    if $debug & (1 << $debug{'work'});
  if ($l eq 'S')
    { # Mersennes with two cofactors (the exponent is an odd multiple of 4) that are both SPRP
      #   are also S work - the L and M cofactors both need to be proved prime.
      my ($Lrows) = $mersDB->fetchRowsByPriority('L', $clientTypes{$clientType}{'lastPri'}, $count);
      warn("MersDB::fetchRowsByPriority 'L' failed - " . $mersDB->dbError()),
        return(-1)
        unless defined $rows;
      while (@{$Lrows})
        { # The SQL query also checks the M table and the M data is also needed.
          push(@{$rows}, [ @{$Lrows->[0]}, ], );
          my ($Mdigits) = $mersDB->value('M', $Lrows->[0][0]);
          splice(@{$Lrows->[0]}, 1, 2, 'M', $Mdigits);
          push(@{$rows}, [ @{$Lrows->[0]}, ], );
          shift(@{$Lrows});
        }
    }
  foreach (@{$rows})
    {
      my ($n) = $_->[0]; # exponent
      next if $n > $lmec;
      my ($pri) = $_->[-1];
      print(" getTasksEoqS from db for clientType $clientType M " . join(' ', @{$_}) . " pri $pri\n")
        if $debug & (1 << $debug{'work'});
      next unless $pri > 0.0;
      $lastPri = $pri if $lastPri < $pri;
      # The 'skip' here tells addWork() the info is from the db.
      unshift(@{$_}, 'skip');
      my ($aw) = addWork(@{$_});
      print("getTasksEoqS 1 addWork '$aw' with '" . join("' '", @{$_}) . "'\n")
        if $aw !~ m/^(added|highExp)$/ || $debug & (1 << $debug{'add'});
    }
  continue
    {
      last if $seenSIG{'IO'} && !$sigIObroken;
    }
  $clientTypes{$clientType}{'lastPri'} = $lastPri,
    if defined $lastPri && $lastPri >= 0.0 && $clientTypes{$clientType}{'lastPri'} < $lastPri;
  print(" getTasksEoqS from db for clientType $clientType lastPri now $lastPri\n")
    if $debug & (1 << $debug{'work'});
  # Eoq work can be "invented" for the smallest Mersenne that has never had Eoq work performed on it.
  #   But S work cannot be "invented" - it is "created" when the or both cofactors are SPRP.
  return(0)
    if $l eq 'S';
  my ($low) = $mersDB->lowestMissingExponent($l) - 1;
  while (++$low <= $mersDB->lowestMissingExponent('c'))
    {
      ++$low 
        while exists $workExponents{$low} && $workExponents{$low} =~ m/$l/i;
      my ($lowPri) = Mersenne::mersPriority($low, $l, 0, 0);
      return(0)
        if $lowPri < 0.0;
      my ($aw) = addWork($low, $l, 0, 0);
      print("getTasksEoqS 2 addWork '$aw' for created " . join(' ', 'skip', $low, $l, 0, 0) . "\n")
        if $aw !~ m/^(added|highExp)$/ || $debug & (1 << $debug{'add'});
      last
        unless $clientTypes{$clientType}{'lastPri'} > $lowPri;
    }
  return(1);
}

sub getWork()
{ # internal work - add work, mostly from the db, to %work
  #!!get G work here as well? reading from results.gaps.sort here instead of as an input only client?
  local $_;
  my ($letters) = $workLetters;
  # HU work is not supported here since that data is not in the db; instead, it is read from files.
  # G work is supported here but gets it from %data.
  $letters =~ s/[HU]//g;
  # Get S work last unless we have less than 5.
  $letters =~ s/S//g
    if $#{$work{'S'}} >= 4;
  # This sort compares the count of work in list/RAM minus need.
  my (@order) = (sort { ($#{$work{$a}} - (defined($need{$a}) ? $need{$a} : -1))
                          <=> ($#{$work{$b}} - (defined($need{$b}) ? $need{$b} : -1)) } split(//, $letters));
  push(@order, 'S')
    if $#{$work{'S'}} < 4;
  foreach (@order)
    {
      next
        unless m/^([$letters])$/ && $#{$work{"$1"}} < $maxTasks;
      my ($l) = "$1";
      if ($l eq 'G')
        {
          next unless exists $data{$l};
          my ($count) = $minTasks*$minTasks;
          foreach my $k (keys(%{$data{$l}}))
            {
              return(1)
                if --$count < 1;
              next if exists $workExponents{$k} && $workExponents{$k} =~ m,$l,;
              delete($data{$l}{$k}),
                next
                unless ref($data{$l}{$k}) eq 'ARRAY' && $#{$data{$l}{$k}} >= 0;
              my ($ref) = $data{$l}{$k}[0];
              shift(@{$data{$l}{$k}}),
                next
                unless $ref && scalar(keys(%{$ref})) > 0;
              my ($aw) = addWork($ref->{'exp'}, $l, $ref->{'one'}, $ref->{'two'});
              print("getWork - addWork '$aw' adding $ref->{'exp'} $l $ref->{'one'} $ref->{'two'}"
                    . (exists($ref->{'created'}) ? " # created $ref->{'created'}" : ' # created missing')
                    . (exists($ref->{'mod1'}) ? " mod1 $ref->{'mod1'}" : '')
                    . (exists($ref->{'mod2'}) ? " mod2 $ref->{'mod2'}" : '')
                    . "\n")
                if $aw !~ m/^(added|highExp)$/ || $debug & (1 << $debug{'add'});
              next
                if $aw eq 'reserved';
              shift(@{$data{$l}{$k}});
              delete($data{$l}{$k})
                unless ref($data{$l}{$k}) eq 'ARRAY' && $#{$data{$l}{$k}} >= 0;
              return(1)
                if $#{$work{$l}} > $maxTasks;
            }
        }
      my ($gT) = getTasksEoqS($l);
      if ($debug & (1 << $debug{'work'}))
        {
          print("getTasks(EoqS) for $l returned $gT - " . ($#{$work{$l}} + 1) . " tasks");
          print(" and " . scalar(keys(%{$data{$l}})). " data")
            if exists $data{$l};
          print("\n");
        }
      return(1)
        if $gT;
      # If no work was found, favor other kinds going forward.  This happens a lot for S.
      $need{$_} += 100
        foreach (grep { $_ ne $l } keys(%need));
    }
  return(0);
}

sub do1strs()
{ # internal work - increase the 'X' data
  local $_;
  my (@xexps) = $mersDB->keyList('X');
  warn("No X data - " . $mersDB->dbError()),
    return(-1)
    unless @xexps;
  foreach (0 .. $#xexps)
    {
      warn("xexps $_ is not numeric - '$xexps[$_]'"),
        delete($clientTypes{'OneStrs'}{'child'}),
        return(-1)
        unless $xexps[$_] =~ m/^([1-9][0-9]*)$/;
      $xexps[$_] = "$1";
    }
  $next{'xIndex'} = 0 unless $next{'xIndex'};
  ++$next{'xIndex'};
  $next{'xIndex'} %= $#xexps + 1;
  # Can't increase M( <n> )X: 0 unless M( <n+1> )[DPS]
  (++$next{'xIndex'})
    if $mersDB->value('X', $xexps[$next{'xIndex'}]) < 1;
  my ($fh);
  if ($next{'xIndex'} > $#xexps)
    { # Read results.x if the db has no or little X data.  This is invoked via init() if the db doesn't have enough X data.
      warn("not enough X data to increase it and results.x is empty or missing"),
        delete($clientTypes{'OneStrs'}{'child'}),
        return(-1)
        unless ($#xexps < 0 || ($#xexps == 0 && $mersDB->value('X', $xexps[0]) < 1)) && -s('results.x');
      if (-l('results.x') && !unlink('results.x'))
        {
          warn("not enough X data to increase it");
          print("could not unlink softlink 'results.x', $!");
          delete($clientTypes{'OneStrs'}{'child'});
          return(-1);
        }
      warn("could not open results.x, $!"),
        delete($clientTypes{'OneStrs'}{'child'}),
        return(-1)
        unless open($fh, '<', 'results.x');
      $clients{fileno($fh)} = { 'fh' => $fh, 'name' => 'results.x', 'uuid' => 'results.x', 'flag' => 'OneStrs', 'connect' => time(), };
      print(" opened results.x fd " . fileno($fh) . ' at ' . localtime() . "\n");
      return(0);
    }
  my ($now) = time();
  # Don't create another child too soon.
  return(0) if $now < $childDelay{'OneStrs'};
  # Defer to doSPRP() if that hasn't run in a bit.
  return(0) if $now < $childDelay{'SPRPChecker'};
  # Only run one child at a time.
  return(0) if grep { exists($clients{$_}{'flag'}) && exists($clientTypes{$clients{$_}{'flag'}}{'child'}) } keys(%clients);
  my ($oneStrsExe) = "./$clientTypes{'OneStrs'}{'script'}";
  warn("$oneStrsExe does not exist or is not executable"),
    delete($clientTypes{'OneStrs'}{'child'}),
    return(-1)
    unless -e("./$oneStrsExe") && -x("./$oneStrsExe");
  my ($minExp) = $xexps[$next{'xIndex'} - 1];
  my ($maxExp) = $xexps[$next{'xIndex'}];
  warn("do1strs() - min exponent $minExp is >= max exponent $maxExp"),
    delete($clientTypes{'OneStrs'}{'child'}),
    return(-1)
    if $minExp >= $maxExp; # shouldn't happen unless the X data is corrupted
  my ($start) = $mersDB->value('X', $xexps[$next{'xIndex'}]);
  warn("do1strs() - start exponent is not numeric - '$start'"),
    delete($clientTypes{'OneStrs'}{'child'}),
    return(-1)
    unless $start =~ m/^([1-9][0-9]*)$/;
  $start = "$1";
  my ($stop) = $start + $clientTypes{'OneStrs'}{'incr'};
  warn("could not run ./$oneStrsExe, $!"),
    delete($clientTypes{'OneStrs'}{'child'}),
    return(-1)
    unless open($fh, '-|', "./$oneStrsExe", "-r$minExp", "$maxExp", "$start", "$stop");
  delete($clientTypes{'OneStrs'}{'child'}),
    return(-1)
    unless unblock($fh, 'OneStrs');
  $fh->autoflush(1);
  # Do not add this descriptor to $fdbitsIn; reading it only thru checkClients() is better as it will likely print factors quite often.
  $clients{fileno($fh)} = { 'fh' => $fh, 'name' => "$oneStrsExe", 'uuid' => "$oneStrsExe", 'flag' => 'OneStrs', 'connect' => time(), };
  print("started '$oneStrsExe $minExp $maxExp $start $stop' fd " . fileno($fh) . " at " . localtime() . "\n");
  return(0);
}

sub checkSPRP()
{ # Internal work - look for exponents not in the MersDB 'sprp' table and add them to it.
  my (@xexps) = $mersDB->keyList('X');
  return(0)
    unless $next{'checkSPRP'} || $#xexps >= 0;
  local $_;
  $next{'checkSPRP'} = 1 + $xexps[0]
    unless $next{'checkSPRP'};
  ++$next{'checkSPRP'},
    return(0)
    if (defined($mersDB->value('sprp', $next{'checkSPRP'}))
        || defined($mersDB->value('done', $next{'checkSPRP'})));
  delete($next{'checkSPRP'}),
    return(0)
    if $next{'checkSPRP'} > $mersDB->lowestMissingExponent('c');
  # No data - mark it as possibly having composite factors.
  warn("update sprp $next{'checkSPRP'} 3 failed, " . $mersDB->dbError())
    unless $mersDB->update('sprp', $next{'checkSPRP'}, 3);
  return(1);
}

sub resetLastPri($)
{
  my ($exp) = @_;
  my ($pri);
  foreach (grep { exists $clientTypes{$_}{'lastPri'} } keys(%clientTypes))
    {
      my ($l) = $clientTypes{$_}{'db'};
      my (@values) = $mersennes{$exp}->get($l) if exists $mersennes{$exp};
      #!!also check $mersDB->value()?  or is that too expensive?
      @values = (0, 0) unless @values;
      $pri = Mersenne::mersPriority($exp, $l, @values);
      if ($clientTypes{$_}{'lastPri'} > $pri)
        {
          $clientTypes{$_}{'lastPri'} = $pri;
          print(" clientType $_ setting lastPri $pri after releasing $exp\n")
            if $debug & (1 << $debug{'pri'});
        }
    }
  return;
}

sub checkComposites()
{ # Internal work - try to factor a composite factor.
  return(0) unless scalar(keys(%composites));
  local $_;
  $next{'CheckComposites'} = 0
    unless $next{'CheckComposites'};
  my ($n) = $next{'CheckComposites'};
  $n = MersInfo::nextKey($n, \%composites)
    unless exists $composites{$n};
  return(0)
    unless defined $n;
  $needData{$n} .= 'C'
    if !exists $mersennes{$n} && exists $needData{$n};
  $needData{$n} = 'C'
    if !exists $mersennes{$n} && !$needData{$n};
  return(getNeededData())
    if exists $needData{$n};
  return(1)
    unless exists $mersennes{$n};
  my ($m) = $mersennes{$n};
  my ($composite);
  foreach ($m->f())
    {
      $composite = $_,
        last
        unless Math::Prime::Util::is_prime($_);
    }
  delete $composites{$n},
    return(scalar(keys(%composites)) > 0)
    unless $composite;
  if (time() < $childDelay{'CheckComposites'})
    {
      print("checkComposites M$n $composite run delayed until " . localtime($childDelay{'CheckComposites'}) . "\n"),
        $lastMsg{'CheckComposites delay'} = $n
        if (($debug & (1 << $debug{'internal'}))
            && (!exists($lastMsg{'CheckComposites delay'}) || $lastMsg{'CheckComposites delay'} != $n));
      return(1);
    }
  if (grep { exists($clients{$_}{'flag'}) && exists($clientTypes{$clients{$_}{'flag'}}{'child'}) } keys(%clients))
    {
      print("checkComposites M$n $composite run delayed until other child completes\n"),
        $lastMsg{'CheckComposites child'} = $n
        if (($debug & (1 << $debug{'internal'}))
            && (!exists($lastMsg{'CheckComposites child'}) || $lastMsg{'CheckComposites child'} != $n));
      return(1);
    }
  my (@r) = $mersDB->reserve($n, 'CheckComposites', '!');
  return(1)
    unless @r && $r[0];
  my ($fh);
  my ($filename) = "$servDir/tmp/composite.txt";
  warn("could not create $filename, $!"),
    return(0)
    unless open($fh, '>', "$filename");
  $m->print($fh);
  warn("could not close $filename, $!"),
    return(0)
    unless close($fh);
  my ($script) = $clientTypes{'CheckComposites'}{'script'};
  warn("could not run $script for M$n, $!"),
    delete($clientTypes{'CheckComposites'}{'child'}),
    return(-1)
    unless open($fh, "$script < $filename |");
  delete($clientTypes{'checkCompoosites'}{'child'}),
    return(-1)
    unless unblock($fh, 'CheckComposites');
  print("$script child started for M$n $composite\n")
    if ($debug & (1 << $debug{'internal'}));
  $clients{fileno($fh)} = { 'fh' => $fh, 'name' => 'CheckComposites', 'uuid' => 'CheckComposites',
                              'flag' => 'CheckComposites', 'connect' => time(), 'tasks' => [],
                              'infile' => $filename, 'exp' => $n, };
  delete $composites{$n},
  return(1);
}

sub doSPRP()
{ # Internal work - do the next SPRP check implied by the 'sprp' table in the db.
  #   All of the logic related to choosing which Mersenne needs an SPRP check is
  #   now in runSPRP.pl, making this a lot simpler and shorter.
  local $_;
  return(1)
    if (2*$mersDB->next('sprp') >= $mersDB->lowestMissingExponent('c')
        && time() < $childDelay{'SPRPChecker'});
  # Only run one child at a time.
  return(1)
    if grep { exists($clients{$_}{'flag'}) && exists($clientTypes{$clients{$_}{'flag'}}{'child'}) } keys(%clients);
  my ($script) = $clientTypes{'SPRPChecker'}{'script'};
  my ($fh);
  warn("could not run $script, $!"),
    delete($clientTypes{'SPRPChecker'}{'child'}),
    return(-1)
    unless open($fh, "$script < /dev/null |");
  delete($clientTypes{'SPRPChecker'}{'child'}),
    return(-1)
    unless unblock($fh, 'SPRPChecker');
  print("$script child started\n")
    if ($debug & (1 << $debug{'internal'}));
  $clients{fileno($fh)} = { 'fh' => $fh, 'name' => 'SPRP', 'uuid' => 'SPRP', 'flag' => 'SPRPChecker',
                              'connect' => time(), 'tasks' => [], };
  return(1);
}

sub trimMersennes()
{ # internal work - trim %mersennes when it gets too large
  local $_;
  $next{'trimM'} = 1 unless $next{'trimM'};
  $next{'trimM'} = MersInfo::nextKey($next{'trimM'}, \%mersennes);
  return(0) unless $next{'trimM'};
  print("trimMersennes - " . scalar(keys(%mersennes)) . " Mersennes; checking $next{'trimM'}\n")
    if $debug & (1 << $debug{'internal'});
  if ($next{'trimM'} > $mersDB->lowestMissingExponent('c'))
    {
      delete($mersennes{$next{'trimM'}}) unless exists $needData{$next{'trimM'}};
      return(1);
    }
  my (@xexps) = $mersDB->keyList('X');
  while ($next{'trimM'} <= $xexps[0])
    {
      delete($mersennes{$next{'trimM'}});
      ++$next{'trimM'};
    }
  if (defined($mersDB->value('done', $next{'trimM'})))
    {
      delete($mersennes{$next{'trimM'}}) if $mersDB->value('done', $next{'trimM'}) eq 'D';
      ++$next{'trimM'};
      return(1);
    }
  $next{'trimM'} = $next{'checkSPRP'} + 10*$minTasks,
    return(1)
    if $next{'trimM'} + 10*$minTasks >= $next{'checkSPRP'} && $next{'trimM'} - 10*$minTasks < $next{'checkSPRP'};
  return(0) if scalar(keys(%mersennes)) < $maxTasks;
  delete($mersennes{$next{'trimM'}})
    unless (exists($needData{$next{'trimM'}}) || $workExponents{$next{'trimM'}}
            || $mersDB->value('sprp', $next{'trimM'}));
  ++$next{'trimM'};
  return(1);
}

sub trimMersDB()
{ # internal work to reduce the 'values' in RAM inside $mersDB
  return($mersDB->reduceMemory());
}

sub checkEcmSaveFiles()
{ # Internal work - check and list the ecm program save files for a Mersenne.
  #   Let the checkECMfiles.pl script do all the real work.
  local $_;
  if (!exists($next{'checkEcmSaveExp'})
      || $next{'checkEcmSaveExp'} > $mersDB->lowestMissingExponent('o')
      || $next{'checkEcmSaveExp'} > $mersDB->lowestMissingExponent('q')
      || $next{'checkEcmSaveExp'} > $mersDB->lowestMissingExponent('c'))
    {
      my (@xexps) = $mersDB->keyList('X');
      return(-1) unless @xexps; #!!should never happen; see init()
      $next{'checkEcmSaveExp'} = $xexps[0] + 1;
    }
  return(0)
    unless scalar(keys(%{$needSaveFiles})) || time() >= $childDelay{'CheckEcmSaveFiles'};
  # Normally, only run one child of any type at a time but this child is disk bound, not,
  #   like all the other children, CPU bound.
  return(1)
    if grep { exists($clients{$_}{'flag'}) && $clientTypes{$clients{$_}{'flag'}} eq 'CheckEcmSaveFiles' } keys(%clients);
  $next{'checkEcmSaveExp'} = MersInfo::nextKey(0, $needSaveFiles)
    if scalar(keys(%{$needSaveFiles}));
  $next{'checkEcmSaveExp'}++
    while exists $saveFileNames{$next{'checkEcmSaveExp'}};
  my ($exp) = $next{'checkEcmSaveExp'}++;
  my (@v) = $mersDB->value('o', $exp);
  @v = $mersDB->value('q', $exp)
    unless @v;
  return(0)
    unless @v;
  my (@r) = $mersDB->reservations($exp);
  return(0)
    if @r;
  @r = $mersDB->reserve($exp, $myUUID, '!');
  return(0)
    unless @r && $r[0];
  my ($script) = $clientTypes{'CheckEcmSaveFiles'}{'script'};
  my (@args) = ( "$script", '--quiet', '--warn', );
  push(@args, '--delete', time() - $saveFileRetainTime)
    if $saveFileRetainTime;
  push(@args, "$exp", );
  my ($fh);
  warn("could not run $script, $!"),
    delete($clientTypes{'CheckEcmSaveFiles'}{'child'}),
    return(-1)
    unless open($fh, '-|', @args);
  delete($clientTypes{'CheckEcmSaveFiles'}{'child'}),
    return(-1)
    unless unblock($fh, 'CheckEcmSaveFiles');
  print("$script child started\n")
    if ($debug & (1 << $debug{'internal'}));
  $clients{fileno($fh)} = { 'fh' => $fh, 'name' => 'CheckEcmSaveFiles',
                              'uuid' => 'CheckEcmSaveFiles', 'connect' => time(),
                              'flag' => 'CheckEcmSaveFiles', 'tasks' => [], 'exp' => $exp, };
  #!!The '20' here is arbitrary to prevent the hash from growing arbitrarily since it is
  #!!  not checked elsewhere and is checked here to prevent deleting data needed right now.
  while (20*scalar(%saveFileNames) > $mersDB->lowestMissingExponent('c'))
    {
      my ($key) = each(%saveFileNames);
      delete $saveFileNames{$key};
    }
  return(0);
}

BEGIN
{ # Restrict these variables to checkInternalWork().
  # The sub()s that do internal work do not have arguments and should return 0, when
  #   there is nothing to do, any positive value (1) if there was (or likely is) more
  #   to do, and any negative value (-1) if something went wrong (in which case it
  #   will not be called again).
  # Exception: the internal work sub()s that create children should return 0 rather than
  #   a positive value after starting a new child process since those children will
  #   likely take considerable time to execute.  do1strs() adjusts its child to take
  #   $childDelay{'all'} seconds (5 minutes); doSPRP() and checkEcmSaveFiles() each can
  #   take any amount of time.
  # 'rtn' is the most recent return value; 'calls' is the number of times
  #   it has been called, 'streak' is calls in a row (without calling any
  #   of the others), 'misses' is calls that returned 0 rather than > 0.
  # All 'streak's are limited to 10 to avoid starving the other sub()s.
  # If 'exit', then skip the internal work if the server is exiting.
  #!!Other possible internal work:
  #!!verify that 0 'mersServer' '!' is reserved and stop db updates and exit if not
  #!!try to close the smallest 'G'ap in a child run of five minutes
  #!!delete $data{'J'} entries for which there are no gaps in $data{'G'}
  #!!  after $data{'J'}{$exp} is deleted, print but otherwise ignore gaps for $exp?
  #!!check 'contacts' for clients that have not connected in, e.g., 30 days?
  my (@internalWorkSubs) = 
    ( { 'name' => 'checkMersData', 'sub' => \&checkMersData,
          'rtn' => 0, 'calls' => 0, 'streak' => 0, 'misses' => 0, 'exit' => 1, },
      { 'name' => 'getWork', 'sub' => \&getWork,
          'rtn' => 0, 'calls' => 0, 'streak' => 0, 'misses' => 0, 'exit' => 0, },
      { 'name' => 'getNeededData', 'sub' => \&getNeededData,
          'rtn' => 0, 'calls' => 0, 'streak' => 0, 'misses' => 0, 'exit' => 0, },
      { 'name' => 'do1strs', 'sub' => \&do1strs, 'childType' => 'OneStrs',
          'rtn' => 0, 'calls' => 0, 'streak' => 0, 'misses' => 0, 'exit' => 1, },
      { 'name' => 'checkSPRP', 'sub' => \&checkSPRP,
          'rtn' => 0, 'calls' => 0, 'streak' => 0, 'misses' => 0, 'exit' => 1, },
      { 'name' => 'trimMersennes', 'sub' => \&trimMersennes,
          'rtn' => 0, 'calls' => 0, 'streak' => 0, 'misses' => 0, 'exit' => 1, },
      { 'name' => 'trimMersDB', 'sub' => \&trimMersDB,
          'rtn' => 0, 'calls' => 0, 'streak' => 0, 'misses' => 0, 'exit' => 1, },
      { 'name' => 'checkWork', 'sub' => \&checkWork,
          'rtn' => 0, 'calls' => 0, 'streak' => 0, 'misses' => 0, 'exit' => 1, },
      { 'name' => 'checkData', 'sub' => \&checkData,
          'rtn' => 0, 'calls' => 0, 'streak' => 0, 'misses' => 0, 'exit' => 1, },
      { 'name' => 'doSPRP', 'sub' => \&doSPRP, 'childType' => 'SPRPChecker',
          'rtn' => 0, 'calls' => 0, 'streak' => 0, 'misses' => 0, 'exit' => 1, },
      { 'name' => 'checkComposites', 'sub' => \&checkComposites, 'childType' => 'CheckComposites',
          'rtn' => 0, 'calls' => 0, 'streak' => 0, 'misses' => 0, 'exit' => 1, },
      { 'name' => 'checkEcmSaveFiles', 'sub' => \&checkEcmSaveFiles, 'childType' => 'CheckEcmSaveFiles',
          'rtn' => 0, 'calls' => 0, 'streak' => 0, 'misses' => 0, 'exit' => 1, },
    ) ;
  # In general, do internal work in rotation, one of each type before another of the same type,
  #   but try to concentrate on the internal work that is needed.
  my ($whichWork) = 0;

  sub checkInternalWork($)
    {
      my ($more) = @_;
      return(scalar(grep { $_->{'rtn'} } @internalWorkSubs)) if (!$more);
      local $_;
      $whichWork = 0 if $whichWork > $#internalWorkSubs;
      ++$whichWork, return(0)
        if ($internalWorkSubs[$whichWork]{'exit'} && -d("$stopDir") || $seenSIG{'INT'});
      print("calling internal work # $whichWork $internalWorkSubs[$whichWork]{'name'}\n") if $debug & (1 << $debug{'internal'});
      $internalWorkSubs[$whichWork]{'rtn'} = $internalWorkSubs[$whichWork]{'sub'}->();
      ++$internalWorkSubs[$whichWork]{'calls'};
      ++$internalWorkSubs[$whichWork]{'streak'};
      print("internal work # $whichWork $internalWorkSubs[$whichWork]{'name'} returned $internalWorkSubs[$whichWork]{'rtn'}\n") if $debug & (1 << $debug{'internal'});
      if ($internalWorkSubs[$whichWork]{'rtn'} < 0)
        { # Error - don't do any more of this internal work - drop it from @internalWorkSubs.
          splice(@internalWorkSubs, $whichWork, 1);
          my ($children) =  0;
          $clientTypes{$_}{'child'} = ++$children
            foreach (grep { exists($clientTypes{$_}{'child'}) } keys(%clientTypes));
        }
      elsif ($internalWorkSubs[$whichWork]{'streak'} >= 10)
        {
          $internalWorkSubs[$whichWork]{'streak'} = 0;
          ++$whichWork;
        }
      elsif ($internalWorkSubs[$whichWork]{'rtn'} < 1)
        {
          ++$internalWorkSubs[$whichWork]{'misses'};
          $internalWorkSubs[$whichWork]{'streak'} = 0;
          ++$whichWork;
        }
      else
        { # Pick the next internal work based on 'rtn's and 'misses' compared to 'calls'
          #   with a slight preference for the one after the one just called.
          my ($next) = 1 + $whichWork;
          $next = 0 if $next >= $#internalWorkSubs;
          my ($check) = -1;
          my ($skipLast);
          my ($leastCalled) = 0;
          my ($mostCalled) = 0;
          while ($internalWorkSubs[$next]{'calls'} && ++$check <= $#internalWorkSubs)
            {
              $leastCalled = $check if $internalWorkSubs[$check]{'calls'} < $internalWorkSubs[$leastCalled]{'calls'};
              $mostCalled = $check if $internalWorkSubs[$check]{'calls'} > $internalWorkSubs[$mostCalled]{'calls'};
              next if $check == $next;
              # Run each of them at least once for each internal sub (and avoid dividing by zero later).
              $next = $check, last unless $internalWorkSubs[$check]{'calls'} >= $#internalWorkSubs;
              $next = $check, next if ($internalWorkSubs[$check]{'misses'} / $internalWorkSubs[$check]{'calls'}
                                       < $internalWorkSubs[$next]{'misses'} / $internalWorkSubs[$next]{'calls'});
            }
          if (2*$internalWorkSubs[$leastCalled]{'calls'} < $internalWorkSubs[$mostCalled]{'calls'})
            {
              $whichWork = $leastCalled;
            }
          else
            {
              $whichWork = $next;
            }
        }
      print("checkInternalWork returning\n")
        if $debug & (1 << $debug{'internal'});
      return(scalar(grep { $_->{'rtn'} } @internalWorkSubs));
    }
}

sub addData(@)
{
  my (@mers) = @_;
  return unless exists $data{$mers[1]};
  local $_;
  my (@newGaps);
  if ($mers[1] eq 'J')
    { # Look for any overlapping 'G' and 'g' work and data and "trim" it appropriately.
      if (exists($workExponents{$mers[0]}) && $workExponents{$mers[0]} =~ m/^G$/i)
        { # This presumes $work{'G'} is the same as $work{'g'}.
          my ($i) = 0;
          $workExponents{$mers[0]} =~ s/G//gi;
          while ($i <= $#{$work{'G'}})
            { #!!find the work entry based on the priority of the %data entry?
              ++$i, next unless $work{'G'}[$i]{'exp'} == $mers[0];
              $workExponents{$mers[0]} .= 'G' unless $workExponents{$mers[0]} =~ m/G/i;
              my ($stop) = $work{'G'}[$i]{'two'};
              # Presume 'g' if $stop is too small since readClient() should
              #   prevent 'G' mers data from having such a small stop value.
              $stop = Math::BigInt->bone()->blsft($stop) if $stop < $mers[0];
              ++$i, next if $stop <= $mers[2]; # gap ends before or as join starts
              my ($start) = $work{'G'}[$i]{'one'};
              $start = Math::BigInt->bone()->blsft($start) if $start < $mers[0];
              ++$i, next if $start >= $mers[3]; # gap starts after or as join ends
              # Join must overlap with this gap to get this far, so:
              if ($start < $mers[2])
                { # Gap start to join start is still a gap.
                  my ($ref) = { 'exp' => $mers[0],
                                'one' => $start,
                                'two' => $mers[2],
                                'created' => __FILE__ . ":" . __LINE__,
                    };
                  push(@newGaps, $ref);
                }
              if ($mers[3] < $stop)
                { # Gap from join end to gap end is still a gap.
                  my ($ref) = { 'exp' => $mers[0],
                                'one' => $mers[3],
                                'two' => $stop,
                                'created' => __FILE__ . ":" . __LINE__,
                    };
                  push(@newGaps, $ref);
                }
              # Delete gap that was covered, trimmed, or split
              splice(@{$work{'G'}}, $i, 1);
            }
        }
      if (exists($data{'g'}{$mers[0]}))
        { # Joins effectively never start and stop at a power of two,
          #   so convert these gaps to G style first.
          while (@{$data{'g'}{$mers[0]}})
            {
              my ($ref) = shift(@{$data{'g'}{$mers[0]}});
              my ($start) = Math::BigInt->bone()->blsft($ref->{'one'});
              my ($stop) = Math::BigInt->bone()->blsft($ref->{'two'});
              $ref->{'one'} = $start;
              $ref->{'two'} = $stop;
              $ref->{'created'} = __FILE__ . ":" . __LINE__;
              push(@{$data{'G'}{$mers[0]}}, $ref);
            }
          delete $data{'g'}{$mers[0]};
        }
      if (exists($data{'G'}{$mers[0]}))
        { #!!similar to code above for the gaps in %work; create a sub?
          #!!but the gap can be changed in place here, unlike above
          my ($i) = 0;
          while ($i <= $#{$data{'G'}{$mers[0]}})
            { # Next gap if join starts after or as gap ends
              ++$i, next if $mers[2] >= $data{'G'}{$mers[0]}[$i]{'two'};
              # Next gap if join ends before or as gap starts
              ++$i, next if $mers[3] <= $data{'G'}{$mers[0]}[$i]{'one'};
              # Delete covered gap
              splice(@{$data{'G'}{$mers[0]}}, $i, 1),
                next
                if ($mers[2] <= $data{'G'}{$mers[0]}[$i]{'one'}
                    && $mers[3] >= $data{'G'}{$mers[0]}[$i]{'two'});
              if ($data{'G'}{$mers[0]}[$i]{'one'} < $mers[2]
                  && $data{'G'}{$mers[0]}[$i]{'two'} > $mers[3])
                { # Split gap that includes all of the new join.
                  my ($ref) = { 'exp' => $mers[0], # exponent
                                'one' => $mers[3], # end of join
                                'two' => $data{'G'}{$mers[0]}[$i]{'two'}, # end of gap
                                'created' => __FILE__ . ":" . __LINE__,
                    };
                  # Splice in a "new" gap that starts after the join.
                  splice(@{$data{'G'}{$mers[0]}}, $i + 1, 0, $ref);
                  # Modify the "old" gap to end at the start of the join.
                  $data{'G'}{$mers[0]}[$i]{'one'} = $mers[2];
                  $data{'G'}{$mers[0]}[$i]{'mod1'} = __FILE__ . ":" . __LINE__,
                  $i += 2; # move past both
                  next;
                }
              # Trim gap that ends within the join
              $data{'G'}{$mers[0]}[$i]{'one'} = $mers[2],
                $data{'G'}{$mers[0]}[$i]{'mod2'} = __FILE__ . ":" . __LINE__,
                if $data{'G'}{$mers[0]}[$i]{'two'} > $mers[2];
              # Trim gap that starts within the join
              $data{'G'}{$mers[0]}[$i]{'two'} = $mers[3],
                $data{'G'}{$mers[0]}[$i]{'mod1'} = __FILE__ . ":" . __LINE__,
                if $data{'G'}{$mers[0]}[$i]{'one'} < $mers[3];
              ++$i;
            }
        }
    }
  elsif ($mers[1] =~ m/^G$/i)
    {
      while (defined($mers[1]) && exists($data{'J'}{$mers[0]}))
        { # Look for overlapping join (J) data.
          my ($start) = ($mers[1] eq 'G' ? $mers[2] : Math::BigInt->bone()->blsft($mers[2]));
          my ($stop) = ($mers[1] eq 'G' ? $mers[3] : Math::BigInt->bone()->blsft($mers[3]));
          foreach (@{$data{'J'}{$mers[0]}})
            { # Next join if gap stops before or as join starts.
              next if $stop <= $_->{'one'};
              # Next join if gap starts after or as join ends.
              next if $start >= $_->{'two'};
              # Done if this gap is "contained" by the join, but check @newGaps.
              $mers[1] = undef, last if ($start >= $_->{'one'} && $stop <= $_->{'two'});
              if ($start < $_->{'one'} && $stop > $_->{'two'})
                { # Split the gap if it "contains" the join.
                  my ($ref) = { 'exp' => $mers[0], 'one' => $start, 'two' => $_->{'one'},
                                  'created' => __FILE__ . ":" . __LINE__,
                    };
                  push(@newGaps, $ref);
                  $mers[3] = $stop, $mers[1] = 'G' if $mers[1] eq 'g';
                  $mers[2] = $_->{'two'};
                  next;
                }
              # Trim gap that starts within the join.
              $start = $_->{'one'} if $start > $_->{'one'};
              # Trim gap that ends within the join.
              $stop = $_->{'two'} if $stop < $_->{'two'};
            }
          # Any remaining Gg data needs to be given to addWork().
          last if defined $mers[1];
          if (@newGaps && !defined $mers[1])
            {
              my ($ref) = pop(@newGaps);
              @mers = ($ref->{'exp'}, 'G', $ref->{'one'}, $ref->{'two'});
            }
        }
    }
  elsif ($mers[1] =~ m/^[HU]$/)
    { #!!skip for now; this is incomplete due to interactions between HUC
      #!!delete($data{$mers[1]}{$mers[0]})
      #!!  if exists $data{$mers[1]}{$mers[0]} && ref($data{$mers[1]}{$mers[0]}) ne 'ARRAY';
      #!!($data{$mers[1]}{$mers[0]}[0] = { 'exp' => $mers[0], 'one' => $mers[2], })
      #!!  unless (exists($data{$mers[1]}{$mers[0]}) && ref($data{$mers[1]}{$mers[0]}) eq 'ARRAY'
      #!!          && $data{$mers[1]}{$mers[0]}[0]{'one'} > $mers[2]);
      return;
    }
  else
    {
      warn("addData() - unsupported " . join(' ', @mers));
      print("  data keys are " . join(' ', keys(%data)) . "\n")
        if $debug;
      return;
    }
  while (@newGaps)
    { # These gaps should have already been checked against existing gap and join data
      my ($ref) = pop(@newGaps);
      push(@{$data{'G'}{$ref->{'exp'}}}, $ref);
      my ($aw) = addWork($ref->{'exp'}, 'G', $ref->{'one'}, $ref->{'two'});
      print("addData 1 addWork '$aw' adding " . join(' ', $ref->{'exp'}, 'G', $ref->{'one'}, $ref->{'two'})
            . ' # ' . (exists($ref->{'created'}) ? ("created " . $ref->{'created'}) : '<created missing>')
            . (exists($ref->{'mod1'}) ? " mod1 " . $ref->{'mod1'} : '')
            . (exists($ref->{'mod2'}) ? " mod2 " . $ref->{'mod1'} : '')
            . "\n")
        if $aw !~ m/^(added|highExp)$/ || $debug & (1 << $debug{'add'});
    }
  return unless defined $mers[1];
  if (exists($data{$mers[1]}{$mers[0]}))
    {{ # Look for overlapping and neighboring data and merge appropriately.
      my ($i) = 0;
      while ($i <= $#{$data{$mers[1]}{$mers[0]}})
        { # Next entry if it ends before the new data starts
          ++$i, next if $data{$mers[1]}{$mers[0]}[$i]{'two'} < $mers[2];
          # Next entry if it starts after the new data ends
          ++$i, next if $data{$mers[1]}{$mers[0]}[$i]{'one'} > $mers[3];
          # If this entry "covers" the new data, there's nothing more to do
          $mers[1] = undef, # "erase" it so it isn't push()d
            last
            if ($data{$mers[1]}{$mers[0]}[$i]{'one'} <= $mers[2]
                && $data{$mers[1]}{$mers[0]}[$i]{'two'} >= $mers[3]);
          # They overlap or meet; combine them and start over.
          ($mers[2] = $data{$mers[1]}{$mers[0]}[$i]{'one'})
            if $mers[2] > $data{$mers[1]}{$mers[0]}[$i]{'one'};
          ($mers[3] = $data{$mers[1]}{$mers[0]}[$i]{'two'})
            if $mers[3] < $data{$mers[1]}{$mers[0]}[$i]{'two'};
          splice(@{$data{$mers[1]}{$mers[0]}}, $i, 1);
          $i = 0;
        }
      if (defined($mers[1]))
        {
          my ($ref) = { 'exp' => $mers[0], 'one' => $mers[2], 'two' => $mers[3],
            'created' => __FILE__ . ":" . __LINE__, };
          push(@{$data{$mers[1]}{$mers[0]}}, $ref);
        }
      # Could check %work for overlaps with this new gap.
      #!!Overlapping gaps should be rare and don't really matter, since any joins
      #!!  returned by clients will be checked against all gaps in %data and %work.
      #!!The code (the while and if just above) is the same except that gaps also need
      #!!  to be checked in %work.
      if (defined($mers[1]) && exists($work{$mers[1]}))
        {
          my ($aw) = addWork(@mers);
          print("addData 2 addWork '$aw' adding " . join(' ', @mers) . "\n")
            if $aw !~ m/^(added|highExp)$/ || $debug & (1 << $debug{'add'});
        }
    }}
  return unless defined $mers[1];
  my ($mersRef) = { 'exp' => $mers[0], 'one' => $mers[2], 'two' => $mers[3],
                      'created' => __FILE__ . ":" . __LINE__,
    };
  push(@{$data{$mers[1]}{$mers[0]}}, $mersRef);
}

sub addGap($@)
{
  my ($exp, @rest) = @_;
  warn("wrong field count for C line for $exp - 'M( $exp )C: " . join(' ', @rest) . "'"),
    return
    if $#rest != 0;
  my (@xexps) = $mersDB->keyList('X');
  return unless $exp > $xexps[0];

  # This doesn't really belong here but it would require calling MersDB::exps('X') again if in the caller.
  $next{'checkSPRP'} = $exp if $exp < $next{'checkSPRP'};

  # Trial factoring (including gap closing) is only done for prime exponent Mersenne numbers.
  return unless Math::Prime::Util::is_prime($exp);
  my ($x) = 1;
  ++$x while ($x <= $#xexps && $xexps[$x] < $exp);
  my (@gap) = ($exp, 'G', );
  push(@gap, ($x <= $#xexps
              ? $mersDB->value('X', $xexps[$x]) # 1strs has been done farther than ...
              : 2*$exp + 1));                   # ... the smallest possible factor.
  return unless $gap[-1] < $rest[0]; # don't create "gaps" consisting of only the factor
  push(@gap, $rest[0]);
  push(@{$mersData{$exp}}, [ @gap, ]);
  return if exists $joinExps{$exp} && $joinExps{$exp} =~ m/J/;
  $needData{$exp} = '' unless exists $needData{$exp};
  $needData{$exp} .= 'J' unless $needData{$exp} =~ m/J/;
}

# Helper for and addWork() - return the index to insert new work
sub binarySearch($$)
{
  my ($list, $pri) = @_;
  my ($low) = 0;
  my ($hi) = $#{$list};
  while ($low + 1 < $hi)
    {
      my ($mid) = $hi + $low;
      --$mid if $mid % 2;
      $mid /= 2;
      $hi = $mid,
        next
        if $pri < $list->[$mid]{'pri'};
      $low = $mid;
      next;
    }
  ++$hi if $list->[$hi]{'pri'} < $pri;
  return($hi);
}

sub addWork
{ # A first argument of 'skip' indicates the info is from the db.
  my (@mers);
  push(@mers, shift(@_))
    while (@_);
  warn("addWork() - no or undef arguments given - caller " . join(' ', caller()) . "\n"),
    return('invalid')
    unless @mers && $mers[0];
  my ($skipUpdate);
  $skipUpdate = 1,
    shift(@mers)
    if $mers[0] eq 'skip';
  warn("addWork() - invalid exponent '$mers[0] - '" . join(' ', @mers) . "' - caller " . join(' ', caller()) . "\n"),
    return('invalid')
    unless ref($mers[0]) eq 'Math::BigInt' || $mers[0] =~ m,^([1-9][0-9]*)$,;
  warn("addWork() - no mers letter given - last index $#mers - '" . join("' '", @mers) . "' caller " . join(' ', caller()) . "\n"),
    return('invalid')
    unless $#mers >= 0;
  warn("addWork() - missing first argument - last index $#mers - '" . join("' '", @mers) . "' caller " . join(' ', caller()) . "\n"),
    return('invalid')
    unless $#mers > 0;
  warn("addWork() - mers letter '$mers[1]' is not in letterPri HASH - '" . join(' ', @mers) . "' caller " . join(' ', caller()) . "\n"),
    return('notWork')
    unless defined $letterPri{$mers[1]};
  $mers[0] = Math::BigInt->new("$1") unless ref($mers[0]) eq 'Math::BigInt';
  warn("addWork() - mers letter '$mers[1]' not in work HASH - '" . join(' ', @mers) . "' caller " . join(' ', caller()) . "\n"),
    delete($need{$mers[1]}),
    return('notWork')
    unless exists $work{$mers[1]};
  # Mersennes that are 'done' (DP) need no more work.
  if ($mersDB->value('done', $mers[0]))
    {
      print("addWork() - M $mers[0] DP so not work - '" . join(' ', @mers) . "' - caller " . join(' ', caller()) . "\n")
        if $debug & (1 << $debug{'add'});
      return('done');
    }
  if (($mers[1] !~ m,^([GgHU])$, || $mers[0] <= $mersDB->lowestMissingExponent('c')) && $mersDB->value('sprp', $mers[0]))
    {
      print("addWork() - M $mers[0] needs SPRP check " . $mersDB->value('sprp', $mers[0]) . " - '" . join(' ', @mers) . "' - caller " . join(' ', caller()) . "\n")
        if $debug & (1 << $debug{'add'});
      return('sprp');
    }
  local $_;
  my ($mersRef) = { 'exp' => $mers[0], 'pri' => Mersenne::mersPriority(@mers), };
  if ($mersRef->{'pri'} < 0.0)
    {
      print("addWork() - '" . join("' '", @mers) . "' priority is $mersRef->{'pri'} - caller " . join(' ', caller()) . "\n")
        if $debug & (1 << $debug{'add'});
      return('negPri');
    }
  $mersRef->{'one'} = $mers[2] if $#mers >= 2;
  $mersRef->{'two'} = $mers[3] if $#mers >= 3;
  $mersRef->{'created'} = __FILE__ . ":" . __LINE__;
  if ($mers[1] !~ m/^[GgHU]$/ && !exists($mersennes{$mers[0]}))
    { # Need more data from the db.
      $needData{$mers[0]} .= 'C' . "$mers[1]" if exists $needData{$mers[0]} && $needData{$mers[0]} !~ m,C,;
      $needData{$mers[0]} = 'C' . "$mers[1]" unless exists $needData{$mers[0]};
    }
  if ($mers[1] !~ m/^[Eoq]$/ && exists($mersennes{$mers[0]}))
    { # Use data from %mersennes if that's "further" or "covers" this work.
      my (@values) = $mersennes{$mers[0]}->get($mers[1]);
      if (@values)
        {
          if ($mers[2] < $values[0])
            {
              $mersRef->{'one'} = $values[0];
              $mersRef->{'mod1'} = __FILE__ . ":" . __LINE__;
              $mersRef->{'two'} = $values[1],
                $mersRef->{'mod2'} = __FILE__ . ":" . __LINE__,
                if $#values > 0;
            }
          elsif ($mers[2] == $values[0] && $#values > 0)
            {
              $mersRef->{'two'} = $values[1],
                $mersRef->{'mod2'} = __FILE__ . ":" . __LINE__,
                if $mers[3] < $values[1];
            }
        }
    }
  $workExponents{$mers[0]} = '' unless exists $workExponents{$mers[0]};
  if ($mers[1] =~ m/^[EoqHU]$/ && $workExponents{$mers[0]} =~ m/$mers[1]/)
    { # Existing EoqHU work needs to be replaced if this work is "farther".
      my ($i) = 0;
      my ($l) = $mers[1];
      while ($i <= $#{$work{$l}})
        {
          ++$i, next unless $work{$l}[$i]{'exp'} == $mers[0];
          warn("addWork - unexpected first value - '" . join(' ', @mers) . "'"),
            return('invalid')
            if $#mers < 2 || $mers[2] !~ m,^[0-9]+$,;
          warn("addWork - invalid work entry - '$work{$l}[$i]{'exp'} <undef>'"),
            splice(@{$work{$l}}, $i, 1),
            next
            unless exists $work{$l}[$i]{'one'};
          warn("addWork - invalid work entry - '$work{$l}[$i]{'exp'} $work{$l}[$i]{'one'}'"),
            splice(@{$work{$l}}, $i, 1),
            next
            if $work{$l}[$i]{'one'} !~ m,^[0-9]+$,;
          if ($work{$l}[$i]{'one'} > $mers[2])
            { # existing work is "farther"
              print("addWork() - existing work is farther - 'one' is $work{$l}[$i]{'one'}\n")
                if $debug & (1 << $debug{'add'});
              return('lesser');
            }
          splice(@{$work{$l}}, $i, 1),
            last
            if $work{$l}[$i]{'one'} < $mers[2]; # "new" work is "farther"
          if ($#mers <= 2)
            { # new has no second value so old must be "farther" or identical (including HU)
              print("addWork() - existing work is farther - no second value\n")
                if $debug & (1 << $debug{'add'});
              return('lesser2');
            }
          warn("addWork - unexpected second value - '" . join(' ', @mers) . "'"),
            return('invalid')
            if $mers[3] !~ m,^[0-9]+$,;
          splice(@{$work{$l}}, $i, 1),
            last
            unless exists $work{$l}[$i]{'two'}; # existing work has no second value
          warn("addWork - invalid work entry - '$work{$l}[$i]{'exp'} $work{$l}[$i]{'one'} $work{$l}[$i]{'two'}'"
               . (exists($work{$l}[$i]{'created'}) ? " # created $work{$l}[$i]{'created'}" : ' # created missing')
               . (exists($work{$l}[$i]{'mod1'}) ? " mod1 $work{$l}[$i]{'mod1'}" : '')
               . (exists($work{$l}[$i]{'mod2'}) ? " mod2 $work{$l}[$i]{'mod2'}" : '')
            ),
            splice(@{$work{$l}}, $i, 1),
            next
            if $work{$l}[$i]{'two'} !~ m,^[0-9]+$,;
          if ($work{$l}[$i]{'two'} >= $mers[3])
            { # E efforts or oq stage 1 bounds must be the same - compare E largest stage 1 bounds or oq stage 2 bounds.
              print("addWork() - existing work is farther - 'two' is $work{$l}[$i]{'two'}"
                    . (exists($work{$l}[$i]{'created'}) ? " # created $work{$l}[$i]{'created'}" : ' # created missing')
                    . (exists($work{$l}[$i]{'mod1'}) ? " mod1 $work{$l}[$i]{'mod1'}" : '')
                    . (exists($work{$l}[$i]{'mod2'}) ? " mod2 $work{$l}[$i]{'mod2'}" : '')
                    . "\n")
                if $debug & (1 << $debug{'add'});
              return('lesser2');
            }
          splice(@{$work{$l}}, $i, 1); # "new" work has larger second value
        }
    }
  else
    {
      $workExponents{$mers[0]} .= $mers[1] unless $workExponents{$mers[0]} =~ m/$mers[1]/i;
    }
  if (!$skipUpdate && $mers[1] !~ m/^[GgHJUe]$/)
    { # MersDB::update() does not support GgJHUe data
      print("addWork() updating db with " . join(' ', @mers) . "\n")
        if $debug & (1 << $debug{'add'});
      $mersDB->update($mers[1], @mers);
    }
  return('highExp') # do not retain Eoq work past the clmLM "boundary"
    if $mers[1] =~ m/^[Eoq]$/ && $mers[0] > 100 + $mersDB->lowestMissingExponent('c');
  my (@r) = $mersDB->reservations($mers[0]);
  while (@r)
    {
      warn("reservation for $mers[0] has no 'kind'"),
        next
        unless exists $r[0]->{'kind'};
      warn("reservation for $mers[0] has unexpected 'kind' '" . $r[0]->{'kind'} . "' not in letterPri HASH"),
        next
        unless exists $letterPri{$r[0]->{'kind'}};
      if ($letterPri{$mers[1]} > $letterPri{$r[0]->{'kind'}})
        {
          print("addWork - " . join(' ', @mers) . " is reserved - $r[0]->{'uuid'} for $r[0]->{'kind'}\n")
            if $debug & (1 << $debug{'add'});
          return('reserved');
        }
      warn("addWork could not release '$r[0]->{'exp'}, $r[0]->{'uuid'}, $r[0]->{'kind'}' - " . $mersDB->dbError())
        unless release($r[0]->{'exp'}, $r[0]->{'uuid'}, $r[0]->{'kind'});
    }
  continue
    {
      shift(@r);
    }
  $need{$mers[1]}--
    if $need{$mers[1]};
  $needSaveFiles->{$mers[0]} = 1
    if ($mers[1] =~ m,^[oq]$, && !exists $saveFileNames{$mers[0]}
        && !exists $saveFileNames{$mers[0]}{$mers[1]});
  push(@{$work{$mers[1]}}, $mersRef),
    return('added')
    if $#{$work{$mers[1]}} < 0 || $work{$mers[1]}[-1]{'pri'} <= $mersRef->{'pri'};
  unshift(@{$work{$mers[1]}}, $mersRef),
    return('added')
    if $mersRef->{'pri'} < $work{$mers[1]}[0]{'pri'};
  my ($where) = binarySearch($work{$mers[1]}, $mersRef->{'pri'});
  print("adding work - " . join(' ', @mers) . " at $where of $#{$work{$mers[1]}}\n")
    if $debug & (1 << $debug{'add'});
  splice(@{$work{$mers[1]}}, $where, 0, $mersRef);
  return('added');
}

sub release($$$)
{ # Call MersDB::release() and put relevant tasks back in %work.
  my ($exp, $uuid, $kind) = @_;
  local $_;
  warn("release() called with undef for exponent - caller " . join(' ', caller()) . "\n"),
    return(undef)
    unless defined $exp;
  warn("release() called with non-numeric exponent - caller " . join(' ', caller()) . "\n"),
    return(undef)
    unless $exp =~ m,^([1-9][0-9]+)$,;
  $exp = $1;
  warn("release called with undef for UUID - caller " . join(' ', caller()) . "\n"),
    return(undef)
    unless defined $uuid;
  warn("release() called with unexpected/invalid UUID - caller " . join(' ', caller()) . "\n"),
    return(undef)
    unless $uuid =~ m,$uuidPattern,;
  $uuid = $1;
  warn("release called with kind undef; using '?'"),
    $kind = '?'
    unless defined($kind);
  # 'c' is SPRP checking; 0 for 'mersServer'; '?' is a default
  warn("release called with kind '$kind'")
    if $kind !~ m,^[c!?$workLetters]$,;
  my (@r) = $mersDB->release($exp, $uuid, $kind);
  return(@r ? $r[0] : undef)
    if $exp < 1;
  warn("could not release $exp $uuid $kind - '" . $mersDB->dbError()),
    return(@r)
    if $#r < 0 || !$r[0];
  if (exists($work{$kind}) && exists($data{$kind}) && exists($data{$kind}{$exp})
      && $#{$data{$kind}{$exp}} >= 0)
    {
      my (@values);
      push(@values, $data{$kind}{$exp}[0]{'one'})
        if exists $data{$kind}{$exp}[0]{'one'};
      push(@values, $data{$kind}{$exp}[0]{'two'})
        if @values && exists $data{$kind}{$exp}[0]{'two'};
      my ($aw) = addWork($exp, $kind, @values);
      print("release - addWork '$aw' adding " . join(' ', $exp, $kind, @values) . "\n")
        if $aw !~ m/^(added|highExp)$/ || $debug & (1 << $debug{'add'});
    }
  return(@r);
}

sub catchSIG
{
  my ($sig) = $_[0];
  $SIG{$sig} = \&catchSIG;
  $DB::signal = 1;
  die("SIGQUIT while debug is $debug")
    if $sig eq 'QUIT' && $debug;
  ++$seenSIG{$sig};
  ++$seenSIG{'INT'} if $sig eq 'TERM';
  ++$seenSIG{'INT'} if $sig eq 'ABRT';
}

sub refetchAliases()
{
  local $_;
  print("refetching aliases\n");
  my (@rows) = $mersDB->fetchRows('aliases');
  foreach (@rows)
    {
      die("MersDB::fetchRows(aliases) returned wrong number of fields - '"
          . join(' ', @{$_}) . "'")
        unless $#{$_} == 1;
      my ($uuid, $alias) = @{$_};
      die("UUID $uuid of alias '$alias' is itself an alias for '$UUIDaliases{$uuid}' in table 'aliases'\n")
        if exists $UUIDaliases{$uuid} && $UUIDaliases{$uuid} ne $uuid;
      $UUIDaliases{$alias} = $uuid;
      $UUIDaliases{$uuid} = $uuid;
    }
}

sub refetchAllowed()
{
  local $_;
  print("refetching allowed\n");
  my (@rows) = $mersDB->fetchRows('allowed');
  foreach (@rows)
    {
      die("MersDB::fetchRows(allowed) returned wrong number of fields - '"
          . join(' ', @{$_}) . "'")
        unless $#{$_} == 2;
      my ($uuid, $letters, $tasks) = @{$_};
      warn("UUID $uuid allowed letters '$letters' includes letters that are not mers work; ignored"),
        next
        unless $letters =~ m/^[$workLetters]+$/;
      warn("UUID $uuid max tasks '$tasks' outside range of 1 to $maxTasks; ignored"),
      next
        unless $tasks > 0 && $tasks <= $maxTasks;
      $allowed{$uuid} = { 'letters' => $letters, 'tasks' => $tasks, };
    }
}

sub refetchContact()
{
  my (@rows) = $mersDB->fetchRows('contact');
  print("refetching connect\n");
  foreach (@rows)
    {
      die("MersDB::fetchRows(contact) returned wrong number of fields - '"
          . join(' ', @{$_}) . "'")
        unless $#{$_} == 1;
      my ($uuid, $time) = @{$_};
      $lastContact{$uuid} = $time;
    }
}

sub shiftToggle
{
  $SIG{$_[0]} = \&shiftToggle;
  ++$debugBit;
  $debugBit = 0
    if $debugBit > $maxDebugBit;
}  

sub toggleDebug
{
  $SIG{$_[0]} = \&toggleDebug;
  $debug ^= (1 << $debugBit);
}

sub init()
{ # SIGINT and SIGTERM tell the server to (safely) exit, sooner for SIGTERM.
  # The others here are ignored unless debugging, in which case 'QUIT' is effective.
  foreach my $sig ('ABRT', 'HUP', 'INT', 'TERM', 'IO', 'QUIT', 'LOST', 'PIPE')
    {
      $SIG{$sig} = \&catchSIG if exists $SIG{$sig};
    }
  if (exists($SIG{'USR1'}) && exists($SIG{'USR2'}))
    { # change $debug if SIGUSR1; SIGUSR2 changes bit that SIGUSR1 will change
      $SIG{'USR1'} = \&toggleDebug;
      $SIG{'USR2'} = \&shiftToggle;
    }
  #!!else ... two other signals?

  # Ubuntu 18.04.4 Linux either does not support SIGIO or it's broken for Perl;
  #   check it and act accordingly later.
  warn("kill(SIGIO, self) failed, $!"),
    $sigIObroken = 1
    unless kill('SIGIO', $$);
  if (!$sigIObroken && !$seenSIG{'IO'})
    { # pretend SIGIO is seen every time select() is called
      warn("SIGIO sent but not delivered to signal handler; marking it broken");
      $sigIObroken = 1;
      $seenSIG{'IO'} = 1;
    }

  my ($port); # TCP port
  my ($file);
  while (@ARGV)
    {
      my ($opt) = shift(@ARGV);
      print($usage), exit(0) if $opt =~ m/^-([hH?]|-help)$/;
      $port = shift(@ARGV), next if $opt eq '-p';
      if ($opt eq '--debug')
        {
          my ($arg) = shift(@ARGV);
          $debug ^= (1 << $debug{$arg}), next if exists $debug{$arg};
          $debug = $1, next if $arg =~ m,^([0-9]+)$,;
          die("'--debug' argument '$arg' unexpected\n");
        }
      next if $opt eq '--check' || $opt eq '--version';
      if ($opt eq '-s')
        {
          $file = shift(@ARGV);
          die("file $file given with -s does not exist or is empty\n")
            unless -f $file && -s $file;
          push(@{$clientTypes{'SortedByPriority'}{'files'}}, $file);
          next;
        }
      if ($opt eq '-S')
        {
          $file = shift(@ARGV);
          die("file $file given with -S does not exist or is empty\n")
            unless -f $file && -s $file;
          push(@{$clientTypes{'SortedByExponent'}{'files'}}, $file);
          next;
        }
      if ($opt eq '-P')
        {
          $file = shift(@ARGV);
          die("file $file given with -P does not exist or is empty\n")
            unless -f $file && -s $file;
          push(@{$clientTypes{'DataToPrint'}{'files'}}, $file);
          next;
        }
      if ($opt eq '--saveFileRetainTime')
        {
          $saveFileRetainTime = shift(@ARGV);
          die("argument '$saveFileRetainTime' given with --saveFileRetainTime is not a number\n")
            unless $saveFileRetainTime =~ m,^([1-9][0-9]*)$,;
          $saveFileRetainTime = Math::BigInt->new("$1");
          $saveFileRetainTime *= 24*60*60;
          next;
        }
      die("unknown option '$opt'\n")
        if substr($opt, 0, 1) eq '-';
      die("file $opt given with no option does not exist or is empty\n")
        unless -f $opt && -s $opt;
      push(@{$clientTypes{'DataFile'}{'files'}}, $opt);
    }

  #!!second server port, one for handing out work and the other for reading results?
  local $_;
  # M2203 and M2281 are prime.
  $port = 2281 if !$port && -f 'mysql.devel';
  $port = 2203 if !$port;
  die("invalid port '$port'; must be alphanumeric\n")
    unless $port =~ /^([[:alnum:]]+)$/;
  $port = $1;
  $port = getservbyname($port, 'tcp') if $port !~ m/^\d+$/;
  die("invalid port '$port'; must be numeric or in /etc/services by name\n")
    unless $port =~ /^\d+$/x;
  socket($serverfh, PF_INET, SOCK_STREAM, $proto)
    or die("socket: $!");
  setsockopt($serverfh, SOL_SOCKET, SO_REUSEADDR, pack("l", 1))
    or die("setsockopt SO_REUSEADDR: $!");
  setsockopt($serverfh, SOL_SOCKET, SO_KEEPALIVE, pack("l", 1))
    or die("setsockopt SO_KEEPALIVE: $!");
  setsockopt($serverfh, SOL_SOCKET, SO_OOBINLINE, pack("l", 1))
    or warn "setsockopt SO_OOBINLINE: $!";

  # We do not read STDIN ...
  close(STDIN)
    or die("could not close STDIN, $!");
  # ... but perl -w complains if STDIN is opened only for writing.
  open(STDIN, '<', '/dev/null')
    or die("could not open /dev/null for STDIN, $!");
  # Flush these every time in case of a crash.
  STDERR->autoflush(1);
  STDOUT->autoflush(1);
  eval
    { # setpriority() is not always implemented
      setpriority(0, 0, 20);
    };

  if (!$checkConfig && $mersDB && ref($mersDB) eq 'MersDB')
    {
      $mersDB->addSortedFiles($clientTypes{'SortedByExponent'}{'files'});
      delete($clientTypes{'SortedByExponent'}{'files'});
      $dbh = $mersDB->connect($myUUID);
      # Release anything that was reserved by a prior run.
      my (@dbResults) = $mersDB->reservations($myUUID);
      push(@dbResults, $mersDB->reservations(MersDB::specialExponent($myUUID)));
      foreach (@dbResults)
        {
          print("reservation ($_->{'exp'}, $_->{'uuid'}, $_->{'kind'}) exists; exiting\n"),
            exit(1)
            unless $_->{'uuid'} eq $myUUID;
          print(" released ($_->{'exp'}, $_->{'uuid'}, $_->{'kind'})\n"),
            next
            if exists($_->{'exp'}) && $mersDB->release($_->{'exp'}, $_->{'uuid'}, $_->{'kind'});
          die(" could not release ($_->{'exp'}, $_->{'uuid'}, $_->{'kind'}), "
              . $mersDB->dbError());
        }
      # If this reserve() fails, let the die() happen; something else is running that overlaps with this.
      @dbResults = $mersDB->reserve(MersDB::specialExponent($myUUID), $myUUID, '!');
      die(" reserve(" . MersDB::specialExponent($myUUID) . ", $myUUID, '!') returned " . join(' ', @dbResults) . "\n"
          . '  ' . $mersDB->dbError())
        unless @dbResults;
      $reserved0 = 1;

      @dbResults = $mersDB->fetchValues('missing', 'done', 'X', 'S', 'L', 'M');
      die("unexpected results from MersDB::fetchValues(missing, done, XSLM) at startup:\n"
          . '  ' . join(' ', @dbResults) . "\n"
          . '  ' . $mersDB->dbError())
        if ($#dbResults != 3*6 - 1 # 'fetch', $table, $cnt for each of 'missing', 'done', XSLM
            || $dbResults[0] ne 'fetch' || $dbResults[1] ne 'missing'
            || $dbResults[3] ne 'fetch' || $dbResults[4] ne 'done'
            || $dbResults[6] ne 'fetch' || $dbResults[7] ne 'X'
            || $dbResults[9] ne 'fetch' || $dbResults[10] ne 'S'
            || $dbResults[12] ne 'fetch' || $dbResults[13] ne 'L'
            || $dbResults[15] ne 'fetch' || $dbResults[16] ne 'M'
            # missing cEoqU, done thru at least 4, and at least one X line (with value 0).
            || $dbResults[2] < 5 || $dbResults[5] < 3 || $dbResults[8] < 1);
      print('initial db fetch results - ' . join(' ', $0, @dbResults) . "\n")
        if @dbResults;
      # Read results.x immediately if there isn't much X (1strs) data.
      my (@xexps) = $mersDB->keyList('X');
      die("not enough X (1strs) data in db or results.x\n")
        if $#xexps < 1 || $xexps[0] < 1200 && do1strs() < 0;
      # Sometimes, the X table has invalid data like:
      # mysql> select * from mers.X;
      # +----------------------+-------------+
      # | x                    | extent      |
      # +----------------------+-------------+
      # |                    1 |           0 |
      # |                 1206 |           0 |
      # | 18446744073709551615 | 66710698623 |
      # +----------------------+-------------+
      # In this case, two rows with the same extent is invalid; the row with the larger
      #   exponent "covers" the other row - the x = 1 row should be deleted in this case.
      # For now, just insist that there are only two rows and that the first has extent = 0.
      die("unexpected X data in db - more than two rows")
        if $#xexps > 1;
      die("unexpected X data in db - the first extent is not 0")
        if $mersDB->value('X', $xexps[0]) != 0;
    }

  -d("$servDir") || mkdir("$servDir") or die("could not mkdir $servDir, $!");
  -d("$ecmSaveDir") || mkdir("$ecmSaveDir") or die("could not mkdir $ecmSaveDir, $!");
  -d("$servDir/sprp") || mkdir("$servDir/sprp") or die("could not mkdir $servDir/sprp, $!");
  -d("$servDir/tmp") || mkdir("$servDir/tmp") or die("could not mkdir $servDir/tmp, $!");
  scalar(grep { $_ eq "$ecmSaveDir" } ECMSave::subdirs())
    or die("ECMSave::subdirs does not include $ecmSaveDir\n");

  my ($neededTasks) = $workLetters;
  $neededTasks =~ s/[UH]//g; #!!UH trial factoring extent work is not currently supported; gaps should be closed first
  $need{$_} = 0
    foreach (split(//, $neededTasks));
  $neededTasks =~ s/[Gg]//g; # Gg gaps tasks are read from a data file
  my ($neededWorkers) = $workLetters;
  $neededWorkers =~ s/gH//g; # g work can be done by G clients; GHU can be done by mersfacgmp clients
  my ($children) = 0;
  foreach (keys(%clientTypes))
    { # several safety checks, partly for security
      die("invalid client type - must be all letters and digits with starting letter: '$_'\n")
        unless m/^([A-Za-z][A-Za-z0-9]+)$/;
      die("invalid client type - 'all' is reserved for other uses\n")
        if $_ eq 'all';
      my ($type) = "$1";
      $neededWorkers =~ s/$clientTypes{$type}{'worker'}//
        if exists $clientTypes{$type}{'worker'};
      my ($l);
      if (exists($clientTypes{$type}{'db'}))
          {
            die("client type '$type' has invalid 'db' entry '$clientTypes{$type}{'db'}' - should be a single mers letter\n")
              if $clientTypes{$type}{'db'} !~ m/^[A-Za-z]$/ || $clientTypes{$type}{'db'} !~ m/^([$workLetters])$/ || !defined($1);
            $l = "$1";
          }
      if (exists($clientTypes{$type}{'input'}))
          {
            die("client type '$type' has invalid 'input' entry '$clientTypes{$type}{'input'}' - should be 0 or 1\n")
              if $clientTypes{$type}{'input'} !~ m/^[01]$/;
          }
      die("'worker' and 'input' entries in clientTypes must be a single mers letter indicating work to do or work read or fetched; client type $type does not have that\n")
        if ((exists($clientTypes{$type}{'worker'}) && $clientTypes{$type}{'worker'} !~ m/^[$workLetters]$/)
            || (exists($clientTypes{$type}{'input'}) && $clientTypes{$type}{'input'} !~ m/^[01]$/)
            || (exists($clientTypes{$type}{'db'}) && $clientTypes{$type}{'db'} !~ m/^[$workLetters]$/));
      die("client type '$type' has 'lastPri' but not 'input'\n")
        if exists($clientTypes{$type}{'lastPri'}) && !exists($clientTypes{$type}{'db'});
      if (defined($l))
        {
          die("clientTypes $type has 'input' entry but no matching 'sort' entry\n")
            if !exists $clientTypes{$type}{'sort'} && !m/^Data/;
          die("client types $type and $inputToClientType{$l} both claim to provide $l work\n")
            if exists $inputToClientType{$l};
          $inputToClientType{$l} = $type;
          $need{$l} = 0;
          $neededTasks =~ s/$l//;
        }
      if (exists($clientTypes{$type}{'child'}))
        {
          $clientTypes{$type}{'child'} = ++$children;
          $childDelay{$type} = 0; # time() this child could next run
          die("client type '$type' has 'child' but not 'script'\n")
            unless exists $clientTypes{$type}{'script'};
          die("client type '$type' 'script' '$clientTypes{$type}{'script'}' contains unexpected characters\n")
            unless $clientTypes{$type}{'script'} =~ m,^([.]/)?[-+.a-zA-Z0-9]+$,;
          $clientTypes{$type}{'script'} = ('./' . $clientTypes{$type}{'script'})
            unless $1;
          die("client type '$type' 'script' '$clientTypes{$type}{'script'}' is not in the current directory\n")
            unless -f "$clientTypes{$type}{'script'}";
          die("client type '$type' 'script' '$clientTypes{$type}{'script'}' cannot be executed\n")
            unless -x "$clientTypes{$type}{'script'}";
          if (exists($clientTypes{$type}{'incrFile'}) && -s($clientTypes{$type}{'incrFile'}))
            {{ # Read the file to get 'incr' from a prior run.
              my ($fh);
              warn("could not open $clientTypes{$type}{'incrFile'}, $!"),
                delete($clients{$type}{'incrFile'}),
                last
                unless open($fh, '<', $clientTypes{$type}{'incrFile'});
              my ($incr) = <$fh>;
              warn("could not read from clientTypes{$type}{'incrFile'}, $!"),
                delete($clients{$type}{'incrFile'}),
                last
                unless defined $incr;
              warn("unexpected data '$incr' in $clientTypes{$type}{'incrFile'}"),
                delete($clients{$type}{'incrFile'}),
                last
                unless $incr =~ m,^([1-9][0-9]*)$,;
              warn("could not close clientTypes{$type}{'incrFile'}, $!"),
                delete($clients{$type}{'incrFile'})
                unless close($fh);
              $clientTypes{$type}{'incr'} = "$1";
              }}
        }
      elsif ($type eq 'OneStrs' || $type eq 'SPRPChecker' || $type eq 'CheckComposites'
             || $type eq 'CheckEcmSaveFiles')
        {
          die("client type $type has no 'child'");
        }
      #!!incomplete?
    }
  die("No client type provides '$neededTasks' tasks\n")
    if $neededTasks;
  die("No client type does '$neededWorkers' work\n")
    if $neededWorkers;

  # Initialize %work.
  ($work{$_} = [ ]) # new ref to empty array
    foreach (split(//, $workLetters));
  # g uses the same list as G.
  $work{'g'} = $work{'G'};
  # LM work is the same as S work.
  $work{'L'} = $work{'S'};
  $work{'M'} = $work{'S'};
  foreach (keys(%work))
    {
      die("work $_ has no entry in letterPri\n")
        unless exists $letterPri{$_};
    }

  # This tells checkWork() what exponent to check next for each work letter.
  $next{'checkWork'}{'letters'} = $workLetters;
  $next{'checkWork'}{'letters'} =~ s/[g]//g; # drop letters with duplicate entries in %work
  $next{'checkWork'}{$_} = 0 # subscript of @{$work{$l}} to check next
    foreach (split(//, $next{'checkWork'}{'letters'}));
  $next{'checkWork'}{'next'} = 0; # index into $next{'checkWork'}{'letters'}

  # Similar for checkData().
  # Only trial factoring gaps, joins, and extents are kept in %data currently;
  #   everything else is in the (mysql) db and perhaps in %mersennes.
  #!!do not store HU for now; see addData() for partial implementation
  #!!$next{'checkData'}{'letters'} = 'GgJHU';
  $next{'checkData'}{'letters'} = 'GgJ';
  $data{$_} = { },
    $next{'checkData'}{$_} = 1
    foreach (split(//, $next{'checkData'}{'letters'}));
  $next{'checkData'}{'next'} = 0; # index into $next{'checkData'}{'letters'}

  refetchAllowed();
  refetchAliases();
  refetchContact();

  #!!add these only while debugging?
  $allowed{'localhost'} = { 'tasks' => $maxTasks, 'letters' => $workLetters, }
    if (scalar(keys(%allowed)) < 1);
  $UUIDaliases{'localhost'} = 'localhost'
    if scalar(keys(%UUIDaliases)) < 1;

  # Always read results.x (1strs data).
  push(@{$clientTypes{'DataFile'}{'files'}}, 'results.x')
    if -s 'results.x'; # M( 18446744073709551615 )X: 19169474139
  # Always read results.gaps.sort for G (trial factoring gaps) work.
  push(@{$clientTypes{'SortedByPriority'}{'files'}}, 'results.gaps.sort')
    if -s 'results.gaps.sort';

  return
    if $checkConfig;

  bind($serverfh, sockaddr_in($port, INADDR_ANY))
    or die("bind: $!");
  listen($serverfh, SOMAXCONN)
    or die("listen: $!");
  # Look for input (new clients) on the server file handle (socket).
  vec($fdbitsIn, fileno($serverfh), 1) = 1;
}

sub unblock($$)
{
  my ($fh, $msg) = @_;
  my ($flags) = fcntl($fh, F_GETFL, 0);
  return(1) if $flags && fcntl($fh, F_SETFL, $flags | O_NONBLOCK);
  warn("fcntl() for $msg, fd " . fileno($fh) . ", could not get flags or could not set non-blocking, $!");
  return(undef);
}
